@@ -46,21 +46,19 @@ let path_of_string s =
       [] -> invalid_arg "path_of_string"
     | l::dir -> LogicalFile {dirpath=dir; basename=l}
 
-let ( / ) = Filename.concat
-
-let get_version_date () =
+let get_version () =
   try
-    let ch = open_in (Envars.coqlib () / "revision") in
+    let env = Boot.Env.init () in
+    let revision = Boot.Env.(Path.to_string (revision env)) in
+    let ch = open_in revision in
     let ver = input_line ch in
     let rev = input_line ch in
     let () = close_in ch in
-    (ver,rev)
-  with _ -> (Coq_config.version,Coq_config.date)
+    Printf.sprintf "%s (%s)" ver rev
+  with _ -> Coq_config.version
 
 let print_header () =
-  let (ver,rev) = (get_version_date ()) in
-  Printf.printf "Welcome to Chicken %s (%s)\n" ver rev;
-  flush stdout
+  Printf.printf "Welcome to Chicken %s\n%!" (get_version ())
 
 (* Adding files to Coq loadpath *)
 
@@ -77,7 +75,7 @@ let convert_string d =
   with CErrors.UserError _ ->
     Flags.if_verbose Feedback.msg_warning
       (str "Directory " ++ str d ++ str " cannot be used as a Coq identifier (skipped)");
-    raise Exit
+    raise_notrace Exit
 
 let add_rec_path ~unix_path ~coq_root =
   if exists_dir unix_path then
@@ -95,26 +93,27 @@ let add_rec_path ~unix_path ~coq_root =
   else
     Feedback.msg_warning (str "Cannot open " ++ str unix_path)
 
-(* By the option -include -I or -R of the command line *)
+(* By the option -R/-Q of the command line *)
 let includes = ref []
 let push_include (s, alias) = includes := (s,alias) :: !includes
 
-let set_default_include d =
-  push_include (d, Check.default_root_prefix)
 let set_include d p =
   let p = dirpath_of_string p in
   push_include (d,p)
 
 (* Initializes the LoadPath *)
 let init_load_path () =
-  let coqlib = Envars.coqlib () in
-  let user_contrib = coqlib/"user-contrib" in
+  let coqenv = Boot.Env.init () in
+  (* the to_string casting won't be necessary once Boot handles
+     include paths *)
+  let plugins = Boot.Env.plugins coqenv |> Boot.Path.to_string in
+  let theories = Boot.Env.stdlib coqenv |> Boot.Path.to_string in
+  let user_contrib = Boot.Env.user_contrib coqenv |> Boot.Path.to_string in
   let xdg_dirs = Envars.xdg_dirs in
   let coqpath = Envars.coqpath in
-  let plugins = coqlib/"plugins" in
   (* NOTE: These directories are searched from last to first *)
   (* first standard library *)
-  add_rec_path ~unix_path:(coqlib/"theories") ~coq_root:(Names.DirPath.make[coq_root]);
+  add_rec_path ~unix_path:theories ~coq_root:(Names.DirPath.make[coq_root]);
   (* then plugins *)
   add_rec_path ~unix_path:plugins ~coq_root:(Names.DirPath.make [coq_root]);
   (* then user-contrib *)
@@ -126,26 +125,23 @@ let init_load_path () =
   (* then directories in COQPATH *)
   List.iter (fun s -> add_rec_path ~unix_path:s ~coq_root:Check.default_root_prefix) coqpath;
   (* then current directory *)
-  add_path ~unix_path:"." ~coq_root:Check.default_root_prefix;
-  (* additional loadpath, given with -I -include -R options *)
-  List.iter
-    (fun (unix_path, coq_root) -> add_rec_path ~unix_path ~coq_root)
-    (List.rev !includes);
-  includes := []
+  add_path ~unix_path:"." ~coq_root:Check.default_root_prefix
 
+let impredicative_set = ref false
+let set_impredicative_set () = impredicative_set := true
 
-let set_debug () = Flags.debug := true
-
-let impredicative_set = ref Declarations.PredicativeSet
-let set_impredicative_set () = impredicative_set := Declarations.ImpredicativeSet
+let boot = ref false
+let set_boot () = boot := true
 
 let indices_matter = ref false
 
+let enable_vm = ref false
+
 let make_senv () =
   let senv = Safe_typing.empty_environment in
-  let senv = Safe_typing.set_engagement !impredicative_set senv in
+  let senv = Safe_typing.set_impredicative_set !impredicative_set senv in
   let senv = Safe_typing.set_indices_matter !indices_matter senv in
-  let senv = Safe_typing.set_VM false senv in
+  let senv = Safe_typing.set_VM !enable_vm senv in
   let senv = Safe_typing.set_allow_sprop true senv in (* be smarter later *)
   Safe_typing.set_native_compiler false senv
 
@@ -172,9 +168,7 @@ let compile_files senv =
     ~check:(List.rev !compile_list)
 
 let version () =
-  Printf.printf "The Coq Proof Checker, version %s (%s)\n"
-    Coq_config.version Coq_config.date;
-  Printf.printf "compiled on %s\n" Coq_config.compile_date;
+  Printf.printf "The Coq Proof Checker, version %s\n" Coq_config.version;
   exit 0
 
 (* print the usage of coqtop (or coqc) on channel co *)
@@ -183,23 +177,28 @@ let print_usage_channel co command =
   output_string co command;
   output_string co "coqchk options are:\n";
   output_string co
-"  -Q dir coqdir          map physical dir to logical coqdir\
-\n  -R dir coqdir          synonymous for -Q\
+"\
+\n  -Q dir coqdir               map physical dir to logical coqdir\
+\n  -R dir coqdir               synonymous for -Q\
+\n  -coqlib dir                 set coqchk's standard library location\
+\n  -boot                       don't initialize the library paths automatically\
 \n\
+\n  -admit module               load module and dependencies without checking\
+\n  -norec module               check module but admit dependencies without checking\
 \n\
-\n  -admit module          load module and dependencies without checking\
-\n  -norec module          check module but admit dependencies without checking\
+\n  -debug                      enable debugging info\
+\n  -where                      print coqchk's standard library location and exit\
+\n  -v, --version               print coqchk version and exit\
+\n  -o, --output-context        print the list of assumptions\
+\n  -m, --memory                print the maximum heap size\
+\n  -silent                     disable trace of constants being checked\
 \n\
-\n  -coqlib dir            set coqchk's standard library location\
-\n  -where                 print coqchk's standard library location and exit\
-\n  -v                     print coqchk version and exit\
-\n  -o, --output-context   print the list of assumptions\
-\n  -m, --memory           print the maximum heap size\
-\n  -silent                disable trace of constants being checked\
+\n  -impredicative-set          set sort Set impredicative\
+\n  -indices-matter             levels of indices (and nonuniform parameters)\
+\n                              contribute to the level of inductives\
+\n  -bytecode-compiler (yes|no) enable the vm_compute reduction machine (default is no)\
 \n\
-\n  -impredicative-set     set sort Set impredicative\
-\n\
-\n  -h, --help             print this list of options\
+\n  -h, --help                  print this list of options\
 \n"
 
 (* print the usage on standard error *)
@@ -221,20 +220,11 @@ let report () = strbrk (". Please report at " ^ Coq_config.wwwbugtracker ^ ".")
 
 let guill s = str "\"" ++ str s ++ str "\""
 
-let where = function
-| None -> mt ()
-| Some s ->
-  if !Flags.debug then  (str"in " ++ str s ++ str":" ++ spc ()) else (mt ())
-
 let explain_exn = function
-  | Stream.Failure ->
-      hov 0 (anomaly_string () ++ str "uncaught Stream.Failure.")
-  | Stream.Error txt ->
-      hov 0 (str "Syntax error: " ++ str txt)
   | Sys_error msg ->
       hov 0 (anomaly_string () ++ str "uncaught exception Sys_error " ++ guill msg ++ report() )
-  | UserError(s,pps) ->
-      hov 1 (str "User error: " ++ where s ++ pps)
+  | UserError pps ->
+      hov 1 (str "User error: " ++ pps)
   | Out_of_memory ->
       hov 0 (str "Out of memory")
   | Stack_overflow ->
@@ -251,11 +241,11 @@ let explain_exn = function
       hov 0 (anomaly_string () ++ str "uncaught exception Invalid_argument " ++ guill s ++ report ())
   | Sys.Break ->
     hov 0 (fnl () ++ str "User interrupt.")
-  | Univ.UniverseInconsistency i ->
+  | UGraph.UniverseInconsistency i ->
     let msg =
-      if !Flags.debug then
+      if CDebug.(get_flag misc) then
         str "." ++ spc() ++
-          Univ.explain_universe_inconsistency Univ.Level.pr i
+          UGraph.explain_universe_inconsistency Univ.Level.pr i
       else
         mt() in
       hov 0 (str "Error: Universe inconsistency" ++ msg ++ str ".")
@@ -291,7 +281,7 @@ let explain_exn = function
                              Constr.debug_print a ++ fnl ());
         Feedback.msg_notice (str"====== universes ====" ++ fnl () ++
                              (UGraph.pr_universes Univ.Level.pr
-                                (ctx.Environ.env_stratification.Environ.env_universes)));
+                                (UGraph.repr (ctx.Environ.env_universes))));
         str "CantApplyBadType at argument " ++ int n
       | CantApplyNonFunctional _ -> str"CantApplyNonFunctional"
       | IllFormedRecBody _ -> str"IllFormedRecBody"
@@ -299,7 +289,10 @@ let explain_exn = function
       | UnsatisfiedConstraints _ -> str"UnsatisfiedConstraints"
       | DisallowedSProp -> str"DisallowedSProp"
       | BadRelevance -> str"BadRelevance"
-      | UndeclaredUniverse _ -> str"UndeclaredUniverse"))
+      | BadInvert -> str"BadInvert"
+      | UndeclaredUniverse _ -> str"UndeclaredUniverse"
+      | BadVariance _ -> str "BadVariance"
+      ))
 
   | InductiveError e ->
       hov 0 (str "Error related to inductive types")
@@ -320,9 +313,6 @@ let explain_exn = function
                report ())
   | e -> CErrors.print e (* for anomalies and other uncaught exceptions *)
 
-let deprecated flag =
-  Feedback.msg_warning (str "Deprecated flag " ++ quote (str flag))
-
 let parse_args argv =
   let rec parse = function
     | [] -> ()
@@ -332,28 +322,30 @@ let parse_args argv =
     | "-indices-matter" :: rem ->
       indices_matter:=true; parse rem
 
+    | "-bytecode-compiler" :: "yes" :: rem ->
+      enable_vm := true; parse rem
+    | "-bytecode-compiler" :: "no" :: rem ->
+      enable_vm := false; parse rem
+
     | "-coqlib" :: s :: rem ->
       if not (exists_dir s) then
         fatal_error (str "Directory '" ++ str s ++ str "' does not exist") false;
-      Envars.set_user_coqlib s;
+      Boot.Env.set_coqlib s;
       parse rem
 
-    | ("-I"|"-include") :: d :: "-as" :: p :: rem -> deprecated "-I"; set_include d p; parse rem
-    | ("-I"|"-include") :: d :: "-as" :: [] -> usage ()
-    | ("-I"|"-include") :: d :: rem -> deprecated "-I"; set_default_include d; parse rem
-    | ("-I"|"-include") :: []       -> usage ()
+    | "-boot" :: rem ->
+      set_boot ();
+      parse rem
 
-    | "-Q" :: d :: p :: rem -> set_include d p;parse rem
-    | "-Q" :: ([] | [_]) -> usage ()
+    | ("-Q"|"-R") :: d :: p :: rem -> set_include d p;parse rem
+    | ("-Q"|"-R") :: ([] | [_]) -> usage ()
 
-    | "-R" :: d :: p :: rem -> set_include d p;parse rem
-    | "-R" :: ([] | [_]) -> usage ()
-
-    | "-debug" :: rem -> set_debug (); parse rem
+    | "-debug" :: rem -> CDebug.set_debug_all true; parse rem
 
     | "-where" :: _ ->
-      Envars.set_coqlib ~fail:(fun x -> CErrors.user_err Pp.(str x));
-      print_endline (Envars.coqlib ());
+      let env = Boot.Env.init () in
+      let coqlib = Boot.Env.coqlib env |> Boot.Path.to_string in
+      print_endline coqlib;
       exit 0
 
     | ("-?"|"-h"|"-H"|"-help"|"--help") :: _ -> usage ()
@@ -387,10 +379,14 @@ let init_with_argv argv =
   try
     parse_args argv;
     CWarnings.set_flags ("+"^Typeops.warn_bad_relevance_name);
-    if !Flags.debug then Printexc.record_backtrace true;
-    Envars.set_coqlib ~fail:(fun x -> CErrors.user_err Pp.(str x));
+    if CDebug.(get_flag misc) then Printexc.record_backtrace true;
     Flags.if_verbose print_header ();
-    init_load_path ();
+    if not !boot then init_load_path ();
+    (* additional loadpath, given with -R/-Q options *)
+    List.iter
+      (fun (unix_path, coq_root) -> add_rec_path ~unix_path ~coq_root)
+      (List.rev !includes);
+    includes := [];
     make_senv ()
   with e ->
     fatal_error (str "Error during initialization :" ++ (explain_exn e)) (is_anomaly e)
@@ -402,7 +398,7 @@ let run senv =
     let senv = compile_files senv in
     flush_all(); senv
   with e ->
-    if !Flags.debug then Printexc.print_backtrace stderr;
+    if CDebug.(get_flag misc) then Printexc.print_backtrace stderr;
     fatal_error (explain_exn e) (is_anomaly e)
 
 let start () =
