@@ -1,7 +1,7 @@
 (************************************************************************)
 (*         *   The Coq Proof Assistant / The Coq Development Team       *)
-(*  v      *   INRIA, CNRS and contributors - Copyright 1999-2019       *)
-(* <O___,, *       (see CREDITS file for the list of authors)           *)
+(*  v      *         Copyright INRIA, CNRS and contributors             *)
+(* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
 (*    //   *    This file is distributed under the terms of the         *)
 (*         *     GNU Lesser General Public License Version 2.1          *)
@@ -66,17 +66,8 @@ sig
       [filter_range] returns the submap of [m] whose keys are in
       range. Note that [in_range] has to define a continouous range. *)
 
-  val update: key -> ('a option -> 'a option) -> 'a t -> 'a t
-  (** [update x f m] returns a map containing the same bindings as
-      [m], except for the binding of [x]. Depending on the value of
-      [y] where [y] is [f (find_opt x m)], the binding of [x] is
-      added, removed or updated. If [y] is [None], the binding is
-      removed if it exists; otherwise, if [y] is [Some z] then [x]
-      is associated to [z] in the resulting map.  If [x] was already
-      bound in [m] to a value that is physically equal to [z], [m]
-      is returned unchanged (the result of the function is then
-      physically equal to [m]).
-    *)
+  val of_list : (key * 'a) list -> 'a t
+  (** Turns an association list into a map *)
 
   module Smart :
   sig
@@ -87,14 +78,6 @@ sig
     (** As [mapi] but tries to preserve sharing. *)
   end
 
-  module Unsafe :
-  sig
-    val map : (key -> 'a -> key * 'b) -> 'a t -> 'b t
-    (** As the usual [map], but also allows modifying the key of a binding.
-        It is required that the mapping function [f] preserves key equality,
-        i.e.: for all (k : key) (x : 'a), compare (fst (f k x)) k = 0. *)
-  end
-
   module Monad(M : MonadS) :
   sig
     val fold : (key -> 'a -> 'b -> 'b M.t) -> 'a t -> 'b -> 'b M.t
