@@ -41,7 +41,27 @@ let string_split s =
   in
   if len == 0 then [] else split 0
 
-let plugin_name = "__coq_plugin_name"
+(* Used to generate unique ids *)
+let file_name = ref ""
+
+(* We used to output __plugin_name = "bla" then refer to __plugin_name in generated code
+   but this is not robust to eg
+
+   DECLARE PLUGIN "bla"
+   open G_ltac
+   use the plugin name
+
+   since g_ltac has no mli its __plugin_name will shadow the local one and wreak havoc
+*)
+let plugin_name = ref None
+
+let force_is_plugin ~what () = match !plugin_name with
+  | Some (Some n) -> n
+  | Some None | None -> fatal ("DECLARE PLUGIN required before "^what)
+
+let check_is_plugin () = match !plugin_name with
+  | Some b -> b
+  | None -> plugin_name := Some None; None
 
 let print_list fmt pr l =
   let rec prl fmt = function
@@ -126,7 +146,6 @@ let print_position fmt pos = match pos with
 | Last -> fprintf fmt "Gramlib.Gramext.Last"
 | Before s -> fprintf fmt "Gramlib.Gramext.Before@ \"%s\"" s
 | After s -> fprintf fmt "Gramlib.Gramext.After@ \"%s\"" s
-| Level s -> fprintf fmt "Gramlib.Gramext.Level@ \"%s\"" s
 
 let print_assoc fmt = function
 | LeftA -> fprintf fmt "Gramlib.Gramext.LeftA"
@@ -201,8 +220,8 @@ function
 | "IDENT", s -> fprintf fmt "Tok.PIDENT (%a)" print_pat s
 | "PATTERNIDENT", s -> fprintf fmt "Tok.PPATTERNIDENT (%a)" print_pat s
 | "FIELD", s -> fprintf fmt "Tok.PFIELD (%a)" print_pat s
-| "NUMERAL", None -> fprintf fmt "Tok.PNUMERAL None"
-| "NUMERAL", Some s -> fprintf fmt "Tok.PNUMERAL (Some (NumTok.Unsigned.of_string %a))" print_string s
+| "NUMBER", None -> fprintf fmt "Tok.PNUMBER None"
+| "NUMBER", Some s -> fprintf fmt "Tok.PNUMBER (Some (NumTok.Unsigned.of_string %a))" print_string s
 | "STRING", s -> fprintf fmt "Tok.PSTRING (%a)" print_pat s
 | "LEFTQMARK", None -> fprintf fmt "Tok.PLEFTQMARK"
 | "BULLET", s -> fprintf fmt "Tok.PBULLET (%a)" print_pat s
@@ -239,11 +258,11 @@ and print_symbol fmt tkn = match tkn with
 | SymbList0 (s, None) ->
   fprintf fmt "(Pcoq.Symbol.list0 %a)" print_symbol s
 | SymbList0 (s, Some sep) ->
-  fprintf fmt "(Pcoq.Symbol.list0sep (%a) (%a) false)" print_symbol s print_symbol sep
+  fprintf fmt "(Pcoq.Symbol.list0sep (%a) (%a) false)" print_symbol s print_anonymized_symbol sep
 | SymbList1 (s, None) ->
   fprintf fmt "(Pcoq.Symbol.list1 (%a))" print_symbol s
 | SymbList1 (s, Some sep) ->
-  fprintf fmt "(Pcoq.Symbol.list1sep (%a) (%a) false)" print_symbol s print_symbol sep
+  fprintf fmt "(Pcoq.Symbol.list1sep (%a) (%a) false)" print_symbol s print_anonymized_symbol sep
 | SymbOpt s ->
   fprintf fmt "(Pcoq.Symbol.opt %a)" print_symbol s
 | SymbRules rules ->
@@ -257,17 +276,45 @@ and print_symbol fmt tkn = match tkn with
 | SymbQuote c ->
   fprintf fmt "(%s)" c
 
+and print_anonymized_symbol fmt tkn = match tkn with
+| SymbToken (t, s) ->
+  fprintf fmt "(Pcoq.Symbol.tokens [Pcoq.TPattern (%a)])" print_tok (t, s)
+| _ -> print_symbol fmt (SymbRules [[None, tkn], mk_code "()"])
+
 let print_rule fmt r =
   let pr_lvl fmt lvl = print_opt fmt print_string lvl in
   let pr_asc fmt asc = print_opt fmt print_assoc asc in
   let pr_prd fmt prd = print_list fmt print_prod prd in
   fprintf fmt "@[(%a,@ %a,@ %a)@]" pr_lvl r.grule_label pr_asc r.grule_assoc pr_prd (List.rev r.grule_prods)
 
-let print_entry fmt e =
-  let print_position_opt fmt pos = print_opt fmt print_position pos in
+let gramext_count = ref 0
+
+let gramext_plugin_uid name =
+  let cnt = !gramext_count in
+  incr gramext_count;
+  " ~plugin_uid:(\""^name^"\", \""^(!file_name)^":"^string_of_int cnt^"\")"
+
+let grammar_extend () =
+  match check_is_plugin () with
+  | Some name -> "Egramml.grammar_extend"^gramext_plugin_uid name
+  | None -> "Pcoq.grammar_extend"
+
+let print_entry fmt e = match e.gentry_rules with
+| GDataReuse (pos, r) ->
+  let rules = List.rev r in
+  let pr_pos fmt pos = print_opt fmt print_string pos in
+  let pr_prd fmt prd = print_list fmt print_prod prd in
+  fprintf fmt "let () =@ @[%s@ %s@ @[(Pcoq.Reuse (%a, %a))@]@]@ in@ "
+    (grammar_extend ()) e.gentry_name pr_pos pos pr_prd rules
+| GDataFresh (pos, rules) ->
   let print_rules fmt rules = print_list fmt print_rule rules in
-  fprintf fmt "let () =@ @[Pcoq.grammar_extend@ %s@ @[{ Pcoq.pos=%a; data=%a}@]@]@ in@ "
-    e.gentry_name print_position_opt e.gentry_pos print_rules e.gentry_rules
+  let pr_check fmt () = match pos with
+  | None -> fprintf fmt "let () =@ @[assert@ (Pcoq.Entry.is_empty@ %s)@]@ in@\n" e.gentry_name
+  | Some _ -> fprintf fmt ""
+  in
+  let pos = match pos with None -> First | Some pos -> pos in
+  fprintf fmt "%alet () =@ @[%s@ %s@ @[(Pcoq.Fresh@ (%a, %a))@]@]@ in@ "
+    pr_check () (grammar_extend ()) e.gentry_name print_position pos print_rules rules
 
 let print_ast fmt ext =
   let () = fprintf fmt "let _ = @[" in
@@ -334,17 +381,21 @@ let print_atts_right fmt = function
     fprintf fmt "(Attributes.parse %s%a atts)" nota aux atts
 
 let understand_state = function
-  | "close_proof" -> "VtCloseProof", false
-  | "open_proof" -> "VtOpenProof", true
-  | "proof" -> "VtModifyProof", false
-  | "proof_opt_query" -> "VtReadProofOpt", false
-  | "proof_query" -> "VtReadProof", false
+  | "close_proof" -> "vtcloseproof", false
+  | "open_proof" -> "vtopenproof", true
+  | "proof" -> "vtmodifyproof", false
+  | "proof_opt_query" -> "vtreadproofopt", false
+  | "proof_query" -> "vtreadproof", false
+  | "read_program" -> "vtreadprogram", false
+  | "program" -> "vtmodifyprogram", false
+  | "declare_program" -> "vtdeclareprogram", false
+  | "program_interactive" -> "vtopenproofprogram", false
   | s -> fatal ("unsupported state specifier: " ^ s)
 
 let print_body_state state fmt r =
   let state = match r.vernac_state with Some _ as s -> s | None -> state in
   match state with
-  | None -> fprintf fmt "Vernacextend.VtDefault (fun () -> %a)" print_code r.vernac_body
+  | None -> fprintf fmt "Vernacextend.vtdefault (fun () -> %a)" print_code r.vernac_body
   | Some "CUSTOM" -> print_code fmt r.vernac_body
   | Some state ->
     let state, unit_wrap = understand_state state in
@@ -356,7 +407,7 @@ let print_body_fun state fmt r =
     print_binders r.vernac_toks print_atts_left r.vernac_atts (print_body_state state) r
 
 let print_body state fmt r =
-  fprintf fmt "@[(%afun %a~atts@ -> coqpp_body %a%a)@]"
+  fprintf fmt "@[(%afun %a?loc ~atts ()@ -> coqpp_body %a%a)@]"
     (print_body_fun state) r print_binders r.vernac_toks
     print_binders r.vernac_toks print_atts_right r.vernac_atts
 
@@ -390,7 +441,8 @@ let print_entry fmt = function
 
 let print_ast fmt ext =
   let pr fmt () =
-    fprintf fmt "Vernacextend.vernac_extend ~command:\"%s\" %a ?entry:%a %a"
+    fprintf fmt "Vernacextend.vernac_extend %s ~command:\"%s\" %a ?entry:%a %a"
+      (match check_is_plugin () with | Some name -> "~plugin:\""^name^"\"" | None -> "")
       ext.vernacext_name print_classifier ext.vernacext_class
       print_entry ext.vernacext_entry (print_rules ext.vernacext_state) ext.vernacext_rules
   in
@@ -429,8 +481,9 @@ let print_ast fmt ext =
   in
   let pr fmt () =
     let level = match ext.tacext_level with None -> 0 | Some i -> i in
-    fprintf fmt "Tacentries.tactic_extend %s \"%s\" ~level:%i %a%a"
-      plugin_name ext.tacext_name level
+    let name = force_is_plugin ~what:"TACTIC EXTEND" () in
+    fprintf fmt "Tacentries.tactic_extend \"%s\" \"%s\" ~level:%i %a%a"
+      name ext.tacext_name level
       deprecation ext.tacext_deprecated
       print_rules ext.tacext_rules
   in
@@ -450,7 +503,7 @@ struct
 
 let terminal s =
   let p =
-    if s <> "" && s.[0] >= '0' && s.[0] <= '9' then "CLexer.terminal_numeral"
+    if s <> "" && s.[0] >= '0' && s.[0] <= '9' then "CLexer.terminal_number"
     else "CLexer.terminal" in
   let c = Printf.sprintf "Pcoq.Symbol.token (%s \"%s\")" p s in
   SymbQuote c
@@ -465,7 +518,7 @@ let rec parse_symb self = function
 | Uentryl (e, l) ->
   assert (e = "tactic");
   if l = 5 then SymbEntry ("Pltac.binder_tactic", None)
-  else SymbEntry ("Pltac.tactic_expr", Some (string_of_int l))
+  else SymbEntry ("Pltac.ltac_expr", Some (string_of_int l))
 
 let parse_token self = function
 | ExtTerminal s -> (terminal s, None)
@@ -498,9 +551,10 @@ let print_printer fmt = function
 let print_ast fmt arg =
   let name = arg.vernacargext_name in
   let pr fmt () =
-    fprintf fmt "Vernacextend.vernac_argument_extend ~name:%a @[{@\n\
+    fprintf fmt "Vernacextend.vernac_argument_extend ~plugin:\"%s\" ~name:%a @[{@\n\
       Vernacextend.arg_parsing = %a;@\n\
       Vernacextend.arg_printer = fun env sigma -> %a;@\n}@]"
+      (force_is_plugin ~what:"VERNAC ARGUMENT EXTEND" ())
       print_string name print_rules (name, arg.vernacargext_rules)
       print_printer arg.vernacargext_printer
   in
@@ -557,8 +611,10 @@ let print_ast fmt arg =
     fprintf fmt "@[Tacentries.ArgSubstFun (fun s v -> v)@]"
   in
   let interp fmt () = match arg.argext_interp, arg.argext_type with
-  | Some f, (None | Some _) ->
-    fprintf fmt "@[Tacentries.ArgInterpLegacy (%a)@]" print_code f
+  | Some (None, f), (None | Some _) ->
+    fprintf fmt "@[Tacentries.ArgInterpSimple (%a)@]" print_code f
+  | Some (Some kind, f), (None | Some _) ->
+    fatal (Printf.sprintf "Unknown kind %s of interpretation function" kind)
   | None, Some t ->
     fprintf fmt "@[Tacentries.ArgInterpWit (%a)@]" print_wit t
   | None, None ->
@@ -580,13 +636,14 @@ let print_ast fmt arg =
   | None -> default_printer
   in
   let pr fmt () =
-    fprintf fmt "Tacentries.argument_extend ~name:%a @[{@\n\
+    fprintf fmt "Tacentries.argument_extend ~plugin:\"%s\" ~name:%a @[{@\n\
       Tacentries.arg_parsing = %a;@\n\
       Tacentries.arg_tag = @[%a@];@\n\
       Tacentries.arg_intern = @[%a@];@\n\
       Tacentries.arg_subst = @[%a@];@\n\
       Tacentries.arg_interp = @[%a@];@\n\
-      Tacentries.arg_printer = @[((fun env sigma -> %a), (fun env sigma -> %a), (fun env sigma -> %a))@];@\n}@]"
+                 Tacentries.arg_printer = @[((fun env sigma -> %a), (fun env sigma -> %a), (fun env sigma -> %a))@];@\n}@]"
+      (force_is_plugin ~what:"ARGUMENT EXTEND" ())
       print_string name
       VernacArgumentExt.print_rules (name, arg.argext_rules)
       pr_tag arg.argext_type
@@ -598,8 +655,13 @@ let print_ast fmt arg =
 end
 
 let declare_plugin fmt name =
-  fprintf fmt "let %s = \"%s\"@\n" plugin_name name;
-  fprintf fmt "let _ = Mltop.add_known_module %s@\n" plugin_name
+  fprintf fmt "let _ = Mltop.add_known_module \"%s\"@\n" name;
+  let () = match !plugin_name with
+    | None -> plugin_name := Some (Some name)
+    | Some (Some _) -> fatal "Multiple DECLARE PLUGIN not allowed";
+    | Some None -> fatal "DECLARE PLUGIN must be before VERNAC / GRAMMAR EXTEND"
+  in
+  ()
 
 let pr_ast fmt = function
 | Code s -> fprintf fmt "%a@\n" print_code s
@@ -611,16 +673,34 @@ let pr_ast fmt = function
 | TacticExt tac -> fprintf fmt "%a@\n" TacticExt.print_ast tac
 | ArgumentExt arg -> fprintf fmt "%a@\n" ArgumentExt.print_ast arg
 
-let () =
+let help () =
+  Format.eprintf "Usage: coqpp file.mlg@\n%!";
+  exit 1
+
+let parse () =
   let () =
-    if Array.length Sys.argv <> 2 then fatal "Expected exactly one command line argument"
+    if Array.length Sys.argv <> 2
+    then help ()
   in
-  let file = Sys.argv.(1) in
-  let output = Filename.chop_extension file ^ ".ml" in
+  match Sys.argv.(1) with
+  | "-help" | "--help" -> help ()
+  | file -> file
+
+let output_name file =
+  try
+    Filename.chop_extension file ^ ".ml"
+  with
+  | Invalid_argument _ ->
+    fatal "Input file must have an extension for coqpp [input.ext -> input.ml]"
+
+let () =
+  let file = parse () in
+  let output = output_name file in
   let ast = parse_file file in
   let chan = open_out output in
+  let () = file_name := Filename.basename file in
   let fmt = formatter_of_out_channel chan in
-  let iter ast = Format.fprintf fmt "@[%a@]%!"pr_ast ast in
+  let iter ast = Format.fprintf fmt "@[%a@]%!" pr_ast ast in
   let () = List.iter iter ast in
   let () = close_out chan in
   exit 0
