@@ -14,6 +14,8 @@ open Names
 open Constr
 open Context
 
+module NamedDecl = Context.Named.Declaration
+
 module ESorts = struct
   include Evd.MiniEConstr.ESorts
 
@@ -35,6 +37,10 @@ include (Evd.MiniEConstr : module type of Evd.MiniEConstr
 type types = t
 type constr = t
 type existential = t pexistential
+type case_return = t pcase_return
+type case_branch = t pcase_branch
+type case_invert = t pcase_invert
+type case = (t, t, EInstance.t) pcase
 type fixpoint = (t, t) pfixpoint
 type cofixpoint = (t, t) pcofixpoint
 type unsafe_judgment = (constr, types) Environ.punsafe_judgment
@@ -69,7 +75,7 @@ let mkInd i = of_kind (Ind (in_punivs i))
 let mkConstructU pc = of_kind (Construct pc)
 let mkConstruct c = of_kind (Construct (in_punivs c))
 let mkConstructUi ((ind,u),i) = of_kind (Construct ((ind,i),u))
-let mkCase (ci, c, r, p) = of_kind (Case (ci, c, r, p))
+let mkCase (ci, u, pms, c, iv, r, p) = of_kind (Case (ci, u, pms, c, iv, r, p))
 let mkFix f = of_kind (Fix f)
 let mkCoFix f = of_kind (CoFix f)
 let mkProj (p, c) = of_kind (Proj (p, c))
@@ -77,6 +83,7 @@ let mkArrow t1 r t2 = of_kind (Prod (make_annot Anonymous r, t1, t2))
 let mkArrowR t1 t2 = mkArrow t1 Sorts.Relevant t2
 let mkInt i = of_kind (Int i)
 let mkFloat f = of_kind (Float f)
+let mkArray (u,t,def,ty) = of_kind (Array (u,t,def,ty))
 
 let mkRef (gr,u) = let open GlobRef in match gr with
   | ConstRef c -> mkConstU (c,u)
@@ -84,6 +91,8 @@ let mkRef (gr,u) = let open GlobRef in match gr with
   | ConstructRef c -> mkConstructU (c,u)
   | VarRef x -> mkVar x
 
+let mkLEvar = Evd.MiniEConstr.mkLEvar
+
 let type1 = mkSort Sorts.type1
 
 let applist (f, arg) = mkApp (f, Array.of_list arg)
@@ -126,9 +135,9 @@ let isRef sigma c = match kind sigma c with
 let isRefX sigma x c =
   let open GlobRef in
   match x, kind sigma c with
-  | ConstRef c, Const (c', _) -> Constant.equal c c'
-  | IndRef i, Ind (i', _) -> eq_ind i i'
-  | ConstructRef i, Construct (i', _) -> eq_constructor i i'
+  | ConstRef c, Const (c', _) -> Constant.CanOrd.equal c c'
+  | IndRef i, Ind (i', _) -> Ind.CanOrd.equal i i'
+  | ConstructRef i, Construct (i', _) -> Construct.CanOrd.equal i i'
   | VarRef id, Var id' -> Id.equal id id'
   | _ -> false
 
@@ -194,7 +203,7 @@ let destCoFix sigma c = match kind sigma c with
 | _ -> raise DestKO
 
 let destCase sigma c = match kind sigma c with
-| Case (ci, t, c, p) -> (ci, t, c, p)
+| Case (ci, u, pms, t, iv, c, p) -> (ci, u, pms, t, iv, c, p)
 | _ -> raise DestKO
 
 let destProj sigma c = match kind sigma c with
@@ -227,7 +236,7 @@ let decompose_lam_assum sigma c =
     match kind sigma c with
     | Lambda (x,t,c)  -> lamdec_rec (Context.Rel.add (LocalAssum (x,t)) l) c
     | LetIn (x,b,t,c) -> lamdec_rec (Context.Rel.add (LocalDef (x,b,t)) l) c
-    | Cast (c,_,_)      -> lamdec_rec l c
+    | Cast (c,_,_)    -> lamdec_rec l c
     | _               -> l,c
   in
   lamdec_rec Context.Rel.empty c
@@ -235,43 +244,33 @@ let decompose_lam_assum sigma c =
 let decompose_lam_n_assum sigma n c =
   let open Rel.Declaration in
   if n < 0 then
-    user_err Pp.(str "decompose_lam_n_assum: integer parameter must be positive");
+    anomaly Pp.(str "decompose_lam_n_assum: integer parameter must be positive.");
   let rec lamdec_rec l n c =
     if Int.equal n 0 then l,c
     else
       match kind sigma c with
       | Lambda (x,t,c)  -> lamdec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> lamdec_rec (Context.Rel.add (LocalDef (x,b,t)) l) n c
-      | Cast (c,_,_)      -> lamdec_rec l n c
-      | c -> user_err Pp.(str "decompose_lam_n_assum: not enough abstractions")
+      | Cast (c,_,_)    -> lamdec_rec l n c
+      | c -> anomaly Pp.(str "decompose_lam_n_assum: not enough abstractions.")
   in
   lamdec_rec Context.Rel.empty n c
 
 let decompose_lam_n_decls sigma n =
   let open Rel.Declaration in
   if n < 0 then
-    user_err Pp.(str "decompose_lam_n_decls: integer parameter must be positive");
+    anomaly Pp.(str "decompose_lam_n_decls: integer parameter must be positive.");
   let rec lamdec_rec l n c =
     if Int.equal n 0 then l,c
     else
       match kind sigma c with
       | Lambda (x,t,c)  -> lamdec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> lamdec_rec (Context.Rel.add (LocalDef (x,b,t)) l) (n-1) c
-      | Cast (c,_,_)      -> lamdec_rec l n c
-      | c -> user_err Pp.(str "decompose_lam_n_decls: not enough abstractions")
+      | Cast (c,_,_)    -> lamdec_rec l n c
+      | c -> anomaly Pp.(str "decompose_lam_n_decls: not enough abstractions.")
   in
   lamdec_rec Context.Rel.empty n
 
-let lamn n env b =
-  let rec lamrec = function
-    | (0, env, b)        -> b
-    | (n, ((v,t)::l), b) -> lamrec (n-1,  l, mkLambda (v,t,b))
-    | _ -> assert false
-  in
-  lamrec (n,env,b)
-
-let compose_lam l b = lamn (List.length l) l b
-
 let rec to_lambda sigma n prod =
   if Int.equal n 0 then
     prod
@@ -279,13 +278,13 @@ let rec to_lambda sigma n prod =
     match kind sigma prod with
       | Prod (na,ty,bd) -> mkLambda (na,ty,to_lambda sigma (n-1) bd)
       | Cast (c,_,_) -> to_lambda sigma n c
-      | _   -> user_err ~hdr:"to_lambda" (Pp.mt ())
+      | _   -> anomaly Pp.(str "Not enough products.")
 
 let decompose_prod sigma c =
   let rec proddec_rec l c = match kind sigma c with
     | Prod (x,t,c) -> proddec_rec ((x,t)::l) c
-    | Cast (c,_,_)     -> proddec_rec l c
-    | _                -> l,c
+    | Cast (c,_,_) -> proddec_rec l c
+    | _            -> l,c
   in
   proddec_rec [] c
 
@@ -293,9 +292,9 @@ let decompose_prod_assum sigma c =
   let open Rel.Declaration in
   let rec proddec_rec l c =
     match kind sigma c with
-    | Prod (x,t,c)  -> proddec_rec (Context.Rel.add (LocalAssum (x,t)) l) c
+    | Prod (x,t,c)    -> proddec_rec (Context.Rel.add (LocalAssum (x,t)) l) c
     | LetIn (x,b,t,c) -> proddec_rec (Context.Rel.add (LocalDef (x,b,t)) l) c
-    | Cast (c,_,_)      -> proddec_rec l c
+    | Cast (c,_,_)    -> proddec_rec l c
     | _               -> l,c
   in
   proddec_rec Context.Rel.empty c
@@ -303,15 +302,15 @@ let decompose_prod_assum sigma c =
 let decompose_prod_n_assum sigma n c =
   let open Rel.Declaration in
   if n < 0 then
-    user_err Pp.(str "decompose_prod_n_assum: integer parameter must be positive");
+    anomaly Pp.(str "decompose_prod_n_assum: integer parameter must be positive.");
   let rec prodec_rec l n c =
     if Int.equal n 0 then l,c
     else
       match kind sigma c with
       | Prod (x,t,c)    -> prodec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> prodec_rec (Context.Rel.add (LocalDef (x,b,t)) l) (n-1) c
-      | Cast (c,_,_)      -> prodec_rec l n c
-      | c -> user_err Pp.(str "decompose_prod_n_assum: not enough assumptions")
+      | Cast (c,_,_)    -> prodec_rec l n c
+      | c -> anomaly Pp.(str "decompose_prod_n_assum: not enough declarations.")
   in
   prodec_rec Context.Rel.empty n c
 
@@ -319,33 +318,139 @@ let existential_type = Evd.existential_type
 
 let lift n c = of_constr (Vars.lift n (unsafe_to_constr c))
 
-let map_under_context f n c =
-  let f c = unsafe_to_constr (f (of_constr c)) in
-  of_constr (Constr.map_under_context f n (unsafe_to_constr c))
-let map_branches f ci br =
-  let f c = unsafe_to_constr (f (of_constr c)) in
-  of_constr_array (Constr.map_branches f ci (unsafe_to_constr_array br))
-let map_return_predicate f ci p =
-  let f c = unsafe_to_constr (f (of_constr c)) in
-  of_constr (Constr.map_return_predicate f ci (unsafe_to_constr p))
+let of_branches : Constr.case_branch array -> case_branch array =
+  match Evd.MiniEConstr.unsafe_eq with
+  | Refl -> fun x -> x
 
-let map_user_view sigma f c =
+let unsafe_to_branches : case_branch array -> Constr.case_branch array =
+  match Evd.MiniEConstr.unsafe_eq with
+  | Refl -> fun x -> x
+
+let of_return : Constr.case_return -> case_return =
+  match Evd.MiniEConstr.unsafe_eq with
+  | Refl -> fun x -> x
+
+let unsafe_to_return : case_return -> Constr.case_return =
+  match Evd.MiniEConstr.unsafe_eq with
+  | Refl -> fun x -> x
+
+let map_branches f br =
   let f c = unsafe_to_constr (f (of_constr c)) in
-  of_constr (Constr.map_user_view f (unsafe_to_constr (whd_evar sigma c)))
+  of_branches (Constr.map_branches f (unsafe_to_branches br))
+let map_return_predicate f p =
+  let f c = unsafe_to_constr (f (of_constr c)) in
+  of_return (Constr.map_return_predicate f (unsafe_to_return p))
+
+let map_instance sigma f evk args =
+  let rec map ctx args = match ctx, SList.view args with
+  | [], None -> SList.empty
+  | decl :: ctx, Some (Some c, rem) ->
+    let c' = f c in
+    let rem' = map ctx rem in
+    if c' == c && rem' == rem then args
+    else if Constr.isVarId (NamedDecl.get_id decl) c' then SList.default rem'
+    else SList.cons c' rem'
+  | decl :: ctx, Some (None, rem) ->
+    let c = Constr.mkVar (NamedDecl.get_id decl) in
+    let c' = f c in
+    let rem' = map ctx rem in
+    if c' == c && rem' == rem then args
+    else SList.cons c' rem'
+  | [], Some _ | _ :: _, None -> assert false
+  in
+  let ctx = Evd.evar_filtered_context @@ Evd.find sigma evk in
+  map ctx args
 
 let map sigma f c =
   let f c = unsafe_to_constr (f (of_constr c)) in
-  of_constr (Constr.map f (unsafe_to_constr (whd_evar sigma c)))
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar (evk, args) ->
+    let args' = map_instance sigma f evk args in
+    if args' == args then of_constr c else of_constr @@ Constr.mkEvar (evk, args')
+  | _ -> of_constr (Constr.map f c)
 
 let map_with_binders sigma g f l c =
   let f l c = unsafe_to_constr (f l (of_constr c)) in
-  of_constr (Constr.map_with_binders g f l (unsafe_to_constr (whd_evar sigma c)))
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar (evk, args) ->
+    let args' = map_instance sigma (fun c -> f l c) evk args in
+    if args' == args then of_constr c else of_constr @@ Constr.mkEvar (evk, args')
+  | _ -> of_constr (Constr.map_with_binders g f l c)
+
+let map_existential sigma f ((evk, args) as ev : existential) =
+  let f c = unsafe_to_constr (f (of_constr c)) in
+  let args : Constr.t SList.t = match Evd.MiniEConstr.unsafe_eq with Refl -> args in
+  let args' = map_instance sigma f evk args in
+  if args' == args then ev
+  else
+    let args' : t SList.t = match Evd.MiniEConstr.unsafe_eq with Refl -> args' in
+    (evk, args')
 
 let iter sigma f c =
   let f c = f (of_constr c) in
-  Constr.iter f (unsafe_to_constr (whd_evar sigma c))
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar ((evk, _) as ev) ->
+    let args = Evd.expand_existential0 sigma ev in
+    List.iter (fun c -> f c) args
+  | _ -> Constr.iter f c
 
-let iter_with_full_binders sigma g f n c =
+let expand_case env _sigma (ci, u, pms, p, iv, c, bl) =
+  let u = EInstance.unsafe_to_instance u in
+  let pms = unsafe_to_constr_array pms in
+  let p = unsafe_to_return p in
+  let iv = unsafe_to_case_invert iv in
+  let c = unsafe_to_constr c in
+  let bl = unsafe_to_branches bl in
+  let (ci, p, iv, c, bl) = Inductive.expand_case env (ci, u, pms, p, iv, c, bl) in
+  let p = of_constr p in
+  let c = of_constr c in
+  let iv = of_case_invert iv in
+  let bl = of_constr_array bl in
+  (ci, p, iv, c, bl)
+
+let annotate_case env sigma (ci, u, pms, p, iv, c, bl as case) =
+  let (_, p, _, _, bl) = expand_case env sigma case in
+  let p =
+    (* Too bad we need to fetch this data in the environment, should be in the
+      case_info instead. *)
+    let (_, mip) = Inductive.lookup_mind_specif env ci.ci_ind in
+    decompose_lam_n_decls sigma (mip.Declarations.mind_nrealdecls + 1) p
+  in
+  let mk_br c n = decompose_lam_n_decls sigma n c in
+  let bl = Array.map2 mk_br bl ci.ci_cstr_ndecls in
+  (ci, u, pms, p, iv, c, bl)
+
+let expand_branch env _sigma u pms (ind, i) (nas, _br) =
+  let open Declarations in
+  let u = EInstance.unsafe_to_instance u in
+  let pms = unsafe_to_constr_array pms in
+  let (mib, mip) = Inductive.lookup_mind_specif env ind in
+  let paramdecl = Vars.subst_instance_context u mib.mind_params_ctxt in
+  let paramsubst = Vars.subst_of_rel_context_instance paramdecl pms in
+  let (ctx, _) = mip.mind_nf_lc.(i - 1) in
+  let (ctx, _) = List.chop mip.mind_consnrealdecls.(i - 1) ctx in
+  let ans = Inductive.instantiate_context u paramsubst nas ctx in
+  let ans : rel_context = match Evd.MiniEConstr.unsafe_eq with Refl -> ans in
+  ans
+
+let contract_case env _sigma (ci, p, iv, c, bl) =
+  let p = unsafe_to_constr p in
+  let iv = unsafe_to_case_invert iv in
+  let c = unsafe_to_constr c in
+  let bl = unsafe_to_constr_array bl in
+  let (ci, u, pms, p, iv, c, bl) = Inductive.contract_case env (ci, p, iv, c, bl) in
+  let u = EInstance.make u in
+  let pms = of_constr_array pms in
+  let p = of_return p in
+  let iv = of_case_invert iv in
+  let c = of_constr c in
+  let bl = of_branches bl in
+  (ci, u, pms, p, iv, c, bl)
+
+let iter_with_full_binders env sigma g f n c =
   let open Context.Rel.Declaration in
   match kind sigma c with
   | (Rel _ | Meta _ | Var _   | Sort _ | Const _ | Ind _
@@ -355,8 +460,13 @@ let iter_with_full_binders sigma g f n c =
   | Lambda (na,t,c) -> f n t; f (g (LocalAssum (na, t)) n) c
   | LetIn (na,b,t,c) -> f n b; f n t; f (g (LocalDef (na, b, t)) n) c
   | App (c,l) -> f n c; Array.Fun1.iter f n l
-  | Evar (_,l) -> List.iter (fun c -> f n c) l
-  | Case (_,p,c,bl) -> f n p; f n c; Array.Fun1.iter f n bl
+  | Evar ((_,l) as ev) ->
+    let l = Evd.expand_existential sigma ev in
+    List.iter (fun c -> f n c) l
+  | Case (ci,u,pms,p,iv,c,bl) ->
+    let (ci, _, pms, p, iv, c, bl) = annotate_case env sigma (ci, u, pms, p, iv, c, bl) in
+    let f_ctx (ctx, c) = f (List.fold_right g ctx n) c in
+    Array.Fun1.iter f n pms; f_ctx p; iter_invert (f n) iv; f n c; Array.iter f_ctx bl
   | Proj (p,c) -> f n c
   | Fix (_,(lna,tl,bl)) ->
     Array.iter (f n) tl;
@@ -366,60 +476,87 @@ let iter_with_full_binders sigma g f n c =
     Array.iter (f n) tl;
     let n' = Array.fold_left2_i (fun i n na t -> g (LocalAssum (na,lift i t)) n) n lna tl in
     Array.iter (f n') bl
+  | Array (_u,t,def,ty) -> Array.Fun1.iter f n t; f n def; f n ty
 
 let iter_with_binders sigma g f n c =
   let f l c = f l (of_constr c) in
-  Constr.iter_with_binders g f n (unsafe_to_constr (whd_evar sigma c))
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar ((evk, _) as ev) ->
+    let args = Evd.expand_existential0 sigma ev in
+    List.iter (fun c -> f n c) args
+  | _ -> Constr.iter_with_binders g f n c
 
 let fold sigma f acc c =
   let f acc c = f acc (of_constr c) in
-  Constr.fold f acc (unsafe_to_constr (whd_evar sigma c))
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar ((evk, _) as ev) ->
+    let args = Evd.expand_existential0 sigma ev in
+    List.fold_left f acc args
+  | _ -> Constr.fold f acc c
 
-let compare_gen k eq_inst eq_sort eq_constr nargs c1 c2 =
-  (c1 == c2) || Constr.compare_head_gen_with k k eq_inst eq_sort eq_constr nargs c1 c2
+let fold_with_binders sigma g f e acc c =
+  let f e acc c = f e acc (of_constr c) in
+  let c = unsafe_to_constr @@ whd_evar sigma c in
+  match Constr.kind c with
+  | Evar ((evk, _) as ev) ->
+    let args = Evd.expand_existential0 sigma ev in
+    List.fold_left (fun acc c -> f e acc c) acc args
+  | _ -> Constr.fold_constr_with_binders g f e acc c
+
+let compare_gen k eq_inst eq_sort eq_constr eq_evars nargs c1 c2 =
+  (c1 == c2) || Constr.compare_head_gen_with k k eq_inst eq_sort eq_constr eq_evars nargs c1 c2
+
+let eq_existential sigma eq (evk1, args1) (evk2, args2) =
+  if Evar.equal evk1 evk2 then
+    let args1 = Evd.expand_existential sigma (evk1, args1) in
+    let args2 = Evd.expand_existential sigma (evk2, args2) in
+    List.equal eq args1 args2
+  else false
 
 let eq_constr sigma c1 c2 =
   let kind c = kind sigma c in
-  let eq_inst _ _ i1 i2 = EInstance.equal sigma i1 i2 in
+  let eq_inst _ i1 i2 = EInstance.equal sigma i1 i2 in
   let eq_sorts s1 s2 = ESorts.equal sigma s1 s2 in
+  let eq_existential eq e1 e2 = eq_existential sigma (eq 0) e1 e2 in
   let rec eq_constr nargs c1 c2 =
-    compare_gen kind eq_inst eq_sorts eq_constr nargs c1 c2
+    compare_gen kind eq_inst eq_sorts (eq_existential eq_constr) eq_constr nargs c1 c2
   in
   eq_constr 0 c1 c2
 
 let eq_constr_nounivs sigma c1 c2 =
   let kind c = kind sigma c in
+  let eq_existential eq e1 e2 = eq_existential sigma (eq 0) e1 e2 in
   let rec eq_constr nargs c1 c2 =
-    compare_gen kind (fun _ _ _ _ -> true) (fun _ _ -> true) eq_constr nargs c1 c2
+    compare_gen kind (fun _ _ _ -> true) (fun _ _ -> true) (eq_existential eq_constr) eq_constr nargs c1 c2
   in
   eq_constr 0 c1 c2
 
 let compare_constr sigma cmp c1 c2 =
   let kind c = kind sigma c in
-  let eq_inst _ _ i1 i2 = EInstance.equal sigma i1 i2 in
+  let eq_inst _ i1 i2 = EInstance.equal sigma i1 i2 in
   let eq_sorts s1 s2 = ESorts.equal sigma s1 s2 in
+  let eq_existential eq e1 e2 = eq_existential sigma (eq 0) e1 e2 in
   let cmp nargs c1 c2 = cmp c1 c2 in
-  compare_gen kind eq_inst eq_sorts cmp 0 c1 c2
+  compare_gen kind eq_inst eq_sorts (eq_existential cmp) cmp 0 c1 c2
 
 let compare_cumulative_instances cv_pb nargs_ok variances u u' cstrs =
   let open UnivProblem in
   if not nargs_ok then enforce_eq_instances_univs false u u' cstrs
   else
+    let make u = Sorts.sort_of_univ @@ Univ.Universe.make u in
     CArray.fold_left3
       (fun cstrs v u u' ->
          let open Univ.Variance in
          match v with
          | Irrelevant -> Set.add (UWeak (u,u')) cstrs
          | Covariant ->
-           let u = Univ.Universe.make u in
-           let u' = Univ.Universe.make u' in
            (match cv_pb with
-            | Reduction.CONV -> Set.add (UEq (u,u')) cstrs
-            | Reduction.CUMUL -> Set.add (ULe (u,u')) cstrs)
+            | Reduction.CONV -> Set.add (UEq (make u, make u')) cstrs
+            | Reduction.CUMUL -> Set.add (ULe (make u, make u')) cstrs)
          | Invariant ->
-           let u = Univ.Universe.make u in
-           let u' = Univ.Universe.make u' in
-           Set.add (UEq (u,u')) cstrs)
+           Set.add (UEq (make u, make u')) cstrs)
       cstrs variances (Univ.Instance.to_array u) (Univ.Instance.to_array u')
 
 let cmp_inductives cv_pb (mind,ind as spec) nargs u1 u2 cstrs =
@@ -442,41 +579,44 @@ let cmp_constructors (mind, ind, cns as spec) nargs u1 u2 cstrs =
       Array.fold_left2 (fun cstrs u1 u2 -> UnivProblem.(Set.add (UWeak (u1,u2)) cstrs))
         cstrs (Univ.Instance.to_array u1) (Univ.Instance.to_array u2)
 
-let eq_universes env sigma cstrs cv_pb ref nargs l l' =
+let eq_universes env sigma cstrs cv_pb refargs l l' =
   if EInstance.is_empty l then (assert (EInstance.is_empty l'); true)
   else
     let l = EInstance.kind sigma l
     and l' = EInstance.kind sigma l' in
     let open GlobRef in
     let open UnivProblem in
-    match ref with
-    | VarRef _ -> assert false (* variables don't have instances *)
-    | ConstRef _ ->
+    match refargs with
+    | Some (ConstRef c, 1) when Environ.is_array_type env c ->
+      cstrs := compare_cumulative_instances cv_pb true [|Univ.Variance.Irrelevant|] l l' !cstrs;
+      true
+    | None | Some (ConstRef _, _) ->
       cstrs := enforce_eq_instances_univs true l l' !cstrs; true
-    | IndRef ind ->
+    | Some (VarRef _, _) -> assert false (* variables don't have instances *)
+    | Some (IndRef ind, nargs) ->
       let mind = Environ.lookup_mind (fst ind) env in
       cstrs := cmp_inductives cv_pb (mind,snd ind) nargs l l' !cstrs;
       true
-    | ConstructRef ((mi,ind),ctor) ->
+    | Some (ConstructRef ((mi,ind),ctor), nargs) ->
       let mind = Environ.lookup_mind mi env in
       cstrs := cmp_constructors (mind,ind,ctor) nargs l l' !cstrs;
       true
 
-let test_constr_universes env sigma leq m n =
+let test_constr_universes env sigma leq ?(nargs=0) m n =
   let open UnivProblem in
   let kind c = kind sigma c in
   if m == n then Some Set.empty
   else
     let cstrs = ref Set.empty in
     let cv_pb = if leq then Reduction.CUMUL else Reduction.CONV in
-    let eq_universes ref nargs l l' = eq_universes env sigma cstrs Reduction.CONV ref nargs l l'
-    and leq_universes ref nargs l l' = eq_universes env sigma cstrs cv_pb ref nargs l l' in
+    let eq_universes refargs l l' = eq_universes env sigma cstrs Reduction.CONV refargs l l'
+    and leq_universes refargs l l' = eq_universes env sigma cstrs cv_pb refargs l l' in
     let eq_sorts s1 s2 =
       let s1 = ESorts.kind sigma s1 in
       let s2 = ESorts.kind sigma s2 in
       if Sorts.equal s1 s2 then true
       else (cstrs := Set.add
-              (UEq (Sorts.univ_of_sort s1,Sorts.univ_of_sort s2)) !cstrs;
+              (UEq (s1, s2)) !cstrs;
             true)
     in
     let leq_sorts s1 s2 =
@@ -485,40 +625,41 @@ let test_constr_universes env sigma leq m n =
       if Sorts.equal s1 s2 then true
       else
         (cstrs := Set.add
-           (ULe (Sorts.univ_of_sort s1,Sorts.univ_of_sort s2)) !cstrs;
+           (ULe (s1, s2)) !cstrs;
          true)
     in
-    let rec eq_constr' nargs m n = compare_gen kind eq_universes eq_sorts eq_constr' nargs m n in
+    let eq_existential eq e1 e2 = eq_existential sigma (eq 0) e1 e2 in
+    let rec eq_constr' nargs m n = compare_gen kind eq_universes eq_sorts (eq_existential eq_constr') eq_constr' nargs m n in
     let res =
       if leq then
         let rec compare_leq nargs m n =
-          Constr.compare_head_gen_leq_with kind kind leq_universes leq_sorts
+          Constr.compare_head_gen_leq_with kind kind leq_universes leq_sorts (eq_existential eq_constr')
             eq_constr' leq_constr' nargs m n
         and leq_constr' nargs m n = m == n || compare_leq nargs m n in
-        compare_leq 0 m n
+        compare_leq nargs m n
       else
-        Constr.compare_head_gen_with kind kind eq_universes eq_sorts eq_constr' 0 m n
+        Constr.compare_head_gen_with kind kind eq_universes eq_sorts (eq_existential eq_constr') eq_constr' nargs m n
     in
     if res then Some !cstrs else None
 
-let eq_constr_universes env sigma m n =
-  test_constr_universes env sigma false m n
-let leq_constr_universes env sigma m n =
-  test_constr_universes env sigma true m n
+let eq_constr_universes env sigma ?nargs m n =
+  test_constr_universes env sigma false ?nargs m n
+let leq_constr_universes env sigma ?nargs m n =
+  test_constr_universes env sigma true ?nargs m n
 
-let compare_head_gen_proj env sigma equ eqs eqc' nargs m n =
+let compare_head_gen_proj env sigma equ eqs eqev eqc' nargs m n =
   let kind c = kind sigma c in
   match kind m, kind n with
   | Proj (p, c), App (f, args)
   | App (f, args), Proj (p, c) ->
       (match kind f with
-      | Const (p', u) when Constant.equal (Projection.constant p) p' ->
+      | Const (p', u) when Environ.QConstant.equal env (Projection.constant p) p' ->
           let npars = Projection.npars p in
           if Array.length args == npars + 1 then
             eqc' 0 c args.(npars)
           else false
       | _ -> false)
-  | _ -> Constr.compare_head_gen_with kind kind equ eqs eqc' nargs m n
+  | _ -> Constr.compare_head_gen_with kind kind equ eqs eqev eqc' nargs m n
 
 let eq_constr_universes_proj env sigma m n =
   let open UnivProblem in
@@ -532,11 +673,12 @@ let eq_constr_universes_proj env sigma m n =
       if Sorts.equal s1 s2 then true
       else
         (cstrs := Set.add
-           (UEq (Sorts.univ_of_sort s1, Sorts.univ_of_sort s2)) !cstrs;
+           (UEq (s1, s2)) !cstrs;
          true)
     in
+    let eq_existential eq e1 e2 = eq_existential sigma (eq 0) e1 e2 in
     let rec eq_constr' nargs m n =
-      m == n || compare_head_gen_proj env sigma eq_universes eq_sorts eq_constr' nargs m n
+      m == n || compare_head_gen_proj env sigma eq_universes eq_sorts (eq_existential eq_constr') eq_constr' nargs m n
     in
     let res = eq_constr' 0 m n in
     if res then Some !cstrs else None
@@ -546,20 +688,25 @@ let universes_of_constr sigma c =
   let rec aux s c =
     match kind sigma c with
     | Const (c, u) ->
-          LSet.fold LSet.add (Instance.levels (EInstance.kind sigma u)) s
+      Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s
     | Ind ((mind,_), u) | Construct (((mind,_),_), u) ->
-          LSet.fold LSet.add (Instance.levels (EInstance.kind sigma u)) s
+      Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s
     | Sort u ->
-       let sort = ESorts.kind sigma u in
-       if Sorts.is_small sort then s
-       else
-         let u = Sorts.univ_of_sort sort in
-         LSet.fold LSet.add (Universe.levels u) s
+      let sort = ESorts.kind sigma u in
+      if Sorts.is_small sort then s
+      else
+        Level.Set.fold Level.Set.add (Sorts.levels sort) s
     | Evar (k, args) ->
-       let concl = Evd.evar_concl (Evd.find sigma k) in
-       fold sigma aux (aux s concl) c
+      let concl = Evd.evar_concl (Evd.find sigma k) in
+      fold sigma aux (aux s concl) c
+    | Array (u,_,_,_) ->
+      let s = Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s in
+      fold sigma aux s c
+    | Case (_,u,_,_,_,_,_) ->
+      let s = Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s in
+      fold sigma aux s c
     | _ -> fold sigma aux s c
-  in aux LSet.empty c
+  in aux Level.Set.empty c
 
 open Context
 open Environ
@@ -567,7 +714,7 @@ open Environ
 let cast_list : type a b. (a,b) eq -> a list -> b list =
   fun Refl x -> x
 
-let cast_list_snd : type a b. (a,b) eq -> ('c * a) list -> ('c * b) list =
+let cast_vect : type a b. (a,b) eq -> a array -> b array =
   fun Refl x -> x
 
 let cast_rel_decl :
@@ -597,6 +744,8 @@ exception LocalOccur
 let to_constr = unsafe_to_constr
 let to_rel_decl = unsafe_to_rel_decl
 
+type instance = t array
+type instance_list = t list
 type substl = t list
 
 (** Operations that commute with evar-normalization *)
@@ -611,19 +760,37 @@ let substnl_decl subst n d = of_rel_decl (Vars.substnl_decl (cast_list unsafe_eq
 let substl_decl subst d = of_rel_decl (Vars.substl_decl (cast_list unsafe_eq subst) (to_rel_decl d))
 let subst1_decl c d = of_rel_decl (Vars.subst1_decl (to_constr c) (to_rel_decl d))
 
-let replace_vars subst c =
-  of_constr (Vars.replace_vars (cast_list_snd unsafe_eq subst) (to_constr c))
-let substn_vars n subst c = of_constr (Vars.substn_vars n subst (to_constr c))
-let subst_vars subst c = of_constr (Vars.subst_vars subst (to_constr c))
-let subst_var subst c = of_constr (Vars.subst_var subst (to_constr c))
+type substituend = Vars.substituend
+let make_substituend c = Vars.make_substituend (unsafe_to_constr c)
+let lift_substituend n s = of_constr (Vars.lift_substituend n s)
+
+let replace_vars = replace_vars
+
+(* (subst_var str t) substitute (Var str) by (Rel 1) in t *)
+let subst_var sigma str t = replace_vars sigma [(str, mkRel 1)] t
+
+(* (subst_vars [id1;...;idn] t) substitute (Var idj) by (Rel j) in t *)
+let substn_vars sigma p vars c =
+  let _,subst =
+    List.fold_left (fun (n,l) var -> ((n+1),(var, mkRel n)::l)) (p,[]) vars
+  in replace_vars sigma (List.rev subst) c
+
+let subst_vars sigma subst c = substn_vars sigma 1 subst c
 
 let subst_univs_level_constr subst c =
   of_constr (Vars.subst_univs_level_constr subst (to_constr c))
 
+let subst_instance_context subst ctx =
+  cast_rel_context (sym unsafe_eq) (Vars.subst_instance_context subst (cast_rel_context unsafe_eq ctx))
+
+let subst_instance_constr subst c =
+  of_constr (Vars.subst_instance_constr subst (to_constr c))
+
 (** Operations that dot NOT commute with evar-normalization *)
 let noccurn sigma n term =
   let rec occur_rec n c = match kind sigma c with
     | Rel m -> if Int.equal m n then raise LocalOccur
+    | Evar (_, l) -> SList.Skip.iter (fun c -> occur_rec n c) l
     | _ -> iter_with_binders sigma succ occur_rec n c
   in
   try occur_rec n term; true with LocalOccur -> false
@@ -631,6 +798,7 @@ let noccurn sigma n term =
 let noccur_between sigma n m term =
   let rec occur_rec n c = match kind sigma c with
     | Rel p -> if n<=p && p<n+m then raise LocalOccur
+    | Evar (_, l) -> SList.Skip.iter (fun c -> occur_rec n c) l
     | _        -> iter_with_binders sigma succ occur_rec n c
   in
   try occur_rec n term; true with LocalOccur -> false
@@ -638,6 +806,7 @@ let noccur_between sigma n m term =
 let closedn sigma n c =
   let rec closed_rec n c = match kind sigma c with
     | Rel m -> if m>n then raise LocalOccur
+    | Evar (_, l) -> SList.Skip.iter (fun c -> closed_rec n c) l
     | _ -> iter_with_binders sigma succ closed_rec n c
   in
   try closed_rec n c; true with LocalOccur -> false
@@ -646,7 +815,34 @@ let closed0 sigma c = closedn sigma 0 c
 
 let subst_of_rel_context_instance ctx subst =
   cast_list (sym unsafe_eq)
-    (Vars.subst_of_rel_context_instance (cast_rel_context unsafe_eq ctx) (cast_list unsafe_eq subst))
+    (Vars.subst_of_rel_context_instance (cast_rel_context unsafe_eq ctx) (cast_vect unsafe_eq subst))
+let subst_of_rel_context_instance_list ctx subst =
+  cast_list (sym unsafe_eq)
+    (Vars.subst_of_rel_context_instance_list (cast_rel_context unsafe_eq ctx) (cast_list unsafe_eq subst))
+
+let liftn_rel_context n k ctx =
+  cast_rel_context (sym unsafe_eq)
+    (Vars.liftn_rel_context n k (cast_rel_context unsafe_eq ctx))
+
+let lift_rel_context n ctx =
+  cast_rel_context (sym unsafe_eq)
+    (Vars.lift_rel_context n (cast_rel_context unsafe_eq ctx))
+
+let substnl_rel_context subst n ctx =
+  cast_rel_context (sym unsafe_eq)
+    (Vars.substnl_rel_context (cast_list unsafe_eq subst) n (cast_rel_context unsafe_eq ctx))
+
+let substl_rel_context subst ctx =
+  cast_rel_context (sym unsafe_eq)
+    (Vars.substl_rel_context (cast_list unsafe_eq subst) (cast_rel_context unsafe_eq ctx))
+
+let smash_rel_context ctx =
+  cast_rel_context (sym unsafe_eq)
+      (Vars.smash_rel_context (cast_rel_context unsafe_eq ctx))
+
+let esubst : (int -> 'a -> t) -> 'a Esubst.subs -> t -> t =
+match unsafe_eq with
+| Refl -> Vars.esubst
 
 end
 
@@ -672,37 +868,70 @@ let destArity sigma =
   in
   prodec_rec []
 
+(* Constructs either [forall x:t, c] or [let x:=b:t in c] *)
 let mkProd_or_LetIn decl c =
   let open Context.Rel.Declaration in
   match decl with
   | LocalAssum (na,t) -> mkProd (na, t, c)
   | LocalDef (na,b,t) -> mkLetIn (na, b, t, c)
 
+(* Constructs either [forall x:t, c] or [c] in which [x] is replaced by [b] *)
+let mkProd_wo_LetIn decl c =
+  let open Context.Rel.Declaration in
+  match decl with
+  | LocalAssum (na,t) -> mkProd (na, t, c)
+  | LocalDef (_,b,_) -> Vars.subst1 b c
+
 let mkLambda_or_LetIn decl c =
   let open Context.Rel.Declaration in
   match decl with
   | LocalAssum (na,t) -> mkLambda (na, t, c)
   | LocalDef (na,b,t) -> mkLetIn (na, b, t, c)
 
-let mkNamedProd id typ c = mkProd (map_annot Name.mk_name id, typ, Vars.subst_var id.binder_name c)
-let mkNamedLambda id typ c = mkLambda (map_annot Name.mk_name id, typ, Vars.subst_var id.binder_name c)
-let mkNamedLetIn id c1 t c2 = mkLetIn (map_annot Name.mk_name id, c1, t, Vars.subst_var id.binder_name c2)
+let mkLambda_wo_LetIn decl c =
+  let open Context.Rel.Declaration in
+  match decl with
+  | LocalAssum (na,t) -> mkLambda (na, t, c)
+  | LocalDef (_,b,_) -> Vars.subst1 b c
 
-let mkNamedProd_or_LetIn decl c =
+let mkNamedProd sigma id typ c = mkProd (map_annot Name.mk_name id, typ, Vars.subst_var sigma id.binder_name c)
+let mkNamedLambda sigma id typ c = mkLambda (map_annot Name.mk_name id, typ, Vars.subst_var sigma id.binder_name c)
+let mkNamedLetIn sigma id c1 t c2 = mkLetIn (map_annot Name.mk_name id, c1, t, Vars.subst_var sigma id.binder_name c2)
+
+let mkNamedProd_or_LetIn sigma decl c =
   let open Context.Named.Declaration in
   match decl with
-    | LocalAssum (id,t) -> mkNamedProd id t c
-    | LocalDef (id,b,t) -> mkNamedLetIn id b t c
+  | LocalAssum (id,t) -> mkNamedProd sigma id t c
+  | LocalDef (id,b,t) -> mkNamedLetIn sigma id b t c
 
-let mkNamedLambda_or_LetIn decl c =
+let mkNamedLambda_or_LetIn sigma decl c =
   let open Context.Named.Declaration in
   match decl with
-    | LocalAssum (id,t) -> mkNamedLambda id t c
-    | LocalDef (id,b,t) -> mkNamedLetIn id b t c
+  | LocalAssum (id,t) -> mkNamedLambda sigma id t c
+  | LocalDef (id,b,t) -> mkNamedLetIn sigma id b t c
+
+let mkNamedProd_wo_LetIn sigma decl c =
+  let open Context.Named.Declaration in
+  match decl with
+  | LocalAssum (id,t) -> mkNamedProd sigma id t c
+  | LocalDef (id,b,t) -> Vars.subst1 b c
+
+let it_mkProd init = List.fold_left (fun c (n,t)  -> mkProd (n, t, c)) init
+let it_mkLambda init = List.fold_left (fun c (n,t)  -> mkLambda (n, t, c)) init
+
+let compose_lam l b = it_mkLambda b l
 
 let it_mkProd_or_LetIn t ctx = List.fold_left (fun c d -> mkProd_or_LetIn d c) t ctx
 let it_mkLambda_or_LetIn t ctx = List.fold_left (fun c d -> mkLambda_or_LetIn d c) t ctx
 
+let it_mkProd_wo_LetIn t ctx = List.fold_left (fun c d -> mkProd_wo_LetIn d c) t ctx
+let it_mkLambda_wo_LetIn t ctx = List.fold_left (fun c d -> mkLambda_wo_LetIn d c) t ctx
+
+let it_mkNamedProd_or_LetIn sigma t ctx = List.fold_left (fun c d -> mkNamedProd_or_LetIn sigma d c) t ctx
+let it_mkNamedLambda_or_LetIn sigma t ctx = List.fold_left (fun c d -> mkNamedLambda_or_LetIn sigma d c) t ctx
+
+let it_mkNamedProd_wo_LetIn sigma t ctx = List.fold_left (fun c d -> mkNamedProd_wo_LetIn sigma d c) t ctx
+
 let push_rel d e = push_rel (cast_rel_decl unsafe_eq d) e
 let push_rel_context d e = push_rel_context (cast_rel_context unsafe_eq d) e
 let push_rec_types d e = push_rec_types (cast_rec_decl unsafe_eq d) e
@@ -717,7 +946,7 @@ let val_of_named_context e = val_of_named_context (cast_named_context unsafe_eq
 let named_context_of_val e = cast_named_context (sym unsafe_eq) (named_context_of_val e)
 
 let of_existential : Constr.existential -> existential =
-  let gen : type a b. (a,b) eq -> 'c * b list -> 'c * a list = fun Refl x -> x in
+  let gen : type a b. (a,b) eq -> 'c * b SList.t -> 'c * a SList.t = fun Refl x -> x in
   gen unsafe_eq
 
 let lookup_rel i e = cast_rel_decl (sym unsafe_eq) (lookup_rel i e)
@@ -733,6 +962,14 @@ let map_rel_context_in_env f env sign =
   in
   aux env [] (List.rev sign)
 
+let match_named_context_val :
+  named_context_val -> (named_declaration * named_context_val) option =
+  match unsafe_eq with
+  | Refl -> match_named_context_val
+
+let identity_subst_val : named_context_val -> t SList.t = fun ctx ->
+  SList.defaultn (List.length ctx.Environ.env_named_ctx) SList.empty
+
 let fresh_global ?loc ?rigid ?names env sigma reference =
   let (evd,t) = Evd.fresh_global ?loc ?rigid ?names env sigma reference in
   evd, t
@@ -757,7 +994,7 @@ let kind_of_type sigma t = match kind sigma t with
   | (Rel _ | Meta _ | Var _ | Evar _ | Const _
   | Proj _ | Case _ | Fix _ | CoFix _ | Ind _)
     -> AtomicType (t,[||])
-  | (Lambda _ | Construct _ | Int _ | Float _) -> failwith "Not a type"
+  | (Lambda _ | Construct _ | Int _ | Float _ | Array _) -> failwith "Not a type"
 
 module Unsafe =
 struct
@@ -772,5 +1009,12 @@ let to_named_context =
     = fun Refl x -> x
   in
   gen unsafe_eq
+let to_rel_context =
+  let gen : type a b. (a, b) eq -> (a,a) Context.Rel.pt -> (b,b) Context.Rel.pt
+    = fun Refl x -> x
+  in
+  gen unsafe_eq
+let to_case_invert = unsafe_to_case_invert
+
 let eq = unsafe_eq
 end
