@@ -21,14 +21,11 @@ open Namegen
 module RelDecl = Context.Rel.Declaration
 module NamedDecl = Context.Named.Declaration
 
-let safe_evar_value sigma ev =
-  let ev = EConstr.of_existential ev in
-  try Some (EConstr.Unsafe.to_constr @@ Evd.existential_value sigma ev)
-  with NotInstantiatedEvar | Not_found -> None
+let create_clos_infos env sigma flags =
+  let open CClosure in
+  let evars = Evd.evar_handler sigma in
+  create_clos_infos ~univs:(Evd.universes sigma) ~evars flags env
 
-let new_global evd x =
-  let (evd, c) = Evd.fresh_global (Global.env()) evd x in
-  (evd, c)
 
 (****************************************************)
 (* Expanding/testing/exposing existential variables *)
@@ -36,11 +33,11 @@ let new_global evd x =
 
 let finalize ?abort_on_undefined_evars sigma f =
   let sigma = minimize_universes sigma in
-  let uvars = ref Univ.LSet.empty in
+  let uvars = ref Univ.Level.Set.empty in
   let v = f (fun c ->
       let varsc = EConstr.universes_of_constr sigma c in
       let c = EConstr.to_constr ?abort_on_undefined_evars sigma c in
-      uvars := Univ.LSet.union !uvars varsc;
+      uvars := Univ.Level.Set.union !uvars varsc;
       c)
   in
   let sigma = restrict_universe_context sigma !uvars in
@@ -64,9 +61,12 @@ let flush_and_check_evars sigma c =
 (** Term exploration up to instantiation. *)
 let kind_of_term_upto = EConstr.kind_upto
 
-let nf_evar0 sigma t = EConstr.to_constr ~abort_on_undefined_evars:false sigma (EConstr.of_constr t)
+let nf_evars_universes sigma t = EConstr.to_constr ~abort_on_undefined_evars:false sigma (EConstr.of_constr t)
 let whd_evar = EConstr.whd_evar
-let nf_evar sigma c = EConstr.of_constr (EConstr.to_constr ~abort_on_undefined_evars:false sigma c)
+
+let nf_evar sigma c =
+  let evar_value ev = Evd.existential_opt_value0 sigma ev in
+  EConstr.of_constr @@ UnivSubst.nf_evars_and_universes_opt_subst evar_value (universe_subst sigma) (EConstr.Unsafe.to_constr c)
 
 let j_nf_evar sigma j =
   { uj_val = nf_evar sigma j.uj_val;
@@ -76,12 +76,8 @@ let jv_nf_evar sigma = Array.map (j_nf_evar sigma)
 let tj_nf_evar sigma {utj_val=v;utj_type=t} =
   {utj_val=nf_evar sigma v;utj_type=t}
 
-let nf_evars_universes evm =
-  UnivSubst.nf_evars_and_universes_opt_subst (safe_evar_value evm)
-    (Evd.universe_subst evm)
-
 let nf_named_context_evar sigma ctx =
-  Context.Named.map (nf_evar0 sigma) ctx
+  Context.Named.map (nf_evars_universes sigma) ctx
 
 let nf_rel_context_evar sigma ctx =
   Context.Rel.map (nf_evar sigma) ctx
@@ -124,17 +120,6 @@ let is_ground_env evd env =
   List.for_all is_ground_rel_decl (rel_context env) &&
   List.for_all is_ground_named_decl (named_context env)
 
-(* Memoization is safe since evar_map and environ are applicative
-   structures *)
-let memo f =
-  let module E = Ephemeron.K2 in
-  let m = E.create () in
-  fun x y -> match E.get_key1 m, E.get_key2 m with
-  | Some x', Some y' when x == x' && y == y' -> Option.get (E.get_data m)
-  | _ -> let r = f x y in E.set_key1 m x; E.set_key2 m y; E.set_data m r; r
-
-let is_ground_env = memo is_ground_env
-
 (* Return the head evar if any *)
 
 exception NoHeadEvar
@@ -144,7 +129,7 @@ let head_evar sigma c =
   let c = EConstr.Unsafe.to_constr c in
   let rec hrec c = match kind c with
     | Evar (evk,_)   -> evk
-    | Case (_,_,c,_) -> hrec c
+    | Case (_, _, _, _, _, c, _) -> hrec c
     | App (c,_)      -> hrec c
     | Cast (c,_,_)   -> hrec c
     | Proj (p, c)    -> hrec c
@@ -183,26 +168,8 @@ let meta_ctr, meta_counter_summary_tag =
 
 let new_meta () = incr meta_ctr; !meta_ctr
 
-let mk_new_meta () = EConstr.mkMeta(new_meta())
-
 (* The list of non-instantiated existential declarations (order is important) *)
 
-let non_instantiated sigma =
-  let listev = Evd.undefined_map sigma in
-  Evar.Map.Smart.map (fun evi -> nf_evar_info sigma evi) listev
-
-(************************)
-(* Manipulating filters *)
-(************************)
-
-let make_pure_subst evi args =
-  snd (List.fold_right
-    (fun decl (args,l) ->
-      match args with
-        | a::rest -> (rest, (NamedDecl.get_id decl, a)::l)
-        | _ -> anomaly (Pp.str "Instance does not match its signature."))
-    (evar_filtered_context evi) (List.rev args,[]))
-
 (*------------------------------------*
  * functional operations on evar sets *
  *------------------------------------*)
@@ -240,11 +207,6 @@ let next_ident_away id avoid =
   let avoid id = Id.Set.mem id avoid in
   next_ident_away_from id avoid
 
-let next_name_away na avoid =
-  let avoid id = Id.Set.mem id avoid in
-  let id = match na with Name id -> id | Anonymous -> default_non_dependent_ident in
-  next_ident_away_from id avoid
-
 type subst_val =
 | SRel of int
 | SVar of Id.t
@@ -272,7 +234,7 @@ let empty_csubst = {
   csubst_rev = Id.Map.empty;
 }
 
-let csubst_subst { csubst_len = k; csubst_var = v; csubst_rel = s } c =
+let csubst_subst sigma { csubst_len = k; csubst_var = v; csubst_rel = s } c =
   (* Safe because this is a substitution *)
   let c = EConstr.Unsafe.to_constr c in
   let rec subst n c = match Constr.kind c with
@@ -282,13 +244,29 @@ let csubst_subst { csubst_len = k; csubst_var = v; csubst_rel = s } c =
     else mkRel (m - k)
   | Var id ->
     begin try Id.Map.find id v with Not_found -> c end
+  | Evar (evk, args) ->
+    let evi = Evd.find sigma evk in
+    let args' = subst_instance n (evar_filtered_context evi) args in
+    if args' == args then c else Constr.mkEvar (evk, args') (* FIXME: preserve sharing *)
   | _ -> Constr.map_with_binders succ subst n c
+
+  and subst_instance n ctx args = match ctx, SList.view args with
+  | [], None -> SList.empty
+  | decl :: ctx, Some (c, args) ->
+    let c' = match c with
+    | None -> begin try Some (Id.Map.find (NamedDecl.get_id decl) v) with Not_found -> c end
+    | Some c ->
+      let c' = subst n c in
+      if isVarId (NamedDecl.get_id decl) c' then None else Some c'
+    in
+    SList.cons_opt c' (subst_instance n ctx args)
+  | _ :: _, None | [], Some _ -> assert false
   in
   let c = if k = 0 && Id.Map.is_empty v then c else subst 0 c in
   EConstr.of_constr c
 
 type ext_named_context =
-  csubst * Id.Set.t * EConstr.named_context
+  csubst * Id.Set.t * named_context_val
 
 let push_var id { csubst_len = n; csubst_var = v; csubst_rel = s; csubst_rev = r } =
   let s = Int.Map.add n (Constr.mkVar id) s in
@@ -317,31 +295,37 @@ let update_var src tgt subst =
       let csubst_var = Id.Map.add id (Constr.mkVar tgt) subst.csubst_var in
       { subst with csubst_var; csubst_rev }
 
+module VarSet =
+struct
+  type t = Id.t -> bool
+  let empty _ = false
+  let full _ = true
+  let variables env id = is_section_variable env id
+end
+
 type naming_mode =
-  | KeepUserNameAndRenameExistingButSectionNames
-  | KeepUserNameAndRenameExistingEvenSectionNames
-  | KeepExistingNames
+  | RenameExistingBut of VarSet.t
   | FailIfConflict
-  | ProgramNaming
+  | ProgramNaming of VarSet.t
 
 let push_rel_decl_to_named_context
-  ?(hypnaming=KeepUserNameAndRenameExistingButSectionNames)
-  sigma decl (subst, avoid, nc) =
+  ~hypnaming
+  sigma decl ((subst, avoid, nc) : ext_named_context) =
   let open EConstr in
   let open Vars in
   let map_decl f d =
     NamedDecl.map_constr f d
   in
-  let rec replace_var_named_declaration id0 id = function
-  | [] -> []
-  | decl :: nc ->
+  let rec replace_var_named_declaration id0 id nc = match match_named_context_val nc with
+  | None -> empty_named_context_val
+  | Some (decl, nc) ->
     if Id.equal id0 (NamedDecl.get_id decl) then
       (* Stop here, the variable cannot occur before its definition *)
-      (NamedDecl.set_id id decl) :: nc
+      push_named_context_val (NamedDecl.set_id id decl) nc
     else
       let nc = replace_var_named_declaration id0 id nc in
       let vsubst = [id0 , mkVar id] in
-      map_decl (fun c -> replace_vars vsubst c) decl :: nc
+      push_named_context_val (map_decl (fun c -> replace_vars sigma vsubst c) decl) nc
   in
   let extract_if_neq id = function
     | Anonymous -> None
@@ -350,128 +334,101 @@ let push_rel_decl_to_named_context
   in
   let na = RelDecl.get_name decl in
   let id =
-    (* ppedrot: we want to infer nicer names for the refine tactic, but
-        keeping at the same time backward compatibility in other code
-        using this function. For now, we only attempt to preserve the
-        old behaviour of Program, but ultimately, one should do something
-        about this whole name generation problem. *)
-    if hypnaming = ProgramNaming then next_name_away na avoid
-    else
-      (* id_of_name_using_hdchar only depends on the rel context which is empty
-         here *)
-      next_ident_away (id_of_name_using_hdchar empty_env sigma (RelDecl.get_type decl) na) avoid
+    (* id_of_name_using_hdchar only depends on the rel context which is empty
+        here *)
+    next_ident_away (id_of_name_using_hdchar empty_env sigma (RelDecl.get_type decl) na) avoid
   in
   match extract_if_neq id na with
-  | Some id0 when hypnaming = KeepUserNameAndRenameExistingEvenSectionNames ||
-                  (hypnaming = KeepUserNameAndRenameExistingButSectionNames ||
-                   hypnaming = ProgramNaming) &&
-                  not (is_section_variable id0) ->
-      (* spiwack: if [id<>id0], rather than introducing a new
-          binding named [id], we will keep [id0] (the name given
-          by the user) and rename [id0] into [id] in the named
-          context. Unless [id] is a section variable. *)
-      let subst = update_var id0 id subst in
-      let d = decl |> NamedDecl.of_rel_decl (fun _ -> id0) |> map_decl (csubst_subst subst) in
-      let nc = replace_var_named_declaration id0 id nc in
-      (push_var id0 subst, Id.Set.add id avoid, d :: nc)
-  | Some id0 when hypnaming = FailIfConflict ->
+  | Some id0 ->
+    begin match hypnaming with
+    | RenameExistingBut f | ProgramNaming f ->
+      if f id0 then
+        (* spiwack: if [id0] is a section variable renaming it is
+            incorrect. We revert to a less robust behaviour where
+            the new binder has name [id]. Which amounts to the same
+            behaviour than when [id=id0]. *)
+        let d = decl |> NamedDecl.of_rel_decl (fun _ -> id) |> map_decl (csubst_subst sigma subst) in
+        (push_var id subst, Id.Set.add id avoid, push_named_context_val d nc)
+      else
+        (* spiwack: if [id<>id0], rather than introducing a new
+            binding named [id], we will keep [id0] (the name given
+            by the user) and rename [id0] into [id] in the named
+            context. Unless [id] is a section variable. *)
+        let subst = update_var id0 id subst in
+        let d = decl |> NamedDecl.of_rel_decl (fun _ -> id0) |> map_decl (csubst_subst sigma subst) in
+        let nc = replace_var_named_declaration id0 id nc in
+        let avoid = Id.Set.add id (Id.Set.add id0 avoid) in
+        (push_var id0 subst, avoid, push_named_context_val d nc)
+    | FailIfConflict ->
        user_err Pp.(Id.print id0 ++ str " is already used.")
-  | _ ->
-      (* spiwack: if [id0] is a section variable renaming it is
-          incorrect. We revert to a less robust behaviour where
-          the new binder has name [id]. Which amounts to the same
-          behaviour than when [id=id0]. *)
-      let d = decl |> NamedDecl.of_rel_decl (fun _ -> id) |> map_decl (csubst_subst subst) in
-      (push_var id subst, Id.Set.add id avoid, d :: nc)
+    end
+  | None ->
+    let d = decl |> NamedDecl.of_rel_decl (fun _ -> id) |> map_decl (csubst_subst sigma subst) in
+    (push_var id subst, Id.Set.add id avoid, push_named_context_val d nc)
 
-let push_rel_context_to_named_context ?hypnaming env sigma typ =
+let csubst_instance subst ctx =
+  let fold decl accu = match Id.Map.find (NamedDecl.get_id decl) subst.csubst_rev with
+  | SRel n -> SList.cons (EConstr.mkRel (subst.csubst_len - n)) accu
+  | SVar id -> SList.cons (EConstr.mkVar id) accu
+  | exception Not_found -> SList.default accu
+  in
+  List.fold_right fold ctx SList.empty
+
+let default_ext_instance (subst, _, ctx) =
+  csubst_instance subst (named_context_of_val ctx)
+
+let push_rel_context_to_named_context ~hypnaming env sigma typ =
   (* compute the instances relative to the named context and rel_context *)
-  let open Context.Named.Declaration in
   let open EConstr in
-  let ids = List.map get_id (named_context env) in
-  let inst_vars = List.map mkVar ids in
+  let ctx = named_context_val env in
   if List.is_empty (Environ.rel_context env) then
-    (named_context_val env, typ, inst_vars, empty_csubst)
+    let inst = SList.defaultn (List.length @@ named_context_of_val ctx) SList.empty in
+    (ctx, typ, inst, empty_csubst)
   else
-    let avoid = List.fold_right Id.Set.add ids Id.Set.empty in
-    let inst_rels = List.rev (rel_list 0 (nb_rel env)) in
+    let avoid = Environ.ids_of_named_context_val (named_context_val env) in
     (* move the rel context to a named context and extend the named instance *)
     (* with vars of the rel context *)
     (* We do keep the instances corresponding to local definition (see above) *)
-    let (subst, _, env) =
-      Context.Rel.fold_outside (fun d acc -> push_rel_decl_to_named_context ?hypnaming sigma d acc)
-        (rel_context env) ~init:(empty_csubst, avoid, named_context env) in
-    (val_of_named_context env, csubst_subst subst typ, inst_rels@inst_vars, subst)
+    let (subst, _, env) as ext =
+      Context.Rel.fold_outside (fun d acc -> push_rel_decl_to_named_context ~hypnaming sigma d acc)
+        (rel_context env) ~init:(empty_csubst, avoid, ctx) in
+    let inst = default_ext_instance ext in
+    (env, csubst_subst sigma subst typ, inst, subst)
 
 (*------------------------------------*
  * Entry points to define new evars   *
  *------------------------------------*)
 
-let default_source = Loc.tag @@ Evar_kinds.InternalHole
+let new_pure_evar = Evd.new_pure_evar
 
-let new_pure_evar_full evd ?typeclass_candidate evi =
-  let (evd, evk) = Evd.new_evar evd ?typeclass_candidate evi in
-  let evd = Evd.declare_future_goal evk evd in
-  (evd, evk)
-
-let new_pure_evar?(src=default_source) ?(filter = Filter.identity) ?(abstract_arguments = Abstraction.identity)
-    ?candidates ?naming ?typeclass_candidate ?(principal=false) sign evd typ =
-  let default_naming = IntroAnonymous in
-  let naming = Option.default default_naming naming in
-  let name = match naming with
-  | IntroAnonymous -> None
-  | IntroIdentifier id -> Some id
-  | IntroFresh id ->
-    let has_name id = try let _ = Evd.evar_key id evd in true with Not_found -> false in
-    let id = Namegen.next_ident_away_from id has_name in
-    Some id
-  in
-  let evi = {
-    evar_hyps = sign;
-    evar_concl = typ;
-    evar_body = Evar_empty;
-    evar_filter = filter;
-    evar_abstract_arguments = abstract_arguments;
-    evar_source = src;
-    evar_candidates = candidates }
-  in
-  let typeclass_candidate = if principal then Some false else typeclass_candidate in
-  let (evd, newevk) = Evd.new_evar evd ?name ?typeclass_candidate evi in
-  let evd =
-    if principal then Evd.declare_principal_goal newevk evd
-    else Evd.declare_future_goal newevk evd
-  in
-  (evd, newevk)
-
-let new_evar_instance ?src ?filter ?abstract_arguments ?candidates ?naming ?typeclass_candidate
-    ?principal sign evd typ instance =
-  let open EConstr in
-  assert (not !Flags.debug ||
-            List.distinct (ids_of_named_context (named_context_of_val sign)));
-  let (evd, newevk) = new_pure_evar sign evd ?src ?filter ?abstract_arguments ?candidates ?naming ?typeclass_candidate ?principal typ in
-  evd, mkEvar (newevk, instance)
-
-let new_evar_from_context ?src ?filter ?candidates ?naming ?typeclass_candidate ?principal sign evd typ =
-  let instance = List.map (NamedDecl.get_id %> EConstr.mkVar) (named_context_of_val sign) in
-  let instance =
-    match filter with
-    | None -> instance
-    | Some filter -> Filter.filter_list filter instance in
-  new_evar_instance sign evd typ ?src ?filter ?candidates ?naming ?principal instance
+let next_evar_name sigma naming = match naming with
+| IntroAnonymous -> None
+| IntroIdentifier id -> Some id
+| IntroFresh id ->
+  let has_name id = try let _ = Evd.evar_key id sigma in true with Not_found -> false in
+  let id = Namegen.next_ident_away_from id has_name in
+  Some id
 
 (* [new_evar] declares a new existential in an env env with type typ *)
 (* Converting the env into the sign of the evar to define *)
-let new_evar ?src ?filter ?abstract_arguments ?candidates ?naming ?typeclass_candidate
+let new_evar ?src ?filter ?abstract_arguments ?candidates ?(naming = IntroAnonymous) ?typeclass_candidate
     ?principal ?hypnaming env evd typ =
-  let sign,typ',instance,subst = push_rel_context_to_named_context ?hypnaming env evd typ in
-  let map c = csubst_subst subst c in
+  let name = next_evar_name evd naming in
+  let hypnaming = match hypnaming with
+  | Some n -> n
+  | None -> RenameExistingBut (VarSet.variables (Global.env ()))
+  in
+  let sign,typ',instance,subst = push_rel_context_to_named_context ~hypnaming env evd typ in
+  let map c = csubst_subst evd subst c in
   let candidates = Option.map (fun l -> List.map map l) candidates in
   let instance =
     match filter with
     | None -> instance
-    | Some filter -> Filter.filter_list filter instance in
-  new_evar_instance sign evd typ' ?src ?filter ?abstract_arguments ?candidates ?naming
-    ?typeclass_candidate ?principal instance
+    | Some filter -> Filter.filter_slist filter instance in
+  let relevance = Sorts.Relevant in (* FIXME: relevant_of_type not defined yet *)
+  let (evd, evk) = new_pure_evar sign evd typ' ?src ?filter ~relevance ?abstract_arguments ?candidates ?name
+    ?typeclass_candidate ?principal in
+  (evd, EConstr.mkEvar (evk, instance))
 
 let new_type_evar ?src ?filter ?naming ?principal ?hypnaming env evd rigid =
   let (evd', s) = new_sort_variable rigid evd in
@@ -503,11 +460,12 @@ let add_unification_pb ?(tail=false) pb evd =
 let generalize_evar_over_rels sigma (ev,args) =
   let open EConstr in
   let evi = Evd.find sigma ev in
-  let sign = named_context_of_val evi.evar_hyps in
+  let args = Evd.expand_existential sigma (ev, args) in
+  let sign = named_context_of_val (Evd.evar_hyps evi) in
   List.fold_left2
     (fun (c,inst as x) a d ->
-      if isRel sigma a then (mkNamedProd_or_LetIn d c,a::inst) else x)
-     (evi.evar_concl,[]) args sign
+      if isRel sigma a then (mkNamedProd_or_LetIn sigma d c,a::inst) else x)
+     (Evd.evar_concl evi,[]) args sign
 
 (************************************)
 (* Removing a dependency in an evar *)
@@ -534,16 +492,9 @@ let restrict_evar evd evk filter ?src candidates =
   let candidates = Option.map (filter_effective_candidates evd evar_info filter) candidates in
   match candidates with
   | Some [] -> raise (ClearDependencyError (*FIXME*)(Id.of_string "blah", (NoCandidatesLeft evk), None))
-  | _ ->
-     let evd, evk' = Evd.restrict evk filter ?candidates ?src evd in
-     (* Mark new evar as future goal, removing previous one,
-        circumventing Proofview.advance but making Proof.run_tactic catch these. *)
-     let future_goals = Evd.save_future_goals evd in
-     let future_goals = Evd.filter_future_goals (fun evk' -> not (Evar.equal evk evk')) future_goals in
-     let evd = Evd.restore_future_goals evd future_goals in
-     (Evd.declare_future_goal evk' evd, evk')
+  | _ -> Evd.restrict evk filter ?candidates ?src evd
 
-let rec check_and_clear_in_constr env evdref err ids global c =
+let rec check_and_clear_in_constr ~is_section_variable env evdref err ids ~global c =
   (* returns a new constr where all the evars have been 'cleaned'
      (ie the hypotheses ids have been removed from the contexts of
      evars). [global] should be true iff there is some variable of [ids] which
@@ -567,7 +518,7 @@ let rec check_and_clear_in_constr env evdref err ids global c =
             (* If evk is already defined we replace it by its definition *)
             let nc = Evd.existential_value !evdref (EConstr.of_existential ev) in
             let nc = EConstr.Unsafe.to_constr nc in
-              (check_and_clear_in_constr env evdref err ids global nc)
+            check_and_clear_in_constr ~is_section_variable env evdref err ids ~global nc
           else
             (* We check for dependencies to elements of ids in the
                evar_info corresponding to e and in the instance of
@@ -576,34 +527,41 @@ let rec check_and_clear_in_constr env evdref err ids global c =
                removed *)
             let evi = Evd.find_undefined !evdref evk in
             let ctxt = Evd.evar_filtered_context evi in
-            let (rids,filter) =
-              List.fold_right2
-                (fun h a (ri,filter) ->
-                  try
-                  (* Check if some id to clear occurs in the instance
-                     a of rid in ev and remember the dependency *)
-                    let check id = if Id.Set.mem id ids then raise (Depends id) in
-                    let () = Id.Set.iter check (collect_vars !evdref (EConstr.of_constr a)) in
-                  (* Check if some rid to clear in the context of ev
-                     has dependencies in another hyp of the context of ev
-                     and transitively remember the dependency *)
-                    let check id _ =
-                      if occur_var_in_decl env !evdref id h
-                      then raise (Depends id)
-                    in
-                    let () = Id.Map.iter check ri in
-                  (* No dependency at all, we can keep this ev's context hyp *)
-                    (ri, true::filter)
-                  with Depends id -> (Id.Map.add (NamedDecl.get_id h) id ri, false::filter))
-                ctxt l (Id.Map.empty,[]) in
+            let rec fold accu ctxt args = match ctxt, SList.view args with
+            | [], Some _ | _ :: _, None -> assert false
+            | [], None -> accu
+            | h :: ctxt, Some (a, args) ->
+              let (ri, filter) = fold accu ctxt args in
+              try
+              (* Check if some id to clear occurs in the instance
+                  a of rid in ev and remember the dependency *)
+                let check id = if Id.Set.mem id ids then raise (Depends id) in
+                let a = match a with
+                | None -> Id.Set.singleton (NamedDecl.get_id h)
+                | Some a -> collect_vars !evdref (EConstr.of_constr a)
+                in
+                let () = Id.Set.iter check a in
+              (* Check if some rid to clear in the context of ev
+                  has dependencies in another hyp of the context of ev
+                  and transitively remember the dependency *)
+                let check id _ =
+                  if occur_var_in_decl env !evdref id h
+                  then raise (Depends id)
+                in
+                let () = Id.Map.iter check ri in
+              (* No dependency at all, we can keep this ev's context hyp *)
+                (ri, true::filter)
+              with Depends id -> (Id.Map.add (NamedDecl.get_id h) id ri, false::filter)
+            in
+            let (rids, filter) = fold (Id.Map.empty, []) ctxt l in
             (* Check if some rid to clear in the context of ev has dependencies
                in the type of ev and adjust the source of the dependency *)
-            let _nconcl =
+            let _nconcl : Constr.t =
               try
                 let nids = Id.Map.domain rids in
                 let global = Id.Set.exists is_section_variable nids in
                 let concl = EConstr.Unsafe.to_constr (evar_concl evi) in
-                check_and_clear_in_constr env evdref (EvarTypingBreak ev) nids global concl
+                check_and_clear_in_constr ~is_section_variable env evdref (EvarTypingBreak ev) nids ~global concl
               with ClearDependencyError (rid,err,where) ->
                 raise (ClearDependencyError (Id.Map.find rid rids,err,where)) in
 
@@ -613,12 +571,11 @@ let rec check_and_clear_in_constr env evdref err ids global c =
               let filter = Evd.Filter.apply_subfilter origfilter filter in
               let evd = !evdref in
               let candidates = Evd.evar_candidates evi in
-              let candidates = Option.map (List.map EConstr.of_constr) candidates in
               let (evd,_) = restrict_evar evd evk filter candidates in
               evdref := evd;
               Evd.existential_value0 !evdref ev
 
-      | _ -> Constr.map (check_and_clear_in_constr env evdref err ids global) c
+      | _ -> Constr.map (check_and_clear_in_constr ~is_section_variable env evdref err ids ~global) c
 
 let clear_hyps_in_evi_main env sigma hyps terms ids =
   (* clear_hyps_in_evi erases hypotheses ids in hyps, checking if some
@@ -626,29 +583,28 @@ let clear_hyps_in_evi_main env sigma hyps terms ids =
      the contexts of the evars occurring in evi *)
   let evdref = ref sigma in
   let terms = List.map EConstr.Unsafe.to_constr terms in
+  let is_section_variable id = is_section_variable (Global.env ()) id in
   let global = Id.Set.exists is_section_variable ids in
   let terms =
-    List.map (check_and_clear_in_constr env evdref (OccurHypInSimpleClause None) ids global) terms in
+    List.map (check_and_clear_in_constr ~is_section_variable env evdref (OccurHypInSimpleClause None) ids ~global) terms in
   let nhyps =
     let check_context decl =
       let err = OccurHypInSimpleClause (Some (NamedDecl.get_id decl)) in
-      NamedDecl.map_constr (check_and_clear_in_constr env evdref err ids global) decl
+      NamedDecl.map_constr (check_and_clear_in_constr ~is_section_variable env evdref err ids ~global) decl
     in
-    let check_value vk = match force_lazy_val vk with
-    | None -> vk
-    | Some (_, d) ->
-      if (Id.Set.for_all (fun e -> not (Id.Set.mem e d)) ids) then
-        (* v does depend on any of ids, it's ok *)
-        vk
-      else
-        (* v depends on one of the cleared hyps:
-            we forget the computed value *)
-        dummy_lazy_val ()
-    in
-      remove_hyps ids check_context check_value hyps
+    remove_hyps ids check_context hyps
   in
   (!evdref, nhyps,List.map EConstr.of_constr terms)
 
+let check_and_clear_in_constr env evd err ids c =
+  let evdref = ref evd in
+  let c = EConstr.Unsafe.to_constr c in
+  let _ : constr = check_and_clear_in_constr
+      ~is_section_variable:(fun _ -> true) ~global:true
+      env evdref err ids c
+  in
+  !evdref
+
 let clear_hyps_in_evi env sigma hyps concl ids =
   match clear_hyps_in_evi_main env sigma hyps [concl] ids with
   | (sigma,nhyps,[nconcl]) -> (sigma,nhyps,nconcl)
@@ -659,54 +615,6 @@ let clear_hyps2_in_evi env sigma hyps t concl ids =
   | (sigma,nhyps,[t;nconcl]) -> (sigma,nhyps,t,nconcl)
   | _ -> assert false
 
-(* spiwack: a few functions to gather evars on which goals depend. *)
-let queue_set q is_dependent set =
-  Evar.Set.iter (fun a -> Queue.push (is_dependent,a) q) set
-let queue_term q is_dependent c =
-  queue_set q is_dependent (evar_nodes_of_term c)
-
-let process_dependent_evar q acc evm is_dependent e =
-  let evi = Evd.find evm e in
-  (* Queues evars appearing in the types of the goal (conclusion, then
-     hypotheses), they are all dependent. *)
-  queue_term q true evi.evar_concl;
-  List.iter begin fun decl ->
-    let open NamedDecl in
-    queue_term q true (NamedDecl.get_type decl);
-    match decl with
-    | LocalAssum _ -> ()
-    | LocalDef (_,b,_) -> queue_term q true b
-  end (EConstr.named_context_of_val evi.evar_hyps);
-  match evi.evar_body with
-  | Evar_empty ->
-      if is_dependent then Evar.Map.add e None acc else acc
-  | Evar_defined b ->
-      let subevars = evar_nodes_of_term b in
-      (* evars appearing in the definition of an evar [e] are marked
-         as dependent when [e] is dependent itself: if [e] is a
-         non-dependent goal, then, unless they are reach from another
-         path, these evars are just other non-dependent goals. *)
-      queue_set q is_dependent subevars;
-      if is_dependent then Evar.Map.add e (Some subevars) acc else acc
-
-let gather_dependent_evars q evm =
-  let acc = ref Evar.Map.empty in
-  while not (Queue.is_empty q) do
-    let (is_dependent,e) = Queue.pop q in
-    (* checks if [e] has already been added to [!acc] *)
-    begin if not (Evar.Map.mem e !acc) then
-        acc := process_dependent_evar q !acc evm is_dependent e
-    end
-  done;
-  !acc
-
-let gather_dependent_evars evm l =
-  let q = Queue.create () in
-  List.iter (fun a -> Queue.add (false,a) q) l;
-  gather_dependent_evars q evm
-
-(* /spiwack *)
-
 (** [advance sigma g] returns [Some g'] if [g'] is undefined and is
     the current avatar of [g] (for instance [g] was changed by [clear]
     into [g']). It returns [None] if [g] has been (partially)
@@ -720,13 +628,25 @@ let gather_dependent_evars evm l =
    goal) it terminates quickly. *)
 let rec advance sigma evk =
   let evi = Evd.find sigma evk in
-  match evi.evar_body with
+  match Evd.evar_body evi with
   | Evar_empty -> Some evk
   | Evar_defined v ->
-      match is_restricted_evar sigma evk with
+      match is_aliased_evar sigma evk with
       | Some evk -> advance sigma evk
       | None -> None
 
+let reachable_from_evars sigma evars =
+  let aliased = Evd.get_aliased_evars sigma in
+  let rec search evk visited =
+    if Evar.Set.mem evk visited then visited
+    else
+      let visited = Evar.Set.add evk visited in
+      match Evar.Map.find evk aliased with
+      | evk' -> search evk' visited
+      | exception Not_found -> visited
+  in
+  Evar.Set.fold (fun evk visited -> search evk visited) evars Evar.Set.empty
+
 (** The following functions return the set of undefined evars
     contained in the object, the defined evars being traversed.
     This is roughly a combination of the previous functions and
@@ -737,7 +657,7 @@ let undefined_evars_of_term evd t =
     match EConstr.kind evd c with
       | Evar (n, l) ->
         let acc = Evar.Set.add n acc in
-        List.fold_left evrec acc l
+        SList.Skip.fold evrec acc l
       | _ -> EConstr.fold evd evrec acc c
   in
   evrec Evar.Set.empty t
@@ -749,13 +669,13 @@ let undefined_evars_of_named_context evd nc =
     ~init:Evar.Set.empty
 
 let undefined_evars_of_evar_info evd evi =
-  Evar.Set.union (undefined_evars_of_term evd evi.evar_concl)
+  Evar.Set.union (undefined_evars_of_term evd (Evd.evar_concl evi))
     (Evar.Set.union
-       (match evi.evar_body with
+       (match Evd.evar_body evi with
          | Evar_empty -> Evar.Set.empty
          | Evar_defined b -> undefined_evars_of_term evd b)
        (undefined_evars_of_named_context evd
-          (named_context_of_val evi.evar_hyps)))
+          (named_context_of_val (Evd.evar_hyps evi))))
 
 type undefined_evars_cache = {
   mutable cache : (EConstr.named_declaration * Evar.Set.t) ref Id.Map.t;
@@ -799,11 +719,11 @@ let filtered_undefined_evars_of_evar_info ?cache sigma evi =
     let fold decl accu = cached_evar_of_hyp cache sigma (EConstr.of_named_decl decl) accu in
     Context.Named.fold_outside fold nc ~init:accu
   in
-  let accu = match evi.evar_body with
+  let accu = match Evd.evar_body evi with
   | Evar_empty -> Evar.Set.empty
   | Evar_defined b -> evars_of_term sigma b
   in
-  let accu = Evar.Set.union (undefined_evars_of_term sigma evi.evar_concl) accu in
+  let accu = Evar.Set.union (undefined_evars_of_term sigma (Evd.evar_concl evi)) accu in
   let ctxt = EConstr.Unsafe.to_named_context (evar_filtered_context evi) in
   evars_of_named_context cache accu ctxt
 
@@ -825,8 +745,8 @@ let occur_evar_upto sigma n c =
 
 let judge_of_new_Type evd =
   let open EConstr in
-  let (evd', s) = new_univ_variable univ_rigid evd in
-  (evd', { uj_val = mkType s; uj_type = mkType (Univ.super s) })
+  let (evd', s) = new_sort_variable univ_rigid evd in
+  (evd', { uj_val = mkSort s; uj_type = mkSort (Sorts.super s) })
 
 let subterm_source evk ?where (loc,k) =
   let evk = match k with
@@ -838,21 +758,21 @@ let subterm_source evk ?where (loc,k) =
    irrelevant positions, unify universes when flexible. *)
 let compare_cumulative_instances cv_pb variances u u' sigma =
   let open UnivProblem in
-  let cstrs = Univ.Constraint.empty in
+  let cstrs = Univ.Constraints.empty in
   let soft = Set.empty in
   let cstrs, soft = Array.fold_left3 (fun (cstrs, soft) v u u' ->
       let open Univ.Variance in
       match v with
       | Irrelevant -> cstrs, Set.add (UWeak (u,u')) soft
       | Covariant when cv_pb == Reduction.CUMUL ->
-        Univ.Constraint.add (u,Univ.Le,u') cstrs, soft
-      | Covariant | Invariant -> Univ.Constraint.add (u,Univ.Eq,u') cstrs, soft)
+        Univ.Constraints.add (u,Univ.Le,u') cstrs, soft
+      | Covariant | Invariant -> Univ.Constraints.add (u,Univ.Eq,u') cstrs, soft)
       (cstrs,soft) variances (Univ.Instance.to_array u) (Univ.Instance.to_array u')
   in
   match Evd.add_constraints sigma cstrs with
   | sigma ->
     Inl (Evd.add_universe_constraints sigma soft)
-  | exception Univ.UniverseInconsistency p -> Inr p
+  | exception UGraph.UniverseInconsistency p -> Inr p
 
 let compare_constructor_instances evd u u' =
   let open UnivProblem in
@@ -872,14 +792,25 @@ let eq_constr_univs_test ~evd ~extended_evd t u =
   let open Evd in
   let t = EConstr.Unsafe.to_constr t
   and u = EConstr.Unsafe.to_constr u in
-  let fold cstr sigma =
-    try Some (add_universe_constraints sigma cstr)
-    with Univ.UniverseInconsistency _ | UniversesDiffer -> None
+  let sigma = ref extended_evd in
+  let eq_universes _ u1 u2 =
+    let u1 = normalize_universe_instance !sigma u1 in
+    let u2 = normalize_universe_instance !sigma u2 in
+    UGraph.check_eq_instances (universes !sigma) u1 u2
   in
-  let ans =
-    UnivProblem.eq_constr_univs_infer_with
-      (fun t -> kind_of_term_upto evd t)
-      (fun u -> kind_of_term_upto extended_evd u)
-      (universes extended_evd) fold t u extended_evd
+  let eq_sorts s1 s2 =
+    if Sorts.equal s1 s2 then true
+    else
+      try sigma := add_universe_constraints !sigma UnivProblem.(Set.singleton (UEq (s1, s2))); true
+      with UGraph.UniverseInconsistency _ | UniversesDiffer -> false
   in
-  match ans with None -> false | Some _ -> true
+  let eq_existential eq e1 e2 =
+    let eq c1 c2 = eq 0 (EConstr.Unsafe.to_constr c1) (EConstr.Unsafe.to_constr c2) in
+    EConstr.eq_existential evd eq (EConstr.of_existential e1) (EConstr.of_existential e2)
+  in
+  let kind1 = kind_of_term_upto evd in
+  let kind2 = kind_of_term_upto extended_evd in
+  let rec eq_constr' nargs m n =
+    Constr.compare_head_gen_with kind1 kind2 eq_universes eq_sorts (eq_existential eq_constr') eq_constr' nargs m n
+  in
+  Constr.compare_head_gen_with kind1 kind2 eq_universes eq_sorts (eq_existential eq_constr') eq_constr' 0 t u
