@@ -14,10 +14,10 @@ open Names
 open Constr
 open Environ
 
-(** This file defines the pervasive unification state used everywhere in Coq
-    tactic engine. It is very low-level and most of the functions exported here
-    are irrelevant to the standard API user. Consider using {!Evarutil},
-    {!Sigma} or {!Proofview} instead.
+(** This file defines the pervasive unification state used everywhere
+    in Coq tactic engine. It is very low-level and most of the
+    functions exported here are irrelevant to the standard API user.
+    Consider using {!Evarutil} or {!Proofview} instead.
 
     A unification state (of type [evar_map]) is primarily a finite mapping
     from existential variables to records containing the type of the evar
@@ -52,6 +52,9 @@ sig
   val filter_array : t -> 'a array -> 'a array
   (** Filter an array. Sizes must coincide. *)
 
+  val filter_slist : t  -> 'a SList.t -> 'a SList.t
+  (** Filter a sparse list. Sizes must coincide. *)
+
   val extend : int -> t -> t
   (** [extend n f] extends [f] on the left with [n]-th times [true]. *)
 
@@ -95,38 +98,47 @@ type evar_body =
   | Evar_empty
   | Evar_defined of econstr
 
-type evar_info = {
-  evar_concl : econstr;
-  (** Type of the evar. *)
-  evar_hyps : named_context_val; (* TODO econstr? *)
-  (** Context of the evar. *)
-  evar_body : evar_body;
-  (** Optional content of the evar. *)
-  evar_filter : Filter.t;
-  (** Boolean mask over {!evar_hyps}. Should have the same length.
-      When filtered out, the corresponding variable is not allowed to occur
-      in the solution *)
-  evar_abstract_arguments : Abstraction.t;
-  (** Boolean information over {!evar_hyps}, telling if an hypothesis instance
-      can be imitated or should stay abstract in HO unification problems
-      and inversion (see [second_order_matching_with_args] for its use). *)
-  evar_source : Evar_kinds.t located;
-  (** Information about the evar. *)
-  evar_candidates : econstr list option;
-  (** List of possible solutions when known that it is a finite list *)
-}
+type evar_info
+
+(** {6 Projections from evar infos} *)
 
-val make_evar : named_context_val -> etypes -> evar_info
 val evar_concl : evar_info -> econstr
+(** Type of the evar. *)
+
 val evar_context : evar_info -> (econstr, etypes) Context.Named.pt
-val evar_filtered_context : evar_info -> (econstr, etypes) Context.Named.pt
+(** Context of the evar. *)
+
 val evar_hyps : evar_info -> named_context_val
-val evar_filtered_hyps : evar_info -> named_context_val
+(** Context of the evar. *)
+
 val evar_body : evar_info -> evar_body
-val evar_candidates : evar_info -> constr list option
+(** Optional content of the evar. *)
+
+val evar_candidates : evar_info -> econstr list option
+(** List of possible solutions when known that it is a finite list *)
+
+val evar_source : evar_info -> Evar_kinds.t located
+
 val evar_filter : evar_info -> Filter.t
+(** Boolean mask over {!evar_hyps}. Should have the same length.
+    When filtered out, the corresponding variable is not allowed to occur
+    in the solution *)
+
+val evar_abstract_arguments : evar_info -> Abstraction.t
+(** Boolean information over {!evar_hyps}, telling if an hypothesis instance
+    can be imitated or should stay abstract in HO unification problems
+    and inversion (see [second_order_matching_with_args] for its use). *)
+
+val evar_relevance : evar_info -> Sorts.relevance
+(** Relevance of the conclusion of the evar. *)
+
+(** {6 Derived projections} *)
+
+val evar_filtered_context : evar_info -> (econstr, etypes) Context.Named.pt
+val evar_filtered_hyps : evar_info -> named_context_val
 val evar_env : env -> evar_info -> env
 val evar_filtered_env : env -> evar_info -> env
+val evar_identity_subst : evar_info -> econstr SList.t
 
 val map_evar_body : (econstr -> econstr) -> evar_body -> evar_body
 val map_evar_info : (econstr -> econstr) -> evar_info -> evar_info
@@ -140,12 +152,18 @@ type evar_map
 val empty : evar_map
 (** The empty evar map. *)
 
-val from_env : env -> evar_map
+val from_env : ?binders:lident list -> env -> evar_map
 (** The empty evar map with given universe context, taking its initial
-    universes from env. *)
+    universes from env, possibly with initial universe binders. This
+    is the main entry point at the beginning of the process of
+    interpreting a declaration (e.g. before entering the
+    interpretation of a Theorem statement). *)
 
 val from_ctx : UState.t -> evar_map
-(** The empty evar map with given universe context *)
+(** The empty evar map with given universe context. This is the main
+    entry point when resuming from a already interpreted declaration
+    (e.g.  after having interpreted a Theorem statement and preparing
+    to open a goal). *)
 
 val is_empty : evar_map -> bool
 (** Whether an evarmap is empty. *)
@@ -154,9 +172,30 @@ val has_undefined : evar_map -> bool
 (** [has_undefined sigma] is [true] if and only if
     there are uninstantiated evars in [sigma]. *)
 
-val new_evar : evar_map ->
-  ?name:Id.t -> ?typeclass_candidate:bool -> evar_info -> evar_map * Evar.t
-(** Creates a fresh evar mapping to the given information. *)
+val has_given_up : evar_map -> bool
+(** [has_given_up sigma] is [true] if and only if
+    there are given up evars in [sigma]. *)
+
+val has_shelved : evar_map -> bool
+(** [has_shelved sigma] is [true] if and only if
+    there are shelved evars in [sigma]. *)
+
+val new_pure_evar :
+  ?src:Evar_kinds.t Loc.located -> ?filter:Filter.t ->
+  ?relevance:Sorts.relevance ->
+  ?abstract_arguments:Abstraction.t -> ?candidates:econstr list ->
+  ?name:Id.t ->
+  ?typeclass_candidate:bool ->
+  ?principal:bool ->
+  named_context_val -> evar_map -> etypes -> evar_map * Evar.t
+(** Low-level interface to create an evar.
+  @param src User-facing source for the evar
+  @param filter See {!Evd.Filter}, must be the same length as [named_context_val]
+  @param name A name for the evar
+  @param principal Whether the evar is the principal goal
+  @param named_context_val The context of the evar
+  @param types The type of conclusion of the evar
+*)
 
 val add : evar_map -> Evar.t -> evar_info -> evar_map
 (** [add sigma ev info] adds [ev] with evar info [info] in sigma.
@@ -172,6 +211,9 @@ val find_undefined : evar_map -> Evar.t -> evar_info
 val remove : evar_map -> Evar.t -> evar_map
 (** Remove an evar from an evar map. Use with caution. *)
 
+val undefine : evar_map -> Evar.t -> evar_map [@@ocaml.deprecated]
+(** Remove the body of an evar. Only there for backward compat, do not use. *)
+
 val mem : evar_map -> Evar.t -> bool
 (** Whether an evar is present in an evarmap. *)
 
@@ -216,7 +258,7 @@ val is_defined : evar_map -> Evar.t-> bool
 val is_undefined : evar_map -> Evar.t-> bool
 (** Whether an evar is not defined in an evarmap. *)
 
-val add_constraints : evar_map -> Univ.Constraint.t -> evar_map
+val add_constraints : evar_map -> Univ.Constraints.t -> evar_map
 (** Add universe constraints in an evar map. *)
 
 val undefined_map : evar_map -> evar_info Evar.Map.t
@@ -246,14 +288,17 @@ val existential_opt_value : evar_map -> econstr pexistential -> econstr option
 
 val existential_opt_value0 : evar_map -> existential -> constr option
 
-val evar_instance_array : (Constr.named_declaration -> 'a -> bool) -> evar_info ->
-  'a list -> (Id.t * 'a) list
+val evar_handler : evar_map -> constr evar_handler
 
-val instantiate_evar_array : evar_info -> econstr -> econstr list -> econstr
+val existential_expand_value0 : evar_map -> existential -> constr Constr.evar_expansion
 
-val evars_reset_evd  : ?with_conv_pbs:bool -> ?with_univs:bool ->
-  evar_map ->  evar_map -> evar_map
-(** spiwack: this function seems to somewhat break the abstraction. *)
+val expand_existential : evar_map -> econstr pexistential -> econstr list
+(** Returns the full evar instance with implicit default variables turned into
+    explicit [Var] nodes. *)
+
+val expand_existential0 : evar_map -> constr pexistential -> constr list
+
+val instantiate_evar_array : evar_map -> evar_info -> econstr -> econstr SList.t -> econstr
 
 (** {6 Misc} *)
 
@@ -263,8 +308,15 @@ val restrict : Evar.t-> Filter.t -> ?candidates:econstr list ->
     possibly limiting the instances to a set of candidates (candidates
     are filtered according to the filter) *)
 
-val is_restricted_evar : evar_map -> Evar.t -> Evar.t option
-(** Tell if an evar comes from restriction of another evar, and if yes, which *)
+val update_source : evar_map -> Evar.t -> Evar_kinds.t located -> evar_map
+(** To update the source a posteriori, e.g. when an evar type of
+    another evar has to refer to this other evar, with a mutual dependency *)
+
+val get_aliased_evars : evar_map -> Evar.t Evar.Map.t
+(** The map of aliased evars *)
+
+val is_aliased_evar : evar_map -> Evar.t -> Evar.t option
+(** Tell if an evar has been aliased to another evar, and if yes, which *)
 
 val set_typeclass_evars : evar_map -> Evar.Set.t -> evar_map
 (** Mark the given set of evars as available for resolution.
@@ -291,10 +343,6 @@ val downcast : Evar.t-> etypes -> evar_map -> evar_map
 (** Change the type of an undefined evar to a new type assumed to be a
     subtype of its current type; subtyping must be ensured by caller *)
 
-val evar_source : Evar.t -> evar_map -> Evar_kinds.t located
-(** Convenience function. Wrapper around {!find} to recover the source of an
-    evar in a given evar map. *)
-
 val evar_ident : Evar.t -> evar_map -> Id.t option
 
 val rename : Evar.t -> Id.t -> evar_map -> evar_map
@@ -330,59 +378,61 @@ val drop_side_effects : evar_map -> evar_map
 
 (** {5 Future goals} *)
 
-type goal_kind = ToShelve | ToGiveUp
-
-val declare_future_goal : ?tag:goal_kind -> Evar.t -> evar_map -> evar_map
+val declare_future_goal : Evar.t -> evar_map -> evar_map
 (** Adds an existential variable to the list of future goals. For
     internal uses only. *)
 
-val declare_principal_goal : ?tag:goal_kind -> Evar.t -> evar_map -> evar_map
+val declare_principal_goal : Evar.t -> evar_map -> evar_map
 (** Adds an existential variable to the list of future goals and make
     it principal. Only one existential variable can be made principal, an
     error is raised otherwise. For internal uses only. *)
 
-val future_goals : evar_map -> Evar.t list
-(** Retrieves the list of future goals. Used by the [refine] primitive
-    of the tactic engine. *)
+module FutureGoals : sig
 
-val principal_future_goal : evar_map -> Evar.t option
-(** Retrieves the name of the principal existential variable if there
-    is one. Used by the [refine] primitive of the tactic engine. *)
+  type t
 
-type future_goals
+  val comb : t -> Evar.t list
 
-val save_future_goals : evar_map -> future_goals
-(** Retrieves the list of future goals including the principal future
-    goal. Used by the [refine] primitive of the tactic engine. *)
+  val principal : t -> Evar.t option
+  (** if [Some e], [e] must be contained in [future_comb]. The evar [e] will
+      inherit properties (now: the name) of the evar which will be instantiated
+      with a term containing [e]. *)
 
-val reset_future_goals : evar_map -> evar_map
-(** Clears the list of future goals (as well as the principal future
-    goal). Used by the [refine] primitive of the tactic engine. *)
+  val map_filter : (Evar.t -> Evar.t option) -> t -> t
+  (** Applies a function on the future goals *)
 
-val restore_future_goals : evar_map -> future_goals -> evar_map
-(** Sets the future goals (including the principal future goal) to a
-    previous value. Intended to be used after a local list of future
-    goals has been consumed. Used by the [refine] primitive of the
-    tactic engine. *)
+  val filter : (Evar.t -> bool) -> t -> t
+  (** Applies a filter on the future goals *)
 
-val fold_future_goals : (evar_map -> Evar.t -> evar_map) -> evar_map -> future_goals -> evar_map
-(** Fold future goals *)
+end
 
-val map_filter_future_goals : (Evar.t -> Evar.t option) -> future_goals -> future_goals
-(** Applies a function on the future goals *)
+val push_future_goals : evar_map -> evar_map
 
-val filter_future_goals : (Evar.t -> bool) -> future_goals -> future_goals
-(** Applies a filter on the future goals *)
+val pop_future_goals : evar_map -> FutureGoals.t * evar_map
 
-val dispatch_future_goals : future_goals -> Evar.t list * Evar.t list * Evar.t list * Evar.t option
-(** Returns the future_goals dispatched into regular, shelved, given_up
-   goals; last argument is the goal tagged as principal if any *)
+val fold_future_goals : (evar_map -> Evar.t -> evar_map) -> evar_map -> evar_map
 
-val extract_given_up_future_goals : future_goals -> Evar.t list * Evar.t list
-(** An ad hoc variant for Proof.proof; not for general use *)
+val remove_future_goal : evar_map -> Evar.t -> evar_map
 
-val shelve_on_future_goals : Evar.t list -> future_goals -> future_goals
-(** Push goals on the shelve of future goals *)
+val pr_future_goals_stack : evar_map -> Pp.t
+
+val push_shelf : evar_map -> evar_map
+
+val pop_shelf : evar_map -> Evar.t list * evar_map
+
+val filter_shelf : (Evar.t -> bool) -> evar_map -> evar_map
+
+val give_up : Evar.t -> evar_map -> evar_map
+
+val shelve : evar_map -> Evar.t list -> evar_map
+
+val unshelve : evar_map -> Evar.t list -> evar_map
+
+val given_up : evar_map -> Evar.Set.t
+
+val shelf : evar_map -> Evar.t list
+
+val pr_shelf : evar_map -> Pp.t
 
 (** {5 Sort variables}
 
@@ -419,13 +469,13 @@ type 'a sigma = {
   (** The base object. *)
   sigma : evar_map
   (** The added unification state. *)
-}
+} [@@ocaml.deprecated]
 (** The type constructor ['a sigma] adds an evar map to an object of type
     ['a]. *)
 
-val sig_it  : 'a sigma -> 'a
-val sig_sig : 'a sigma -> evar_map
-val on_sig : 'a sigma -> (evar_map -> evar_map * 'b) -> 'a sigma * 'b
+val sig_it  : 'a sigma -> 'a  [@@ocaml.warning "-3"] [@@ocaml.deprecated]
+val sig_sig : 'a sigma -> evar_map [@@ocaml.warning "-3"] [@@ocaml.deprecated]
+val on_sig : 'a sigma -> (evar_map -> evar_map * 'b) -> 'a sigma * 'b [@@ocaml.warning "-3"] [@@ocaml.deprecated]
 
 (** {5 The state monad with state an evar map} *)
 
@@ -509,26 +559,18 @@ val loc_of_conv_pb : evar_map -> evar_constraint -> Loc.t option
 val evars_of_term : evar_map -> econstr -> Evar.Set.t
   (** including evars in instances of evars *)
 
-val evar_nodes_of_term : econstr -> Evar.Set.t
-  (** same as evars_of_term but also including defined evars.
-      For use in printing dependent evars *)
-
 val evars_of_named_context : evar_map -> (econstr, etypes) Context.Named.pt -> Evar.Set.t
 
 val evars_of_filtered_evar_info : evar_map -> evar_info -> Evar.Set.t
 
 (** Metas *)
-val meta_list : evar_map -> (metavariable * clbinding) list
-val meta_defined : evar_map -> metavariable -> bool
+val meta_list : evar_map -> clbinding Metamap.t
 
 val meta_value     : evar_map -> metavariable -> econstr
 (** [meta_fvalue] raises [Not_found] if meta not in map or [Anomaly] if
    meta has no value *)
 
-val meta_fvalue    : evar_map -> metavariable -> econstr freelisted * instance_status
 val meta_opt_fvalue : evar_map -> metavariable -> (econstr freelisted * instance_status) option
-val meta_type      : evar_map -> metavariable -> etypes
-val meta_type0 : evar_map -> metavariable -> types
 val meta_ftype     : evar_map -> metavariable -> etypes freelisted
 val meta_name      : evar_map -> metavariable -> Name.t
 val meta_declare   :
@@ -539,9 +581,8 @@ val meta_reassign  : metavariable -> econstr * instance_status -> evar_map -> ev
 val clear_metas : evar_map -> evar_map
 
 (** [meta_merge evd1 evd2] returns [evd2] extended with the metas of [evd1] *)
-val meta_merge : ?with_univs:bool -> evar_map -> evar_map -> evar_map
+val meta_merge : clbinding Metamap.t -> evar_map -> evar_map
 
-val undefined_metas : evar_map -> metavariable list
 val map_metas_fvalue : (econstr -> econstr) -> evar_map -> evar_map
 val map_metas : (econstr -> econstr) -> evar_map -> evar_map
 
@@ -576,7 +617,7 @@ val univ_flexible_alg : rigid
 
 type 'a in_evar_universe_context = 'a * UState.t
 
-val restrict_universe_context : evar_map -> Univ.LSet.t -> evar_map
+val restrict_universe_context : evar_map -> Univ.Level.Set.t -> evar_map
 
 (** Raises Not_found if not a name for a universe in this map. *)
 val universe_of_name : evar_map -> Id.t -> Univ.Level.t
@@ -584,7 +625,6 @@ val universe_of_name : evar_map -> Id.t -> Univ.Level.t
 val universe_binders : evar_map -> UnivNames.universe_binders
 
 val new_univ_level_variable : ?loc:Loc.t -> ?name:Id.t -> rigid -> evar_map -> evar_map * Univ.Level.t
-val new_univ_variable : ?loc:Loc.t -> ?name:Id.t -> rigid -> evar_map -> evar_map * Univ.Universe.t
 val new_sort_variable : ?loc:Loc.t -> ?name:Id.t -> rigid -> evar_map -> evar_map * Sorts.t
 
 val add_global_univ : evar_map -> Univ.Level.t -> evar_map
@@ -602,8 +642,6 @@ val is_sort_variable : evar_map -> Sorts.t -> Univ.Level.t option
 
 val is_flexible_level : evar_map -> Univ.Level.t -> bool
 
-(* val normalize_universe_level : evar_map -> Univ.Level.t -> Univ.Level.t *)
-val normalize_universe : evar_map -> Univ.Universe.t -> Univ.Universe.t
 val normalize_universe_instance : evar_map -> Univ.Instance.t -> Univ.Instance.t
 
 val set_leq_sort : env -> evar_map -> Sorts.t -> Sorts.t -> evar_map
@@ -613,10 +651,10 @@ val set_leq_level : evar_map -> Univ.Level.t -> Univ.Level.t -> evar_map
 val set_eq_instances : ?flex:bool ->
   evar_map -> Univ.Instance.t -> Univ.Instance.t -> evar_map
 
-val check_eq : evar_map -> Univ.Universe.t -> Univ.Universe.t -> bool
-val check_leq : evar_map -> Univ.Universe.t -> Univ.Universe.t -> bool
+val check_eq : evar_map -> Sorts.t -> Sorts.t -> bool
+val check_leq : evar_map -> Sorts.t -> Sorts.t -> bool
 
-val check_constraints : evar_map -> Univ.Constraint.t -> bool
+val check_constraints : evar_map -> Univ.Constraints.t -> bool
 
 val evar_universe_context : evar_map -> UState.t
 val universe_context_set : evar_map -> Univ.ContextSet.t
@@ -628,28 +666,26 @@ val universes : evar_map -> UGraph.t
     [Univ.ContextSet.to_context]. *)
 val to_universe_context : evar_map -> Univ.UContext.t
 
-val univ_entry : poly:bool -> evar_map -> Entries.universes_entry
+val univ_entry : poly:bool -> evar_map -> UState.named_universes_entry
 
-val check_univ_decl : poly:bool -> evar_map -> UState.universe_decl -> Entries.universes_entry
+val check_univ_decl : poly:bool -> evar_map -> UState.universe_decl -> UState.named_universes_entry
 
 val merge_universe_context : evar_map -> UState.t -> evar_map
 val set_universe_context : evar_map -> UState.t -> evar_map
 
 val merge_context_set : ?loc:Loc.t -> ?sideff:bool -> rigid -> evar_map -> Univ.ContextSet.t -> evar_map
-val merge_universe_subst : evar_map -> UnivSubst.universe_opt_subst -> evar_map
 
 val with_context_set : ?loc:Loc.t -> rigid -> evar_map -> 'a Univ.in_universe_context_set -> evar_map * 'a
 
-val nf_univ_variables : evar_map -> evar_map * Univ.universe_subst
+val nf_univ_variables : evar_map -> evar_map
 
 val fix_undefined_variables : evar_map -> evar_map
 
-val refresh_undefined_universes : evar_map -> evar_map * Univ.universe_level_subst
-
 (** Universe minimization *)
 val minimize_universes : evar_map -> evar_map
 
-val update_sigma_env : evar_map -> env -> evar_map
+(** Lift [UState.update_sigma_univs] *)
+val update_sigma_univs : UGraph.t -> evar_map -> evar_map
 
 (** Polymorphic universes *)
 
@@ -661,6 +697,8 @@ val fresh_inductive_instance : ?loc:Loc.t -> ?rigid:rigid
   -> env -> evar_map -> inductive -> evar_map * pinductive
 val fresh_constructor_instance : ?loc:Loc.t -> ?rigid:rigid
   -> env -> evar_map -> constructor -> evar_map * pconstructor
+val fresh_array_instance : ?loc:Loc.t -> ?rigid:rigid
+  -> env -> evar_map  -> evar_map * Univ.Instance.t
 
 val fresh_global : ?loc:Loc.t -> ?rigid:rigid -> ?names:Univ.Instance.t -> env ->
   evar_map -> GlobRef.t -> evar_map * econstr
@@ -714,6 +752,10 @@ module MiniEConstr : sig
 
   val whd_evar : evar_map -> t -> t
 
+  val mkLEvar : evar_map -> Evar.t * t list -> t
+
+  val replace_vars : evar_map -> (Id.t * t) list -> t -> t
+
   val of_kind : (t, t, ESorts.t, EInstance.t) Constr.kind_of_term -> t
 
   val of_constr : Constr.t -> t
@@ -733,8 +775,13 @@ module MiniEConstr : sig
     (Constr.t, Constr.types) Context.Named.Declaration.pt
   val unsafe_to_rel_decl : (t, t) Context.Rel.Declaration.pt ->
     (Constr.t, Constr.types) Context.Rel.Declaration.pt
+  val of_case_invert : constr pcase_invert -> econstr pcase_invert
+  val unsafe_to_case_invert : econstr pcase_invert -> constr pcase_invert
   val of_rel_decl : (Constr.t, Constr.types) Context.Rel.Declaration.pt ->
     (t, t) Context.Rel.Declaration.pt
   val to_rel_decl : evar_map -> (t, t) Context.Rel.Declaration.pt ->
     (Constr.t, Constr.types) Context.Rel.Declaration.pt
+
+  val of_named_context : (Constr.t, Constr.types) Context.Named.pt -> (t, t) Context.Named.pt
+  val of_rel_context : (Constr.t, Constr.types) Context.Rel.pt -> (t, t) Context.Rel.pt
 end
