@@ -94,8 +94,8 @@ let is_constructor id =
   with Not_found ->
     false
 
-let is_section_variable id =
-  try let _ = Global.lookup_named id in true
+let is_section_variable env id =
+  try let _ = Environ.lookup_named id env in true
   with Not_found -> false
 
 (**********************************************************************)
@@ -118,7 +118,7 @@ let head_name sigma c = (* Find the head constant of a constr if any *)
         Some (Nametab.basename_of_global (global_of_constr c))
     | Fix ((_,i),(lna,_,_)) | CoFix (i,(lna,_,_)) ->
         Some (match lna.(i).binder_name with Name id -> id | _ -> assert false)
-    | Sort _ | Rel _ | Meta _|Evar _|Case (_, _, _, _) | Int _ | Float _ -> None
+    | Sort _ | Rel _ | Meta _|Evar _|Case _ | Int _ | Float _ | Array _ -> None
   in
   hdrec c
 
@@ -163,9 +163,10 @@ let hdchar env sigma c =
         let id = match lna.(i).binder_name with Name id -> id | _ -> assert false in
         lowercase_first_char id
     | Evar _ (* We could do better... *)
-    | Meta _ | Case (_, _, _, _) -> "y"
+    | Meta _ | Case _ -> "y"
     | Int _ -> "i"
     | Float _ -> "f"
+    | Array _ -> "a"
   in
   hdrec 0 c
 
@@ -211,7 +212,8 @@ let it_mkLambda_or_LetIn_name env sigma b hyps =
 (* Fresh names *)
 
 (* Introduce a mode where auto-generated names are mangled
-   to test dependence of scripts on auto-generated names *)
+   to test dependence of scripts on auto-generated names.
+   We also supply a version which only adds a prefix. *)
 
 let get_mangle_names =
   Goptions.declare_bool_option_and_ref
@@ -219,30 +221,49 @@ let get_mangle_names =
     ~key:["Mangle";"Names"]
     ~value:false
 
+let get_mangle_names_light =
+  Goptions.declare_bool_option_and_ref
+    ~depr:false
+    ~key:["Mangle";"Names";"Light"]
+    ~value:false
+
 let mangle_names_prefix =
   Goptions.declare_interpreted_string_option_and_ref
     ~depr:false
     ~key:["Mangle";"Names";"Prefix"]
-    ~value:(Id.of_string "_0")
+    ~value:("_")
     (fun x ->
+      Id.to_string
       (try
          Id.of_string x
        with
        | CErrors.UserError _ ->
-          CErrors.user_err Pp.(str ("Not a valid identifier: \"" ^ x ^ "\"."))
-      ) |> forget_subscript
+         CErrors.user_err Pp.(str ("Not a valid identifier: \"" ^ x ^ "\"."))
+      )
     )
-    (fun x -> Id.to_string x)
+    (fun x -> x)
 
-let mangle_id id = if get_mangle_names () then mangle_names_prefix () else id
+(** The name "foo" becomes "_0" if we get_mangle_names and "_foo" if
+    get_mangle_names_light is also set. Otherwise it is left alone. *)
+
+let mangle_id id =
+  let prfx = mangle_names_prefix () in
+  if get_mangle_names () then
+    if get_mangle_names_light () then
+      Id.of_string (prfx ^ Id.to_string id)
+    else Id.of_string (prfx ^ "0")
+  else id
 
 (* Looks for next "good" name by lifting subscript *)
 
-let next_ident_away_from id bad =
-  let id = mangle_id id in
+let next_ident_away_from_post_mangling id bad =
   let rec name_rec id = if bad id then name_rec (increment_subscript id) else id in
   name_rec id
 
+let next_ident_away_from id bad =
+  let id = mangle_id id in
+  next_ident_away_from_post_mangling id bad
+
 (* Restart subscript from x0 if name starts with xN, or x00 if name
    starts with x0N, etc *)
 
@@ -272,8 +293,8 @@ let visible_ids sigma (nenv, c) =
       accu := (gseen, vseen, ids)
   | Rel p ->
     let (gseen, vseen, ids) = !accu in
-    if p > n && not (Int.Set.mem p vseen) then
-      let vseen = Int.Set.add p vseen in
+    if p > n && not (Int.Set.mem (p - n) vseen) then
+      let vseen = Int.Set.add (p - n) vseen in
       let name =
         try Some (List.nth nenv (p - n - 1))
         with Invalid_argument _ | Failure _ ->
@@ -287,9 +308,17 @@ let visible_ids sigma (nenv, c) =
       | _ -> ids
       in
       accu := (gseen, vseen, ids)
+  | Evar (_,args as ev) ->
+    (* Useful for at least debugger: do the same as in iter_with_binders *)
+    (* except that Not_found is not fatal *)
+    begin match Evd.expand_existential sigma ev with
+    | args -> List.iter (visible_ids n) args
+    | exception Not_found when !Flags.in_debugger ->
+      SList.Skip.iter (visible_ids n) args
+    end
   | _ -> EConstr.iter_with_binders sigma succ visible_ids n c
   in
-  let () = visible_ids 1 c in
+  let () = visible_ids 1 c in (* n = 1 to count the binder to rename *)
   let (_, _, ids) = !accu in
   ids
 
@@ -313,16 +342,16 @@ let next_name_away_in_cases_pattern sigma env_t na avoid =
    - but if the name to use is not fresh in the current context, the fresh
      name is taken by finding a free subscript starting from 0 *)
 
-let next_ident_away_in_goal id avoid =
+let next_ident_away_in_goal env id avoid =
   let id = if Id.Set.mem id avoid then restart_subscript id else id in
-  let bad id = Id.Set.mem id avoid || (is_global id && not (is_section_variable id)) in
+  let bad id = Id.Set.mem id avoid || (is_global id && not (is_section_variable env id)) in
   next_ident_away_from id bad
 
-let next_name_away_in_goal na avoid =
+let next_name_away_in_goal env na avoid =
   let id = match na with
     | Name id -> id
     | Anonymous -> default_non_dependent_ident in
-  next_ident_away_in_goal id avoid
+  next_ident_away_in_goal env id avoid
 
 (* 3- Looks for next fresh name outside a list that is moreover valid
    as a global identifier; the legacy algorithm is that if the name is
@@ -342,7 +371,7 @@ let next_global_ident_away id avoid =
 let next_ident_away id avoid =
   let id = mangle_id id in
   if Id.Set.mem id avoid then
-    next_ident_away_from (restart_subscript id) (fun id -> Id.Set.mem id avoid)
+    next_ident_away_from_post_mangling (restart_subscript id) (fun id -> Id.Set.mem id avoid)
   else id
 
 let next_name_away_with_default default na avoid =
@@ -362,19 +391,24 @@ let next_name_away_with_default_using_types default na avoid t =
 
 let next_name_away = next_name_away_with_default default_non_dependent_string
 
-let make_all_name_different env sigma =
-  (* FIXME: this is inefficient, but only used in printing *)
-  let avoid = ref (ids_of_named_context_val (named_context_val env)) in
-  let sign = named_context_val env in
-  let rels = rel_context env in
-  let env0 = reset_with_named_context sign env in
+let make_all_rel_context_name_different env sigma ctx =
+  let avoid = ref (Id.Set.union (Context.Rel.to_vars (Environ.rel_context env)) (ids_of_named_context_val (named_context_val env))) in
   Context.Rel.fold_outside
-    (fun decl newenv ->
+    (fun decl (newenv,ctx) ->
        let na = named_hd newenv sigma (RelDecl.get_type decl) (RelDecl.get_name decl) in
        let id = next_name_away na !avoid in
        avoid := Id.Set.add id !avoid;
-       push_rel (RelDecl.set_name (Name id) decl) newenv)
-    rels ~init:env0
+       let decl = RelDecl.set_name (Name id) decl in
+       push_rel decl newenv, decl :: ctx)
+    ctx ~init:(env,[])
+
+let make_all_name_different env sigma =
+  (* FIXME: this is inefficient, but only used in printing *)
+  let sign = named_context_val env in
+  let rels = rel_context env in
+  let env0 = reset_with_named_context sign env in
+  let env,_ = make_all_rel_context_name_different env0 sigma rels in
+  env
 
 (* 5- Looks for next fresh name outside a list; avoids also to use names that
    would clash with short name of global references; if name is already used,
@@ -415,59 +449,53 @@ let next_name_away_for_default_printing sigma env_t na avoid =
 *)
 
 type renaming_flags =
+  (* The term is the body of a binder and the environment excludes this binder *)
+  (* so, there is a missing binder in the environment *)
   | RenamingForCasesPattern of (Name.t list * constr)
   | RenamingForGoal
   | RenamingElsewhereFor of (Name.t list * constr)
 
-let next_name_for_display sigma flags =
+let next_name_for_display env sigma flags =
   match flags with
   | RenamingForCasesPattern env_t -> next_name_away_in_cases_pattern sigma env_t
-  | RenamingForGoal -> next_name_away_in_goal
+  | RenamingForGoal -> next_name_away_in_goal env
   | RenamingElsewhereFor env_t -> next_name_away_for_default_printing sigma env_t
 
 (* Remark: Anonymous var may be dependent in Evar's contexts *)
-let compute_displayed_name_in_gen_poly noccurn_fun sigma flags avoid na c =
+let compute_displayed_name_in_gen_poly noccurn_fun env sigma flags avoid na c =
   match na with
   | Anonymous when noccurn_fun sigma 1 c ->
     (Anonymous,avoid)
   | _ ->
-    let fresh_id = next_name_for_display sigma flags na avoid in
+    let fresh_id = next_name_for_display env sigma flags na avoid in
     let idopt = if noccurn_fun sigma 1 c then Anonymous else Name fresh_id in
     (idopt, Id.Set.add fresh_id avoid)
 
 let compute_displayed_name_in = compute_displayed_name_in_gen_poly noccurn
 
-let compute_displayed_name_in_gen f sigma =
+let compute_displayed_name_in_gen f env sigma =
   (* only flag which does not need a constr, maybe to be refined *)
   let flag = RenamingForGoal in
-  compute_displayed_name_in_gen_poly f sigma flag
+  compute_displayed_name_in_gen_poly f env sigma flag
 
-let compute_and_force_displayed_name_in sigma flags avoid na c =
-  match na with
-  | Anonymous when noccurn sigma 1 c ->
-    (Anonymous,avoid)
-  | _ ->
-    let fresh_id = next_name_for_display sigma flags na avoid in
-    (Name fresh_id, Id.Set.add fresh_id avoid)
-
-let compute_displayed_let_name_in sigma flags avoid na c =
-  let fresh_id = next_name_for_display sigma flags na avoid in
+let compute_displayed_let_name_in env sigma flags avoid na =
+  let fresh_id = next_name_for_display env sigma flags na avoid in
   (Name fresh_id, Id.Set.add fresh_id avoid)
 
-let rename_bound_vars_as_displayed sigma avoid env c =
-  let rec rename avoid env c =
+let rename_bound_vars_as_displayed env sigma avoid tenv c =
+  let rec rename avoid tenv c =
     match EConstr.kind sigma c with
     | Prod (na,c1,c2)  ->
         let na',avoid' =
-          compute_displayed_name_in sigma
-            (RenamingElsewhereFor (env,c2)) avoid na.binder_name c2 in
-        mkProd ({na with binder_name=na'}, c1, rename avoid' (na' :: env) c2)
+          compute_displayed_name_in env sigma
+            (RenamingElsewhereFor (tenv,c2)) avoid na.binder_name c2 in
+        mkProd ({na with binder_name=na'}, c1, rename avoid' (na' :: tenv) c2)
     | LetIn (na,c1,t,c2) ->
         let na',avoid' =
-          compute_displayed_let_name_in sigma
-            (RenamingElsewhereFor (env,c2)) avoid na.binder_name c2 in
-        mkLetIn ({na with binder_name=na'},c1,t, rename avoid' (na' :: env) c2)
-    | Cast (c,k,t) -> mkCast (rename avoid env c, k,t)
+          compute_displayed_let_name_in env sigma
+            (RenamingElsewhereFor (tenv,c2)) avoid na.binder_name in
+        mkLetIn ({na with binder_name=na'},c1,t, rename avoid' (na' :: tenv) c2)
+    | Cast (c,k,t) -> mkCast (rename avoid tenv c, k,t)
     | _ -> c
   in
-  rename avoid env c
+  rename avoid tenv c
