@@ -34,19 +34,19 @@ let set_print_constr f = term_printer := f
 
 module EvMap = Evar.Map
 
-let evar_suggested_name evk sigma =
+let evar_suggested_name env sigma evk =
   let open Evd in
   let base_id evk' evi =
   match evar_ident evk' sigma with
   | Some id -> id
-  | None -> match evi.evar_source with
+  | None -> match Evd.evar_source evi with
   | _,Evar_kinds.ImplicitArg (c,(n,Some id),b) -> id
   | _,Evar_kinds.VarInstance id -> id
   | _,Evar_kinds.QuestionMark {Evar_kinds.qm_name = Name id} -> id
   | _,Evar_kinds.GoalEvar -> Id.of_string "Goal"
   | _ ->
-      let env = reset_with_named_context evi.evar_hyps (Global.env()) in
-      Namegen.id_of_name_using_hdchar env sigma evi.evar_concl Anonymous
+      let env = reset_with_named_context (Evd.evar_hyps evi) env in
+      Namegen.id_of_name_using_hdchar env sigma (Evd.evar_concl evi) Anonymous
   in
   let names = EvMap.mapi base_id (undefined_map sigma) in
   let id = EvMap.find evk names in
@@ -59,11 +59,11 @@ let evar_suggested_name evk sigma =
   let (_, n) = EvMap.fold fold names (false, 0) in
   if n = 0 then id else Nameops.add_suffix id (string_of_int (pred n))
 
-let pr_existential_key sigma evk =
+let pr_existential_key env sigma evk =
 let open Evd in
 match evar_ident evk sigma with
 | None ->
-  str "?" ++ Id.print (evar_suggested_name evk sigma)
+  str "?" ++ Id.print (evar_suggested_name env sigma evk)
 | Some id ->
   str "?" ++ Id.print id
 
@@ -105,7 +105,7 @@ let pr_meta_map env sigma =
            str " : " ++ print_constr env sigma t.rebus ++
            spc () ++ pr_instance_status s ++ fnl ())
   in
-  prlist pr_meta_binding (meta_list sigma)
+  prlist pr_meta_binding (Evd.Metamap.bindings (meta_list sigma))
 
 let pr_decl env sigma (decl,ok) =
   let open NamedDecl in
@@ -123,6 +123,13 @@ let pr_evar_source env sigma = function
       str "subterm of pattern-matching return predicate"
   | Evar_kinds.BinderType (Name id) -> str "type of " ++ Id.print id
   | Evar_kinds.BinderType Anonymous -> str "type of anonymous binder"
+  | Evar_kinds.EvarType (ido,evk) ->
+      let pp = match ido with
+        | Some id -> str "?" ++ Id.print id
+        | None ->
+          try pr_existential_key env sigma evk
+          with (* defined *) Not_found -> str "an internal placeholder" in
+     str "type of " ++ pp
   | Evar_kinds.ImplicitArg (c,(n,ido),b) ->
       let open Globnames in
       let print_constr = print_kconstr in
@@ -155,21 +162,21 @@ let pr_evar_info env sigma evi =
       in
       prlist_with_sep spc (pr_decl env sigma) (List.rev decls)
     with Invalid_argument _ -> str "Ill-formed filtered context" in
-  let pty = print_constr env sigma evi.evar_concl in
+  let pty = print_constr env sigma (Evd.evar_concl evi) in
   let pb =
-    match evi.evar_body with
+    match Evd.evar_body evi with
       | Evar_empty -> mt ()
       | Evar_defined c -> spc() ++ str"=> "  ++ print_constr env sigma c
   in
   let candidates =
-    match evi.evar_body, evi.evar_candidates with
+    match Evd.evar_body evi, Evd.evar_candidates evi with
       | Evar_empty, Some l ->
            spc () ++ str "{" ++
            prlist_with_sep (fun () -> str "|") (print_constr env sigma) l ++ str "}"
       | _ ->
           mt ()
   in
-  let src = str "(" ++ pr_evar_source env sigma (snd evi.evar_source) ++ str ")" in
+  let src = str "(" ++ pr_evar_source env sigma (snd (Evd.evar_source evi)) ++ str ")" in
   hov 2
     (str"["  ++ phyps ++ spc () ++ str"|- "  ++ pty ++ pb ++ str"]" ++
        candidates ++ spc() ++ src)
@@ -221,7 +228,7 @@ let evar_dependency_closure n sigma =
   EvMap.bindings map
 
 let has_no_evar sigma =
-  try let () = Evd.fold (fun _ _ () -> raise Exit) sigma () in true
+  try let () = Evd.fold (fun _ _ () -> raise_notrace Exit) sigma () in true
   with Exit -> false
 
 let pr_evd_level sigma = UState.pr_uctx_level (Evd.evar_universe_context sigma)
@@ -232,14 +239,14 @@ let pr_evar_universe_context ctx =
   let prl = pr_uctx_level ctx in
   if UState.is_empty ctx then mt ()
   else
-    (str"UNIVERSES:"++brk(0,1)++
-       h 0 (Univ.pr_universe_context_set prl (UState.context_set ctx)) ++ fnl () ++
+    v 0 (str"UNIVERSES:"++brk(0,1)++
+       h (Univ.pr_universe_context_set prl (UState.context_set ctx)) ++ fnl () ++
      str"ALGEBRAIC UNIVERSES:"++brk(0,1)++
-     h 0 (Univ.LSet.pr prl (UState.algebraics ctx)) ++ fnl() ++
+     h (Univ.Level.Set.pr prl (UState.algebraics ctx)) ++ fnl() ++
      str"UNDEFINED UNIVERSES:"++brk(0,1)++
-     h 0 (UnivSubst.pr_universe_opt_subst (UState.subst ctx)) ++ fnl() ++
+     h (UState.pr_universe_opt_subst (UState.subst ctx)) ++ fnl() ++
      str "WEAK CONSTRAINTS:"++brk(0,1)++
-     h 0 (UState.pr_weak prl ctx) ++ fnl ())
+     h (UState.pr_weak prl ctx) ++ fnl ())
 
 let print_env_short env sigma =
   let print_constr = print_kconstr in
@@ -298,38 +305,42 @@ let pr_evar_map_gen with_univs pr_evars env sigma =
       str "OBLIGATIONS:" ++ brk (0, 1) ++
       prlist_with_sep spc Evar.print (Evar.Set.elements evars) ++ fnl ()
   and metas =
-    if List.is_empty (Evd.meta_list sigma) then mt ()
+    if Evd.Metamap.is_empty (Evd.meta_list sigma) then mt ()
     else
       str "METAS:" ++ brk (0, 1) ++ pr_meta_map env sigma
+  and shelf =
+    str "SHELF:" ++ brk (0, 1) ++ Evd.pr_shelf sigma ++ fnl ()
+  and future_goals =
+    str "FUTURE GOALS STACK:" ++ brk (0, 1) ++ Evd.pr_future_goals_stack sigma ++ fnl ()
   in
-  evs ++ svs ++ cstrs ++ typeclasses ++ obligations ++ metas
+  evs ++ svs ++ cstrs ++ typeclasses ++ obligations ++ metas ++ shelf ++ future_goals
 
 let pr_evar_list env sigma l =
   let open Evd in
-  let pr_restrict ev =
-    match is_restricted_evar sigma ev with
+  let pr_alias ev =
+    match is_aliased_evar sigma ev with
     | None -> mt ()
-    | Some ev' -> str " (restricted to " ++ Evar.print ev' ++ str ")"
+    | Some ev' -> str " (aliased to " ++ Evar.print ev' ++ str ")"
   in
   let pr (ev, evi) =
-    h 0 (Evar.print ev ++
+    h (Evar.print ev ++
       str "==" ++ pr_evar_info env sigma evi ++
-      pr_restrict ev ++
-      (if evi.evar_body == Evar_empty
-       then str " {" ++ pr_existential_key sigma ev ++ str "}"
+      pr_alias ev ++
+      (if Evd.evar_body evi == Evar_empty
+       then str " {" ++ pr_existential_key env sigma ev ++ str "}"
        else mt ()))
   in
-  h 0 (prlist_with_sep fnl pr l)
+  hv 0 (prlist_with_sep fnl pr l)
 
 let to_list d =
   let open Evd in
   (* Workaround for change in Map.fold behavior in ocaml 3.08.4 *)
   let l = ref [] in
-  let fold_def evk evi () = match evi.evar_body with
+  let fold_def evk evi () = match Evd.evar_body evi with
     | Evar_defined _ -> l := (evk, evi) :: !l
     | Evar_empty -> ()
   in
-  let fold_undef evk evi () = match evi.evar_body with
+  let fold_undef evk evi () = match Evd.evar_body evi with
     | Evar_empty -> l := (evk, evi) :: !l
     | Evar_defined _ -> ()
   in
@@ -352,7 +363,7 @@ let pr_evar_by_filter filter env sigma =
   let open Evd in
   let elts = Evd.fold (fun evk evi accu -> (evk, evi) :: accu) sigma [] in
   let elts = List.rev elts in
-  let is_def (_, evi) = match evi.evar_body with
+  let is_def (_, evi) = match Evd.evar_body evi with
   | Evar_defined _ -> true
   | Evar_empty -> false
   in
@@ -480,28 +491,21 @@ let lookup_rel_id id sign =
   in
   lookrec 1 sign
 
-(* Constructs either [forall x:t, c] or [let x:=b:t in c] *)
 let mkProd_or_LetIn = EConstr.mkProd_or_LetIn
-(* Constructs either [forall x:t, c] or [c] in which [x] is replaced by [b] *)
-let mkProd_wo_LetIn decl c =
-  let open EConstr in
-  let open RelDecl in
-  match decl with
-    | LocalAssum (na,t) -> mkProd (na, t, c)
-    | LocalDef (_,b,_) -> Vars.subst1 b c
+let mkProd_wo_LetIn = EConstr.mkProd_wo_LetIn
 
-let it_mkProd init = List.fold_left (fun c (n,t)  -> EConstr.mkProd (n, t, c)) init
-let it_mkLambda init = List.fold_left (fun c (n,t)  -> EConstr.mkLambda (n, t, c)) init
+let it_mkProd = EConstr.it_mkProd
+let it_mkLambda = EConstr.it_mkLambda
 
-let it_named_context_quantifier f ~init =
-  List.fold_left (fun c d -> f d c) init
+let it_mkProd_or_LetIn = EConstr.it_mkProd_or_LetIn
+let it_mkProd_wo_LetIn = EConstr.it_mkProd_wo_LetIn
+let it_mkLambda_or_LetIn = Term.it_mkLambda_or_LetIn
+let it_mkNamedProd_or_LetIn = EConstr.it_mkNamedProd_or_LetIn
+let it_mkNamedLambda_or_LetIn = EConstr.it_mkNamedLambda_or_LetIn
 
-let it_mkProd_or_LetIn init = it_named_context_quantifier mkProd_or_LetIn ~init
-let it_mkProd_wo_LetIn init = it_named_context_quantifier mkProd_wo_LetIn ~init
-let it_mkLambda_or_LetIn init = it_named_context_quantifier mkLambda_or_LetIn ~init
-let it_mkNamedProd_or_LetIn init = it_named_context_quantifier EConstr.mkNamedProd_or_LetIn ~init
+(* On Constr *)
+let it_named_context_quantifier f ~init = List.fold_left (fun c d -> f d c) init
 let it_mkNamedProd_wo_LetIn init = it_named_context_quantifier mkNamedProd_wo_LetIn ~init
-let it_mkNamedLambda_or_LetIn init = it_named_context_quantifier EConstr.mkNamedLambda_or_LetIn ~init
 
 let it_mkLambda_or_LetIn_from_no_LetIn c decls =
   let open RelDecl in
@@ -595,7 +599,7 @@ let map_left2 f a g b =
     r, s
   end
 
-let map_constr_with_binders_left_to_right sigma g f l c =
+let map_constr_with_binders_left_to_right env sigma g f l c =
   let open RelDecl in
   let open EConstr in
   match EConstr.kind sigma c with
@@ -635,17 +639,23 @@ let map_constr_with_binders_left_to_right sigma g f l c =
     let b' = f l b in
       if b' == b then c
       else mkProj (p, b')
-  | Evar (e,al) ->
-    let al' = List.map_left (f l) al in
-      if List.for_all2 (==) al' al then c
-      else mkEvar (e, al')
-  | Case (ci,p,b,bl) ->
+  | Evar ev ->
+    let ev' = EConstr.map_existential sigma (fun c -> f l c) ev in
+    if ev' == ev then c else mkEvar ev'
+  | Case (ci,u,pms,p,iv,b,bl) ->
+      let (ci, _, pms, p0, _, b, bl0) = annotate_case env sigma (ci, u, pms, p, iv, b, bl) in
+      let f_ctx (nas, _ as r) (ctx, c) =
+        let c' = f (List.fold_right g ctx l) c in
+        if c' == c then r else (nas, c')
+      in
       (* In v8 concrete syntax, predicate is after the term to match! *)
       let b' = f l b in
-      let p' = f l p in
-      let bl' = Array.map_left (f l) bl in
-        if b' == b && p' == p && bl' == bl then c
-        else mkCase (ci, p', b', bl')
+      let pms' = Array.map_left (f l) pms in
+      let p' = f_ctx p p0 in
+      let iv' = map_invert (f l) iv in
+      let bl' = Array.map_left (fun (c, c0) -> f_ctx c c0) (Array.map2 (fun x y -> (x, y)) bl bl0) in
+        if b' == b && pms' == pms && p' == p && iv' == iv && bl' == bl then c
+        else mkCase (ci, u, pms', p', iv', b', bl')
   | Fix (ln,(lna,tl,bl as fx)) ->
       let l' = fold_rec_types g fx l in
       let (tl', bl') = map_left2 (f l) tl (f l') bl in
@@ -658,26 +668,15 @@ let map_constr_with_binders_left_to_right sigma g f l c =
         if Array.for_all2 (==) tl tl' && Array.for_all2 (==) bl bl'
         then c
         else mkCoFix (ln,(lna,tl',bl'))
-
-let map_under_context_with_full_binders sigma g f l n d =
-  let open EConstr in
-  let f l c = Unsafe.to_constr (f l (of_constr c)) in
-  let g d l = g (of_rel_decl d) l in
-  let d = EConstr.Unsafe.to_constr (EConstr.whd_evar sigma d) in
-  EConstr.of_constr (Constr.map_under_context_with_full_binders g f l n d)
-
-let map_branches_with_full_binders sigma g f l ci bl =
-  let tags = Array.map List.length ci.ci_pp_info.cstr_tags in
-  let bl' = Array.map2 (map_under_context_with_full_binders sigma g f l) tags bl in
-  if Array.for_all2 (==) bl' bl then bl else bl'
-
-let map_return_predicate_with_full_binders sigma g f l ci p =
-  let n = List.length ci.ci_pp_info.ind_tags in
-  let p' = map_under_context_with_full_binders sigma g f l n p in
-  if p' == p then p else p'
+  | Array(u,t,def,ty) ->
+      let t' = Array.map_left (f l) t in
+      let def' = f l def in
+      let ty' = f l ty in
+      if def' == def && t' == t && ty' == ty then c
+      else mkArray(u,t',def',ty')
 
 (* strong *)
-let map_constr_with_full_binders_gen userview sigma g f l cstr =
+let map_constr_with_full_binders env sigma g f l cstr =
   let open EConstr in
   match EConstr.kind sigma cstr with
   | (Rel _ | Meta _ | Var _   | Sort _ | Const _ | Ind _
@@ -706,21 +705,22 @@ let map_constr_with_full_binders_gen userview sigma g f l cstr =
   | Proj (p,c) ->
       let c' = f l c in
         if c' == c then cstr else mkProj (p, c')
-  | Evar (e,al) ->
-      let al' = List.map (f l) al in
-      if List.for_all2 (==) al al' then cstr else mkEvar (e, al')
-  | Case (ci,p,c,bl) when userview ->
-      let p' = map_return_predicate_with_full_binders sigma g f l ci p in
+  | Evar ev ->
+    let ev' = EConstr.map_existential sigma (fun c -> f l c) ev in
+    if ev' == ev then cstr else mkEvar ev'
+  | Case (ci, u, pms, p, iv, c, bl) ->
+      let (ci, _, pms, p0, _, c, bl0) = annotate_case env sigma (ci, u, pms, p, iv, c, bl) in
+      let f_ctx (nas, _ as r) (ctx, c) =
+        let c' = f (List.fold_right g ctx l) c in
+        if c' == c then r else (nas, c')
+      in
+      let pms' = Array.Smart.map (f l) pms in
+      let p' = f_ctx p p0 in
+      let iv' = map_invert (f l) iv in
       let c' = f l c in
-      let bl' = map_branches_with_full_binders sigma g f l ci bl in
-      if p==p' && c==c' && bl'==bl then cstr else
-        mkCase (ci, p', c', bl')
-  | Case (ci,p,c,bl) ->
-      let p' = f l p in
-      let c' = f l c in
-      let bl' = Array.map (f l) bl in
-      if p==p' && c==c' && Array.for_all2 (==) bl bl' then cstr else
-        mkCase (ci, p', c', bl')
+      let bl' = Array.map2 f_ctx bl bl0 in
+      if pms==pms' && p==p' && iv'==iv && c==c' && Array.for_all2 (==) bl bl' then cstr else
+        mkCase (ci, u, pms', p', iv', c', bl')
   | Fix (ln,(lna,tl,bl as fx)) ->
       let tl' = Array.map (f l) tl in
       let l' = fold_rec_types g fx l in
@@ -735,12 +735,11 @@ let map_constr_with_full_binders_gen userview sigma g f l cstr =
       if Array.for_all2 (==) tl tl' && Array.for_all2 (==) bl bl'
       then cstr
       else mkCoFix (ln,(lna,tl',bl'))
-
-let map_constr_with_full_binders sigma g f =
-  map_constr_with_full_binders_gen false sigma g f
-
-let map_constr_with_full_binders_user_view sigma g f =
-  map_constr_with_full_binders_gen true sigma g f
+  | Array(u,t,def,ty) ->
+      let t' = Array.Smart.map (f l) t in
+      let def' = f l def in
+      let ty' = f l ty in
+      if def==def' && t == t' && ty==ty' then cstr else mkArray (u,t', def',ty')
 
 (* [fold_constr_with_binders g f n acc c] folds [f n] on the immediate
    subterms of [c] starting from [acc] and proceeding from left to
@@ -749,18 +748,33 @@ let map_constr_with_full_binders_user_view sigma g f =
    index) which is processed by [g] (which typically add 1 to [n]) at
    each binder traversal; it is not recursive *)
 
-let fold_constr_with_full_binders sigma g f n acc c =
-  let open EConstr in
-  let f l acc c = f l acc (of_constr c) in
-  let g d l  = g (of_rel_decl d) l in
-  let c = Unsafe.to_constr (whd_evar sigma c) in
-  Constr.fold_with_full_binders g f n acc c
-
-let fold_constr_with_binders sigma g f n acc c =
-  let open EConstr in
-  let f l acc c = f l acc (of_constr c) in
-  let c = Unsafe.to_constr (whd_evar sigma c) in
-  Constr.fold_constr_with_binders g f n acc c
+let fold_constr_with_full_binders env sigma g f n acc c =
+  let open EConstr.Vars in
+  let open Context.Rel.Declaration in
+  match EConstr.kind sigma c with
+  | Rel _ | Meta _ | Var _   | Sort _ | Const _ | Ind _ | Construct _  | Int _ | Float _ -> acc
+  | Cast (c,_, t) -> f n (f n acc c) t
+  | Prod (na,t,c) -> f (g (LocalAssum (na,t)) n) (f n acc t) c
+  | Lambda (na,t,c) -> f (g (LocalAssum (na,t)) n) (f n acc t) c
+  | LetIn (na,b,t,c) -> f (g (LocalDef (na,b,t)) n) (f n (f n acc b) t) c
+  | App (c,l) -> Array.fold_left (f n) (f n acc c) l
+  | Proj (_,c) -> f n acc c
+  | Evar ev ->
+    let args = Evd.expand_existential sigma ev in
+    List.fold_left (fun c -> f n c) acc args
+  | Case (ci, u, pms, p, iv, c, bl) ->
+    let (ci, _, pms, p, _, c, bl) = EConstr.annotate_case env sigma (ci, u, pms, p, iv, c, bl) in
+    let f_ctx acc (ctx, c) = f (List.fold_right g ctx n) acc c in
+    Array.fold_left f_ctx (f n (fold_invert (f n) (f_ctx (Array.fold_left (f n) acc pms) p) iv) c) bl
+  | Fix (_,(lna,tl,bl)) ->
+      let n' = CArray.fold_left2_i (fun i c n t -> g (LocalAssum (n,lift i t)) c) n lna tl in
+      let fd = Array.map2 (fun t b -> (t,b)) tl bl in
+      Array.fold_left (fun acc (t,b) -> f n' (f n acc t) b) acc fd
+  | CoFix (_,(lna,tl,bl)) ->
+      let n' = CArray.fold_left2_i (fun i c n t -> g (LocalAssum (n,lift i t)) c) n lna tl in
+      let fd = Array.map2 (fun t b -> (t,b)) tl bl in
+      Array.fold_left (fun acc (t,b) -> f n' (f n acc t) b) acc fd
+  | Array(_u,t,def,ty) -> f n (f n (Array.fold_left (f n) acc t) def) ty
 
 (***************************)
 (* occurs check functions  *)
@@ -771,6 +785,7 @@ exception Occur
 let occur_meta sigma c =
   let rec occrec c = match EConstr.kind sigma c with
     | Meta _ -> raise Occur
+    | Evar (_, args) -> SList.Skip.iter occrec args
     | _ -> EConstr.iter sigma occrec c
   in try occrec c; false with Occur -> true
 
@@ -790,13 +805,16 @@ let occur_meta_or_existential sigma c =
 let occur_metavariable sigma m c =
   let rec occrec c = match EConstr.kind sigma c with
   | Meta m' -> if Int.equal m m' then raise Occur
+  | Evar (_, args) -> SList.Skip.iter occrec args
   | _ -> EConstr.iter sigma occrec c
   in
   try occrec c; false with Occur -> true
 
 let occur_evar sigma n c =
   let rec occur_rec c = match EConstr.kind sigma c with
-    | Evar (sp,_) when Evar.equal sp n -> raise Occur
+    | Evar (sp, args) ->
+      if Evar.equal sp n then raise Occur
+      else SList.Skip.iter occur_rec args
     | _ -> EConstr.iter sigma occur_rec c
   in
   try occur_rec c; false with Occur -> true
@@ -813,6 +831,16 @@ let occur_var env sigma id c =
   in
   try occur_rec c; false with Occur -> true
 
+let occur_vars env sigma ids c =
+  let rec occur_rec c =
+    match EConstr.destRef sigma c with
+    | gr, _ ->
+      let vars = vars_of_global env gr in
+      if not (Id.Set.is_empty (Id.Set.inter ids vars)) then raise Occur
+    | exception DestKO -> EConstr.iter sigma occur_rec c
+  in
+  try occur_rec c; false with Occur -> true
+
 exception OccurInGlobal of GlobRef.t
 
 let occur_var_indirectly env sigma id c =
@@ -827,6 +855,9 @@ let occur_var_indirectly env sigma id c =
 let occur_var_in_decl env sigma hyp decl =
   NamedDecl.exists (occur_var env sigma hyp) decl
 
+let occur_vars_in_decl env sigma hyps decl =
+  NamedDecl.exists (occur_vars env sigma hyps) decl
+
 let local_occur_var sigma id c =
   let rec occur c = match EConstr.kind sigma c with
   | Var id' -> if Id.equal id id' then raise Occur
@@ -842,10 +873,39 @@ let local_occur_var_in_decl sigma hyp decl =
 let free_rels sigma m =
   let rec frec depth acc c = match EConstr.kind sigma c with
     | Rel n       -> if n >= depth then Int.Set.add (n-depth+1) acc else acc
-    | _ -> fold_constr_with_binders sigma succ frec depth acc c
+    | Evar (_, args) -> SList.Skip.fold (fun acc c -> frec depth acc c) acc args
+    | _ -> EConstr.fold_with_binders sigma succ frec depth acc c
   in
   frec 1 Int.Set.empty m
 
+let free_rels_and_unqualified_refs sigma t =
+  let rec aux k (gseen, vseen, ids as accu) t =
+    match EConstr.kind sigma t with
+    | Const _ | Ind _ | Construct _ | Var _ ->
+      let g, _ = EConstr.destRef sigma t in
+      if not (GlobRef.Set_env.mem g gseen) then begin
+        try
+          let gseen = GlobRef.Set_env.add g gseen in
+          let short = Nametab.shortest_qualid_of_global Id.Set.empty g in
+          let dir, id = Libnames.repr_qualid short in
+          let ids = if DirPath.is_empty dir then Id.Set.add id ids else ids in
+          (gseen, vseen, ids)
+        with Not_found when !Flags.in_debugger || !Flags.in_toplevel ->
+          accu
+      end else
+        accu
+    | Rel p ->
+      if p > k && not (Int.Set.mem (p - k) vseen) then
+        let vseen = Int.Set.add (p - k) vseen in
+        (gseen, vseen, ids)
+      else
+        accu
+    | _ ->
+      EConstr.fold_with_binders sigma succ aux k accu t in
+  let accu = (GlobRef.Set_env.empty, Int.Set.empty, Id.Set.empty) in
+  let (_, rels, ids) = aux 0 accu t in
+  rels, ids
+
 (* collects all metavar occurrences, in left-to-right order, preserving
  * repetitions and all. *)
 
@@ -853,6 +913,7 @@ let collect_metas sigma c =
   let rec collrec acc c =
     match EConstr.kind sigma c with
       | Meta mv -> List.add_set Int.equal mv acc
+      | Evar (_, args) -> SList.Skip.fold collrec acc args
       | _       -> EConstr.fold sigma collrec acc c
   in
   List.rev (collrec [] c)
@@ -938,84 +999,52 @@ let rec strip_outer_cast sigma c = match EConstr.kind sigma c with
   | Cast (c,_,_) -> strip_outer_cast sigma c
   | _ -> c
 
-(* flattens application lists throwing casts in-between *)
-let collapse_appl sigma c = match EConstr.kind sigma c with
-  | App (f,cl) ->
-    if EConstr.isCast sigma f then
-      let rec collapse_rec f cl2 =
-        match EConstr.kind sigma (strip_outer_cast sigma f) with
-        | App (g,cl1) -> collapse_rec g (Array.append cl1 cl2)
-        | _ -> EConstr.mkApp (f,cl2)
-      in
-      collapse_rec f cl
-    else c
-  | _ -> c
-
 (* First utilities for avoiding telescope computation for subst_term *)
 
-let prefix_application sigma eq_fun (k,c) t =
+let prefix_application sigma eq_fun k l1 t =
   let open EConstr in
-  let c' = collapse_appl sigma c and t' = collapse_appl sigma t in
-  match EConstr.kind sigma c', EConstr.kind sigma t' with
-    | App (f1,cl1), App (f2,cl2) ->
-        let l1 = Array.length cl1
-        and l2 = Array.length cl2 in
+  if 0 < l1 then match EConstr.kind sigma t with
+    | App (f2,cl2) ->
+        let l2 = Array.length cl2 in
         if l1 <= l2
-           && eq_fun sigma c' (mkApp (f2, Array.sub cl2 0 l1)) then
-          Some (mkApp (mkRel k, Array.sub cl2 l1 (l2 - l1)))
+           && eq_fun sigma k (mkApp (f2, Array.sub cl2 0 l1)) then
+          Some (Array.sub cl2 l1 (l2 - l1))
         else
           None
     | _ -> None
+  else None
 
-let my_prefix_application sigma eq_fun (k,c) by_c t =
-  let open EConstr in
-  let c' = collapse_appl sigma c and t' = collapse_appl sigma t in
-  match EConstr.kind sigma c', EConstr.kind sigma t' with
-    | App (f1,cl1), App (f2,cl2) ->
-        let l1 = Array.length cl1
-        and l2 = Array.length cl2 in
-        if l1 <= l2
-           && eq_fun sigma c' (mkApp (f2, Array.sub cl2 0 l1)) then
-          Some (mkApp ((Vars.lift k by_c), Array.sub cl2 l1 (l2 - l1)))
-        else
-          None
-    | _ -> None
-
-(* Recognizing occurrences of a given subterm in a term: [subst_term c t]
-   substitutes [(Rel 1)] for all occurrences of term [c] in a term [t];
-   works if [c] has rels *)
-
-let subst_term_gen sigma eq_fun c t =
-  let open EConstr in
-  let open Vars in
-  let rec substrec (k,c as kc) t =
-    match prefix_application sigma eq_fun kc t with
-      | Some x -> x
-      | None ->
-    if eq_fun sigma c t then mkRel k
-    else
-      EConstr.map_with_binders sigma (fun (k,c) -> (k+1,lift 1 c)) substrec kc t
+let eq_upto_lift cache c sigma k t =
+  let c =
+    try Int.Map.find k !cache
+    with Not_found ->
+      let c = EConstr.Vars.lift k c in
+      let () = cache := Int.Map.add k c !cache in
+      c
   in
-  substrec (1,c) t
-
-let subst_term sigma c t = subst_term_gen sigma EConstr.eq_constr c t
+  EConstr.eq_constr sigma c t
 
 (* Recognizing occurrences of a given subterm in a term :
    [replace_term c1 c2 t] substitutes [c2] for all occurrences of
    term [c1] in a term [t]; works if [c1] and [c2] have rels *)
 
-let replace_term_gen sigma eq_fun c by_c in_t =
-  let rec substrec (k,c as kc) t =
-    match my_prefix_application sigma eq_fun kc by_c t with
-      | Some x -> x
+let replace_term_gen sigma eq_fun ar by_c in_t =
+  let rec substrec k t =
+    match prefix_application sigma eq_fun k ar t with
+      | Some args -> EConstr.mkApp (EConstr.Vars.lift k by_c, args)
       | None ->
-    (if eq_fun sigma c t then (EConstr.Vars.lift k by_c) else
-      EConstr.map_with_binders sigma (fun (k,c) -> (k+1,EConstr.Vars.lift 1 c))
-        substrec kc t)
+    (if eq_fun sigma k t then (EConstr.Vars.lift k by_c) else
+      EConstr.map_with_binders sigma succ substrec k t)
   in
-  substrec (0,c) in_t
+  substrec 0 in_t
 
-let replace_term sigma c byc t = replace_term_gen sigma EConstr.eq_constr c byc t
+let replace_term sigma c byc t =
+  let cache = ref Int.Map.empty in
+  let ar = Array.length (snd (decompose_app_vect sigma c)) in
+  let eq sigma k t = eq_upto_lift cache c sigma k t in
+  replace_term_gen sigma eq ar byc t
+
+let subst_term sigma c t = replace_term sigma c (EConstr.mkRel 1) t
 
 let vars_of_env env =
   let s = Environ.ids_of_named_context_val (Environ.named_context_val env) in
@@ -1062,8 +1091,8 @@ let ids_of_context env =
 let names_of_rel_context env =
   List.map RelDecl.get_name (rel_context env)
 
-let is_section_variable id =
-  try let _ = Global.lookup_named id in true
+let is_section_variable env id =
+  try let _ = Environ.lookup_named id env in true
   with Not_found -> false
 
 let global_of_constr sigma c =
@@ -1127,9 +1156,9 @@ let compare_constr_univ sigma f cv_pb t1 t2 =
       Sort s1, Sort s2 -> base_sort_cmp cv_pb (ESorts.kind sigma s1) (ESorts.kind sigma s2)
     | Prod (_,t1,c1), Prod (_,t2,c2) ->
         f Reduction.CONV t1 t2 && f cv_pb c1 c2
-    | Const (c, u), Const (c', u') -> Constant.equal c c'
-    | Ind (i, _), Ind (i', _) -> eq_ind i i'
-    | Construct (i, _), Construct (i', _) -> eq_constructor i i'
+    | Const (c, u), Const (c', u') -> Constant.CanOrd.equal c c'
+    | Ind (i, _), Ind (i', _) -> Ind.CanOrd.equal i i'
+    | Construct (i, _), Construct (i', _) -> Construct.CanOrd.equal i i'
     | _ -> EConstr.compare_constr sigma (fun t1 t2 -> f Reduction.CONV t1 t2) t1 t2
 
 let constr_cmp sigma cv_pb t1 t2 =
@@ -1138,64 +1167,6 @@ let constr_cmp sigma cv_pb t1 t2 =
 
 let eq_constr sigma t1 t2 = constr_cmp sigma Reduction.CONV t1 t2
 
-(* App(c,[t1,...tn]) -> ([c,t1,...,tn-1],tn)
-   App(c,[||]) -> ([],c) *)
-let split_app sigma c = match EConstr.kind sigma c with
-    App(c,l) ->
-      let len = Array.length l in
-      if Int.equal len 0 then ([],c) else
-        let last = Array.get l (len-1) in
-        let prev = Array.sub l 0 (len-1) in
-        c::(Array.to_list prev), last
-  | _ -> assert false
-
-type subst = (EConstr.rel_context * EConstr.constr) Evar.Map.t
-
-exception CannotFilter
-
-let filtering sigma env cv_pb c1 c2 =
-  let open EConstr in
-  let open Vars in
-  let evm = ref Evar.Map.empty in
-  let define cv_pb e1 ev c1 =
-    try let (e2,c2) = Evar.Map.find ev !evm in
-    let shift = List.length e1 - List.length e2 in
-    if constr_cmp sigma cv_pb c1 (lift shift c2) then () else raise CannotFilter
-    with Not_found ->
-      evm := Evar.Map.add ev (e1,c1) !evm
-  in
-  let rec aux env cv_pb c1 c2 =
-    match EConstr.kind sigma c1, EConstr.kind sigma c2 with
-      | App _, App _ ->
-        let ((p1,l1),(p2,l2)) = (split_app sigma c1),(split_app sigma c2) in
-        let () = aux env cv_pb l1 l2 in
-        begin match p1, p2 with
-        | [], [] -> ()
-        | (h1 :: p1), (h2 :: p2) ->
-          aux env cv_pb (applist (h1, p1)) (applist (h2, p2))
-        | _ -> assert false
-        end
-      | Prod (n,t1,c1), Prod (_,t2,c2) ->
-          aux env cv_pb t1 t2;
-          aux (RelDecl.LocalAssum (n,t1) :: env) cv_pb c1 c2
-      | _, Evar (ev,_) -> define cv_pb env ev c1
-      | Evar (ev,_), _ -> define cv_pb env ev c2
-      | _ ->
-          if compare_constr_univ sigma
-          (fun pb c1 c2 -> aux env pb c1 c2; true) cv_pb c1 c2 then ()
-          else raise CannotFilter
-          (* TODO: le reste des binders *)
-  in
-  aux env cv_pb c1 c2; !evm
-
-let decompose_prod_letin sigma c =
-  let rec prodec_rec i l c = match EConstr.kind sigma c with
-    | Prod (n,t,c)    -> prodec_rec (succ i) (RelDecl.LocalAssum (n,t)::l) c
-    | LetIn (n,d,t,c) -> prodec_rec (succ i) (RelDecl.LocalDef (n,d,t)::l) c
-    | Cast (c,_,_)    -> prodec_rec i l c
-    | _               -> i,l,c in
-  prodec_rec 0 [] c
-
 (* (nb_lam [na1:T1]...[nan:Tan]c) where c is not an abstraction
  * gives n (casts are ignored) *)
 let nb_lam sigma c =
@@ -1224,13 +1195,6 @@ let nb_prod_modulo_zeta sigma x =
       | _ -> n
   in count 0 x
 
-let align_prod_letin sigma c a =
-  let (lc,_,_) = decompose_prod_letin sigma c in
-  let (la,l,a) = decompose_prod_letin sigma a in
-  if not (la >= lc) then invalid_arg "align_prod_letin";
-  let (l1,l2) = Util.List.chop lc l in
-  l2,it_mkProd_or_LetIn a l1
-
 (* We reduce a series of head eta-redex or nothing at all   *)
 (* [x1:c1;...;xn:cn]@(f;a1...an;x1;...;xn) --> @(f;a1...an) *)
 (* Remplace 2 earlier buggish versions                      *)
@@ -1283,27 +1247,10 @@ let map_rel_context_in_env f env sign =
   in
   aux env [] (List.rev sign)
 
-let map_rel_context_with_binders f sign =
-  let rec aux k = function
-    | d::sign -> RelDecl.map_constr (f k) d :: aux (k-1) sign
-    | [] -> []
-  in
-  aux (Context.Rel.length sign) sign
-
-let substl_rel_context l =
-  map_rel_context_with_binders (fun k -> substnl l (k-1))
-
-let lift_rel_context n =
-  map_rel_context_with_binders (liftn n)
-
-let smash_rel_context sign =
-  let rec aux acc = function
-  | [] -> acc
-  | (RelDecl.LocalAssum _ as d) :: l -> aux (d::acc) l
-  | RelDecl.LocalDef (_,b,_) :: l ->
-      (* Quadratic in the number of let but there are probably a few of them *)
-      aux (List.rev (substl_rel_context [b] (List.rev acc))) l
-  in List.rev (aux [] sign)
+let map_rel_context_with_binders = Context.Rel.map_with_binders
+let substl_rel_context = Vars.substl_rel_context
+let lift_rel_context = Vars.lift_rel_context
+let smash_rel_context = Vars.smash_rel_context
 
 let fold_named_context_both_sides f l ~init = List.fold_right_and_left f l init
 
@@ -1355,8 +1302,6 @@ let global_vars_set env sigma constr =
   in
   filtrec Id.Set.empty constr
 
-let global_vars env sigma ids = Id.Set.elements (global_vars_set env sigma ids)
-
 let global_vars_set_of_decl env sigma = function
   | NamedDecl.LocalAssum (_,t) -> global_vars_set env sigma t
   | NamedDecl.LocalDef (_,c,t) ->
