@@ -31,17 +31,38 @@ val lookup_rel_id : Id.t -> ('c, 't) Context.Rel.pt -> int * 'c option * 't
 val rel_vect : int -> int -> Constr.constr array
 val rel_list : int -> int -> constr list
 
-(** iterators/destructors on terms *)
+(** Prod/Lambda/LetIn destructors on econstr *)
+
 val mkProd_or_LetIn : rel_declaration -> types -> types
+  [@@ocaml.deprecated "Use synonymous [EConstr.mkProd_or_LetIn]."]
+
 val mkProd_wo_LetIn : rel_declaration -> types -> types
+  [@@ocaml.deprecated "Use synonymous [EConstr.mkProd_wo_LetIn]."]
+
 val it_mkProd : types -> (Name.t Context.binder_annot * types) list -> types
+  [@@ocaml.deprecated "Use synonymous [EConstr.it_mkProd]."]
+
 val it_mkLambda : constr -> (Name.t Context.binder_annot * types) list -> constr
+  [@@ocaml.deprecated "Use synonymous [EConstr.it_mkLambda]."]
+
 val it_mkProd_or_LetIn : types -> rel_context -> types
+  [@@ocaml.deprecated "Use synonymous [EConstr.it_mkProd_or_LetIn]."]
+
 val it_mkProd_wo_LetIn : types -> rel_context -> types
+  [@@ocaml.deprecated "Use synonymous [EConstr.it_mkProd_wo_LetIn]."]
+
 val it_mkLambda_or_LetIn : Constr.constr -> Constr.rel_context -> Constr.constr
-val it_mkNamedProd_or_LetIn : types -> named_context -> types
+  [@@ocaml.deprecated "Use synonymous [Term.it_mkLambda_or_LetIn]."]
+
+val it_mkNamedProd_or_LetIn : Evd.evar_map -> types -> named_context -> types
+  [@@ocaml.deprecated "Use synonymous [EConstr.it_mkNamedProd_or_LetIn]."]
+
+val it_mkNamedLambda_or_LetIn : Evd.evar_map -> constr -> named_context -> constr
+  [@@ocaml.deprecated "Use synonymous [EConstr.it_mkNamedLambda_or_LetIn]."]
+
+(** Prod/Lambda/LetIn destructors on constr *)
+
 val it_mkNamedProd_wo_LetIn : Constr.types -> Constr.named_context -> Constr.types
-val it_mkNamedLambda_or_LetIn : constr -> named_context -> constr
 
 (* Ad hoc version reinserting letin, assuming the body is defined in
    the context where the letins are expanded *)
@@ -50,30 +71,16 @@ val it_mkLambda_or_LetIn_from_no_LetIn : Constr.constr -> Constr.rel_context ->
 (** {6 Generic iterators on constr} *)
 
 val map_constr_with_binders_left_to_right :
-  Evd.evar_map ->
+  Environ.env -> Evd.evar_map ->
   (rel_declaration -> 'a -> 'a) ->
   ('a -> constr -> constr) ->
     'a -> constr -> constr
 val map_constr_with_full_binders :
-  Evd.evar_map ->
-  (rel_declaration -> 'a -> 'a) ->
-  ('a -> constr -> constr) -> 'a -> constr -> constr
-val map_constr_with_full_binders_user_view :
-  Evd.evar_map ->
+  Environ.env -> Evd.evar_map ->
   (rel_declaration -> 'a -> 'a) ->
   ('a -> constr -> constr) -> 'a -> constr -> constr
 
-(** [fold_constr_with_binders g f n acc c] folds [f n] on the immediate
-   subterms of [c] starting from [acc] and proceeding from left to
-   right according to the usual representation of the constructions as
-   [fold_constr] but it carries an extra data [n] (typically a lift
-   index) which is processed by [g] (which typically add 1 to [n]) at
-   each binder traversal; it is not recursive *)
-
-val fold_constr_with_binders : Evd.evar_map ->
-  ('a -> 'a) -> ('a -> 'b -> constr -> 'b) -> 'a -> 'b -> constr -> 'b
-
-val fold_constr_with_full_binders : Evd.evar_map ->
+val fold_constr_with_full_binders : Environ.env -> Evd.evar_map ->
   (rel_declaration -> 'a -> 'a) ->
   ('a -> 'b -> constr -> 'b) ->
   'a -> 'b -> constr -> 'b
@@ -93,9 +100,9 @@ val occur_metavariable : Evd.evar_map -> metavariable -> constr -> bool
 val occur_evar : Evd.evar_map -> Evar.t -> constr -> bool
 val occur_var : env -> Evd.evar_map -> Id.t -> constr -> bool
 val occur_var_indirectly : env -> Evd.evar_map -> Id.t -> constr -> GlobRef.t option
-val occur_var_in_decl :
-  env -> Evd.evar_map ->
-  Id.t -> named_declaration -> bool
+val occur_var_in_decl : env -> Evd.evar_map -> Id.t -> named_declaration -> bool
+val occur_vars : env -> Evd.evar_map -> Id.Set.t -> constr -> bool
+val occur_vars_in_decl : env -> Evd.evar_map -> Id.Set.t -> named_declaration -> bool
 
 (** As {!occur_var} but assume the identifier not to be a section variable *)
 val local_occur_var : Evd.evar_map -> Id.t -> constr -> bool
@@ -103,6 +110,10 @@ val local_occur_var_in_decl : Evd.evar_map -> Id.t -> named_declaration -> bool
 
 val free_rels : Evd.evar_map -> constr -> Int.Set.t
 
+(* Return the list of unbound rels and unqualified reference (same
+   strategy as in Namegen) *)
+val free_rels_and_unqualified_refs : Evd.evar_map -> constr -> Int.Set.t * Id.Set.t
+
 (** [dependent m t] tests whether [m] is a subterm of [t] *)
 val dependent : Evd.evar_map -> constr -> constr -> bool
 val dependent_no_evar : Evd.evar_map -> constr -> constr -> bool
@@ -126,16 +137,12 @@ val pop : constr -> constr
 (** Substitution of an arbitrary large term. Uses equality modulo
    reduction of let *)
 
-(** [subst_term_gen eq d c] replaces [d] by [Rel 1] in [c] using [eq]
-   as equality *)
-val subst_term_gen : Evd.evar_map ->
-  (Evd.evar_map -> constr -> constr -> bool) -> constr -> constr -> constr
-
-(** [replace_term_gen eq d e c] replaces [d] by [e] in [c] using [eq]
-   as equality *)
+(** [replace_term_gen eq arity e c] replaces matching subterms according to
+    [eq] by [e] in [c]. If [arity] is non-zero applications of larger length
+    are handled atomically. *)
 val replace_term_gen :
-  Evd.evar_map -> (Evd.evar_map -> constr -> constr -> bool) ->
-    constr -> constr -> constr -> constr
+  Evd.evar_map -> (Evd.evar_map -> int -> constr -> bool) ->
+    int -> constr -> constr -> constr
 
 (** [subst_term d c] replaces [d] by [Rel 1] in [c] *)
 val subst_term : Evd.evar_map -> constr -> constr -> constr
@@ -152,9 +159,6 @@ val eq_constr : Evd.evar_map -> constr -> constr -> bool (* FIXME rename: erases
 
 val eta_reduce_head : Evd.evar_map -> constr -> constr
 
-(** Flattens application lists *)
-val collapse_appl : Evd.evar_map -> constr -> constr
-
 (** [prod_applist] [forall (x1:B1;...;xn:Bn), B] [a1...an] @return [B[a1...an]] *)
 val prod_applist : Evd.evar_map -> constr -> constr list -> constr
 
@@ -171,20 +175,6 @@ val prod_applist_assum : Evd.evar_map -> int -> constr -> constr list -> constr
    [strip_outer_cast (Cast (Cast ... (Cast c, t) ... ))] is [c]. *)
 val strip_outer_cast : Evd.evar_map -> constr -> constr
 
-exception CannotFilter
-
-(** Lightweight first-order filtering procedure. Unification
-   variables ar represented by (untyped) Evars.
-   [filtering c1 c2] returns the substitution n'th evar ->
-   (context,term), or raises [CannotFilter].
-   Warning: Outer-kernel sort subtyping are taken into account: c1 has
-   to be smaller than c2 wrt. sorts. *)
-type subst = (rel_context * constr) Evar.Map.t
-val filtering : Evd.evar_map -> rel_context -> Reduction.conv_pb -> constr -> constr -> subst
-
-val decompose_prod_letin : Evd.evar_map -> constr -> int * rel_context * constr
-val align_prod_letin : Evd.evar_map -> constr -> constr -> rel_context * constr
-
 (** [nb_lam] {% $ %}[x_1:T_1]...[x_n:T_n]c{% $ %} where {% $ %}c{% $ %} is not an abstraction
    gives {% $ %}n{% $ %} (casts are ignored) *)
 val nb_lam : Evd.evar_map -> constr -> int
@@ -221,6 +211,7 @@ val names_of_rel_context : env -> names_context
    [n] hypotheses, excluding local definitions, and [Γ₁], if not empty,
    starts with an hypothesis (i.e. [Γ₁] has the form empty or [x:A;Γ₁'] *)
 val context_chop : int -> Constr.rel_context -> Constr.rel_context * Constr.rel_context
+  [@@ocaml.deprecated "Use synonymous [Context.Rel.chop_nhyps]."]
 
 (* [env_rel_context_chop n env] extracts out the [n] top declarations
    of the rel_context part of [env], counting both local definitions and
@@ -234,14 +225,19 @@ val add_vname : Id.Set.t -> Name.t -> Id.Set.t
 (** other signature iterators *)
 val process_rel_context : (rel_declaration -> env -> env) -> env -> env
 val assums_of_rel_context : ('c, 't) Context.Rel.pt -> (Name.t Context.binder_annot * 't) list
+
 val lift_rel_context : int -> Constr.rel_context -> Constr.rel_context
+  [@@ocaml.deprecated "Use synonymous [Vars.lift_rel_context]."]
 val substl_rel_context : Constr.constr list -> Constr.rel_context -> Constr.rel_context
-val smash_rel_context : Constr.rel_context -> Constr.rel_context (** expand lets in context *)
+  [@@ocaml.deprecated "Use synonymous [Vars.substl_rel_context]."]
+val smash_rel_context : Constr.rel_context -> Constr.rel_context
+  [@@ocaml.deprecated "Use synonymous [Vars.smash_rel_context]."]
+val map_rel_context_with_binders :
+  (int -> 'c -> 'c) -> ('c, 'c) Context.Rel.pt -> ('c, 'c) Context.Rel.pt
+  [@@ocaml.deprecated "Use synonymous [Context.Rel.map_with_binders]."]
 
 val map_rel_context_in_env :
   (env -> Constr.constr -> Constr.constr) -> env -> Constr.rel_context -> Constr.rel_context
-val map_rel_context_with_binders :
-  (int -> 'c -> 'c) -> ('c, 'c) Context.Rel.pt -> ('c, 'c) Context.Rel.pt
 val fold_named_context_both_sides :
   ('a -> Constr.named_declaration -> Constr.named_declaration list -> 'a) ->
     Constr.named_context -> init:'a -> 'a
@@ -253,7 +249,6 @@ val map_named_decl : ('a -> 'b) -> ('a, 'a) Context.Named.Declaration.pt -> ('b,
 
 val clear_named_body : Id.t -> env -> env
 
-val global_vars : env -> Evd.evar_map -> constr -> Id.t list
 val global_vars_set : env -> Evd.evar_map -> constr -> Id.Set.t
 val global_vars_set_of_decl : env -> Evd.evar_map -> named_declaration -> Id.Set.t
 val global_app_of_constr : Evd.evar_map -> constr -> (GlobRef.t * EInstance.t) * constr option
@@ -263,7 +258,7 @@ val global_app_of_constr : Evd.evar_map -> constr -> (GlobRef.t * EInstance.t) *
 val dependency_closure : env -> Evd.evar_map -> named_context -> Id.Set.t -> Id.t list
 
 (** Test if an identifier is the basename of a global reference *)
-val is_section_variable : Id.t -> bool
+val is_section_variable : env -> Id.t -> bool
 
 val global_of_constr : Evd.evar_map -> constr -> GlobRef.t * EInstance.t
 [@@ocaml.deprecated "Use [EConstr.destRef] instead (throws DestKO instead of Not_found)."]
@@ -286,9 +281,9 @@ val reference_of_level : Evd.evar_map -> Univ.Level.t -> Libnames.qualid option
 
 open Evd
 
-val pr_existential_key : evar_map -> Evar.t -> Pp.t
+val pr_existential_key : env -> evar_map -> Evar.t -> Pp.t
 
-val evar_suggested_name : Evar.t -> evar_map -> Id.t
+val evar_suggested_name : env -> evar_map -> Evar.t -> Id.t
 
 val pr_evar_info : env -> evar_map -> evar_info -> Pp.t
 val pr_evar_constraints : evar_map -> evar_constraint list -> Pp.t
