@@ -24,14 +24,20 @@ type mayrec
 module type S = sig
   type te
   type 'c pattern
+  type ty_pattern = TPattern : 'a pattern -> ty_pattern
 
   module Parsable : sig
     type t
-    val make : ?loc:Loc.t -> char Stream.t -> t
-    val comments : t -> ((int * int) * string) list
-  end
+    val make : ?loc:Loc.t -> ?fix_loc:(Loc.t -> Loc.t) -> char Stream.t -> t
+    (** [make ?loc ?fix_loc stream] builds a parsable starting at pos
+        [loc], and will apply [fix_loc] to each token. The use of the
+        [fix_loc] parameter is internal and will be removed in the
+        future as it is reserved to the XML protocol. *)
 
-  val tokens : string -> (string option * int) list
+    val comments : t -> ((int * int) * string) list
+    val loc : t -> Loc.t
+    val consume : t -> int -> unit
+  end
 
   module Entry : sig
     type 'a t
@@ -39,9 +45,14 @@ module type S = sig
     val create : string -> 'a t (* compat *)
     val parse : 'a t -> Parsable.t -> 'a
     val name : 'a t -> string
-    val of_parser : string -> (Plexing.location_function -> te Stream.t -> 'a) -> 'a t
-    val parse_token_stream : 'a t -> te Stream.t -> 'a
+    type 'a parser_fun = { parser_fun : te LStream.t -> 'a }
+    val of_parser : string -> 'a parser_fun -> 'a t
+    val parse_token_stream : 'a t -> te LStream.t -> 'a
     val print : Format.formatter -> 'a t -> unit
+    val is_empty : 'a t -> bool
+
+    type any_t = Any : 'a t -> any_t
+    val accumulate_in : 'a t -> any_t list CString.Map.t
   end
 
   module rec Symbol : sig
@@ -51,16 +62,17 @@ module type S = sig
     val nterml : 'a Entry.t -> string -> ('self, norec, 'a) t
     val list0 : ('self, 'trec, 'a) t -> ('self, 'trec, 'a list) t
     val list0sep :
-      ('self, 'trec, 'a) t -> ('self, norec, 'b) t -> bool ->
+      ('self, 'trec, 'a) t -> ('self, norec, unit) t -> bool ->
       ('self, 'trec, 'a list) t
     val list1 : ('self, 'trec, 'a) t -> ('self, 'trec, 'a list) t
     val list1sep :
-      ('self, 'trec, 'a) t -> ('self, norec, 'b) t -> bool ->
+      ('self, 'trec, 'a) t -> ('self, norec, unit) t -> bool ->
       ('self, 'trec, 'a list) t
     val opt : ('self, 'trec, 'a) t -> ('self, 'trec, 'a option) t
     val self : ('self, mayrec, 'self) t
     val next : ('self, mayrec, 'self) t
     val token : 'c pattern -> ('self, norec, 'c) t
+    val tokens : ty_pattern list -> ('self, norec, unit) t
     val rules : 'a Rules.t list -> ('self, norec, 'a) t
 
   end and Rule : sig
@@ -91,14 +103,13 @@ module type S = sig
     string option * Gramext.g_assoc option * 'a Production.t list
 
   type 'a extend_statement =
-    { pos : Gramext.position option
-    ; data : 'a single_extend_statement list
-    }
+  | Reuse of string option * 'a Production.t list
+    (** Extend an existing level by its optional given name. If None, picks the topmost level. *)
+  | Fresh of Gramext.position * 'a single_extend_statement list
+    (** Create a level at the given position. *)
 
   val generalize_symbol : ('a, 'tr, 'c) Symbol.t -> ('a, norec, 'c) Symbol.t option
 
-  val mk_rule : 'a pattern list -> string Rules.t
-
   (* Used in custom entries, should tweak? *)
   val level_of_nonterm : ('a, norec, 'c) Symbol.t -> string option
 
