@@ -17,7 +17,6 @@ open Glob_term
 open Pattern
 open Constrexpr
 open Notation_term
-open Notation
 open Ltac_pretype
 
 (** Translation of pattern, cases pattern, glob_constr and term into syntax
@@ -53,8 +52,6 @@ val extern_rel_context : constr option -> env -> Evd.evar_map ->
 (** Printing options *)
 val print_implicits : bool ref
 val print_implicits_defensive : bool ref
-val print_arguments : bool ref
-val print_evar_arguments : bool ref
 val print_coercions : bool ref
 val print_parentheses : bool ref
 val print_universes : bool ref
@@ -71,11 +68,11 @@ val get_extern_reference :
 (** WARNING: The following functions are evil due to
     side-effects. Think of the following case as used in the printer:
 
-    without_specific_symbols [SynDefRule kn] (pr_glob_constr_env env) c
+    without_specific_symbols [AbbrevRule kn] (pr_glob_constr_env env) c
 
     vs
 
-    without_specific_symbols [SynDefRule kn] pr_glob_constr_env env c
+    without_specific_symbols [AbbrevRule kn] pr_glob_constr_env env c
 
     which one is wrong? We should turn this kind of state into an
     explicit argument.
@@ -87,23 +84,8 @@ val with_universes : ('a -> 'b) -> 'a -> 'b
 (** This suppresses printing of primitive tokens and notations *)
 val without_symbols : ('a -> 'b) -> 'a -> 'b
 
-(** This suppresses printing of specific notations only *)
-val without_specific_symbols : interp_rule list -> ('a -> 'b) -> 'a -> 'b
-
 (** This prints metas as anonymous holes *)
 val with_meta_as_hole : ('a -> 'b) -> 'a -> 'b
 
-(** Fine-grained activation and deactivation of notation printing.
- *)
-val toggle_scope_printing :
-  scope:Notation_term.scope_name -> activate:bool -> unit
-
-val toggle_notation_printing
-  : ?scope:Notation_term.scope_name
-  -> notation:Constrexpr.notation
-  -> activate:bool
-  -> unit
-  -> unit
-
 (** Probably shouldn't be used *)
 val empty_extern_env : extern_env
