@@ -13,18 +13,18 @@ open Libnames
 open Constrexpr
 open Glob_term
 open Notation_term
+open Notationextern
 
 (** Notations *)
 
+val notation_cat : Libobject.category
+
 val pr_notation : notation -> Pp.t
 (** Printing *)
 
 val notation_entry_eq : notation_entry -> notation_entry -> bool
 (** Equality on [notation_entry]. *)
 
-val notation_entry_level_eq : notation_entry_level -> notation_entry_level -> bool
-(** Equality on [notation_entry_level]. *)
-
 val notation_with_optional_scope_eq : notation_with_optional_scope -> notation_with_optional_scope -> bool
 
 val notation_eq : notation -> notation -> bool
@@ -56,9 +56,11 @@ val scope_is_open_in_scopes : scope_name -> scopes -> bool
 val scope_is_open : scope_name -> bool
 
 (** Open scope *)
+val open_scope : scope_name -> unit
+val close_scope : scope_name -> unit
 
-val open_close_scope :
-  (* locality *) bool * (* open *) bool * scope_name -> unit
+(** Return a scope taking either a scope name or delimiter *)
+val normalize_scope : string -> scope_name
 
 (** Extend a list of scopes *)
 val empty_scope_stack : scopes
@@ -74,7 +76,7 @@ val find_delimiters_scope : ?loc:Loc.t -> delimiters -> scope_name
 
 (** {6 Declare and uses back and forth an interpretation of primitive token } *)
 
-(** A numeral interpreter is the pair of an interpreter for **(hexa)decimal**
+(** A number interpreter is the pair of an interpreter for **(hexa)decimal**
    numbers in terms and an optional interpreter in pattern, if
    non integer or negative numbers are not supported, the interpreter
    must fail with an appropriate error message *)
@@ -84,7 +86,7 @@ type required_module = full_path * string list
 type rawnum = NumTok.Signed.t
 
 (** The unique id string below will be used to refer to a particular
-    registered interpreter/uninterpreter of numeral or string notation.
+    registered interpreter/uninterpreter of number or string notation.
     Using the same uid for different (un)interpreters will fail.
     If at most one interpretation of prim token is used per scope,
     then the scope name could be used as unique id. *)
@@ -101,12 +103,12 @@ val register_rawnumeral_interpretation :
   ?allow_overwrite:bool -> prim_token_uid -> rawnum prim_token_interpretation -> unit
 
 val register_bignumeral_interpretation :
-  ?allow_overwrite:bool -> prim_token_uid -> Bigint.bigint prim_token_interpretation -> unit
+  ?allow_overwrite:bool -> prim_token_uid -> Z.t prim_token_interpretation -> unit
 
 val register_string_interpretation :
   ?allow_overwrite:bool -> prim_token_uid -> string prim_token_interpretation -> unit
 
-(** * Numeral notation *)
+(** * Number notation *)
 
 type prim_token_notation_error =
   | UnexpectedTerm of Constr.t
@@ -131,21 +133,22 @@ type z_pos_ty =
   { z_ty : Names.inductive;
     pos_ty : Names.inductive }
 
-type numeral_ty =
+type number_ty =
   { int : int_ty;
     decimal : Names.inductive;
     hexadecimal : Names.inductive;
-    numeral : Names.inductive }
+    number : Names.inductive }
+
+type pos_neg_int63_ty =
+  { pos_neg_int63_ty : Names.inductive }
 
 type target_kind =
-  | Int of int_ty (* Coq.Init.Numeral.int + uint *)
-  | UInt of int_ty (* Coq.Init.Numeral.uint *)
+  | Int of int_ty (* Coq.Init.Number.int + uint *)
+  | UInt of int_ty (* Coq.Init.Number.uint *)
   | Z of z_pos_ty (* Coq.Numbers.BinNums.Z and positive *)
-  | Int63 (* Coq.Numbers.Cyclic.Int63.Int63.int *)
-  | Numeral of numeral_ty (* Coq.Init.Numeral.numeral + uint + int *)
-  | DecimalInt of int_ty (* Coq.Init.Decimal.int + uint (deprecated) *)
-  | DecimalUInt of int_ty (* Coq.Init.Decimal.uint (deprecated) *)
-  | Decimal of numeral_ty (* Coq.Init.Decimal.Decimal + uint + int (deprecated) *)
+  | Int63 of pos_neg_int63_ty (* Coq.Numbers.Cyclic.Int63.PrimInt63.pos_neg_int63 *)
+  | Float64 (* Coq.Floats.PrimFloat.float *)
+  | Number of number_ty (* Coq.Init.Number.number + uint + int *)
 
 type string_target_kind =
   | ListByte
@@ -154,26 +157,43 @@ type string_target_kind =
 type option_kind = Option | Direct
 type 'target conversion_kind = 'target * option_kind
 
+(** A postprocessing translation [to_post] can be done after execution
+   of the [to_ty] interpreter. The reverse translation is performed
+   before the [of_ty] uninterpreter.
+
+   [to_post] is an array of [n] lists [l_i] of tuples [(f, t,
+   args)]. When the head symbol of the translated term matches one of
+   the [f] in the list [l_0] it is replaced by [t] and its arguments
+   are translated acording to [args] where [ToPostCopy] means that the
+   argument is kept unchanged and [ToPostAs k] means that the
+   argument is recursively translated according to [l_k].
+   [ToPostHole] introduces an additional implicit argument hole
+   (in the reverse translation, the corresponding argument is removed).
+   [ToPostCheck r] behaves as [ToPostCopy] except in the reverse
+   translation which fails if the copied term is not [r].
+   When [n] is null, no translation is performed. *)
+type to_post_arg = ToPostCopy | ToPostAs of int | ToPostHole | ToPostCheck of Constr.t
 type ('target, 'warning) prim_token_notation_obj =
   { to_kind : 'target conversion_kind;
     to_ty : GlobRef.t;
+    to_post : ((GlobRef.t * GlobRef.t * to_post_arg list) list) array;
     of_kind : 'target conversion_kind;
     of_ty : GlobRef.t;
     ty_name : Libnames.qualid; (* for warnings / error messages *)
     warning : 'warning }
 
-type numeral_notation_obj = (target_kind, numnot_option) prim_token_notation_obj
+type number_notation_obj = (target_kind, numnot_option) prim_token_notation_obj
 type string_notation_obj = (string_target_kind, unit) prim_token_notation_obj
 
 type prim_token_interp_info =
     Uid of prim_token_uid
-  | NumeralNotation of numeral_notation_obj
+  | NumberNotation of number_notation_obj
   | StringNotation of string_notation_obj
 
 type prim_token_infos = {
   pt_local : bool; (** Is this interpretation local? *)
   pt_scope : scope_name; (** Concerned scope *)
-  pt_interp_info : prim_token_interp_info; (** Unique id "pointing" to (un)interp functions, OR a numeral notation object describing (un)interp functions *)
+  pt_interp_info : prim_token_interp_info; (** Unique id "pointing" to (un)interp functions, OR a number notation object describing (un)interp functions *)
   pt_required : required_module; (** Module that should be loaded first *)
   pt_refs : GlobRef.t list; (** Entry points during uninterpretation *)
   pt_in_match : bool (** Is this prim token legal in match patterns ? *)
@@ -196,8 +216,8 @@ val enable_prim_token_interpretation : prim_token_infos -> unit
 *)
 
 val declare_numeral_interpreter : ?local:bool -> scope_name -> required_module ->
-  Bigint.bigint prim_token_interpreter ->
-  glob_constr list * Bigint.bigint prim_token_uninterpreter * bool -> unit
+  Z.t prim_token_interpreter ->
+  glob_constr list * Z.t prim_token_uninterpreter * bool -> unit
 val declare_string_interpreter : ?local:bool -> scope_name -> required_module ->
   string prim_token_interpreter ->
   glob_constr list * string prim_token_uninterpreter * bool -> unit
@@ -224,54 +244,56 @@ val availability_of_prim_token :
 
 (** {6 Declare and interpret back and forth a notation } *)
 
-(** Binds a notation in a given scope to an interpretation *)
-type interp_rule =
-  | NotationRule of specific_notation
-  | SynDefRule of KerName.t
+type entry_coercion_kind =
+  | IsEntryCoercion of notation_entry_level
+  | IsEntryGlobal of string * int
+  | IsEntryIdent of string * int
 
-val declare_notation_interpretation : notation -> scope_name option ->
-  interpretation -> notation_location -> onlyprint:bool ->
+val declare_notation : notation_with_optional_scope * notation ->
+  interpretation -> notation_location -> use:notation_use ->
+  also_in_cases_pattern:bool ->
+  entry_coercion_kind option ->
   Deprecation.t option -> unit
 
-val declare_uninterpretation : interp_rule -> interpretation -> unit
 
 (** Return the interpretation bound to a notation *)
 val interp_notation : ?loc:Loc.t -> notation -> subscopes ->
       interpretation * (notation_location * scope_name option)
 
-type notation_applicative_status =
-  | AppBoundedNotation of int
-  | AppUnboundedNotation
-  | NotAppNotation
-
-type notation_rule = interp_rule * interpretation * notation_applicative_status
-
-(** Return the possible notations for a given term *)
-val uninterp_notations : 'a glob_constr_g -> notation_rule list
-val uninterp_cases_pattern_notations : 'a cases_pattern_g -> notation_rule list
-val uninterp_ind_pattern_notations : inductive -> notation_rule list
-
 (** Test if a notation is available in the scopes
    context [scopes]; if available, the result is not None; the first
    argument is itself not None if a delimiters is needed *)
 val availability_of_notation : specific_notation -> subscopes ->
   (scope_name option * delimiters option) option
 
+val is_printing_inactive_rule : Notationextern.interp_rule -> interpretation -> bool
+
+type 'a notation_query_pattern_gen = {
+    notation_entry_pattern : notation_entry list;
+    interp_rule_key_pattern : (notation_key, 'a) Util.union option;
+    use_pattern : notation_use;
+    scope_pattern : notation_with_optional_scope option;
+    interpretation_pattern : interpretation option;
+  }
+
+type notation_query_pattern = qualid notation_query_pattern_gen
+
+val toggle_notations : on:bool -> all:bool -> (glob_constr -> Pp.t) -> notation_query_pattern -> unit
+
 (** {6 Miscellaneous} *)
 
+(** Take a notation string and turn it into a notation key. eg. ["x +  y"] becomes ["_ + _"]. *)
+val interpret_notation_string : string -> string
+
 (** If head is true, also allows applied global references. *)
 val interp_notation_as_global_reference : ?loc:Loc.t -> head:bool -> (GlobRef.t -> bool) ->
       notation_key -> delimiters option -> GlobRef.t
 
-(** Checks for already existing notations *)
-val exists_notation_in_scope : scope_name option -> notation ->
-      bool -> interpretation -> bool
-
 (** Declares and looks for scopes associated to arguments of a global ref *)
 val declare_arguments_scope :
-  bool (** true=local *) -> GlobRef.t -> scope_name option list -> unit
+  bool (** true=local *) -> GlobRef.t -> scope_name list list -> unit
 
-val find_arguments_scope : GlobRef.t -> scope_name option list
+val find_arguments_scope : GlobRef.t -> scope_name list list
 
 type scope_class
 
@@ -281,14 +303,16 @@ val scope_class_compare : scope_class -> scope_class -> int
 val subst_scope_class :
   Environ.env -> Mod_subst.substitution -> scope_class -> scope_class option
 
-val declare_scope_class : scope_name -> scope_class -> unit
+type add_scope_where = AddScopeTop | AddScopeBottom
+(** add new scope at top or bottom of existing stack (default is reset) *)
+val declare_scope_class : scope_name -> ?where:add_scope_where -> scope_class -> unit
 val declare_ref_arguments_scope : Evd.evar_map -> GlobRef.t -> unit
 
-val compute_arguments_scope : Environ.env -> Evd.evar_map -> EConstr.types -> scope_name option list
-val compute_type_scope : Environ.env -> Evd.evar_map -> EConstr.types -> scope_name option
+val compute_arguments_scope : Environ.env -> Evd.evar_map -> EConstr.types -> scope_name list list
+val compute_type_scope : Environ.env -> Evd.evar_map -> EConstr.types -> scope_name list
 
-(** Get the current scope bound to Sortclass, if it exists *)
-val current_type_scope_name : unit -> scope_name option
+(** Get the current scopes bound to Sortclass *)
+val current_type_scope_names : unit -> scope_name list
 
 val scope_class_of_class : Coercionops.cl_typ -> scope_class
 
@@ -306,8 +330,16 @@ val symbol_eq : symbol -> symbol -> bool
 val make_notation_key : notation_entry -> symbol list -> notation
 val decompose_notation_key : notation -> notation_entry * symbol list
 
-(** Decompose a notation of the form "a 'U' b" *)
-val decompose_raw_notation : string -> symbol list
+type notation_symbols = {
+  recvars : (Id.t * Id.t) list; (* pairs (x,y) as in [ x ; .. ; y ] *)
+  mainvars : Id.t list; (* variables non involved in a recursive pattern *)
+  symbols : symbol list; (* the decomposition of the notation into terminals and nonterminals *)
+}
+
+(** Decompose a notation of the form "a 'U' b" together with the lists
+    of pairs of recursive variables and the list of all variables
+    binding in the notation *)
+val decompose_raw_notation : string -> notation_symbols
 
 (** Prints scopes (expects a pure aconstr printer) *)
 val pr_scope_class : scope_class -> Pp.t
@@ -349,4 +381,4 @@ val level_of_notation : notation -> level
 val with_notation_protection : ('a -> 'b) -> 'a -> 'b
 
 (** Conversion from bigint to int63 *)
-val int63_of_pos_bigint : Bigint.bigint -> Uint63.t
+val int63_of_pos_bigint : Z.t -> Uint63.t
