@@ -65,23 +65,21 @@ let mind_check_names mie =
 (************************************************************************)
 
 type univ_info = { ind_squashed : bool; ind_has_relevant_arg : bool;
-                   ind_min_univ : Universe.t option; (* Some for template *)
-                   ind_univ : Universe.t;
-                   missing : Universe.Set.t; (* missing u <= ind_univ constraints *)
+                   ind_template : bool;
+                   ind_univ : Sorts.t;
+                   missing : Sorts.t list; (* missing u <= ind_univ constraints *)
                  }
 
 let check_univ_leq ?(is_real_arg=false) env u info =
   let ind_univ = info.ind_univ in
-  let info = if not info.ind_has_relevant_arg && is_real_arg && not (Univ.Universe.is_sprop u)
+  let info = if not info.ind_has_relevant_arg && is_real_arg && not (Sorts.is_sprop u)
     then {info with ind_has_relevant_arg=true}
     else info
   in
   (* Inductive types provide explicit lifting from SProp to other universes, so allow SProp <= any. *)
-  if type_in_type env || Univ.Universe.is_sprop u || UGraph.check_leq (universes env) u ind_univ
-  then { info with ind_min_univ = Option.map (Universe.sup u) info.ind_min_univ }
-  else if is_impredicative_univ env ind_univ
-       && Option.is_empty info.ind_min_univ then { info with ind_squashed = true }
-  else {info with missing = Universe.Set.add u info.missing}
+  if Sorts.is_sprop u || UGraph.check_leq_sort (universes env) u ind_univ then info
+  else if is_impredicative_sort env ind_univ then { info with ind_squashed = true }
+  else {info with missing = u :: info.missing}
 
 let check_context_univs ~ctor env info ctx =
   let check_one d (info,env) =
@@ -89,7 +87,7 @@ let check_context_univs ~ctor env info ctx =
       | LocalAssum (_,t) ->
         (* could be retyping if it becomes available in the kernel *)
         let tj = Typeops.infer_type env t in
-        check_univ_leq ~is_real_arg:ctor env (Sorts.univ_of_sort tj.utj_type) info
+        check_univ_leq ~is_real_arg:ctor env tj.utj_type info
       | LocalDef _ -> info
     in
     info, push_rel d env
@@ -104,13 +102,12 @@ let check_indices_matter env_params info indices =
 let check_arity ~template env_params env_ar ind =
   let {utj_val=arity;utj_type=_} = Typeops.infer_type env_params ind.mind_entry_arity in
   let indices, ind_sort = Reduction.dest_arity env_params arity in
-  let ind_min_univ = if template then Some Universe.type0m else None in
   let univ_info = {
     ind_squashed=false;
     ind_has_relevant_arg=false;
-    ind_min_univ;
-    ind_univ=Sorts.univ_of_sort ind_sort;
-    missing=Universe.Set.empty;
+    ind_template = template;
+    ind_univ=ind_sort;
+    missing=[];
   }
   in
   let univ_info = check_indices_matter env_params univ_info indices in
@@ -134,14 +131,21 @@ let check_constructors env_ar_par isrecord params lc (arity,indices,univ_info) =
     (* Empty type: all OK *)
     | 0 -> univ_info
 
-    (* SProp primitive records are OK, if we squash and become fakerecord also OK *)
-    | 1 when isrecord -> univ_info
-
-    (* Unit and identity types must squash if SProp *)
-    | 1 -> check_univ_leq env_ar_par Univ.Universe.type0m univ_info
+    | 1 ->
+      (* SProp primitive records are OK, if we squash and become fakerecord also OK *)
+      if isrecord then univ_info
+      (* 1 constructor with no arguments also OK in SProp (to make
+         things easier on ourselves when reducing we forbid letins) *)
+      else if (Environ.typing_flags env_ar_par).allow_uip
+           && fst (splayed_lc.(0)) = []
+           && List.for_all Context.Rel.Declaration.is_local_assum params
+      then univ_info
+      (* 1 constructor with arguments must squash if SProp
+         (we could allow arguments in SProp but the reduction rule is a pain) *)
+      else check_univ_leq env_ar_par Sorts.prop univ_info
 
     (* More than 1 constructor: must squash if Prop/SProp *)
-    | _ -> check_univ_leq env_ar_par Univ.Universe.type0 univ_info
+    | _ -> check_univ_leq env_ar_par Sorts.set univ_info
   in
   let univ_info = Array.fold_left (check_constructor_univs env_ar_par) univ_info splayed_lc in
   (* generalize the constructors over the parameters *)
@@ -153,7 +157,7 @@ let check_record data =
       (* records must have all projections definable -> equivalent to not being squashed *)
       not info.ind_squashed
       (* relevant records must have at least 1 relevant argument *)
-      && (Univ.Universe.is_sprop info.ind_univ
+      && (Sorts.is_sprop info.ind_univ
           || info.ind_has_relevant_arg)
       && (match splayed_lc with
           (* records must have 1 constructor with at least 1 argument, and no anonymous fields *)
@@ -176,9 +180,9 @@ let check_record data =
 (* - all_sorts in case of small, unitary Prop (not smashed) *)
 (* - logical_sorts in case of large, unitary Prop (smashed) *)
 
-let allowed_sorts {ind_squashed;ind_univ;ind_min_univ=_;ind_has_relevant_arg=_;missing=_} =
+let allowed_sorts {ind_squashed;ind_univ;ind_template=_;ind_has_relevant_arg=_;missing=_} =
   if not ind_squashed then InType
-  else Sorts.family (Sorts.sort_of_univ ind_univ)
+  else Sorts.family ind_univ
 
 (* For a level to be template polymorphic, it must be introduced
    by the definition (so have no constraint except lbound <= l)
@@ -186,103 +190,61 @@ let allowed_sorts {ind_squashed;ind_univ;ind_min_univ=_;ind_has_relevant_arg=_;m
    can be used as an instance of l. All bounds from above, i.e.
    l <=/< r will be valid for any l' <= l. *)
 let unbounded_from_below u cstrs =
-  Univ.Constraint.for_all (fun (l, d, r) ->
+  Univ.Constraints.for_all (fun (l, d, r) ->
       match d with
       | Eq -> not (Univ.Level.equal l u) && not (Univ.Level.equal r u)
       | Lt | Le -> not (Univ.Level.equal r u))
     cstrs
 
-(* Returns the list [x_1, ..., x_n] of levels contributing to template
-   polymorphism. The elements x_k is None if the k-th parameter
-   (starting from the most recent and ignoring let-definitions) is not
-   contributing to the inductive type's sort or is Some u_k if its level
-   is u_k and is contributing. *)
-let template_polymorphic_univs ~ctor_levels uctx paramsctxt concl =
-  let check_level l =
-    Univ.LSet.mem l (Univ.ContextSet.levels uctx) &&
-    (let () = assert (not @@ Univ.Level.is_small l) in true) &&
-    unbounded_from_below l (Univ.ContextSet.constraints uctx) &&
-    not (Univ.LSet.mem l ctor_levels)
-  in
-  let univs = Univ.Universe.levels concl in
-  let univs = Univ.LSet.filter (fun l -> check_level l) univs in
-  let fold acc = function
-    | (LocalAssum (_, p)) ->
-      (let c = Term.strip_prod_assum p in
-      match kind c with
-        | Sort (Type u) ->
-            (match Univ.Universe.level u with
-             | Some l -> if Univ.LSet.mem l univs then Some l else None
-             | None -> None)
-        | _ -> None) :: acc
-    | LocalDef _ -> acc
-  in
-  let params = List.fold_left fold [] paramsctxt in
-  if Universe.is_type0m concl then Some (univs, params)
-  else if not @@ Univ.LSet.is_empty univs then Some (univs, params)
-  else None
-
-let get_param_levels ctx params arity splayed_lc =
-  let min_univ = match arity with
-  | RegularArity _ ->
-    CErrors.user_err
-      Pp.(strbrk "Ill-formed template mutual inductive declaration: all types must be template.")
-  | TemplateArity ar -> ar.template_level
-  in
-  let ctor_levels =
-    let add_levels c levels = Univ.LSet.union levels (Vars.universes_of_constr c) in
-    let param_levels =
-      List.fold_left (fun levels d -> match d with
-          | LocalAssum _ -> levels
-          | LocalDef (_,b,t) -> add_levels b (add_levels t levels))
-        Univ.LSet.empty params
+let get_template univs ~env_params ~env_ar_par ~params entries =
+  match univs with
+  | Polymorphic_ind_entry _ | Monomorphic_ind_entry -> None
+  | Template_ind_entry ctx ->
+    let entry = match entries with
+      | [entry] -> entry
+      | _ -> CErrors.user_err Pp.(str "Template-polymorphism not allowed with mutual inductives.")
     in
-    Array.fold_left
-      (fun levels (d,c) ->
-          let levels =
-            List.fold_left (fun levels d ->
-                Context.Rel.Declaration.fold_constr add_levels d levels)
-              levels d
-          in
-          add_levels c levels)
-      param_levels
-      splayed_lc
-  in
-  match template_polymorphic_univs ~ctor_levels ctx params min_univ with
-  | None ->
-    CErrors.user_err
-      Pp.(strbrk "Ill-formed template inductive declaration: not polymorphic on any universe.")
-  | Some (_, param_levels) ->
-    param_levels
-
-let get_template univs params data =
-  let ctx = match univs with
-      | Monomorphic ctx -> ctx
-      | Polymorphic _ ->
-        CErrors.anomaly ~label:"polymorphic_template_ind"
-          Pp.(strbrk "Template polymorphism and full polymorphism are incompatible.") in
-  (* For each type in the block, compute potential template parameters *)
-  let params = List.map (fun ((arity, _), (_, splayed_lc), _) -> get_param_levels ctx params arity splayed_lc) data in
-  (* Pick the lower bound of template parameters. Note that in particular, if
-     one of the the inductive types from the block is Prop-valued, then no
-     parameters are template. *)
-  let fold min params =
-    let map u v = match u, v with
-    | (None, _) | (_, None) -> None
-    | Some u, Some v ->
-      let () = assert (Univ.Level.equal u v) in
-      Some u
+    (* Compute potential template parameters *)
+    let map decl = match decl with
+    | LocalAssum (_, p) ->
+      let c = Term.strip_prod_assum p in
+      let s = match kind c with
+      | Sort (Type u) ->
+        begin match Universe.level u with
+        | Some l -> if Level.Set.mem l (fst ctx) then Some l else None
+        | None -> None
+        end
+      | _ -> None
+      in
+      Some s
+    | LocalDef _ -> None
     in
-    List.map2 map min params
-  in
-  let params = match params with
-  | [] -> assert false
-  | hd :: rem -> List.fold_left fold hd rem
-  in
-  { template_param_levels = params; template_context = ctx }
+    let params = List.map_filter map params in
+    let fold accu u = match u with None -> accu | Some u -> Level.Set.add u accu in
+    let plevels = List.fold_left fold Level.Set.empty params in
+    let unbound = Level.Set.diff (fst ctx) plevels in
+    let plevels =
+      if not (Level.Set.is_empty unbound) then
+        CErrors.user_err Pp.(strbrk "The following template universes are not \
+          bound by parameters: " ++ pr_sequence Level.pr (Level.Set.elements unbound))
+      else Level.Set.elements plevels
+    in
+    let check_bound l =
+      if not (unbounded_from_below l (snd ctx)) then
+        CErrors.user_err Pp.(strbrk "Universe level " ++ Level.pr l ++ strbrk " has a lower bound")
+    in
+    let () = List.iter check_bound plevels in
+    (* We reuse the same code as the one for variance inference. *)
+    let init_variance = Array.map_of_list (fun l -> l, Some Variance.Irrelevant) plevels in
+    let _variance = InferCumulativity.infer_inductive ~env_params ~env_ar_par init_variance
+        ~arities:[entry.mind_entry_arity]
+        ~ctors:[entry.mind_entry_lc]
+    in
+    let params = List.rev params in
+    Some { template_param_levels = params; template_context = ctx }
 
 let abstract_packets usubst ((arity,lc),(indices,splayed_lc),univ_info) =
-  if not (Universe.Set.is_empty univ_info.missing)
+  if not (List.is_empty univ_info.missing)
   then raise (InductiveError (MissingConstraints (univ_info.missing,univ_info.ind_univ)));
   let arity = Vars.subst_univs_level_constr usubst arity in
   let lc = Array.map (Vars.subst_univs_level_constr usubst) lc in
@@ -293,11 +255,16 @@ let abstract_packets usubst ((arity,lc),(indices,splayed_lc),univ_info) =
       args,out)
       splayed_lc
   in
-  let ind_univ = Univ.subst_univs_level_universe usubst univ_info.ind_univ in
+  let ind_univ = match univ_info.ind_univ with
+  | Prop | SProp | Set -> univ_info.ind_univ
+  | Type u -> Sorts.sort_of_univ (Univ.subst_univs_level_universe usubst u)
+  in
 
-  let arity = match univ_info.ind_min_univ with
-    | None -> RegularArity {mind_user_arity = arity; mind_sort = Sorts.sort_of_univ ind_univ}
-    | Some min_univ -> TemplateArity { template_level = min_univ; }
+  let arity =
+    if univ_info.ind_template then
+      TemplateArity { template_level = univ_info.ind_univ; }
+    else
+      RegularArity {mind_user_arity = arity; mind_sort = ind_univ}
   in
 
   let kelim = allowed_sorts univ_info in
@@ -312,21 +279,21 @@ let typecheck_inductive env ~sec_univs (mie:mutual_inductive_entry) =
   mind_check_names mie;
   assert (List.is_empty (Environ.rel_context env));
 
-  let has_template_poly = mie.mind_entry_template in
-
   (* universes *)
   let env_univs =
     match mie.mind_entry_universes with
-    | Monomorphic_entry ctx ->
-      if has_template_poly then
+    | Template_ind_entry ctx ->
         (* For that particular case, we typecheck the inductive in an environment
            where the universes introduced by the definition are only [>= Prop] *)
         let env = set_universes_lbound env UGraph.Bound.Prop in
         push_context_set ~strict:false ctx env
-      else
-        (* In the regular case, all universes are [> Set] *)
-        push_context_set ~strict:true ctx env
-    | Polymorphic_entry (_, ctx) -> push_context ctx env
+    | Monomorphic_ind_entry -> env
+    | Polymorphic_ind_entry ctx -> push_context ctx env
+  in
+
+  let has_template_poly = match mie.mind_entry_universes with
+  | Template_ind_entry _ -> true
+  | Monomorphic_ind_entry | Polymorphic_ind_entry _ -> false
   in
 
   (* Params *)
@@ -356,37 +323,48 @@ let typecheck_inductive env ~sec_univs (mie:mutual_inductive_entry) =
         (* if someone tried to declare a record as SProp but it can't
            be primitive we must squash. *)
         let data = List.map (fun (a,b,univs) ->
-            a,b,check_univ_leq env_ar_par Univ.Universe.type0m univs)
+            a,b,check_univ_leq env_ar_par Sorts.prop univs)
             data
         in
         data, Some None
   in
 
-  let variance = if not mie.mind_entry_cumulative then None
-    else match mie.mind_entry_universes with
-      | Monomorphic_entry _ ->
+  let variance = match mie.mind_entry_variance with
+    | None -> None
+    | Some variances ->
+      match mie.mind_entry_universes with
+      | Monomorphic_ind_entry | Template_ind_entry _ ->
         CErrors.user_err Pp.(str "Inductive cannot be both monomorphic and universe cumulative.")
-      | Polymorphic_entry (_,uctx) ->
+      | Polymorphic_ind_entry uctx ->
         let univs = Instance.to_array @@ UContext.instance uctx in
+        let univs = Array.map2 (fun a b -> a,b) univs variances in
         let univs = match sec_univs with
           | None -> univs
-          | Some sec_univs -> Array.append sec_univs univs
+          | Some sec_univs ->
+            let sec_univs = Array.map (fun u -> u, None) sec_univs in
+            Array.append sec_univs univs
+        in
+        let variances = InferCumulativity.infer_inductive ~env_params ~env_ar_par
+            ~arities:(List.map (fun e -> e.mind_entry_arity) mie.mind_entry_inds)
+            ~ctors:(List.map (fun e -> e.mind_entry_lc) mie.mind_entry_inds)
+            univs
         in
-        let variances = InferCumulativity.infer_inductive ~env_params univs mie.mind_entry_inds in
         Some variances
   in
 
+  let template = get_template mie.mind_entry_universes ~env_params ~env_ar_par ~params mie.mind_entry_inds in
+
   (* Abstract universes *)
-  let usubst, univs = Declareops.abstract_universes mie.mind_entry_universes in
+  let usubst, univs = match mie.mind_entry_universes with
+  | Monomorphic_ind_entry | Template_ind_entry _ ->
+    Univ.empty_level_subst, Monomorphic
+  | Polymorphic_ind_entry uctx ->
+    let (inst, auctx) = Univ.abstract_universes uctx in
+    let inst = Univ.make_instance_subst inst in
+    (inst, Polymorphic auctx)
+  in
   let params = Vars.subst_univs_level_context usubst params in
   let data = List.map (abstract_packets usubst) data in
-  let template =
-    let check ((arity, _), _, _) = match arity with
-    | TemplateArity _ -> true
-    | RegularArity _ -> false
-    in
-    if List.exists check data then Some (get_template univs params data) else None
-  in
 
   let env_ar_par =
     let ctx = Environ.rel_context env_ar_par in
