@@ -30,18 +30,23 @@ type mind_specif = mutual_inductive_body * one_inductive_body
 
 (** {6 ... } *)
 (** Fetching information in the environment about an inductive type.
-    Raises [Not_found] if the inductive type is not found. *)
+    Raises an anomaly if the inductive type is not found. *)
 val lookup_mind_specif : env -> inductive -> mind_specif
 
 (** {6 Functions to build standard types related to inductive } *)
-val ind_subst : MutInd.t -> mutual_inductive_body -> Instance.t -> constr list
 
+(** Returns the parameters of an inductive type with universes instantiated *)
 val inductive_paramdecls : mutual_inductive_body puniverses -> Constr.rel_context
 
-val instantiate_inductive_constraints :
-  mutual_inductive_body -> Instance.t -> Constraint.t
+(** Returns the parameters of an inductive type with universes
+    instantiated, splitting it into the contexts of recursively
+    uniform and recursively non-uniform parameters *)
+val inductive_nonrec_rec_paramdecls : mutual_inductive_body puniverses -> Constr.rel_context * Constr.rel_context
 
-type param_univs = (unit -> Universe.t) list
+val instantiate_inductive_constraints :
+  mutual_inductive_body -> Instance.t -> Constraints.t
+
+type param_univs = (unit -> Sorts.t) list
 
 val make_param_univs : Environ.env -> constr array -> param_univs
 (** The constr array is the types of the arguments to a template
@@ -74,21 +79,32 @@ val arities_of_constructors : pinductive -> mind_specif -> types array
 (** Return constructor types in user form *)
 val type_of_constructors : pinductive -> mind_specif -> types array
 
-(** Transforms inductive specification into types (in nf) *)
-val arities_of_specif : MutInd.t puniverses -> mind_specif -> types array
+(** Turns a constructor type recursively referring to inductive types
+    into the same constructor type referring instead to a context made
+    from the abstract declaration of the inductive types (e.g. turns
+    [nat->nat] into [mkArrowR (Rel 1) (Rel 2)]); takes as arguments the number
+    of inductive types in the block and the name of the block *)
+val abstract_constructor_type_relatively_to_inductive_types_context :
+  int -> MutInd.t -> types -> types
 
 val inductive_params : mind_specif -> int
 
-(** [type_case_branches env (I,args) (p:A) c] computes useful types
-   about the following Cases expression:
-      <p>Cases (c :: (I args)) of b1..bn end
-   It computes the type of every branch (pattern variables are
-   introduced by products), the type for the whole expression, and
-   the universe constraints generated.
- *)
-val type_case_branches :
-  env -> pinductive * constr list -> unsafe_judgment -> constr
-    -> types array * types
+(** Given a pattern-matching represented compactly, expands it so as to produce
+    lambda and let abstractions in front of the return clause and the pattern
+    branches. *)
+val expand_case : env -> case -> (case_info * constr * case_invert * constr * constr array)
+
+val expand_case_specif : mutual_inductive_body -> case -> (case_info * constr * case_invert * constr * constr array)
+
+(** Dual operation of the above. Fails if the return clause or branch has not
+    the expected form. *)
+val contract_case : env -> (case_info * constr * case_invert * constr * constr array) -> case
+
+(** [instantiate_context u subst nas ctx] applies both [u] and [subst]
+    to [ctx] while replacing names using [nas] (order reversed). In particular,
+    assumes that [ctx] and [nas] have the same length. *)
+val instantiate_context : Instance.t -> Vars.substl -> Name.t Context.binder_annot array ->
+  rel_context -> rel_context
 
 val build_branches_type :
   pinductive -> mutual_inductive_body * one_inductive_body ->
@@ -105,6 +121,13 @@ val check_case_info : env -> pinductive -> Sorts.relevance -> case_info -> unit
 
 (** {6 Guard conditions for fix and cofix-points. } *)
 
+(** [is_primitive_positive_container env c] tells if the constant [c] is
+    registered as a primitive type that can be seen as a container where the
+    occurrences of its parameters are positive, in which case the positivity and
+    guard conditions are extended to allow inductive types to nest their subterms
+    in these containers. *)
+val is_primitive_positive_container : env -> Constant.t -> bool
+
 (** When [chk] is false, the guard condition is not actually
     checked. *)
 val check_fix : env -> fixpoint -> unit
@@ -121,27 +144,4 @@ val check_cofix : env -> cofixpoint -> unit
 
 exception SingletonInductiveBecomesProp of Id.t
 
-val max_inductive_sort : Sorts.t array -> Universe.t
-
-(** {6 Debug} *)
-
-type size = Large | Strict
-type subterm_spec =
-    Subterm of (size * wf_paths)
-  | Dead_code
-  | Not_subterm
-type guard_env =
-  { env     : env;
-    (** dB of last fixpoint *)
-    rel_min : int;
-    (** dB of variables denoting subterms *)
-    genv    : subterm_spec Lazy.t list;
-  }
-
-type stack_element = |SClosure of guard_env*constr |SArg of subterm_spec Lazy.t
-
-val subterm_specif : guard_env -> stack_element list -> constr -> subterm_spec
-
-val lambda_implicit_lift : int -> constr -> constr
-
-val abstract_mind_lc : int -> Int.t -> (rel_context * constr) array -> constr array
+val abstract_mind_lc : int -> int -> MutInd.t -> (rel_context * constr) array -> constr array
