@@ -11,12 +11,44 @@
 open Util
 open CErrors
 open Names
+open Values
 open Constr
 
 (** This module defines the representation of values internally used by
 the native compiler *)
 
-type t = t -> t
+type ('a,'b) eq = ('a,'b) Util.eq = Refl : ('a,'a) eq
+
+type t
+let t_eq : (t, t -> t) eq = Obj.magic ()
+let t_eq2 : (t, t -> t -> t) eq = Obj.magic ()
+let t_eq3 : (t, t -> t -> t -> t) eq = Obj.magic ()
+let t_eq4 : (t, t -> t -> t -> t -> t) eq = Obj.magic ()
+
+let cast_gen (type a) (type b) (e:(a,b) eq) (x:a) : b =
+  let Refl = e in x
+
+let apply (f:t) : t -> t =
+  cast_gen t_eq f
+
+(* Writing this as [apply2 f x y = apply (apply f x) y] seems to
+   confuse ocaml and it produces less efficient code.
+
+   When written with a direct [cast_gen t_eq2] ocaml seems smart
+   enough to fully reduce it where used. *)
+let apply2 (f:t) : t -> t -> t =
+  cast_gen t_eq2 f
+
+let apply3 (f:t) : t -> t -> t -> t =
+  cast_gen t_eq3 f
+
+let apply4 (f:t) : t -> t -> t -> t -> t =
+  cast_gen t_eq4 f
+
+let of_fun (f:t->t) : t =
+  cast_gen (sym t_eq) f
+
+let eta_expand f = of_fun (fun x -> apply f x)
 
 type accumulator (* = t (* a block [0:code;atom;arguments] *) *)
 
@@ -36,13 +68,13 @@ type annot_sw = {
 
 (* We compare only what is relevant for generation of ml code *)
 let eq_annot_sw asw1 asw2 =
-  eq_ind asw1.asw_ind asw2.asw_ind &&
+  Ind.CanOrd.equal asw1.asw_ind asw2.asw_ind &&
   String.equal asw1.asw_prefix asw2.asw_prefix
 
 open Hashset.Combine
 
 let hash_annot_sw asw =
-  combine (ind_hash asw.asw_ind) (String.hash asw.asw_prefix)
+  combine (Ind.CanOrd.hash asw.asw_ind) (String.hash asw.asw_prefix)
 
 type sort_annot = string * int
 
@@ -50,19 +82,18 @@ type rec_pos = int array
 
 let eq_rec_pos = Array.equal Int.equal
 
+type vcofix = CofixLazy of t | CofixValue of t
+
 type atom =
   | Arel of int
   | Aconstant of pconstant
   | Aind of pinductive
   | Asort of Sorts.t
   | Avar of Id.t
-  | Acase of annot_sw * accumulator * t * (t -> t)
+  | Acase of annot_sw * accumulator * t * t
   | Afix of t array * t array * rec_pos * int
             (* types, bodies, rec_pos, pos *)
-  | Acofix of t array * t array * int * t
-  | Acofixe of t array * t array * int * t
-  | Aprod of Name.t * t * (t -> t)
-  | Ameta of metavariable * t
+  | Acofix of t array * t array * int * vcofix
   | Aevar of Evar.t * t array
   | Aproj of (inductive * int) * accumulator
 
@@ -73,11 +104,12 @@ type symbol =
   | SymbConst of Constant.t
   | SymbMatch of annot_sw
   | SymbInd of inductive
-  | SymbMeta of metavariable
   | SymbEvar of Evar.t
   | SymbLevel of Univ.Level.t
   | SymbProj of (inductive * int)
 
+type block
+
 type symbols = symbol array
 
 let empty_symbols = [| |]
@@ -90,20 +122,22 @@ let ret_accu = Obj.repr (ref ())
 
 type accu_val = { mutable acc_atm : atom; acc_arg : Obj.t list }
 
+external set_tag : Obj.t -> int -> unit = "coq_obj_set_tag"
+
 let mk_accu (a : atom) : t =
   let rec accumulate data x =
     if x == ret_accu then Obj.repr data
     else
       let data = { data with acc_arg = x :: data.acc_arg } in
       let ans = Obj.repr (accumulate data) in
-      let () = Obj.set_tag ans accumulate_tag [@ocaml.warning "-3"] in
+      let () = set_tag ans accumulate_tag in
       ans
   in
   let acc = { acc_atm = a; acc_arg = [] } in
   let ans = Obj.repr (accumulate acc) in
   (** FIXME: use another representation for accumulators, this causes naked
       pointers. *)
-  let () = Obj.set_tag ans accumulate_tag [@ocaml.warning "-3"] in
+  let () = set_tag ans accumulate_tag in
   (Obj.obj ans : t)
 
 let get_accu (k : accumulator) =
@@ -113,17 +147,28 @@ let mk_rel_accu i =
   mk_accu (Arel i)
 
 let rel_tbl_size = 100
-let rel_tbl = Array.init rel_tbl_size mk_rel_accu
+let rel_tbl_init = ref false
+(* Initialize the table lazily not to generate naked pointers at startup *)
+let rel_tbl = Array.make rel_tbl_size (Obj.magic 0 : t)
 
 let mk_rel_accu i =
-  if i < rel_tbl_size then rel_tbl.(i)
+  if i < rel_tbl_size then
+    let () =
+      if not !rel_tbl_init then begin
+        for i = 0 to rel_tbl_size - 1 do
+          rel_tbl.(i) <- mk_rel_accu i
+        done;
+        rel_tbl_init := true
+      end
+    in
+    rel_tbl.(i)
   else mk_rel_accu i
 
 let mk_rels_accu lvl len =
   Array.init len (fun i -> mk_rel_accu (lvl + i))
 
 let napply (f:t) (args: t array) =
-  Array.fold_left (fun f a -> f a) f args
+  Array.fold_left (fun f a -> apply f a) f args
 
 let mk_constant_accu kn u =
   mk_accu (Aconstant (kn,Univ.Instance.of_array u))
@@ -146,11 +191,18 @@ let mk_var_accu id =
 let mk_sw_accu annot c p ac =
   mk_accu (Acase(annot,c,p,ac))
 
-let mk_prod_accu s dom codom =
-  mk_accu (Aprod (s,dom,codom))
+let prod_tag =
+  (* We rely on the tag of Vprod! *)
+  let () = assert (Obj.tag (Obj.repr (Vprod (Anonymous, Obj.magic 0, Obj.magic 0))) == 2) in
+  2
 
-let mk_meta_accu mv ty =
-  mk_accu (Ameta (mv,ty))
+let mk_prod s dom codom =
+  (* [Prod (s, dom, codom)] is coded as [tag:0|[tag:2|s; dom; codom]]
+     This looks like a PArray but has a tag distinct from all PArray values on
+     the inner block. This cannot be an accumulator because all accumulators
+     have length >= 2. *)
+  let block = Obj.repr (Vprod (s, dom, codom)) in
+  (Obj.magic (ref block) : t)
 
 let mk_evar_accu ev args =
   mk_accu (Aevar (ev, args))
@@ -175,26 +227,26 @@ let mk_fix_accu rec_pos pos types bodies =
   mk_accu (Afix(types,bodies,rec_pos, pos))
 
 let mk_cofix_accu pos types norm =
-  mk_accu (Acofix(types,norm,pos,(Obj.magic 0 : t)))
+  mk_accu (Acofix (types, norm, pos, CofixLazy (Obj.magic 0 : t)))
 
 let upd_cofix (cofix :t) (cofix_fun : t) =
   let atom = atom_of_accu (Obj.magic cofix) in
   match atom with
   | Acofix (typ,norm,pos,_) ->
-      set_atom_of_accu (Obj.magic cofix) (Acofix(typ,norm,pos,cofix_fun))
+    set_atom_of_accu (Obj.magic cofix) (Acofix (typ, norm, pos, CofixLazy cofix_fun))
   | _ -> assert false
 
 let force_cofix (cofix : t) =
   let accu = (Obj.magic cofix : accumulator) in
   let atom = atom_of_accu accu in
   match atom with
-  | Acofix(typ,norm,pos,f) ->
+  | Acofix (typ, norm, pos, CofixLazy f) ->
     let args = args_of_accu accu in
-    let f = Array.fold_right (fun arg f -> f arg) args f in
-    let v = f (Obj.magic ()) in
-    set_atom_of_accu accu (Acofixe(typ,norm,pos,v));
+    let f = Array.fold_right (fun arg f -> apply f arg) args f in
+    let v = apply f (Obj.magic ()) in
+    let () = set_atom_of_accu accu (Acofix (typ, norm, pos, CofixValue v)) in
       v
-  | Acofixe(_,_,_,v) -> v
+  | Acofix (_, _, _, CofixValue v) -> v
   | _ -> cofix
 
 let mk_const tag = Obj.magic tag
@@ -210,7 +262,7 @@ let mk_block tag args =
 (* Two instances of dummy_value should not be pointer equal, otherwise
  comparing them as terms would succeed *)
 let dummy_value : unit -> t =
-  fun () _ -> anomaly ~label:"native" (Pp.str "Evaluation failed.")
+  fun () -> of_fun (fun _ -> anomaly ~label:"native" (Pp.str "Evaluation failed."))
 
 let cast_accu v = (Obj.magic v:accumulator)
 [@@ocaml.inline always]
@@ -228,8 +280,6 @@ let mk_uint (x : Uint63.t) = (Obj.magic x : t)
 let mk_float (x : Float64.t) = (Obj.magic x : t)
 [@@ocaml.inline always]
 
-type block
-
 let block_size (b:block) =
   Obj.size (Obj.magic b)
 
@@ -238,29 +288,27 @@ let block_field (b:block) i = (Obj.magic (Obj.field (Obj.magic b) i) : t)
 let block_tag (b:block) =
   Obj.tag (Obj.magic b)
 
-type kind_of_value =
-  | Vaccu of accumulator
-  | Vfun of (t -> t)
-  | Vconst of int
-  | Vint64 of int64
-  | Vfloat64 of float
-  | Vblock of block
+type kind = (t, accumulator, t -> t, Name.t * t * t, Empty.t, Empty.t, block) Values.kind
 
 let kind_of_value (v:t) =
   let o = Obj.repr v in
   if Obj.is_int o then Vconst (Obj.magic v)
-  else if Obj.tag o == Obj.double_tag then Vfloat64 (Obj.magic v)
   else
     let tag = Obj.tag o in
     if Int.equal tag accumulate_tag then
-      Vaccu (Obj.magic v)
+      if Int.equal (Obj.size o) 1 then
+        let w = Obj.field o 0 in
+        let tag = Obj.tag w in
+        if Int.equal tag prod_tag then Obj.magic w
+        else Varray (Obj.magic v)
+      else Vaccu (Obj.magic v)
     else if Int.equal tag Obj.custom_tag then Vint64 (Obj.magic v)
     else if Int.equal tag Obj.double_tag then Vfloat64 (Obj.magic v)
     else if (tag < Obj.lazy_tag) then Vblock (Obj.magic v)
       else
         (* assert (tag = Obj.closure_tag || tag = Obj.infix_tag);
            or ??? what is 1002*)
-        Vfun v
+        Vfun (apply v)
 
 (** Support for machine integers *)
 
@@ -281,7 +329,7 @@ let no_check_head0 x =
 
 let head0 accu x =
  if is_int x then  no_check_head0 x
- else accu x
+ else apply accu x
 
 let no_check_tail0 x =
   mk_uint (Uint63.tail0 (to_uint x))
@@ -289,7 +337,7 @@ let no_check_tail0 x =
 
 let tail0 accu x =
  if is_int x then no_check_tail0 x
- else accu x
+ else apply accu x
 
 let no_check_add  x y =
   mk_uint (Uint63.add (to_uint x) (to_uint y))
@@ -297,7 +345,7 @@ let no_check_add  x y =
 
 let add accu x y =
   if is_int x && is_int y then no_check_add x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_sub x y =
   mk_uint (Uint63.sub (to_uint x) (to_uint y))
@@ -305,7 +353,7 @@ let no_check_sub x y =
 
 let sub accu x y =
   if is_int x && is_int y then no_check_sub x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_mul x y =
   mk_uint (Uint63.mul (to_uint x) (to_uint y))
@@ -313,7 +361,7 @@ let no_check_mul x y =
 
 let mul accu x y =
   if is_int x && is_int y then no_check_mul x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_div x y =
   mk_uint (Uint63.div (to_uint x) (to_uint y))
@@ -321,7 +369,7 @@ let no_check_div x y =
 
 let div accu x y =
   if is_int x && is_int y then no_check_div x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_rem x y =
   mk_uint (Uint63.rem (to_uint x) (to_uint y))
@@ -329,7 +377,23 @@ let no_check_rem x y =
 
 let rem accu x y =
   if is_int x && is_int y then no_check_rem x y
-  else accu x y
+  else apply2 accu x y
+
+let no_check_divs x y =
+  mk_uint (Uint63.divs (to_uint x) (to_uint y))
+[@@ocaml.inline always]
+
+let divs accu x y =
+  if is_int x && is_int y then no_check_divs x y
+  else apply2 accu x y
+
+let no_check_rems x y =
+  mk_uint (Uint63.rems (to_uint x) (to_uint y))
+[@@ocaml.inline always]
+
+let rems accu x y =
+  if is_int x && is_int y then no_check_rems x y
+  else apply2 accu x y
 
 let no_check_l_sr x y =
   mk_uint (Uint63.l_sr (to_uint x) (to_uint y))
@@ -337,7 +401,7 @@ let no_check_l_sr x y =
 
 let l_sr accu x y =
   if is_int x && is_int y then no_check_l_sr x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_l_sl x y =
   mk_uint (Uint63.l_sl (to_uint x) (to_uint y))
@@ -345,7 +409,15 @@ let no_check_l_sl x y =
 
 let l_sl accu x y =
   if is_int x && is_int y then no_check_l_sl x y
-  else accu x y
+  else apply2 accu x y
+
+let no_check_a_sr x y =
+  mk_uint (Uint63.a_sr (to_uint x) (to_uint y))
+[@@ocaml.inline always]
+
+let a_sr accu x y =
+  if is_int x && is_int y then no_check_a_sr x y
+  else apply2 accu x y
 
 let no_check_l_and x y =
   mk_uint (Uint63.l_and (to_uint x) (to_uint y))
@@ -353,7 +425,7 @@ let no_check_l_and x y =
 
 let l_and accu x y =
   if is_int x && is_int y then no_check_l_and x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_l_xor x y =
   mk_uint (Uint63.l_xor (to_uint x) (to_uint y))
@@ -361,7 +433,7 @@ let no_check_l_xor x y =
 
 let l_xor accu x y =
   if is_int x && is_int y then no_check_l_xor x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_l_or x y =
   mk_uint (Uint63.l_or (to_uint x) (to_uint y))
@@ -369,7 +441,7 @@ let no_check_l_or x y =
 
 let l_or accu x y =
   if is_int x && is_int y then no_check_l_or x y
-  else accu x y
+  else apply2 accu x y
 
 [@@@ocaml.warning "-37"]
 type coq_carry =
@@ -392,7 +464,7 @@ let no_check_addc x y =
 
 let addc accu x y =
   if is_int x && is_int y then no_check_addc x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_subc x y =
   let s = Uint63.sub (to_uint x) (to_uint y) in
@@ -401,7 +473,7 @@ let no_check_subc x y =
 
 let subc accu x y =
   if is_int x && is_int y then no_check_subc x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_addCarryC x y =
   let s =
@@ -412,7 +484,7 @@ let no_check_addCarryC x y =
 
 let addCarryC accu x y =
   if is_int x && is_int y then no_check_addCarryC x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_subCarryC x y =
   let s =
@@ -423,7 +495,7 @@ let no_check_subCarryC x y =
 
 let subCarryC accu x y =
   if is_int x && is_int y then no_check_subCarryC x y
-  else accu x y
+  else apply2 accu x y
 
 let of_pair (x, y) =
   (Obj.magic (PPair(mk_uint x, mk_uint y)):t)
@@ -435,7 +507,7 @@ let no_check_mulc x y =
 
 let mulc accu x y =
   if is_int x && is_int y then no_check_mulc x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_diveucl x y =
   let i1, i2 = to_uint x, to_uint y in
@@ -444,7 +516,7 @@ let no_check_diveucl x y =
 
 let diveucl accu x y =
   if is_int x && is_int y then no_check_diveucl x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_div21 x y z =
   let i1, i2, i3 = to_uint x, to_uint y, to_uint z in
@@ -453,7 +525,7 @@ let no_check_div21 x y z =
 
 let div21 accu x y z =
   if is_int x && is_int y && is_int z then no_check_div21 x y z
-  else accu x y z
+  else apply3 accu x y z
 
 let no_check_addMulDiv x y z =
   let p, i, j = to_uint x, to_uint y, to_uint z in
@@ -462,7 +534,7 @@ let no_check_addMulDiv x y z =
 
 let addMulDiv accu x y z =
   if is_int x && is_int y && is_int z then no_check_addMulDiv x y z
-  else accu x y z
+  else apply3 accu x y z
 
 [@@@ocaml.warning "-34"]
 type coq_bool =
@@ -482,7 +554,7 @@ let no_check_eq x y =
 
 let eq accu x y =
   if is_int x && is_int y then no_check_eq x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_lt x y =
   mk_bool (Uint63.lt (to_uint x) (to_uint y))
@@ -490,7 +562,7 @@ let no_check_lt x y =
 
 let lt accu x y =
   if is_int x && is_int y then no_check_lt x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_le x y =
   mk_bool (Uint63.le (to_uint x) (to_uint y))
@@ -498,7 +570,23 @@ let no_check_le x y =
 
 let le accu x y =
   if is_int x && is_int y then no_check_le x y
-  else accu x y
+  else apply2 accu x y
+
+let no_check_lts x y =
+  mk_bool (Uint63.lts (to_uint x) (to_uint y))
+[@@ocaml.inline always]
+
+let lts accu x y =
+  if is_int x && is_int y then no_check_lts x y
+  else apply2 accu x y
+
+let no_check_les x y =
+  mk_bool (Uint63.les (to_uint x) (to_uint y))
+[@@ocaml.inline always]
+
+let les accu x y =
+  if is_int x && is_int y then no_check_les x y
+  else apply2 accu x y
 
 let no_check_compare x y =
   match Uint63.compare (to_uint x) (to_uint y) with
@@ -508,7 +596,17 @@ let no_check_compare x y =
 
 let compare accu x y =
   if is_int x && is_int y then no_check_compare x y
-  else accu x y
+  else apply2 accu x y
+
+let no_check_compares x y =
+  match Uint63.compares (to_uint x) (to_uint y) with
+  | x when x < 0 -> (Obj.magic CmpLt:t)
+  | 0 -> (Obj.magic CmpEq:t)
+  | _ -> (Obj.magic CmpGt:t)
+
+let compares accu x y =
+  if is_int x && is_int y then no_check_compares x y
+  else apply2 accu x y
 
 let print x =
   Printf.fprintf stderr "%s" (Uint63.to_string (to_uint x));
@@ -529,7 +627,7 @@ let no_check_fopp x =
 
 let fopp accu x =
   if is_float x then no_check_fopp x
-  else accu x
+  else apply accu x
 
 let no_check_fabs x =
   mk_float (Float64.abs (to_float x))
@@ -537,28 +635,28 @@ let no_check_fabs x =
 
 let fabs accu x =
   if is_float x then no_check_fabs x
-  else accu x
+  else apply accu x
 
 let no_check_feq x y =
   mk_bool (Float64.eq (to_float x) (to_float y))
 
 let feq accu x y =
   if is_float x && is_float y then no_check_feq x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_flt x y =
   mk_bool (Float64.lt (to_float x) (to_float y))
 
 let flt accu x y =
   if is_float x && is_float y then no_check_flt x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_fle x y =
   mk_bool (Float64.le (to_float x) (to_float y))
 
 let fle accu x y =
   if is_float x && is_float y then no_check_fle x y
-  else accu x y
+  else apply2 accu x y
 
 type coq_fcmp =
   | CFcmpAccu of t
@@ -574,7 +672,14 @@ let no_check_fcompare x y =
 
 let fcompare accu x y =
   if is_float x && is_float y then no_check_fcompare x y
-  else accu x y
+  else apply2 accu x y
+
+let no_check_fequal x y =
+  mk_bool (Float64.equal (to_float x) (to_float y))
+
+let fequal accu x y =
+  if is_float x && is_float y then no_check_fequal x y
+  else apply2 accu x y
 
 type coq_fclass =
   | CFclassAccu of t
@@ -595,7 +700,7 @@ let no_check_fclassify x =
 
 let fclassify accu x =
   if is_float x then no_check_fclassify x
-  else accu x
+  else apply accu x
 
 let no_check_fadd x y =
   mk_float (Float64.add (to_float x) (to_float y))
@@ -603,7 +708,7 @@ let no_check_fadd x y =
 
 let fadd accu x y =
   if is_float x && is_float y then no_check_fadd x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_fsub x y =
   mk_float (Float64.sub (to_float x) (to_float y))
@@ -611,7 +716,7 @@ let no_check_fsub x y =
 
 let fsub accu x y =
   if is_float x && is_float y then no_check_fsub x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_fmul x y =
   mk_float (Float64.mul (to_float x) (to_float y))
@@ -619,7 +724,7 @@ let no_check_fmul x y =
 
 let fmul accu x y =
   if is_float x && is_float y then no_check_fmul x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_fdiv x y =
   mk_float (Float64.div (to_float x) (to_float y))
@@ -627,7 +732,7 @@ let no_check_fdiv x y =
 
 let fdiv accu x y =
   if is_float x && is_float y then no_check_fdiv x y
-  else accu x y
+  else apply2 accu x y
 
 let no_check_fsqrt x =
   mk_float (Float64.sqrt (to_float x))
@@ -635,15 +740,15 @@ let no_check_fsqrt x =
 
 let fsqrt accu x =
   if is_float x then no_check_fsqrt x
-  else accu x
+  else apply accu x
 
 let no_check_float_of_int x =
-  mk_float (Float64.of_int63 (to_uint x))
+  mk_float (Float64.of_uint63 (to_uint x))
 [@@ocaml.inline always]
 
 let float_of_int accu x =
   if is_int x then no_check_float_of_int x
-  else accu x
+  else apply accu x
 
 let no_check_normfr_mantissa x =
   mk_uint (Float64.normfr_mantissa (to_float x))
@@ -651,7 +756,7 @@ let no_check_normfr_mantissa x =
 
 let normfr_mantissa accu x =
   if is_float x then no_check_normfr_mantissa x
-  else accu x
+  else apply accu x
 
 let no_check_frshiftexp x =
   let f, e = Float64.frshiftexp (to_float x) in
@@ -660,7 +765,7 @@ let no_check_frshiftexp x =
 
 let frshiftexp accu x =
   if is_float x then no_check_frshiftexp x
-  else accu x
+  else apply accu x
 
 let no_check_ldshiftexp x e =
   mk_float (Float64.ldshiftexp (to_float x) (to_uint e))
@@ -668,7 +773,7 @@ let no_check_ldshiftexp x e =
 
 let ldshiftexp accu x e =
   if is_float x && is_int e then no_check_ldshiftexp x e
-  else accu x e
+  else apply2 accu x e
 
 let no_check_next_up x =
   mk_float (Float64.next_up (to_float x))
@@ -676,7 +781,7 @@ let no_check_next_up x =
 
 let next_up accu x =
   if is_float x then no_check_next_up x
-  else accu x
+  else apply accu x
 
 let no_check_next_down x =
   mk_float (Float64.next_down (to_float x))
@@ -684,7 +789,73 @@ let no_check_next_down x =
 
 let next_down accu x =
   if is_float x then no_check_next_down x
-  else accu x
+  else apply accu x
+
+let is_parray t =
+  (* This is only used over values known to inhabit an array type, so we just
+     have to discriminate between actual arrays and accumulators. The latter
+     are always closures with the tag set to 0, so they have size >= 2. *)
+  let t = Obj.magic t in
+  Obj.is_block t && Obj.size t = 1
+
+let to_parray t = Obj.magic t
+let of_parray t = Obj.magic t
+
+let no_check_arraymake n def =
+  of_parray (Parray.make (to_uint n) def)
+
+let arraymake accu vA n def =
+  if is_int n then
+    no_check_arraymake n def
+  else apply3 accu vA n def
+
+let no_check_arrayget t n =
+   Parray.get (to_parray t) (to_uint n)
+[@@ocaml.inline always]
+
+let arrayget accu vA t n =
+  if is_parray t && is_int n then
+    no_check_arrayget t n
+  else apply3 accu vA t n
+
+let no_check_arraydefault t =
+  Parray.default (to_parray t)
+[@@ocaml.inline always]
+
+let arraydefault accu vA t =
+  if is_parray t then
+    no_check_arraydefault t
+  else apply2 accu vA t
+
+let no_check_arrayset t n v =
+  of_parray (Parray.set (to_parray t) (to_uint n) v)
+[@@ocaml.inline always]
+
+let arrayset accu vA t n v =
+  if is_parray t && is_int n then
+    no_check_arrayset t n v
+  else apply4 accu vA t n v
+
+let no_check_arraycopy t =
+  of_parray (Parray.copy (to_parray t))
+[@@ocaml.inline always]
+
+let arraycopy accu vA t =
+  if is_parray t then
+    no_check_arraycopy t
+  else apply2 accu vA t
+
+let no_check_arraylength t =
+  mk_uint (Parray.length (to_parray t))
+[@@ocaml.inline always]
+
+let arraylength accu vA t =
+  if is_parray t then
+    no_check_arraylength t
+  else apply2 accu vA t
+
+let parray_of_array t def =
+  (Obj.magic (Parray.unsafe_of_obj (Obj.repr t) def) : t)
 
 let hobcnv = Array.init 256 (fun i -> Printf.sprintf "%02x" i)
 let bohcnv = Array.init 256 (fun i -> i -
