@@ -11,17 +11,19 @@
 open Names
 open Constr
 open Environ
-open Reduction
 open Univ
 
 (* Type errors. *)
 
-type 'constr pguard_error =
+type 'constr pfix_guard_error =
   (* Fixpoints *)
   | NotEnoughAbstractionInFixBody
   | RecursionNotOnInductiveType of 'constr
-  | RecursionOnIllegalTerm of int * (env * 'constr) * int list * int list
+  | RecursionOnIllegalTerm of int * (env * 'constr) * (int list * int list) Lazy.t
   | NotEnoughArgumentsForFixCall of int
+  | FixpointOnIrrelevantInductive
+
+type 'constr pcofix_guard_error =
   (* CoFixpoints *)
   | CodomainNotInductiveType of 'constr
   | NestedRecursiveOccurrences
@@ -34,23 +36,23 @@ type 'constr pguard_error =
   | RecCallInCasePred of 'constr
   | NotGuardedForm of 'constr
   | ReturnPredicateNotCoInductive of 'constr
-  | FixpointOnIrrelevantInductive
 
+type 'constr pguard_error =
+  | FixGuardError of 'constr pfix_guard_error
+  | CoFixGuardError of 'constr pcofix_guard_error
+
+type fix_guard_error = constr pfix_guard_error
+type cofix_guard_error = constr pcofix_guard_error
 type guard_error = constr pguard_error
 
-type arity_error =
-  | NonInformativeToInformative
-  | StrongEliminationOnNonSmallType
-  | WrongArity
-
 type ('constr, 'types) ptype_error =
   | UnboundRel of int
   | UnboundVar of variable
   | NotAType of ('constr, 'types) punsafe_judgment
   | BadAssumption of ('constr, 'types) punsafe_judgment
   | ReferenceVariables of Id.t * GlobRef.t
-  | ElimArity of pinductive * 'constr * ('constr, 'types) punsafe_judgment
-      * (Sorts.family * Sorts.family * Sorts.family * arity_error) option
+  | ElimArity of pinductive * 'constr *
+    (('constr, 'types) punsafe_judgment * Sorts.family * Sorts.family * Sorts.family) option
   | CaseNotInductive of ('constr, 'types) punsafe_judgment
   | WrongCaseInfo of pinductive * case_info
   | NumberBranches of ('constr, 'types) punsafe_judgment * int
@@ -64,10 +66,12 @@ type ('constr, 'types) ptype_error =
   | IllFormedRecBody of 'constr pguard_error * Name.t Context.binder_annot array * int * env * ('constr, 'types) punsafe_judgment array
   | IllTypedRecBody of
       int * Name.t Context.binder_annot array * ('constr, 'types) punsafe_judgment array * 'types array
-  | UnsatisfiedConstraints of Constraint.t
+  | UnsatisfiedConstraints of Constraints.t
   | UndeclaredUniverse of Level.t
   | DisallowedSProp
   | BadRelevance
+  | BadInvert
+  | BadVariance of { lev : Level.t; expected : Variance.t; actual : Variance.t }
 
 type type_error = (constr, types) ptype_error
 
@@ -84,13 +88,10 @@ type inductive_error =
   | NotAnArity of env * constr
   | BadEntry
   | LargeNonPropInductiveNotInType
-  | MissingConstraints of (Universe.Set.t * Universe.t)
+  | MissingConstraints of (Sorts.t list * Sorts.t)
 
 exception InductiveError of inductive_error
 
-let nfj env {uj_val=c;uj_type=ct} =
-  {uj_val=c;uj_type=nf_betaiota env ct}
-
 let error_unbound_rel env n =
   raise (TypeError (env, UnboundRel n))
 
@@ -106,18 +107,17 @@ let error_assumption env j =
 let error_reference_variables env id c =
   raise (TypeError (env, ReferenceVariables (id,c)))
 
-let error_elim_arity env ind c pj okinds =
-  raise (TypeError (env, ElimArity (ind,c,pj,okinds)))
+let error_elim_arity env ind c okinds =
+  raise (TypeError (env, ElimArity (ind, c, okinds)))
 
 let error_case_not_inductive env j =
   raise (TypeError (env, CaseNotInductive j))
 
 let error_number_branches env cj expn =
-  raise (TypeError (env, NumberBranches (nfj env cj,expn)))
+  raise (TypeError (env, NumberBranches (cj, expn)))
 
 let error_ill_formed_branch env c i actty expty =
-  raise (TypeError (env,
-    IllFormedBranch (c,i,nf_betaiota env actty, nf_betaiota env expty)))
+  raise (TypeError (env, IllFormedBranch (c, i, actty, expty)))
 
 let error_generalization env nvar c =
   raise (TypeError (env, Generalization (nvar,c)))
@@ -140,13 +140,6 @@ let error_ill_formed_rec_body env why lna i fixenv vdefj =
 let error_ill_typed_rec_body env i lna vdefj vargs =
   raise (TypeError (env, IllTypedRecBody (i,lna,vdefj,vargs)))
 
-let error_elim_explain kp ki =
-  let open Sorts in
-  match kp,ki with
-  | (InType | InSet), InProp -> NonInformativeToInformative
-  | InType, InSet -> StrongEliminationOnNonSmallType (* if Set impredicative *)
-  | _ -> WrongArity
-
 let error_unsatisfied_constraints env c =
   raise (TypeError (env, UnsatisfiedConstraints c))
 
@@ -159,11 +152,20 @@ let error_disallowed_sprop env =
 let error_bad_relevance env =
   raise (TypeError (env, BadRelevance))
 
-let map_pguard_error f = function
+let error_bad_invert env =
+  raise (TypeError (env, BadInvert))
+
+let error_bad_variance env ~lev ~expected ~actual =
+  raise (TypeError (env, BadVariance {lev;expected;actual}))
+
+let map_pfix_guard_error f = function
 | NotEnoughAbstractionInFixBody -> NotEnoughAbstractionInFixBody
 | RecursionNotOnInductiveType c -> RecursionNotOnInductiveType (f c)
-| RecursionOnIllegalTerm (n, (env, c), l1, l2) -> RecursionOnIllegalTerm (n, (env, f c), l1, l2)
+| RecursionOnIllegalTerm (n, (env, c), l1_l2) -> RecursionOnIllegalTerm (n, (env, f c), l1_l2)
 | NotEnoughArgumentsForFixCall n -> NotEnoughArgumentsForFixCall n
+| FixpointOnIrrelevantInductive -> FixpointOnIrrelevantInductive
+
+let map_pcofix_guard_error f = function
 | CodomainNotInductiveType c -> CodomainNotInductiveType (f c)
 | NestedRecursiveOccurrences -> NestedRecursiveOccurrences
 | UnguardedRecursiveCall c -> UnguardedRecursiveCall (f c)
@@ -175,7 +177,10 @@ let map_pguard_error f = function
 | RecCallInCasePred c -> RecCallInCasePred (f c)
 | NotGuardedForm c -> NotGuardedForm (f c)
 | ReturnPredicateNotCoInductive c -> ReturnPredicateNotCoInductive (f c)
-| FixpointOnIrrelevantInductive -> FixpointOnIrrelevantInductive
+
+let map_pguard_error f = function
+| FixGuardError e -> FixGuardError (map_pfix_guard_error f e)
+| CoFixGuardError e -> CoFixGuardError (map_pcofix_guard_error f e)
 
 let map_ptype_error f = function
 | UnboundRel n -> UnboundRel n
@@ -183,7 +188,8 @@ let map_ptype_error f = function
 | NotAType j -> NotAType (on_judgment f j)
 | BadAssumption j -> BadAssumption (on_judgment f j)
 | ReferenceVariables (id, c) -> ReferenceVariables (id, c)
-| ElimArity (pi, c, j, ar) -> ElimArity (pi, f c, on_judgment f j, ar)
+| ElimArity (pi, c, ar) ->
+  ElimArity (pi, f c, Option.map (fun (j, s1, s2, s3) -> (on_judgment f j, s1, s2, s3)) ar)
 | CaseNotInductive j -> CaseNotInductive (on_judgment f j)
 | WrongCaseInfo (pi, ci) -> WrongCaseInfo (pi, ci)
 | NumberBranches (j, n) -> NumberBranches (on_judgment f j, n)
@@ -202,3 +208,5 @@ let map_ptype_error f = function
 | UndeclaredUniverse l -> UndeclaredUniverse l
 | DisallowedSProp -> DisallowedSProp
 | BadRelevance -> BadRelevance
+| BadInvert -> BadInvert
+| BadVariance u -> BadVariance u
