@@ -17,13 +17,16 @@ open Univ
 
 (*i Rem: NotEnoughAbstractionInFixBody should only occur with "/i" Fix
     notation i*)
-type 'constr pguard_error =
-  (** Fixpoints *)
+type 'constr pfix_guard_error =
+  (* Fixpoints *)
   | NotEnoughAbstractionInFixBody
   | RecursionNotOnInductiveType of 'constr
-  | RecursionOnIllegalTerm of int * (env * 'constr) * int list * int list
+  | RecursionOnIllegalTerm of int * (env * 'constr) * (int list * int list) Lazy.t
   | NotEnoughArgumentsForFixCall of int
-  (** CoFixpoints *)
+  | FixpointOnIrrelevantInductive
+
+type 'constr pcofix_guard_error =
+  (* CoFixpoints *)
   | CodomainNotInductiveType of 'constr
   | NestedRecursiveOccurrences
   | UnguardedRecursiveCall of 'constr
@@ -35,23 +38,23 @@ type 'constr pguard_error =
   | RecCallInCasePred of 'constr
   | NotGuardedForm of 'constr
   | ReturnPredicateNotCoInductive of 'constr
-  | FixpointOnIrrelevantInductive
 
+type 'constr pguard_error =
+  | FixGuardError of 'constr pfix_guard_error
+  | CoFixGuardError of 'constr pcofix_guard_error
+
+type fix_guard_error = constr pfix_guard_error
+type cofix_guard_error = constr pcofix_guard_error
 type guard_error = constr pguard_error
 
-type arity_error =
-  | NonInformativeToInformative
-  | StrongEliminationOnNonSmallType
-  | WrongArity
-
 type ('constr, 'types) ptype_error =
   | UnboundRel of int
   | UnboundVar of variable
   | NotAType of ('constr, 'types) punsafe_judgment
   | BadAssumption of ('constr, 'types) punsafe_judgment
   | ReferenceVariables of Id.t * GlobRef.t
-  | ElimArity of pinductive * 'constr * ('constr, 'types) punsafe_judgment
-      * (Sorts.family * Sorts.family * Sorts.family * arity_error) option
+  | ElimArity of pinductive * 'constr *
+      (('constr, 'types) punsafe_judgment * Sorts.family * Sorts.family * Sorts.family) option
   | CaseNotInductive of ('constr, 'types) punsafe_judgment
   | WrongCaseInfo of pinductive * case_info
   | NumberBranches of ('constr, 'types) punsafe_judgment * int
@@ -65,10 +68,12 @@ type ('constr, 'types) ptype_error =
   | IllFormedRecBody of 'constr pguard_error * Name.t Context.binder_annot array * int * env * ('constr, 'types) punsafe_judgment array
   | IllTypedRecBody of
       int * Name.t Context.binder_annot array * ('constr, 'types) punsafe_judgment array * 'types array
-  | UnsatisfiedConstraints of Constraint.t
+  | UnsatisfiedConstraints of Constraints.t
   | UndeclaredUniverse of Level.t
   | DisallowedSProp
   | BadRelevance
+  | BadInvert
+  | BadVariance of { lev : Level.t; expected : Variance.t; actual : Variance.t }
 
 type type_error = (constr, types) ptype_error
 
@@ -87,7 +92,7 @@ type inductive_error =
   | NotAnArity of env * constr
   | BadEntry
   | LargeNonPropInductiveNotInType
-  | MissingConstraints of (Universe.Set.t * Universe.t)
+  | MissingConstraints of (Sorts.t list * Sorts.t)
   (* each universe in the set should have been <= the other one *)
 
 exception InductiveError of inductive_error
@@ -105,8 +110,8 @@ val error_assumption : env -> unsafe_judgment -> 'a
 val error_reference_variables : env -> Id.t -> GlobRef.t -> 'a
 
 val error_elim_arity :
-  env -> pinductive -> constr -> unsafe_judgment ->
-      (Sorts.family * Sorts.family * Sorts.family * arity_error) option -> 'a
+  env -> pinductive -> constr ->
+    (unsafe_judgment * Sorts.family * Sorts.family * Sorts.family) option -> 'a
 
 val error_case_not_inductive : env -> unsafe_judgment -> 'a
 
@@ -133,9 +138,7 @@ val error_ill_formed_rec_body :
 val error_ill_typed_rec_body  :
   env -> int -> Name.t Context.binder_annot array -> unsafe_judgment array -> types array -> 'a
 
-val error_elim_explain : Sorts.family -> Sorts.family -> arity_error
-
-val error_unsatisfied_constraints : env -> Constraint.t -> 'a
+val error_unsatisfied_constraints : env -> Constraints.t -> 'a
 
 val error_undeclared_universe : env -> Level.t -> 'a
 
@@ -143,5 +146,9 @@ val error_disallowed_sprop : env -> 'a
 
 val error_bad_relevance : env -> 'a
 
+val error_bad_invert : env -> 'a
+
+val error_bad_variance : env -> lev:Level.t -> expected:Variance.t -> actual:Variance.t -> 'a
+
 val map_pguard_error : ('c -> 'd) -> 'c pguard_error -> 'd pguard_error
 val map_ptype_error : ('c -> 'd) -> ('c, 'c) ptype_error -> ('d, 'd) ptype_error
