@@ -28,12 +28,12 @@ module NamedDecl = Context.Named.Declaration
 
 exception NotConvertibleVect of int
 
-let conv_leq l2r env x y = default_conv CUMUL ~l2r env x y
+let conv_leq env x y = default_conv CUMUL env x y
 
 let conv_leq_vecti env v1 v2 =
   Array.fold_left2_i
     (fun i _ t1 t2 ->
-      try conv_leq false env t1 t2
+      try conv_leq env t1 t2
       with NotConvertible -> raise (NotConvertibleVect i))
     ()
     v1
@@ -77,6 +77,35 @@ let check_assumption env x t ty =
   in
   x
 
+let check_binding_relevance na1 na2 =
+  (* Since we know statically the relevance here, we are stricter *)
+  assert (Sorts.relevance_equal (binder_relevance na1) (binder_relevance na2))
+
+let esubst u s c =
+  Vars.esubst Vars.lift_substituend s (subst_instance_constr u c)
+
+exception ArgumentsMismatch
+
+let instantiate_context u subst nas ctx =
+  let open Context.Rel.Declaration in
+  let rec instantiate i ctx = match ctx with
+  | [] -> if 0 <= i then raise ArgumentsMismatch else []
+  | LocalAssum (na, ty) :: ctx ->
+    let ctx = instantiate (pred i) ctx in
+    let subst = Esubst.subs_liftn i subst in
+    let ty = esubst u subst ty in
+    let () = check_binding_relevance na nas.(i) in
+    LocalAssum (nas.(i), ty) :: ctx
+  | LocalDef (na, ty, bdy) :: ctx ->
+    let ctx = instantiate (pred i) ctx in
+    let subst = Esubst.subs_liftn i subst in
+    let ty = esubst u subst ty in
+    let bdy = esubst u subst bdy in
+    let () = check_binding_relevance na nas.(i) in
+    LocalDef (nas.(i), ty, bdy) :: ctx
+  in
+  instantiate (Array.length nas - 1) ctx
+
 (************************************************)
 (* Incremental typing rules: builds a typing judgment given the *)
 (* judgments for the subterms. *)
@@ -197,8 +226,8 @@ let type_of_apply env func funt argsv argstv =
         let arg = argsv.(i) in
         let argt = argstv.(i) in
         let c1 = term_of_fconstr c1 in
-        begin match conv_leq false env argt c1 with
-        | () -> apply_rec (i+1) (mk_clos (Esubst.subs_cons ([| inject arg |], e)) c2)
+        begin match conv_leq env argt c1 with
+        | () -> apply_rec (i+1) (mk_clos (CClosure.usubs_cons (inject arg) e) c2)
         | exception NotConvertible ->
           error_cant_apply_bad_type env
             (i+1,c1,argt)
@@ -212,10 +241,43 @@ let type_of_apply env func funt argsv argstv =
   in
   apply_rec 0 (inject funt)
 
+(* Checks that an array of terms has the type of a telescope. We assume that all
+   inputs are well-typed separately. *)
+let type_of_parameters env ctx u argsv argstv =
+  let open Context.Rel.Declaration in
+  let ctx = List.rev ctx in
+  let rec apply_rec i subst ctx = match ctx with
+  | [] -> if Int.equal i (Array.length argsv) then subst else raise ArgumentsMismatch
+  | LocalAssum (_, t) :: ctx ->
+    let arg = argsv.(i) in
+    let argt = argstv.(i) in
+    let t = esubst u subst t in
+    begin match conv_leq env argt t with
+    | () -> apply_rec (i + 1) (Esubst.subs_cons (Vars.make_substituend arg) subst) ctx
+    | exception NotConvertible ->
+      error_actual_type env (make_judge arg argt) t
+    end
+  | LocalDef (_, b, _) :: ctx ->
+    let b = esubst u subst b in
+    apply_rec i (Esubst.subs_cons (Vars.make_substituend b) subst) ctx
+  in
+  apply_rec 0 (Esubst.subs_id 0) ctx
+
 (* Type of primitive constructs *)
-let type_of_prim_type _env = function
-  | CPrimitives.PT_int63 -> Constr.mkSet
-  | CPrimitives.PT_float64 -> Constr.mkSet
+let type_of_prim_type _env u (type a) (prim : a CPrimitives.prim_type) = match prim with
+  | CPrimitives.PT_int63 ->
+    assert (Univ.Instance.is_empty u);
+    Constr.mkSet
+  | CPrimitives.PT_float64 ->
+    assert (Univ.Instance.is_empty u);
+    Constr.mkSet
+  | CPrimitives.PT_array ->
+    begin match Univ.Instance.to_array u with
+    | [|u|] ->
+      let ty = Constr.mkType (Univ.Universe.make u) in
+      Constr.mkProd(Context.anonR, ty , ty)
+    | _ -> anomaly Pp.(str"universe instance for array type should have length 1")
+    end
 
 let type_of_int env =
   match env.retroknowledge.Retroknowledge.retro_int63 with
@@ -225,74 +287,13 @@ let type_of_int env =
 let type_of_float env =
   match env.retroknowledge.Retroknowledge.retro_float64 with
   | Some c -> mkConst c
-  | None -> raise
-        (Invalid_argument "Typeops.type_of_float: float64 not_defined")
+  | None -> CErrors.user_err Pp.(str"The type float must be registered before this construction can be typechecked.")
 
-let type_of_prim env t =
-  let int_ty () = type_of_int env in
-  let float_ty () = type_of_float env in
-  let bool_ty () =
-    match env.retroknowledge.Retroknowledge.retro_bool with
-    | Some ((ind,_),_) -> Constr.mkInd ind
-    | None -> CErrors.user_err Pp.(str"The type bool must be registered before this primitive.")
-  in
-  let compare_ty () =
-    match env.retroknowledge.Retroknowledge.retro_cmp with
-    | Some ((ind,_),_,_) -> Constr.mkInd ind
-    | None -> CErrors.user_err Pp.(str"The type compare must be registered before this primitive.")
-  in
-  let f_compare_ty () =
-    match env.retroknowledge.Retroknowledge.retro_f_cmp with
-    | Some ((ind,_),_,_,_) -> Constr.mkInd ind
-    | None -> CErrors.user_err Pp.(str"The type float_comparison must be registered before this primitive.")
-  in
-  let f_class_ty () =
-    match env.retroknowledge.Retroknowledge.retro_f_class with
-    | Some ((ind,_),_,_,_,_,_,_,_,_) -> Constr.mkInd ind
-    | None -> CErrors.user_err Pp.(str"The type float_class must be registered before this primitive.")
-  in
-  let pair_ty fst_ty snd_ty =
-    match env.retroknowledge.Retroknowledge.retro_pair with
-    | Some (ind,_) -> Constr.mkApp(Constr.mkInd ind, [|fst_ty;snd_ty|])
-    | None -> CErrors.user_err Pp.(str"The type pair must be registered before this primitive.")
-  in
-  let carry_ty int_ty =
-    match env.retroknowledge.Retroknowledge.retro_carry with
-    | Some ((ind,_),_) -> Constr.mkApp(Constr.mkInd ind, [|int_ty|])
-    | None -> CErrors.user_err Pp.(str"The type carry must be registered before this primitive.")
-  in
-  let open CPrimitives in
-  let tr_prim_type = function
-    | PT_int63 -> int_ty ()
-    | PT_float64 -> float_ty () in
-  let tr_ind (type t) (i : t prim_ind) (a : t) = match i, a with
-    | PIT_bool, () -> bool_ty ()
-    | PIT_carry, t -> carry_ty (tr_prim_type t)
-    | PIT_pair, (t1, t2) -> pair_ty (tr_prim_type t1) (tr_prim_type t2)
-    | PIT_cmp, () -> compare_ty ()
-    | PIT_f_cmp, () -> f_compare_ty ()
-    | PIT_f_class, () -> f_class_ty () in
-  let tr_type = function
-    | PITT_ind (i, a) -> tr_ind i a
-    | PITT_type t -> tr_prim_type t in
-  let rec nary_op = function
-    | [] -> assert false
-    | [ret_ty] -> tr_type ret_ty
-    | arg_ty :: r ->
-       let arg_ty = tr_type arg_ty in
-       Constr.mkProd(Context.nameR (Id.of_string "x"), arg_ty, nary_op r) in
-  nary_op (types t)
-
-let type_of_prim_or_type env = let open CPrimitives in
-  function
-  | OT_type t -> type_of_prim_type env t
-  | OT_op op -> type_of_prim env op
-
-let judge_of_int env i =
-  make_judge (Constr.mkInt i) (type_of_int env)
-
-let judge_of_float env f =
-  make_judge (Constr.mkFloat f) (type_of_float env)
+let type_of_array env u =
+  assert (Univ.Instance.length u = 1);
+  match env.retroknowledge.Retroknowledge.retro_array with
+  | Some c -> mkConstU (c,u)
+  | None -> CErrors.user_err Pp.(str"The type array must be registered before this construction can be typechecked.")
 
 (* Type of product *)
 
@@ -345,15 +346,25 @@ let check_cast env c ct k expected_type =
     | VMcast ->
       Vconv.vm_conv CUMUL env ct expected_type
     | DEFAULTcast ->
-      default_conv ~l2r:false CUMUL env ct expected_type
-    | REVERTcast ->
-      default_conv ~l2r:true CUMUL env ct expected_type
+      default_conv CUMUL env ct expected_type
     | NATIVEcast ->
-      let sigma = Nativelambda.empty_evars in
+      let sigma = Genlambda.empty_evars in
       Nativeconv.native_conv CUMUL sigma env ct expected_type
   with NotConvertible ->
     error_actual_type env (make_judge c ct) expected_type
 
+let judge_of_int env i =
+  make_judge (Constr.mkInt i) (type_of_int env)
+
+let judge_of_float env f =
+  make_judge (Constr.mkFloat f) (type_of_float env)
+
+let judge_of_array env u tj defj =
+  let def = defj.uj_val in
+  let ty = defj.uj_type in
+  Array.iter (fun j -> check_cast env j.uj_val j.uj_type DEFAULTcast ty) tj;
+  make_judge (mkArray(u, Array.map j_val tj, def, ty)) (mkApp (type_of_array env u, [|ty|]))
+
 (* Inductive types. *)
 
 (* The type is parametric over the uniform parameters whose conclusion
@@ -387,40 +398,111 @@ let type_of_inductive env (ind,u) =
 (* Constructors. *)
 
 let type_of_constructor env (c,_u as cu) =
-  let () =
-    let ((kn,_),_) = c in
-    let mib = lookup_mind kn env in
-    check_hyps_inclusion env (GlobRef.ConstructRef c) mib.mind_hyps
-  in
-  let specif = lookup_mind_specif env (inductive_of_constructor c) in
+  let (mib, _ as specif) = lookup_mind_specif env (inductive_of_constructor c) in
+  let () = check_hyps_inclusion env (GlobRef.ConstructRef c) mib.mind_hyps in
   let t,cst = constrained_type_of_constructor cu specif in
   let () = check_constraints cst env in
   t
 
 (* Case. *)
 
-let check_branch_types env (ind,u) c ct lft explft =
-  try conv_leq_vecti env lft explft
-  with
-      NotConvertibleVect i ->
-        error_ill_formed_branch env c ((ind,i+1),u) lft.(i) explft.(i)
-    | Invalid_argument _ ->
-        error_number_branches env (make_judge c ct) (Array.length explft)
+exception NotConvertibleBranch of int * rel_context * types * types
 
-let type_of_case env ci p pt c ct _lf lft =
-  let (pind, _ as indspec) =
+let check_branch_types env (_mib, mip) ci u pms c _ct lft (pctx, p) =
+  let open Context.Rel.Declaration in
+  let rec instantiate ctx args subst = match ctx, args with
+  | [], [] -> subst
+  | LocalAssum _ :: ctx, a :: args ->
+    let subst = Esubst.subs_cons (Vars.make_substituend a) subst in
+    instantiate ctx args subst
+  | LocalDef (_, a, _) :: ctx, args ->
+    let a = Vars.esubst Vars.lift_substituend subst a in
+    let subst = Esubst.subs_cons (Vars.make_substituend a) subst in
+    instantiate ctx args subst
+  | _ -> assert false
+  in
+  let iter i (brctx, brt, constrty) =
+    let brenv = push_rel_context brctx env in
+    let nargs = List.length brctx in
+    let pms = Array.map (fun c -> lift nargs c) pms in
+    let cargs = Context.Rel.instance mkRel 0 brctx in
+    let cstr = mkApp (mkConstructU ((ci.ci_ind, i + 1), u), Array.append pms cargs) in
+    let (_, retargs) = find_rectype brenv constrty in
+    let indices = List.lastn mip.mind_nrealargs retargs in
+    let subst = instantiate (List.rev pctx) (indices @ [cstr]) (Esubst.subs_shft (nargs, Esubst.subs_id 0)) in
+    let expbrt = Vars.esubst Vars.lift_substituend subst p in
+    try conv_leq brenv brt expbrt
+    with NotConvertible -> raise (NotConvertibleBranch (i, brctx, brt, expbrt))
+  in
+  try Array.iteri iter lft
+  with NotConvertibleBranch (i, brctx, brt, expbrt) ->
+    let brt = it_mkLambda_or_LetIn brt brctx in
+    let expbrt = it_mkLambda_or_LetIn expbrt brctx in
+    error_ill_formed_branch env c ((ci.ci_ind, i + 1), u) brt expbrt
+
+let should_invert_case env ci =
+  ci.ci_relevance == Sorts.Relevant &&
+  let mib,mip = lookup_mind_specif env ci.ci_ind in
+  mip.mind_relevance == Sorts.Irrelevant &&
+  (* NB: it's possible to have 2 ctors or arguments to 1 ctor by unsetting univ checks
+     but we don't do special reduction in such cases
+
+     XXX Someday consider more carefully what happens with letin params and arguments
+     (currently they're squashed, see indtyping)
+ *)
+  Array.length mip.mind_nf_lc = 1 &&
+  List.length (fst mip.mind_nf_lc.(0)) = List.length mib.mind_params_ctxt
+
+let type_case_scrutinee env (mib, _mip) (u', largs) u pms (pctx, p) c =
+  let (params, realargs) = List.chop mib.mind_nparams largs in
+  (* Check that the type of the scrutinee is <= the expected argument type *)
+  let () = Array.iter2 (fun p1 p2 -> Reduction.conv ~l2r:true env p1 p2) (Array.of_list params) pms in
+  (* We use l2r:true for compat with old versions which used CONV with arguments
+     flipped. It is relevant for performance eg in bedrock / Kami. *)
+  let cst = match mib.mind_variance with
+  | None -> Univ.enforce_eq_instances u u' Univ.Constraints.empty
+  | Some variance -> Univ.enforce_leq_variance_instances variance u' u Univ.Constraints.empty
+  in
+  let () = check_constraints cst env in
+  let subst = Vars.subst_of_rel_context_instance_list pctx (realargs @ [c]) in
+  Vars.substl subst p
+
+let type_of_case env (mib, mip) ci u pms (pctx, p, pt) iv c ct _lf lft =
+  let ((ind, u'), largs) =
     try find_rectype env ct
     with Not_found -> error_case_not_inductive env (make_judge c ct) in
-  let _, sp = try dest_arity env pt
-    with NotArity -> error_elim_arity env pind c (make_judge p pt) None in
+  (* Various well-formedness conditions *)
+  let sp = match destSort (whd_all (push_rel_context pctx env) pt) with
+  | sp -> sp
+  | exception DestKO ->
+    error_elim_arity env (ind, u') c None
+  in
   let rp = Sorts.relevance_of_sort sp in
   let ci = if ci.ci_relevance == rp then ci
     else (warn_bad_relevance_ci (); {ci with ci_relevance=rp})
   in
-  let () = check_case_info env pind rp ci in
-  let (bty,rslty) =
-    type_case_branches env indspec (make_judge p pt) c in
-  let () = check_branch_types env pind c ct lft bty in
+  let () = check_case_info env (ind, u') rp ci in
+  let () =
+    let is_inversion = match iv with
+      | NoInvert -> false
+      | CaseInvert _ -> true (* contents already checked *)
+    in
+    if not (is_inversion = should_invert_case env ci)
+    then error_bad_invert env
+  in
+  let () =
+    let ksort = Sorts.family sp in
+    if not (Sorts.family_leq ksort mip.mind_kelim) then
+      let s = inductive_sort_family mip in
+      let pj = make_judge (it_mkLambda_or_LetIn p pctx) (it_mkProd_or_LetIn pt pctx) in
+      let kinds = Some (pj, mip.mind_kelim, ksort, s) in
+      error_elim_arity env (ind, u') c kinds
+  in
+  (* Check that the scrutinee has the right type *)
+  let rslty = type_case_scrutinee env (mib, mip) (u', largs) u pms (pctx, p) c in
+  (* We return the "higher" inductive universe instance from the predicate,
+     the branches must be typeable using these universes. *)
+  let () = check_branch_types env (mib, mip) ci u pms c ct lft (pctx, p) in
   ci, rslty
 
 let type_of_projection env p c ct =
@@ -429,7 +511,7 @@ let type_of_projection env p c ct =
     try find_rectype env ct
     with Not_found -> error_case_not_inductive env (make_judge c ct)
   in
-  assert(eq_ind (Projection.inductive p) ind);
+  assert(Ind.CanOrd.equal (Projection.inductive p) ind);
   let ty = Vars.subst_instance_constr u pty in
   substl (c :: CList.rev args) ty
 
@@ -452,7 +534,7 @@ let check_fixpoint env lna lar vdef vdeft =
 let type_of_global_in_context env r =
   let open Names.GlobRef in
   match r with
-  | VarRef id -> Environ.named_type id env, Univ.AUContext.empty
+  | VarRef id -> Environ.named_type id env, Univ.AbstractContext.empty
   | ConstRef c ->
     let cb = Environ.lookup_constant c env in
     let univs = Declareops.constant_polymorphic_context cb in
@@ -564,13 +646,70 @@ let rec execute env cstr =
     | Construct c ->
       cstr, type_of_constructor env c
 
-    | Case (ci,p,c,lf) ->
+    | Case (ci, u, pms, p, iv, c, lf) ->
         let c', ct = execute env c in
-        let p', pt = execute env p in
-        let lf', lft = execute_array env lf in
-        let ci', t = type_of_case env ci p' pt c' ct lf' lft in
-        let cstr = if ci == ci' && c == c' && p == p' && lf == lf' then cstr
-          else mkCase(ci',p',c',lf')
+        let iv' = match iv with
+          | NoInvert -> NoInvert
+          | CaseInvert {indices} ->
+            let args = Array.append pms indices in
+            let ct' = mkApp (mkIndU (ci.ci_ind,u), args) in
+            let (ct', _) : constr * Sorts.t = execute_is_type env ct' in
+            let () = conv_leq env ct ct' in
+            let _, args' = decompose_appvect ct' in
+            if args == args' then iv
+            else CaseInvert {indices=Array.sub args' (Array.length pms) (Array.length indices)}
+        in
+        let mib, mip = Inductive.lookup_mind_specif env ci.ci_ind in
+        let cst = Inductive.instantiate_inductive_constraints mib u in
+        let () = check_constraints cst env in
+        let pms', pmst = execute_array env pms in
+        let paramsubst =
+          try type_of_parameters env mib.mind_params_ctxt u pms' pmst
+          with ArgumentsMismatch -> error_elim_arity env (ci.ci_ind, u) c' None
+        in
+        let (pctx, p', pt) =
+          let (nas, p) = p in
+          let realdecls, _ = List.chop mip.mind_nrealdecls mip.mind_arity_ctxt in
+          let self =
+            let args = Context.Rel.instance mkRel 0 mip.mind_arity_ctxt in
+            let inst = Instance.of_array (Array.init (Instance.length u) Level.var) in
+            mkApp (mkIndU (ci.ci_ind, inst), args)
+          in
+          let realdecls = LocalAssum (Context.make_annot Anonymous mip.mind_relevance, self) :: realdecls in
+          let realdecls =
+            try instantiate_context u paramsubst nas realdecls
+            with ArgumentsMismatch -> error_elim_arity env (ci.ci_ind, u) c' None
+          in
+          let p_env = Environ.push_rel_context realdecls env in
+          let p', pt = execute p_env p in
+          (realdecls, p', pt)
+        in
+        let () =
+          let nbranches = Array.length mip.mind_nf_lc in
+          if not (Int.equal (Array.length lf) nbranches) then
+            error_number_branches env (make_judge c ct) nbranches
+        in
+        let lft = Array.make (Array.length lf) ([], mkProp, mkProp) in
+        let build_one_branch i (nas, br as b) =
+          let (ctx, cty) = mip.mind_nf_lc.(i) in
+          let ctx, _ = List.chop mip.mind_consnrealdecls.(i) ctx in
+          let ctx =
+            try instantiate_context u paramsubst nas ctx
+            with ArgumentsMismatch ->
+              (* Despite the name, the toplevel message is reasonable *)
+              error_elim_arity env (ci.ci_ind, u) c' None
+          in
+          let br_env = Environ.push_rel_context ctx env in
+          let br', brt = execute br_env br in
+          let cty = esubst u (Esubst.subs_liftn mip.mind_consnrealdecls.(i) paramsubst) cty in
+          let () = lft.(i) <- (ctx, brt, cty) in
+          if br == br' then b else (nas, br')
+        in
+        let lf' = Array.Smart.map_i build_one_branch lf in
+        let ci', t = type_of_case env (mib, mip) ci u pms' (pctx, p', pt) iv' c' ct lf' lft in
+        let eqbr (_, br1) (_, br2) = br1 == br2 in
+        let cstr = if ci == ci' && pms == pms' && c == c' && snd p == p' && iv == iv' && Array.equal eqbr lf lf' then cstr
+          else mkCase (ci', u, pms', (fst p, p'), iv', c', lf')
         in
         cstr, t
 
@@ -591,6 +730,23 @@ let rec execute env cstr =
     (* Primitive types *)
     | Int _ -> cstr, type_of_int env
     | Float _ -> cstr, type_of_float env
+    | Array(u,t,def,ty) ->
+      (* ty : Type@{u} and all of t,def : ty *)
+      let ulev = match Univ.Instance.to_array u with
+        | [|u|] -> u
+        | _ -> assert false
+      in
+      let ty',tyty = execute env ty in
+      check_cast env ty' tyty DEFAULTcast (mkType (Universe.make ulev));
+      let def', def_ty = execute env def in
+      check_cast env def' def_ty DEFAULTcast ty';
+      let ta = type_of_array env u in
+      let t' = Array.Smart.map (fun x ->
+        let x', xt = execute env x in
+        check_cast env x' xt DEFAULTcast ty';
+        x') t in
+      let cstr = if def'==def && t'==t && ty'==ty then cstr else mkArray(u, t',def',ty') in
+      cstr, mkApp(ta, [|ty'|])
 
     (* Partial proofs: unsupported by the kernel *)
     | Meta _ ->
@@ -630,12 +786,6 @@ let infer env constr =
   let constr, t = execute env constr in
   make_judge constr t
 
-let infer =
-  if Flags.profile then
-    let infer_key = CProfile.declare_profile "Fast_infer" in
-      CProfile.profile2 infer_key (fun b c -> infer b c)
-  else (fun b c -> infer b c)
-
 let assumption_of_judgment env {uj_val=c; uj_type=t} =
   infer_assumption env c t
 
@@ -662,7 +812,7 @@ let check_context env rels =
       | LocalDef (x,bd,ty) ->
         let j1 = infer env bd in
         let jty = infer_type env ty in
-        conv_leq false env j1.uj_type ty;
+        conv_leq env j1.uj_type ty;
         let x = check_binder_annot jty.utj_type x in
         push_rel d env, LocalDef (x,j1.uj_val,jty.utj_val) :: rels)
     rels ~init:(env,[])
@@ -701,7 +851,7 @@ let judge_of_apply env funj argjv =
 
 let judge_of_cast env cj k tj =
   let () = check_cast env cj.uj_val cj.uj_type k tj.utj_val in
-  let c = match k with | REVERTcast -> cj.uj_val | _ -> mkCast (cj.uj_val, k, tj.utj_val) in
+  let c = mkCast (cj.uj_val, k, tj.utj_val) in
   make_judge c tj.utj_val
 
 let judge_of_inductive env indu =
@@ -710,14 +860,80 @@ let judge_of_inductive env indu =
 let judge_of_constructor env cu =
   make_judge (mkConstructU cu) (type_of_constructor env cu)
 
-let judge_of_case env ci pj cj lfj =
-  let lf, lft = dest_judgev lfj in
-  let ci, t = type_of_case env ci pj.uj_val pj.uj_type cj.uj_val cj.uj_type lf lft in
-  make_judge (mkCase (ci, (*nf_betaiota*) pj.uj_val, cj.uj_val, lft)) t
-
 (* Building type of primitive operators and type *)
 
-let check_primitive_type env op_t t =
-  let inft = type_of_prim_or_type env op_t in
-  try default_conv ~l2r:false CUMUL env inft t
-  with NotConvertible -> error_incorrect_primitive env (make_judge op_t inft) t
+let type_of_prim_const env _u c =
+  let int_ty () = type_of_int env in
+  match c with
+  | CPrimitives.Arraymaxlength ->
+    int_ty ()
+
+let type_of_prim env u t =
+  let int_ty () = type_of_int env in
+  let float_ty () = type_of_float env in
+  let array_ty u a = mkApp(type_of_array env u, [|a|]) in
+  let bool_ty () =
+    match env.retroknowledge.Retroknowledge.retro_bool with
+    | Some ((ind,_),_) -> Constr.mkInd ind
+    | None -> CErrors.user_err Pp.(str"The type bool must be registered before this primitive.")
+  in
+  let compare_ty () =
+    match env.retroknowledge.Retroknowledge.retro_cmp with
+    | Some ((ind,_),_,_) -> Constr.mkInd ind
+    | None -> CErrors.user_err Pp.(str"The type compare must be registered before this primitive.")
+  in
+  let f_compare_ty () =
+    match env.retroknowledge.Retroknowledge.retro_f_cmp with
+    | Some ((ind,_),_,_,_) -> Constr.mkInd ind
+    | None -> CErrors.user_err Pp.(str"The type float_comparison must be registered before this primitive.")
+  in
+  let f_class_ty () =
+    match env.retroknowledge.Retroknowledge.retro_f_class with
+    | Some ((ind,_),_,_,_,_,_,_,_,_) -> Constr.mkInd ind
+    | None -> CErrors.user_err Pp.(str"The type float_class must be registered before this primitive.")
+  in
+  let pair_ty fst_ty snd_ty =
+    match env.retroknowledge.Retroknowledge.retro_pair with
+    | Some (ind,_) -> Constr.mkApp(Constr.mkInd ind, [|fst_ty;snd_ty|])
+    | None -> CErrors.user_err Pp.(str"The type pair must be registered before this primitive.")
+  in
+  let carry_ty int_ty =
+    match env.retroknowledge.Retroknowledge.retro_carry with
+    | Some ((ind,_),_) -> Constr.mkApp(Constr.mkInd ind, [|int_ty|])
+    | None -> CErrors.user_err Pp.(str"The type carry must be registered before this primitive.")
+  in
+  let open CPrimitives in
+  let tr_prim_type (tr_type : ind_or_type -> constr) (type a) (ty : a prim_type) (t : a) = match ty with
+    | PT_int63 -> int_ty t
+    | PT_float64 -> float_ty t
+    | PT_array -> array_ty (fst t) (tr_type (snd t))
+  in
+  let tr_ind (tr_type : ind_or_type -> constr) (type t) (i : t prim_ind) (a : t) = match i, a with
+    | PIT_bool, () -> bool_ty ()
+    | PIT_carry, t -> carry_ty (tr_type t)
+    | PIT_pair, (t1, t2) -> pair_ty (tr_type t1) (tr_type t2)
+    | PIT_cmp, () -> compare_ty ()
+    | PIT_f_cmp, () -> f_compare_ty ()
+    | PIT_f_class, () -> f_class_ty ()
+  in
+  let rec tr_type n = function
+    | PITT_ind (i, a) -> tr_ind (tr_type n) i a
+    | PITT_type (ty,t) -> tr_prim_type (tr_type n) ty t
+    | PITT_param i -> Constr.mkRel (n+i)
+  in
+  let rec nary_op n ret_ty = function
+    | [] -> tr_type n ret_ty
+    | arg_ty :: r ->
+        Constr.mkProd (Context.nameR (Id.of_string "x"),
+                       tr_type n arg_ty, nary_op (n + 1) ret_ty r)
+  in
+  let params, args_ty, ret_ty = types t in
+  assert (AbstractContext.size (univs t) = Instance.length u);
+  Vars.subst_instance_constr u
+    (Term.it_mkProd_or_LetIn (nary_op 0 ret_ty args_ty) params)
+
+let type_of_prim_or_type env u = let open CPrimitives in
+  function
+  | OT_type t -> type_of_prim_type env u t
+  | OT_op op -> type_of_prim env u op
+  | OT_const c -> type_of_prim_const env u c
