@@ -92,14 +92,9 @@ val judge_of_cast :
 val judge_of_inductive : env -> inductive puniverses -> unsafe_judgment
 val judge_of_constructor : env -> constructor puniverses -> unsafe_judgment
 
-(** {6 Type of Cases. } *)
-val judge_of_case : env -> case_info
-  -> unsafe_judgment -> unsafe_judgment -> unsafe_judgment array
-    -> unsafe_judgment
-
 (** {6 Type of global references. } *)
 
-val type_of_global_in_context : env -> GlobRef.t -> types * Univ.AUContext.t
+val type_of_global_in_context : env -> GlobRef.t -> types * Univ.AbstractContext.t
 (** Returns the type of the global reference, by creating a fresh
     instance of polymorphic references and computing their
     instantiated universe context. The type should not be used
@@ -110,11 +105,9 @@ val type_of_global_in_context : env -> GlobRef.t -> types * Univ.AUContext.t
 (** {6 Miscellaneous. } *)
 
 (** Check that hyps are included in env and fails with error otherwise *)
-val check_hyps_inclusion : env -> ?evars:((existential->constr option) * UGraph.t) ->
+val check_hyps_inclusion : env -> ?evars:constr evar_handler ->
   GlobRef.t -> Constr.named_context -> unit
 
-val check_primitive_type : env -> CPrimitives.op_or_type -> types -> unit
-
 (** Types for primitives *)
 
 val type_of_int : env -> types
@@ -123,8 +116,16 @@ val judge_of_int : env -> Uint63.t -> unsafe_judgment
 val type_of_float : env -> types
 val judge_of_float : env -> Float64.t -> unsafe_judgment
 
-val type_of_prim_type : env -> CPrimitives.prim_type -> types
-val type_of_prim : env -> CPrimitives.t -> types
+val type_of_array : env -> Univ.Instance.t -> types
+val judge_of_array : env -> Univ.Instance.t -> unsafe_judgment array -> unsafe_judgment -> unsafe_judgment
+
+val type_of_prim_type : env -> Univ.Instance.t -> 'a CPrimitives.prim_type -> types
+val type_of_prim : env -> Univ.Instance.t -> CPrimitives.t -> types
+val type_of_prim_or_type : env -> Univ.Instance.t -> CPrimitives.op_or_type -> types
 
 val warn_bad_relevance_name : string
 (** Allow the checker to make this warning into an error. *)
+
+val should_invert_case : env -> case_info -> bool
+(** We have case inversion exactly when going from irrelevant nonempty
+   (ie 1 constructor) inductive to relevant type. *)
