@@ -15,6 +15,6 @@ open Reduction
   s conversion functions *)
 val vm_conv : conv_pb -> types kernel_conversion_function
 
-(** A conversion function parametrized by a universe comparator. Used outside of
-    the kernel. *)
-val vm_conv_gen : conv_pb -> (types, 'a) generic_conversion_function
+(** A conversion function parametrized by a universe comparator and
+   evar normalizer. Used outside of the kernel. *)
+val vm_conv_gen : conv_pb -> Genlambda.evars -> (types, 'a) generic_conversion_function
