@@ -7,6 +7,8 @@
 (*         *     GNU Lesser General Public License Version 2.1          *)
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
+
+open Values
 open Vmvalues
 
 external set_drawinstr : unit -> unit = "coq_set_drawinstr"
@@ -38,13 +40,7 @@ external push_val : values -> unit = "coq_push_val"
 external push_arguments : arguments -> unit = "coq_push_arguments"
 external push_vstack : vstack -> int -> unit = "coq_push_vstack"
 
-
-(* interpreteur *)
-external coq_interprete : tcode -> values -> atom array -> vm_global -> vm_env -> int -> values =
-  "coq_interprete_byte" "coq_interprete_ml"
-
-let interprete code v env k =
-  coq_interprete code v (get_atom_rel ()) (Csymtable.get_global_data ()) env k
+let interprete = Vmsymtable.vm_interp
 
 (* Functions over arguments *)
 
@@ -95,7 +91,7 @@ let reduce_fix k vf =
   (* computing types *)
   let fc_typ = fix_types fb in
   let ndef = Array.length fc_typ in
-  let et = offset_closure_fix fb (2*(ndef - 1)) in
+  let et = fix_env fb in
   let ftyp =
     Array.map
       (fun c -> interprete c crazy_val et 0) fc_typ in
@@ -169,7 +165,7 @@ let rec apply_stack a stk v =
 let apply_whd k whd =
   let v = val_of_rel k in
   match whd with
-  | Vprod _ | Vconstr_const _ | Vconstr_block _ | Vint64 _ | Vfloat64 _ ->
+  | Vprod _ | Vconst _ | Vblock _ | Vint64 _ | Vfloat64 _ | Varray _ ->
      assert false
   | Vfun f -> reduce_fun k f
   | Vfix(f, None) ->
@@ -185,7 +181,6 @@ let apply_whd k whd =
       push_ra stop;
       push_val v;
       interprete (cofix_upd_code to_up) (cofix_upd_val to_up) (cofix_upd_env to_up) 0
-  | Vatom_stk(a,stk) ->
+  | Vaccu (a, stk) ->
       apply_stack (val_of_atom a) stk v
-  | Vuniv_level _lvl -> assert false
 
