@@ -9,7 +9,7 @@
 (************************************************************************)
 open Names
 open Univ
-open Constr
+open Values
 
 (********************************************)
 (* Initialization of the abstract machine ***)
@@ -35,8 +35,6 @@ let crazy_val = (val_of_obj (Obj.repr 0))
 
 type tag = int
 
-let accu_tag = 0
-
 let type_atom_tag = 2
 let max_atom_tag = 2
 let proj_tag = 3
@@ -61,8 +59,9 @@ type structured_constant =
 
 type reloc_table = (tag * int) array
 
+(** When changing this, adapt Is_tailrec_switch in coq_values.h accordingly *)
 type annot_switch =
-   {ci : case_info; rtbl : reloc_table; tailcall : bool; max_stack_size : int}
+   { rtbl : reloc_table; tailcall : bool; max_stack_size : int }
 
 let rec eq_structured_values v1 v2 =
   v1 == v2 ||
@@ -98,7 +97,7 @@ let hash_structured_values (v : structured_values) =
 let eq_structured_constant c1 c2 = match c1, c2 with
 | Const_sort s1, Const_sort s2 -> Sorts.equal s1 s2
 | Const_sort _, _ -> false
-| Const_ind i1, Const_ind i2 -> eq_ind i1 i2
+| Const_ind i1, Const_ind i2 -> Ind.CanOrd.equal i1 i2
 | Const_ind _, _ -> false
 | Const_b0 t1, Const_b0 t2 -> Int.equal t1 t2
 | Const_b0 _, _ -> false
@@ -115,7 +114,7 @@ let hash_structured_constant c =
   let open Hashset.Combine in
   match c with
   | Const_sort s -> combinesmall 1 (Sorts.hash s)
-  | Const_ind i -> combinesmall 2 (ind_hash i)
+  | Const_ind i -> combinesmall 2 (Ind.CanOrd.hash i)
   | Const_b0 t -> combinesmall 3 (Int.hash t)
   | Const_univ_level l -> combinesmall 4 (Univ.Level.hash l)
   | Const_val v -> combinesmall 5 (hash_structured_values v)
@@ -123,22 +122,16 @@ let hash_structured_constant c =
   | Const_float f -> combinesmall 7 (Float64.hash f)
 
 let eq_annot_switch asw1 asw2 =
-  let eq_ci ci1 ci2 =
-    eq_ind ci1.ci_ind ci2.ci_ind &&
-    Int.equal ci1.ci_npar ci2.ci_npar &&
-    CArray.equal Int.equal ci1.ci_cstr_ndecls ci2.ci_cstr_ndecls
-  in
   let eq_rlc (i1, j1) (i2, j2) = Int.equal i1 i2 && Int.equal j1 j2 in
-  eq_ci asw1.ci asw2.ci &&
   CArray.equal eq_rlc asw1.rtbl asw2.rtbl &&
-  (asw1.tailcall : bool) == asw2.tailcall
+  (asw1.tailcall : bool) == asw2.tailcall &&
+  Int.equal asw1.max_stack_size asw2.max_stack_size
 
 let hash_annot_switch asw =
   let open Hashset.Combine in
-  let h1 = Constr.case_info_hash asw.ci in
-  let h2 = Array.fold_left (fun h (t, i) -> combine3 h t i) 0 asw.rtbl in
-  let h3 = if asw.tailcall then 1 else 0 in
-  combine3 h1 h2 h3
+  let h1 = Array.fold_left (fun h (t, i) -> combine3 h t i) 0 asw.rtbl in
+  let h2 = if asw.tailcall then 1 else 0 in
+  combine3 h1 h2 asw.max_stack_size
 
 let pp_sort s =
   let open Sorts in
@@ -146,7 +139,7 @@ let pp_sort s =
   | SProp -> Pp.str "SProp"
   | Prop -> Pp.str "Prop"
   | Set -> Pp.str "Set"
-  | Type u -> Pp.(str "Type@{" ++ Univ.pr_uni u ++ str "}")
+  | Type u -> Pp.(str "Type@{" ++ Univ.Universe.pr u ++ str "}")
 
 let pp_struct_const = function
   | Const_sort s -> pp_sort s
@@ -170,17 +163,14 @@ let fix_val v = (Obj.magic v : values)
 let cofix_upd_val v = (Obj.magic v : values)
 
 type vm_env
-type vm_global
+let inj_env v = (Obj.magic v : vm_env)
 let fun_env v = (Obj.magic v : vm_env)
-let fix_env v = (Obj.magic v : vm_env)
 let cofix_env v = (Obj.magic v : vm_env)
 let cofix_upd_env v = (Obj.magic v : vm_env)
 type vstack = values array
 
 let fun_of_val v = (Obj.magic v : vfun)
 
-let vm_global (v : values array) = (Obj.magic v : vm_global)
-
 (*******************************************)
 (* Machine code *** ************************)
 (*******************************************)
@@ -213,13 +203,13 @@ type vswitch = {
 (*             dom : values, codom : vfun                                 *)
 (*                                                                        *)
 (* + Functions have two representations :                                 *)
-(*   - unapplied fun :  vf = Ct_[ C | fv1 | ... | fvn]                    *)
+(*   - unapplied fun :  vf = Ct_[ C | envofs=2 | fv1 | ... | fvn]         *)
 (*                                       C:tcode, fvi : values            *)
 (*     Remark : a function and its environment is the same value.         *)
-(*   - partially applied fun : Ct_[Restart:C| vf | arg1 | ... argn]       *)
+(*   - partially applied fun : Ct_[ Restart::C | envofs=2 | vf | arg1 | ... | argn] *)
 (*                                                                        *)
 (* + Fixpoints :                                                          *)
-(*   -        Ct_[C1|Infix_t|C2|...|Infix_t|Cn|fv1|...|fvn]               *)
+(*   - Ct_[C1|envofs1=3*n-1 | Infix_t|C2|envofs2 | ... | Infix_t|Cn|envofsn=2 | fv1|...|fvn] *)
 (*     One single block to represent all of the fixpoints, each fixpoint  *)
 (*     is the pointer to the field holding the pointer to its code, and   *)
 (*     the infix tag is used to know where the block starts.              *)
@@ -232,13 +222,12 @@ type vswitch = {
 (* + Cofixpoints : see cbytegen.ml                                        *)
 (*                                                                        *)
 (* + vblock's encode (non constant) constructors as in Ocaml, but         *)
-(*   starting from 0 up. tag 0 ( = accu_tag) is reserved for              *)
-(*   accumulators.                                                        *)
+(*   starting from 0 up.                                                  *)
 (*                                                                        *)
 (* + vm_env is the type of the machine environments (i.e. a function or   *)
 (*   a fixpoint)                                                          *)
 (*                                                                        *)
-(* + Accumulators : At_[accumulate| accu | arg1 | ... | argn ]            *)
+(* + Accumulators : At_[accumulate | envofs=2 | accu | arg1 | ... | argn ] *)
 (*   - representation of [accu] : tag_[....]                              *)
 (*     -- tag <= 3 : encoding atom type (sorts, free vars, etc.)          *)
 (*     -- 10_[accu|proj name] : a projection blocked by an accu           *)
@@ -260,7 +249,7 @@ type id_key =
 | EvarKey of Evar.t
 
 let eq_id_key (k1 : id_key) (k2 : id_key) = match k1, k2 with
-| ConstKey c1, ConstKey c2 -> Constant.equal c1 c2
+| ConstKey c1, ConstKey c2 -> Constant.CanOrd.equal c1 c2
 | VarKey id1, VarKey id2 -> Id.equal id1 id2
 | RelKey n1, RelKey n2 -> Int.equal n1 n2
 | EvarKey evk1, EvarKey evk2 -> Evar.equal evk1 evk2
@@ -277,29 +266,20 @@ type zipper =
   | Zapp of arguments
   | Zfix of vfix*arguments  (* Possibly empty *)
   | Zswitch of vswitch
-  | Zproj of Projection.Repr.t (* name of the projection *)
+  | Zproj of int (* index of the projection as in Projection.Repr *)
 
 type stack = zipper list
 
 type to_update = values
+type accumulator = atom * stack
 
-type whd =
-  | Vprod of vprod
-  | Vfun of vfun
-  | Vfix of vfix * arguments option
-  | Vcofix of vcofix * to_update * arguments option
-  | Vconstr_const of int
-  | Vconstr_block of vblock
-  | Vint64 of int64
-  | Vfloat64 of float
-  | Vatom_stk of atom * stack
-  | Vuniv_level of Univ.Level.t
+type kind = (values, accumulator, vfun, vprod, vfix * arguments option, vcofix * to_update * arguments option, vblock) Values.kind
 
 (* Functions over arguments *)
-let nargs : arguments -> int = fun args -> (Obj.size (Obj.repr args)) - 2
+let nargs : arguments -> int = fun args -> Obj.size (Obj.repr args) - 3
 let arg args i =
   if  0 <= i && i < (nargs args) then
-    val_of_obj (Obj.field (Obj.repr args) (i+2))
+    val_of_obj (Obj.field (Obj.repr args) (i + 3))
   else invalid_arg
                 ("Vm.arg size = "^(string_of_int (nargs args))^
                  " acces "^(string_of_int i))
@@ -308,38 +288,17 @@ let arg args i =
 (* Destructors ***********************************)
 (*************************************************)
 
-let uni_lvl_val (v : values) : Univ.Level.t =
-    let whd = Obj.magic v in
-    match whd with
-    | Vuniv_level lvl -> lvl
-    | _ ->
-      let pr =
-        let open Pp in
-        match whd with
-        | Vprod _ -> str "Vprod"
-        | Vfun _ -> str "Vfun"
-        | Vfix _ -> str "Vfix"
-        | Vcofix _ -> str "Vcofix"
-        | Vconstr_const _i -> str "Vconstr_const"
-        | Vconstr_block _b -> str "Vconstr_block"
-        | Vint64 _ -> str "Vint64"
-        | Vfloat64 _ -> str "Vfloat64"
-        | Vatom_stk (_a,_stk) -> str "Vatom_stk"
-        | Vuniv_level _ -> assert false
-      in
-      CErrors.anomaly
-        Pp.(   strbrk "Parsing virtual machine value expected universe level, got "
-            ++ pr ++ str ".")
+let uni_lvl_val (v : values) : Univ.Level.t = Obj.magic v
 
 let rec whd_accu a stk =
   let stk =
-    if Int.equal (Obj.size a) 2 then stk
+    if Int.equal (Obj.size a) 3 then stk
     else Zapp (Obj.obj a) :: stk in
-  let at = Obj.field a 1 in
+  let at = Obj.field a 2 in
   match Obj.tag at with
   | i when Int.equal i type_atom_tag ->
      begin match stk with
-     | [] -> Vatom_stk(Obj.magic at, stk)
+     | [] -> Vaccu (Obj.magic at, stk)
      | [Zapp args] ->
         let args = Array.init (nargs args) (arg args) in
         let s = Obj.obj (Obj.field at 0) in
@@ -347,20 +306,20 @@ let rec whd_accu a stk =
         | Sorts.Type u ->
           let inst = Instance.of_array (Array.map uni_lvl_val args) in
           let u = Univ.subst_instance_universe inst u in
-          Vatom_stk (Asort (Sorts.sort_of_univ u), [])
+          Vaccu (Asort (Sorts.sort_of_univ u), [])
         | _ -> assert false
         end
      | _ -> assert false
      end
   | i when i <= max_atom_tag ->
-      Vatom_stk(Obj.magic at, stk)
+      Vaccu (Obj.magic at, stk)
   | i when Int.equal i proj_tag ->
      let zproj = Zproj (Obj.obj (Obj.field at 0)) in
      whd_accu (Obj.field at 1) (zproj :: stk)
   | i when Int.equal i fix_app_tag ->
       let fa = Obj.field at 1 in
       let zfix  =
-        Zfix (Obj.obj (Obj.field fa 1), Obj.obj fa) in
+        Zfix (Obj.obj (Obj.field fa 2), Obj.obj fa) in
       whd_accu (Obj.field at 0) (zfix :: stk)
   | i when Int.equal i switch_tag ->
       let zswitch = Zswitch (Obj.obj (Obj.field at 1)) in
@@ -381,42 +340,46 @@ let rec whd_accu a stk =
       | [Zapp args] -> Vcofix(vcofix, res, Some args)
       | _           -> assert false
       end
-  | i when Int.equal i Obj.custom_tag ->
-    Vint64 (Obj.magic i)
-  | i when Int.equal i Obj.double_tag ->
-    Vfloat64 (Obj.magic i)
   | tg ->
     CErrors.anomaly
       Pp.(strbrk "Failed to parse VM value. Tag = " ++ int tg ++ str ".")
 
-external kind_of_closure : Obj.t -> int = "coq_kind_of_closure"
+[@@@warning "-37"]
+type vm_closure_kind =
+  | VCfun (** closure, or fixpoint applied past the recursive argument *)
+  | VCfix (** unapplied fixpoint *)
+  | VCfix_partial (** partially applied fixpoint, but not sufficiently for recursion *)
+  | VCaccu (** accumulator *)
+[@@@warning "+37"]
+
+external kind_of_closure : Obj.t -> vm_closure_kind = "coq_kind_of_closure"
 external is_accumulate : tcode -> bool = "coq_is_accumulate_code"
 external int_tcode : tcode -> int -> int = "coq_int_tcode"
 external accumulate : unit -> tcode = "accumulate_code"
 external set_bytecode_field : Obj.t -> int -> tcode -> unit = "coq_set_bytecode_field"
 let accumulate = accumulate ()
 
-let whd_val : values -> whd =
-  fun v ->
-    let o = Obj.repr v in
-    if Obj.is_int o then Vconstr_const (Obj.obj o)
+let whd_val (v: values) =
+  let o = Obj.repr v in
+  if Obj.is_int o then Vconst (Obj.obj o)
+  else
+    let tag = Obj.tag o in
+    if Int.equal tag 0 then
+      if Int.equal (Obj.size o) 1 then
+        Varray (Obj.obj o)
+      else Vprod (Obj.obj o)
+    else if Int.equal tag Obj.closure_tag && is_accumulate (fun_code o) then
+      whd_accu o []
+    else if Int.equal tag Obj.closure_tag || Int.equal tag Obj.infix_tag then
+      match kind_of_closure o with
+      | VCfun -> Vfun (Obj.obj o)
+      | VCfix -> Vfix (Obj.obj o, None)
+      | VCfix_partial -> Vfix (Obj.obj (Obj.field o 2), Some (Obj.obj o))
+      | VCaccu -> Vaccu (Aid (RelKey (int_tcode (fun_code o) 1)), [])
+    else if Int.equal tag Obj.custom_tag then Vint64 (Obj.magic v)
+    else if Int.equal tag Obj.double_tag then Vfloat64 (Obj.magic v)
     else
-      let tag = Obj.tag o in
-      if tag = accu_tag then
-        if is_accumulate (fun_code o) then whd_accu o []
-        else Vprod(Obj.obj o)
-      else
-        if tag = Obj.closure_tag || tag = Obj.infix_tag then
-          (match kind_of_closure o with
-           | 0 -> Vfun(Obj.obj o)
-           | 1 -> Vfix(Obj.obj o, None)
-           | 2 -> Vfix(Obj.obj (Obj.field o 1), Some (Obj.obj o))
-           | 3 -> Vatom_stk(Aid(RelKey(int_tcode (fun_code o) 1)), [])
-           | _ -> CErrors.anomaly ~label:"Vm.whd " (Pp.str "kind_of_closure does not work."))
-        else if Int.equal tag Obj.custom_tag then Vint64 (Obj.magic v)
-        else if Int.equal tag Obj.double_tag then Vfloat64 (Obj.magic v)
-        else
-           Vconstr_block(Obj.obj o)
+      Vblock (Obj.obj o)
 
 (**********************************************)
 (* Constructors *******************************)
@@ -424,9 +387,10 @@ let whd_val : values -> whd =
 
 let obj_of_atom : atom -> Obj.t =
   fun a ->
-    let res = Obj.new_block accu_tag 2 in
+    let res = Obj.new_block Obj.closure_tag 3 in
     set_bytecode_field res 0 accumulate;
-    Obj.set_field res 1 (Obj.repr a);
+    Obj.set_field res 1 (Obj.repr 2);
+    Obj.set_field res 2 (Obj.repr a);
     res
 
 (* obj_of_str_const : structured_constant -> Obj.t *)
@@ -435,7 +399,7 @@ let obj_of_str_const str =
   | Const_sort s -> obj_of_atom (Asort s)
   | Const_ind ind -> obj_of_atom (Aind ind)
   | Const_b0 tag -> Obj.repr tag
-  | Const_univ_level l -> Obj.repr (Vuniv_level l)
+  | Const_univ_level l -> Obj.repr l
   | Const_val v -> Obj.repr v
   | Const_uint i -> Obj.repr i
   | Const_float f -> Obj.repr f
@@ -456,7 +420,7 @@ let val_of_atom a = val_of_obj (obj_of_atom a)
 
 let val_of_int i = (Obj.magic i : values)
 
-let val_of_uint i = (Obj.magic i : values)
+let val_of_uint i = (Obj.magic i : structured_values)
 
 let atom_of_proj kn v =
   let r = Obj.new_block proj_tag 2 in
@@ -473,7 +437,7 @@ struct
   let equal = eq_id_key
   open Hashset.Combine
   let hash : t -> tag = function
-  | ConstKey c -> combinesmall 1 (Constant.hash c)
+  | ConstKey c -> combinesmall 1 (Constant.CanOrd.hash c)
   | VarKey id -> combinesmall 2 (Id.hash id)
   | RelKey i -> combinesmall 3 (Int.hash i)
   | EvarKey evk -> combinesmall 4 (Evar.hash evk)
@@ -499,7 +463,6 @@ let val_of_constant c = val_of_idkey (ConstKey c)
 let val_of_evar evk = val_of_idkey (EvarKey evk)
 
 external val_of_annot_switch : annot_switch -> values = "%identity"
-external val_of_proj_name : Projection.Repr.t -> values = "%identity"
 
 (*************************************************)
 (** Operations manipulating data types ***********)
@@ -516,29 +479,23 @@ external closure_arity : vfun -> int = "coq_closure_arity"
 
 (* Functions over fixpoint *)
 
-external offset : Obj.t -> int = "coq_offset"
-external offset_closure : Obj.t -> int -> Obj.t = "coq_offset_closure"
-external offset_closure_fix : vfix -> int -> vm_env = "coq_offset_closure"
+external current_fix : vfix -> int = "coq_current_fix"
+external shift_fix : vfix -> int -> vfix = "coq_shift_fix"
+external last_fix : vfix -> vfix = "coq_last_fix"
 external tcode_array : tcode_array -> tcode array = "coq_tcode_array"
 
-let first o = (offset_closure o (offset o))
-let first_fix (v:vfix) = (Obj.magic (first (Obj.repr v)) : vfix)
+let first_fix o = shift_fix o (- current_fix o)
+let fix_env v = (Obj.magic (last_fix v) : vm_env)
 
 let last o = (Obj.field o (Obj.size o - 1))
 let fix_types (v:vfix) = tcode_array (Obj.magic (last (Obj.repr v)) : tcode_array)
 let cofix_types (v:vcofix) = tcode_array (Obj.magic (last (Obj.repr v)) : tcode_array)
 
-let current_fix vf = - (offset (Obj.repr vf) / 2)
-
-let unsafe_fb_code fb i =
-  let off = (2 * i) * (Sys.word_size / 8) in
-  Obj.obj (Obj.add_offset (Obj.repr fb) (Int32.of_int off))
-
-let unsafe_rec_arg fb i = int_tcode (unsafe_fb_code fb i) 1
+let unsafe_rec_arg fb i = int_tcode (Obj.magic (shift_fix fb i)) 1
 
 let rec_args vf =
-  let fb = first (Obj.repr vf) in
-  let size = Obj.size (last fb) in
+  let fb = first_fix vf in
+  let size = Obj.size (last (Obj.repr fb)) in
   Array.init size (unsafe_rec_arg fb)
 
 exception FALSE
@@ -547,10 +504,10 @@ let check_fix f1 f2 =
   let i1, i2 = current_fix f1, current_fix f2 in
   (* Checking starting point *)
   if i1 = i2 then
-    let fb1,fb2 = first (Obj.repr f1), first (Obj.repr f2) in
-    let n = Obj.size (last fb1) in
+    let fb1,fb2 = first_fix f1, first_fix f2 in
+    let n = Obj.size (last (Obj.repr fb1)) in
     (* Checking number of definitions *)
-    if n = Obj.size (last fb2) then
+    if n = Obj.size (last (Obj.repr fb2)) then
       (* Checking recursive arguments *)
       try
         for i = 0 to n - 1 do
@@ -593,21 +550,23 @@ let relaccu_code i =
 
 let mk_fix_body k ndef fb =
   let e = Obj.dup (Obj.repr fb) in
+  let env = Obj.repr (fix_env (Obj.obj e)) in
   for i = 0 to ndef - 1 do
-    set_bytecode_field e (2 * i) (relaccu_code (k + i))
+    set_bytecode_field e (3 * i) (relaccu_code (k + i))
   done;
   let fix_body i =
-    let c = offset_tcode (unsafe_fb_code fb i) 2 in
-    let res = Obj.new_block Obj.closure_tag 2 in
+    let c = offset_tcode (Obj.magic (shift_fix fb i)) 2 in
+    let res = Obj.new_block Obj.closure_tag 3 in
     set_bytecode_field res 0 c;
-    Obj.set_field res 1 (offset_closure e (2*i));
+    Obj.set_field res 1 (Obj.repr 2);
+    Obj.set_field res 2 env;
     ((Obj.obj res) : vfun)  in
   Array.init ndef fix_body
 
 (* Functions over vcofix *)
 
 let get_fcofix vcf i =
-  match whd_val (Obj.obj (Obj.field (Obj.repr vcf) (i+1))) with
+  match whd_val (Obj.obj (Obj.field (Obj.repr vcf) (i+2))) with
   | Vcofix(vcfi, _, _) -> vcfi
   | _ -> assert false
 
@@ -628,7 +587,7 @@ let check_cofix vcf1 vcf2 =
 let mk_cofix_body apply_varray k ndef vcf =
   let e = Obj.dup (Obj.repr vcf) in
   for i = 0 to ndef - 1 do
-    Obj.set_field e (i+1) (Obj.repr (val_of_rel (k+i)))
+    Obj.set_field e (i+2) (Obj.repr (val_of_rel (k+i)))
   done;
 
   let cofix_body i =
@@ -636,9 +595,7 @@ let mk_cofix_body apply_varray k ndef vcf =
     let c = Obj.field (Obj.repr vcfi) 0 in
     Obj.set_field e 0 c;
     let atom = Obj.new_block cofix_tag 1 in
-    let self = Obj.new_block accu_tag 2 in
-    set_bytecode_field self 0 accumulate;
-    Obj.set_field self 1 (Obj.repr atom);
+    let self = obj_of_atom (Obj.obj atom) in
     apply_varray (Obj.obj e) [|Obj.obj self|] in
   Array.init ndef cofix_body
 
@@ -679,18 +636,19 @@ let rec pr_atom a =
                             | _ -> str "...") ++ str ")"
   | Aind (mi,i) -> str "Aind(" ++ MutInd.print mi ++ str "#" ++ int i ++ str ")"
   | Asort _ -> str "Asort(")
-and pr_whd w =
-  Pp.(match w with
+and pr_kind w =
+  let open Pp in
+  match w with
   | Vprod _ -> str "Vprod"
   | Vfun _ -> str "Vfun"
   | Vfix _ -> str "Vfix"
   | Vcofix _ -> str "Vcofix"
-  | Vconstr_const i -> str "Vconstr_const(" ++ int i ++ str ")"
-  | Vconstr_block _b -> str "Vconstr_block"
+  | Vconst i -> str "Vconst(" ++ int i ++ str ")"
+  | Vblock _b -> str "Vblock"
   | Vint64 i -> i |> Format.sprintf "Vint64(%LiL)" |> str
   | Vfloat64 f -> str "Vfloat64(" ++ str (Float64.(to_string (of_float f))) ++ str ")"
-  | Vatom_stk (a,stk) -> str "Vatom_stk(" ++ pr_atom a ++ str ", " ++ pr_stack stk ++ str ")"
-  | Vuniv_level _ -> assert false)
+  | Varray _ -> str "Varray"
+  | Vaccu (a, stk) -> str "Vaccu(" ++ pr_atom a ++ str ", " ++ pr_stack stk ++ str ")"
 and pr_stack stk =
   Pp.(match stk with
       | [] -> str "[]"
@@ -700,4 +658,13 @@ and pr_zipper z =
   | Zapp args -> str "Zapp(len = " ++ int (nargs args) ++ str ")"
   | Zfix (_f,args) -> str "Zfix(..., len=" ++ int (nargs args) ++ str ")"
   | Zswitch _s -> str "Zswitch(...)"
-  | Zproj c -> str "Zproj(" ++ Projection.Repr.print c ++ str ")")
+  | Zproj n -> str "Zproj(" ++ int n ++ str ")")
+
+(** Primitives implemented in OCaml *)
+
+let parray_make = Obj.magic Parray.make
+let parray_get = Obj.magic Parray.get
+let parray_get_default = Obj.magic Parray.default
+let parray_set = Obj.magic Parray.set
+let parray_copy = Obj.magic Parray.copy
+let parray_length = Obj.magic Parray.length
