@@ -16,14 +16,9 @@ module type Point = sig
   module Set : CSig.SetS with type elt = t
   module Map : CMap.ExtS with type key = t and module Set := Set
 
-  module Constraint : CSet.S with type elt = (t * constraint_type * t)
-
   val equal : t -> t -> bool
   val compare : t -> t -> int
 
-  type explanation = (constraint_type * t) list
-  val error_inconsistency : constraint_type -> t -> t -> explanation lazy_t option -> 'a
-
   val pr : t -> Pp.t
 end
 
@@ -58,55 +53,96 @@ module Make (Point:Point) = struct
 
   *)
 
-  module PMap = Point.Map
-  module PSet = Point.Set
-  module Constraint = Point.Constraint
+  module Index :
+  sig
+    type t
+    val equal : t -> t -> bool
+    module Set : CSig.SetS with type elt = t
+    module Map : CMap.ExtS with type key = t and module Set := Set
+    type table
+    val empty : table
+    val fresh : Point.t -> table -> t * table
+    val mem : Point.t -> table -> bool
+    val find : Point.t -> table -> t
+    val repr : t -> table -> Point.t
+    val hash : t -> int
+  end =
+  struct
+    type t = int
+    let equal = Int.equal
+    module Set = Int.Set
+    module Map = Int.Map
 
-  type status = NoMark | Visited | WeakVisited | ToMerge
+    type table = {
+      tab_len : int;
+      tab_fwd : Point.t Int.Map.t;
+      tab_bwd : int Point.Map.t
+    }
+
+    let empty = {
+      tab_len = 0;
+      tab_fwd = Int.Map.empty;
+      tab_bwd = Point.Map.empty;
+    }
+    let mem x t = Point.Map.mem x t.tab_bwd
+    let find x t = Point.Map.find x t.tab_bwd
+    let repr n t = Int.Map.find n t.tab_fwd
+
+    let fresh x t =
+      let () = assert (not @@ mem x t) in
+      let n = t.tab_len in
+      n, {
+        tab_len = n + 1;
+        tab_fwd = Int.Map.add n x t.tab_fwd;
+        tab_bwd = Point.Map.add x n t.tab_bwd;
+      }
+    let hash x = x
+  end
+
+  module PMap = Index.Map
+  module PSet = Index.Set
 
   (* Comparison on this type is pointer equality *)
   type canonical_node =
-    { canon: Point.t;
+    { canon: Index.t;
       ltle: bool PMap.t;  (* true: strict (lt) constraint.
                              false: weak  (le) constraint. *)
       gtge: PSet.t;
       rank : int;
       klvl: int;
       ilvl: int;
-      mutable status: status
     }
 
-  let big_rank = 1000000
-
   (* A Point.t is either an alias for another one, or a canonical one,
      for which we know the points that are above *)
 
   type entry =
     | Canonical of canonical_node
-    | Equiv of Point.t
+    | Equiv of Index.t
 
   type t =
     { entries : entry PMap.t;
       index : int;
-      n_nodes : int; n_edges : int }
+      n_nodes : int; n_edges : int;
+      table : Index.table }
 
-  (** Used to cleanup mutable marks if a traversal function is
-      interrupted before it has the opportunity to do it itself. *)
-  let unsafe_cleanup_marks g =
-    let iter _ n = match n with
-      | Equiv _ -> ()
-      | Canonical n -> n.status <- NoMark
-    in
-    PMap.iter iter g.entries
+  module CN = struct
+    type t = canonical_node
+    let equal x y = x.canon == y.canon
+    let hash x = Index.hash x.canon
+  end
 
-  let rec cleanup_marks g =
-    try unsafe_cleanup_marks g
-    with e ->
-      (* The only way unsafe_cleanup_marks may raise an exception is when
-         a serious error (stack overflow, out of memory) occurs, or a signal is
-         sent. In this unlikely event, we relaunch the cleanup until we finally
-         succeed. *)
-      cleanup_marks g; raise e
+  module Status = struct
+    module Internal = Hashtbl.Make(CN)
+
+    (** we could experiment with creation size based on the size of [g] *)
+    let create (g:t) = Internal.create 17
+
+    let mem = Internal.mem
+    let find = Internal.find
+    let replace = Internal.replace
+    let fold = Internal.fold
+  end
 
   (* Every Point.t has a unique canonical arc representative *)
 
@@ -118,12 +154,12 @@ module Make (Point:Point) = struct
         PMap.modify u (fun _ a ->
             match a with
             | Canonical n ->
-              n.status <- NoMark;
               Equiv v
             | _ -> assert false) g.entries;
       index = g.index;
       n_nodes = g.n_nodes - 1;
-      n_edges = g.n_edges }
+      n_edges = g.n_edges;
+      table = g.table }
 
   (* Low-level function : changes data associated with a canonical node.
      Resets the mutable fields in the old record, in order to avoid breaking
@@ -134,8 +170,7 @@ module Make (Point:Point) = struct
                PMap.modify n.canon
                  (fun _ a ->
                     match a with
-                    | Canonical n' ->
-                      n'.status <- NoMark;
+                    | Canonical _ ->
                       Canonical n
                     | _ -> assert false)
                  g.entries }
@@ -145,7 +180,10 @@ module Make (Point:Point) = struct
     match PMap.find u g.entries with
     | Equiv v -> repr g v
     | Canonical arc -> arc
-    | exception Not_found ->
+
+  let repr_node g u =
+    try repr g (Index.find u g.table)
+    with Not_found ->
       CErrors.anomaly ~label:"Univ.repr"
         Pp.(str"Universe " ++ Point.pr u ++ str" undefined.")
 
@@ -158,30 +196,6 @@ module Make (Point:Point) = struct
     assert (g.index > min_int);
     { g with index = g.index - 1 }
 
-  (* [safe_repr] is like [repr] but if the graph doesn't contain the
-     searched point, we add it. *)
-  let safe_repr g u =
-    let rec safe_repr_rec entries u =
-      match PMap.find u entries with
-      | Equiv v -> safe_repr_rec entries v
-      | Canonical arc -> arc
-    in
-    try g, safe_repr_rec g.entries u
-    with Not_found ->
-      let can =
-        { canon = u;
-          ltle = PMap.empty; gtge = PSet.empty;
-          rank = 0;
-          klvl = 0; ilvl = 0;
-          status = NoMark }
-      in
-      let g = { g with
-                entries = PMap.add u (Canonical can) g.entries;
-                n_nodes = g.n_nodes + 1 }
-      in
-      let g = use_index g u in
-      g, repr g u
-
   (* Returns 1 if u is higher than v in topological order.
              -1        lower
              0 if u = v *)
@@ -194,6 +208,7 @@ module Make (Point:Point) = struct
 
   (* Checks most of the invariants of the graph. For debugging purposes. *)
   let check_invariants ~required_canonical g =
+    let required_canonical u = required_canonical (Index.repr u g.table) in
     let n_edges = ref 0 in
     let n_nodes = ref 0 in
     PMap.iter (fun l u ->
@@ -213,8 +228,7 @@ module Make (Point:Point) = struct
                       (PMap.mem u.canon v.ltle ||
                        PMap.exists (fun l _ -> u == repr g l) v.ltle))
             ) u.gtge;
-          assert (u.status = NoMark);
-          assert (Point.equal l u.canon);
+          assert (Index.equal l u.canon);
           assert (u.ilvl > g.index);
           assert (not (PMap.mem u.canon u.ltle));
           incr n_nodes
@@ -226,7 +240,7 @@ module Make (Point:Point) = struct
   let clean_ltle g ltle =
     PMap.fold (fun u strict acc ->
         let uu = (repr g u).canon in
-        if Point.equal uu u then acc
+        if Index.equal uu u then acc
         else (
           let acc = PMap.remove u (fst acc) in
           if not strict && PMap.mem uu acc then (acc, true)
@@ -236,7 +250,7 @@ module Make (Point:Point) = struct
   let clean_gtge g gtge =
     PSet.fold (fun u acc ->
         let uu = (repr g u).canon in
-        if Point.equal uu u then acc
+        if Index.equal uu u then acc
         else PSet.add uu (PSet.remove u (fst acc)), true)
       gtge (gtge, false)
 
@@ -263,16 +277,6 @@ module Make (Point:Point) = struct
       let g = change_node g u in
       u.gtge, u, g
 
-  (* [revert_graph] rollbacks the changes made to mutable fields in
-     nodes in the graph.
-     [to_revert] contains the touched nodes. *)
-  let revert_graph to_revert g =
-    List.iter (fun t ->
-        match PMap.find t g.entries with
-        | Equiv _ -> ()
-        | Canonical t ->
-          t.status <- NoMark) to_revert
-
   exception AbortBackward of t
   exception CycleDetected
 
@@ -286,32 +290,25 @@ module Make (Point:Point) = struct
      corresponding step numbers of the algorithm described in Section
      5.1 of this paper.  *)
 
-  (* [delta] is the timeout for backward search. It might be
-      useful to tune a multiplicative constant. *)
-  let get_delta g =
-    int_of_float
-      (min (float_of_int g.n_edges ** 0.5)
-         (float_of_int g.n_nodes ** (2./.3.)))
-
-  let rec backward_traverse to_revert b_traversed count g x =
-    let x = repr g x in
+  let rec backward_traverse status b_traversed count g x =
     let count = count - 1 in
     if count < 0 then begin
-      revert_graph to_revert g;
-      raise (AbortBackward g)
+      raise_notrace (AbortBackward g)
     end;
-    if x.status = NoMark then begin
-      x.status <- Visited;
-      let to_revert = x.canon::to_revert in
+    if Status.mem status x then b_traversed, count, g
+    else begin
+      Status.replace status x ();
       let gtge, x, g = get_gtge g x in
-      let to_revert, b_traversed, count, g =
-        PSet.fold (fun y (to_revert, b_traversed, count, g) ->
-            backward_traverse to_revert b_traversed count g y)
-          gtge (to_revert, b_traversed, count, g)
+      let b_traversed, count, g =
+        PSet.fold (fun y (b_traversed, count, g) ->
+            let y = repr g y in
+            backward_traverse status b_traversed count g y)
+          gtge (b_traversed, count, g)
       in
-      to_revert, x.canon::b_traversed, count, g
+      x.canon::b_traversed, count, g
     end
-    else to_revert, b_traversed, count, g
+
+  let backward_traverse count g x = backward_traverse (Status.create g) [] count g x
 
   let rec forward_traverse f_traversed g v_klvl x y =
     let y = repr g y in
@@ -334,79 +331,80 @@ module Make (Point:Point) = struct
       f_traversed, g
     else f_traversed, g
 
-  let rec find_to_merge to_revert g x v =
+  let rec find_to_merge status g x v =
     let x = repr g x in
-    match x.status with
-    | Visited -> false, to_revert   | ToMerge -> true, to_revert
-    | NoMark ->
-      let to_revert = x::to_revert in
-      if Point.equal x.canon v then
-        begin x.status <- ToMerge; true, to_revert end
+    match Status.find status x with
+    | merge -> merge
+    | exception Not_found ->
+      if Index.equal x.canon v then begin
+        Status.replace status x true;
+        true
+      end
       else
         begin
-          let merge, to_revert = PSet.fold
-              (fun y (merge, to_revert) ->
-                 let merge', to_revert = find_to_merge to_revert g y v in
-                 merge' || merge, to_revert) x.gtge (false, to_revert)
+          let merge = PSet.fold
+              (fun y merge ->
+                 let merge' = find_to_merge status g y v in
+                 merge' || merge) x.gtge false
           in
-          x.status <- if merge then ToMerge else Visited;
-          merge, to_revert
+          Status.replace status x merge;
+          merge
         end
-    | _ -> assert false
+
+  let find_to_merge g x v =
+    let status = Status.create g in
+    status, find_to_merge status g x v
 
   let get_new_edges g to_merge =
     (* Computing edge sets. *)
-    let to_merge_lvl =
-      List.fold_left (fun acc u -> PMap.add u.canon u acc)
-        PMap.empty to_merge
-    in
     let ltle =
-      let fold _ n acc =
+      let fold acc n =
         let fold u strict acc =
-          if strict then PMap.add u strict acc
-          else if PMap.mem u acc then acc
-          else PMap.add u false acc
+          match PMap.find u acc with
+          | true -> acc
+          | false -> if strict then PMap.add u true acc else acc
+          | exception Not_found -> PMap.add u strict acc
         in
         PMap.fold fold n.ltle acc
       in
-      PMap.fold fold to_merge_lvl PMap.empty
+      match to_merge with
+      | [] -> assert false
+      | hd :: tl -> List.fold_left fold hd.ltle tl
     in
     let ltle, _ = clean_ltle g ltle in
-    let ltle =
-      PMap.merge (fun _ a strict ->
-          match a, strict with
-          | Some _, Some true ->
-            (* There is a lt edge inside the new component. This is a
-                "bad cycle". *)
-            raise CycleDetected
-          | Some _, Some false -> None
-          | _, _ -> strict
-        ) to_merge_lvl ltle
+    let fold accu a =
+      match PMap.find a.canon ltle with
+      | true ->
+        (* There is a lt edge inside the new component. This is a
+            "bad cycle". *)
+        raise_notrace CycleDetected
+      | false -> PMap.remove a.canon accu
+      | exception Not_found -> accu
     in
+    let ltle = List.fold_left fold ltle to_merge in
     let gtge =
-      PMap.fold (fun _ n acc -> PSet.union acc n.gtge)
-        to_merge_lvl PSet.empty
+      List.fold_left (fun acc n -> PSet.union acc n.gtge)
+        PSet.empty to_merge
     in
     let gtge, _ = clean_gtge g gtge in
-    let gtge = PSet.diff gtge (PMap.domain to_merge_lvl) in
+    let gtge = List.fold_left (fun acc n -> PSet.remove n.canon acc) gtge to_merge in
     (ltle, gtge)
 
 
   let reorder g u v =
     (* STEP 2: backward search in the k-level of u. *)
-    let delta = get_delta g in
 
     (* [v_klvl] is the chosen future level for u, v and all
         traversed nodes. *)
     let b_traversed, v_klvl, g =
+      let u = repr g u in
       try
-        let to_revert, b_traversed, _, g = backward_traverse [] [] delta g u in
-        revert_graph to_revert g;
-        let v_klvl = (repr g u).klvl in
+        let b_traversed, _, g = backward_traverse (u.klvl + 1) g u in
+        let v_klvl = u.klvl in
         b_traversed, v_klvl, g
       with AbortBackward g ->
         (* Backward search was too long, use the next k-level. *)
-        let v_klvl = (repr g u).klvl + 1 in
+        let v_klvl = u.klvl + 1 in
         [], v_klvl, g
     in
     let f_traversed, g =
@@ -421,16 +419,13 @@ module Make (Point:Point) = struct
     let to_merge, b_reindex, f_reindex =
       if (repr g u).klvl = v_klvl then
         begin
-          let merge, to_revert = find_to_merge [] g u v in
-          let r =
-            if merge then
-              List.filter (fun u -> u.status = ToMerge) to_revert,
-              List.filter (fun u -> (repr g u).status <> ToMerge) b_traversed,
-              List.filter (fun u -> (repr g u).status <> ToMerge) f_traversed
-            else [], b_traversed, f_traversed
-          in
-          List.iter (fun u -> u.status <- NoMark) to_revert;
-          r
+          let status, merge = find_to_merge g u v in
+          if merge then
+            let not_merged u = try not (Status.find status (repr g u)) with Not_found -> true in
+            Status.fold (fun u merged acc -> if merged then u::acc else acc) status [],
+            List.filter not_merged b_traversed,
+            List.filter not_merged f_traversed
+          else [], b_traversed, f_traversed
         end
       else [], b_traversed, f_traversed
     in
@@ -453,7 +448,7 @@ module Make (Point:Point) = struct
 
         (* Inserting shortcuts for old nodes. *)
         let g = List.fold_left (fun g n ->
-            if Point.equal n.canon root.canon then g else enter_equiv g n.canon root.canon)
+            if Index.equal n.canon root.canon then g else enter_equiv g n.canon root.canon)
             g to_merge
         in
 
@@ -486,7 +481,7 @@ module Make (Point:Point) = struct
       let u = repr g u in
       let v = repr g v in
       if u == v then
-        if strict then raise CycleDetected else g
+        if strict then raise_notrace CycleDetected else g
       else
         let g =
           try let oldstrict = PMap.find v.canon u.ltle in
@@ -502,18 +497,13 @@ module Make (Point:Point) = struct
           let v = { v with gtge = PSet.add u.canon v.gtge } in
           change_node g v
     with
-    | CycleDetected as e -> raise e
-    | e ->
-      (* Unlikely event: fatal error or signal *)
-      let () = cleanup_marks g in
-      raise e
+    | CycleDetected as e -> raise_notrace e
 
   let add ?(rank=0) v g =
-    try
-      let _arcv = PMap.find v g.entries in
-      raise AlreadyDeclared
-    with Not_found ->
-      assert (g.index > min_int);
+    if Index.mem v g.table then raise AlreadyDeclared
+    else
+      let () = assert (g.index > min_int) in
+      let v, table = Index.fresh v g.table in
       let node = {
         canon = v;
         ltle = PMap.empty;
@@ -521,21 +511,22 @@ module Make (Point:Point) = struct
         rank;
         klvl = 0;
         ilvl = g.index;
-        status = NoMark;
       }
       in
       let entries = PMap.add v (Canonical node) g.entries in
-      { entries; index = g.index - 1; n_nodes = g.n_nodes + 1; n_edges = g.n_edges }
+      { entries; index = g.index - 1; n_nodes = g.n_nodes + 1; n_edges = g.n_edges; table }
 
   exception Undeclared of Point.t
   let check_declared g us =
-    let check l = if not (PMap.mem l g.entries) then raise (Undeclared l) in
-    PSet.iter check us
+    let check l = if not (Index.mem l g.table) then raise (Undeclared l) in
+    Point.Set.iter check us
 
   exception Found_explanation of (constraint_type * Point.t) list
 
-  let get_explanation strict u v g =
-    let v = repr g v in
+  type explanation = Point.t * (constraint_type * Point.t) list
+
+  let get_explanation strict pu pv g =
+    let v = repr_node g pv in
     let visited_strict = ref PMap.empty in
     let rec traverse strict u =
       if u == v then
@@ -555,18 +546,21 @@ module Make (Point:Point) = struct
                 | None -> ()
                 | Some exp ->
                   let typ = if strictu' then Lt else Le in
-                  raise (Found_explanation ((typ, u') :: exp)))
+                  let exp = if CList.is_empty exp then [typ, pv] else
+                      let u' = Index.repr u' g.table in
+                      (typ, u') :: exp
+                  in
+                  raise_notrace (Found_explanation exp))
               u.ltle;
             None
           with Found_explanation exp -> Some exp
         end
     in
-    let u = repr g u in
-    if u == v then [(Eq, v.canon)]
+    let u = repr_node g pu in
+    if u == v then begin assert (not strict); [(Eq, pv)] end
     else match traverse strict u with Some exp -> exp | None -> assert false
 
-  let get_explanation strict u v g =
-    Some (lazy (get_explanation strict u v g))
+  let get_explanation strict u v g = u, get_explanation strict u v g
 
   (* To compare two nodes, we simply do a forward search.
      We implement two improvements:
@@ -574,50 +568,45 @@ module Make (Point:Point) = struct
      - we do a BFS rather than a DFS because we expect to have a short
          path (typically, the shortest path has length 1)
   *)
-  exception Found of canonical_node list
+  exception Found
+  type visited = WeakVisited | Visited
   let search_path strict u v g =
-    let rec loop to_revert todo next_todo =
+    let rec loop status todo next_todo =
       match todo, next_todo with
-      | [], [] -> to_revert (* No path found *)
-      | [], _ -> loop to_revert next_todo []
+      | [], [] -> () (* No path found *)
+      | [], _ -> loop status next_todo []
       | (u, strict)::todo, _ ->
-        if u.status = Visited || (u.status = WeakVisited && strict)
-        then loop to_revert todo next_todo
-        else
-          let to_revert =
-            if u.status = NoMark then u::to_revert else to_revert
-          in
-          u.status <- if strict then WeakVisited else Visited;
+        let is_visited = match Status.find status u with
+          | Visited -> true
+          | WeakVisited -> strict
+          | exception Not_found -> false
+        in
+        if is_visited
+        then loop status todo next_todo
+        else begin
+          Status.replace status u (if strict then WeakVisited else Visited);
           if try PMap.find v.canon u.ltle || not strict
             with Not_found -> false
-          then raise (Found to_revert)
+          then raise_notrace Found
           else
             begin
               let next_todo =
                 PMap.fold (fun u strictu next_todo ->
                     let strict = not strictu && strict in
                     let u = repr g u in
-                    if u == v && not strict then raise (Found to_revert)
+                    if u == v && not strict then raise_notrace Found
                     else if topo_compare u v = 1 then next_todo
                     else (u, strict)::next_todo)
                   u.ltle next_todo
               in
-              loop to_revert todo next_todo
+              loop status todo next_todo
             end
+        end
     in
     if u == v then not strict
     else
-      try
-        let res, to_revert =
-          try false, loop [] [u, strict] []
-          with Found to_revert -> true, to_revert
-        in
-        List.iter (fun u -> u.status <- NoMark) to_revert;
-        res
-      with e ->
-        (* Unlikely event: fatal error or signal *)
-        let () = cleanup_marks g in
-        raise e
+      try loop (Status.create g) [u, strict] []; false
+      with Found -> true
 
   (** Uncomment to debug the cycle detection algorithm. *)
   (*let insert_edge strict ucan vcan g =
@@ -636,100 +625,97 @@ module Make (Point:Point) = struct
 
   let check_eq g u v =
     u == v ||
-    let arcu = repr g u and arcv = repr g v in
+    let arcu = repr_node g u and arcv = repr_node g v in
     arcu == arcv
 
   let check_smaller g strict u v =
-    search_path strict (repr g u) (repr g v) g
+    search_path strict (repr_node g u) (repr_node g v) g
 
   let check_leq g u v = check_smaller g false u v
   let check_lt g u v = check_smaller g true u v
 
+  let get_explanation (u, c, v) g = match c with
+  | Eq ->
+    (* Redo the search, not important because this is only used for display. *)
+    if check_lt g u v then get_explanation true u v g else get_explanation true v u g
+  | Le -> get_explanation true v u g
+  | Lt -> get_explanation false v u g
+
   (* enforce_eq g u v will force u=v if possible, will fail otherwise *)
 
-  let rec enforce_eq u v g =
-    let ucan = repr g u in
-    let vcan = repr g v in
-    if topo_compare ucan vcan = 1 then enforce_eq v u g
+  let enforce_eq u v g =
+    let ucan = repr_node g u in
+    let vcan = repr_node g v in
+    if ucan == vcan then Some g
+    else if topo_compare ucan vcan = 1 then
+      let ucan = vcan and vcan = ucan in
+      let g = insert_edge false ucan vcan g in  (* Cannot fail *)
+      try Some (insert_edge false vcan ucan g)
+      with CycleDetected -> None
     else
       let g = insert_edge false ucan vcan g in  (* Cannot fail *)
-      try insert_edge false vcan ucan g
-      with CycleDetected ->
-        Point.error_inconsistency Eq v u (get_explanation true u v g)
+      try Some (insert_edge false vcan ucan g)
+      with CycleDetected -> None
 
   (* enforce_leq g u v will force u<=v if possible, will fail otherwise *)
   let enforce_leq u v g =
-    let ucan = repr g u in
-    let vcan = repr g v in
-    try insert_edge false ucan vcan g
-    with CycleDetected ->
-      Point.error_inconsistency Le u v (get_explanation true v u g)
+    let ucan = repr_node g u in
+    let vcan = repr_node g v in
+    try Some (insert_edge false ucan vcan g)
+    with CycleDetected -> None
 
   (* enforce_lt u v will force u<v if possible, will fail otherwise *)
   let enforce_lt u v g =
-    let ucan = repr g u in
-    let vcan = repr g v in
-    try insert_edge true ucan vcan g
-    with CycleDetected ->
-      Point.error_inconsistency Lt u v (get_explanation false v u g)
+    let ucan = repr_node g u in
+    let vcan = repr_node g v in
+    try Some (insert_edge true ucan vcan g)
+    with CycleDetected -> None
 
   let empty =
-    { entries = PMap.empty; index = 0; n_nodes = 0; n_edges = 0 }
+    { entries = PMap.empty; index = 0; n_nodes = 0; n_edges = 0; table = Index.empty }
 
   (* Normalization *)
 
-  (** [normalize g] returns a graph where all edges point
-      directly to the canonical representent of their target. The output
-      graph should be equivalent to the input graph from a logical point
-      of view, but optimized. We maintain the invariant that the key of
-      a [Canonical] element is its own name, by keeping [Equiv] edges. *)
-  let normalize g =
-    let g =
-      { g with
-        entries = PMap.map (fun entry ->
-            match entry with
-            | Equiv u -> Equiv ((repr g u).canon)
-            | Canonical ucan -> Canonical { ucan with rank = 1 })
-            g.entries }
-    in
-    PMap.fold (fun _ u g ->
-        match u with
-        | Equiv _u -> g
-        | Canonical u ->
-          let _, u, g = get_ltle g u in
-          let _, _, g = get_gtge g u in
-          g)
-      g.entries g
+  type 'a constraint_fold = Point.t * constraint_type * Point.t -> 'a -> 'a
 
-  let constraints_of g =
-    let module UF = Unionfind.Make (PSet) (PMap) in
+  let constraints_of g fold accu =
+    let module UF = Unionfind.Make (Point.Set) (Point.Map) in
     let uf = UF.create () in
     let constraints_of u v acc =
       match v with
       | Canonical {canon=u; ltle; _} ->
         PMap.fold (fun v strict acc->
             let typ = if strict then Lt else Le in
-            Constraint.add (u,typ,v) acc) ltle acc
-      | Equiv v -> UF.union u v uf; acc
+            let u = Index.repr u g.table in
+            let v = Index.repr v g.table in
+            fold (u,typ,v) acc) ltle acc
+      | Equiv v ->
+        let u = Index.repr u g.table in
+        let v = Index.repr v g.table in
+        UF.union u v uf; acc
     in
-    let csts = PMap.fold constraints_of g.entries Constraint.empty in
+    let csts = PMap.fold constraints_of g.entries accu in
     csts, UF.partition uf
 
   (* domain g.entries = kept + removed *)
-  let constraints_for ~kept g =
+  let constraints_for ~kept g fold accu =
     (* rmap: partial map from canonical points to kept points *)
+    let add_cst u knd v cst =
+      fold (Index.repr u g.table, knd, Index.repr v g.table) cst
+    in
+    let kept = Point.Set.fold (fun u accu -> PSet.add (Index.find u g.table) accu) kept PSet.empty in
     let rmap, csts = PSet.fold (fun u (rmap,csts) ->
         let arcu = repr g u in
         if PSet.mem arcu.canon kept then
-          let csts = if Point.equal u arcu.canon then csts
-            else Constraint.add (u,Eq,arcu.canon) csts
+          let csts = if Index.equal u arcu.canon then csts
+            else add_cst u Eq arcu.canon csts
           in
           PMap.add arcu.canon arcu.canon rmap, csts
         else
           match PMap.find arcu.canon rmap with
-          | v -> rmap, Constraint.add (u,Eq,v) csts
+          | v -> rmap, add_cst u Eq v csts
           | exception Not_found -> PMap.add arcu.canon u rmap, csts)
-        kept (PMap.empty,Constraint.empty)
+        kept (PMap.empty, accu)
     in
     let rec add_from u csts todo = match todo with
       | [] -> csts
@@ -738,7 +724,7 @@ module Make (Point:Point) = struct
         (match PMap.find v.canon rmap with
          | v ->
            let d = if strict then Lt else Le in
-           let csts = Constraint.add (u,d,v) csts in
+           let csts = add_cst u d v csts in
            add_from u csts todo
          | exception Not_found ->
            (* v is not equal to any kept point *)
@@ -754,102 +740,42 @@ module Make (Point:Point) = struct
           arc.ltle csts)
       kept csts
 
-  let domain g = PMap.domain g.entries
+  let domain g =
+    let fold u _ accu = Point.Set.add (Index.repr u g.table) accu in
+    PMap.fold fold g.entries Point.Set.empty
 
   let choose p g u =
     let exception Found of Point.t in
-    let ru = (repr g u).canon in
-    if p ru then Some ru
+    let ru = (repr_node g u).canon in
+    let ruv = Index.repr ru g.table in
+    if p ruv then Some ruv
     else
       try PMap.iter (fun v -> function
           | Canonical _ -> () (* we already tried [p ru] *)
           | Equiv v' ->
             let rv = (repr g v').canon in
-            if rv == ru && p v then raise (Found v)
+            if rv == ru then
+              let v = Index.repr v g.table in
+              if p v then raise_notrace (Found v)
             (* NB: we could also try [p v'] but it will come up in the
                rest of the iteration regardless. *)
         ) g.entries; None
       with Found v -> Some v
 
-  let sort make_dummy first g =
-    let cans =
-      PMap.fold (fun _ u l ->
-          match u with
-          | Equiv _ -> l
-          | Canonical can -> can :: l
-        ) g.entries []
-    in
-    let cans = List.sort topo_compare cans in
-    let lowest =
-      PMap.mapi (fun u _ -> if CList.mem_f Point.equal u first then 0 else 2)
-        (PMap.filter
-           (fun _ u -> match u with Equiv _ -> false | Canonical _ -> true)
-           g.entries)
-    in
-    let lowest =
-      List.fold_left (fun lowest can ->
-          let lvl = PMap.find can.canon lowest in
-          PMap.fold (fun u' strict lowest ->
-              let cost = if strict then 1 else 0 in
-              let u' = (repr g u').canon in
-              PMap.modify u' (fun _ lvl0 -> max lvl0 (lvl+cost)) lowest)
-            can.ltle lowest)
-        lowest cans
-    in
-    let max_lvl = PMap.fold (fun _ a b -> max a b) lowest 0 in
-    let types = Array.init (max_lvl + 1) (fun i ->
-        match List.nth_opt first i with
-        | Some u -> u
-        | None -> make_dummy (i-2))
-    in
-    let g = Array.fold_left (fun g u ->
-        let g, u = safe_repr g u in
-        change_node g { u with rank = big_rank }) g types
-    in
-    let g = if max_lvl > List.length first && not (CList.is_empty first) then
-        enforce_lt (CList.last first) types.(List.length first) g
-      else g
-    in
-    let g =
-      PMap.fold (fun u lvl g -> enforce_eq u (types.(lvl)) g)
-        lowest g
-    in
-    normalize g
+  type node = Alias of Point.t | Node of bool Point.Map.t
+  type repr = node Point.Map.t
 
-  (** Pretty-printing *)
-
-  let pr_pmap sep pr map =
-    let cmp (u,_) (v,_) = Point.compare u v in
-    Pp.prlist_with_sep sep pr (List.sort cmp (PMap.bindings map))
-
-  let pr_arc prl = let open Pp in
-    function
-    | _, Canonical {canon=u; ltle; _} ->
-      if PMap.is_empty ltle then mt ()
-      else
-        prl u ++ str " " ++
-        v 0
-          (pr_pmap spc (fun (v, strict) ->
-               (if strict then str "< " else str "<= ") ++ prl v)
-              ltle) ++
-        fnl ()
-    | u, Equiv v ->
-      prl u  ++ str " = " ++ prl v ++ fnl ()
-
-  let pr prl g =
-    pr_pmap Pp.mt (pr_arc prl) g.entries
-
-  (* Dumping constraints to a file *)
-
-  let dump output g =
-    let dump_arc u = function
-      | Canonical {canon=u; ltle; _} ->
-        PMap.iter (fun v strict ->
-            let typ = if strict then Lt else Le in
-            output typ u v) ltle;
-      | Equiv v ->
-        output Eq u v
+  let repr g =
+    let fold u n accu =
+      let n = match n with
+      | Canonical n ->
+        let fold u lt accu = Point.Map.add (Index.repr u g.table) lt accu in
+        let ltle = PMap.fold fold n.ltle Point.Map.empty in
+        Node ltle
+      | Equiv u -> Alias (Index.repr u g.table)
+      in
+      Point.Map.add (Index.repr u g.table) n accu
     in
-    PMap.iter dump_arc g.entries
+    PMap.fold fold g.entries Point.Map.empty
 
 end
