@@ -11,7 +11,8 @@
 open Pp
 open Util
 open Tok
-open Gramlib
+
+module Stream = Gramlib.Stream
 
 (* Dictionaries: trees annotated with string options, each node being a map
    from chars to dictionaries (the subtrees). A trie, in other words. *)
@@ -56,24 +57,6 @@ let ttree_find ttree str =
   in
   proc_rec ttree 0
 
-(* Removes a string from a dictionary: returns an equal dictionary
-   if the word not present. *)
-let ttree_remove ttree str =
-  let rec remove tt i =
-    if i == String.length str then
-      {node = None; branch = tt.branch}
-    else
-      let c = str.[i] in
-      let br =
-        match try Some (CharMap.find c tt.branch) with Not_found -> None with
-          | Some tt' ->
-              CharMap.add c (remove tt' (i + 1)) (CharMap.remove c tt.branch)
-          | None -> tt.branch
-      in
-      { node = tt.node; branch = br }
-  in
-  remove ttree 0
-
 let ttree_elements ttree =
   let rec elts tt accu =
     let accu = match tt.node with
@@ -115,7 +98,7 @@ let bad_token str = raise (Error.E (Bad_token str))
 
 (* Update a loc without allocating an intermediate pair *)
 let set_loc_pos loc bp ep =
-  Ploc.sub loc (bp - loc.Loc.bp) (ep - bp)
+  Loc.sub loc (bp - loc.Loc.bp) (ep - bp)
 
 (* Increase line number by 1 and update position of beginning of line *)
 let bump_loc_line loc bol_pos =
@@ -156,7 +139,6 @@ let after loc =
 
 type token_kind =
   | Utf8Token of (Unicode.status * int)
-  | AsciiChar
   | EmptyStream
 
 let error_utf8 loc cs =
@@ -165,14 +147,6 @@ let error_utf8 loc cs =
   let loc = set_loc_pos loc bp (bp+1) in
   err loc Illegal_character
 
-let utf8_char_size loc cs = function
-  (* Utf8 leading byte *)
-  | '\x00'..'\x7F' -> 1
-  | '\xC0'..'\xDF' -> 2
-  | '\xE0'..'\xEF' -> 3
-  | '\xF0'..'\xF7' -> 4
-  | _ (* '\x80'..\xBF'|'\xF8'..'\xFF' *) -> error_utf8 loc cs
-
 let njunk n = Util.repeat n Stream.junk
 
 let check_utf8_trailing_byte loc cs c =
@@ -180,43 +154,61 @@ let check_utf8_trailing_byte loc cs c =
 
 (* Recognize utf8 blocks (of length less than 4 bytes) *)
 (* but don't certify full utf8 compliance (e.g. no emptyness check) *)
-let lookup_utf8_tail loc c cs =
-  let c1 = Char.code c in
-  if Int.equal (c1 land 0x40) 0 || Int.equal (c1 land 0x38) 0x38 then error_utf8 loc cs
-  else
-    let n, unicode =
+let lookup_utf8_char loc nj cs =
+  match try Some (List.nth (Stream.npeek (nj+1) cs) nj) with Failure _ -> None with
+  | None -> []
+  | Some c1 ->
+    match c1 with
+    | '\x00'..'\x7F' -> [c1]
+    | c1 ->
+      let c1 = Char.code c1 in
+      if Int.equal (c1 land 0x40) 0 || Int.equal (c1 land 0x38) 0x38 then error_utf8 loc cs else
       if Int.equal (c1 land 0x20) 0 then
-      match Stream.npeek 2 cs with
-      | [_;c2] ->
-          check_utf8_trailing_byte loc cs c2;
-          2, (c1 land 0x1F) lsl 6 + (Char.code c2 land 0x3F)
-      | _ -> error_utf8 loc cs
+        match List.skipn nj (Stream.npeek (nj+2) cs) with
+        | [_;c2] as l -> check_utf8_trailing_byte loc cs c2; l
+        | _ -> error_utf8 loc cs
       else if Int.equal (c1 land 0x10) 0 then
-      match Stream.npeek 3 cs with
-      | [_;c2;c3] ->
+        match List.skipn nj (Stream.npeek (nj+3) cs) with
+        | [_;c2;c3] as l ->
           check_utf8_trailing_byte loc cs c2;
           check_utf8_trailing_byte loc cs c3;
-          3, (c1 land 0x0F) lsl 12 + (Char.code c2 land 0x3F) lsl 6 +
-          (Char.code c3 land 0x3F)
-      | _ -> error_utf8 loc cs
-      else match Stream.npeek 4 cs with
-      | [_;c2;c3;c4] ->
+          l
+        | _ -> error_utf8 loc cs
+      else match List.skipn nj (Stream.npeek (nj+4) cs) with
+      | [_;c2;c3;c4] as l ->
           check_utf8_trailing_byte loc cs c2;
           check_utf8_trailing_byte loc cs c3;
           check_utf8_trailing_byte loc cs c4;
-          4, (c1 land 0x07) lsl 18 + (Char.code c2 land 0x3F) lsl 12 +
-          (Char.code c3 land 0x3F) lsl 6 + (Char.code c4 land 0x3F)
+          l
       | _ -> error_utf8 loc cs
+
+let status_of_utf8 = function
+  | [] -> EmptyStream
+  | l ->
+    let n, unicode = match l with
+    | [c1] -> 1, Char.code c1
+    | [c1;c2] -> 2, (Char.code c1 land 0x1F) lsl 6 + (Char.code c2 land 0x3F)
+    | [c1;c2;c3] ->
+       3, (Char.code c1 land 0x0F) lsl 12 + (Char.code c2 land 0x3F) lsl 6 +
+            (Char.code c3 land 0x3F)
+    | [c1;c2;c3;c4] ->
+       4, (Char.code c1 land 0x07) lsl 18 + (Char.code c2 land 0x3F) lsl 12 +
+            (Char.code c3 land 0x3F) lsl 6 + (Char.code c4 land 0x3F)
+    | _ -> assert false
     in
     Utf8Token (Unicode.classify unicode, n)
 
 let lookup_utf8 loc cs =
-  match Stream.peek cs with
-    | Some ('\x00'..'\x7F') -> AsciiChar
-    | Some ('\x80'..'\xFF' as c) -> lookup_utf8_tail loc c cs
-    | None -> EmptyStream
+  status_of_utf8 (lookup_utf8_char loc 0 cs)
 
-let unlocated f x = f x
+let is_letter l =
+  match status_of_utf8 l with
+  | EmptyStream -> false
+  | Utf8Token (st,_) -> Unicode.is_letter st
+
+let unlocated f x =
+  let dummy_loc = Loc.(initial ToplevelInput) in
+  f dummy_loc x
   (** FIXME: should we still unloc the exception? *)
 (*   try f x with Loc.Exc_located (_, exc) -> raise exc *)
 
@@ -226,29 +218,21 @@ let check_keyword str =
         Stream.junk s;
         bad_token str
     | _ ->
-        match unlocated lookup_utf8 Ploc.dummy s with
+        match unlocated lookup_utf8 s with
         | Utf8Token (_,n) -> njunk n s; loop_symb s
-        | AsciiChar -> Stream.junk s; loop_symb s
         | EmptyStream -> ()
   in
   loop_symb (Stream.of_string str)
 
 let check_ident str =
-  let rec loop_id intail s = match Stream.peek s with
-    | Some ('a'..'z' | 'A'..'Z' | '_') ->
-        Stream.junk s;
+  let rec loop_id intail s =
+    match unlocated lookup_utf8 s with
+    | Utf8Token (st, n) when not intail && Unicode.is_valid_ident_initial st -> njunk n s; loop_id true s
+    | Utf8Token (st, n) when intail && Unicode.is_valid_ident_trailing st ->
+        njunk n s;
         loop_id true s
-    | Some ('0'..'9' | '\'') when intail ->
-        Stream.junk s;
-        loop_id true s
-    | _ ->
-        match unlocated lookup_utf8 Ploc.dummy s with
-        | Utf8Token (st, n) when not intail && Unicode.is_valid_ident_initial st -> njunk n s; loop_id true s
-        | Utf8Token (st, n) when intail && Unicode.is_valid_ident_trailing st ->
-          njunk n s;
-          loop_id true s
-        | EmptyStream -> ()
-        | Utf8Token _ | AsciiChar -> bad_token str
+    | EmptyStream -> ()
+    | Utf8Token _ -> bad_token str
   in
   loop_id false (Stream.of_string str)
 
@@ -269,9 +253,6 @@ let add_keyword ?(quotation=NoQuotation) str =
       token_tree := ttree_add !token_tree (str,quotation)
     end
 
-let remove_keyword str =
-  token_tree := ttree_remove !token_tree str
-
 let keywords () = ttree_elements !token_tree
 
 (* Freeze and unfreeze the state of the lexer *)
@@ -304,19 +285,15 @@ let warn_unrecognized_unicode =
           strbrk (Printf.sprintf "Not considering unicode character \"%s\" of unknown \
                                   lexical status as part of identifier \"%s\"." u id))
 
-let rec ident_tail loc len s = match Stream.peek s with
-  | Some ('a'..'z' | 'A'..'Z' | '0'..'9' | '\'' | '_' as c) ->
-      Stream.junk s;
-      ident_tail loc (store len c) s
-  | _ ->
-      match lookup_utf8 loc s with
-      | Utf8Token (st, n) when Unicode.is_valid_ident_trailing st ->
-          ident_tail loc (nstore n len s) s
-      | Utf8Token (st, n) when Unicode.is_unknown st ->
-          let id = get_buff len in
-          let u = String.concat "" (List.map (String.make 1) (Stream.npeek n s)) in
-          warn_unrecognized_unicode ~loc (u,id); len
-      | _ -> len
+let rec ident_tail loc len s =
+  match lookup_utf8 loc s with
+  | Utf8Token (st, n) when Unicode.is_valid_ident_trailing st ->
+      ident_tail loc (nstore n len s) s
+  | Utf8Token (st, n) when Unicode.is_unknown st ->
+      let id = get_buff len in
+      let u = String.concat "" (List.map (String.make 1) (Stream.npeek n s)) in
+      warn_unrecognized_unicode ~loc (u,id); len
+  | _ -> len
 
 let warn_comment_terminator_in_string =
   CWarnings.create ~name:"comment-terminator-in-string" ~category:"parsing"
@@ -377,10 +354,10 @@ let rec string loc ~comm_level bp len s = match Stream.peek s with
       Stream.junk s;
       string loc ~comm_level bp (store len c) s
   | _ ->
-      let _ = Stream.empty s in
-      let ep = Stream.count s in
-     let loc = set_loc_pos loc bp ep in
-     err loc Unterminated_string
+    let () = if not (Stream.is_empty s) then raise Stream.Failure in
+    let ep = Stream.count s in
+    let loc = set_loc_pos loc bp ep in
+    err loc Unterminated_string
 
 (* Utilities for comments in beautify *)
 let comment_begin = ref None
@@ -459,12 +436,12 @@ let rec comment loc bp s =
       push_string "\""; push_string (get_buff len); push_string "\"";
       comment loc bp s
   | _ ->
-    match try Some (Stream.empty s) with Stream.Failure -> None with
-    | Some _ ->
+    match Stream.is_empty s with
+    | true ->
       let ep = Stream.count s in
       let loc = set_loc_pos loc bp ep in
       err loc Unterminated_comment
-    | _ ->
+    | false ->
           match Stream.peek s with
             Some ('\n' as z) ->
               Stream.junk s;
@@ -477,46 +454,53 @@ let rec comment loc bp s =
 
 (* Parse a special token, using the [token_tree] *)
 
-(* Peek as much utf-8 lexemes as possible *)
-(* and retain the longest valid special token obtained *)
-let rec progress_further loc last nj tt cs =
-  try progress_from_byte loc last nj tt cs (List.nth (Stream.npeek (nj+1) cs) nj)
-  with Failure _ -> last
+(* Below, [last] is the last valid token found in the table *)
+(* [nj] is the number of bytes read since then and *)
+(* these bytes form a complete utf-8 sequence *)
+(* [tt] is the tree after having taken into account these [nj] bytes *)
 
-and update_longest_valid_token loc last nj tt cs =
+let update_longest_valid_token last nj tt cs =
   match tt.node with
   | Some _ as last' ->
-    stream_njunk nj cs;
-    progress_further loc last' 0 tt cs
+    (* [last] extended with the [nj] bytes constitutes a valid token *)
+    (* we update cs, last and nj *)
+    Stream.njunk nj cs; 0, last'
   | None ->
-    progress_further loc last nj tt cs
+    (* [last] extended with the [nj] bytes does not form a full valid token *)
+    nj, last
 
-(* nj is the number of char peeked since last valid token *)
-(* n the number of char in utf8 block *)
-and progress_utf8 loc last nj n c tt cs =
+(* try to progress by peeking the next utf-8 lexeme *)
+(* and retain the longest valid special token obtained *)
+let rec progress_further loc last nj last_is_letter tt cs =
+  match lookup_utf8_char loc nj cs with
+  | [] -> snd (update_longest_valid_token last nj tt cs)
+  | l -> progress_utf8 loc last nj last_is_letter tt cs l
+
+(* under the same assumptions as [update_longest_valid_token], *)
+(* read the [n] bytes of the current utf-8 lexeme whose first byte is [c] *)
+and progress_utf8 loc last nj last_is_letter tt cs l =
+  let is_letter' = is_letter l in
+  (* compute longest match before considering utf8 block [l] *)
+  (* not allowing update if in the middle of an ident part *)
+  let nj, last = if last_is_letter && is_letter' then nj, last else update_longest_valid_token last nj tt cs in
   try
-    let tt = CharMap.find c tt.branch in
-    if Int.equal n 1 then
-      update_longest_valid_token loc last (nj+n) tt cs
-    else
-      match Util.List.skipn (nj+1) (Stream.npeek (nj+n) cs) with
-      | l when Int.equal (List.length l) (n - 1) ->
-         List.iter (check_utf8_trailing_byte loc cs) l;
-         let tt = List.fold_left (fun tt c -> CharMap.find c tt.branch) tt l in
-         update_longest_valid_token loc last (nj+n) tt cs
-      | _ ->
-          error_utf8 loc cs
+    (* descend in tree according to current utf8 block [l] *)
+    let tt = List.fold_left (fun tt c -> CharMap.find c tt.branch) tt l in
+    progress_further loc last (nj + List.length l) is_letter' tt cs
   with Not_found ->
     last
 
-and progress_from_byte loc last nj tt cs c =
-  progress_utf8 loc last nj (utf8_char_size loc cs c) c tt cs
+let blank_or_eof cs =
+  match Stream.peek cs with
+    | None -> true
+    | Some (' ' | '\t' | '\n' |'\r') -> true
+    | _ -> false
 
 type marker = Delimited of int * char list * char list | ImmediateAsciiIdent
 
 let peek_marker_len b e s =
   let rec peek n =
-    match stream_nth n s with
+    match Stream.nth n s with
     | c -> if c = b then peek (n+1) else n, List.make n b, List.make n e
     | exception Stream.Failure -> n, List.make n b, List.make n e
   in
@@ -525,7 +509,7 @@ let peek_marker_len b e s =
   else Delimited (len, start, stop)
 
 let peek_marker s =
-  match stream_nth 0 s with
+  match Stream.nth 0 s with
     | '(' -> peek_marker_len '(' ')' s
     | '[' -> peek_marker_len '[' ']' s
     | '{' -> peek_marker_len '{' '}' s
@@ -542,6 +526,11 @@ let parse_quotation loc bp s =
       in
       get_buff len, set_loc_pos loc bp (Stream.count s)
   | Delimited (lenmarker, bmarker, emarker) ->
+      let dot_gobbling =
+        (* only quotations starting with two curly braces can gobble sentences *)
+        match bmarker with
+        | '{' :: '{' :: _ -> true
+        | _ -> false in
       let b = Buffer.create 80 in
       let commit1 c = Buffer.add_char b c; Stream.junk s in
       let commit l = List.iter commit1 l in
@@ -557,6 +546,10 @@ let parse_quotation loc bp s =
               commit1 '\n';
               let loc = bump_loc_line_last loc (Stream.count s) in
               quotation loc depth
+        | '.' :: _ ->
+              commit1 '.';
+              if not dot_gobbling && blank_or_eof s then raise Stream.Failure;
+              quotation loc depth
         | c :: cs ->
               commit1 c;
               quotation loc depth
@@ -565,10 +558,28 @@ let parse_quotation loc bp s =
       let loc = quotation loc 0 in
       Buffer.contents b, set_loc_pos loc bp (Stream.count s)
 
+let peek_string v s =
+  let l = String.length v in
+  let rec aux i =
+    if Int.equal i l then true
+    else
+      let l' = Stream.npeek (i + 1) s in
+      match List.nth l' i with
+      | c -> Char.equal c v.[i] && aux (i + 1)
+      | exception _ -> false (* EOF *) in
+  aux 0
 
 let find_keyword loc id bp s =
+  if peek_string ":{{" s then
+    begin
+      (* "xxx:{{" always starts a sentence-gobbling quotation, whether registered or not *)
+      Stream.junk s;
+      let txt, loc = parse_quotation loc bp s in
+      QUOTATION (id ^ ":", txt), loc
+    end
+  else
   let tt = ttree_find !token_tree id in
-  match progress_further loc tt.node 0 tt s with
+  match progress_further loc tt.node 0 true tt s with
   | None -> raise Not_found
   | Some (c,NoQuotation) ->
       let ep = Stream.count s in
@@ -586,8 +597,8 @@ let process_sequence loc bp c cs =
   aux 1 cs
 
 (* Must be a special token *)
-let process_chars ~diff_mode loc bp c cs =
-  let t = progress_from_byte loc None (-1) !token_tree cs c in
+let process_chars ~diff_mode loc bp l cs =
+  let t = progress_utf8 loc None (- (List.length l)) false !token_tree cs l in
   let ep = Stream.count cs in
   match t with
     | Some (t,NoQuotation) -> (KEYWORD t, set_loc_pos loc bp ep)
@@ -595,61 +606,35 @@ let process_chars ~diff_mode loc bp c cs =
         let txt, loc = parse_quotation loc bp cs in
         QUOTATION(c, txt), loc
     | None ->
-        let ep' = bp + utf8_char_size loc cs c in
         if diff_mode then begin
-          let len = ep' - bp in
-          ignore (store 0 c);
-          ignore (nstore (len - 1) 1 cs);
-          IDENT (get_buff len), set_loc_pos loc bp ep
+          let s = String.concat "" (List.map (String.make 1) l) in
+          IDENT s, set_loc_pos loc bp ep
         end else begin
-          njunk (ep' - ep) cs;
-          let loc = set_loc_pos loc bp ep' in
+          let loc = set_loc_pos loc bp ep in
           err loc Undefined_token
         end
 
 (* Parse what follows a dot *)
 
-let parse_after_dot ~diff_mode loc c bp s = match Stream.peek s with
-  | Some ('a'..'z' | 'A'..'Z' | '_' as d) ->
-      Stream.junk s;
-      let len =
-        try ident_tail loc (store 0 d) s with
-          Stream.Failure -> raise (Stream.Error "")
-      in
+let parse_after_dot ~diff_mode loc c bp s =
+  match lookup_utf8 loc s with
+  | Utf8Token (st, n) when Unicode.is_valid_ident_initial st ->
+      let len = ident_tail loc (nstore n 0 s) s in
       let field = get_buff len in
       begin try find_keyword loc ("."^field) bp s
-      with Not_found ->
-        let ep = Stream.count s in
-        FIELD field, set_loc_pos loc bp ep end
-  | _ ->
-      match lookup_utf8 loc s with
-      | Utf8Token (st, n) when Unicode.is_valid_ident_initial st ->
-          let len = ident_tail loc (nstore n 0 s) s in
-          let field = get_buff len in
-          begin try find_keyword loc ("."^field) bp s
-          with Not_found ->
-            let ep = Stream.count s in
-            FIELD field, set_loc_pos loc bp ep end
-      | AsciiChar | Utf8Token _ | EmptyStream ->
-          process_chars ~diff_mode loc bp c s
+            with Not_found ->
+              let ep = Stream.count s in
+              FIELD field, set_loc_pos loc bp ep end
+  | Utf8Token _ | EmptyStream ->
+      process_chars ~diff_mode loc bp [c] s
 
 (* Parse what follows a question mark *)
 
 let parse_after_qmark ~diff_mode loc bp s =
-  match Stream.peek s with
-    | Some ('a'..'z' | 'A'..'Z' | '_') -> LEFTQMARK
-    | None -> KEYWORD "?"
-    | _ ->
-        match lookup_utf8 loc s with
-          | Utf8Token (st, _) when Unicode.is_valid_ident_initial st -> LEFTQMARK
-          | AsciiChar | Utf8Token _ | EmptyStream ->
-            fst (process_chars ~diff_mode loc bp '?' s)
-
-let blank_or_eof cs =
-  match Stream.peek cs with
-    | None -> true
-    | Some (' ' | '\t' | '\n' |'\r') -> true
-    | _ -> false
+  match lookup_utf8 loc s with
+  | Utf8Token (st, _) when Unicode.is_valid_ident_initial st -> LEFTQMARK
+  | EmptyStream -> KEYWORD "?"
+  | Utf8Token _ -> fst (process_chars ~diff_mode loc bp ['?'] s)
 
 (* Parse a token in a char stream *)
 
@@ -686,7 +671,7 @@ let rec next_token ~diff_mode loc s =
       Stream.junk s;
       let t,new_between_commands =
         if !between_commands then process_sequence loc bp c s, true
-        else process_chars ~diff_mode loc bp c s,false
+        else process_chars ~diff_mode loc bp [c] s,false
       in
       comment_stop bp; between_commands := new_between_commands; t
   | Some '?' ->
@@ -708,9 +693,11 @@ let rec next_token ~diff_mode loc s =
         IDENT id, set_loc_pos loc bp ep end
   | Some ('0'..'9') ->
       let n = NumTok.Unsigned.parse s in
-      let ep = Stream.count s in
       comment_stop bp;
-      (NUMERAL n, set_loc_pos loc bp ep)
+      begin try find_keyword loc (NumTok.Unsigned.sprint n) bp s
+      with Not_found ->
+        let ep = Stream.count s in
+        NUMBER n, set_loc_pos loc bp ep end
   | Some '\"' ->
       Stream.junk s;
       let (loc, len) =
@@ -733,7 +720,7 @@ let rec next_token ~diff_mode loc s =
             comm_loc bp;
             push_string "(*";
             let loc = comment loc bp s in next_token ~diff_mode loc s
-        | _ -> let t = process_chars ~diff_mode loc bp c s in comment_stop bp; t
+        | _ -> let t = process_chars ~diff_mode loc bp [c] s in comment_stop bp; t
       with Stream.Failure -> raise (Stream.Error "")
       end
   | Some ('{' | '}' as c) ->
@@ -741,11 +728,12 @@ let rec next_token ~diff_mode loc s =
       let ep = Stream.count s in
       let t,new_between_commands =
         if !between_commands then (KEYWORD (String.make 1 c), set_loc_pos loc bp ep), true
-        else process_chars ~diff_mode loc bp c s, false
+        else process_chars ~diff_mode loc bp [c] s, false
       in
       comment_stop bp; between_commands := new_between_commands; t
   | _ ->
-      match lookup_utf8 loc s with
+      let l = lookup_utf8_char loc 0 s in
+      match status_of_utf8 l with
         | Utf8Token (st, n) when Unicode.is_valid_ident_initial st ->
             let len = ident_tail loc (nstore n 0 s) s in
             let id = get_buff len in
@@ -754,73 +742,23 @@ let rec next_token ~diff_mode loc s =
             with Not_found ->
               let ep = Stream.count s in
               IDENT id, set_loc_pos loc bp ep end
-        | AsciiChar | Utf8Token _ ->
-            let t = process_chars ~diff_mode loc bp (Stream.next s) s in
+        | Utf8Token (_, n) ->
+            Stream.njunk n s;
+            let t = process_chars ~diff_mode loc bp l s in
             comment_stop bp; t
         | EmptyStream ->
             comment_stop bp; (EOI, set_loc_pos loc bp (bp+1))
 
-(* (* Debug: uncomment this for tracing tokens seen by coq...*)
-let next_token ~diff_mode loc s =
-  let (t,loc as r) = next_token ~diff_mode loc s in
-  Printf.eprintf "(line %i, %i-%i)[%s]\n%!" (Ploc.line_nb loc) (Ploc.first_pos loc) (Ploc.last_pos loc) (Tok.to_string t);
-  r *)
-
-(* Location table system for creating tables associating a token count
-   to its location in a char stream (the source) *)
-
-let locerr i =
-  let m = "Lexer: location function called on token "^string_of_int i in
-  invalid_arg m
-
-let loct_create () = Hashtbl.create 207
-
-let loct_func loct i =
-  try Hashtbl.find loct i with Not_found -> locerr i
-
-let loct_add loct i loc = Hashtbl.add loct i loc
-
 (** {6 The lexer of Coq} *)
 
-(** Note: removing a token.
-   We do nothing because [remove_token] is called only when removing a grammar
-   rule with [Grammar.delete_rule]. The latter command is called only when
-   unfreezing the state of the grammar entries (see GRAMMAR summary, file
-   env/metasyntax.ml). Therefore, instead of removing tokens one by one,
-   we unfreeze the state of the lexer. This restores the behaviour of the
-   lexer. B.B. *)
-
-(** Names of tokens, for this lexer, used in Grammar error messages *)
-
-let token_text : type c. c Tok.p -> string = function
-  | PKEYWORD t -> "'" ^ t ^ "'"
-  | PIDENT None -> "identifier"
-  | PIDENT (Some t) -> "'" ^ t ^ "'"
-  | PNUMERAL None -> "numeral"
-  | PNUMERAL (Some n) -> "'" ^ NumTok.Unsigned.sprint n ^ "'"
-  | PSTRING None -> "string"
-  | PSTRING (Some s) -> "STRING \"" ^ s ^ "\""
-  | PLEFTQMARK -> "LEFTQMARK"
-  | PEOI -> "end of input"
-  | PPATTERNIDENT None -> "PATTERNIDENT"
-  | PPATTERNIDENT (Some s) -> "PATTERNIDENT \"" ^ s ^ "\""
-  | PFIELD None -> "FIELD"
-  | PFIELD (Some s) -> "FIELD \"" ^ s ^ "\""
-  | PBULLET None -> "BULLET"
-  | PBULLET (Some s) -> "BULLET \"" ^ s ^ "\""
-  | PQUOTATION lbl -> "QUOTATION \"" ^ lbl ^ "\""
-
-let func next_token ?loc cs =
-  let loct = loct_create () in
-  let cur_loc = ref (Option.default Loc.(initial ToplevelInput) loc) in
-  let ts =
-    Stream.from
-      (fun i ->
-         let (tok, loc) = next_token !cur_loc cs in
-         cur_loc := after loc;
-         loct_add loct i loc; Some tok)
-  in
-  (ts, loct_func loct)
+let func next_token ?(loc=Loc.(initial ToplevelInput)) ?(fix_loc=(fun l -> l)) cs =
+  let cur_loc = ref loc in
+  Gramlib.LStream.from ~loc
+    (fun () ->
+       let (tok, loc) = next_token !cur_loc cs in
+       cur_loc := after loc;
+       let loc = fix_loc loc in
+       Some (tok,loc))
 
 module MakeLexer (Diff : sig val mode : bool end) = struct
   type te = Tok.t
@@ -832,9 +770,8 @@ module MakeLexer (Diff : sig val mode : bool end) = struct
     | PKEYWORD s -> add_keyword ~quotation:NoQuotation s
     | PQUOTATION s -> add_keyword ~quotation:Quotation s
     | _ -> ()
-  let tok_removing = (fun _ -> ())
   let tok_match = Tok.match_pattern
-  let tok_text = token_text
+  let tok_text = Tok.token_text
 
   (* The state of the lexer visible from outside *)
   module State = struct
@@ -889,7 +826,7 @@ let terminal s =
   if is_ident_not_keyword s then PIDENT (Some s)
   else PKEYWORD s
 
-(* Precondition: the input is a numeral (c.f. [NumTok.t]) *)
-let terminal_numeral s = match NumTok.Unsigned.parse_string s with
-  | Some n -> PNUMERAL (Some n)
-  | None -> failwith "numeral token expected."
+(* Precondition: the input is a number (c.f. [NumTok.t]) *)
+let terminal_number s = match NumTok.Unsigned.parse_string s with
+  | Some n -> PNUMBER (Some n)
+  | None -> failwith "number token expected."
