@@ -98,7 +98,7 @@ module EntryDataMap = EntryCommand.Map(EntryData)
 
 type ext_kind =
   | ByGrammar of extend_rule
-  | ByEXTEND of (unit -> unit) * (unit -> unit)
+  | ByEXTEND of string * (unit -> unit) * (unit -> unit)
   | ByEntry : ('a * 'b) EntryCommand.tag * string * 'b Entry.t -> ext_kind
 
 (** The list of extensions *)
@@ -134,7 +134,7 @@ let grammar_delete_reinit e reinit d =
 let grammar_extend e ext =
   let undo () = grammar_delete e ext in
   let redo () = safe_extend e ext in
-  camlp5_state := ByEXTEND (undo, redo) :: !camlp5_state;
+  camlp5_state := ByEXTEND (Entry.name e, undo, redo) :: !camlp5_state;
   redo ()
 
 let grammar_extend_sync e ext =
@@ -159,15 +159,17 @@ let rec remove_grammars n =
            camlp5_state := t;
            remove_grammars (n-1)
        | ByGrammar (ExtendRule (g, ext)) :: t ->
-           grammar_delete g ext;
-           camlp5_state := t;
-           remove_grammars (n-1)
-       | ByEXTEND (undo,redo)::t ->
+         (try grammar_delete g ext
+          with Not_found -> Feedback.msg_info Pp.(str "failed on " ++ str (Entry.name g));
+            raise Not_found);
+         camlp5_state := t;
+         remove_grammars (n-1)
+       | ByEXTEND (name, undo,redo)::t ->
            undo();
            camlp5_state := t;
            remove_grammars n;
            redo();
-           camlp5_state := ByEXTEND (undo,redo) :: !camlp5_state
+           camlp5_state := ByEXTEND (name, undo,redo) :: !camlp5_state
        | ByEntry (tag, name, e) :: t ->
            Unsafe.clear_entry e;
            camlp5_state := t;
@@ -322,7 +324,10 @@ let epsilon_value (type s tr a) f (e : (s, tr, a) Symbol.t) =
 
 module GramState = Store.Make ()
 
-type 'a grammar_extension = 'a -> GramState.t -> extend_rule list * GramState.t
+type 'a grammar_extension = {
+  gext_fun : 'a -> GramState.t -> extend_rule list * GramState.t;
+  gext_eq : 'a -> 'a -> bool;
+}
 
 module GrammarCommand = Dyn.Make ()
 module GrammarInterp = struct type 'a t = 'a grammar_extension end
@@ -330,7 +335,10 @@ module GrammarInterpMap = GrammarCommand.Map(GrammarInterp)
 
 let grammar_interp = ref GrammarInterpMap.empty
 
-type ('a, 'b) entry_extension = 'a -> GramState.t -> string list * GramState.t
+type ('a, 'b) entry_extension = {
+  eext_fun : 'a -> GramState.t -> string list * GramState.t;
+  eext_eq : 'a -> 'a -> bool;
+}
 
 module EntryInterp = struct type _ t = Ex : ('a, 'b) entry_extension -> ('a * 'b) t end
 module EntryInterpMap = EntryCommand.Map(EntryInterp)
@@ -368,7 +376,7 @@ let extend_grammar_command tag g =
   | [] -> GramState.empty
   | (_, st) :: _ -> st
   in
-  let (rules, st) = modify g grammar_state in
+  let (rules, st) = modify.gext_fun g grammar_state in
   let () = List.iter iter_extend_sync rules in
   let nb = List.length rules in
   grammar_stack := (GramExt (nb, GrammarCommand.Dyn (tag, g)), st) :: !grammar_stack
@@ -379,7 +387,7 @@ let extend_entry_command (type a) (type b) (tag : (a, b) entry_command) (g : a)
   | [] -> GramState.empty
   | (_, st) :: _ -> st
   in
-  let (names, st) = modify g grammar_state in
+  let (names, st) = modify.eext_fun g grammar_state in
   let entries = List.map (fun name -> Entry.make name) names in
   let iter name e =
     camlp5_state := ByEntry (tag, name, e) :: !camlp5_state;
@@ -406,9 +414,7 @@ let extend_dyn_grammar (e, _) = match e with
 
 (** Registering extra grammar *)
 
-type any_entry = AnyEntry : 'a Entry.t -> any_entry
-
-let grammar_names : any_entry list String.Map.t ref = ref String.Map.empty
+let grammar_names : Entry.any_t list String.Map.t ref = ref String.Map.empty
 
 let register_grammars_by_name name grams =
   grammar_names := String.Map.add name grams !grammar_names
@@ -417,7 +423,7 @@ let find_grammars_by_name name =
   try String.Map.find name !grammar_names
   with Not_found ->
     let fold (EntryDataMap.Any (tag, EntryData.Ex map)) accu =
-      try AnyEntry (String.Map.find name map) :: accu
+      try Entry.Any (String.Map.find name map) :: accu
       with Not_found -> accu
     in
     EntryDataMap.fold fold !camlp5_entries []
@@ -432,10 +438,27 @@ type frozen_t =
 let freeze ~marshallable : frozen_t =
   (!grammar_stack, CLexer.get_keyword_state ())
 
+let eq_grams (g1, _) (g2, _) = match g1, g2 with
+| GramExt (_, GrammarCommand.Dyn (t1, v1)), GramExt (_, GrammarCommand.Dyn (t2, v2)) ->
+  begin match GrammarCommand.eq t1 t2 with
+  | None -> false
+  | Some Refl ->
+    let data = GrammarInterpMap.find t1 !grammar_interp in
+    data.gext_eq v1 v2
+  end
+| EntryExt (_, t1, v1), EntryExt (_, t2, v2) ->
+  begin match EntryCommand.eq t1 t2 with
+  | None -> false
+  | Some Refl ->
+    let EntryInterp.Ex data = EntryInterpMap.find t1 !entry_interp in
+    data.eext_eq v1 v2
+  end
+| (GramExt _, EntryExt _) | (EntryExt _, GramExt _) -> false
+
 (* We compare the current state of the grammar and the state to unfreeze,
    by computing the longest common suffixes *)
 let factorize_grams l1 l2 =
-  if l1 == l2 then ([], [], l1) else List.share_tails l1 l2
+  if l1 == l2 then ([], [], l1) else List.share_tails eq_grams l1 l2
 
 let rec number_of_entries accu = function
 | [] -> accu
@@ -457,7 +480,8 @@ let unfreeze (grams, lex) =
 
 let parser_summary_tag =
   Summary.declare_summary_tag "GRAMMAR_LEXER"
-    { Summary.freeze_function = freeze;
+    { stage = Summary.Stage.Synterp;
+      Summary.freeze_function = freeze;
       Summary.unfreeze_function = unfreeze;
       Summary.init_function = Summary.nop }
 
