@@ -20,9 +20,7 @@ open Environ
 open Reduction
 open Reductionops
 open Inductive
-open Termops
 open Inductiveops
-open Recordops
 open Namegen
 open Miniml
 open Table
@@ -42,11 +40,11 @@ let current_fixpoints = ref ([] : Constant.t list)
 
 let type_of env sg c =
   let polyprop = (lang() == Haskell) in
-  Retyping.get_type_of ~polyprop env sg (strip_outer_cast sg c)
+  Retyping.get_type_of ~polyprop env sg (Termops.strip_outer_cast sg c)
 
 let sort_of env sg c =
   let polyprop = (lang() == Haskell) in
-  Retyping.get_sort_family_of ~polyprop env sg (strip_outer_cast sg c)
+  Retyping.get_sort_family_of ~polyprop env sg (Termops.strip_outer_cast sg c)
 
 (*S Generation of flags and signatures. *)
 
@@ -110,11 +108,11 @@ let push_rel_assum (n, t) env =
 let push_rels_assum assums =
   EConstr.push_rel_context (List.map (fun (x,t) -> LocalAssum (x,t)) assums)
 
-let get_body lconstr = EConstr.of_constr (Mod_subst.force_constr lconstr)
+let get_body lconstr = EConstr.of_constr lconstr
 
 let get_opaque env c =
   EConstr.of_constr
-    (fst (Opaqueproof.force_proof Library.indirect_accessor (Environ.opaque_tables env) c))
+    (fst (Global.force_proof Library.indirect_accessor c))
 
 let applistc c args = EConstr.mkApp (c, Array.of_list args)
 
@@ -291,7 +289,7 @@ let rec extract_type env sg db j c args =
                       let reason = if lvl == TypeScheme then Ktype else Kprop in
                       Tarr (Tdummy reason, mld)))
     | Sort _ -> Tdummy Ktype (* The two logical cases. *)
-    | _ when sort_of env sg (applistc c args) == InProp -> Tdummy Kprop
+    | _ when info_of_family (sort_of env sg (applistc c args)) == Logic -> Tdummy Kprop
     | Rel n ->
         (match EConstr.lookup_rel n env with
            | LocalDef (_,t,_) ->
@@ -307,19 +305,8 @@ let rec extract_type env sg db j c args =
         (match flag_of_type env sg typ with
            | (Logic,_) -> assert false (* Cf. logical cases above *)
            | (Info, TypeScheme) ->
-               let mlt = extract_type_app env sg db (r, type_sign env sg typ) args in
-               (match (lookup_constant kn env).const_body with
-                 | Undef _ | OpaqueDef _ | Primitive _ -> mlt
-                 | Def _ when is_custom (GlobRef.ConstRef kn) -> mlt
-                 | Def lbody ->
-                      let newc = applistc (get_body lbody) args in
-                      let mlt' = extract_type env sg db j newc [] in
-                      (* ML type abbreviations interact badly with Coq *)
-                      (* reduction, so [mlt] and [mlt'] might be different: *)
-                      (* The more precise is [mlt'], extracted after reduction *)
-                      (* The shortest is [mlt], which use abbreviations *)
-                      (* If possible, we take [mlt], otherwise [mlt']. *)
-                      if eq_ml_type (expand env mlt) (expand env mlt') then mlt else mlt')
+             let mlt = extract_type_app env sg db (r, type_sign env sg typ) args in
+             mlt
            | (Info, Default) ->
                (* Not an ML type, for example [(c:forall X, X->X) Type nat] *)
                (match (lookup_constant kn env).const_body with
@@ -351,7 +338,7 @@ let rec extract_type env sg db j c args =
             | (Info, TypeScheme) ->
               extract_type_app env sg db (r, type_sign env sg ty) args
             | (Info, Default) -> Tunknown))
-    | Cast _ | LetIn _ | Construct _ | Int _ | Float _ -> assert false
+    | Cast _ | LetIn _ | Construct _ | Int _ | Float _ | Array _ -> assert false
 
 (*s Auxiliary function dealing with type application.
   Precondition: [r] is a type scheme represented by the signature [s],
@@ -428,6 +415,7 @@ and extract_really_ind env kn mib =
     (* Everything concerning parameters. *)
     (* We do that first, since they are common to all the [mib]. *)
     let mip0 = mib.mind_packets.(0) in
+    let ndecls = List.length mib.mind_params_ctxt in
     let npar = mib.mind_nparams in
     let epar = push_rel_context mib.mind_params_ctxt env in
     let sg = Evd.from_env env in
@@ -463,17 +451,17 @@ and extract_really_ind env kn mib =
       if not p.ip_logical then
         let types = arities_of_constructors env ((kn,i),u) in
         for j = 0 to Array.length types - 1 do
-          let t = snd (decompose_prod_n npar types.(j)) in
+          let t = snd (decompose_prod_n_assum ndecls types.(j)) in
           let prods,head = dest_prod epar t in
           let nprods = List.length prods in
           let args = match Constr.kind head with
             | App (f,args) -> args (* [Constr.kind f = Ind ip] *)
             | _ -> [||]
           in
-          let dbmap = parse_ind_args p.ip_sign args (nprods + npar) in
-          let db = db_from_ind dbmap npar in
+          let dbmap = parse_ind_args p.ip_sign args (nprods + ndecls) in
+          let db = db_from_ind dbmap ndecls in
           p.ip_types.(j) <-
-            extract_type_cons epar sg db dbmap (EConstr.of_constr t) (npar+1)
+            extract_type_cons epar sg db dbmap (EConstr.of_constr t) (ndecls+1)
         done
     done;
     (* Third pass: we determine special cases. *)
@@ -488,7 +476,7 @@ and extract_really_ind env kn mib =
         if p.ip_logical then raise (I Standard);
         if not (Int.equal (Array.length p.ip_types) 1) then raise (I Standard);
         let typ = p.ip_types.(0) in
-        let l = List.filter (fun t -> not (isTdummy (expand env t))) typ in
+        let l = if conservative_types () then [] else List.filter (fun t -> not (isTdummy (expand env t))) typ in
         if not (keep_singleton ()) &&
             Int.equal (List.length l) 1 && not (type_mem_kn kn (List.hd l))
         then raise (I Singleton);
@@ -531,7 +519,7 @@ and extract_really_ind env kn mib =
           let n = nb_default_params env sg (EConstr.of_constr ty) in
           let check_proj kn = if Cset.mem kn !projs then add_projection n kn ip
           in
-          List.iter (Option.iter check_proj) (lookup_projections ip)
+          List.iter (Option.iter check_proj) (Structures.Structure.find_projections ip)
         with Not_found -> ()
         end;
         Record field_glob
@@ -672,7 +660,10 @@ let rec extract_term env sg mle mlt c args =
         (* we unify it with an fresh copy of the stored type of [Rel n]. *)
         let extract_rel mlt = put_magic (mlt, Mlenv.get mle n) (MLrel n)
         in extract_app env sg mle mlt extract_rel args
-    | Case ({ci_ind=ip},_,c0,br) ->
+    | Case (ci, u, pms, r, iv, c0, br) ->
+        (* If invert_case then this is a match that will get erased later, but right now we don't care. *)
+        let (ip, r, iv, c0, br) = EConstr.expand_case env sg (ci, u, pms, r, iv, c0, br) in
+        let ip = ci.ci_ind in
         extract_app env sg mle mlt (extract_case env sg mle (ip,c0,br)) args
     | Fix ((_,i),recd) ->
         extract_app env sg mle mlt (extract_fix env sg mle i recd) args
@@ -692,6 +683,12 @@ let rec extract_term env sg mle mlt c args =
        extract_app env sg mle mlt extract_var args
     | Int i -> assert (args = []); MLuint i
     | Float f -> assert (args = []); MLfloat f
+    | Array (_u,t,def,_ty) ->
+      assert (args = []);
+      let a = new_meta () in
+      let ml_arr = Array.map (fun c -> extract_term env sg mle a c []) t in
+      let def = extract_term env sg mle a def [] in
+      MLparray(ml_arr, def)
     | Ind _ | Prod _ | Sort _ -> assert false
 
 (*s [extract_maybe_term] is [extract_term] for usual terms, else [MLdummy] *)
@@ -734,7 +731,7 @@ and extract_cst_app env sg mle mlt kn args =
   (* Can we instantiate types variables for this constant ? *)
   (* In Ocaml, inside the definition of this constant, the answer is no. *)
   let instantiated =
-    if lang () == Ocaml && List.mem_f Constant.equal kn !current_fixpoints
+    if lang () == Ocaml && List.mem_f Constant.CanOrd.equal kn !current_fixpoints
     then var2var' (snd schema)
     else instantiation schema
   in
@@ -852,8 +849,8 @@ and extract_case env sg mle ((kn,i) as ip,c,br) mlt =
   end else
     (* [c] has an inductive type, and is not a type scheme type. *)
     let t = type_of env sg c in
-    (* The only non-informative case: [c] is of sort [Prop] *)
-    if (sort_of env sg t) == InProp then
+    (* The only non-informative case: [c] is of sort [Prop]/[SProp] *)
+    if info_of_family (sort_of env sg t) == Logic then
       begin
         add_recursors env kn; (* May have passed unseen if logical ... *)
         (* Logical singleton case: *)
@@ -1016,7 +1013,7 @@ let extract_fixpoint env sg vkn (fi,ti,ci) =
   (* for replacing recursive calls [Rel ..] by the corresponding [Const]: *)
   let sub = List.rev_map EConstr.mkConst kns in
   for i = 0 to n-1 do
-    if sort_of env sg ti.(i) != InProp then
+    if info_of_family (sort_of env sg ti.(i)) != Logic then
       try
         let e,t = extract_std_constant env sg vkn.(i)
                    (EConstr.Vars.substl sub ci.(i)) ti.(i) in
@@ -1061,19 +1058,24 @@ let fake_match_projection env p =
       let x, _, _, _ = info.(snd ind) in
       make_annot (Name x) mip.mind_relevance
   in
-  let indty = mkApp (indu, Context.Rel.to_extended_vect mkRel 0 paramslet) in
+  let indty = mkApp (indu, Context.Rel.instance mkRel 0 paramslet) in
   let rec fold arg j subst = function
     | [] -> assert false
     | LocalAssum (na,ty) :: rem ->
       let ty = Vars.substl subst (liftn 1 j ty) in
       if arg != proj_arg then
         let lab = match na.binder_name with Name id -> Label.of_id id | Anonymous -> assert false in
-        let kn = Projection.Repr.make ind ~proj_npars:mib.mind_nparams ~proj_arg:arg lab in
+        let proj_relevant = match na.binder_relevance with Sorts.Irrelevant -> false | Sorts.Relevant -> true in
+        let kn = Projection.Repr.make ind ~proj_relevant ~proj_npars:mib.mind_nparams ~proj_arg:arg lab in
         fold (arg+1) (j+1) (mkProj (Projection.make kn false, mkRel 1)::subst) rem
       else
-        let p = mkLambda (x, lift 1 indty, liftn 1 2 ty) in
-        let branch = lift 1 (it_mkLambda_or_LetIn (mkRel (List.length ctx - (j-1))) ctx) in
-        let body = mkCase (ci, p, mkRel 1, [|branch|]) in
+        let p = ([|x|], liftn 1 2 ty) in
+        let branch =
+          let nas = Array.of_list (List.rev_map Context.Rel.Declaration.get_annot ctx) in
+          (nas, mkRel (List.length ctx - (j - 1)))
+        in
+        let params = Context.Rel.instance mkRel 1 paramslet in
+        let body = mkCase (ci, u, params, p, NoInvert, mkRel 1, [|branch|]) in
         it_mkLambda_or_LetIn (mkLambda (x,indty,body)) mib.mind_params_ctxt
     | LocalDef (_,c,t) :: rem ->
       let c = liftn 1 j c in
@@ -1110,13 +1112,15 @@ let extract_constant env kn cb =
   in
   try
     match flag_of_type env sg typ with
-    | (Logic,TypeScheme) -> warn_log (); Dtype (r, [], Tdummy Ktype)
+    | (Logic,TypeScheme) -> warn_log ();
+        let s,vl = type_sign_vl env sg typ in
+        Dtype (r, vl, Tdummy Ktype)
     | (Logic,Default) -> warn_log (); Dterm (r, MLdummy Kprop, Tdummy Kprop)
     | (Info,TypeScheme) ->
         (match cb.const_body with
           | Primitive _ | Undef _ -> warn_info (); mk_typ_ax ()
           | Def c ->
-             (match Recordops.find_primitive_projection kn with
+             (match Structures.PrimitiveProjections.find_opt kn with
               | None -> mk_typ (get_body c)
               | Some p ->
                 let body = fake_match_projection env p in
@@ -1129,7 +1133,7 @@ let extract_constant env kn cb =
         (match cb.const_body with
           | Primitive _ | Undef _ -> warn_info (); mk_ax ()
           | Def c ->
-             (match Recordops.find_primitive_projection kn with
+             (match Structures.PrimitiveProjections.find_opt kn with
               | None -> mk_def (get_body c)
               | Some p ->
                 let body = fake_match_projection env p in
@@ -1147,7 +1151,9 @@ let extract_constant_spec env kn cb =
   let typ = EConstr.of_constr cb.const_type in
   try
     match flag_of_type env sg typ with
-    | (Logic, TypeScheme) -> Stype (r, [], Some (Tdummy Ktype))
+    | (Logic, TypeScheme) ->
+        let s,vl = type_sign_vl env sg typ in
+        Stype (r, vl, Some (Tdummy Ktype))
     | (Logic, Default) -> Sval (r, Tdummy Kprop)
     | (Info, TypeScheme) ->
         let s,vl = type_sign_vl env sg typ in
@@ -1212,7 +1218,7 @@ let extract_inductive env kn =
 
 let logical_decl = function
   | Dterm (_,MLdummy _,Tdummy _) -> true
-  | Dtype (_,[],Tdummy _) -> true
+  | Dtype (_,_,Tdummy _) -> true
   | Dfix (_,av,tv) ->
       (Array.for_all isMLdummy av) &&
       (Array.for_all isTdummy tv)
@@ -1222,7 +1228,7 @@ let logical_decl = function
 (*s Is a [ml_spec] logical ? *)
 
 let logical_spec = function
-  | Stype (_, [], Some (Tdummy _)) -> true
+  | Stype (_, _, Some (Tdummy _)) -> true
   | Sval (_,Tdummy _) -> true
   | Sind (_,i) -> Array.for_all (fun ip -> ip.ip_logical) i.ind_packets
   | _ -> false
