@@ -14,9 +14,9 @@ open CErrors
 open Util
 open EConstr
 open Vars
-open Tacmach.New
+open Tacmach
 open Tactics
-open Tacticals.New
+open Tacticals
 open Proofview.Notations
 open Reductionops
 open Formula
@@ -38,11 +38,11 @@ let compare_gr id1 id2 =
   if id1==id2 then 0 else
     if id1==dummy_id then 1
     else if id2==dummy_id then -1
-    else GlobRef.Ordered.compare id1 id2
+    else GlobRef.CanOrd.compare id1 id2
 
 module OrderedInstance=
 struct
-  type t=instance * GlobRef.t
+  type t = Unify.instance * GlobRef.t
   let compare (inst1,id1) (inst2,id2)=
     (compare_instance =? compare_gr) inst2 inst1 id2 id1
     (* we want a __decreasing__ total order *)
@@ -125,7 +125,7 @@ let mk_open_instance env sigma id idc m t =
 
 (* tactics   *)
 
-let left_instance_tac (inst,id) continue seq=
+let left_instance_tac ~flags (inst,id) continue seq=
   let open EConstr in
   Proofview.Goal.enter begin fun gl ->
   let sigma = project gl in
@@ -133,7 +133,7 @@ let left_instance_tac (inst,id) continue seq=
   match inst with
       Phantom dom->
         if lookup env sigma (id,None) seq then
-          tclFAIL 0 (Pp.str "already done")
+          tclFAIL (Pp.str "already done")
         else
           tclTHENS (cut dom)
             [tclTHENLIST
@@ -144,13 +144,13 @@ let left_instance_tac (inst,id) continue seq=
                   generalize [mkApp(idc, [|mkVar id0|])]
                 end);
                 introf;
-                tclSOLVE [wrap 1 false continue
+                tclSOLVE [wrap ~flags 1 false continue
                             (deepen (record (id,None) seq))]];
             tclTRY assumption]
     | Real((m,t),_)->
-        let c = (m, EConstr.to_constr sigma t) in
+        let c = (m, EConstr.to_constr ~abort_on_undefined_evars:false sigma t) in
         if lookup env sigma (id,Some c) seq then
-          tclFAIL 0 (Pp.str "already done")
+          tclFAIL (Pp.str "already done")
         else
           let special_generalize=
             if m>0 then
@@ -174,10 +174,10 @@ let left_instance_tac (inst,id) continue seq=
               [special_generalize;
                introf;
                tclSOLVE
-                 [wrap 1 false continue (deepen (record (id,Some c) seq))]]
+                 [wrap ~flags 1 false continue (deepen (record (id,Some c) seq))]]
   end
 
-let right_instance_tac inst continue seq=
+let right_instance_tac ~flags inst continue seq=
   let open EConstr in
   Proofview.Goal.enter begin fun gl ->
   match inst with
@@ -189,26 +189,26 @@ let right_instance_tac inst continue seq=
               let id0 = List.nth (pf_ids_of_hyps gl) 0 in
               split (Tactypes.ImplicitBindings [mkVar id0])
             end;
-            tclSOLVE [wrap 0 true continue (deepen seq)]];
+            tclSOLVE [wrap ~flags 0 true continue (deepen seq)]];
          tclTRY assumption]
     | Real ((0,t),_) ->
         (tclTHEN (split (Tactypes.ImplicitBindings [t]))
-           (tclSOLVE [wrap 0 true continue (deepen seq)]))
+           (tclSOLVE [wrap ~flags 0 true continue (deepen seq)]))
     | Real ((m,t),_) ->
-        tclFAIL 0 (Pp.str "not implemented ... yet")
+        tclFAIL (Pp.str "not implemented ... yet")
   end
 
-let instance_tac inst=
+let instance_tac ~flags inst=
   if (snd inst)==dummy_id then
-    right_instance_tac (fst inst)
+    right_instance_tac ~flags (fst inst)
   else
-    left_instance_tac inst
+    left_instance_tac ~flags inst
 
-let quantified_tac lf backtrack continue seq =
+let quantified_tac ~flags lf backtrack continue seq =
   Proofview.Goal.enter begin fun gl ->
   let env = Proofview.Goal.env gl in
   let insts=give_instances env (project gl) lf seq in
     tclORELSE
-      (tclFIRST (List.map (fun inst->instance_tac inst continue seq) insts))
+      (tclFIRST (List.map (fun inst->instance_tac ~flags inst continue seq) insts))
       backtrack
   end
