@@ -64,6 +64,8 @@ let change_vars =
           GVar new_id | GEvar _ as x -> x | GPatVar _ as x -> x
         | GApp (rt', rtl) ->
           GApp (change_vars mapping rt', List.map (change_vars mapping) rtl)
+        | GProj (f, rtl, rt) ->
+          GProj (f, List.map (change_vars mapping) rtl, change_vars mapping rt)
         | GLambda (name, k, t, b) ->
           GLambda
             ( name
@@ -106,10 +108,14 @@ let change_vars =
         | GRec _ -> user_err ?loc Pp.(str "Local (co)fixes are not supported")
         | GSort _ as x -> x | GHole _ as x -> x | GInt _ as x -> x
         | GFloat _ as x -> x
-        | GCast (b, c) ->
-          GCast
-            ( change_vars mapping b
-            , Glob_ops.map_cast_type (change_vars mapping) c ))
+        | GCast (b, k, c) ->
+          GCast (change_vars mapping b, k, change_vars mapping c)
+        | GArray (u, t, def, ty) ->
+          GArray
+            ( u
+            , Array.map (change_vars mapping) t
+            , change_vars mapping def
+            , change_vars mapping ty ))
       rt
   and change_vars_br mapping ({CAst.loc; v = idl, patl, res} as br) =
     let new_mapping = List.fold_right Id.Map.remove idl mapping in
@@ -278,10 +284,18 @@ let rec alpha_rt excluded rt =
         , alpha_rt excluded rhs )
     | GRec _ -> user_err Pp.(str "Not handled GRec")
     | (GSort _ | GInt _ | GFloat _ | GHole _) as rt -> rt
-    | GCast (b, c) ->
-      GCast (alpha_rt excluded b, Glob_ops.map_cast_type (alpha_rt excluded) c)
+    | GCast (b, k, c) ->
+      GCast (alpha_rt excluded b, k, alpha_rt excluded c)
     | GApp (f, args) ->
       GApp (alpha_rt excluded f, List.map (alpha_rt excluded) args)
+    | GProj (f, args, c) ->
+      GProj (f, List.map (alpha_rt excluded) args, alpha_rt excluded c)
+    | GArray (u, t, def, ty) ->
+      GArray
+        ( u
+        , Array.map (alpha_rt excluded) t
+        , alpha_rt excluded def
+        , alpha_rt excluded ty )
   in
   new_rt
 
@@ -301,7 +315,7 @@ let is_free_in id =
     DAst.with_loc_val
       (fun ?loc -> function GRef _ -> false | GVar id' -> Id.compare id' id == 0
         | GEvar _ -> false | GPatVar _ -> false
-        | GApp (rt, rtl) -> List.exists is_free_in (rt :: rtl)
+        | GApp (rt, rtl) | GProj (_, rtl, rt) -> List.exists is_free_in (rt :: rtl)
         | GLambda (n, _, t, b) | GProd (n, _, t, b) ->
           let check_in_b =
             match n with Name id' -> not (Id.equal id' id) | _ -> true
@@ -329,9 +343,11 @@ let is_free_in id =
           is_free_in cond || is_free_in br1 || is_free_in br2
         | GRec _ -> user_err Pp.(str "Not handled GRec") | GSort _ -> false
         | GHole _ -> false
-        | GCast (b, (CastConv t | CastVM t | CastNative t)) ->
-          is_free_in b || is_free_in t | GCast (b, CastCoerce) -> is_free_in b
-        | GInt _ | GFloat _ -> false)
+        | GCast (b, _, t) ->
+          is_free_in b || is_free_in t
+        | GInt _ | GFloat _ -> false
+        | GArray (_u, t, def, ty) ->
+          Array.exists is_free_in t || is_free_in def || is_free_in ty)
       x
   and is_free_in_br {CAst.v = ids, _, rt} =
     (not (Id.List.mem id ids)) && is_free_in rt
@@ -364,6 +380,8 @@ let replace_var_by_term x_id term =
         | (GRef _ | GVar _ | GEvar _ | GPatVar _) as rt -> rt
         | GApp (rt', rtl) ->
           GApp (replace_var_by_pattern rt', List.map replace_var_by_pattern rtl)
+        | GProj (f, rtl, rt) ->
+          GProj (f, List.map replace_var_by_pattern rtl, replace_var_by_pattern rt)
         | GLambda (Name id, _, _, _) as rt when Id.compare id x_id == 0 -> rt
         | GLambda (name, k, t, b) ->
           GLambda (name, k, replace_var_by_pattern t, replace_var_by_pattern b)
@@ -404,10 +422,14 @@ let replace_var_by_term x_id term =
         | (GSort _ | GHole _) as rt -> rt
         | GInt _ as rt -> rt
         | GFloat _ as rt -> rt
-        | GCast (b, c) ->
-          GCast
-            ( replace_var_by_pattern b
-            , Glob_ops.map_cast_type replace_var_by_pattern c ))
+        | GArray (u, t, def, ty) ->
+          GArray
+            ( u
+            , Array.map replace_var_by_pattern t
+            , replace_var_by_pattern def
+            , replace_var_by_pattern ty )
+        | GCast (b, k, c) ->
+          GCast (replace_var_by_pattern b, k, replace_var_by_pattern c))
       x
   and replace_var_by_pattern_br ({CAst.loc; v = idl, patl, res} as br) =
     if List.exists (fun id -> Id.compare id x_id == 0) idl then br
@@ -424,7 +446,8 @@ let rec are_unifiable_aux = function
     match (DAst.get l, DAst.get r) with
     | PatVar _, _ | _, PatVar _ -> are_unifiable_aux eqs
     | PatCstr (constructor1, cpl1, _), PatCstr (constructor2, cpl2, _) ->
-      if not (eq_constructor constructor2 constructor1) then raise NotUnifiable
+      if not (Construct.CanOrd.equal constructor2 constructor1) then
+        raise NotUnifiable
       else
         let eqs' =
           try List.combine cpl1 cpl2 @ eqs
@@ -444,7 +467,8 @@ let rec eq_cases_pattern_aux = function
     match (DAst.get l, DAst.get r) with
     | PatVar _, PatVar _ -> eq_cases_pattern_aux eqs
     | PatCstr (constructor1, cpl1, _), PatCstr (constructor2, cpl2, _) ->
-      if not (eq_constructor constructor2 constructor1) then raise NotUnifiable
+      if not (Construct.CanOrd.equal constructor2 constructor1) then
+        raise NotUnifiable
       else
         let eqs' =
           try List.combine cpl1 cpl2 @ eqs
@@ -484,6 +508,7 @@ let expand_as =
       | GVar id as rt -> (
         try DAst.get (Id.Map.find id map) with Not_found -> rt )
       | GApp (f, args) -> GApp (expand_as map f, List.map (expand_as map) args)
+      | GProj (f, args, c) -> GProj (f, List.map (expand_as map) args, expand_as map c)
       | GLambda (na, k, t, b) ->
         GLambda (na, k, expand_as map t, expand_as map b)
       | GProd (na, k, t, b) -> GProd (na, k, expand_as map t, expand_as map b)
@@ -503,14 +528,17 @@ let expand_as =
           , expand_as map br1
           , expand_as map br2 )
       | GRec _ -> user_err Pp.(str "Not handled GRec")
-      | GCast (b, c) ->
-        GCast (expand_as map b, Glob_ops.map_cast_type (expand_as map) c)
+      | GCast (b, k, c) ->
+        GCast (expand_as map b, k, expand_as map c)
       | GCases (sty, po, el, brl) ->
         GCases
           ( sty
           , Option.map (expand_as map) po
           , List.map (fun (rt, t) -> (expand_as map rt, t)) el
-          , List.map (expand_as_br map) brl ))
+          , List.map (expand_as_br map) brl )
+      | GArray (u, t, def, ty) ->
+        GArray
+          (u, Array.map (expand_as map) t, expand_as map def, expand_as map ty))
   and expand_as_br map {CAst.loc; v = idl, cpl, rt} =
     CAst.make ?loc (idl, cpl, expand_as (List.fold_left add_as map cpl) rt)
   in
@@ -547,7 +575,7 @@ let resolve_and_replace_implicits ?(flags = Pretyping.all_and_fail_flags)
         (* we scan the new evar map to find the evar corresponding to this hole (by looking the source *)
         Evd.fold (* to simulate an iter *)
           (fun _ evi _ ->
-            match evi.evar_source with
+            match Evd.evar_source evi with
             | loc_evi, ImplicitArg (gr_evi, p_evi, b_evi) ->
               if
                 GlobRef.equal grk gr_evi && pk = p_evi && bk = b_evi
@@ -559,7 +587,7 @@ let resolve_and_replace_implicits ?(flags = Pretyping.all_and_fail_flags)
         rt
       with Found evi -> (
         (* we found the evar corresponding to this hole *)
-        match evi.evar_body with
+        match Evd.evar_body evi with
         | Evar_defined c ->
           (* we just have to lift the solution in glob_term *)
           Detyping.detype Detyping.Now false Id.Set.empty env ctx (f c)
@@ -571,7 +599,7 @@ let resolve_and_replace_implicits ?(flags = Pretyping.all_and_fail_flags)
           (* we scan the new evar map to find the evar corresponding to this hole (by looking the source *)
           Evd.fold (* to simulate an iter *)
             (fun _ evi _ ->
-              match evi.evar_source with
+              match Evd.evar_source evi with
               | loc_evi, BinderType na' ->
                 if Name.equal na na' && rt.CAst.loc = loc_evi then
                   raise (Found evi)
@@ -581,7 +609,7 @@ let resolve_and_replace_implicits ?(flags = Pretyping.all_and_fail_flags)
           rt
         with Found evi -> (
           (* we found the evar corresponding to this hole *)
-          match evi.evar_body with
+          match Evd.evar_body evi with
           | Evar_defined c ->
             (* we just have to lift the solution in glob_term *)
             Detyping.detype Detyping.Now false Id.Set.empty env ctx (f c)
