@@ -69,15 +69,10 @@ val pr_info : Environ.env -> Evd.evar_map -> function_info -> Pp.t
 val pr_table : Environ.env -> Evd.evar_map -> Pp.t
 
 val observe_tac :
-  (Environ.env -> Evd.evar_map -> Pp.t) -> Tacmach.tactic -> Tacmach.tactic
-
-module New : sig
-  val observe_tac :
-       header:Pp.t
-    -> (Environ.env -> Evd.evar_map -> Pp.t)
-    -> unit Proofview.tactic
-    -> unit Proofview.tactic
-end
+  header:Pp.t
+  -> (Environ.env -> Evd.evar_map -> Pp.t)
+  -> unit Proofview.tactic
+  -> unit Proofview.tactic
 
 (* val function_debug : bool ref  *)
 val observe : Pp.t -> unit
@@ -90,7 +85,7 @@ exception Defining_principle of exn
 exception ToShow of exn
 
 val is_strict_tcc : unit -> bool
-val h_intros : Names.Id.t list -> Tacmach.tactic
+val h_intros : Names.Id.t list -> unit Proofview.tactic
 val h_id : Names.Id.t
 val hrec_id : Names.Id.t
 val acc_inv_id : EConstr.constr Util.delayed
@@ -100,9 +95,9 @@ val acc_rel : EConstr.constr Util.delayed
 val well_founded : EConstr.constr Util.delayed
 
 val evaluable_of_global_reference :
-  GlobRef.t -> Names.evaluable_global_reference
+  GlobRef.t -> Tacred.evaluable_global_reference
 
-val list_rewrite : bool -> (EConstr.constr * bool) list -> Tacmach.tactic
+val list_rewrite : bool -> (EConstr.constr * bool) list -> unit Proofview.tactic
 
 val decompose_lam_n :
      Evd.evar_map
@@ -119,6 +114,3 @@ val compose_prod :
 type tcc_lemma_value = Undefined | Value of Constr.t | Not_needed
 
 val funind_purify : ('a -> 'b) -> 'a -> 'b
-
-val tac_type_of :
-  Goal.goal Evd.sigma -> EConstr.constr -> Goal.goal Evd.sigma * EConstr.types
