@@ -25,11 +25,11 @@ exception CannotCoerceTo of string
 let base_val_typ wit =
   match val_tag (topwit wit) with Val.Base t -> t | _ -> CErrors.anomaly (Pp.str "Not a base val.")
 
-let (wit_constr_context : (Empty.t, Empty.t, EConstr.constr) Genarg.genarg_type) =
+let (wit_constr_context : (Empty.t, Empty.t, Constr_matching.context) Genarg.genarg_type) =
   let wit = Genarg.create_arg "constr_context" in
   let () = register_val0 wit None in
-  let () = Genprint.register_val_print0 (base_val_typ wit)
-    (Pptactic.make_constr_printer Printer.pr_econstr_n_env) in
+  let pr env sigma lev c : Pp.t = Printer.pr_econstr_n_env env sigma lev (Constr_matching.repr_context c) in
+  let () = Genprint.register_val_print0 (base_val_typ wit) (Pptactic.make_constr_printer pr) in
   wit
 
 (* includes idents known to be bound and references *)
@@ -91,6 +91,13 @@ let to_int v =
     Some (out_gen (topwit wit_int) v)
   else None
 
+let of_ident id = in_gen (topwit wit_ident) id
+
+let to_ident v =
+  if has_type v (topwit wit_ident) then
+    Some (out_gen (topwit wit_ident) v)
+  else None
+
 let to_list v = prj Val.typ_list v
 
 let to_option v = prj Val.typ_opt v
@@ -161,8 +168,8 @@ let coerce_var_to_ident fresh env sigma v =
     match out_gen (topwit wit_intro_pattern) v with
     | { CAst.v=IntroNaming (IntroIdentifier id)} -> id
     | _ -> fail ()
-  else if has_type v (topwit wit_var) then
-    out_gen (topwit wit_var) v
+  else if has_type v (topwit wit_hyp) then
+    out_gen (topwit wit_hyp) v
   else match Value.to_constr v with
   | None -> fail ()
   | Some c ->
@@ -184,8 +191,8 @@ let id_of_name = function
   | Some (IntroNaming (IntroIdentifier id)) -> id
   | Some _ -> fail ()
   | None ->
-  if has_type v (topwit wit_var) then
-    out_gen (topwit wit_var) v
+  if has_type v (topwit wit_hyp) then
+    out_gen (topwit wit_hyp) v
   else
     match Value.to_constr v with
     | None -> fail ()
@@ -222,8 +229,8 @@ let coerce_to_intro_pattern sigma v =
   match is_intro_pattern v with
   | Some pat -> pat
   | None ->
-  if has_type v (topwit wit_var) then
-    let id = out_gen (topwit wit_var) v in
+  if has_type v (topwit wit_hyp) then
+    let id = out_gen (topwit wit_hyp) v in
     IntroNaming (IntroIdentifier id)
   else match Value.to_constr v with
   | Some c when isVar sigma c ->
@@ -259,8 +266,8 @@ let coerce_to_constr env v =
     ([], c)
   else if has_type v (topwit wit_constr_under_binders) then
     out_gen (topwit wit_constr_under_binders) v
-  else if has_type v (topwit wit_var) then
-    let id = out_gen (topwit wit_var) v in
+  else if has_type v (topwit wit_hyp) then
+    let id = out_gen (topwit wit_hyp) v in
     (try [], constr_of_id env id with Not_found -> fail ())
   else fail ()
 
@@ -276,14 +283,15 @@ let coerce_to_closed_constr env v =
   c
 
 let coerce_to_evaluable_ref env sigma v =
+  let open Tacred in
   let fail () = raise (CannotCoerceTo "an evaluable reference") in
   let ev =
   match is_intro_pattern v with
   | Some (IntroNaming (IntroIdentifier id)) when is_variable env id -> EvalVarRef id
   | Some _ -> fail ()
   | None ->
-  if has_type v (topwit wit_var) then
-    let id = out_gen (topwit wit_var) v in
+  if has_type v (topwit wit_hyp) then
+    let id = out_gen (topwit wit_hyp) v in
     if Id.List.mem id (Termops.ids_of_context env) then EvalVarRef id
     else fail ()
   else if has_type v (topwit wit_ref) then
@@ -328,8 +336,8 @@ let coerce_to_hyp env sigma v =
   | Some (IntroNaming (IntroIdentifier id)) when is_variable env id -> id
   | Some _ -> fail ()
   | None ->
-  if has_type v (topwit wit_var) then
-    let id = out_gen (topwit wit_var) v in
+  if has_type v (topwit wit_hyp) then
+    let id = out_gen (topwit wit_hyp) v in
     if is_variable env id then id else fail ()
   else match Value.to_constr v with
   | Some c when isVar sigma c -> destVar sigma c
@@ -357,16 +365,16 @@ let coerce_to_reference sigma v =
 (* (as in Inversion) *)
 let coerce_to_quantified_hypothesis sigma v =
   match is_intro_pattern v with
-  | Some (IntroNaming (IntroIdentifier id)) -> NamedHyp id
+  | Some (IntroNaming (IntroIdentifier id)) -> NamedHyp (CAst.make id)
   | Some _ -> raise (CannotCoerceTo "a quantified hypothesis")
   | None ->
-  if has_type v (topwit wit_var) then
-    let id = out_gen (topwit wit_var) v in
-    NamedHyp id
+  if has_type v (topwit wit_hyp) then
+    let id = out_gen (topwit wit_hyp) v in
+    NamedHyp (CAst.make id)
   else if has_type v (topwit wit_int) then
     AnonHyp (out_gen (topwit wit_int) v)
   else match Value.to_constr v with
-  | Some c when isVar sigma c -> NamedHyp (destVar sigma c)
+  | Some c when isVar sigma c -> NamedHyp (CAst.make @@ destVar sigma c)
   | _ -> raise (CannotCoerceTo "a quantified hypothesis")
 
 (* Quantified named or numbered hypothesis or hypothesis in context *)
@@ -394,8 +402,13 @@ type appl =
 
 (* Values for interpretation *)
 type tacvalue =
-  | VFun of appl*Tacexpr.ltac_trace * Val.t Id.Map.t *
-      Name.t list * Tacexpr.glob_tactic_expr
+  | VFun of
+      appl *
+      Tacexpr.ltac_trace *
+      Loc.t option * (* when executing a global Ltac function: the location where this function was called *)
+      Val.t Id.Map.t * (* closure *)
+      Name.t list * (* binders *)
+      Tacexpr.glob_tactic_expr (* body *)
   | VRec of Val.t Id.Map.t ref * Tacexpr.glob_tactic_expr
 
 let (wit_tacvalue : (Empty.t, tacvalue, tacvalue) Genarg.genarg_type) =
@@ -424,7 +437,15 @@ let pr_value env v =
   | TopPrinterNeedsContextAndLevel { default_already_surrounded; printer } ->
      pr_with_env (fun env sigma -> printer env sigma default_already_surrounded)
 
-let error_ltac_variable ?loc id env v s =
-   CErrors.user_err ?loc  (str "Ltac variable " ++ Id.print id ++
+exception CoercionError of Id.t * (Environ.env * Evd.evar_map) option * Val.t * string
+
+let () = CErrors.register_handler begin function
+| CoercionError (id, env, v, s) ->
+  Some (str "Ltac variable " ++ Id.print id ++
    strbrk " is bound to" ++ spc () ++ pr_value env v ++ spc () ++
    strbrk "which cannot be coerced to " ++ str s ++ str".")
+| _ -> None
+end
+
+let error_ltac_variable ?loc id env v s =
+  Loc.raise ?loc (CoercionError (id, env, v, s))
