@@ -27,8 +27,7 @@ open Ssrmatching
 open Ssrast
 open Ssrprinters
 open Ssrcommon
-open Tacticals
-open Tacmach
+open Proofview.Notations
 
 let ssroldreworder = Summary.ref ~name:"SSR:oldreworder" false
 let () =
@@ -67,16 +66,16 @@ let safe_simpltac n =
 
 let simpltac = function
   | Simpl n -> safe_simpltac n
-  | Cut n -> Tacticals.New.tclTRY (donetac n)
-  | SimplCut (n,m) -> Tacticals.New.tclTHEN (safe_simpltac m) (Tacticals.New.tclTRY (donetac n))
-  | Nop -> Tacticals.New.tclIDTAC
+  | Cut n -> Tacticals.tclTRY (donetac n)
+  | SimplCut (n,m) -> Tacticals.tclTHEN (safe_simpltac m) (Tacticals.tclTRY (donetac n))
+  | Nop -> Tacticals.tclIDTAC
 
 let simpltac s = Proofview.Goal.enter (fun _ -> simpltac s)
 
 (** The "congr" tactic *)
 
-let interp_congrarg_at ist gl n rf ty m =
-  ppdebug(lazy Pp.(str"===interp_congrarg_at==="));
+let interp_congrarg_at env sigma ist ~concl n rf ty m =
+  debug_ssr (fun () -> Pp.(str"===interp_congrarg_at==="));
   let congrn, _ = mkSsrRRef "nary_congruence" in
   let args1 = mkRnat n :: mkRHoles n @ [ty] in
   let args2 = mkRHoles (3 * n) in
@@ -84,82 +83,100 @@ let interp_congrarg_at ist gl n rf ty m =
     if i + n > m then None else
     try
       let rt = mkRApp congrn (args1 @  mkRApp rf (mkRHoles i) :: args2) in
-      ppdebug(lazy Pp.(str"rt=" ++ Printer.pr_glob_constr_env (pf_env gl) rt));
-      Some (interp_refine ist gl rt)
+      debug_ssr (fun () -> Pp.(str"rt=" ++ Printer.pr_glob_constr_env env sigma rt));
+      Some (interp_refine env sigma ist ~concl rt)
     with _ -> loop (i + 1) in
   loop 0
 
 let pattern_id = mk_internal_id "pattern value"
 
-let congrtac ((n, t), ty) ist gl =
-  ppdebug(lazy (Pp.str"===congr==="));
-  ppdebug(lazy Pp.(str"concl=" ++ Printer.pr_econstr_env (pf_env gl) (project gl) (Tacmach.pf_concl gl)));
-  let sigma, _ as it = interp_term (pf_env gl) (project gl) ist t in
-  let gl = pf_merge_uc_of sigma gl in
-  let _, f, _, _ucst = pf_abs_evars gl it in
+let congrtac ((n, t), ty) ist =
+  Proofview.Goal.enter begin fun gl ->
+  let env = Proofview.Goal.env gl in
+  let sigma = Proofview.Goal.sigma gl in
+  let concl = Proofview.Goal.concl gl in
+  debug_ssr (fun () -> (Pp.str"===congr==="));
+  debug_ssr (fun () -> Pp.(str"concl=" ++ Printer.pr_econstr_env env sigma concl));
+  let nsigma, _ as it = interp_term env sigma ist t in
+  let sigma = Evd.merge_universe_context sigma (Evd.evar_universe_context nsigma) in
+  let f, _, _ucst = abs_evars env sigma it in
   let ist' = {ist with lfun =
     Id.Map.add pattern_id (Tacinterp.Value.of_constr f) Id.Map.empty } in
   let rf = mkRltacVar pattern_id in
-  let m = pf_nbargs (pf_env gl) (project gl) f in
-  let _, cf = if n > 0 then
-    match interp_congrarg_at ist' gl n rf ty m with
+  let m = pf_nbargs env sigma f in
+  let cf = if n > 0 then
+    match interp_congrarg_at env sigma ist' ~concl n rf ty m with
     | Some cf -> cf
     | None -> errorstrm Pp.(str "No " ++ int n ++ str "-congruence with "
                          ++ pr_term t)
     else let rec loop i =
       if i > m then errorstrm Pp.(str "No congruence with " ++ pr_term t)
-      else match interp_congrarg_at ist' gl i rf ty m with
+      else match interp_congrarg_at env sigma ist' ~concl i rf ty m with
       | Some cf -> cf
       | None -> loop (i + 1) in
       loop 1 in
-  Proofview.V82.of_tactic Tacticals.New.(tclTHEN (refine_with cf) (tclTRY Tactics.reflexivity)) gl
+  Tacticals.(tclTHEN (refine_with cf) (tclTRY Tactics.reflexivity))
+  end
 
-let pf_typecheck t gl =
-  let it = sig_it gl in
-  let sigma,_  = pf_type_of gl t in
-  re_sig [it] sigma
+let pf_typecheck t =
+  Proofview.Goal.enter begin fun gl ->
+  let sigma, _  = Typing.type_of (Proofview.Goal.env gl) (Proofview.Goal.sigma gl) t in
+  Proofview.Unsafe.tclEVARS sigma
+  end
+
+let tclMATCH_GOAL env sigma c t_ok t_fail =
+  Proofview.Goal.enter begin fun gl ->
+  match unify_HO env sigma (Proofview.Goal.concl gl) c with
+  | sigma ->
+    Proofview.Unsafe.tclEVARS sigma <*>
+    convert_concl ~check:true (Reductionops.nf_evar sigma c) <*> t_ok sigma
+  | exception exn when CErrors.noncritical exn -> t_fail ()
+  end
+
+let mk_evar env sigma ty =
+  let sigma = Evd.create_evar_defs sigma in
+  let (sigma, x) = Evarutil.new_evar env sigma ty in
+  sigma, x
 
 let newssrcongrtac arg ist =
   let open Proofview.Notations in
   Proofview.Goal.enter_one ~__LOC__ begin fun _g ->
     (Ssrcommon.tacMK_SSR_CONST "ssr_congr_arrow") end >>= fun arr ->
-  Proofview.V82.tactic begin fun gl ->
-  ppdebug(lazy Pp.(str"===newcongr==="));
-  ppdebug(lazy Pp.(str"concl=" ++ Printer.pr_econstr_env (pf_env gl) (project gl) (pf_concl gl)));
+  Proofview.Goal.enter begin fun gl ->
+  let env = Proofview.Goal.env gl in
+  let sigma = Proofview.Goal.sigma gl in
+  let concl = Proofview.Goal.concl gl in
+  debug_ssr (fun () -> Pp.(str"===newcongr==="));
+  debug_ssr (fun () -> Pp.(str"concl=" ++ Printer.pr_econstr_env env sigma concl));
   (* utils *)
-  let fs gl t = Reductionops.nf_evar (project gl) t in
-  let tclMATCH_GOAL (c, gl_c) proj t_ok t_fail gl =
-    match try Some (pf_unify_HO gl_c (pf_concl gl) c)
-          with exn when CErrors.noncritical exn -> None with
-    | Some gl_c ->
-        tclTHEN (Proofview.V82.of_tactic (convert_concl ~check:true (fs gl_c c)))
-          (t_ok (proj gl_c)) gl
-    | None -> t_fail () gl in
-  let mk_evar gl ty =
-    let env, sigma, si = pf_env gl, project gl, sig_it gl in
-    let sigma = Evd.create_evar_defs sigma in
-    let (sigma, x) = Evarutil.new_evar env sigma ty in
-    x, re_sig si sigma in
+  let fs sigma t = Reductionops.nf_evar sigma t in
   let ssr_congr lr = EConstr.mkApp (arr, lr) in
-  let eq, gl = pf_fresh_global Coqlib.(lib_ref "core.eq.type") gl in
+  let sigma, eq = Evd.fresh_global env sigma Coqlib.(lib_ref "core.eq.type") in
   (* here the two cases: simple equality or arrow *)
-  let equality, _, eq_args, gl' = pf_saturate gl (EConstr.of_constr eq) 3 in
-  tclMATCH_GOAL (equality, gl') (fun gl' -> fs gl' (List.assoc 0 eq_args))
-  (fun ty -> congrtac (arg, Detyping.detype Detyping.Now false Id.Set.empty (pf_env gl) (project gl) ty) ist)
+  let equality, _, eq_args, sigma' = saturate env sigma eq 3 in
+  Proofview.Unsafe.tclEVARS sigma <*>
+  tclMATCH_GOAL env sigma' equality
+  (fun sigma' ->
+    let x = List.find_map (fun (n, x, _) -> if n = 0 then Some x else None) eq_args in
+    let ty = fs sigma' x in
+    congrtac (arg, Detyping.detype Detyping.Now false Id.Set.empty env sigma ty) ist)
   (fun () ->
-    let gl', t_lhs = pfe_new_type gl in
-    let gl', t_rhs = pfe_new_type gl' in
-    let lhs, gl' = mk_evar gl' t_lhs in
-    let rhs, gl' = mk_evar gl' t_rhs in
+    try
+    let sigma, t_lhs = Evarutil.new_Type sigma in
+    let sigma, t_rhs = Evarutil.new_Type sigma in
+    let sigma, lhs = mk_evar env sigma t_lhs in
+    let sigma, rhs = mk_evar env sigma t_rhs in
     let arrow = EConstr.mkArrow lhs Sorts.Relevant (EConstr.Vars.lift 1 rhs) in
-    tclMATCH_GOAL (arrow, gl') (fun gl' -> [|fs gl' lhs;fs gl' rhs|])
-    (fun lr ->
+    tclMATCH_GOAL env sigma arrow
+    (fun sigma ->
+      let lr = [|fs sigma lhs;fs sigma rhs|] in
       let a = ssr_congr lr in
-      tclTHENLIST [ pf_typecheck a
-                  ; Proofview.V82.of_tactic (Tactics.apply a)
+      Tacticals.tclTHENLIST [ pf_typecheck a
+                  ; Tactics.apply a
                   ; congrtac (arg, mkRType) ist ])
-    (fun _ _ -> errorstrm Pp.(str"Conclusion is not an equality nor an arrow")))
-    gl
+    (fun _ -> Tacticals.tclZEROMSG Pp.(str"Conclusion is not an equality nor an arrow"))
+    with e -> Proofview.tclZERO e (* FIXME *)
+    )
   end
 
 (** 7. Rewriting tactics (rewrite, unlock) *)
@@ -210,11 +227,11 @@ let simplintac occ rdx sim =
       simpltac (Simpl m)
     end else
       let sigma0, concl0, env0 = Proofview.Goal.(sigma gl, concl gl, env gl) in
-      let simp env c _ _ = EConstr.Unsafe.to_constr (red_safe Tacred.simpl env sigma0 (EConstr.of_constr c)) in
-      convert_concl_no_check (EConstr.of_constr (eval_pattern env0 sigma0 (EConstr.to_constr ~abort_on_undefined_evars:false sigma0 concl0) rdx occ simp))
+      let simp env c _ _ = red_safe Tacred.simpl env sigma0 c in
+      convert_concl_no_check (eval_pattern env0 sigma0 (Reductionops.nf_evar sigma0 concl0) rdx occ simp)
     end
   in
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter begin fun _ ->
   match sim with
   | Simpl m -> simptac m
@@ -223,16 +240,16 @@ let simplintac occ rdx sim =
   end
 
 let rec get_evalref env sigma c = match EConstr.kind sigma c with
-  | Var id -> EvalVarRef id
-  | Const (k,_) -> EvalConstRef k
+  | Var id -> Tacred.EvalVarRef id
+  | Const (k,_) -> Tacred.EvalConstRef k
   | App (c', _) -> get_evalref env sigma c'
   | Cast (c', _, _) -> get_evalref env sigma c'
-  | Proj(c,_) -> EvalConstRef(Projection.constant c)
+  | Proj(c,_) -> Tacred.EvalConstRef(Projection.constant c)
   | _ -> errorstrm Pp.(str "The term " ++ pr_econstr_pat (Global.env ()) sigma c ++ str " is not unfoldable")
 
 (* Strip a pattern generated by a prenex implicit to its constant. *)
 let strip_unfold_term _ ((sigma, t) as p) kt = match EConstr.kind sigma t with
-  | App (f, a) when kt = xNoFlag && Array.for_all (EConstr.isEvar sigma) a && EConstr.isConst sigma f ->
+  | App (f, a) when kt = NoFlag && Array.for_all (EConstr.isEvar sigma) a && EConstr.isConst sigma f ->
     (sigma, f), true
   | Const _ | Var _ -> p, true
   | Proj _ -> p, true
@@ -240,39 +257,39 @@ let strip_unfold_term _ ((sigma, t) as p) kt = match EConstr.kind sigma t with
 
 let same_proj sigma t1 t2 =
   match EConstr.kind sigma t1, EConstr.kind sigma t2 with
-  | Proj(c1,_), Proj(c2, _) -> Projection.equal c1 c2
+  | Proj(c1,_), Proj(c2, _) -> Projection.CanOrd.equal c1 c2
   | _ -> false
 
-let all_ok _ _ = true
-
-let fake_pmatcher_end () =
-  mkProp, L2R, (Evd.empty, UState.empty, mkProp)
+let classify_pattern p = match p with
+| None -> None
+| Some p -> redex_of_pattern p
 
 let unfoldintac occ rdx t (kt,_) =
-  Proofview.V82.tactic begin fun gl ->
+  Proofview.Goal.enter begin fun gl ->
   let fs sigma x = Reductionops.nf_evar sigma x in
-  let sigma0, concl0, env0 = project gl, pf_concl gl, pf_env gl in
+  let sigma0 = Proofview.Goal.sigma gl in
+  let env0 = Proofview.Goal.env gl in
+  let concl0 = Proofview.Goal.concl gl in
   let (sigma, t), const = strip_unfold_term env0 t kt in
   let body env t c =
     Tacred.unfoldn [AllOccurrences, get_evalref env sigma t] env sigma0 c in
   let easy = occ = None && rdx = None in
   let red_flags = if easy then CClosure.betaiotazeta else CClosure.betaiota in
   let beta env = Reductionops.clos_norm_flags red_flags env sigma0 in
-  let unfold, conclude = match rdx with
-  | Some (_, (In_T _ | In_X_In_T _)) | None ->
-    let ise = Evd.create_evar_defs sigma in
-    let ise, u = mk_tpattern env0 sigma0 (ise,EConstr.Unsafe.to_constr t) all_ok L2R (EConstr.Unsafe.to_constr t) in
-    let find_T, end_T =
-      mk_tpattern_matcher ~raise_NoMatch:true sigma0 occ (ise,[u]) in
+  let unfold, conclude = match classify_pattern rdx with
+  | None ->
+    let rigid ev = Evd.mem sigma0 ev in
+    let u = mk_tpattern env0 ~rigid t L2R t (empty_tpatterns (Evd.create_evar_defs sigma)) in
+    let find_T, end_T = mk_tpattern_matcher ~raise_NoMatch:true sigma0 occ u in
     (fun env c _ h ->
-      try find_T env c h ~k:(fun env c _ _ -> EConstr.Unsafe.to_constr (body env t (EConstr.of_constr c)))
+      try find_T env c h ~k:(fun env c _ _ -> body env t c)
       with NoMatch when easy -> c
       | NoMatch | NoProgress -> errorstrm Pp.(str"No occurrence of "
-        ++ pr_econstr_pat env sigma0 t ++ spc() ++ str "in " ++ Printer.pr_constr_env env sigma c)),
-    (fun () -> try end_T () with
-      | NoMatch when easy -> fake_pmatcher_end ()
+        ++ pr_econstr_pat env sigma0 t ++ spc() ++ str "in " ++ Printer.pr_econstr_env env sigma c)),
+    (fun () -> try ignore @@ end_T () with
+      | NoMatch when easy -> ()
       | NoMatch -> anomaly "unfoldintac")
-  | _ ->
+  | Some _ ->
     (fun env (c as orig_c) _ h ->
       if const then
         let rec aux c =
@@ -288,48 +305,47 @@ let unfoldintac occ rdx t (kt,_) =
             | Proj _ when same_proj sigma0 c t -> body env t c
             | Const f -> aux (body env c c)
             | App (f, a) -> aux (EConstr.mkApp (body env f f, a))
-            | _ -> errorstrm Pp.(str "The term "++ pr_constr_env env sigma orig_c++
+            | _ -> errorstrm Pp.(str "The term "++ pr_econstr_env env sigma orig_c ++
                 str" contains no " ++ pr_econstr_env env sigma t ++ str" even after unfolding")
-          in EConstr.Unsafe.to_constr @@ aux (EConstr.of_constr c)
+          in aux c
       else
-        try EConstr.Unsafe.to_constr @@ body env t (fs (unify_HO env sigma (EConstr.of_constr c) t) t)
+        try body env t (fs (unify_HO env sigma c t) t)
         with _ -> errorstrm Pp.(str "The term " ++
-          pr_constr_env env sigma c ++spc()++ str "does not unify with " ++ pr_econstr_pat env sigma t)),
-    fake_pmatcher_end in
+          pr_econstr_env env sigma c ++spc()++ str "does not unify with " ++ pr_econstr_pat env sigma t)),
+    ignore in
   let concl =
-    let concl0 = EConstr.Unsafe.to_constr concl0 in
-    try beta env0 (EConstr.of_constr (eval_pattern env0 sigma0 concl0 rdx occ unfold))
+    try beta env0 (eval_pattern env0 sigma0 concl0 rdx occ unfold)
     with Option.IsNone -> errorstrm Pp.(str"Failed to unfold " ++ pr_econstr_pat env0 sigma t) in
-  let _ = conclude () in
-  Proofview.V82.of_tactic (convert_concl ~check:true concl) gl
+  let () = conclude () in
+  convert_concl ~check:true concl
   end
 
 let foldtac occ rdx ft =
-  Proofview.V82.tactic begin fun gl ->
-  let sigma0, concl0, env0 = project gl, pf_concl gl, pf_env gl in
+  Proofview.Goal.enter begin fun gl ->
+  let sigma0 = Proofview.Goal.sigma gl in
+  let env0 = Proofview.Goal.env gl in
+  let concl0 = Proofview.Goal.concl gl in
   let sigma, t = ft in
-  let t = EConstr.to_constr ~abort_on_undefined_evars:false sigma t in
-  let fold, conclude = match rdx with
-  | Some (_, (In_T _ | In_X_In_T _)) | None ->
-    let ise = Evd.create_evar_defs sigma in
-    let ut = EConstr.Unsafe.to_constr (red_product_skip_id env0 sigma (EConstr.of_constr t)) in
-    let ise, ut = mk_tpattern env0 sigma0 (ise,t) all_ok L2R ut in
-    let find_T, end_T =
-      mk_tpattern_matcher ~raise_NoMatch:true sigma0 occ (ise,[ut]) in
+  let t = Reductionops.nf_evar sigma t in
+  let fold, conclude = match classify_pattern rdx with
+  | None ->
+    let ut = red_product_skip_id env0 sigma t in
+    let rigid ev = Evd.mem sigma0 ev in
+    let ut = mk_tpattern ~rigid env0 t L2R ut (empty_tpatterns (Evd.create_evar_defs sigma)) in
+    let find_T, end_T = mk_tpattern_matcher ~raise_NoMatch:true sigma0 occ ut in
     (fun env c _ h -> try find_T env c h ~k:(fun env t _ _ -> t) with NoMatch ->c),
-    (fun () -> try end_T () with NoMatch -> fake_pmatcher_end ())
-  | _ ->
+    (fun () -> try ignore @@ end_T () with NoMatch -> ())
+  | Some _ ->
     (fun env c _ h ->
        try
-         let sigma = unify_HO env sigma (EConstr.of_constr c) (EConstr.of_constr t) in
-         EConstr.to_constr ~abort_on_undefined_evars:false sigma (EConstr.of_constr t)
-    with _ -> errorstrm Pp.(str "fold pattern " ++ pr_constr_pat env sigma t ++ spc ()
-      ++ str "does not match redex " ++ pr_constr_pat env sigma c)),
-    fake_pmatcher_end in
-  let concl0 = EConstr.Unsafe.to_constr concl0 in
+         let sigma = unify_HO env sigma c t in
+         Reductionops.nf_evar sigma t
+    with _ -> errorstrm Pp.(str "fold pattern " ++ pr_econstr_pat env sigma t ++ spc ()
+      ++ str "does not match redex " ++ pr_econstr_pat env sigma c)),
+    ignore in
   let concl = eval_pattern env0 sigma0 concl0 rdx occ fold in
-  let _ = conclude () in
-  Proofview.V82.of_tactic (convert_concl ~check:true (EConstr.of_constr concl)) gl
+  let () = conclude () in
+  convert_concl ~check:true concl
   end
 
 let converse_dir = function L2R -> R2L | R2L -> L2R
@@ -354,42 +370,62 @@ exception PRtype_error of (Environ.env * Evd.evar_map * Pretype_errors.pretype_e
 
 let id_map_redex _ sigma ~before:_ ~after = sigma, after
 
+(* Invariants expected from the arguments:
+    ⊢ rdx : rdx_ty
+    pattern_id : rdx_ty ⊢ pred : Type@{s}
+    ⊢ new_rdx : rdx_ty
+    ⊢ c : c_ty
+    ⊢ c_ty ≡ EQN rdx_ty rdx new_rdx
+*)
 let pirrel_rewrite ?(under=false) ?(map_redex=id_map_redex) pred rdx rdx_ty new_rdx dir (sigma, c) c_ty =
-  Proofview.V82.tactic begin fun gl ->
+  let open Tacmach in
+  let open Tacticals in
+  Proofview.Goal.enter begin fun gl ->
 (*   ppdebug(lazy(str"sigma@pirrel_rewrite=" ++ pr_evar_map None sigma)); *)
   let env = pf_env gl in
   let beta = Reductionops.clos_norm_flags CClosure.beta env sigma in
   let sigma, new_rdx = map_redex env sigma ~before:rdx ~after:new_rdx in
   let sigma, p = (* The resulting goal *)
     Evarutil.new_evar env sigma (beta (EConstr.Vars.subst1 new_rdx pred)) in
-  let pred = EConstr.mkNamedLambda (make_annot pattern_id Sorts.Relevant) rdx_ty pred in
+  let pred = EConstr.mkNamedLambda sigma (make_annot pattern_id Sorts.Relevant) rdx_ty pred in
   let sigma, elim =
-    let sort = elimination_sort_of_goal gl in
+    let sort = Tacticals.elimination_sort_of_goal gl in
     match Equality.eq_elimination_ref (dir = L2R) sort with
     | Some r -> Evd.fresh_global env sigma r
     | None ->
-      let ((kn, i) as ind, _), unfolded_c_ty = Tacred.reduce_to_quantified_ind env sigma c_ty in
-      let sort = elimination_sort_of_goal gl in
+      let ((kn, i) as ind, _) = Tacred.eval_to_quantified_ind env sigma c_ty in
+      let sort = Tacticals.elimination_sort_of_goal gl in
       let sigma, elim = Evd.fresh_global env sigma (Indrec.lookup_eliminator env ind sort) in
       if dir = R2L then sigma, elim else
       let elim, _ = EConstr.destConst sigma elim in
-      let mp,l = Constant.repr2 (Constant.make1 (Constant.canonical elim)) in
+      let mp,l = KerName.repr (Constant.canonical elim) in
       let l' = Label.of_id (Nameops.add_suffix (Label.to_id l) "_r")  in
       let c1' = Global.constant_of_delta_kn (Constant.canonical (Constant.make2 mp l')) in
       sigma, EConstr.of_constr (mkConst c1')
   in
-  let proof = EConstr.mkApp (elim, [| rdx_ty; new_rdx; pred; p; rdx; c |]) in
-  (* We check the proof is well typed *)
-  let sigma, proof_ty =
-    try Typing.type_of env sigma proof with
+  (* We check the proof is well typed. We assume that the type of [elim] is of
+     the form [forall (A : Type) (x : A) (P : A -> Type@{s}), T] s.t. the only
+     universes to unify are by checking the [A] and [P] arguments. *)
+  let sigma =
+    try
+      let open EConstr in
+      let elimT = Retyping.get_type_of env sigma elim in
+      let (idA, tA, elimT) = destProd sigma elimT in
+      let (_, _, elimT) = destProd sigma elimT in
+      let (idP, tP, _) = destProd sigma elimT in
+      let sigma = Typing.check env sigma rdx_ty tA in
+      let tP = mkLetIn (idA, rdx_ty, tA, mkLetIn (anonR, mkProp, mkType Univ.Universe.type1, tP)) in
+      let sigma = Typing.check env sigma pred tP in
+      sigma
+    with
     | Pretype_errors.PretypeError (env, sigma, te) -> raise (PRtype_error (Some (env, sigma, te)))
     | e when CErrors.noncritical e -> raise (PRtype_error None)
   in
-  ppdebug(lazy Pp.(str"pirrel_rewrite: proof term: " ++ pr_econstr_env env sigma proof));
-  ppdebug(lazy Pp.(str"pirrel_rewrite of type: " ++ pr_econstr_env env sigma proof_ty));
-  try Proofview.V82.of_tactic (refine_with
-    ~first_goes_last:(not !ssroldreworder || under) ~with_evars:under (sigma, proof)) gl
-  with e when CErrors.noncritical e ->
+  let proof = EConstr.mkApp (elim, [| rdx_ty; new_rdx; pred; p; rdx; c |]) in
+  debug_ssr (fun () -> Pp.(str"pirrel_rewrite: proof term: " ++ pr_econstr_env env sigma proof));
+  Proofview.tclORELSE (refine_with
+    ~first_goes_last:(not !ssroldreworder || under) ~with_evars:under (sigma, proof))
+  (fun e ->
     (* we generate a msg like: "Unable to find an instance for the variable" *)
     let hd_ty, miss = match EConstr.kind sigma c with
     | App (hd, args) ->
@@ -413,41 +449,38 @@ let pirrel_rewrite ?(under=false) ?(map_redex=id_map_redex) pred rdx rdx_ty new_
           if open_evs <> [] then Some name else None)
           (List.combine (Array.to_list args) names)
     | _ -> anomaly "rewrite rule not an application" in
-    errorstrm Pp.(Himsg.explain_refiner_error env sigma (Logic.UnresolvedBindings miss)++
-      (Pp.fnl()++str"Rule's type:" ++ spc() ++ pr_econstr_env env sigma hd_ty))
+      tclZEROMSG Pp.(Himsg.explain_refiner_error env sigma (Logic.UnresolvedBindings miss)++
+      (Pp.fnl()++str"Rule's type:" ++ spc() ++ pr_econstr_env env sigma hd_ty)))
   end
 
 let pf_merge_uc_of s sigma =
   Evd.merge_universe_context sigma (Evd.evar_universe_context s)
 
-let rwcltac ?under ?map_redex cl rdx dir sr =
+let rwcltac ?under ?map_redex cl rdx dir (sigma, r) =
   let open Proofview.Notations in
   Proofview.Goal.enter begin fun gl ->
   let env = Proofview.Goal.env gl in
   let sigma0 = Proofview.Goal.sigma gl in
-  let sr =
-    let sigma, r = sr in
-    let sigma = resolve_typeclasses ~where:r ~fail:false env sigma in
-    sigma, r in
-  let n, r_n,_, ucst = abs_evars env sigma0 sr in
+  let sigma = resolve_typeclasses ~where:r ~fail:false env sigma in
+  let r_n, evs, ucst = abs_evars env sigma0 (sigma, r) in
+  let n = List.length evs in
   let r_n' = abs_cterm env sigma0 n r_n in
-  let r' = EConstr.Vars.subst_var pattern_id r_n' in
+  let r' = EConstr.Vars.subst_var sigma pattern_id r_n' in
   let sigma0 = Evd.set_universe_context sigma0 ucst in
-  let rdxt = Retyping.get_type_of env (fst sr) rdx in
-(*         ppdebug(lazy(str"sigma@rwcltac=" ++ pr_evar_map None (fst sr))); *)
-        ppdebug(lazy Pp.(str"r@rwcltac=" ++ pr_econstr_env env sigma0 (snd sr)));
+  let sigma, rdxt = Typing.type_of env sigma rdx in
+  let () = debug_ssr (fun () -> Pp.(str"r@rwcltac=" ++ pr_econstr_env env sigma r)) in
   let cvtac, rwtac, sigma0 =
     if EConstr.Vars.closed0 sigma0 r' then
-      let sigma, c, c_eq = fst sr, snd sr, Coqlib.(lib_ref "core.eq.type") in
-      let sigma, c_ty = Typing.type_of env sigma c in
-        ppdebug(lazy Pp.(str"c_ty@rwcltac=" ++ pr_econstr_env env sigma c_ty));
+      let c_eq = Coqlib.(lib_ref "core.eq.type") in
+      let sigma, c_ty = Typing.type_of env sigma r in
+      let () = debug_ssr (fun () -> Pp.(str"c_ty@rwcltac=" ++ pr_econstr_env env sigma c_ty)) in
       let open EConstr in
       match kind_of_type sigma (Reductionops.whd_all env sigma c_ty) with
       | AtomicType(e, a) when Ssrcommon.is_ind_ref sigma e c_eq ->
           let new_rdx = if dir = L2R then a.(2) else a.(1) in
-          pirrel_rewrite ?under ?map_redex cl rdx rdxt new_rdx dir (sigma,c) c_ty, Tacticals.New.tclIDTAC, sigma0
+          pirrel_rewrite ?under ?map_redex cl rdx a.(0) new_rdx dir (sigma, r) c_ty, Tacticals.tclIDTAC, sigma0
       | _ ->
-          let cl' = EConstr.mkApp (EConstr.mkNamedLambda (make_annot pattern_id Sorts.Relevant) rdxt cl, [|rdx|]) in
+          let cl' = EConstr.mkApp (EConstr.mkNamedLambda sigma (make_annot pattern_id Sorts.Relevant) rdxt cl, [|rdx|]) in
           let sigma, _ = Typing.type_of env sigma cl' in
           let sigma0 = pf_merge_uc_of sigma sigma0 in
           convert_concl ~check:true cl', rewritetac ?under dir r', sigma0
@@ -455,32 +488,40 @@ let rwcltac ?under ?map_redex cl rdx dir sr =
       let dc, r2 = EConstr.decompose_lam_n_assum sigma0 n r' in
       let r3, _, r3t  =
         try EConstr.destCast sigma0 r2 with _ ->
-        errorstrm Pp.(str "no cast from " ++ pr_econstr_pat env sigma0 (snd sr)
+        errorstrm Pp.(str "no cast from " ++ pr_econstr_pat env sigma0 r
                     ++ str " to " ++ pr_econstr_env env sigma0 r2) in
-      let cl' = EConstr.mkNamedProd (make_annot rule_id Sorts.Relevant) (EConstr.it_mkProd_or_LetIn r3t dc) (EConstr.Vars.lift 1 cl) in
-      let cl'' = EConstr.mkNamedProd (make_annot pattern_id Sorts.Relevant) rdxt cl' in
+      let cl' = EConstr.mkNamedProd sigma (make_annot rule_id Sorts.Relevant) (EConstr.it_mkProd_or_LetIn r3t dc) (EConstr.Vars.lift 1 cl) in
+      let cl'' = EConstr.mkNamedProd sigma (make_annot pattern_id Sorts.Relevant) rdxt cl' in
       let itacs = [introid pattern_id; introid rule_id] in
       let cltac = Tactics.clear [pattern_id; rule_id] in
-      let rwtacs = [rewritetac ?under dir (EConstr.mkVar rule_id); cltac] in
-      Tactics.apply_type ~typecheck:true cl'' [rdx; EConstr.it_mkLambda_or_LetIn r3 dc], Tacticals.New.tclTHENLIST (itacs @ rwtacs), sigma0
+      let rwtacs = [
+        Tacticals.tclTHENLIST [
+          rewritetac ?under dir (EConstr.mkVar rule_id);
+          if !ssroldreworder || Option.default false under then
+            Proofview.tclUNIT ()
+          else
+            Proofview.cycle 1
+          ];
+        cltac] in
+      Tactics.apply_type ~typecheck:true cl'' [rdx; EConstr.it_mkLambda_or_LetIn r3 dc], Tacticals.tclTHENLIST (itacs @ rwtacs), sigma0
   in
   let cvtac' =
-    Proofview.tclOR cvtac begin function
+    Proofview.tclORELSE cvtac begin function
     | (PRtype_error e, _) ->
       let error = Option.cata (fun (env, sigma, te) ->
           Pp.(fnl () ++ str "Type error was: " ++ Himsg.explain_pretype_error env sigma te))
           (Pp.mt ()) e in
-      if occur_existential sigma0 (Tacmach.New.pf_concl gl)
-      then Tacticals.New.tclZEROMSG Pp.(str "Rewriting impacts evars" ++ error)
-      else Tacticals.New.tclZEROMSG Pp.(str "Dependent type error in rewrite of "
+      if occur_existential sigma0 (Tacmach.pf_concl gl)
+      then Tacticals.tclZEROMSG Pp.(str "Rewriting impacts evars" ++ error)
+      else Tacticals.tclZEROMSG Pp.(str "Dependent type error in rewrite of "
         ++ pr_econstr_env env sigma0
-          (EConstr.mkNamedLambda (make_annot pattern_id Sorts.Relevant) rdxt cl)
+          (EConstr.mkNamedLambda sigma (make_annot pattern_id Sorts.Relevant) rdxt cl)
         ++ error)
     | (e, info) -> Proofview.tclZERO ~info e
     end
   in
   Proofview.Unsafe.tclEVARS sigma0 <*>
-  Tacticals.New.tclTHEN cvtac' rwtac
+  Tacticals.tclTHEN cvtac' rwtac
   end
 
 [@@@ocaml.warning "-3"]
@@ -494,7 +535,7 @@ let lz_setoid_relation =
   | Some (env', srel) when env' == env -> srel
   | _ ->
     let srel =
-       try Some (UnivGen.constr_of_monomorphic_global @@
+       try Some (UnivGen.constr_of_monomorphic_global (Global.env ()) @@
                  Coqlib.find_reference "Class_setoid" ("Coq"::sdir) "RewriteRelation" [@ocaml.warning "-3"])
        with _ -> None in
     last_srel := Some (env, srel); srel
@@ -508,8 +549,8 @@ let ssr_is_setoid env =
       sigma [] (EConstr.mkApp (r, args)) <> None
 
 let closed0_check env sigma cl p =
-  if closed0 cl then
-    errorstrm Pp.(str"No occurrence of redex "++ pr_constr_env env sigma p)
+  if EConstr.Vars.closed0 sigma cl then
+    errorstrm Pp.(str"No occurrence of redex "++ pr_econstr_env env sigma p)
 
 let dir_org = function L2R -> 1 | R2L -> 2
 
@@ -520,33 +561,31 @@ let rwprocess_rule env dir rule =
     let rec loop d sigma r t0 rs red =
       let t =
         if red = 1 then Tacred.hnf_constr env sigma t0
-        else Reductionops.whd_betaiotazeta env sigma t0 in
-      ppdebug(lazy Pp.(str"rewrule="++pr_econstr_pat env sigma t));
+        else Reductionops.clos_whd_flags CClosure.betaiotazeta env sigma t0 in
+      debug_ssr (fun () -> Pp.(str"rewrule="++pr_econstr_pat env sigma t));
       match EConstr.kind sigma t with
       | Prod (_, xt, at) ->
         let sigma = Evd.create_evar_defs sigma in
         let (sigma, x) = Evarutil.new_evar env sigma xt in
         loop d sigma EConstr.(mkApp (r, [|x|])) (EConstr.Vars.subst1 x at) rs 0
       | App (pr, a) when is_ind_ref sigma pr coq_prod.Coqlib.typ ->
-        let sr sigma = match EConstr.kind sigma (Tacred.hnf_constr env sigma r) with
+        let r0 = Reductionops.clos_whd_flags CClosure.all env sigma r in
+        let sigma, pL, pR = match EConstr.kind sigma r0 with
         | App (c, ra) when is_construct_ref sigma c coq_prod.Coqlib.intro ->
-          fun i -> ra.(i + 1), sigma
-        | _ -> let ra = Array.append a [|r|] in
-          function 1 ->
-            let sigma, pi1 = Evd.fresh_global env sigma coq_prod.Coqlib.proj1 in
-            EConstr.mkApp (pi1, ra), sigma
-          | _ ->
-            let sigma, pi2 = Evd.fresh_global env sigma coq_prod.Coqlib.proj2 in
-            EConstr.mkApp (pi2, ra), sigma in
-        let sigma,trty = Evd.fresh_global env sigma Coqlib.(lib_ref "core.True.type") in
-        if EConstr.eq_constr sigma a.(0) trty then
-         let s, sigma = sr sigma 2 in
-         loop (converse_dir d) sigma s a.(1) rs 0
+          (sigma, ra.(2), ra.(3))
+        | _ ->
+          let ra = Array.append a [|r|] in
+          let sigma, pi1 = Evd.fresh_global env sigma coq_prod.Coqlib.proj1 in
+          let sigma, pi2 = Evd.fresh_global env sigma coq_prod.Coqlib.proj2 in
+          let pL = EConstr.mkApp (pi1, ra) in
+          let pR = EConstr.mkApp (pi2, ra) in
+          (sigma, pL, pR)
+        in
+        if EConstr.isRefX sigma Coqlib.(lib_ref "core.True.type") a.(0) then
+         loop (converse_dir d) sigma pR a.(1) rs 0
         else
-         let s, sigma = sr sigma 2 in
-         let sigma, rs2 = loop d sigma s a.(1) rs 0 in
-         let s, sigma = sr sigma 1 in
-         loop d sigma s a.(0) rs2 0
+         let sigma, rs2 = loop d sigma pR a.(1) rs 0 in
+         loop d sigma pL a.(0) rs2 0
       | App (r_eq, a) when Hipattern.match_with_equality_type env sigma t != None ->
         let (ind, u) = EConstr.destInd sigma r_eq and rhs = Array.last a in
         let np = Inductiveops.inductive_nparamdecls env ind in
@@ -610,28 +649,29 @@ let rwrxtac ?under ?map_redex occ rdx_pat dir rule =
      rwtac rules in
   let env0 = env in
   let concl0 = Proofview.Goal.concl gl in
-  let find_R, conclude = match rdx_pat with
-  | Some (_, (In_T _ | In_X_In_T _)) | None ->
-      let upats_origin = dir, EConstr.Unsafe.to_constr (snd rule) in
-      let rpat env sigma0 (sigma, pats) (d, r, lhs, rhs) =
-        let sigma, pat =
-          let rw_progress rhs t evd = rw_progress rhs (EConstr.of_constr t) evd in
-          mk_tpattern env sigma0 (sigma, EConstr.to_constr ~abort_on_undefined_evars:false sigma r) (rw_progress rhs) d (EConstr.to_constr ~abort_on_undefined_evars:false sigma lhs) in
-        sigma, pats @ [pat] in
-      let rpats = List.fold_left (rpat env0 sigma0) (r_sigma,[]) rules in
+  let rigid ev = Evd.mem sigma0 ev in
+  let find_R, conclude = match classify_pattern rdx_pat with
+  | None ->
+      let upats_origin = dir, (snd rule) in
+      let rpat pats (d, r, lhs, rhs) =
+        let r = Reductionops.nf_evar r_sigma r in
+        let lhs = Reductionops.nf_evar r_sigma lhs in
+        mk_tpattern ~ok:(rw_progress rhs) ~rigid env0 r d lhs pats
+      in
+      let rpats = List.fold_left rpat (empty_tpatterns r_sigma) rules in
       let find_R, end_R = mk_tpattern_matcher sigma0 occ ~upats_origin rpats in
-      (fun e c _ i -> find_R ~k:(fun _ _ _ h -> mkRel h) e c i),
+      (fun e c _ i -> find_R ~k:(fun _ _ _ h -> EConstr.mkRel h) e c i),
       fun cl -> let rdx,d,r = end_R () in closed0_check env0 sigma0 cl rdx; (d,r),rdx
-  | Some(_, (T e | X_In_T (_,e) | E_As_X_In_T (e,_,_) | E_In_X_In_T (e,_,_))) ->
+  | Some (_, e) ->
       let r = ref None in
-      (fun env c _ h -> do_once r (fun () -> find_rule (EConstr.of_constr c), c); mkRel h),
+      (fun env c _ h -> do_once r (fun () -> find_rule c, c); EConstr.mkRel h),
       (fun concl -> closed0_check env0 sigma0 concl e;
-        let (d,(ev,ctx,c)) , x = assert_done r in (d,(ev,ctx, EConstr.to_constr ~abort_on_undefined_evars:false ev c)) , x) in
-  let concl0 = EConstr.to_constr ~abort_on_undefined_evars:false sigma0 concl0 in
+        let (d,(ev,ctx,c)) , x = assert_done r in (d,(true, ev,ctx, Reductionops.nf_evar ev c)) , x) in
+  let concl0 = Reductionops.nf_evar sigma0 concl0 in
   let concl = eval_pattern env0 sigma0 concl0 rdx_pat occ find_R in
-  let (d, r), rdx = conclude concl in
-  let r = Evd.merge_universe_context (pi1 r) (pi2 r), EConstr.of_constr (pi3 r) in
-  rwcltac ?under ?map_redex (EConstr.of_constr concl) (EConstr.of_constr rdx) d r
+  let (d, (_, sigma, uc, t)), rdx = conclude concl in
+  let r = Evd.merge_universe_context sigma uc, t in
+  rwcltac ?under ?map_redex concl rdx d r
   end
 
 let ssrinstancesofrule ist dir arg =
@@ -639,27 +679,25 @@ let ssrinstancesofrule ist dir arg =
   let env0 = Proofview.Goal.env gl in
   let sigma0 = Proofview.Goal.sigma gl in
   let concl0 = Proofview.Goal.concl gl in
+  let rigid ev = Evd.mem sigma0 ev in
   let rule = interp_term env0 sigma0 ist arg in
   let r_sigma, rules = rwprocess_rule env0 dir rule in
   let find, conclude =
-    let upats_origin = dir, EConstr.Unsafe.to_constr (snd rule) in
-    let rpat env sigma0 (sigma, pats) (d, r, lhs, rhs) =
-      let sigma, pat =
-        let rw_progress rhs t evd = rw_progress rhs (EConstr.of_constr t) evd in
-        mk_tpattern env sigma0
-          (sigma,EConstr.to_constr ~abort_on_undefined_evars:false sigma r)
-          (rw_progress rhs) d
-          (EConstr.to_constr ~abort_on_undefined_evars:false sigma lhs) in
-      sigma, pats @ [pat] in
-    let rpats = List.fold_left (rpat env0 sigma0) (r_sigma,[]) rules in
+    let upats_origin = dir, (snd rule) in
+    let rpat pats (d, r, lhs, rhs) =
+      let r = Reductionops.nf_evar r_sigma r in
+      let lhs = Reductionops.nf_evar r_sigma lhs in
+      mk_tpattern ~ok:(rw_progress rhs) ~rigid env0 r d lhs pats
+    in
+    let rpats = List.fold_left rpat (empty_tpatterns r_sigma) rules in
     mk_tpattern_matcher ~all_instances:true ~raise_NoMatch:true sigma0 None ~upats_origin rpats in
-  let print env p c _ = Feedback.msg_info Pp.(hov 1 (str"instance:" ++ spc() ++ pr_constr_env env r_sigma p ++ spc() ++ str "matches:" ++ spc() ++ pr_constr_env env r_sigma c)); c in
+  let print env p c _ = Feedback.msg_info Pp.(hov 1 (str"instance:" ++ spc() ++ pr_econstr_env env r_sigma p ++ spc() ++ str "matches:" ++ spc() ++ pr_econstr_env env r_sigma c)); c in
   Feedback.msg_info Pp.(str"BEGIN INSTANCES");
   try
     while true do
-      ignore(find env0 (EConstr.to_constr ~abort_on_undefined_evars:false sigma0 concl0) 1 ~k:print)
+      ignore(find env0 (Reductionops.nf_evar sigma0 concl0) 1 ~k:print)
     done; raise NoMatch
-  with NoMatch -> Feedback.msg_info Pp.(str"END INSTANCES"); Tacticals.New.tclIDTAC
+  with NoMatch -> Feedback.msg_info Pp.(str"END INSTANCES"); Tacticals.tclIDTAC
   end
 
 let ipat_rewrite occ dir c = Proofview.Goal.enter begin fun gl ->
@@ -674,7 +712,7 @@ let rwargtac ?under ?map_redex ist ((dir, mult), (((oclr, occ), grx), (kind, gt)
   let fail = ref false in
   let interp_rpattern env sigma gc =
     try interp_rpattern env sigma gc
-    with _ when snd mult = May -> fail := true; sigma, T mkProp in
+    with _ when snd mult = May -> fail := true; { pat_sigma = sigma; pat_pat = T EConstr.mkProp } in
   let interp env sigma gc =
     try interp_term env sigma ist gc
     with _ when snd mult = May -> fail := true; (sigma, EConstr.mkProp) in
@@ -683,11 +721,13 @@ let rwargtac ?under ?map_redex ist ((dir, mult), (((oclr, occ), grx), (kind, gt)
     let env = Proofview.Goal.env gl in
     let sigma = Proofview.Goal.sigma gl in
     let rx = Option.map (interp_rpattern env sigma) grx in
+    (* Evarmaps below are extensions of sigma, so setting the universe context is correct *)
     let sigma = match rx with
-      | None -> sigma
-      | Some (s,_) -> pf_merge_uc_of s sigma in
+    | None -> sigma
+    | Some { pat_sigma = s } -> Evd.set_universe_context sigma (Evd.evar_universe_context s)
+    in
     let t = interp env sigma gt in
-    let sigma = pf_merge_uc_of (fst t) sigma in
+    let sigma = Evd.set_universe_context sigma  (Evd.evar_universe_context (fst t)) in
     Proofview.Unsafe.tclEVARS sigma <*>
     (match kind with
     | RWred sim -> simplintac occ rx sim
@@ -696,7 +736,7 @@ let rwargtac ?under ?map_redex ist ((dir, mult), (((oclr, occ), grx), (kind, gt)
     end
   in
   let ctac = cleartac (interp_clr sigma (oclr, (fst gt, snd (interp env sigma gt)))) in
-  if !fail then ctac else Tacticals.New.tclTHEN (tclMULT mult rwtac) ctac
+  if !fail then ctac else Tacticals.tclTHEN (tclMULT mult rwtac) ctac
   end
 
 (** Rewrite argument sequence *)
@@ -707,7 +747,7 @@ let rwargtac ?under ?map_redex ist ((dir, mult), (((oclr, occ), grx), (kind, gt)
 
 let ssrrewritetac ?under ?map_redex ist rwargs =
   Proofview.Goal.enter begin fun _ ->
-  Tacticals.New.tclTHENLIST (List.map (rwargtac ?under ?map_redex ist) rwargs)
+  Tacticals.tclTHENLIST (List.map (rwargtac ?under ?map_redex ist) rwargs)
   end
 
 (** The "unlock" tactic *)
@@ -736,7 +776,7 @@ let unlocktac ist args =
   Ssrcommon.tacMK_SSR_CONST "locked" >>= fun locked ->
   Ssrcommon.tacMK_SSR_CONST "master_key" >>= fun key ->
   let ktacs = [
-    (Proofview.tclEVARMAP >>= fun sigma -> unfoldtac None None (sigma, locked) xInParens);
+    (Proofview.tclEVARMAP >>= fun sigma -> unfoldtac None None (sigma, locked) InParens);
     Ssrelim.casetac key (fun ?seed:_ k -> k)
   ] in
-  Tacticals.New.tclTHENLIST (List.map utac args @ ktacs)
+  Tacticals.tclTHENLIST (List.map utac args @ ktacs)
