@@ -52,7 +52,8 @@ type cbv_value =
   | FIXP of fixpoint * cbv_value subs * cbv_value array
   | COFIXP of cofixpoint * cbv_value subs * cbv_value array
   | CONSTR of constructor Univ.puniverses * cbv_value array
-  | PRIMITIVE of CPrimitives.t * constr * cbv_value array
+  | PRIMITIVE of CPrimitives.t * pconstant * cbv_value array
+  | ARRAY of Univ.Instance.t * cbv_value Parray.t * cbv_value
 
 (* type of terms with a hole. This hole can appear only under App or Case.
  *   TOP means the term is considered without context
@@ -74,8 +75,8 @@ type cbv_value =
  *)
 and cbv_stack =
   | TOP
-  | APP of cbv_value array * cbv_stack
-  | CASE of constr * constr array * case_info * cbv_value subs * cbv_stack
+  | APP of cbv_value list * cbv_stack
+  | CASE of Univ.Instance.t * constr array * case_return * case_branch array * Constr.case_invert * case_info * cbv_value subs * cbv_stack
   | PROJ of Projection.t * cbv_stack
 
 (* les vars pourraient etre des constr,
@@ -97,6 +98,8 @@ let rec shift_value n = function
       CONSTR (c, Array.map (shift_value n) args)
   | PRIMITIVE(op,c,args) ->
       PRIMITIVE(op,c,Array.map (shift_value n) args)
+  | ARRAY (u,t,ty) ->
+      ARRAY(u, Parray.map (shift_value n) t, shift_value n ty)
 
 let shift_value n v =
   if Int.equal n 0 then v else shift_value n v
@@ -107,15 +110,20 @@ let shift_value n v =
  * (S, (fix Fi {F0 := T0 .. Fn-1 := Tn-1}))
  *    -> (S. [S]F0 . [S]F1 ... . [S]Fn-1, Ti)
  *)
+
+let rec mk_fix_subs make_body n env i =
+  if Int.equal i n then env
+  else mk_fix_subs make_body n (subs_cons (make_body i) env) (i + 1)
+
 let contract_fixp env ((reci,i),(_,_,bds as bodies)) =
   let make_body j = FIXP(((reci,j),bodies), env, [||]) in
   let n = Array.length bds in
-  subs_cons(Array.init n make_body, env), bds.(i)
+  mk_fix_subs make_body n env 0, bds.(i)
 
 let contract_cofixp env (i,(_,_,bds as bodies)) =
   let make_body j = COFIXP((j,bodies), env, [||]) in
   let n = Array.length bds in
-  subs_cons(Array.init n make_body, env), bds.(i)
+  mk_fix_subs make_body n env 0, bds.(i)
 
 let make_constr_ref n k t =
   match k with
@@ -124,17 +132,23 @@ let make_constr_ref n k t =
   | ConstKey cst -> t
 
 (* Adds an application list. Collapse APPs! *)
-let stack_app appl stack =
+let stack_vect_app appl stack =
   if Int.equal (Array.length appl) 0 then stack else
     match stack with
-    | APP(args,stk) -> APP(Array.append appl args,stk)
+    | APP(args,stk) -> APP(Array.fold_right (fun v accu -> v :: accu) appl args,stk)
+    | _             -> APP(Array.to_list appl, stack)
+
+let stack_app appl stack =
+  if List.is_empty appl then stack else
+    match stack with
+    | APP(args,stk) -> APP(appl @ args,stk)
     | _             -> APP(appl, stack)
 
 let rec stack_concat stk1 stk2 =
   match stk1 with
       TOP -> stk2
     | APP(v,stk1') -> APP(v,stack_concat stk1' stk2)
-    | CASE(c,b,i,s,stk1') -> CASE(c,b,i,s,stack_concat stk1' stk2)
+    | CASE(u,pms,c,b,iv,i,s,stk1') -> CASE(u,pms,c,b,iv,i,s,stack_concat stk1' stk2)
     | PROJ (p,stk1') -> PROJ (p,stack_concat stk1' stk2)
 
 (* merge stacks when there is no shifts in between *)
@@ -165,11 +179,11 @@ let red_set_ref flags = function
  *)
 let strip_appl head stack =
   match head with
-    | FIXP (fix,env,app) -> (FIXP(fix,env,[||]), stack_app app stack)
-    | COFIXP (cofix,env,app) -> (COFIXP(cofix,env,[||]), stack_app app stack)
-    | CONSTR (c,app) -> (CONSTR(c,[||]), stack_app app stack)
-    | PRIMITIVE(op,c,app) -> (PRIMITIVE(op,c,[||]), stack_app app stack)
-    | VAL _ | STACK _ | CBN _ | LAM _ -> (head, stack)
+    | FIXP (fix,env,app) -> (FIXP(fix,env,[||]), stack_vect_app app stack)
+    | COFIXP (cofix,env,app) -> (COFIXP(cofix,env,[||]), stack_vect_app app stack)
+    | CONSTR (c,app) -> (CONSTR(c,[||]), stack_vect_app app stack)
+    | PRIMITIVE(op,c,app) -> (PRIMITIVE(op,c,[||]), stack_vect_app app stack)
+    | VAL _ | STACK _ | CBN _ | LAM _ | ARRAY _ -> (head, stack)
 
 
 (* Tests if fixpoint reduction is possible. *)
@@ -177,10 +191,15 @@ let fixp_reducible flgs ((reci,i),_) stk =
   if red_set flgs fFIX then
     match stk with
       | APP(appl,_) ->
-          Array.length appl > reci.(i) &&
-          (match appl.(reci.(i)) with
-              CONSTR _ -> true
-            | _ -> false)
+        let rec check n = function
+        | [] -> false
+        | v :: appl ->
+          if Int.equal n 0 then match v with
+          | CONSTR _ -> true
+          | _ -> false
+          else check (n - 1) appl
+        in
+        check reci.(i) appl
       | _ -> false
   else
     false
@@ -193,10 +212,7 @@ let cofixp_reducible flgs _ stk =
   else
     false
 
-let get_debug_cbv = Goptions.declare_bool_option_and_ref
-    ~depr:false
-    ~value:false
-    ~key:["Debug";"Cbv"]
+let debug_cbv = CDebug.create ~name:"Cbv" ()
 
 (* Reduction of primitives *)
 
@@ -208,6 +224,7 @@ module VNativeEntries =
     type elem = cbv_value
     type args = cbv_value array
     type evd = unit
+    type uinstance = Univ.Instance.t
 
     let get = Array.get
 
@@ -227,6 +244,11 @@ module VNativeEntries =
         | _ -> raise Primred.NativeDestKO)
       | _ -> raise Primred.NativeDestKO
 
+    let get_parray () e =
+      match e with
+      | ARRAY(_u,t,_ty) -> t
+      | _ -> raise Primred.NativeDestKO
+
     let mkInt env i = VAL(0, mkInt i)
 
     let mkFloat env f = VAL(0, mkFloat f)
@@ -326,6 +348,9 @@ module VNativeEntries =
       let (_pNormal,_nNormal,_pSubn,_nSubn,_pZero,_nZero,_pInf,_nInf,nan) =
         get_f_class_constructors env in
       CONSTR(Univ.in_punivs nan, [||])
+
+    let mkArray env u t ty =
+      ARRAY (u,t,ty)
   end
 
 module VredNative = RedNative(VNativeEntries)
@@ -338,10 +363,10 @@ let debug_pr_key = function
 let rec reify_stack t = function
   | TOP -> t
   | APP (args,st) ->
-      reify_stack (mkApp(t,Array.map reify_value args)) st
-  | CASE (ty,br,ci,env,st) ->
+      reify_stack (mkApp(t,Array.map_of_list reify_value args)) st
+  | CASE (u,pms,ty,br,iv,ci,env,st) ->
       reify_stack
-        (mkCase (ci, ty, t,br))
+        (mkCase (ci, u, pms, ty, iv, t,br))
         st
   | PROJ (p, st) ->
        reify_stack (mkProj (p, t)) st
@@ -367,7 +392,10 @@ and reify_value = function (* reduction under binders *)
   | CONSTR (c,args) ->
       mkApp(mkConstructU c, Array.map reify_value args)
   | PRIMITIVE(op,c,args) ->
-      mkApp(c, Array.map reify_value args)
+      mkApp(mkConstU c, Array.map reify_value args)
+  | ARRAY (u,t,ty) ->
+    let t, def = Parray.to_array t in
+      mkArray(u, Array.map reify_value t, reify_value def, reify_value ty)
 
 and apply_env env t =
   match kind t with
@@ -381,6 +409,32 @@ and apply_env env t =
   | _ ->
     map_with_binders subs_lift apply_env env t
 
+let rec strip_app = function
+  | APP (args,st) -> APP (args,strip_app st)
+  | s -> TOP
+
+(* TODO: share the common parts with EConstr *)
+let expand_branch env u pms (ind, i) br =
+  let open Declarations in
+  let nas, _br = br.(i - 1) in
+  let (mib, mip) = Inductive.lookup_mind_specif env ind in
+  let paramdecl = Vars.subst_instance_context u mib.mind_params_ctxt in
+  let paramsubst = Vars.subst_of_rel_context_instance paramdecl pms in
+  let (ctx, _) = mip.mind_nf_lc.(i - 1) in
+  let (ctx, _) = List.chop mip.mind_consnrealdecls.(i - 1) ctx in
+  Inductive.instantiate_context u paramsubst nas ctx
+
+let cbv_subst_of_rel_context_instance_list mkclos sign args env =
+  let rec aux subst sign l =
+    let open Context.Rel.Declaration in
+    match sign, l with
+    | LocalAssum _ :: sign', a::args' -> aux (subs_cons a subst) sign' args'
+    | LocalDef (_,c,_)::sign', args' ->
+        aux (subs_cons (mkclos subst c) subst) sign' args'
+    | [], [] -> subst
+    | _ -> CErrors.anomaly (Pp.str "Instance and signature do not match.")
+  in aux env (List.rev sign) args
+
 (* The main recursive functions
  *
  * Go under applications and cases/projections (pushed in the stack),
@@ -398,9 +452,14 @@ let rec norm_head info env t stack =
   (* stack grows (remove casts) *)
   | App (head,args) -> (* Applied terms are normalized immediately;
                         they could be computed when getting out of the stack *)
-      let nargs = Array.map (cbv_stack_term info TOP env) args in
-      norm_head info env head (stack_app nargs stack)
-  | Case (ci,p,c,v) -> norm_head info env c (CASE(p,v,ci,env,stack))
+    let fold c accu = cbv_stack_term info TOP env c :: accu in
+    let rem, stack = match stack with
+    | APP (nargs, stack) -> nargs, stack
+    | _ -> [], stack
+    in
+    let stack = APP (Array.fold_right fold args rem, stack) in
+    norm_head info env head stack
+  | Case (ci,u,pms,p,iv,c,v) -> norm_head info env c (CASE(u,pms,p,v,iv,ci,env,stack))
   | Cast (ct,_,_) -> norm_head info env ct stack
 
   | Proj (p, c) ->
@@ -426,7 +485,7 @@ let rec norm_head info env t stack =
 
   | Const sp ->
     Reductionops.reduction_effect_hook info.env info.sigma
-      (fst sp) (lazy (reify_stack t stack));
+      (fst sp) (lazy (reify_stack t (strip_app stack)));
     norm_head_ref 0 info env stack (ConstKey sp) t
 
   | LetIn (_, b, _, c) ->
@@ -436,18 +495,19 @@ let rec norm_head info env t stack =
         (* New rule: for Cbv, Delta does not apply to locally bound variables
            or red_set info.reds fDELTA
          *)
-        let env' = subs_cons ([|cbv_stack_term info TOP env b|],env) in
+        let env' = subs_cons (cbv_stack_term info TOP env b) env in
         norm_head info env' c stack
       else
         (CBN(t,env), stack) (* Should we consider a commutative cut ? *)
 
-  | Evar ev ->
-      (match Reductionops.safe_evar_value info.sigma ev with
+  | Evar ((e, _) as ev) ->
+      (match Evd.existential_opt_value0 info.sigma ev with
           Some c -> norm_head info env c stack
         | None ->
-          let e, xs = ev in
-          let xs' = List.map (apply_env env) xs in
-          (VAL(0, mkEvar (e,xs')), stack))
+          let ev = EConstr.of_existential ev in
+          let map c = EConstr.of_constr @@ apply_env env (EConstr.Unsafe.to_constr c) in
+          let ev' = EConstr.map_existential info.sigma map ev in
+          (VAL(0, EConstr.Unsafe.to_constr @@ EConstr.mkEvar ev'), stack))
 
   (* non-neutral cases *)
   | Lambda _ ->
@@ -457,6 +517,15 @@ let rec norm_head info env t stack =
   | CoFix cofix -> (COFIXP(cofix,env,[||]), stack)
   | Construct c -> (CONSTR(c, [||]), stack)
 
+  | Array(u,t,def,ty) ->
+    let ty = cbv_stack_term info TOP env ty in
+    let len = Array.length t in
+    let t =
+      Parray.init (Uint63.of_int len)
+      (fun i -> cbv_stack_term info TOP env t.(i))
+      (cbv_stack_term info TOP env def) in
+    (ARRAY (u,t,ty), stack)
+
   (* neutral cases *)
   | (Sort _ | Meta _ | Ind _ | Int _ | Float _) -> (VAL(0, t), stack)
   | Prod _ -> (CBN(t,env), stack)
@@ -465,15 +534,20 @@ and norm_head_ref k info env stack normt t =
   if red_set_ref info.reds normt then
     match cbv_value_cache info normt with
       | Declarations.Def body ->
-         if get_debug_cbv () then Feedback.msg_debug Pp.(str "Unfolding " ++ debug_pr_key normt);
+         debug_cbv (fun () -> Pp.(str "Unfolding " ++ debug_pr_key normt));
          strip_appl (shift_value k body) stack
-      | Declarations.Primitive op -> (PRIMITIVE(op,t,[||]),stack)
+      | Declarations.Primitive op ->
+        let c = match normt with
+          | ConstKey c -> c
+          | RelKey _ | VarKey _ -> assert false
+        in
+        (PRIMITIVE(op,c,[||]),stack)
       | Declarations.OpaqueDef _ | Declarations.Undef _ ->
-         if get_debug_cbv () then Feedback.msg_debug Pp.(str "Not unfolding " ++ debug_pr_key normt);
+         debug_cbv (fun () -> Pp.(str "Not unfolding " ++ debug_pr_key normt));
          (VAL(0,make_constr_ref k normt t),stack)
   else
     begin
-      if get_debug_cbv () then Feedback.msg_debug Pp.(str "Not unfolding " ++ debug_pr_key normt);
+      debug_cbv (fun () -> Pp.(str "Not unfolding " ++ debug_pr_key normt));
       (VAL(0,make_constr_ref k normt t),stack)
     end
 
@@ -490,17 +564,19 @@ and cbv_stack_value info env = function
   (* a lambda meets an application -> BETA *)
   | (LAM (nlams,ctxt,b,env), APP (args, stk))
       when red_set info.reds fBETA ->
-    let nargs = Array.length args in
-      if nargs == nlams then
-          cbv_stack_term info stk (subs_cons(args,env)) b
-        else if nlams < nargs then
-          let env' = subs_cons(Array.sub args 0 nlams, env) in
-          let eargs = Array.sub args nlams (nargs-nlams) in
-          cbv_stack_term info (APP(eargs,stk)) env' b
-        else
-          let ctxt' = List.skipn nargs ctxt in
-          LAM(nlams-nargs,ctxt', b, subs_cons(args,env))
-
+    let rec apply env lams args =
+      if Int.equal lams 0 then
+        let stk = if List.is_empty args then stk else APP (args, stk) in
+        cbv_stack_term info stk env b
+      else match args with
+      | [] ->
+        let ctxt' = List.skipn (nlams - lams) ctxt in
+        LAM (lams, ctxt', b, env)
+      | v :: args ->
+        let env = subs_cons v env in
+        apply env (lams - 1) args
+    in
+    apply env nlams args
     (* a Fix applied enough -> IOTA *)
     | (FIXP(fix,env,[||]), stk)
         when fixp_reducible info.reds fix stk ->
@@ -514,49 +590,61 @@ and cbv_stack_value info env = function
         cbv_stack_term info stk envf redfix
 
     (* constructor in a Case -> IOTA *)
-    | (CONSTR(((sp,n),u),[||]), APP(args,CASE(_,br,ci,env,stk)))
+    | (CONSTR(((sp,n),_),[||]), APP(args,CASE(u,pms,_p,br,iv,ci,env,stk)))
             when red_set info.reds fMATCH ->
-        let cargs =
-          Array.sub args ci.ci_npar (Array.length args - ci.ci_npar) in
-        cbv_stack_term info (stack_app cargs stk) env br.(n-1)
+        let cargs = List.skipn ci.ci_npar args in
+        let env =
+          if (Int.equal ci.ci_cstr_ndecls.(n - 1) ci.ci_cstr_nargs.(n - 1)) then (* no lets *)
+            List.fold_left (fun accu v -> subs_cons v accu) env cargs
+          else
+            let mkclos env c = cbv_stack_term info TOP env c in
+            let ctx = expand_branch info.env u pms (sp, n) br in
+            cbv_subst_of_rel_context_instance_list mkclos ctx cargs env
+        in
+        cbv_stack_term info stk env (snd br.(n-1))
 
     (* constructor of arity 0 in a Case -> IOTA *)
-    | (CONSTR(((_,n),u),[||]), CASE(_,br,_,env,stk))
+    | (CONSTR(((sp, n), _),[||]), CASE(u,pms,_,br,_,ci,env,stk))
             when red_set info.reds fMATCH ->
-                    cbv_stack_term info stk env br.(n-1)
+        let env =
+          if (Int.equal ci.ci_cstr_ndecls.(n - 1) ci.ci_cstr_nargs.(n - 1)) then (* no lets *)
+            env
+          else
+            let mkclos env c = cbv_stack_term info TOP env c in
+            let ctx = expand_branch info.env u pms (sp, n) br in
+            cbv_subst_of_rel_context_instance_list mkclos ctx [] env
+        in
+        cbv_stack_term info stk env (snd br.(n-1))
 
     (* constructor in a Projection -> IOTA *)
     | (CONSTR(((sp,n),u),[||]), APP(args,PROJ(p,stk)))
         when red_set info.reds fMATCH && Projection.unfolded p ->
-      let arg = args.(Projection.npars p + Projection.arg p) in
+      let arg = List.nth args (Projection.npars p + Projection.arg p) in
         cbv_stack_value info env (strip_appl arg stk)
 
     (* may be reduced later by application *)
-    | (FIXP(fix,env,[||]), APP(appl,TOP)) -> FIXP(fix,env,appl)
-    | (COFIXP(cofix,env,[||]), APP(appl,TOP)) -> COFIXP(cofix,env,appl)
-    | (CONSTR(c,[||]), APP(appl,TOP)) -> CONSTR(c,appl)
+    | (FIXP(fix,env,[||]), APP(appl,TOP)) -> FIXP(fix,env,Array.of_list appl)
+    | (COFIXP(cofix,env,[||]), APP(appl,TOP)) -> COFIXP(cofix,env,Array.of_list appl)
+    | (CONSTR(c,[||]), APP(appl,TOP)) -> CONSTR(c,Array.of_list appl)
 
     (* primitive apply to arguments *)
-    | (PRIMITIVE(op,c,[||]), APP(appl,stk)) ->
+    | (PRIMITIVE(op,(_,u as c),[||]), APP(appl,stk)) ->
       let nargs = CPrimitives.arity op in
-      let len = Array.length appl in
-      if nargs <= len then
-        let args =
-          if len = nargs then appl
-          else Array.sub appl 0 nargs in
-        let stk =
-          if nargs < len then
-            stack_app (Array.sub appl nargs (len - nargs)) stk
-          else stk in
-        match VredNative.red_prim info.env () op args with
+      begin match List.chop nargs appl with
+      | (args, appl) ->
+        let stk = if List.is_empty appl then stk else stack_app appl stk in
+        begin match VredNative.red_prim info.env () op u (Array.of_list args) with
         | Some (CONSTR (c, args)) ->
           (* args must be moved to the stack to allow future reductions *)
-          cbv_stack_value info env (CONSTR(c, [||]), stack_app args stk)
+          cbv_stack_value info env (CONSTR(c, [||]), stack_vect_app args stk)
         | Some v ->  cbv_stack_value info env (v,stk)
-        | None -> mkSTACK(PRIMITIVE(op,c,args), stk)
-      else (* partial application *)
+        | None -> mkSTACK(PRIMITIVE(op,c,Array.of_list args), stk)
+        end
+      | exception Failure _ ->
+        (* partial application *)
               (assert (stk = TOP);
-               PRIMITIVE(op,c,appl))
+               PRIMITIVE(op,c,Array.of_list appl))
+      end
 
     (* definitely a value *)
     | (head,stk) -> mkSTACK(head, stk)
@@ -584,7 +672,7 @@ and cbv_value_cache info ref =
         let v = cbv_stack_term info TOP (subs_id 0) body in
         Declarations.Def v
       with
-      | Environ.NotEvaluableConst (Environ.IsPrimitive op) -> Declarations.Primitive op
+      | Environ.NotEvaluableConst (Environ.IsPrimitive (_u,op)) -> Declarations.Primitive op
       | Not_found | Environ.NotEvaluableConst _ -> Declarations.Undef None
     in
     KeyTable.add info.tab ref v; v
@@ -596,11 +684,32 @@ and cbv_value_cache info ref =
 let rec apply_stack info t = function
   | TOP -> t
   | APP (args,st) ->
-      apply_stack info (mkApp(t,Array.map (cbv_norm_value info) args)) st
-  | CASE (ty,br,ci,env,st) ->
+      apply_stack info (mkApp(t,Array.map_of_list (cbv_norm_value info) args)) st
+  | CASE (u,pms,ty,br,iv,ci,env,st) ->
+    (* FIXME: Prevent this expansion by caching whether an inductive contains let-bindings *)
+    let (_, ty, _, _, br) = Inductive.expand_case info.env (ci, u, pms, ty, iv, mkProp, br) in
+    let ty =
+      let (_, mip) = Inductive.lookup_mind_specif info.env ci.ci_ind in
+      Term.decompose_lam_n_decls (mip.Declarations.mind_nrealdecls + 1) ty
+    in
+    let mk_br c n = Term.decompose_lam_n_decls n c in
+    let br = Array.map2 mk_br br ci.ci_cstr_ndecls in
+    let map_ctx (nas, c) =
+      let open Context.Rel.Declaration in
+      let fold decl e = match decl with
+      | LocalAssum _ -> subs_lift e
+      | LocalDef (_, b, _) ->
+        let b = cbv_stack_term info TOP e b in
+        (* The let-binding persists, so we have to shift *)
+        subs_shft (1, subs_cons b e)
+      in
+      let env = List.fold_right fold nas env in
+      let nas = Array.of_list (List.rev_map get_annot nas) in
+      (nas, cbv_norm_term info env c)
+    in
       apply_stack info
-        (mkCase (ci, cbv_norm_term info env ty, t,
-                    Array.map (cbv_norm_term info env) br))
+        (mkCase (ci, u, Array.map (cbv_norm_term info env) pms, map_ctx ty, iv, t,
+                    Array.map map_ctx br))
         st
   | PROJ (p, st) ->
        apply_stack info (mkProj (p, t)) st
@@ -642,12 +751,17 @@ and cbv_norm_value info = function (* reduction under binders *)
   | CONSTR (c,args) ->
       mkApp(mkConstructU c, Array.map (cbv_norm_value info) args)
   | PRIMITIVE(op,c,args) ->
-      mkApp(c,Array.map (cbv_norm_value info) args)
+      mkApp(mkConstU c,Array.map (cbv_norm_value info) args)
+  | ARRAY (u,t,ty) ->
+    let ty = cbv_norm_value info ty in
+    let t, def = Parray.to_array t in
+    let def = cbv_norm_value info def in
+      mkArray(u, Array.map (cbv_norm_value info) t, def, ty)
 
 (* with profiling *)
 let cbv_norm infos constr =
   let constr = EConstr.Unsafe.to_constr constr in
-  EConstr.of_constr (with_stats (lazy (cbv_norm_term infos (subs_id 0) constr)))
+  EConstr.of_constr (cbv_norm_term infos (subs_id 0) constr)
 
 (* constant bodies are normalized at the first expansion *)
 let create_cbv_infos reds env sigma =
