@@ -8,8 +8,6 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
-open Sorts
-open Util
 open Pp
 open Names
 open Constr
@@ -77,10 +75,10 @@ let define_pure_evar_as_product env evd evk =
   let open Context.Named.Declaration in
   let evi = Evd.find_undefined evd evk in
   let evenv = evar_env env evi in
-  let id = next_ident_away idx (Environ.ids_of_named_context_val evi.evar_hyps) in
-  let concl = Reductionops.whd_all evenv evd evi.evar_concl in
+  let id = next_ident_away idx (Environ.ids_of_named_context_val (Evd.evar_hyps evi)) in
+  let concl = Reductionops.whd_all evenv evd (Evd.evar_concl evi) in
   let s = destSort evd concl in
-  let evksrc = evar_source evk evd in
+  let evksrc = evar_source evi in
   let src = subterm_source evk ~where:Domain evksrc in
   let evd1,(dom,u1) =
     new_type_evar evenv evd univ_flexible_alg ~src ~filter:(evar_filter evi)
@@ -98,11 +96,11 @@ let define_pure_evar_as_product env evd evk =
         let evd3, (rng, srng) =
           new_type_evar newenv evd1 status ~src ~filter
         in
-        let prods = Univ.sup (univ_of_sort u1) (univ_of_sort srng) in
-        let evd3 = Evd.set_leq_sort evenv evd3 (Sorts.sort_of_univ prods) (ESorts.kind evd1 s) in
+        let prods = Typeops.sort_of_product env u1 srng in
+        let evd3 = Evd.set_leq_sort evenv evd3 prods (ESorts.kind evd1 s) in
           evd3, rng
   in
-  let prod = mkProd (make_annot (Name id) rdom, dom, subst_var id rng) in
+  let prod = mkProd (make_annot (Name id) rdom, dom, subst_var evd2 id rng) in
   let evd3 = Evd.define evk prod evd2 in
     evd3,prod
 
@@ -113,7 +111,7 @@ let define_evar_as_product env evd (evk,args) =
   (* Quick way to compute the instantiation of evk with args *)
   let na,dom,rng = destProd evd prod in
   let evdom = mkEvar (fst (destEvar evd dom), args) in
-  let evrngargs = mkRel 1 :: List.map (lift 1) args in
+  let evrngargs = SList.cons (mkRel 1) (SList.Skip.map (lift 1) args) in
   let evrng =  mkEvar (fst (destEvar evd rng), evrngargs) in
   evd, mkProd (na, evdom, evrng)
 
@@ -135,24 +133,24 @@ let define_pure_evar_as_lambda env evd evk =
   | Prod (na,dom,rng) -> (evd,(na,dom,rng))
   | Evar ev' -> let evd,typ = define_evar_as_product env evd ev' in evd,destProd evd typ
   | _ -> error_not_product env evd typ in
-  let avoid = Environ.ids_of_named_context_val evi.evar_hyps in
+  let avoid = Environ.ids_of_named_context_val (Evd.evar_hyps evi) in
   let id =
     map_annot (fun na -> next_name_away_with_default_using_types "x" na avoid
       (Reductionops.whd_evar evd dom)) na
   in
   let newenv = push_named (LocalAssum (id, dom)) evenv in
   let filter = Filter.extend 1 (evar_filter evi) in
-  let src = subterm_source evk ~where:Body (evar_source evk evd1) in
-  let abstract_arguments = Abstraction.abstract_last evi.evar_abstract_arguments in
+  let src = subterm_source evk ~where:Body (evar_source (Evd.find evd1 evk)) in
+  let abstract_arguments = Abstraction.abstract_last (Evd.evar_abstract_arguments evi) in
   let evd2,body = new_evar newenv evd1 ~src (subst1 (mkVar id.binder_name) rng) ~filter ~abstract_arguments in
-  let lam = mkLambda (map_annot Name.mk_name id, dom, subst_var id.binder_name body) in
+  let lam = mkLambda (map_annot Name.mk_name id, dom, subst_var evd2 id.binder_name body) in
   Evd.define evk lam evd2, lam
 
 let define_evar_as_lambda env evd (evk,args) =
   let evd,lam = define_pure_evar_as_lambda env evd evk in
   (* Quick way to compute the instantiation of evk with args *)
   let na,dom,body = destLambda evd lam in
-  let evbodyargs = mkRel 1 :: List.map (lift 1) args in
+  let evbodyargs = SList.cons (mkRel 1) (SList.Skip.map (lift 1) args) in
   let evbody = mkEvar (fst (destEvar evd body), evbodyargs) in
   evd, mkLambda (na, dom, evbody)
 
@@ -163,43 +161,67 @@ let rec evar_absorb_arguments env evd (evk,args as ev) = function
       let evd,lam = define_pure_evar_as_lambda env evd evk in
       let _,_,body = destLambda evd lam in
       let evk = fst (destEvar evd body) in
-      evar_absorb_arguments env evd (evk, a :: args) l
+      evar_absorb_arguments env evd (evk, SList.cons a args) l
 
 (* Refining an evar to a sort *)
 
 let define_evar_as_sort env evd (ev,args) =
   let evd, s = new_sort_variable univ_rigid evd in
   let evi = Evd.find_undefined evd ev in
-  let concl = Reductionops.whd_all (evar_env env evi) evd evi.evar_concl in
+  let concl = Reductionops.whd_all (evar_env env evi) evd (Evd.evar_concl evi) in
   let sort = destSort evd concl in
   let evd' = Evd.define ev (mkSort s) evd in
   Evd.set_leq_sort env evd' (Sorts.super s) (ESorts.kind evd' sort), s
 
-(* Propagation of constraints through application and abstraction:
-   Given a type constraint on a functional term, returns the type
-   constraint on its domain and codomain. If the input constraint is
-   an evar instantiate it with the product of 2 new evars. *)
+(* Unify with unknown array *)
 
-let split_tycon ?loc env evd tycon =
-  let rec real_split evd c =
-    let t = Reductionops.whd_all env evd c in
-      match EConstr.kind evd t with
-        | Prod (na,dom,rng) -> evd, (na, dom, rng)
-        | Evar ev (* ev is undefined because of whd_all *) ->
-            let (evd',prod) = define_evar_as_product env evd ev in
-            let (na,dom,rng) = destProd evd prod in
-            let anon = {na with binder_name = Anonymous} in
-              evd',(anon, dom, rng)
-        | App (c,args) when isEvar evd c ->
-            let (evd',lam) = define_evar_as_lambda env evd (destEvar evd c) in
-            real_split evd' (mkApp (lam,args))
-        | _ -> error_not_product ?loc env evd c
-  in
-    match tycon with
-      | None -> evd,(make_annot Anonymous Relevant,None,None)
-      | Some c ->
-          let evd', (n, dom, rng) = real_split evd c in
-            evd', (n, mk_tycon dom, mk_tycon rng)
+let rec presplit env sigma c =
+  let c = Reductionops.whd_all env sigma c in
+  match EConstr.kind sigma c with
+  | App (h,args) when isEvar sigma h ->
+    let sigma, lam = define_evar_as_lambda env sigma (destEvar sigma h) in
+    (* XXX could be just whd_all -> no recursion? *)
+    presplit env sigma (mkApp (lam, args))
+  | _ -> sigma, c
+
+let define_pure_evar_as_array env sigma evk =
+  let evi = Evd.find_undefined sigma evk in
+  let evenv = evar_env env evi in
+  let evksrc = evar_source evi in
+  let src = subterm_source evk ~where:Domain evksrc in
+  let sigma, u = new_univ_level_variable univ_flexible sigma in
+  let s' = Sorts.sort_of_univ @@ Univ.Universe.make u in
+  let sigma, ty = new_evar evenv sigma ~typeclass_candidate:false ~src ~filter:(evar_filter evi) (mkSort s') in
+  let concl = Reductionops.whd_all evenv sigma (Evd.evar_concl evi) in
+  let s = destSort sigma concl in
+  (* array@{u} ty : Type@{u} <= Type@{s} *)
+  let sigma = Evd.set_leq_sort env sigma s' (ESorts.kind sigma s) in
+  let ar = Typeops.type_of_array env (Univ.Instance.of_array [|u|]) in
+  let sigma = Evd.define evk (mkApp (EConstr.of_constr ar, [| ty |])) sigma in
+  sigma
+
+let is_array_const env sigma c =
+  match EConstr.kind sigma c with
+  | Const (cst,_) ->
+    (match env.Environ.retroknowledge.Retroknowledge.retro_array with
+     | None -> false
+     | Some cst' -> Environ.QConstant.equal env cst cst')
+  | _ -> false
+
+let split_as_array env sigma0 = function
+  | None -> sigma0, None
+  | Some c ->
+    let sigma, c = presplit env sigma0 c in
+    match EConstr.kind sigma c with
+    | App (h,[|ty|]) when is_array_const env sigma h -> sigma, Some ty
+    | Evar ev ->
+      let sigma = define_pure_evar_as_array env sigma (fst ev) in
+      let ty = match EConstr.kind sigma c with
+        | App (_,[|ty|]) -> ty
+        | _ -> assert false
+      in
+      sigma, Some ty
+    | _ -> sigma0, None
 
 let valcon_of_tycon x = x
 let lift_tycon n = Option.map (lift n)
