@@ -21,42 +21,15 @@ module NamedDecl = Context.Named.Declaration
 
 (** Processing occurrences *)
 
-type occurrence_error =
-  | InvalidOccurrence of int list
-  | IncorrectInValueOccurrence of Id.t
-
-let explain_invalid_occurrence l =
-  let l = List.sort_uniquize Int.compare l in
-  str ("Invalid occurrence " ^ String.plural (List.length l) "number" ^": ")
-  ++ prlist_with_sep spc int l ++ str "."
-
 let explain_incorrect_in_value_occurrence id =
   Id.print id ++ str " has no value."
 
-let explain_occurrence_error = function
-  | InvalidOccurrence l -> explain_invalid_occurrence l
-  | IncorrectInValueOccurrence id -> explain_incorrect_in_value_occurrence id
-
-let error_occurrences_error e =
-  user_err  (explain_occurrence_error e)
-
-let error_invalid_occurrence occ =
-  error_occurrences_error (InvalidOccurrence occ)
-
-let check_used_occurrences nbocc (nowhere_except_in,locs) =
-  let rest = List.filter (fun o -> o >= nbocc) locs in
-  match rest with
-  | [] -> ()
-  | _ -> error_occurrences_error (InvalidOccurrence rest)
-
 let proceed_with_occurrences f occs x =
   match occs with
   | NoOccurrences -> x
   | occs ->
-    let plocs = Locusops.convert_occs occs in
-    assert (List.for_all (fun x -> x >= 0) (snd plocs));
-    let (nbocc,x) = f 1 x in
-    check_used_occurrences nbocc plocs;
+    let (occs,x) = f (Locusops.initialize_occurrence_counter occs) x in
+    Locusops.check_used_occurrences occs;
     x
 
 (** Applying a function over a named_declaration with an hypothesis
@@ -70,7 +43,7 @@ let map_named_declaration_with_hyploc f hyploc acc decl =
   in
   match decl,hyploc with
   | LocalAssum (id,_), InHypValueOnly ->
-      error_occurrences_error (IncorrectInValueOccurrence id.Context.binder_name)
+      user_err (explain_incorrect_in_value_occurrence id.Context.binder_name)
   | LocalAssum (id,typ), _ ->
       let acc,typ = f acc typ in acc, LocalAssum (id,typ)
   | LocalDef (id,body,typ), InHypTypeOnly ->
@@ -100,57 +73,57 @@ type 'a testing_function = {
    (b,l), b=true means no occurrence except the ones in l and b=false,
    means all occurrences except the ones in l *)
 
-let replace_term_occ_gen_modulo sigma occs like_first test bywhat cl occ t =
-  let (nowhere_except_in,locs) = Locusops.convert_occs occs in
-  let maxocc = List.fold_right max locs 0 in
-  let pos = ref occ in
+let replace_term_occ_gen_modulo env sigma like_first test bywhat cl count t =
+  let count = ref count in
   let nested = ref false in
-  let add_subst t subst =
+  let add_subst pos t subst =
     try
       test.testing_state <- test.merge_fun subst test.testing_state;
-      test.last_found <- Some ((cl,!pos),t)
+      test.last_found <- Some ((cl,pos),t)
     with NotUnifiable e when not like_first ->
       let lastpos = Option.get test.last_found in
-     raise (SubtermUnificationError (!nested,((cl,!pos),t),lastpos,e)) in
+     raise (SubtermUnificationError (!nested,((cl,pos),t),lastpos,e)) in
   let rec substrec k t =
-    if nowhere_except_in && !pos > maxocc then t else
+    if Locusops.occurrences_done !count then t else
     try
       let subst = test.match_fun test.testing_state t in
-      if Locusops.is_selected !pos occs then
+      let selected, count' = Locusops.update_occurrence_counter !count in count := count';
+      if selected then
+        let pos = Locusops.current_occurrence !count in
         (if !nested then begin
           (* in case it is nested but not later detected as unconvertible,
              as when matching "id _" in "id (id 0)" *)
           let lastpos = Option.get test.last_found in
-          raise (SubtermUnificationError (!nested,((cl,!pos),t),lastpos,None))
+          raise (SubtermUnificationError (!nested,((cl,pos),t),lastpos,None))
          end;
-         add_subst t subst; incr pos;
+         add_subst pos t subst;
          (* Check nested matching subterms *)
-         if not (Locusops.is_all_occurrences occs) && occs != Locus.NoOccurrences then
+         if Locusops.more_specific_occurrences !count then
            begin nested := true; ignore (subst_below k t); nested := false end;
          (* Do the effective substitution *)
          Vars.lift k (bywhat ()))
       else
-        (incr pos; subst_below k t)
+        subst_below k t
     with NotUnifiable _ ->
       subst_below k t
   and subst_below k t =
-    map_constr_with_binders_left_to_right sigma (fun d k -> k+1) substrec k t
+    map_constr_with_binders_left_to_right env sigma (fun d k -> k+1) substrec k t
   in
   let t' = substrec 0 t in
-  (!pos, t')
+  (!count, t')
 
-let replace_term_occ_modulo evd occs test bywhat t =
+let replace_term_occ_modulo env evd occs test bywhat t =
   let occs',like_first =
     match occs with AtOccs occs -> occs,false | LikeFirst -> AllOccurrences,true in
   proceed_with_occurrences
-    (replace_term_occ_gen_modulo evd occs' like_first test bywhat None) occs' t
+    (replace_term_occ_gen_modulo env evd like_first test bywhat None) occs' t
 
-let replace_term_occ_decl_modulo evd occs test bywhat d =
+let replace_term_occ_decl_modulo env evd occs test bywhat d =
   let (plocs,hyploc),like_first =
     match occs with AtOccs occs -> occs,false | LikeFirst -> (AllOccurrences,InHyp),true in
   proceed_with_occurrences
     (map_named_declaration_with_hyploc
-       (replace_term_occ_gen_modulo evd plocs like_first test bywhat)
+       (replace_term_occ_gen_modulo env evd like_first test bywhat)
        hyploc)
     plocs d
 
@@ -162,7 +135,7 @@ let make_eq_univs_test env evd c =
     | None -> raise (NotUnifiable None)
     | Some cst ->
         try Evd.add_universe_constraints evd cst
-        with Evd.UniversesDiffer -> raise (NotUnifiable None)
+        with Evd.UniversesDiffer | UGraph.UniverseInconsistency _ -> raise (NotUnifiable None)
     );
   merge_fun = (fun evd _ -> evd);
   testing_state = evd;
@@ -172,7 +145,7 @@ let make_eq_univs_test env evd c =
 let subst_closed_term_occ env evd occs c t =
   let test = make_eq_univs_test env evd c in
   let bywhat () = mkRel 1 in
-  let t' = replace_term_occ_modulo evd occs test bywhat t in
+  let t' = replace_term_occ_modulo env evd occs test bywhat t in
     t', test.testing_state
 
 let subst_closed_term_occ_decl env evd occs c d =
@@ -182,6 +155,6 @@ let subst_closed_term_occ_decl env evd occs c d =
   let bywhat () = mkRel 1 in
   proceed_with_occurrences
     (map_named_declaration_with_hyploc
-       (fun _ -> replace_term_occ_gen_modulo evd plocs like_first test bywhat None)
+       (fun _ -> replace_term_occ_gen_modulo env evd like_first test bywhat None)
        hyploc) plocs d,
   test.testing_state
