@@ -42,7 +42,7 @@ let make ~hypnaming env sigma lvar =
   let get_extra env sigma =
     let avoid = Environ.ids_of_named_context_val (Environ.named_context_val env) in
     Context.Rel.fold_outside (fun d acc -> push_rel_decl_to_named_context ~hypnaming sigma d acc)
-      (rel_context env) ~init:(empty_csubst, avoid, named_context env) in
+      (rel_context env) ~init:(empty_csubst, avoid, named_context_val env) in
   {
     static_env = env;
     renamed_env = env;
@@ -51,6 +51,8 @@ let make ~hypnaming env sigma lvar =
   }
 
 let env env = env.static_env
+let renamed_env env = env.renamed_env
+let lfun env = env.lvar.ltac_genargs
 
 let vars_of_env env =
   Id.Set.union (Id.Map.domain env.lvar.ltac_genargs) (vars_of_env env.static_env)
@@ -67,7 +69,8 @@ let ltac_interp_id { ltac_idents ; ltac_genargs } id =
 let ltac_interp_name lvar = Nameops.Name.map (ltac_interp_id lvar)
 
 let push_rel ~hypnaming sigma d env =
-  let d' = Context.Rel.Declaration.map_name (ltac_interp_name env.lvar) d in
+  let open Context.Rel.Declaration in
+  let d' = map_name (ltac_interp_name env.lvar) d in
   let env = {
     static_env = push_rel d env.static_env;
     renamed_env = push_rel d' env.renamed_env;
@@ -94,15 +97,13 @@ let push_rec_types ~hypnaming sigma (lna,typarray) env =
   let env,ctx = Array.fold_left_map (fun e assum -> let (d,e) = push_rel sigma assum e ~hypnaming in (e,d)) env ctxt in
   Array.map get_annot ctx, env
 
-let new_evar env sigma ?src ?naming typ =
-  let open Context.Named.Declaration in
-  let inst_vars = List.map (get_id %> mkVar) (named_context env.renamed_env) in
-  let inst_rels = List.rev (rel_list 0 (nb_rel env.renamed_env)) in
-  let (subst, _, nc) = Lazy.force env.extra in
-  let typ' = csubst_subst subst typ in
-  let instance = inst_rels @ inst_vars in
-  let sign = val_of_named_context nc in
-  new_evar_instance sign sigma typ' ?src ?naming instance
+let new_evar env sigma ?src ?(naming = Namegen.IntroAnonymous) typ =
+  let (subst, _, sign) as ext = Lazy.force env.extra in
+  let instance = Evarutil.default_ext_instance ext in
+  let typ' = csubst_subst sigma subst typ in
+  let name = Evarutil.next_evar_name sigma naming in
+  let (sigma, evk) = new_pure_evar sign sigma typ' ?src ?name in
+  (sigma, mkEvar (evk, instance))
 
 let new_type_evar env sigma ~src =
   let sigma, s = Evd.new_sort_variable Evd.univ_flexible_alg sigma in
@@ -160,7 +161,7 @@ let interp_ltac_variable ?loc typing_fun env sigma id : Evd.evar_map * unsafe_ju
       ltac_constrs = closure.typed;
       ltac_uconstrs = closure.untyped;
       ltac_idents = closure.idents;
-      ltac_genargs = Id.Map.empty; }
+      ltac_genargs = closure.genargs; }
     in
     (* spiwack: I'm catching [Not_found] potentially too eagerly
        here, as the call to the main pretyping function is caught
@@ -176,14 +177,23 @@ let interp_ltac_variable ?loc typing_fun env sigma id : Evd.evar_map * unsafe_ju
      (str "Variable " ++ Id.print id ++ str " should be bound to a term but is \
       bound to a " ++ Geninterp.Val.pr typ ++ str ".")
   end;
+  if Id.Map.mem id env.lvar.ltac_idents then begin
+    let bnd = Id.Map.find id env.lvar.ltac_idents in
+    user_err ?loc
+     (str "Variable " ++ Id.print id ++ str " should be bound to a term but is \
+      bound to the identifier " ++ quote (Id.print bnd) ++ str ".")
+  end;
   raise Not_found
 
 let interp_ltac_id env id = ltac_interp_id env.lvar id
 
+type 'a obj_interp_fun =
+  ?loc:Loc.t -> poly:bool -> t -> Evd.evar_map -> Evardefine.type_constraint ->
+  'a -> unsafe_judgment * Evd.evar_map
+
 module ConstrInterpObj =
 struct
-  type ('r, 'g, 't) obj =
-    unbound_ltac_var_map -> bool -> env -> Evd.evar_map -> types -> 'g -> constr * Evd.evar_map
+  type ('r, 'g, 't) obj = 'g obj_interp_fun
   let name = "constr_interp"
   let default _ = None
 end
@@ -192,8 +202,8 @@ module ConstrInterp = Genarg.Register(ConstrInterpObj)
 
 let register_constr_interp0 = ConstrInterp.register0
 
-let interp_glob_genarg env poly sigma ty arg =
+let interp_glob_genarg ?loc ~poly env sigma ty arg =
   let open Genarg in
   let GenArg (Glbwit tag, arg) = arg in
   let interp = ConstrInterp.obj tag in
-  interp env.lvar.ltac_genargs poly env.renamed_env sigma ty arg
+  interp ?loc ~poly env sigma ty arg
