@@ -19,6 +19,7 @@ open Names
 open Inductive
 open Util
 open Nativecode
+open Values
 open Nativevalues
 open Context.Rel.Declaration
 
@@ -65,17 +66,8 @@ let get_available_profile_filename () =
   let profile_filename = get_profile_filename () in
   let dir = Filename.dirname profile_filename in
   let base = Filename.basename profile_filename in
-  (* starting with OCaml 4.04, could use Filename.remove_extension and Filename.extension, which
-     gets rid of need for exception-handling here
-  *)
-  let (name,ext) =
-    try
-      let nm = Filename.chop_extension base in
-      let nm_len = String.length nm in
-      let ex = String.sub base nm_len (String.length base - nm_len) in
-      (nm,ex)
-    with Invalid_argument _ -> (base,"")
-  in
+  let name = Filename.remove_extension base in
+  let ext = Filename.extension base in
   try
     (* unlikely race: fn deleted, another process uses fn *)
     Filename.temp_file ~temp_dir:dir (name ^ "_") ext
@@ -102,43 +94,34 @@ let decompose_prod env t =
   in
   (name,dom,codom)
 
-let app_type env c =
-  let t = whd_all env c in
-  try destApp t with DestKO -> (t,[||])
+let e_whd_all = Reductionops.clos_whd_flags CClosure.all
 
+let app_type env sigma c =
+  let t = e_whd_all env sigma c in
+  decompose_appvect (EConstr.Unsafe.to_constr t)
 
-let find_rectype_a env c =
-  let (t, l) = app_type env c in
+let find_rectype_a env sigma c =
+  let (t, l) = app_type env sigma c in
   match kind t with
   | Ind ind -> (ind, l)
   | _ -> raise Not_found
 
 (* Instantiate inductives and parameters in constructor type *)
 
-let type_constructor mind mib u (ctx, typ) params =
-  let typ = it_mkProd_or_LetIn typ ctx in
-  let s = ind_subst mind mib u in
-  let ctyp = substl s typ in
-  let nparams = Array.length params in
-  if Int.equal nparams 0 then ctyp
-  else
-    let _,ctyp = decompose_prod_n nparams ctyp in
-    substl (List.rev (Array.to_list params)) ctyp
-
-let construct_of_constr_notnative const env tag (mind, _ as ind) u allargs =
+let construct_of_constr_notnative const env tag (ind,u) allargs =
   let mib,mip = lookup_mind_specif env ind in
   let nparams = mib.mind_nparams in
   let params = Array.sub allargs 0 nparams in
   let i = invert_tag const tag mip.mind_reloc_tbl in
-  let ctyp = type_constructor mind mib u (mip.mind_nf_lc.(i-1)) params in
+  let ctyp = Inductiveops.instantiate_constructor_params ((ind,i),u) (mib,mip) (Array.to_list params) in
   (mkApp(mkConstructU((ind,i),u), params), ctyp)
 
-
 let construct_of_constr const env sigma tag typ =
-  let t, l = app_type env typ in
-  match EConstr.kind_upto sigma t with
-  | Ind (ind,u) ->
-      construct_of_constr_notnative const env tag ind u l
+  let typ = Reductionops.clos_whd_flags CClosure.all env sigma (EConstr.of_constr typ) in
+  let t, l = decompose_appvect (EConstr.Unsafe.to_constr typ) in
+  match Constr.kind t with
+  | Ind indu ->
+      construct_of_constr_notnative const env tag indu l
   | _ ->
     assert (Constr.equal t (Typeops.type_of_int env));
     (mkInt (Uint63.of_int tag), t)
@@ -148,17 +131,20 @@ let construct_of_constr_const env sigma tag typ =
 
 let construct_of_constr_block = construct_of_constr false
 
-let build_branches_type env sigma (mind,_ as _ind) mib mip u params p =
+let get_case_annot decls =
+  Array.map_of_list (fun decl -> get_annot decl) (List.rev decls)
+
+let build_branches_type env sigma mib mip (ind,u) params (pctx, p) =
   let rtbl = mip.mind_reloc_tbl in
+  let paramsl = Array.to_list params in
   (* [build_one_branch i cty] construit le type de la ieme branche (commence
      a 0) et les lambda correspondant aux realargs *)
-  let build_one_branch i cty =
-    let typi = type_constructor mind mib u cty params in
+  let p = it_mkLambda_or_LetIn p pctx in (* TODO: prevent useless cut? *)
+  let build_one_branch i (ctx, _) =
+    let typi = Inductiveops.instantiate_constructor_params ((ind,i+1),u) (mib,mip) paramsl in
     let decl,indapp = Reductionops.splay_prod env sigma (EConstr.of_constr typi) in
     let decl = List.map (on_snd EConstr.Unsafe.to_constr) decl in
-    let indapp = EConstr.Unsafe.to_constr indapp in
-    let decl_with_letin,_ = decompose_prod_assum typi in
-    let ind,cargs = find_rectype_a env indapp in
+    let ind,cargs = find_rectype_a env sigma indapp in
     let nparams = Array.length params in
     let carity = snd (rtbl.(i)) in
     let crealargs = Array.sub cargs nparams (Array.length cargs - nparams) in
@@ -171,10 +157,19 @@ let build_branches_type env sigma (mind,_ as _ind) mib mip u params p =
       let dep_cstr = mkApp(mkApp(mkConstructU (cstr,snd ind),params),relargs) in
       mkApp(papp,[|dep_cstr|])
     in
-    decl, decl_with_letin, codom
-  in Array.mapi build_one_branch mip.mind_nf_lc
+    let decl_with_letin = List.firstn mip.mind_consnrealdecls.(i) ctx in
+    let nas = get_case_annot decl_with_letin in
+    let rec get_lift decls = match decls with
+    | [] -> Esubst.el_id
+    | LocalDef _ :: decls -> Esubst.el_shft 1 (get_lift decls)
+    | LocalAssum _ :: decls -> Esubst.el_lift (get_lift decls)
+    in
+    decl, nas, get_lift decl_with_letin, codom
+  in
+  Array.mapi build_one_branch mip.mind_nf_lc
 
-let build_case_type p realargs c =
+let build_case_type (pctx, p) realargs c =
+  let p = it_mkLambda_or_LetIn p pctx in (* TODO: prevent useless cut? *)
   mkApp(mkApp(p, realargs), [|c|])
 
 (* normalisation of values *)
@@ -186,20 +181,18 @@ let branch_of_switch lvl ans bs =
     let ci =
       if Int.equal arity 0 then mk_const tag
       else mk_block tag (mk_rels_accu lvl arity) in
-    bs ci in
+    apply bs ci in
   Array.init (Array.length tbl) branch
 
 let get_proj env (ind, proj_arg) =
-  let mib = Environ.lookup_mind (fst ind) env in
-  match Declareops.inductive_make_projection ind mib ~proj_arg with
-  | None ->
-    CErrors.anomaly (Pp.strbrk "Return type is not a primitive record")
-  | Some p ->
-    Projection.make p true
+  let p = Environ.get_projection env ind ~proj_arg in
+  Projection.make p true
 
 let rec nf_val env sigma v typ =
   match kind_of_value v with
   | Vaccu accu -> nf_accu env sigma accu
+  | Vprod (na, dom, codom) -> fst @@ nf_prod env sigma (na, dom, codom)
+  | Vfix e | Vcofix e -> Empty.abort e
   | Vfun f ->
       let lvl = nb_rel env in
       let name,dom,codom =
@@ -214,6 +207,7 @@ let rec nf_val env sigma v typ =
   | Vconst n -> construct_of_constr_const env sigma n typ
   | Vint64 i -> i |> Uint63.of_int64 |> mkInt
   | Vfloat64 f -> f |> Float64.of_float |> mkFloat
+  | Varray t -> nf_array env sigma t typ
   | Vblock b ->
       let capp,ctyp = construct_of_constr_block env sigma (block_tag b) typ in
       let args = nf_bargs env sigma b ctyp in
@@ -222,6 +216,7 @@ let rec nf_val env sigma v typ =
 and nf_type env sigma v =
   match kind_of_value v with
   | Vaccu accu -> nf_accu env sigma accu
+  | Vprod (na, dom, codom) -> fst @@ nf_prod env sigma (na, dom, codom)
   | _ -> assert false
 
 and nf_type_sort env sigma v =
@@ -235,6 +230,7 @@ and nf_type_sort env sigma v =
           CErrors.anomaly (Pp.str "Value should be a sort")
       in
       t, s
+  | Vprod (na, dom, codom) -> nf_prod env sigma (na, dom, codom)
   | _ -> assert false
 
 and nf_accu env sigma accu =
@@ -281,6 +277,15 @@ and nf_bargs env sigma b t =
       let c = nf_val env sigma (block_field b i) dom in
       t := subst1 c codom; c)
 
+and nf_prod env sigma (na, dom, codom) =
+  let dom, sdom = nf_type_sort env sigma dom in
+  let rdom = Sorts.relevance_of_sort sdom in
+  let na = make_annot na rdom in
+  let vn = mk_rel_accu (nb_rel env) in
+  let env = push_rel (LocalAssum (na, dom)) env in
+  let codom, scodom = nf_type_sort env sigma (apply codom vn) in
+  mkProd (na, dom, codom), destSort (Typeops.type_of_product env na sdom scodom)
+
 and nf_atom env sigma atom =
   match atom with
   | Arel i -> mkRel (nb_rel env - i)
@@ -288,15 +293,6 @@ and nf_atom env sigma atom =
   | Aind ind -> mkIndU ind
   | Asort s -> mkSort s
   | Avar id -> mkVar id
-  | Aprod(n,dom,codom) ->
-    let dom, sdom = nf_type_sort env sigma dom in
-    let rdom = Sorts.relevance_of_sort sdom in
-    let n = make_annot n rdom in
-    let vn = mk_rel_accu (nb_rel env) in
-    let env = push_rel (LocalAssum (n,dom)) env in
-    let codom = nf_type env sigma (codom vn) in
-    mkProd(n,dom,codom)
-  | Ameta (mv,_) -> mkMeta mv
   | Aproj (p, c) ->
       let c = nf_accu env sigma c in
       let p = get_proj env p in
@@ -318,28 +314,33 @@ and nf_atom_type env sigma atom =
       mkVar id, Typeops.type_of_variable env id
   | Acase(ans,accu,p,bs) ->
       let a,ta = nf_accu_type env sigma accu in
-      let ((mind,_),u as ind),allargs = find_rectype_a env ta in
+      let ((mind,_),u as ind),allargs = find_rectype_a env sigma (EConstr.of_constr ta) in
       let (mib,mip) = Inductive.lookup_mind_specif env (fst ind) in
       let nparams = mib.mind_nparams in
       let params,realargs = Array.chop nparams allargs in
+      let iv = if Typeops.should_invert_case env ans.asw_ci then
+          CaseInvert {indices=realargs}
+        else NoInvert
+      in
       let nparamdecls = Context.Rel.length (Inductive.inductive_paramdecls (mib,u)) in
       let pT =
         hnf_prod_applist_assum env nparamdecls
           (Inductiveops.type_of_inductive env ind) (Array.to_list params) in
-      let p = nf_predicate env sigma ind mip params p pT in
+      let pctx, p = nf_predicate env sigma ind mip params [] p pT in
       (* Calcul du type des branches *)
-      let btypes = build_branches_type env sigma (fst ind) mib mip u params p in
+      let btypes = build_branches_type env sigma mib mip ind params (pctx, p) in
       (* calcul des branches *)
       let bsw = branch_of_switch (nb_rel env) ans bs in
       let mkbranch i v =
-       let decl,decl_with_letin,codom = btypes.(i) in
-       let b = nf_val (Termops.push_rels_assum decl env) sigma v codom in
-        Termops.it_mkLambda_or_LetIn_from_no_LetIn b decl_with_letin
+        let decl, nas, lft, codom = btypes.(i) in
+        let b = nf_val (Termops.push_rels_assum decl env) sigma v codom in
+        nas, exliftn lft b
       in
       let branchs = Array.mapi mkbranch bsw in
-      let tcase = build_case_type p realargs a in
+      let tcase = build_case_type (pctx, p) realargs a in
+      let p = (get_case_annot pctx, p) in
       let ci = ans.asw_ci in
-      mkCase(ci, p, a, branchs), tcase
+      mkCase (ci, u, params, p, iv, a, branchs), tcase
   | Afix(tt,ft,rp,s) ->
       let tt = Array.map (fun t -> nf_type_sort env sigma t) tt in
       let tt = Array.map fst tt and rt = Array.map snd tt in
@@ -355,7 +356,7 @@ and nf_atom_type env sigma atom =
       let norm_body i v = nf_val env sigma (napply v fargs) (lift nbfix tt.(i)) in
       let ft = Array.mapi norm_body ft in
       mkFix((rp,s),(names,tt,ft)), tt.(s)
-  | Acofix(tt,ft,s,_) | Acofixe(tt,ft,s,_) ->
+  | Acofix(tt,ft,s,_) ->
       let tt = Array.map (fun t -> nf_type_sort env sigma t) tt in
       let tt = Array.map fst tt and rt = Array.map snd tt in
       let name = Name (Id.of_string "Fcofix") in
@@ -365,19 +366,8 @@ and nf_atom_type env sigma atom =
       let env = push_rec_types (names,tt,[||]) env in
       let ft = Array.mapi (fun i v -> nf_val env sigma (napply v fargs) tt.(i)) ft in
       mkCoFix(s,(names,tt,ft)), tt.(s)
-  | Aprod(n,dom,codom) ->
-      let dom,s1 = nf_type_sort env sigma dom in
-      let r1 = Sorts.relevance_of_sort s1 in
-      let n = make_annot n r1 in
-      let vn = mk_rel_accu (nb_rel env) in
-      let env = push_rel (LocalAssum (n,dom)) env in
-      let codom,s2 = nf_type_sort env sigma (codom vn) in
-      mkProd(n,dom,codom), Typeops.type_of_product env n s1 s2
   | Aevar(evk,args) ->
     nf_evar env sigma evk args
-  | Ameta(mv,ty) ->
-     let ty = nf_type env sigma ty in
-     mkMeta mv, ty
   | Aproj(p,c) ->
       let c,tc = nf_accu_type env sigma c in
       let cj = make_judge c tc in
@@ -386,21 +376,19 @@ and nf_atom_type env sigma atom =
       uj.uj_val, uj.uj_type
 
 
-and  nf_predicate env sigma ind mip params v pT =
+and nf_predicate env sigma ind mip params pctx v pT =
   match kind (whd_allnolet env pT) with
   | LetIn (name,b,t,pT) ->
-      let body =
-        nf_predicate (push_rel (LocalDef (name,b,t)) env) sigma ind mip params v pT in
-      mkLetIn (name,b,t,body)
+    let decl = LocalDef (name, b, t) in
+    nf_predicate (push_rel decl env) sigma ind mip params (decl :: pctx) v pT
   | Prod (name,dom,codom) -> begin
     match kind_of_value v with
     | Vfun f ->
       let k = nb_rel env in
       let vb = f (mk_rel_accu k) in
-      let body =
-        nf_predicate (push_rel (LocalAssum (name,dom)) env) sigma ind mip params vb codom in
-      mkLambda(name,dom,body)
-    | _ -> nf_type env sigma v
+      let decl = LocalAssum (name, dom) in
+      nf_predicate (push_rel decl env) sigma ind mip params (decl :: pctx) vb codom
+    | _ -> assert false
     end
   | _ ->
     match kind_of_value v with
@@ -414,33 +402,46 @@ and  nf_predicate env sigma ind mip params v pT =
       let dom = mkApp(mkIndU ind,Array.append params rargs) in
       let r = Inductive.relevance_of_inductive env (fst ind) in
       let name = make_annot name r in
-      let body = nf_type (push_rel (LocalAssum (name,dom)) env) sigma vb in
-      mkLambda(name,dom,body)
-    | _ -> nf_type env sigma v
+      let decl = LocalAssum (name, dom) in
+      let env = push_rel decl env in
+      let body = nf_type env sigma vb in
+      decl :: pctx, body
+    | _ -> assert false
 
 and nf_evar env sigma evk args =
   let evi = try Evd.find sigma evk with Not_found -> assert false in
-  let hyps = Environ.named_context_of_val (Evd.evar_filtered_hyps evi) in
-  let ty = EConstr.to_constr ~abort_on_undefined_evars:false sigma @@ Evd.evar_concl evi in
+  let hyps = EConstr.named_context_of_val (Evd.evar_filtered_hyps evi) in
   if List.is_empty hyps then begin
     assert (Array.is_empty args);
-    mkEvar (evk, []), ty
+    let ty = EConstr.to_constr ~abort_on_undefined_evars:false sigma @@ Evd.evar_concl evi in
+    mkEvar (evk, SList.empty), ty
   end
   else
     (* Let-bound arguments are present in the evar arguments but not
        in the type, so we turn the let into a product. *)
+    let ty = Evd.evar_concl evi in
     let hyps = Context.Named.drop_bodies hyps in
-    let fold accu d = Term.mkNamedProd_or_LetIn d accu in
+    let fold accu d = EConstr.mkNamedProd_or_LetIn sigma d accu in
     let t = List.fold_left fold ty hyps in
+    let t = EConstr.to_constr ~abort_on_undefined_evars:false sigma t in
     let ty, args = nf_args env sigma args t in
     (* nf_args takes arguments in the reverse order but produces them
        in the correct one, so we have to reverse them again for the
        evar node *)
-    mkEvar (evk, List.rev args), ty
+    EConstr.(Unsafe.to_constr @@ mkLEvar sigma (evk, List.rev_map of_constr args)), ty
+
+and nf_array env sigma t typ =
+  let ty, allargs = app_type env sigma (EConstr.of_constr typ) in
+  let typ_elem = allargs.(0) in
+  let vdef = Parray.default t in
+  (* Do not cast into an array out of fear that floats may sneak in *)
+  let init i = nf_val env sigma (Parray.get t (Uint63.of_int i)) typ_elem in
+  let t = Array.init (Parray.length_int t) init in
+  let u = snd (destConst ty) in
+  mkArray(u, t, nf_val env sigma vdef typ_elem, typ_elem)
 
 let evars_of_evar_map sigma =
-  { Nativelambda.evars_val = Evd.existential_opt_value0 sigma;
-    Nativelambda.evars_metas = Evd.meta_type0 sigma }
+  { Genlambda.evars_val = Evd.evar_handler sigma }
 
 (* fork perf process, return profiler's process id *)
 let start_profiler_linux profile_fn =
@@ -455,16 +456,24 @@ let start_profiler_linux profile_fn =
       [|perf; "record"; "-g"; "-o"; profile_fn; "-p"; string_of_int coq_pid |]
       Unix.stdin dev_null dev_null
   in
-  (* doesn't seem to be a way to test whether process creation succeeded *)
-  if !Flags.debug then
-    Feedback.msg_debug (Pp.str (Format.sprintf "Native compute profiler started, pid = %d, output to: %s" profiler_pid profile_fn));
+  (* doesn't seem to be a way to test whether process creation succeeded
+     (create_process doesn't raise until OCaml 4.12.0) *)
+  debug_native_compiler (fun () ->
+    Pp.str (Format.sprintf "Native compute profiler started, pid = %d, output to: %s" profiler_pid profile_fn));
   Some profiler_pid
 
+let start_profiler_linux fn =
+  try start_profiler_linux fn
+  with Unix.Unix_error _ as e ->
+    Feedback.msg_info
+      Pp.(str "Could not start native code profiler: " ++ str (Printexc.to_string e));
+    None
+
 (* kill profiler via SIGINT *)
 let stop_profiler_linux m_pid =
   match m_pid with
   | Some pid -> (
-    let _ = if !Flags.debug then Feedback.msg_debug (Pp.str "Stopping native code profiler") in
+    let _ = debug_native_compiler (fun () -> Pp.str "Stopping native code profiler") in
     try
       Unix.kill pid Sys.sigint;
       let _ = Unix.waitpid [] pid in ()
@@ -489,15 +498,9 @@ let stop_profiler m_pid =
   | _ -> ()
 
 let native_norm env sigma c ty =
+  Nativelib.link_libraries ();
   let c = EConstr.Unsafe.to_constr c in
   let ty = EConstr.Unsafe.to_constr ty in
-  if not (Flags.get_native_compiler ()) then
-    user_err Pp.(str "Native_compute reduction has been disabled.")
-  else
-  (*
-  Format.eprintf "Numbers of free variables (named): %i\n" (List.length vl1);
-  Format.eprintf "Numbers of free variables (rel): %i\n" (List.length vl2);
-  *)
     let profile = get_profiling_enabled () in
     let print_timing = get_timing_enabled () in
     let ml_filename, prefix = Nativelib.get_ml_filename () in
@@ -513,20 +516,18 @@ let native_norm env sigma c ty =
     if print_timing then Feedback.msg_info (Pp.str time_info);
     let profiler_pid = if profile then start_profiler () else None in
     let t0 = Unix.gettimeofday () in
-    Nativelib.call_linker ~fatal:true env ~prefix fn (Some upd);
+    let (rt1, _) = Nativelib.execute_library ~prefix fn upd in
     let t1 = Unix.gettimeofday () in
     if profile then stop_profiler profiler_pid;
     let time_info = Format.sprintf "native_compute: Evaluation done in %.5f" (t1 -. t0) in
     if print_timing then Feedback.msg_info (Pp.str time_info);
-    let res = nf_val env sigma !Nativelib.rt1 ty in
+    let res = nf_val env sigma rt1 ty in
     let t2 = Unix.gettimeofday () in
     let time_info = Format.sprintf "native_compute: Reification done in %.5f" (t2 -. t1) in
     if print_timing then Feedback.msg_info (Pp.str time_info);
     EConstr.of_constr res
 
-let native_conv_generic pb sigma t =
-  Nativeconv.native_conv_gen pb (evars_of_evar_map sigma) t
-
-let native_infer_conv ?(pb=Reduction.CUMUL) env sigma t1 t2 =
-  Reductionops.infer_conv_gen (fun pb ~l2r sigma ts -> native_conv_generic pb sigma)
-    ~catch_incon:true ~pb env sigma t1 t2
+let native_norm env sigma c ty =
+  if not (Flags.get_native_compiler ()) then
+    user_err Pp.(str "Native_compute reduction has been disabled.");
+  native_norm env sigma c ty
