@@ -23,8 +23,7 @@ open Environ
 
 let case_info_pattern_eq i1 i2 =
   i1.cip_style == i2.cip_style &&
-  Option.equal eq_ind i1.cip_ind i2.cip_ind &&
-  Option.equal (List.equal (==)) i1.cip_ind_tags i2.cip_ind_tags &&
+  Option.equal Ind.CanOrd.equal i1.cip_ind i2.cip_ind &&
   i1.cip_extensible == i2.cip_extensible
 
 let rec constr_pattern_eq p1 p2 = match p1, p2 with
@@ -51,7 +50,7 @@ let rec constr_pattern_eq p1 p2 = match p1, p2 with
   constr_pattern_eq t1 t2 && constr_pattern_eq l1 l2 && constr_pattern_eq r1 r2
 | PCase (info1, p1, r1, l1), PCase (info2, p2, r2, l2) ->
   case_info_pattern_eq info1 info2 &&
-  constr_pattern_eq p1 p2 &&
+  Option.equal (fun (nas1, p1) (nas2, p2) -> Array.equal Name.equal nas1 nas2 && constr_pattern_eq p1 p2) p1 p2 &&
   constr_pattern_eq r1 r2 &&
   List.equal pattern_eq l1 l2
 | PFix ((ln1,i1),f1), PFix ((ln2,i2),f2) ->
@@ -59,19 +58,22 @@ let rec constr_pattern_eq p1 p2 = match p1, p2 with
 | PCoFix (i1,f1), PCoFix (i2,f2) ->
   Int.equal i1 i2 && rec_declaration_eq f1 f2
 | PProj (p1, t1), PProj (p2, t2) ->
-   Projection.equal p1 p2 && constr_pattern_eq t1 t2
+   Projection.CanOrd.equal p1 p2 && constr_pattern_eq t1 t2
 | PInt i1, PInt i2 ->
    Uint63.equal i1 i2
 | PFloat f1, PFloat f2 ->
    Float64.equal f1 f2
+| PArray (t1, def1, ty1), PArray (t2, def2, ty2) ->
+  Array.equal constr_pattern_eq t1 t2 && constr_pattern_eq def1 def2
+  && constr_pattern_eq ty1 ty2
 | (PRef _ | PVar _ | PEvar _ | PRel _ | PApp _ | PSoApp _
    | PLambda _ | PProd _ | PLetIn _ | PSort _ | PMeta _
    | PIf _ | PCase _ | PFix _ | PCoFix _ | PProj _ | PInt _
-   | PFloat _), _ -> false
+   | PFloat _ | PArray _), _ -> false
 (** FIXME: fixpoint and cofixpoint should be relativized to pattern *)
 
 and pattern_eq (i1, j1, p1) (i2, j2, p2) =
-  Int.equal i1 i2 && List.equal (==) j1 j2 && constr_pattern_eq p1 p2
+  Int.equal i1 i2 && Array.equal Name.equal j1 j2 && constr_pattern_eq p1 p2
 
 and rec_declaration_eq (n1, c1, r1) (n2, c2, r2) =
   Array.equal Name.equal n1 n2 &&
@@ -89,10 +91,12 @@ let rec occur_meta_pattern = function
   | PIf (c,c1,c2)  ->
       (occur_meta_pattern c) ||
       (occur_meta_pattern c1) || (occur_meta_pattern c2)
-  | PCase(_,p,c,br) ->
-      (occur_meta_pattern p) ||
+  | PCase(_, p,c,br) ->
+      Option.cata (fun (_, p) -> occur_meta_pattern p) false p ||
       (occur_meta_pattern c) ||
       (List.exists (fun (_,_,p) -> occur_meta_pattern p) br)
+  | PArray (t,def,ty) ->
+      Array.exists occur_meta_pattern t || occur_meta_pattern def || occur_meta_pattern ty
   | PMeta _ | PSoApp _ -> true
   | PEvar _ | PVar _ | PRef _ | PRel _ | PSort _ | PFix _ | PCoFix _
     | PInt _ | PFloat _ -> false
@@ -110,10 +114,10 @@ let rec occurn_pattern n = function
   | PIf (c,c1,c2)  ->
       (occurn_pattern n c) ||
       (occurn_pattern n c1) || (occurn_pattern n c2)
-  | PCase(_,p,c,br) ->
-      (occurn_pattern n p) ||
+  | PCase(_, p, c, br) ->
+      Option.cata (fun (nas, p) -> occurn_pattern (Array.length nas + n) p) false p ||
       (occurn_pattern n c) ||
-      (List.exists (fun (_,_,p) -> occurn_pattern n p) br)
+      (List.exists (fun (_, nas, p) -> occurn_pattern (Array.length nas + n) p) br)
   | PMeta _ | PSoApp _ -> true
   | PEvar (_,args) -> List.exists (occurn_pattern n) args
   | PVar _ | PRef _ | PSort _ | PInt _ | PFloat _ -> false
@@ -121,6 +125,8 @@ let rec occurn_pattern n = function
      Array.exists (occurn_pattern n) tl || Array.exists (occurn_pattern (n+Array.length tl)) bl
   | PCoFix (_,(_,tl,bl)) ->
      Array.exists (occurn_pattern n) tl || Array.exists (occurn_pattern (n+Array.length tl)) bl
+  | PArray (t,def,ty) ->
+    Array.exists (occurn_pattern n) t || occurn_pattern n def || occurn_pattern n ty
 
 let noccurn_pattern n c = not (occurn_pattern n c)
 
@@ -139,7 +145,8 @@ let rec head_pattern_bound t =
         -> raise BoundPattern
     (* Perhaps they were arguments, but we don't beta-reduce *)
     | PLambda _ -> raise BoundPattern
-    | PCoFix _ | PInt _ | PFloat _ -> anomaly ~label:"head_pattern_bound" (Pp.str "not a type.")
+    | PCoFix _ | PInt _ | PFloat _ | PArray _ ->
+      anomaly ~label:"head_pattern_bound" (Pp.str "not a type.")
 
 let head_of_constr_reference sigma c = match EConstr.kind sigma c with
   | Const (sp,_) -> GlobRef.ConstRef sp
@@ -148,7 +155,9 @@ let head_of_constr_reference sigma c = match EConstr.kind sigma c with
   | Var id -> GlobRef.VarRef id
   | _ -> anomaly (Pp.str "Not a rigid reference.")
 
-let pattern_of_constr env sigma t =
+let pattern_of_constr ~broken env sigma t =
+  let t = EConstr.Unsafe.to_constr t in
+  let kind = if broken then Constr.kind else fun c -> EConstr.kind_upto sigma c in
   let rec pattern_of_constr env t =
   let open Context.Rel.Declaration in
   match kind t with
@@ -170,7 +179,7 @@ let pattern_of_constr env sigma t =
         (match
           match kind f with
           | Evar (evk,args) ->
-            (match snd (Evd.evar_source evk sigma) with
+            (match snd (Evd.evar_source (Evd.find sigma evk)) with
               Evar_kinds.MatchingVar (Evar_kinds.SecondOrderPatVar id) -> Some id
             | _ -> None)
           | _ -> None
@@ -183,29 +192,39 @@ let pattern_of_constr env sigma t =
     | Proj (p, c) ->
       pattern_of_constr env (EConstr.Unsafe.to_constr (Retyping.expand_projection env sigma p (EConstr.of_constr c) []))
     | Evar (evk,ctxt as ev) ->
-      (match snd (Evd.evar_source evk sigma) with
+      (match snd (Evd.evar_source (Evd.find sigma evk)) with
       | Evar_kinds.MatchingVar (Evar_kinds.FirstOrderPatVar id) ->
         PMeta (Some id)
       | Evar_kinds.GoalEvar | Evar_kinds.VarInstance _ ->
         (* These are the two evar kinds used for existing goals *)
         (* see Proofview.mark_in_evm *)
          if Evd.is_defined sigma evk then pattern_of_constr env (Evd.existential_value0 sigma ev)
-         else PEvar (evk,List.map (pattern_of_constr env) ctxt)
+         else
+          let ctxt = Evd.expand_existential sigma (EConstr.of_existential ev) in
+          PEvar (evk,List.map (fun c -> pattern_of_constr env (EConstr.Unsafe.to_constr c)) ctxt)
       | Evar_kinds.MatchingVar (Evar_kinds.SecondOrderPatVar ido) -> assert false
       | _ ->
-         PMeta None)
-    | Case (ci,p,a,br) ->
+        PMeta None)
+    | Case (ci, u, pms, p0, iv, a, br0) ->
+        let (ci, p, iv, a, br) = Inductive.expand_case env (ci, u, pms, p0, iv, a, br0) in
+        let pattern_of_ctx c (nas, c0) =
+          let ctx, c = Term.decompose_lam_n_decls (Array.length nas) c in
+          let env = push_rel_context ctx env in
+          let c = pattern_of_constr env c in
+          (Array.map Context.binder_name nas, c)
+        in
+        let p = pattern_of_ctx p p0 in
         let cip =
           { cip_style = ci.ci_pp_info.style;
             cip_ind = Some ci.ci_ind;
-            cip_ind_tags = Some ci.ci_pp_info.ind_tags;
             cip_extensible = false }
         in
         let branch_of_constr i c =
-          (i, ci.ci_pp_info.cstr_tags.(i), pattern_of_constr env c)
+          let nas, c = pattern_of_ctx c br0.(i) in
+          (i, nas, c)
         in
-        PCase (cip, pattern_of_constr env p, pattern_of_constr env a,
-               Array.to_list (Array.mapi branch_of_constr br))
+        PCase (cip, Some p, pattern_of_constr env a,
+                Array.to_list (Array.mapi branch_of_constr br))
     | Fix (lni,(lna,tl,bl)) ->
        let push env na2 c2 = push_rel (LocalAssum (na2,c2)) env in
        let env' = Array.fold_left2 push env lna tl in
@@ -217,9 +236,15 @@ let pattern_of_constr env sigma t =
        PCoFix (ln,(Array.map binder_name lna,Array.map (pattern_of_constr env) tl,
                   Array.map (pattern_of_constr env') bl))
     | Int i -> PInt i
-    | Float f -> PFloat f in
+    | Float f -> PFloat f
+    | Array (_u, t, def, ty) ->
+      PArray (Array.map (pattern_of_constr env) t, pattern_of_constr env def, pattern_of_constr env ty)
+    in
   pattern_of_constr env t
 
+let legacy_bad_pattern_of_constr env sigma c : constr_pattern = pattern_of_constr ~broken:true env sigma c
+let pattern_of_constr env sigma c : constr_pattern = pattern_of_constr ~broken:false env sigma c
+
 (* To process patterns, we need a translation without typing at all. *)
 
 let map_pattern_with_binders g f l = function
@@ -230,7 +255,10 @@ let map_pattern_with_binders g f l = function
   | PLetIn (n,a,t,b) -> PLetIn (n,f l a,Option.map (f l) t,f (g n l) b)
   | PIf (c,b1,b2) -> PIf (f l c,f l b1,f l b2)
   | PCase (ci,po,p,pl) ->
-    PCase (ci,f l po,f l p, List.map (fun (i,n,c) -> (i,n,f l c)) pl)
+    let fold nas l = Array.fold_left (fun l na -> g na l) l nas in
+    let map_branch (i, n, c) = (i, n, f (fold n l) c) in
+    let po = Option.map (fun (nas, po) -> nas, (f (fold nas l) po)) po in
+    PCase (ci,po,f l p, List.map map_branch pl)
   | PProj (p,pc) -> PProj (p, f l pc)
   | PFix (lni,(lna,tl,bl)) ->
      let l' = Array.fold_left (fun l na -> g na l) l lna in
@@ -238,46 +266,11 @@ let map_pattern_with_binders g f l = function
   | PCoFix (ln,(lna,tl,bl)) ->
      let l' = Array.fold_left (fun l na -> g na l) l lna in
      PCoFix (ln,(lna,Array.map (f l) tl,Array.map (f l') bl))
+  | PArray (t,def,ty) -> PArray (Array.map (f l) t, f l def, f l ty)
   (* Non recursive *)
   | (PVar _ | PEvar _ | PRel _ | PRef _  | PSort _  | PMeta _ | PInt _
      | PFloat _ as x) -> x
 
-let error_instantiate_pattern id l =
-  let is = match l with
-  | [_] -> "is"
-  | _ -> "are"
-  in
-  user_err  (str "Cannot substitute the term bound to " ++ Id.print id
-    ++ strbrk " in pattern because the term refers to " ++ pr_enum Id.print l
-    ++ strbrk " which " ++ str is ++ strbrk " not bound in the pattern.")
-
-let instantiate_pattern env sigma lvar c =
-  let open EConstr in
-  let open Vars in
-  let rec aux vars = function
-  | PVar id as x ->
-      (try
-        let ctx,c = Id.Map.find id lvar in
-        try
-          let inst =
-            List.map
-              (fun id -> mkRel (List.index Name.equal (Name id) vars))
-              ctx
-          in
-          let c = substl inst c in
-          (* FIXME: Stupid workaround to pattern_of_constr being evar sensitive *)
-          let c = Evarutil.nf_evar sigma c in
-          pattern_of_constr env sigma (EConstr.Unsafe.to_constr c)
-        with Not_found (* List.index failed *) ->
-          let vars =
-            List.map_filter (function Name id -> Some id | _ -> None) vars in
-          error_instantiate_pattern id (List.subtract Id.equal ctx vars)
-       with Not_found (* Map.find failed *) ->
-         x)
-  | c ->
-      map_pattern_with_binders (fun id vars -> id::vars) aux vars c in
-  aux [] c
-
 let rec liftn_pattern k n = function
   | PRel i as x -> if i >= n then PRel (i+k) else x
   | c -> map_pattern_with_binders (fun _ -> succ) (liftn_pattern k) n c
@@ -291,7 +284,7 @@ let rec subst_pattern env sigma subst pat =
     if ref' == ref then pat else (match t with
         | None -> PRef ref'
         | Some t ->
-          pattern_of_constr env sigma t.Univ.univ_abstracted_value)
+          pattern_of_constr env sigma (EConstr.of_constr t.Univ.univ_abstracted_value))
   | PVar _
   | PEvar _
   | PRel _
@@ -339,7 +332,11 @@ let rec subst_pattern env sigma subst pat =
       let ind = cip.cip_ind in
       let ind' = Option.Smart.map (subst_ind subst) ind in
       let cip' = if ind' == ind then cip else { cip with cip_ind = ind' } in
-      let typ' = subst_pattern env sigma subst typ in
+      let map ((nas, typ) as t) =
+        let typ' = subst_pattern env sigma subst typ in
+        if typ' == typ then t else (nas, typ')
+      in
+      let typ' = Option.Smart.map map typ in
       let c' = subst_pattern env sigma subst c in
       let subst_branch ((i,n,c) as br) =
         let c' = subst_pattern env sigma subst c in
@@ -359,12 +356,16 @@ let rec subst_pattern env sigma subst pat =
       let bl' = Array.Smart.map (subst_pattern env sigma subst) bl in
       if bl' == bl && tl' == tl then pat
       else PCoFix (ln,(lna,tl',bl'))
+  | PArray (t,def,ty) ->
+      let t' = Array.Smart.map (subst_pattern env sigma subst) t in
+      let def' = subst_pattern env sigma subst def in
+      let ty' = subst_pattern env sigma subst ty in
+      if def' == def && t' == t && ty' == ty then pat
+      else PArray (t',def',ty')
 
 let mkPLetIn na b t c = PLetIn(na,b,t,c)
 let mkPProd na t u = PProd(na,t,u)
 let mkPLambda na t b = PLambda(na,t,b)
-let mkPLambdaUntyped na b = PLambda(na,PMeta None,b)
-let rev_it_mkPLambdaUntyped = List.fold_right mkPLambdaUntyped
 
 let mkPProd_or_LetIn (na,_,bo,t) c =
   match bo with
@@ -379,7 +380,7 @@ let mkPLambda_or_LetIn (na,_,bo,t) c =
 let it_mkPProd_or_LetIn = List.fold_left (fun c d -> mkPProd_or_LetIn d c)
 let it_mkPLambda_or_LetIn = List.fold_left (fun c d -> mkPLambda_or_LetIn d c)
 
-let err ?loc pp = user_err ?loc ~hdr:"pattern_of_glob_constr" pp
+let err ?loc pp = user_err ?loc pp
 
 let warn_cast_in_pattern =
   CWarnings.create ~name:"cast-in-pattern" ~category:"automation"
@@ -402,6 +403,13 @@ let rec pat_of_raw metas vars = DAst.with_loc_val (fun ?loc -> function
       PApp (pat_of_raw metas vars c,
             Array.of_list (List.map (pat_of_raw metas vars) cl))
     end
+  | GProj ((p,_), cl, c) ->
+    if Structures.PrimitiveProjections.mem p then
+      let p = Option.get @@ Structures.PrimitiveProjections.find_opt p in
+      let p = Projection.make p false in
+      PProj (p, pat_of_raw metas vars c)
+    else
+      PApp (PRef (GlobRef.ConstRef p), Array.map_of_list (pat_of_raw metas vars) (cl @ [c]))
   | GLambda (na,bk,c1,c2) ->
       Name.iter (fun n -> metas := n::!metas) na;
       PLambda (na, pat_of_raw metas vars c1,
@@ -420,25 +428,21 @@ let rec pat_of_raw metas vars = DAst.with_loc_val (fun ?loc -> function
       with Glob_ops.ComplexSort -> user_err ?loc (str "Unexpected universe in pattern."))
   | GHole _ ->
       PMeta None
-  | GCast (c,_) ->
+  | GCast (c,_,_) ->
       warn_cast_in_pattern ();
       pat_of_raw metas vars c
   | GIf (c,(_,None),b1,b2) ->
       PIf (pat_of_raw metas vars c,
            pat_of_raw metas vars b1,pat_of_raw metas vars b2)
   | GLetTuple (nal,(_,None),b,c) ->
-      let mkGLambda na c = DAst.make ?loc @@
-        GLambda (na,Explicit, DAst.make @@ GHole (Evar_kinds.InternalHole, Namegen.IntroAnonymous, None),c) in
-      let c = List.fold_right mkGLambda nal c in
       let cip =
         { cip_style = LetStyle;
           cip_ind = None;
-          cip_ind_tags = None;
           cip_extensible = false }
       in
-      let tags = List.map (fun _ -> false) nal (* Approximation which can be without let-ins... *) in
-      PCase (cip, PMeta None, pat_of_raw metas vars b,
-             [0,tags,pat_of_raw metas vars c])
+      let tags = Array.of_list nal (* Approximation which can be without let-ins... *) in
+      PCase (cip, None, pat_of_raw metas vars b,
+             [0,tags,pat_of_raw metas (List.rev_append (Array.to_list tags) vars) c])
   | GCases (sty,p,[c,(na,indnames)],brs) ->
       let get_ind p = match DAst.get p with
       | PatCstr((ind,_),_,_) -> Some ind
@@ -457,18 +461,17 @@ let rec pat_of_raw metas vars = DAst.with_loc_val (fun ?loc -> function
       let pred = match p,indnames with
         | Some p, Some {CAst.v=(_,nal)} ->
           let nvars = na :: List.rev nal @ vars in
-          rev_it_mkPLambdaUntyped nal (mkPLambdaUntyped na (pat_of_raw metas nvars p))
-        | None, _ -> PMeta None
+          Some (Array.rev_of_list (na :: List.rev nal), pat_of_raw metas nvars p)
+        | None, _ -> None
         | Some p, None ->
           match DAst.get p with
-          | GHole _ -> PMeta None
+          | GHole _ -> None
           | _ ->
             user_err ?loc  (strbrk "Clause \"in\" expected in patterns over \"match\" expressions with an explicit \"return\" clause.")
       in
       let info =
         { cip_style = sty;
           cip_ind = ind;
-          cip_ind_tags = None;
           cip_extensible = ext }
       in
       (* Nota : when we have a non-trivial predicate,
@@ -502,7 +505,7 @@ let rec pat_of_raw metas vars = DAst.with_loc_val (fun ?loc -> function
 
   | GInt i -> PInt i
   | GFloat f -> PFloat f
-  | GPatVar _ | GIf _ | GLetTuple _ | GCases _ | GEvar _ ->
+  | GPatVar _ | GIf _ | GLetTuple _ | GCases _ | GEvar _ | GArray _ ->
       err ?loc (Pp.str "Non supported pattern."))
 
 and pat_of_glob_in_context metas vars decls c =
@@ -529,7 +532,7 @@ and pats_of_glob_branches loc metas vars ind brs =
         true, [] (* ends with _ => _ *)
       | PatCstr((indsp,j),lv,_), _, _ ->
         let () = match ind with
-        | Some sp when eq_ind sp indsp -> ()
+        | Some sp when Ind.CanOrd.equal sp indsp -> ()
         | _ ->
           err ?loc (Pp.str "All constructors must be in the same inductive type.")
         in
@@ -537,10 +540,10 @@ and pats_of_glob_branches loc metas vars ind brs =
           err ?loc
             (str "No unique branch for " ++ int j ++ str"-th constructor.");
         let lna = List.map get_arg lv in
-        let vars' = List.rev lna @ vars in
-        let pat = rev_it_mkPLambdaUntyped lna (pat_of_raw metas vars' br) in
+        let vars' = List.rev_append lna vars in
+        let tags = Array.of_list lna in
+        let pat = pat_of_raw metas vars' br in
         let ext,pats = get_pat (Int.Set.add (j-1) indexes) brs in
-        let tags = List.map (fun _ -> false) lv (* approximation, w/o let-in *) in
         ext, ((j-1, tags, pat) :: pats)
       | _ ->
         err ?loc:loc' (Pp.str "Non supported pattern.")
