@@ -20,9 +20,11 @@ type unification_error =
   | NotSameHead
   | NoCanonicalStructure
   | ConversionFailed of env * constr * constr (* Non convertible closed terms *)
+  | IncompatibleInstances of env * existential * constr * constr
   | MetaOccurInBody of Evar.t
   | InstanceNotSameType of Evar.t * env * types * types
-  | UnifUnivInconsistency of Univ.univ_inconsistency
+  | InstanceNotFunctionalType of Evar.t * env * constr * types
+  | UnifUnivInconsistency of UGraph.univ_inconsistency
   | CannotSolveConstraint of Evd.evar_constraint * unification_error
   | ProblemBeyondCapabilities
 
@@ -47,14 +49,14 @@ type pretype_error =
   | CannotUnifyBindingType of constr * constr
   | CannotGeneralize of constr
   | NoOccurrenceFound of constr * Id.t option
-  | CannotFindWellTypedAbstraction of constr * constr list * (env * type_error) option
+  | CannotFindWellTypedAbstraction of constr * constr list * (env * pretype_error) option
   | WrongAbstractionType of Name.t * constr * types * types
   | AbstractionOverMeta of Name.t * Name.t
   | NonLinearUnification of Name.t * constr
   (* Pretyping *)
   | VarNotFound of Id.t
   | EvarNotFound of Id.t
-  | UnexpectedType of constr * constr
+  | UnexpectedType of constr * constr * unification_error
   | NotProduct of constr
   | TypingError of type_error
   | CannotUnifyOccurrences of subterm_unification_error
@@ -66,6 +68,7 @@ exception PretypeError of env * Evd.evar_map * pretype_error
 
 let precatchable_exception = function
   | CErrors.UserError _ | TypeError _ | PretypeError _
+  | Reductionops.AnomalyInConversion _
   | Nametab.GlobalizationError _ -> true
   | _ -> false
 
@@ -110,9 +113,9 @@ let error_ill_typed_rec_body ?loc env sigma i na jl tys =
   raise_type_error ?loc
     (env, sigma, IllTypedRecBody (i, na, jl, tys))
 
-let error_elim_arity ?loc env sigma pi c j a =
+let error_elim_arity ?loc env sigma pi c a =
   raise_type_error ?loc
-    (env, sigma, ElimArity (pi, c, j, a))
+    (env, sigma, ElimArity (pi, c, a))
 
 let error_not_a_type ?loc env sigma j =
   raise_type_error ?loc (env, sigma, NotAType j)
@@ -160,8 +163,8 @@ let error_cant_find_case_type ?loc env sigma expr =
 
 (*s Pretyping errors *)
 
-let error_unexpected_type ?loc env sigma actty expty =
-  raise_pretype_error ?loc (env, sigma, UnexpectedType (actty, expty))
+let error_unexpected_type ?loc env sigma actty expty e =
+  raise_pretype_error ?loc (env, sigma, UnexpectedType (actty, expty, e))
 
 let error_not_product ?loc env sigma c =
   raise_pretype_error ?loc (env, sigma, NotProduct c)
@@ -185,7 +188,7 @@ let unsatisfiable_constraints env evd ev comp =
     let err = UnsatisfiableConstraints (None, comp) in
     raise (PretypeError (env,evd,err))
   | Some ev ->
-    let loc, kind = Evd.evar_source ev evd in
+    let loc, kind = Evd.evar_source (Evd.find evd ev) in
     let err = UnsatisfiableConstraints (Some (ev, kind), comp) in
     Loc.raise ?loc (PretypeError (env,evd,err))
 
