@@ -23,9 +23,11 @@ type unification_error =
   | NotSameHead
   | NoCanonicalStructure
   | ConversionFailed of env * constr * constr
+  | IncompatibleInstances of env * existential * constr * constr
   | MetaOccurInBody of Evar.t
   | InstanceNotSameType of Evar.t * env * types * types
-  | UnifUnivInconsistency of Univ.univ_inconsistency
+  | InstanceNotFunctionalType of Evar.t * env * constr * types
+  | UnifUnivInconsistency of UGraph.univ_inconsistency
   | CannotSolveConstraint of Evd.evar_constraint * unification_error
   | ProblemBeyondCapabilities
 
@@ -53,14 +55,14 @@ type pretype_error =
   | CannotUnifyBindingType of constr * constr
   | CannotGeneralize of constr
   | NoOccurrenceFound of constr * Id.t option
-  | CannotFindWellTypedAbstraction of constr * constr list * (env * type_error) option
+  | CannotFindWellTypedAbstraction of constr * constr list * (env * pretype_error) option
   | WrongAbstractionType of Name.t * constr * types * types
   | AbstractionOverMeta of Name.t * Name.t
   | NonLinearUnification of Name.t * constr
   (** Pretyping *)
   | VarNotFound of Id.t
   | EvarNotFound of Id.t
-  | UnexpectedType of constr * constr
+  | UnexpectedType of constr * constr * unification_error
   | NotProduct of constr
   | TypingError of type_error
   | CannotUnifyOccurrences of subterm_unification_error
@@ -107,7 +109,7 @@ val error_ill_typed_rec_body :
 val error_elim_arity :
   ?loc:Loc.t -> env -> Evd.evar_map ->
       pinductive -> constr ->
-      unsafe_judgment -> (Sorts.family * Sorts.family * Sorts.family * arity_error) option -> 'b
+      (unsafe_judgment * Sorts.family * Sorts.family * Sorts.family) option -> 'b
 
 val error_not_a_type :
   ?loc:Loc.t -> env -> Evd.evar_map -> unsafe_judgment -> 'b
@@ -131,7 +133,7 @@ val error_cannot_unify : ?loc:Loc.t -> env -> Evd.evar_map ->
 val error_cannot_unify_local : env -> Evd.evar_map -> constr * constr * constr -> 'b
 
 val error_cannot_find_well_typed_abstraction : env -> Evd.evar_map ->
-      constr -> constr list -> (env * type_error) option -> 'b
+      constr -> constr list -> (env * pretype_error) option -> 'b
 
 val error_wrong_abstraction_type :  env -> Evd.evar_map ->
       Name.t -> constr -> types -> types -> 'b
@@ -150,7 +152,7 @@ val error_cant_find_case_type :
 (** {6 Pretyping errors } *)
 
 val error_unexpected_type :
-  ?loc:Loc.t -> env -> Evd.evar_map -> constr -> constr -> 'b
+  ?loc:Loc.t -> env -> Evd.evar_map -> constr -> constr -> unification_error -> 'b
 
 val error_not_product :
   ?loc:Loc.t -> env -> Evd.evar_map -> constr -> 'b
