@@ -15,9 +15,23 @@ open EConstr
 open Reductionops
 open Pattern
 open Locus
-open Univ
 open Ltac_pretype
 
+(* XXX: Move to a module *)
+type evaluable_global_reference =
+  | EvalVarRef of Id.t
+  | EvalConstRef of Constant.t
+
+val eq_egr : evaluable_global_reference ->  evaluable_global_reference -> bool
+
+(** Here the semantics is completely unclear.
+   What does "Hint Unfold t" means when "t" is a parameter?
+   Does the user mean "Unfold X.t" or does she mean "Unfold y"
+   where X.t is later on instantiated with y? I choose the first
+   interpretation (i.e. an evaluable reference is never expanded). *)
+val subst_evaluable_reference :
+  Mod_subst.substitution -> evaluable_global_reference -> evaluable_global_reference
+
 type reduction_tactic_error =
     InvalidAbstraction of env * evar_map * constr * (env * Type_errors.type_error)
 
@@ -55,6 +69,9 @@ val whd_simpl : reduction_function
    not immediately hide a non reducible fix or cofix *)
 val hnf_constr : reduction_function
 
+(** Variant of the above that does not perform nf-βι *)
+val hnf_constr0 : reduction_function
+
 (** Unfold *)
 val unfoldn :
   (occurrences * evaluable_global_reference) list ->  reduction_function
@@ -84,13 +101,19 @@ val reduce_to_atomic_ind : env ->  evar_map -> types -> (inductive * EInstance.t
    returns [I] and [t'] or fails with a user error *)
 val reduce_to_quantified_ind : env ->  evar_map -> types -> (inductive * EInstance.t) * types
 
+(** Same as {!reduce_to_quantified_ind} but more efficient because it does not
+    return the normalized type. *)
+val eval_to_quantified_ind : env -> evar_map -> types -> (inductive * EInstance.t)
+
 (** [reduce_to_quantified_ref env sigma ref t] try to put [t] in the form
-   [t'=(x1:A1)..(xn:An)(ref args)] and fails with user error if not possible *)
+    [t'=(x1:A1)..(xn:An)(ref args)]. When this is not possible, if [allow_failure]
+    is specified, [t] is unfolded until the point where this impossibility is plainly
+    visible. Otherwise, it fails with user error. *)
 val reduce_to_quantified_ref :
-  env ->  evar_map -> GlobRef.t -> types -> types
+  ?allow_failure:bool -> env ->  evar_map -> GlobRef.t -> types -> types
 
 val reduce_to_atomic_ref :
-  env ->  evar_map -> GlobRef.t -> types -> types
+  ?allow_failure:bool -> env ->  evar_map -> GlobRef.t -> types -> types
 
 val find_hnf_rectype :
   env ->  evar_map -> types -> (inductive * EInstance.t) * constr list
@@ -101,10 +124,5 @@ val contextually : bool -> occurrences * constr_pattern ->
 val e_contextually : bool -> occurrences * constr_pattern ->
   (patvar_map -> e_reduction_function) -> e_reduction_function
 
-(** Returns the same inductive if it is allowed for pattern-matching
-    raises an error otherwise. **)
-val check_privacy : env -> inductive puniverses -> inductive puniverses
-
-(** Returns the same inductive if it is not a primitive record
-    raises an error otherwise. **)
-val check_not_primitive_record : env -> inductive puniverses -> inductive puniverses
+(** Errors if the inductive is not allowed for pattern-matching. **)
+val check_privacy : env -> inductive -> unit
