@@ -11,7 +11,6 @@
 (*i*)
 open Names
 open Globnames
-open Term
 open Constr
 open Vars
 open Evd
@@ -34,33 +33,35 @@ let get_typeclasses_unique_solutions =
     ~key:["Typeclasses";"Unique";"Solutions"]
     ~value:false
 
-let (classes_transparent_state, classes_transparent_state_hook) = Hook.make ()
-let classes_transparent_state () = Hook.get classes_transparent_state ()
-
-let get_solve_one_instance, solve_one_instance_hook = Hook.make ()
+let solve_one_instance = ref (fun env evm t unique -> assert false)
 
 let resolve_one_typeclass ?(unique=get_typeclasses_unique_solutions ()) env evm t =
-  Hook.get get_solve_one_instance env evm t unique
+  !solve_one_instance env evm t unique
 
-type direction = Forward | Backward
+let set_solve_one_instance f = solve_one_instance := f
+
+type class_method = {
+  meth_name : Name.t;
+  meth_info : hint_info option;
+  meth_const : Constant.t option;
+}
 
 (* This module defines type-classes *)
 type typeclass = {
   (* Universe quantification *)
-  cl_univs : Univ.AUContext.t;
+  cl_univs : Univ.AbstractContext.t;
 
   (* The class implementation *)
   cl_impl : GlobRef.t;
 
   (* Context in which the definitions are typed. Includes both typeclass parameters and superclasses. *)
-  cl_context : GlobRef.t option list * Constr.rel_context;
+  cl_context : Constr.rel_context;
 
   (* Context of definitions and properties on defs, will not be shared *)
   cl_props : Constr.rel_context;
 
   (* The method implementations as projections. *)
-  cl_projs : (Name.t * (direction * hint_info) option
-              * Constant.t option) list;
+  cl_projs : class_method list;
 
   cl_strict : bool;
 
@@ -68,13 +69,11 @@ type typeclass = {
 }
 
 type typeclasses = typeclass GlobRef.Map.t
+(* Invariant: for any pair (gr, tc) in the map, gr and tc.cl_impl are equal *)
 
 type instance = {
   is_class: GlobRef.t;
   is_info: hint_info;
-  (* Sections where the instance should be redeclared,
-     None for discard, Some 0 for none. *)
-  is_global: int option;
   is_impl: GlobRef.t;
 }
 
@@ -92,9 +91,9 @@ let classes : typeclasses ref = Summary.ref GlobRef.Map.empty ~name:"classes"
 let instances : instances ref = Summary.ref GlobRef.Map.empty ~name:"instances"
 
 let typeclass_univ_instance (cl, u) =
-  assert (Univ.AUContext.size cl.cl_univs == Univ.Instance.length u);
+  assert (Univ.AbstractContext.size cl.cl_univs == Univ.Instance.length u);
   let subst_ctx c = Context.Rel.map (subst_instance_constr u) c in
-    { cl with cl_context = on_snd subst_ctx cl.cl_context;
+    { cl with cl_context = subst_ctx cl.cl_context;
       cl_props = subst_ctx cl.cl_props}
 
 let class_info env sigma c =
@@ -107,14 +106,22 @@ let global_class_of_constr env sigma c =
     GlobRef.Map.find gr !classes, u
   with DestKO | Not_found -> not_a_class env sigma c
 
+let decompose_class_app env sigma c =
+  let hd, args = EConstr.decompose_app sigma c in
+  match EConstr.kind sigma hd with
+  | Proj (p, c) ->
+    let expp = Retyping.expand_projection env sigma p c args in
+    EConstr.decompose_app sigma expp
+  | _ -> hd, args
+
 let dest_class_app env sigma c =
-  let cl, args = EConstr.decompose_app sigma c in
+  let cl, args = decompose_class_app env sigma c in
     global_class_of_constr env sigma cl, (List.map EConstr.Unsafe.to_constr args)
 
 let dest_class_arity env sigma c =
   let open EConstr in
   let rels, c = decompose_prod_assum sigma c in
-    rels, dest_class_app env sigma c
+    rels, dest_class_app (push_rel_context rels env) sigma c
 
 let class_of_constr env sigma c =
   try Some (dest_class_arity env sigma c)
@@ -130,15 +137,11 @@ let rec is_class_type evd c =
     match EConstr.kind evd c with
     | Prod (_, _, t) -> is_class_type evd t
     | Cast (t, _, _) -> is_class_type evd t
+    | Proj (p, c) -> GlobRef.(Map.mem (ConstRef (Projection.constant p))) !classes
     | _ -> is_class_constr evd c
 
 let is_class_evar evd evi =
-  is_class_type evd evi.Evd.evar_concl
-
-let is_class_constr sigma c =
-  try let gr, u = EConstr.destRef sigma c in
-    GlobRef.Map.mem gr !classes
-  with DestKO | Not_found -> false
+  is_class_type evd (Evd.evar_concl evi)
 
 let rec is_maybe_class_type evd c =
   let c, _ = Termops.decompose_app_vect evd c in
@@ -146,6 +149,7 @@ let rec is_maybe_class_type evd c =
     | Prod (_, _, t) -> is_maybe_class_type evd t
     | Cast (t, _, _) -> is_maybe_class_type evd t
     | Evar _ -> true
+    | Proj (p, c) -> GlobRef.(Map.mem (ConstRef (Projection.constant p))) !classes
     | _ -> is_class_constr evd c
 
 let () = Hook.set Evd.is_maybe_typeclass_hook (fun evd c -> is_maybe_class_type evd (EConstr.of_constr c))
@@ -155,66 +159,6 @@ let load_class cl =
 
 (** Build the subinstances hints. *)
 
-let check_instance env sigma c =
-  try
-    let (evd, c) = resolve_one_typeclass env sigma
-      (Retyping.get_type_of env sigma c) in
-      not (Evd.has_undefined evd)
-  with e when CErrors.noncritical e -> false
-
-let build_subclasses ~check env sigma glob { hint_priority = pri } =
-  let _id = Nametab.basename_of_global glob in
-  let _next_id =
-    let i = ref (-1) in
-      (fun () -> incr i;
-        Nameops.add_suffix _id ("_subinstance_" ^ string_of_int !i))
-  in
-  let ty, ctx = Typeops.type_of_global_in_context env glob in
-  let inst, ctx = UnivGen.fresh_instance_from ctx None in
-  let ty = Vars.subst_instance_constr inst ty in
-  let ty = EConstr.of_constr ty in
-  let sigma = Evd.merge_context_set Evd.univ_rigid sigma ctx in
-  let rec aux pri c ty path =
-      match class_of_constr env sigma ty with
-      | None -> []
-      | Some (rels, ((tc,u), args)) ->
-        let instapp =
-          Reductionops.whd_beta env sigma (EConstr.of_constr (appvectc c (Context.Rel.to_extended_vect mkRel 0 rels)))
-        in
-        let instapp = EConstr.Unsafe.to_constr instapp in
-        let projargs = Array.of_list (args @ [instapp]) in
-        let projs = List.map_filter
-          (fun (n, b, proj) ->
-           match b with
-           | None -> None
-           | Some (Backward, _) -> None
-           | Some (Forward, info) ->
-             let proj = Option.get proj in
-             let rels = List.map (fun d -> Termops.map_rel_decl EConstr.Unsafe.to_constr d) rels in
-             let u = EConstr.EInstance.kind sigma u in
-             let body = it_mkLambda_or_LetIn (mkApp (mkConstU (proj,u), projargs)) rels in
-               if check && check_instance env sigma (EConstr.of_constr body) then None
-               else
-                 let newpri =
-                   match pri, info.hint_priority with
-                   | Some p, Some p' -> Some (p + p')
-                   | Some p, None -> Some (p + 1)
-                   | _, _ -> None
-                 in
-                   Some (GlobRef.ConstRef proj, { info with hint_priority = newpri }, body)) tc.cl_projs
-        in
-        let declare_proj hints (cref, info, body) =
-          let path' = cref :: path in
-          let ty = Retyping.get_type_of env sigma (EConstr.of_constr body) in
-          let rest = aux pri body ty path' in
-            hints @ (path', info, body) :: rest
-        in List.fold_left declare_proj [] projs
-  in
-  let term = Constr.mkRef (glob, inst) in
-    (*FIXME subclasses should now get substituted for each particular instance of
-      the polymorphic superclass *)
-    aux pri term ty [glob]
-
 (*
  * interface functions
  *)
@@ -235,7 +179,7 @@ let remove_instance inst =
 
 
 let instance_constructor (cl,u) args =
-  let lenpars = List.count is_local_assum (snd cl.cl_context) in
+  let lenpars = List.count is_local_assum cl.cl_context in
   let open EConstr in
   let pars = fst (List.chop lenpars args) in
     match cl.cl_impl with
@@ -268,7 +212,7 @@ let instances env sigma r =
   let cl = class_info env sigma r in instances_of cl
 
 let is_class gr =
-  GlobRef.Map.exists (fun _ v -> GlobRef.equal v.cl_impl gr) !classes
+  GlobRef.Map.mem gr !classes
 
 open Evar_kinds
 type evar_filter = Evar.t -> Evar_kinds.t Lazy.t -> bool
@@ -291,19 +235,40 @@ let no_goals_or_obligations _ source =
 
 let has_typeclasses filter evd =
   let tcs = get_typeclass_evars evd in
-  let check ev = filter ev (lazy (snd (Evd.find evd ev).evar_source)) in
+  let check ev = filter ev (lazy (snd (Evd.evar_source (Evd.find evd ev)))) in
   Evar.Set.exists check tcs
 
-let get_solve_all_instances, solve_all_instances_hook = Hook.make ()
+let get_filtered_typeclass_evars filter evd =
+  let tcs = get_typeclass_evars evd in
+  let check ev = filter ev (lazy (snd (Evd.evar_source (Evd.find evd ev)))) in
+  Evar.Set.filter check tcs
+
+let solve_all_instances_hook = ref (fun env evd filter unique split fail -> assert false)
 
 let solve_all_instances env evd filter unique split fail =
-  Hook.get get_solve_all_instances env evd filter unique split fail
+  !solve_all_instances_hook env evd filter unique split fail
 
-(** Profiling resolution of typeclasses *)
-(* let solve_classeskey = CProfile.declare_profile "solve_typeclasses" *)
-(* let solve_problem = CProfile.profile5 solve_classeskey solve_problem *)
+let set_solve_all_instances f = solve_all_instances_hook := f
 
 let resolve_typeclasses ?(filter=no_goals) ?(unique=get_typeclasses_unique_solutions ())
     ?(split=true) ?(fail=true) env evd =
   if not (has_typeclasses filter evd) then evd
   else solve_all_instances env evd filter unique split fail
+
+(** In case of unsatisfiable constraints, build a nice error message *)
+
+let error_unresolvable env evd comp =
+  let is_part ev = match comp with
+  | None -> true
+  | Some s -> Evar.Set.mem ev s
+  in
+  let fold ev evi (found, accu) =
+    let ev_class = class_of_constr env evd (Evd.evar_concl evi) in
+    if not (Option.is_empty ev_class) && is_part ev then
+      (* focus on one instance if only one was searched for *)
+      if not found then (true, Some ev)
+      else (found, None)
+    else (found, accu)
+   in
+  let (_, ev) = Evd.fold_undefined fold evd (true, None) in
+  Pretype_errors.unsatisfiable_constraints env evd ev comp
