@@ -33,14 +33,14 @@ let meta_type env evd mv =
   let ty =
     try Evd.meta_ftype evd mv
     with Not_found -> anomaly (str "unknown meta ?" ++ str (Nameops.string_of_meta mv) ++ str ".") in
-  meta_instance env evd ty
+  meta_instance env (create_meta_instance_subst evd) ty
 
 let inductive_type_knowing_parameters env sigma (ind,u) jl =
   let u = Unsafe.to_instance u in
   let mspec = lookup_mind_specif env ind in
   let paramstyp = Array.map_to_list (fun j () ->
       let s = Reductionops.sort_of_arity env sigma j.uj_type in
-      Sorts.univ_of_sort (EConstr.ESorts.kind sigma s)) jl
+      EConstr.ESorts.kind sigma s) jl
   in
   Inductive.type_of_inductive_knowing_parameters (mspec,u) paramstyp
 
@@ -59,57 +59,54 @@ let assumption_of_judgment env sigma j =
   with Type_errors.TypeError _ | PretypeError _ ->
     error_assumption env sigma j
 
-let judge_of_applied_inductive_knowing_parameters env sigma funj ind argjv =
-  let rec apply_rec sigma n typ = function
-  | [] ->
-      sigma, { uj_val  = mkApp (j_val funj, Array.map j_val argjv);
-               uj_type =
-                 let ar = inductive_type_knowing_parameters env sigma ind argjv in
-                 hnf_prod_appvect env sigma (EConstr.of_constr ar) (Array.map j_val argjv) }
-  | hj::restjl ->
-       let sigma, (c1,c2) =
-         match EConstr.kind sigma (whd_all env sigma typ) with
-         | Prod (_,c1,c2) -> sigma, (c1,c2)
-         | Evar ev ->
-             let (sigma,t) = Evardefine.define_evar_as_product env sigma ev in
-             let (_,c1,c2) = destProd sigma t in
-             sigma, (c1,c2)
-         | _ ->
-             error_cant_apply_not_functional env sigma funj argjv
-       in
-       begin match Evarconv.unify_leq_delay env sigma hj.uj_type c1 with
-         | sigma ->
-           apply_rec sigma (n+1) (subst1 hj.uj_val c2) restjl
-         | exception Evarconv.UnableToUnify _ ->
-           error_cant_apply_bad_type env sigma (n, c1, hj.uj_type) funj argjv
-       end
-  in
-  apply_rec sigma 1 funj.uj_type (Array.to_list argjv)
-
 let judge_of_apply env sigma funj argjv =
-  let rec apply_rec sigma n typ = function
+  let rec apply_rec sigma n subs typ = function
   | [] ->
+    let typ = Vars.esubst Vars.lift_substituend subs typ in
     sigma, { uj_val  = mkApp (j_val funj, Array.map j_val argjv);
              uj_type = typ }
   | hj::restjl ->
-       let sigma, (c1,c2) =
-         match EConstr.kind sigma (whd_all env sigma typ) with
-         | Prod (_,c1,c2) -> sigma, (c1,c2)
-         | Evar ev ->
-             let (sigma,t) = Evardefine.define_evar_as_product env sigma ev in
-             let (_,c1,c2) = destProd sigma t in
-             sigma, (c1,c2)
-         | _ ->
-            error_cant_apply_not_functional env sigma funj argjv
-       in
-       begin match Evarconv.unify_leq_delay env sigma hj.uj_type c1 with
-         | sigma ->
-           apply_rec sigma (n+1) (subst1 hj.uj_val c2) restjl
-         | exception Evarconv.UnableToUnify _ ->
-           error_cant_apply_bad_type env sigma (n, c1, hj.uj_type) funj argjv
-       end
+    let sigma, c1, subs, c2 = match EConstr.kind sigma typ with
+    | Prod (_, c1, c2) ->
+      (* Fast path *)
+      let c1 = Vars.esubst Vars.lift_substituend subs c1 in
+      let subs = Esubst.subs_cons (Vars.make_substituend hj.uj_val) subs in
+      sigma, c1, subs, c2
+    | _ ->
+      let typ = Vars.esubst Vars.lift_substituend subs typ in
+      let subs = Esubst.subs_cons (Vars.make_substituend hj.uj_val) (Esubst.subs_id 0) in
+      match EConstr.kind sigma (whd_all env sigma typ) with
+      | Prod (_, c1, c2) -> sigma, c1, subs, c2
+      | Evar ev ->
+        let (sigma,t) = Evardefine.define_evar_as_product env sigma ev in
+        let (_, c1, c2) = destProd sigma t in
+        sigma, c1, subs, c2
+      | _ ->
+        error_cant_apply_not_functional env sigma funj argjv
+    in
+    match Evarconv.unify_leq_delay env sigma hj.uj_type c1 with
+    | sigma ->
+      apply_rec sigma (n+1) subs c2 restjl
+    | exception Evarconv.UnableToUnify _ ->
+      error_cant_apply_bad_type env sigma (n, c1, hj.uj_type) funj argjv
   in
-  apply_rec sigma 1 funj.uj_type (Array.to_list argjv)
+  apply_rec sigma 1 (Esubst.subs_id 0) funj.uj_type (Array.to_list argjv)
+
+let checked_appvect env sigma f args =
+  let mk c = Retyping.get_judgment_of env sigma c in
+  let sigma, j = judge_of_apply env sigma (mk f) (Array.map mk args) in
+  sigma, j.uj_val
+
+let checked_applist env sigma f args = checked_appvect env sigma f (Array.of_list args)
+
+let judge_of_applied_inductive_knowing_parameters_nocheck env sigma funj ind argjv =
+  let ar = inductive_type_knowing_parameters env sigma ind argjv in
+  let typ = hnf_prod_appvect env sigma (EConstr.of_constr ar) (Array.map j_val argjv) in
+  sigma, { uj_val = (mkApp (j_val funj, Array.map j_val argjv)); uj_type = typ }
+
+let judge_of_applied_inductive_knowing_parameters env sigma funj ind argjv =
+  let (sigma, j) = judge_of_apply env sigma funj argjv in
+  judge_of_applied_inductive_knowing_parameters_nocheck env sigma funj ind argjv
 
 let check_branch_types env sigma (ind,u) cj (lfj,explft) =
   if not (Int.equal (Array.length lfj) (Array.length explft)) then
@@ -128,7 +125,7 @@ let max_sort l =
 let is_correct_arity env sigma c pj ind specif params =
   let arsign = make_arity_signature env sigma true (make_ind_family (ind,params)) in
   let allowed_sorts = sorts_below (elim_sort specif) in
-  let error () = Pretype_errors.error_elim_arity env sigma ind c pj None in
+  let error () = Pretype_errors.error_elim_arity env sigma ind c None in
   let rec srec env sigma pt ar =
     let pt' = whd_all env sigma pt in
     match EConstr.kind sigma pt', ar with
@@ -178,7 +175,7 @@ let type_case_branches env sigma (ind,largs) pj c =
   let ty = whd_betaiota env sigma (lambda_applist_assum sigma (n+1) p (realargs@[c])) in
   sigma, (lc, ty, Sorts.relevance_of_sort ps)
 
-let judge_of_case env sigma ci pj cj lfj =
+let judge_of_case env sigma case ci pj iv cj lfj =
   let ((ind, u), spec) =
     try find_mrectype env sigma cj.uj_type
     with Not_found -> error_case_not_inductive env sigma cj in
@@ -186,7 +183,10 @@ let judge_of_case env sigma ci pj cj lfj =
   let sigma, (bty,rslty,rci) = type_case_branches env sigma indspec pj cj.uj_val in
   let () = check_case_info env (fst indspec) rci ci in
   let sigma = check_branch_types env sigma (fst indspec) cj (lfj,bty) in
-  sigma, { uj_val  = mkCase (ci, pj.uj_val, cj.uj_val, Array.map j_val lfj);
+  let () = if (match iv with | NoInvert -> false | CaseInvert _ -> true) != should_invert_case env ci
+    then Type_errors.error_bad_invert env
+  in
+  sigma, { uj_val  = mkCase case;
            uj_type = rslty }
 
 let check_type_fixpoint ?loc env sigma lna lar vdefj =
@@ -209,22 +209,22 @@ let check_allowed_sort env sigma ind c p =
   let _, s = splay_prod env sigma pj.uj_type in
   let ksort = match EConstr.kind sigma s with
   | Sort s -> Sorts.family (ESorts.kind sigma s)
-  | _ -> error_elim_arity env sigma ind c pj None in
+  | _ -> error_elim_arity env sigma ind c None in
   if not (Sorts.family_leq ksort sorts) then
     let s = inductive_sort_family (snd specif) in
-    error_elim_arity env sigma ind c pj
-      (Some(sorts,ksort,s,Type_errors.error_elim_explain ksort s))
+    error_elim_arity env sigma ind c (Some (pj, sorts, ksort, s))
   else
     Sorts.relevance_of_sort_family ksort
 
+let check_actual_type env sigma cj t =
+  try Evarconv.unify_leq_delay env sigma cj.uj_type t
+  with Evarconv.UnableToUnify (sigma,e) -> error_actual_type env sigma cj t e
+
 let judge_of_cast env sigma cj k tj =
   let expected_type = tj.utj_val in
-  match Evarconv.unify_leq_delay env sigma cj.uj_type expected_type with
-  | exception Evarconv.UnableToUnify _ ->
-    error_actual_type_core env sigma cj expected_type;
-  | sigma ->
-    sigma, { uj_val = mkCast (cj.uj_val, k, expected_type);
-             uj_type = expected_type }
+  let sigma = check_actual_type env sigma cj expected_type in
+  sigma, { uj_val = mkCast (cj.uj_val, k, expected_type);
+           uj_type = expected_type }
 
 let check_fix env sigma pfix =
   let inj c = EConstr.to_constr ~abort_on_undefined_evars:false sigma c in
@@ -293,7 +293,8 @@ let judge_of_letin env name defj typj j =
     uj_type = subst1 defj.uj_val j.uj_type }
 
 let check_hyps_inclusion env sigma x hyps =
-  let evars = Evarutil.safe_evar_value sigma, Evd.universes sigma in
+  let env = Environ.set_universes (Evd.universes sigma) env in
+  let evars = Evd.evar_handler sigma in
   Typeops.check_hyps_inclusion env ~evars x hyps
 
 let type_of_constant env sigma (c,u) =
@@ -329,6 +330,23 @@ let judge_of_int env v =
 let judge_of_float env v =
   Environ.on_judgment EConstr.of_constr (judge_of_float env v)
 
+let judge_of_array env sigma u tj defj tyj =
+  let ulev = match Univ.Instance.to_array u with
+    | [|u|] -> u
+    | _ -> assert false
+  in
+  let sigma = Evd.set_leq_sort env sigma tyj.utj_type
+      (Sorts.sort_of_univ (Univ.Universe.make ulev))
+  in
+  let check_one sigma j = check_actual_type env sigma j tyj.utj_val in
+  let sigma = check_one sigma defj in
+  let sigma = Array.fold_left check_one sigma tj in
+  let arr = EConstr.of_constr @@ type_of_array env u in
+  let j = make_judge (mkArray(EInstance.make u, Array.map j_val tj, defj.uj_val, tyj.utj_val))
+      (mkApp (arr, [|tyj.utj_val|]))
+  in
+  sigma, j
+
 (* cstr must be in n.f. w.r.t. evars and execute returns a judgement
    where both the term and type are in n.f. *)
 let rec execute env sigma cstr =
@@ -361,11 +379,23 @@ let rec execute env sigma cstr =
         let sigma, ty = type_of_constructor env sigma ctor in
         sigma, make_judge cstr ty
 
-    | Case (ci,p,c,lf) ->
+    | Case (ci, u, pms, p, iv, c, lf) ->
+        let case = (ci, u, pms, p, iv, c, lf) in
+        let (ci, p, iv, c, lf) = EConstr.expand_case env sigma case in
         let sigma, cj = execute env sigma c in
         let sigma, pj = execute env sigma p in
         let sigma, lfj = execute_array env sigma lf in
-        judge_of_case env sigma ci pj cj lfj
+        let sigma = match iv with
+          | NoInvert -> sigma
+          | CaseInvert {indices} ->
+            let args = Array.append pms indices in
+            let t = mkApp (mkIndU (ci.ci_ind,u), args) in
+            let sigma, tj = execute env sigma t in
+            let sigma, tj = type_judgment env sigma tj in
+            let sigma = check_actual_type env sigma cj tj.utj_val in
+            sigma
+        in
+        judge_of_case env sigma case ci pj iv cj lfj
 
     | Fix ((vn,i as vni),recdef) ->
         let sigma, (_,tys,_ as recdef') = execute_recdef env sigma recdef in
@@ -443,6 +473,13 @@ let rec execute env sigma cstr =
     | Float f ->
         sigma, judge_of_float env f
 
+    | Array(u,t,def,ty) ->
+      let sigma, tyj = execute env sigma ty in
+      let sigma, tyj = type_judgment env sigma tyj in
+      let sigma, defj = execute env sigma def in
+      let sigma, tj = execute_array env sigma t in
+      judge_of_array env sigma (EInstance.kind sigma u) tj defj tyj
+
 and execute_recdef env sigma (names,lar,vdef) =
   let sigma, larj = execute_array env sigma lar in
   let sigma, lara = Array.fold_left_map (assumption_of_judgment env) sigma larj in
@@ -456,16 +493,7 @@ and execute_array env = Array.fold_left_map (execute env)
 
 let check env sigma c t =
   let sigma, j = execute env sigma c in
-  match Evarconv.unify_leq_delay env sigma j.uj_type t with
-  | exception Evarconv.UnableToUnify _ ->
-    error_actual_type_core env sigma j t
-  | sigma -> sigma
-
-(* Type of a constr *)
-
-let unsafe_type_of env sigma c =
-  let sigma, j = execute env sigma c in
-  j.uj_type
+  check_actual_type env sigma j t
 
 (* Sort of a type *)
 
@@ -489,3 +517,214 @@ let solve_evars env sigma c =
   sigma, nf_evar sigma j.uj_val
 
 let _ = Evarconv.set_solve_evars (fun env sigma c -> solve_evars env sigma c)
+
+type change_kind = Same | Changed of {bodyonly : bool Lazy.t }
+
+let merge_changes a b = match a,b with
+  | Same, v | v, Same -> v
+  | Changed {bodyonly=a}, Changed {bodyonly=b} ->
+    Changed {bodyonly = lazy (Lazy.force a && Lazy.force b)}
+
+let unchanged = function
+  | Same -> true
+  | Changed _ -> false
+
+let bodyonly = function
+  | Same -> true
+  | Changed {bodyonly} -> Lazy.force bodyonly
+
+let judge_of_apply_against env sigma changedf funj argjv =
+  let rec apply_rec sigma changedf n subs typ = function
+  | [] ->
+    let typ = Vars.esubst Vars.lift_substituend subs typ in
+    sigma,
+    { uj_val  = mkApp (j_val funj, Array.map (fun (_,j) -> j_val j) argjv);
+      uj_type = typ }
+  | (changedh,hj)::restjl ->
+    let sigma, c1, subs, c2 = match EConstr.kind sigma typ with
+    | Prod (_, c1, c2) ->
+      (* Fast path *)
+      let c1 = Vars.esubst Vars.lift_substituend subs c1 in
+      let subs = Esubst.subs_cons (Vars.make_substituend hj.uj_val) subs in
+      sigma, c1, subs, c2
+    | _ ->
+      let typ = Vars.esubst Vars.lift_substituend subs typ in
+      let subs = Esubst.subs_cons (Vars.make_substituend hj.uj_val) (Esubst.subs_id 0) in
+      match EConstr.kind sigma (whd_all env sigma typ) with
+      | Prod (_, c1, c2) -> sigma, c1, subs, c2
+      | Evar ev ->
+        let (sigma,t) = Evardefine.define_evar_as_product env sigma ev in
+        let (_, c1, c2) = destProd sigma t in
+        sigma, c1, subs, c2
+      | _ ->
+        error_cant_apply_not_functional env sigma funj (Array.map snd argjv)
+    in
+    if bodyonly changedf then begin match changedh with
+      | Same -> apply_rec sigma changedf (n+1) subs c2 restjl
+      | Changed {bodyonly=lazy true} ->
+        (* TODO if non dependent product then changedf else bodyonly = false *)
+        apply_rec sigma (Changed {bodyonly=lazy false}) (n+1) subs c2 restjl
+      | Changed {bodyonly=lazy false} ->
+        match Evarconv.unify_leq_delay env sigma hj.uj_type c1 with
+        | sigma ->
+          apply_rec sigma (Changed {bodyonly=lazy false}) (n+1) subs c2 restjl
+        | exception Evarconv.UnableToUnify _ ->
+          error_cant_apply_bad_type env sigma (n, c1, hj.uj_type) funj (Array.map snd argjv)
+    end
+    else
+      match Evarconv.unify_leq_delay env sigma hj.uj_type c1 with
+      | sigma ->
+        apply_rec sigma changedf (n+1) subs c2 restjl
+      | exception Evarconv.UnableToUnify _ ->
+        error_cant_apply_bad_type env sigma (n, c1, hj.uj_type) funj (Array.map snd argjv)
+  in
+  apply_rec sigma changedf 1 (Esubst.subs_id 0) funj.uj_type (Array.to_list argjv)
+
+(* Assuming "env |- good : some type", infer "env |- c : other type" *)
+let rec recheck_against env sigma good c =
+  if EConstr.eq_constr sigma good c then sigma, Same, make_judge c (Retyping.get_type_of env sigma c)
+  else
+    let assume_unchanged_type sigma =
+      let gt = Retyping.get_type_of env sigma good in
+      sigma, Changed {bodyonly=Lazy.from_val true}, make_judge c gt
+    in
+    let maybe_changed (sigma, j) =
+      let bodyonly = lazy (EConstr.eq_constr sigma (Retyping.get_type_of env sigma good) j.uj_type) in
+      let change = Changed {bodyonly} in
+      sigma, change, j
+    in
+    let default () = maybe_changed (execute env sigma c) in
+    match kind sigma good, kind sigma c with
+    (* No subterms *)
+    | _, (Meta _ | Rel _ | Var _ | Const _ | Ind _ | Construct _
+         | Sort _ | Int _ | Float _) ->
+      default ()
+
+    (* Evar (todo deal with Evar differently??? execute recurses on its type)
+       others: too annoying for now *)
+    | _, (Evar _ | Fix _ | CoFix _ | LetIn _ | Array _) ->
+     default ()
+
+    | Case (gci, gu, gpms, gp, giv, gc, glf),
+      Case (ci, u, pms, p, iv, c, lf) ->
+      let (gci, gp, giv, gc, glf) = EConstr.expand_case env sigma (gci, gu, gpms, gp, giv, gc, glf) in
+      let case = (ci, u, pms, p, iv, c, lf) in
+      let (ci, p, iv, c, lf) = EConstr.expand_case env sigma case in
+      let sigma, changedc, cj = recheck_against env sigma gc c in
+      let sigma, changedp, pj = recheck_against env sigma gp p in
+      let (sigma, changedlf), lfj =
+        if Array.length glf <> Array.length lf then
+          Array.fold_left_map (fun (sigma,changed) c ->
+              let sigma, j = execute env sigma c in
+              (sigma, changed), j)
+            (sigma, Changed {bodyonly=lazy false})
+            lf
+        else
+          Array.fold_left2_map (fun (sigma,changed) good c ->
+              let sigma, changed', t = recheck_against env sigma good c in
+              (sigma, merge_changes changed changed'), t)
+            (sigma,Same) glf lf
+      in
+      let sigma, changediv = match giv, iv with
+        | _, NoInvert -> sigma, Same
+        | NoInvert, CaseInvert {indices} ->
+          (* likely bug but accept for now *)
+          let args = Array.append pms indices in
+          let t = mkApp (mkIndU (ci.ci_ind,u), args) in
+          let sigma, tj = execute env sigma t in
+          let sigma, tj = type_judgment env sigma tj in
+          let sigma = check_actual_type env sigma cj tj.utj_val in
+          sigma, Changed {bodyonly=Lazy.from_val false}
+        | CaseInvert {indices=gindices}, CaseInvert {indices} ->
+          let gargs = Array.append gpms gindices in
+          let args = Array.append pms indices in
+          let gt = mkApp (mkIndU (gci.ci_ind,u), gargs) in
+          let t = mkApp (mkIndU (ci.ci_ind,u), args) in
+          let sigma, changediv, tj = recheck_against env sigma gt t in
+          let sigma, tj = type_judgment env sigma tj in
+          let sigma = check_actual_type env sigma cj tj.utj_val in
+          sigma, changediv
+      in
+      if unchanged changedc && unchanged changedp && unchanged changediv && bodyonly changedlf
+      then assume_unchanged_type sigma
+      else maybe_changed (judge_of_case env sigma case ci pj iv cj lfj)
+
+    | Proj (gp, gc),
+      Proj (p, c) ->
+      if not (QProjection.equal env gp p) then default ()
+      else
+        let sigma, changed, c = recheck_against env sigma gc c in
+        maybe_changed (sigma, judge_of_projection env sigma p c)
+
+    | App (gf, gargs),
+      App (f, args) ->
+      if Array.length gargs <> Array.length args then
+        let sigma, _, fj = recheck_against env sigma gf f in
+        let sigma, jl = execute_array env sigma args in
+        (match EConstr.kind sigma f with
+         | Ind (ind, u) when EInstance.is_empty u && Environ.template_polymorphic_ind ind env ->
+           maybe_changed (judge_of_applied_inductive_knowing_parameters env sigma fj (ind, u) jl)
+         | _ ->
+           (* No template polymorphism *)
+           maybe_changed (judge_of_apply env sigma fj jl))
+      else begin
+        let (sigma, changedargs), jl =
+          Array.fold_left2_map (fun (sigma,changed) good c ->
+              let sigma, changed', t = recheck_against env sigma good c in
+              (sigma, merge_changes changed changed'), (changed', t))
+            (sigma,Same) gargs args
+        in
+        let sigma, changedf, fj = recheck_against env sigma gf f in
+        if unchanged changedargs && bodyonly changedf
+        then assume_unchanged_type sigma
+        else
+          (match EConstr.kind sigma f with
+           | Ind (ind, u) when EInstance.is_empty u && Environ.template_polymorphic_ind ind env ->
+             let sigma, _ = judge_of_apply_against env sigma changedf fj jl in
+             let jl = Array.map snd jl in
+             maybe_changed (judge_of_applied_inductive_knowing_parameters_nocheck env sigma fj (ind, u) jl)
+           | _ ->
+             (* No template polymorphism *)
+             maybe_changed (judge_of_apply_against env sigma changedf fj jl))
+      end
+
+    | Lambda (_, gc1, gc2),
+      Lambda (name, c1, c2) ->
+      let sigma, changedj, j = recheck_against env sigma gc1 c1 in
+      let sigma, var = type_judgment env sigma j in
+      let name = check_binder_annot var.utj_type name in
+      let env1 = push_rel (LocalAssum (name, var.utj_val)) env in
+      let sigma, changedj', j' = if unchanged changedj then recheck_against env1 sigma gc2 c2
+        else let sigma, j' = execute env1 sigma c2 in
+          sigma, Changed {bodyonly=lazy false}, j'
+      in
+      sigma, merge_changes changedj' changedj, judge_of_abstraction env1 name.binder_name var j'
+
+    | Prod (_, gc1, gc2),
+      Prod (name, c1, c2) ->
+      let sigma, changedj, j = recheck_against env sigma gc1 c1 in
+      let sigma, var = type_judgment env sigma j in
+      let name = check_binder_annot var.utj_type name in
+      let env1 = push_rel (LocalAssum (name, var.utj_val)) env in
+      let sigma, changedj', j' = if unchanged changedj then recheck_against env1 sigma gc2 c2
+        else let sigma, j' = execute env1 sigma c2 in
+          sigma, Changed {bodyonly=lazy false}, j'
+      in
+      let sigma, j' = type_judgment env1 sigma j' in
+      sigma, merge_changes changedj' changedj, judge_of_product env1 name.binder_name var j'
+
+    | Cast (gc, _, gt),
+      Cast (c, k, t) ->
+      let sigma, changedc, cj = recheck_against env sigma gc c in
+      let sigma, changedt, tj = recheck_against env sigma gt t in
+      if unchanged changedt && bodyonly changedc
+      then assume_unchanged_type sigma
+      else
+        let sigma, tj = type_judgment env sigma tj in
+        maybe_changed (judge_of_cast env sigma cj k tj)
+
+    | _, (Case _ | App _ | Lambda _ | Prod _ | Cast _ | Proj _) -> default ()
+
+let recheck_against env sigma a b =
+  let sigma, _, j = recheck_against env sigma a b in
+  sigma, j.uj_type
