@@ -17,10 +17,6 @@ open Evd
 (** This module provides the typing machine with existential variables
     and universes. *)
 
-(** Typecheck a term and return its type. May trigger an evarmap leak. *)
-val unsafe_type_of : env -> evar_map -> constr -> types
-[@@ocaml.deprecated "Use [type_of] or retyping according to your needs."]
-
 (** Typecheck a term and return its type + updated evars, optionally refreshing
     universes *)
 val type_of : ?refresh:bool -> env -> evar_map -> constr -> evar_map * types
@@ -50,9 +46,15 @@ val check_allowed_sort : env -> evar_map -> pinductive -> constr -> constr ->
 val check_type_fixpoint : ?loc:Loc.t -> env -> evar_map ->
   Names.Name.t Context.binder_annot array -> types array -> unsafe_judgment array -> evar_map
 
+(** Variant of {!check} that assumes that the argument term is well-typed. *)
+val check_actual_type : env -> evar_map -> unsafe_judgment -> types -> evar_map
+
+val type_judgment : env -> evar_map -> unsafe_judgment -> evar_map * unsafe_type_judgment
+
 val judge_of_sprop : unsafe_judgment
 val judge_of_prop : unsafe_judgment
 val judge_of_set : unsafe_judgment
+val judge_of_variable : env -> Id.t -> unsafe_judgment
 val judge_of_apply : env -> evar_map -> unsafe_judgment -> unsafe_judgment array ->
   evar_map * unsafe_judgment
 val judge_of_abstraction : Environ.env -> Name.t ->
@@ -62,3 +64,9 @@ val judge_of_product : Environ.env -> Name.t ->
 val judge_of_projection : env -> evar_map -> Projection.t -> unsafe_judgment -> unsafe_judgment
 val judge_of_int : Environ.env -> Uint63.t -> unsafe_judgment
 val judge_of_float : Environ.env -> Float64.t -> unsafe_judgment
+
+val checked_appvect : env -> evar_map -> constr -> constr array -> evar_map * constr
+val checked_applist : env -> evar_map -> constr -> constr list -> evar_map * constr
+
+(** hack *)
+val recheck_against : Environ.env -> evar_map -> constr -> constr -> evar_map * types
