@@ -24,10 +24,6 @@ open Evd
   check that the name exist and that its type is convertible\\
 *)
 
-(** The primitive refiner. *)
-
-val refiner : check:bool -> constr -> unit Proofview.tactic
-
 (** {6 Refiner errors. } *)
 
 type refiner_error =
@@ -47,9 +43,6 @@ exception RefinerError of Environ.env * evar_map * refiner_error
 
 val error_no_such_hypothesis : Environ.env -> evar_map -> Id.t -> 'a
 
-val catchable_exception : exn -> bool
-[@@ocaml.deprecated "This function does not scale in the presence of dynamically added exceptions. Use instead CErrors.noncritical, or the exact name of the exception that matters in the corresponding case."]
-
 (** Move destination for hypothesis *)
 
 type 'id move_location =
@@ -64,6 +57,9 @@ val pr_move_location :
 val convert_hyp : check:bool -> reorder:bool -> Environ.env -> evar_map ->
   EConstr.named_declaration -> Environ.named_context_val
 
+type cannot_move_hyp
+exception CannotMoveHyp of cannot_move_hyp
+
 val move_hyp_in_named_context : Environ.env -> Evd.evar_map -> Id.t -> Id.t move_location ->
   Environ.named_context_val -> Environ.named_context_val
 
