@@ -19,7 +19,7 @@ module AsyncOpts : sig
   type async_proofs = APoff
                     | APonLazy (* Delays proof checking, but does it in master *)
                     | APon
-  type tac_error_filter = [ `None | `Only of string list | `All ]
+  type tac_error_filter = FNone | FOnly of string list | FAll
 
   type stm_opt = {
     async_proofs_n_workers : int;
@@ -42,29 +42,13 @@ module AsyncOpts : sig
 
 end
 
-type interactive_top = TopLogical of DirPath.t | TopPhysical of string
-
 (** The STM document type [stm_doc_type] determines some properties
    such as what uncompleted proofs are allowed and what gets recorded
    to aux files. *)
 type stm_doc_type =
   | VoDoc       of string       (* file path *)
   | VioDoc      of string       (* file path *)
-  | Interactive of interactive_top    (* module path *)
-
-type option_command = OptionSet of string option | OptionUnset
-
-type injection_command =
-  | OptionInjection of (Goptions.option_name * option_command)
-  (** Set flags or options before the initial state is ready. *)
-  | RequireInjection of (string * string option * bool option)
-  (** Require libraries before the initial state is
-     ready. Parameters follow [Library], that is to say,
-     [lib,prefix,import_export] means require library [lib] from
-     optional [prefix] and [import_export] if [Some false/Some true]
-     is used.  *)
-  (* -load-vernac-source interleaving is not supported yet *)
-  (* | LoadInjection of (string * bool) *)
+  | Interactive of Coqargs.top    (* module path *)
 
 (** STM initialization options: *)
 type stm_init_options =
@@ -73,26 +57,19 @@ type stm_init_options =
      the specified [doc_type]. This distinction should disappear at
      some some point. *)
 
-  ; ml_load_path : CUnix.physical_path list
-  (** OCaml load paths for the document. *)
-
-  ; vo_load_path   : Loadpath.vo_path list
-  (** [vo] load paths for the document. Usually extracted from -R
-     options / _CoqProject *)
-
-  ; injections : injection_command list
+  ; injections : Coqargs.injection_command list
   (** Injects Require and Set/Unset commands before the initial
      state is ready *)
 
-  ; stm_options  : AsyncOpts.stm_opt
-  (** Low-level STM options *)
   }
 
 (** The type of a STM document *)
 type doc
 
-(** [init_core] performs some low-level initialization; should go away
-   in future releases. *)
+(** [init_process] performs some low-level initialization, call early *)
+val init_process : AsyncOpts.stm_opt -> unit
+
+(** [init_core] snapshorts the initial system state *)
 val init_core : unit -> unit
 
 (** [new_doc opt] Creates a new document with options [opt] *)
@@ -109,6 +86,8 @@ val parse_sentence :
 (* Reminder: A parsable [pa] is constructed using
    [Pcoq.Parsable.t stream], where [stream : char Stream.t]. *)
 
+type add_focus = NewAddTip | Unfocus of Stateid.t
+
 (* [add ~ontop ?newtip verbose cmd] adds a new command [cmd] ontop of
    the state [ontop].
    The [ontop] parameter just asserts that the GUI is on
@@ -117,7 +96,7 @@ val parse_sentence :
    to be [newtip] *)
 val add : doc:doc -> ontop:Stateid.t -> ?newtip:Stateid.t ->
   bool -> Vernacexpr.vernac_control ->
-  doc * Stateid.t * [ `NewTip | `Unfocus of Stateid.t ]
+  doc * Stateid.t * add_focus
 
 (* Returns the proof state before the last tactic that was applied at or before
 the specified state AND that has differences in the underlying proof (i.e.,
@@ -132,7 +111,7 @@ val get_proof : doc:doc -> Stateid.t -> Proof.t option
 val query : doc:doc ->
   at:Stateid.t -> route:Feedback.route_id -> Pcoq.Parsable.t -> unit
 
-(* [edit_at id] is issued to change the editing zone.  [`NewTip] is returned if
+(* [edit_at id] is issued to change the editing zone.  [NewTip] is returned if
    the requested id is the new document tip hence the document portion following
    [id] is dropped by Coq.  [`Focus fo] is returned to say that [fo.tip] is the
    new document tip, the document between [id] and [fo.stop] has been dropped.
@@ -142,13 +121,14 @@ val query : doc:doc ->
    If Flags.async_proofs_full is set, then [id] is not [observe]d, else it is.
 *)
 type focus = { start : Stateid.t; stop : Stateid.t; tip : Stateid.t }
-val edit_at : doc:doc -> Stateid.t -> doc * [ `NewTip | `Focus of focus ]
+type edit_focus = NewTip | Focus of focus
+val edit_at : doc:doc -> Stateid.t -> doc * edit_focus
 
 (* [observe doc sid]] Check / execute span [sid] *)
 val observe : doc:doc -> Stateid.t -> doc
 
 (* [finish doc] Fully checks a document up to the "current" tip. *)
-val finish : doc:doc -> doc
+val finish : doc:doc -> doc * Vernacstate.t
 
 (* Internal use (fake_ide) only, do not use *)
 val wait : doc:doc -> doc
@@ -171,8 +151,10 @@ val snapshot_vio : create_vos:bool -> doc:doc -> output_native_objects:bool -> D
  * after having built a .vio in batch mode *)
 val reset_task_queue : unit -> unit
 
+type document
+
 (* A .vio contains tasks to be completed *)
-type tasks
+type tasks = (Opaqueproof.opaque_handle option, document) Library.tasks
 val check_task : string -> tasks -> int -> bool
 val info_tasks : tasks -> (string * float * int) list
 val finish_tasks : string ->
@@ -195,7 +177,6 @@ val set_perspective : doc:doc -> Stateid.t list -> unit
 (** workers **************************************************************** **)
 
 module ProofTask : AsyncTaskQueue.Task
-module TacTask   : AsyncTaskQueue.Task
 module QueryTask : AsyncTaskQueue.Task
 
 (** document structure customization *************************************** **)
@@ -261,12 +242,14 @@ type static_block_detection =
 
 type recovery_action = {
   base_state : Stateid.t;
-  goals_to_admit : Goal.goal list;
+  goals_to_admit : Evar.t list;
   recovery_command : Vernacexpr.vernac_control option;
 }
 
+type block_classification = ValidBlock of recovery_action | Leaks
+
 type dynamic_block_error_recovery =
-  doc -> static_block_declaration -> [ `ValidBlock of recovery_action | `Leaks ]
+  doc -> static_block_declaration -> block_classification
 
 val register_proof_block_delimiter :
   Vernacextend.proof_block_name ->
@@ -281,33 +264,34 @@ val register_proof_block_delimiter :
  * the alternative toploop for the worker can be selected by changing
  * the name of the Task(s) above) *)
 
-val state_computed_hook : (doc:doc -> Stateid.t -> in_cache:bool -> unit) Hook.t
-val unreachable_state_hook :
-  (doc:doc -> Stateid.t -> Exninfo.iexn -> unit) Hook.t
+val state_computed_hook : (doc:doc -> Stateid.t -> in_cache:bool -> unit) -> unit
+val unreachable_state_hook : (doc:doc -> Stateid.t -> Exninfo.iexn -> unit) -> unit
 
 (* ready means that master has it at hand *)
-val state_ready_hook : (doc:doc -> Stateid.t -> unit) Hook.t
+val state_ready_hook : (doc:doc -> Stateid.t -> unit) -> unit
 
 (* Messages from the workers to the master *)
-val forward_feedback_hook : (Feedback.feedback -> unit) Hook.t
+val forward_feedback_hook : (Feedback.feedback -> unit) -> unit
 
 (*
  * Hooks into the UI for plugins (not for general use)
  *)
 
 (** User adds a sentence to the document (after parsing) *)
-val document_add_hook : (Vernacexpr.vernac_control -> Stateid.t -> unit) Hook.t
+val document_add_hook : (Vernacexpr.vernac_control -> Stateid.t -> unit) -> unit
 
 (** User edits a sentence in the document *)
-val document_edit_hook : (Stateid.t -> unit) Hook.t
+val document_edit_hook : (Stateid.t -> unit) -> unit
 
 (** User requests evaluation of a sentence *)
-val sentence_exec_hook : (Stateid.t -> unit) Hook.t
+val sentence_exec_hook : (Stateid.t -> unit) -> unit
 
 val get_doc : Feedback.doc_id -> doc
 
+type state = Valid of Vernacstate.t option | Expired | Error of exn
+
 val state_of_id : doc:doc ->
-  Stateid.t -> [ `Valid of Vernacstate.t option | `Expired | `Error of exn ]
+  Stateid.t -> state
 
 (* Queries for backward compatibility *)
 val current_proof_depth : doc:doc -> int
@@ -316,6 +300,5 @@ val get_all_proof_names : doc:doc -> Id.t list
 (** Enable STM debugging *)
 val stm_debug : bool ref
 
-type document
 val backup : unit -> document
 val restore : document -> unit
