@@ -17,7 +17,7 @@ open Equality
 type raw_rew_rule = (constr Univ.in_universe_context_set * bool * Genarg.raw_generic_argument option) CAst.t
 
 (** To add rewriting rules to a base *)
-val add_rew_rules : string -> raw_rew_rule list -> unit
+val add_rew_rules : locality:Hints.hint_locality -> string -> raw_rew_rule list -> unit
 
 (** The AutoRewrite tactic.
    The optional conditions tell rewrite how to handle matching and side-condition solving.
@@ -27,16 +27,17 @@ val autorewrite : ?conds:conditions -> unit Proofview.tactic -> string list -> u
 val autorewrite_in : ?conds:conditions -> Names.Id.t -> unit Proofview.tactic -> string list -> unit Proofview.tactic
 
 (** Rewriting rules *)
-type rew_rule = { rew_lemma: constr;
-                  rew_type: types;
-                  rew_pat: constr;
-                  rew_ctx: Univ.ContextSet.t;
-                  rew_l2r: bool;
-                  rew_tac: Genarg.glob_generic_argument option }
+module RewRule :
+sig
+   type t
+   val rew_lemma : t -> Univ.ContextSet.t * constr
+   val rew_l2r : t -> bool
+   val rew_tac : t -> Genarg.glob_generic_argument option
+end
 
-val find_rewrites : string -> rew_rule list
+val find_rewrites : string -> RewRule.t list
 
-val find_matches : string -> constr -> rew_rule list
+val find_matches : string -> constr -> RewRule.t list
 
 val auto_multi_rewrite : ?conds:conditions -> string list -> Locus.clause -> unit Proofview.tactic
 
@@ -44,21 +45,4 @@ val auto_multi_rewrite_with : ?conds:conditions -> unit Proofview.tactic -> stri
 
 val print_rewrite_hintdb : string -> Pp.t
 
-open Clenv
-
-
-type hypinfo = {
-  hyp_cl : clausenv;
-  hyp_prf : constr;
-  hyp_ty : types;
-  hyp_car : constr;
-  hyp_rel : constr;
-  hyp_l2r : bool;
-  hyp_left : constr;
-  hyp_right : constr;
-}
-
-val find_applied_relation :
-  ?loc:Loc.t -> bool ->
-  Environ.env -> Evd.evar_map -> constr -> bool -> hypinfo
-
+val default_hint_rewrite_locality : unit -> Hints.hint_locality
