@@ -22,8 +22,8 @@ open Namegen
 open Inductiveops
 open Printer
 open Retyping
-open Tacmach.New
-open Tacticals.New
+open Tacmach
+open Tacticals
 open Tactics
 open Elim
 open Equality
@@ -90,7 +90,7 @@ let make_inv_predicate env evd indf realargs id status concl =
             (hyps_arity,concl)
       | Dep dflt_concl ->
           if not (occur_var env !evd id concl) then
-            user_err ~hdr:"make_inv_predicate"
+            user_err
               (str "Current goal does not depend on " ++ Id.print id ++ str".");
           (* We abstract the conclusion of goal with respect to
              realargs and c to * be concl in order to rewrite and have
@@ -116,7 +116,11 @@ let make_inv_predicate env evd indf realargs id status concl =
   (* Now, we can recurse down this list, for each ai,(mkRel k) whether to
      push <Ai>(mkRel k)=ai (when   Ai is closed).
    In any case, we carry along the rest of pairs *)
-  let eqdata = Coqlib.build_coq_eq_data () in
+  let eqdata =
+    try Coqlib.build_coq_eq_data ()
+    with UserError _ ->
+    try Coqlib.build_coq_identity_data ()
+    with UserError _ -> user_err (str "No registered equality.") in
   let rec build_concl eqns args n = function
     | [] -> it_mkProd concl eqns, Array.rev_of_list args
     | ai :: restlist ->
@@ -351,8 +355,12 @@ let remember_first_eq id x = if !x == Logic.MoveLast then x := Logic.MoveAfter i
 let dest_nf_eq env sigma t = match EConstr.kind sigma t with
 | App (r, [| t; x; y |]) ->
   let open Reductionops in
-  let eq = Coqlib.lib_ref "core.eq.type" in
-  if isRefX sigma eq r then
+  let is_global_exists gr c =
+    Coqlib.has_ref gr && isRefX sigma (Coqlib.lib_ref gr) c
+  in
+  let is_eq = is_global_exists "core.eq.type" r in
+  let is_identity = is_global_exists "core.identity.type" r in
+  if is_eq || is_identity then
     (t, whd_all env sigma x, whd_all env sigma y)
   else user_err Pp.(str "Not an equality.")
 | _ ->
@@ -375,8 +383,7 @@ let projectAndApply as_mode thin avoid id eqname names depids =
     | _ -> tac id
     end
   in
-  let deq_trailer id clear_flag _ neqns =
-    assert (clear_flag == None);
+  let deq_trailer id neqns =
     tclTHENLIST
       [if as_mode then clear [id] else tclIDTAC;
        (tclMAP_i (false,false) neqns (function (idopt,_) ->
@@ -409,7 +416,7 @@ let nLastDecls i tac =
 
 let rewrite_equations as_mode othin neqns names ba =
   Proofview.Goal.enter begin fun gl ->
-  let (depids,nodepids) = split_dep_and_nodep ba.Tacticals.assums gl in
+  let (depids,nodepids) = split_dep_and_nodep ba gl in
   let first_eq = ref Logic.MoveLast in
   let avoid = if as_mode then Id.Set.of_list (List.map NamedDecl.get_id nodepids) else Id.Set.empty in
   match othin with
@@ -463,7 +470,7 @@ let raw_inversion inv_kind id status names =
     let env = Proofview.Goal.env gl in
     let concl = Proofview.Goal.concl gl in
     let c = mkVar id in
-    let (ind, t) =
+    let ((ind, u), t) =
       try pf_apply Tacred.reduce_to_atomic_ind gl (pf_get_type_of gl c)
       with UserError _ ->
         let msg = str "The type of " ++ Id.print id ++ str " is not inductive." in
@@ -474,13 +481,12 @@ let raw_inversion inv_kind id status names =
     let (elim_predicate, args) =
       make_inv_predicate env evdref indf realargs id status concl in
     let sigma = !evdref in
-    let (cut_concl,case_tac) =
-      if status != NoDep && (local_occur_var sigma id concl) then
-        Reductionops.beta_applist sigma (elim_predicate, realargs@[c]),
-        case_then_using
+    let dep = status != NoDep && (local_occur_var sigma id concl) in
+    let cut_concl =
+      if dep then
+        Reductionops.beta_applist sigma (elim_predicate, realargs@[c])
       else
-        Reductionops.beta_applist sigma (elim_predicate, realargs),
-        case_nodep_then_using
+        Reductionops.beta_applist sigma (elim_predicate, realargs)
     in
     let refined id =
       let prf = mkApp (mkVar id, args) in
@@ -488,13 +494,11 @@ let raw_inversion inv_kind id status names =
     in
     let neqns = List.length realargs in
     let as_mode = names != None in
+    let (_, args) = decompose_app_vect sigma t in
     tclTHEN (Proofview.Unsafe.tclEVARS sigma)
       (tclTHENS
         (assert_before Anonymous cut_concl)
-        [case_tac names
-            (introCaseAssumsThen false (* ApplyOn not supported by inversion *)
-               (rewrite_equations_tac as_mode inv_kind id neqns))
-            (Some elim_predicate) ind (c,t);
+        [case_tac dep names (rewrite_equations_tac as_mode inv_kind id neqns) elim_predicate (ind, u, args) id;
         onLastHypId (fun id -> tclTHEN (refined id) reflexivity)])
   end
 
@@ -524,13 +528,13 @@ let inv_gen thin status names =
 
 let inv k = inv_gen k NoDep
 
-let inv_tac id       = inv FullInversion None (NamedHyp id)
-let inv_clear_tac id = inv FullInversionClear None (NamedHyp id)
+let inv_tac id       = inv FullInversion None (NamedHyp (CAst.make id))
+let inv_clear_tac id = inv FullInversionClear None (NamedHyp (CAst.make id))
 
 let dinv k c = inv_gen k (Dep c)
 
-let dinv_tac id       = dinv FullInversion None None (NamedHyp id)
-let dinv_clear_tac id = dinv FullInversionClear None None (NamedHyp id)
+let dinv_tac id       = dinv FullInversion None None (NamedHyp (CAst.make id))
+let dinv_clear_tac id = dinv FullInversionClear None None (NamedHyp (CAst.make id))
 
 (* InvIn will bring the specified clauses into the conclusion, and then
  * perform inversion on the named hypothesis.  After, it will intro them
