@@ -14,7 +14,6 @@ open Util
 open Names
 open Constr
 open EConstr
-open Declarations
 open Genredexpr
 open Pattern
 open Reductionops
@@ -23,13 +22,20 @@ open CClosure
 open RedFlags
 open Libobject
 
+let warn_vm_disabled =
+  CWarnings.create ~name:"vm-compute-disabled" ~category:"bytecode-compiler"
+  (fun () ->
+   strbrk "vm_compute disabled at configure time; falling back to cbv.")
+
 (* call by value normalisation function using the virtual machine *)
 let cbv_vm env sigma c =
-  if Coq_config.bytecode_compiler then
+  if (Environ.typing_flags env).enable_VM then
     let ctyp = Retyping.get_type_of env sigma c in
     Vnorm.cbv_vm env sigma c ctyp
-  else
+  else begin
+    warn_vm_disabled ();
     compute env sigma c
+  end
 
 let warn_native_compute_disabled =
   CWarnings.create ~name:"native-compute-disabled" ~category:"native-compiler"
@@ -44,14 +50,7 @@ let cbv_native env sigma c =
     (warn_native_compute_disabled ();
      cbv_vm env sigma c)
 
-let whd_cbn flags env sigma t =
-  let (state,_) =
-    (whd_state_gen ~refold:true ~tactic_mode:true flags env sigma (t,Reductionops.Stack.empty))
-  in
-  Reductionops.Stack.zip ~refold:true sigma state
-
-let strong_cbn flags =
-  strong_with_flags whd_cbn flags
+let whd_cbn = Cbn.whd_cbn
 
 let simplIsCbn =
   Goptions.declare_bool_option_and_ref ~depr:false ~key:["SimplIsCbn"] ~value:false
@@ -61,20 +60,16 @@ let set_strategy_one ref l  =
     match ref with
       | EvalConstRef sp -> ConstKey sp
       | EvalVarRef id -> VarKey id in
-  Global.set_strategy k l;
-  match k,l with
-      ConstKey sp, Conv_oracle.Opaque ->
-        Csymtable.set_opaque_const sp
-    | ConstKey sp, _ ->
-        let cb = Global.lookup_constant sp in
-        (match cb.const_body with
-          | OpaqueDef _ ->
-            user_err ~hdr:"set_transparent_const"
-              (str "Cannot make" ++ spc () ++
-                 Nametab.pr_global_env Id.Set.empty (GlobRef.ConstRef sp) ++
-                 spc () ++ str "transparent because it was declared opaque.");
-          | _ -> Csymtable.set_transparent_const sp)
-    | _ -> ()
+  let () = Global.set_strategy k l in
+  match k, l with
+  | ConstKey sp, Conv_oracle.Opaque -> ()
+  | ConstKey sp, _ ->
+    if Declareops.is_opaque (Global.lookup_constant sp) then
+      user_err
+        (str "Cannot make" ++ spc () ++
+            Nametab.pr_global_env Id.Set.empty (GlobRef.ConstRef sp) ++
+            spc () ++ str "transparent because it was declared opaque.")
+  | _ -> ()
 
 let cache_strategy (_,str) =
   List.iter
@@ -85,7 +80,7 @@ let subst_strategy (subs,(local,obj)) =
   local,
   List.Smart.map
     (fun (k,ql as entry) ->
-      let ql' = List.Smart.map (Mod_subst.subst_evaluable_reference subs) ql in
+      let ql' = List.Smart.map (Tacred.subst_evaluable_reference subs) ql in
       if ql==ql' then entry else (k,ql'))
     obj
 
@@ -101,15 +96,15 @@ let map_strategy f l =
       if List.is_empty ql' then str else (lev,ql')::str) l [] in
   if List.is_empty l' then None else Some (false,l')
 
-let classify_strategy (local,_ as obj) =
-  if local then Dispose else Substitute obj
+let classify_strategy (local,_) =
+  if local then Dispose else Substitute
 
 let disch_ref ref =
   match ref with
       EvalConstRef c -> Some ref
     | EvalVarRef id -> if Lib.is_in_section (GlobRef.VarRef id) then None else Some ref
 
-let discharge_strategy (_,(local,obj)) =
+let discharge_strategy (local,obj) =
   if local then None else
   map_strategy disch_ref obj
 
@@ -117,22 +112,27 @@ type strategy_obj =
     bool * (Conv_oracle.level * evaluable_global_reference list) list
 
 let inStrategy : strategy_obj -> obj =
-  declare_object {(default_object "STRATEGY") with
-                    cache_function = (fun (_,obj) -> cache_strategy obj);
-                    load_function = (fun _ (_,obj) -> cache_strategy obj);
-                    subst_function = subst_strategy;
-                    discharge_function = discharge_strategy;
-                    classify_function = classify_strategy }
+  declare_object
+    {(default_object "STRATEGY") with
+     cache_function = cache_strategy;
+     load_function = (fun _ obj -> cache_strategy obj);
+     subst_function = subst_strategy;
+     discharge_function = discharge_strategy;
+     classify_function = classify_strategy;
+    }
 
 
 let set_strategy local str =
-  Lib.add_anonymous_leaf (inStrategy (local,str))
+  Lib.add_leaf (inStrategy (local,str))
 
 (* Generic reduction: reduction functions used in reduction tactics *)
 
 type red_expr =
     (constr, evaluable_global_reference, constr_pattern) red_expr_gen
 
+type red_expr_val =
+  (constr, evaluable_global_reference, constr_pattern, CClosure.RedFlags.reds) red_expr_gen0
+
 let make_flag_constant = function
   | EvalVarRef id -> fVAR id
   | EvalConstRef sp -> fCONST sp
@@ -153,7 +153,7 @@ let make_flag env f =
           (fun v red -> red_sub red (make_flag_constant v))
           f.rConst red
     else (* Only rConst *)
-        let red = red_add_transparent (red_add red fDELTA) TransparentState.empty in
+        let red = red_add red fDELTA in
         List.fold_right
           (fun v red -> red_add red (make_flag_constant v))
           f.rConst red
@@ -169,20 +169,20 @@ let red_expr_tab = Summary.ref String.Map.empty ~name:"Declare Reduction"
 
 let declare_reduction s f =
   if String.Map.mem s !reduction_tab || String.Map.mem s !red_expr_tab
-  then user_err ~hdr:"Redexpr.declare_reduction"
-    (str "There is already a reduction expression of name " ++ str s)
+  then user_err
+    (str "There is already a reduction expression of name " ++ str s ++ str ".")
   else reduction_tab := String.Map.add s f !reduction_tab
 
 let check_custom = function
   | ExtraRedExpr s ->
       if not (String.Map.mem s !reduction_tab || String.Map.mem s !red_expr_tab)
-      then user_err ~hdr:"Redexpr.check_custom" (str "Reference to undefined reduction expression " ++ str s)
+      then user_err (str "Reference to undefined reduction expression " ++ str s ++ str ".")
   |_ -> ()
 
 let decl_red_expr s e =
   if String.Map.mem s !reduction_tab || String.Map.mem s !red_expr_tab
-  then user_err ~hdr:"Redexpr.decl_red_expr"
-    (str "There is already a reduction expression of name " ++ str s)
+  then user_err
+    (str "There is already a reduction expression of name " ++ str s ++ str ".")
   else begin
     check_custom e;
     red_expr_tab := String.Map.add s e !red_expr_tab
@@ -225,38 +225,120 @@ let warn_simpl_unfolding_modifiers =
          (fun () ->
           Pp.strbrk "The legacy simpl ignores constant unfolding modifiers.")
 
-let reduction_of_red_expr env =
-  let make_flag = make_flag env in
-  let rec reduction_of_red_expr = function
+let rec eval_red_expr env = function
+| Simpl (f, o) ->
+  let () =
+    if not (simplIsCbn () || List.is_empty f.rConst) then
+      warn_simpl_unfolding_modifiers () in
+  let f = if simplIsCbn () then make_flag env f else CClosure.all (* dummy *) in
+  Simpl (f, o)
+| Cbv f -> Cbv (make_flag env f)
+| Cbn f -> Cbn (make_flag env f)
+| Lazy f -> Lazy (make_flag env f)
+| ExtraRedExpr s ->
+  begin match String.Map.find s !red_expr_tab with
+  | e -> eval_red_expr env e
+  | exception Not_found -> ExtraRedExpr s (* delay to runtime interpretation *)
+  end
+| (Red _ | Hnf | Unfold _ | Fold _ | Pattern _ | CbvVm _ | CbvNative _) as e -> e
+
+let reduction_of_red_expr_val = function
   | Red internal ->
       if internal then (e_red try_red_product,DEFAULTcast)
       else (e_red red_product,DEFAULTcast)
   | Hnf -> (e_red hnf_constr,DEFAULTcast)
   | Simpl (f,o) ->
-     let whd_am = if simplIsCbn () then whd_cbn (make_flag f) else whd_simpl in
-     let am = if simplIsCbn () then strong_cbn (make_flag f) else simpl in
-     let () =
-       if not (simplIsCbn () || List.is_empty f.rConst) then
-         warn_simpl_unfolding_modifiers () in
+     let whd_am = if simplIsCbn () then whd_cbn f else whd_simpl in
+     let am = if simplIsCbn () then Cbn.norm_cbn f else simpl in
      (contextualize (if head_style then whd_am else am) am o,DEFAULTcast)
-  | Cbv f -> (e_red (cbv_norm_flags (make_flag f)),DEFAULTcast)
+  | Cbv f -> (e_red (cbv_norm_flags f),DEFAULTcast)
   | Cbn f ->
-     (e_red (strong_cbn (make_flag f)), DEFAULTcast)
-  | Lazy f -> (e_red (clos_norm_flags (make_flag f)),DEFAULTcast)
+     (e_red (Cbn.norm_cbn f), DEFAULTcast)
+  | Lazy f -> (e_red (clos_norm_flags f),DEFAULTcast)
   | Unfold ubinds -> (e_red (unfoldn (List.map out_with_occurrences ubinds)),DEFAULTcast)
   | Fold cl -> (e_red (fold_commands cl),DEFAULTcast)
   | Pattern lp -> (pattern_occs (List.map out_with_occurrences lp),DEFAULTcast)
   | ExtraRedExpr s ->
       (try (e_red (String.Map.find s !reduction_tab),DEFAULTcast)
       with Not_found ->
-        (try reduction_of_red_expr (String.Map.find s !red_expr_tab)
-         with Not_found ->
-           user_err ~hdr:"Redexpr.reduction_of_red_expr"
-             (str "unknown user-defined reduction \"" ++ str s ++ str "\"")))
+           user_err
+             (str "Unknown user-defined reduction \"" ++ str s ++ str "\"."))
   | CbvVm o -> (contextualize cbv_vm cbv_vm o, VMcast)
   | CbvNative o -> (contextualize cbv_native cbv_native o, NATIVEcast)
+
+let reduction_of_red_expr env r =
+  reduction_of_red_expr_val (eval_red_expr env r)
+
+(* Possibly equip a reduction with the occurrences mentioned in an
+   occurrence clause *)
+
+let error_illegal_clause () =
+  CErrors.user_err Pp.(str "\"at\" clause not supported in presence of an occurrence clause.")
+
+let error_multiple_patterns () =
+  CErrors.user_err Pp.(str "\"at\" clause in occurences not supported with multiple patterns or references.")
+
+let error_illegal_non_atomic_clause () =
+  CErrors.user_err Pp.(str "\"at\" clause not supported in presence of a non atomic \"in\" clause.")
+
+let error_at_in_occurrences_not_supported () =
+  CErrors.user_err Pp.(str "\"at\" clause not supported for this reduction tactic.")
+
+let bind_red_expr_occurrences occs nbcl redexp =
+  let open Locus in
+  let has_at_clause = function
+    | Unfold l -> List.exists (fun (occl,_) -> occl != AllOccurrences) l
+    | Pattern l -> List.exists (fun (occl,_) -> occl != AllOccurrences) l
+    | Simpl (_,Some (occl,_)) -> occl != AllOccurrences
+    | _ -> false in
+  if occs == AllOccurrences then
+    if nbcl > 1 && has_at_clause redexp then
+      error_illegal_non_atomic_clause ()
+    else
+      redexp
+  else
+    match redexp with
+    | Unfold (_::_::_) ->
+        error_multiple_patterns ()
+    | Unfold [(occl,c)] ->
+        if occl != AllOccurrences then
+          error_illegal_clause ()
+        else
+          Unfold [(occs,c)]
+    | Pattern (_::_::_) ->
+        error_multiple_patterns ()
+    | Pattern [(occl,c)] ->
+        if occl != AllOccurrences then
+          error_illegal_clause ()
+        else
+          Pattern [(occs,c)]
+    | Simpl (f,Some (occl,c)) ->
+        if occl != AllOccurrences then
+          error_illegal_clause ()
+        else
+          Simpl (f,Some (occs,c))
+    | CbvVm (Some (occl,c)) ->
+        if occl != AllOccurrences then
+          error_illegal_clause ()
+        else
+          CbvVm (Some (occs,c))
+    | CbvNative (Some (occl,c)) ->
+        if occl != AllOccurrences then
+          error_illegal_clause ()
+        else
+          CbvNative (Some (occs,c))
+    | Red _ | Hnf | Cbv _ | Lazy _ | Cbn _
+    | ExtraRedExpr _ | Fold _ | Simpl (_,None) | CbvVm None | CbvNative None ->
+        error_at_in_occurrences_not_supported ()
+    | Unfold [] | Pattern [] ->
+        assert false
+
+let reduction_of_red_expr_val ?occs r =
+  let r = match occs with
+  | None -> r
+  | Some (occs, nbcl) -> bind_red_expr_occurrences occs nbcl r
   in
-    reduction_of_red_expr
+  reduction_of_red_expr_val r
 
 let subst_mps subst c =
   EConstr.of_constr (Mod_subst.subst_mps subst (EConstr.Unsafe.to_constr c))
@@ -266,18 +348,18 @@ let subst_red_expr subs =
   let sigma = Evd.from_env env in
   Redops.map_red_expr_gen
     (subst_mps subs)
-    (Mod_subst.subst_evaluable_reference subs)
+    (Tacred.subst_evaluable_reference subs)
     (Patternops.subst_pattern env sigma subs)
 
 let inReduction : bool * string * red_expr -> obj =
   declare_object
     {(default_object "REDUCTION") with
-       cache_function = (fun (_,(_,s,e)) -> decl_red_expr s e);
-       load_function = (fun _ (_,(_,s,e)) -> decl_red_expr s e);
-       subst_function =
-        (fun (subs,(b,s,e)) -> b,s,subst_red_expr subs e);
-       classify_function =
-        (fun ((b,_,_) as obj) -> if b then Dispose else Substitute obj) }
+     cache_function = (fun (_,s,e) -> decl_red_expr s e);
+     load_function = (fun _ (_,s,e) -> decl_red_expr s e);
+     subst_function =
+       (fun (subs,(b,s,e)) -> b,s,subst_red_expr subs e);
+     classify_function =
+       (fun ((b,_,_)) -> if b then Dispose else Substitute) }
 
 let declare_red_expr locality s expr =
-    Lib.add_anonymous_leaf (inReduction (locality,s,expr))
+    Lib.add_leaf (inReduction (locality,s,expr))
