@@ -13,7 +13,6 @@ open Constr
 open EConstr
 open Environ
 open Evd
-open Clenv
 open Redexpr
 open Pattern
 open Unification
@@ -33,12 +32,8 @@ val is_quantified_hypothesis : Id.t -> Proofview.Goal.t -> bool
 (** {6 Primitive tactics. } *)
 
 val introduction    : Id.t -> unit Proofview.tactic
-val convert_concl   : check:bool -> types -> cast_kind -> unit Proofview.tactic
+val convert_concl   : cast:bool -> check:bool -> types -> cast_kind -> unit Proofview.tactic
 val convert_hyp     : check:bool -> reorder:bool -> named_declaration -> unit Proofview.tactic
-val convert_concl_no_check : types -> cast_kind -> unit Proofview.tactic
-[@@ocaml.deprecated "use [Tactics.convert_concl]"]
-val convert_hyp_no_check : named_declaration -> unit Proofview.tactic
-[@@ocaml.deprecated "use [Tactics.convert_hyp]"]
 val mutual_fix      :
   Id.t -> int -> (Id.t * int * constr) list -> int -> unit Proofview.tactic
 val fix             : Id.t -> int -> unit Proofview.tactic
@@ -51,8 +46,7 @@ val convert_leq     : constr -> constr -> unit Proofview.tactic
 (** {6 Introduction tactics. } *)
 
 val fresh_id_in_env : Id.Set.t -> Id.t -> env -> Id.t
-val fresh_id : Id.Set.t -> Id.t -> Goal.goal sigma -> Id.t
-val find_intro_names : rel_context -> Goal.goal sigma -> Id.t list
+val find_intro_names : env -> Evd.evar_map -> rel_context -> Id.t list
 
 val intro                : unit Proofview.tactic
 val introf               : unit Proofview.tactic
@@ -65,9 +59,14 @@ val intro_avoiding       : Id.Set.t -> unit Proofview.tactic
 
 val intro_replacing      : Id.t -> unit Proofview.tactic
 val intro_using          : Id.t -> unit Proofview.tactic
+[@@ocaml.deprecated "Prefer [intro_using_then] to avoid renaming issues."]
+val intro_using_then     : Id.t -> (Id.t -> unit Proofview.tactic) -> unit Proofview.tactic
 val intro_mustbe_force   : Id.t -> unit Proofview.tactic
+val intros_mustbe_force  : Id.t list -> unit Proofview.tactic
 val intro_then           : (Id.t -> unit Proofview.tactic) -> unit Proofview.tactic
 val intros_using         : Id.t list -> unit Proofview.tactic
+[@@ocaml.deprecated "Prefer [intros_using_then] to avoid renaming issues."]
+val intros_using_then    : Id.t list -> (Id.t list -> unit Proofview.tactic) -> unit Proofview.tactic
 val intros_replacing     : Id.t list -> unit Proofview.tactic
 val intros_possibly_replacing : Id.t list -> unit Proofview.tactic
 
@@ -76,11 +75,6 @@ val auto_intros_tac : Names.Name.t list -> unit Proofview.tactic
 
 val intros               : unit Proofview.tactic
 
-(** [depth_of_quantified_hypothesis b h g] returns the index of [h] in
-   the conclusion of goal [g], up to head-reduction if [b] is [true] *)
-val depth_of_quantified_hypothesis :
-  bool -> quantified_hypothesis -> Proofview.Goal.t -> int
-
 val intros_until         : quantified_hypothesis -> unit Proofview.tactic
 
 val intros_clearing      : bool list -> unit Proofview.tactic
@@ -118,6 +112,9 @@ val force_destruction_arg : evars_flag -> env -> evar_map ->
     delayed_open_constr_with_bindings destruction_arg ->
     evar_map * constr with_bindings destruction_arg
 
+val finish_evar_resolution : ?flags:Pretyping.inference_flags ->
+  env -> evar_map -> (evar_map * constr) -> evar_map * constr
+
 (** Tell if a used hypothesis should be cleared by default or not *)
 
 val use_clear_hyp_by_default : unit -> bool
@@ -154,8 +151,8 @@ type change_arg = patvar_map -> env -> evar_map -> evar_map * constr
 val make_change_arg   : constr -> change_arg
 val reduct_in_hyp     : check:bool -> reorder:bool -> tactic_reduction -> hyp_location -> unit Proofview.tactic
 val reduct_option     : check:bool -> tactic_reduction * cast_kind -> goal_location -> unit Proofview.tactic
-val reduct_in_concl   : check:bool -> tactic_reduction * cast_kind -> unit Proofview.tactic
-val e_reduct_in_concl   : check:bool -> e_tactic_reduction * cast_kind -> unit Proofview.tactic
+val reduct_in_concl : cast:bool -> check:bool -> tactic_reduction * cast_kind -> unit Proofview.tactic
+val e_reduct_in_concl : cast:bool -> check:bool -> e_tactic_reduction * cast_kind -> unit Proofview.tactic
 val change_in_concl   : check:bool -> (occurrences * constr_pattern) option -> change_arg -> unit Proofview.tactic
 val change_concl      : constr -> unit Proofview.tactic
 val change_in_hyp     : check:bool -> (occurrences * constr_pattern) option -> change_arg ->
@@ -174,11 +171,11 @@ val normalise_in_hyp  : hyp_location -> unit Proofview.tactic
 val normalise_option  : goal_location -> unit Proofview.tactic
 val normalise_vm_in_concl : unit Proofview.tactic
 val unfold_in_concl   :
-  (occurrences * evaluable_global_reference) list -> unit Proofview.tactic
+  (occurrences * Tacred.evaluable_global_reference) list -> unit Proofview.tactic
 val unfold_in_hyp     :
-  (occurrences * evaluable_global_reference) list -> hyp_location -> unit Proofview.tactic
+  (occurrences * Tacred.evaluable_global_reference) list -> hyp_location -> unit Proofview.tactic
 val unfold_option     :
-  (occurrences * evaluable_global_reference) list -> goal_location -> unit Proofview.tactic
+  (occurrences * Tacred.evaluable_global_reference) list -> goal_location -> unit Proofview.tactic
 val change            :
   check:bool -> constr_pattern option -> change_arg -> clause -> unit Proofview.tactic
 val pattern_option    :
@@ -192,7 +189,7 @@ val clear         : Id.t list -> unit Proofview.tactic
 val clear_body    : Id.t list -> unit Proofview.tactic
 val unfold_body   : Id.t -> unit Proofview.tactic
 val keep          : Id.t list -> unit Proofview.tactic
-val apply_clear_request : clear_flag -> bool -> constr -> unit Proofview.tactic
+val apply_clear_request : clear_flag -> bool -> Id.t option -> unit Proofview.tactic
 
 val specialize    : constr with_bindings -> intro_pattern option -> unit Proofview.tactic
 
@@ -207,16 +204,16 @@ val apply_type : typecheck:bool -> constr -> constr list -> unit Proofview.tacti
 val bring_hyps : named_context -> unit Proofview.tactic
 
 val apply                 : constr -> unit Proofview.tactic
-val eapply                : constr -> unit Proofview.tactic
+val eapply : ?with_classes:bool -> constr -> unit Proofview.tactic
 
 val apply_with_bindings_gen :
-  advanced_flag -> evars_flag -> (clear_flag * constr with_bindings CAst.t) list -> unit Proofview.tactic
+  ?with_classes:bool -> advanced_flag -> evars_flag -> (clear_flag * constr with_bindings CAst.t) list -> unit Proofview.tactic
 
 val apply_with_delayed_bindings_gen :
   advanced_flag -> evars_flag -> (clear_flag * delayed_open_constr_with_bindings CAst.t) list -> unit Proofview.tactic
 
 val apply_with_bindings   : constr with_bindings -> unit Proofview.tactic
-val eapply_with_bindings  : constr with_bindings -> unit Proofview.tactic
+val eapply_with_bindings : ?with_classes:bool -> constr with_bindings -> unit Proofview.tactic
 
 val cut_and_apply         : constr -> unit Proofview.tactic
 
@@ -228,7 +225,7 @@ val apply_in :
 val apply_delayed_in :
   advanced_flag -> evars_flag -> Id.t ->
     (clear_flag * delayed_open_constr_with_bindings CAst.t) list ->
-    intro_pattern option -> unit Proofview.tactic
+    intro_pattern option -> unit Proofview.tactic -> unit Proofview.tactic
 
 (** {6 Elimination tactics. } *)
 
@@ -258,7 +255,6 @@ val apply_delayed_in :
 (** [rel_contexts] and [rel_declaration] actually contain triples, and
    lists are actually in reverse order to fit [compose_prod]. *)
 type elim_scheme = {
-  elimc: constr with_bindings option;
   elimt: types;
   indref: GlobRef.t option;
   params: rel_context;      (** (prm1,tprm1);(prm2,tprm2)...(prmp,tprmp) *)
@@ -277,19 +273,10 @@ type elim_scheme = {
   farg_in_concl: bool;        (** true if (f...) appears at the end of conclusion *)
 }
 
-val compute_elim_sig : evar_map -> ?elimc:constr with_bindings -> types -> elim_scheme
-
-(** elim principle with the index of its inductive arg *)
-type eliminator = {
-  elimindex : int option;  (** None = find it automatically *)
-  elimbody : constr with_bindings
-}
-
-val general_elim  : evars_flag -> clear_flag ->
-  constr with_bindings -> eliminator -> unit Proofview.tactic
+val compute_elim_sig : evar_map -> types -> elim_scheme
 
 val general_elim_clause : evars_flag -> unify_flags -> Id.t option ->
-  clausenv -> eliminator -> unit Proofview.tactic
+  (EConstr.t * EConstr.types) -> EConstr.t -> unit Proofview.tactic
 
 val default_elim  : evars_flag -> clear_flag -> constr with_bindings ->
   unit Proofview.tactic
@@ -458,13 +445,9 @@ end
 
 (** {6 Tacticals defined directly in term of Proofview} *)
 
-module New : sig
+val refine : typecheck:bool -> (evar_map -> evar_map * constr) -> unit Proofview.tactic
+(** [refine ~typecheck c] is [Refine.refine ~typecheck c]
+    followed by beta-iota-reduction of the conclusion. *)
 
-  val refine : typecheck:bool -> (evar_map -> evar_map * constr) -> unit Proofview.tactic
-  (** [refine ~typecheck c] is [Refine.refine ~typecheck c]
-      followed by beta-iota-reduction of the conclusion. *)
-
-  val reduce_after_refine : unit Proofview.tactic
-  (** The reducing tactic called after {!refine}. *)
-
-end
+val reduce_after_refine : unit Proofview.tactic
+(** The reducing tactic called after {!refine}. *)
