@@ -8,19 +8,22 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
+[@@@ocaml.warning "+a"]
+
 (**********************************)
 (**  Configuration script for Coq *)
 (**********************************)
 open Printf
 open Conf
 open Util
-open CmdArgs
+(* cprintf is shadowed here... *)
+open! CmdArgs
 open CmdArgs.Prefs
 
 let (/) = Filename.concat
 
-let coq_version = "8.15.2"
-let vo_magic = 81500
+let coq_version = "8.17.1"
+let vo_magic = 81700
 let is_a_released_version = true
 
 (** Default OCaml binaries *)
@@ -30,14 +33,6 @@ let camlexec = { find = "ocamlfind" }
 
 let reset_caml_find c o = c.find <- o
 
-let coq_annot_flag prefs = if prefs.annot then "-annot" else ""
-let coq_bin_annot_flag prefs = if prefs.bin_annot then "-bin-annot" else ""
-
-(* This variable can be overridden only for debug purposes, use with
-   care. *)
-let coq_safe_string = "-safe-string"
-let coq_strict_sequence = "-strict-sequence"
-
 (* Query for the architecture *)
 let arch prefs = arch prefs.arch
 
@@ -84,7 +79,7 @@ module CamlConf = struct
     }
 end
 
-let resolve_caml prefs =
+let resolve_caml () =
   let () =
     try reset_caml_find camlexec (which camlexec.find)
     with Not_found ->
@@ -102,29 +97,33 @@ let resolve_caml prefs =
     { CamlConf.camlbin; caml_version; camllib; findlib_version }
 
 (** Caml version as a list of ints [4;0;1] *)
-let caml_version_nums { CamlConf.caml_version } =
+let caml_version_nums { CamlConf.caml_version; _ } =
   generic_version_nums ~name:"the OCaml compiler" caml_version
 
 let check_caml_version prefs caml_version caml_version_nums =
-  if caml_version_nums >= [4;5;0] then
+  if caml_version_nums >= [5;0;0] && prefs.nativecompiler <> NativeNo then
+    let () = cprintf prefs "Your version of OCaml is %s." caml_version in
+    die "You have enabled Coq's native compiler, however it is not compatible with OCaml >= 5.0.0"
+  else if caml_version_nums >= [4;9;0] then
     cprintf prefs "You have OCaml %s. Good!" caml_version
   else
     let () = cprintf prefs "Your version of OCaml is %s." caml_version in
-    die "You need OCaml 4.05.0 or later."
+    die "You need OCaml 4.09.0 or later."
 
-let check_findlib_version prefs { CamlConf.findlib_version } =
+let check_findlib_version prefs { CamlConf.findlib_version; _ } =
   let findlib_version_nums = generic_version_nums ~name:"findlib" findlib_version in
-  if findlib_version_nums >= [1;8;0] then
+  if findlib_version_nums >= [1;8;1] then
     cprintf prefs "You have OCamlfind %s. Good!" findlib_version
   else
     let () = cprintf prefs "Your version of OCamlfind is %s." findlib_version in
-    die "You need OCamlfind 1.8.0 or later."
+    die "You need OCamlfind 1.8.1 or later."
 
 (** Note, these warnings are only used in Coq Makefile *)
 (** Explanation of enabled/disabled warnings:
     4: fragile pattern matching: too common in the code and too annoying to avoid in general
     9: missing fields in a record pattern: too common in the code and not worth the bother
     27: innocuous unused variable: innocuous
+    40: constructor or label name used out of scope: gets in the way of putting types in modules
     41: ambiguous constructor or label: too common
     42: disambiguated counstructor or label: too common
     44: "open" shadowing already defined identifier: too common, especially when some are aliases
@@ -137,15 +136,15 @@ let check_findlib_version prefs { CamlConf.findlib_version } =
 *)
 
 (* Note, we list all warnings to be complete *)
-let coq_warnings = "-w -a+1..3-4+5..8-9+10..26-27+28..40-41-42+43-44-45+46..47-48+49..57-58+59..66-67-68+69-70"
+let coq_warnings = "-w -a+1..3-4+5..8-9+10..26-27+28..39-40-41-42+43-44-45+46..47-48+49..57-58+59..66-67-68+69-70"
 let coq_warn_error prefs =
     if prefs.warn_error
     then "-warn-error +a"
     else ""
 
 (* Flags used to compile Coq and plugins (via coq_makefile) *)
-let caml_flags coq_annot_flag coq_bin_annot_flag =
-  Printf.sprintf "-thread -rectypes %s %s %s %s %s" coq_warnings coq_annot_flag coq_bin_annot_flag coq_safe_string coq_strict_sequence
+let caml_flags =
+  Printf.sprintf "-thread -rectypes -bin-annot -strict-sequence %s" coq_warnings
 
 (* Flags used to compile Coq but _not_ plugins (via coq_makefile) *)
 let coq_caml_flags = coq_warn_error
@@ -165,14 +164,13 @@ let msg_no_dynlink_cmxa prefs =
   cprintf prefs "and then run ./configure -natdynlink no"
 
 let check_native prefs camlenv =
-  let () = if prefs.byteonly then raise Not_found in
   let version, _ = tryrun camlexec.find ["opt";"-version"] in
   if version = "" then let () = msg_no_ocamlopt () in raise Not_found
   else if fst (tryrun camlexec.find ["query";"dynlink"]) = ""
   then let () = msg_no_dynlink_cmxa prefs in raise Not_found
   else
     let () =
-      let { CamlConf.caml_version } = camlenv in
+      let { CamlConf.caml_version; _ } = camlenv in
       if version <> caml_version then
         warn "Native and bytecode compilers do not have the same version!"
     in cprintf prefs "You have native-code compilation. Good!"
@@ -186,25 +184,6 @@ let hasnatdynlink prefs best_compiler = prefs.natdynlink && best_compiler = "opt
 
 (** * OS dependent libraries *)
 
-(** Check for dune *)
-
-let dune_install_warning () =
-  warn "You are using Dune < 2.9, the install procedure will not respect the -docdir and -configdir configure directives; please see dev/doc/INSTALL.make.md for more information"
-
-(* returns true if dune >= 2.9 *)
-let check_for_dune_29 () =
-  let dune_version, _  = tryrun "dune" ["--version"] in
-  let dune_version = generic_version_nums ~name:"dune" dune_version in
-  match dune_version with
-  (* Development version, consider it >= 2.9 *)
-  | [] -> true
-  | _ ->
-    if dune_version < [2;9;0] then
-      (dune_install_warning ();
-       false)
-    else
-      true
-
 (** Zarith library *)
 
 let check_for_zarith prefs =
@@ -218,23 +197,10 @@ let check_for_zarith prefs =
     die "Zarith library installed but no development files found (try installing the -dev package)"
   | _   ->
     let zarith_version_int = generic_version_nums ~name:"Zarith" zarith_version in
-    if zarith_version_int >= [1;10;0] then
+    if zarith_version_int >= [1;11;0] then
       cprintf prefs "You have the Zarith library %s installed. Good!" zarith_version
     else
-      die ("Zarith version 1.10 is required, you have " ^ zarith_version)
-
-(** * Documentation : do we have latex, hevea, ... *)
-
-let check_sphinx_deps () =
-  ignore (run (which "python3") ["doc/tools/coqrst/checkdeps.py"])
-
-let check_doc () =
-  let err s =
-    die (sprintf "A documentation build was requested, but %s was not found." s);
-  in
-  if not (program_in_path "python3") then err "python3";
-  if not (program_in_path "sphinx-build") then err "sphinx-build";
-  check_sphinx_deps ()
+      die ("Zarith version 1.11 is required, you have " ^ zarith_version)
 
 (** * Installation directories : bindir, libdir, mandir, docdir, etc *)
 
@@ -269,11 +235,11 @@ end
 
 let install prefs =
   [ InstallDir.make "COQPREFIX" "Coq" prefs.prefix (Relative "") (Relative "")
-  ; InstallDir.make "COQLIBINSTALL" "the Coq library" prefs.libdir (Relative "lib") (Relative "lib/coq")
+  ; InstallDir.make "COQLIBINSTALL" "the Coq library" prefs.libdir (Relative "lib/coq") (Relative "lib/coq")
   ; InstallDir.make "CONFIGDIR" "the Coqide configuration files" prefs.configdir (Relative "config") (Absolute "/etc/xdg/coq")
-  ; InstallDir.make "DATADIR" "the Coqide data files" prefs.datadir (Relative "share") (Relative "share/coq")
-  ; InstallDir.make "MANDIR" "the Coq man pages" prefs.mandir (Relative "man") (Relative "share/man")
-  ; InstallDir.make "DOCDIR" "documentation prefix path for all Coq packages" prefs.docdir (Relative "doc") (Relative "share/doc")
+  ; InstallDir.make "DATADIR" "the Coqide data files" prefs.datadir (Relative "share/coq") (Relative "share/coq")
+  ; InstallDir.make "MANDIR" "the Coq man pages" prefs.mandir (Relative "share/man") (Relative "share/man")
+  ; InstallDir.make "DOCDIR" "documentation prefix path for all Coq packages" prefs.docdir (Relative "share/doc") (Relative "share/doc")
   ]
 
 let strip_trailing_slash_if_any p =
@@ -401,8 +367,8 @@ let esc s = if String.contains s ' ' then "\"" ^ s ^ "\"" else s
 let pr_native = function
   | NativeYes -> "yes" | NativeNo -> "no" | NativeOndemand -> "ondemand"
 
-let print_summary prefs arch camlenv best_compiler install_dirs coqide lablgtkdir hasnatdynlink idearchdef browser =
-  let { CamlConf.caml_version; camlbin; camllib } = camlenv in
+let print_summary prefs arch camlenv install_dirs browser =
+  let { CamlConf.caml_version; camlbin; camllib; _ } = camlenv in
   let pr s = printf s in
   pr "\n";
   pr "  Architecture                : %s\n" arch;
@@ -410,44 +376,24 @@ let print_summary prefs arch camlenv best_compiler install_dirs coqide lablgtkdi
   pr "  OCaml version               : %s\n" caml_version;
   pr "  OCaml binaries in           : %s\n" (esc camlbin);
   pr "  OCaml library in            : %s\n" (esc camllib);
-  if best_compiler = "opt" then
-    pr "  Native dynamic link support : %B\n" hasnatdynlink;
-  if coqide <> "no" then
-    pr "  Lablgtk3 library in         : %s\n" (esc lablgtkdir);
-  if idearchdef = "QUARTZ" then
-    pr "  Mac OS integration is on\n";
-  pr "  CoqIDE                      : %s\n" coqide;
-  pr "  Documentation               : %s\n"
-    (if prefs.withdoc then "All" else "None");
   pr "  Web browser                 : %s\n" browser;
   pr "  Coq web site                : %s\n" prefs.coqwebsite;
   pr "  Bytecode VM enabled         : %B\n" prefs.bytecodecompiler;
   pr "  Native Compiler enabled     : %s\n\n" (pr_native prefs.nativecompiler);
-  (pr "  Paths for true installation:\n";
+  (pr "  Paths where installation is expected by Coq Makefile:\n";
    List.iter
-     (fun ((_,msg),(dir,_)) -> pr "  - %s will be copied in %s\n" msg (esc dir))
+     (fun ((_,msg),(dir,_)) -> pr "  - %s is expected in %s\n" msg (esc dir))
      install_dirs);
   pr "\n";
   pr "If anything is wrong above, please restart './configure'.\n\n";
   pr "*Warning* To compile the system for a new architecture\n";
   pr "          don't forget to do a 'make clean' before './configure'.\n"
 
-(** * Build the dev/ocamldebug-coq file *)
-let write_dbg_wrapper camlenv o =
-  let { CamlConf.camlbin } = camlenv in
-  let pr s = fprintf o s in
-  pr "#!/bin/sh\n\n";
-  pr "###### ocamldebug-coq : a wrapper around ocamldebug for Coq ######\n\n";
-  pr "# DO NOT EDIT THIS FILE: automatically generated by ../configure #\n\n";
-  pr "export COQTOP=%S\n" coqsrc;
-  pr "OCAMLDEBUG=%S\n" (camlbin^"/ocamldebug");
-  pr ". $COQTOP/dev/ocamldebug-coq.run\n"
-
 (** * Build the config/coq_config.ml file *)
 
-let write_configml camlenv coqenv caml_flags caml_version_nums arch arch_is_win32 hasnatdynlink browser idearchdef prefs o =
+let write_configml camlenv coqenv caml_flags caml_version_nums arch arch_is_win32 hasnatdynlink browser prefs o =
   let { CoqEnv.coqlib; coqlibsuffix; configdir; configdirsuffix; docdir; docdirsuffix; datadir; datadirsuffix } = coqenv in
-  let { CamlConf.caml_version } = camlenv in
+  let { CamlConf.caml_version; _ } = camlenv in
   let pr s = fprintf o s in
   let pr_s = pr "let %s = %S\n" in
   let pr_b = pr "let %s = %B\n" in
@@ -474,7 +420,6 @@ let write_configml camlenv coqenv caml_flags caml_version_nums arch arch_is_win3
   pr_s "arch" arch;
   pr_b "arch_is_win32" arch_is_win32;
   pr_s "exec_extension" !exe;
-  pr "let gtk_platform = `%s\n" idearchdef;
   pr_b "has_natdynlink" hasnatdynlink;
   pr_i32 "vo_version" vo_magic;
   pr_s "browser" browser;
@@ -482,7 +427,6 @@ let write_configml camlenv coqenv caml_flags caml_version_nums arch arch_is_win3
   pr_s "wwwbugtracker" (prefs.coqwebsite ^ "bugs/");
   pr_s "wwwrefman" (prefs.coqwebsite ^ "distrib/V" ^ coq_version ^ "/refman/");
   pr_s "wwwstdlib" (prefs.coqwebsite ^ "distrib/V" ^ coq_version ^ "/stdlib/");
-  pr_s "localwwwrefman"  ("file:/" ^ docdir ^ "/html/refman");
   pr_b "bytecode_compiler" prefs.bytecodecompiler;
   pr "type native_compiler = NativeOff | NativeOn of { ondemand : bool }\n";
   pr "let native_compiler = %s\n"
@@ -493,6 +437,7 @@ let write_configml camlenv coqenv caml_flags caml_version_nums arch arch_is_win3
   let core_src_dirs = [ "boot"; "config"; "lib"; "clib"; "kernel"; "library";
                         "engine"; "pretyping"; "interp"; "gramlib"; "parsing"; "proofs";
                         "tactics"; "toplevel"; "printing"; "ide"; "stm"; "vernac" ] in
+
   let core_src_dirs = List.fold_left (fun acc core_src_subdir -> acc ^ "  \"" ^ core_src_subdir ^ "\";\n")
                                     ""
                                     core_src_dirs in
@@ -521,7 +466,8 @@ let write_configml camlenv coqenv caml_flags caml_version_nums arch arch_is_win3
 
 (** * Build the config/Makefile file *)
 
-let write_makefile prefs install_dirs best_compiler caml_flags coq_caml_flags coqide arch exe dune_29 o =
+(** This Makefile is only used in the test-suite now, remove eventually. *)
+let write_makefile install_dirs best_compiler caml_flags coq_caml_flags o =
   let pr s = fprintf o s in
   pr "###### config/Makefile : Configuration file for Coq ##############\n";
   pr "#                                                                #\n";
@@ -531,40 +477,19 @@ let write_makefile prefs install_dirs best_compiler caml_flags coq_caml_flags co
   pr "# with the good options (see the file INSTALL).                  #\n";
   pr "#                                                                #\n";
   pr "##################################################################\n\n";
-  pr "#Variable used to detect whether ./configure has run successfully.\n";
-  pr "COQ_CONFIGURED=yes\n\n";
+
+  (* XXX: Not used anymore, or only in the test suite makefile *)
   pr "# Paths for true installation\n";
   List.iter (fun ((v,msg),_) -> pr "# %s: path for %s\n" v msg) install_dirs;
   List.iter (fun ((v,_),(dir,_)) -> pr "%s=%S\n" v dir) install_dirs;
-  pr "\n# Coq version\n";
-  pr "VERSION=%s\n" coq_version;
+
+  (* XXX: Only used in the test suite: BEST OCAMLFIND CAMLFLAGS ARCH *)
   pr "# The best compiler: native (=opt) or bytecode (=byte)\n";
   pr "BEST=%s\n\n" best_compiler;
-  (* Only used in the test suite: OCAMLFIND CAMLFLAGS *)
   pr "# Findlib command\n";
   pr "OCAMLFIND=%S\n" camlexec.find;
   pr "# Caml flags\n";
   pr "CAMLFLAGS=%s %s\n" caml_flags coq_caml_flags;
-  pr "# Your architecture\n";
-  pr "# Can be obtain by UNIX command arch\n";
-  pr "ARCH=%s\n" arch;
-  pr "# executable files extension, currently:\n";
-  pr "#  Unix systems:\n";
-  pr "#  Win32 systems : .exe\n";
-  pr "EXE=%s\n" exe;
-  pr "# the command MKDIR (try to use mkdirhier if you have problems)\n";
-  pr "MKDIR=mkdir -p\n\n";
-  pr "# CoqIDE (no/byte/opt)\n";
-  pr "HASCOQIDE=%s\n" coqide;
-  pr "# Defining REVISION\n";
-  pr "# Option to control compilation and installation of the documentation\n";
-  pr "WITHDOC=%s\n\n" (if prefs.withdoc then "all" else "no");
-  pr "# Option to produce precompiled files for native_compute\n";
-  pr "NATIVECOMPUTE=%s\n" (if prefs.nativecompiler = NativeYes then "-native-compiler yes" else "");
-  pr "COQWARNERROR=%s\n" (if prefs.warn_error then "-w +default" else "");
-  pr "CONFIGURE_DPROFILE=%s\n" prefs.dune_profile;
-  pr "COQ_INSTALL_ENABLED=%b\n" prefs.install_enabled;
-  if dune_29 then pr "DUNE_29_PLUS=yes\n";
   ()
 
 let write_dune_c_flags cflags o =
@@ -581,38 +506,30 @@ let write_configpy o =
 let main () =
   let prefs = CmdArgs.parse_args () in
   Util.debug := prefs.debug;
-  let dune_29 = check_for_dune_29 () in
-  let coq_annot_flag = coq_annot_flag prefs in
-  let coq_bin_annot_flag = coq_bin_annot_flag prefs in
   let arch = arch prefs in
   let arch_is_win32 = arch_is_win32 arch in
   let exe = resolve_binary_suffix arch in
   Util.exe := exe;
   install_precommit_hook prefs;
   let browser = browser prefs arch in
-  let camlenv = resolve_caml prefs in
+  let camlenv = resolve_caml () in
   let caml_version_nums = caml_version_nums camlenv in
   check_caml_version prefs camlenv.CamlConf.caml_version caml_version_nums;
   check_findlib_version prefs camlenv;
   let best_compiler = best_compiler prefs camlenv in
-  let caml_flags = caml_flags coq_annot_flag coq_bin_annot_flag in
   let coq_caml_flags = coq_caml_flags prefs in
   let hasnatdynlink = hasnatdynlink prefs best_compiler in
   check_for_zarith prefs;
-  let coqide, lablgtkdir = Coqide.coqide camlexec.find prefs best_compiler camlenv.CamlConf.camllib in
-  let idearchdef = Coqide.idearchdef camlexec.find prefs coqide arch in
-  (if prefs.withdoc then check_doc ());
   let install_dirs = install_dirs prefs arch in
   let coqenv = resolve_coqenv install_dirs in
   let cflags, sse2_math = compute_cflags () in
   check_fmath sse2_math;
   if prefs.interactive then
-    print_summary prefs arch camlenv best_compiler install_dirs coqide lablgtkdir hasnatdynlink idearchdef browser;
-  write_config_file ~file:"dev/ocamldebug-coq" ~bin:true (write_dbg_wrapper camlenv);
+    print_summary prefs arch camlenv install_dirs browser;
   write_config_file ~file:"config/coq_config.ml"
-    (write_configml camlenv coqenv caml_flags caml_version_nums arch arch_is_win32 hasnatdynlink browser idearchdef prefs);
+    (write_configml camlenv coqenv caml_flags caml_version_nums arch arch_is_win32 hasnatdynlink browser prefs);
   write_config_file ~file:"config/Makefile"
-    (write_makefile prefs install_dirs best_compiler caml_flags coq_caml_flags coqide arch exe dune_29);
+    (write_makefile install_dirs best_compiler caml_flags coq_caml_flags);
   write_config_file ~file:"config/dune.c_flags" (write_dune_c_flags cflags);
   write_config_file ~file:"config/coq_config.py" write_configpy;
   ()
