@@ -35,10 +35,10 @@ let usage_coq_makefile () =
   output_string stderr "Usage summary:\
 \n\
 \ncoq_makefile .... [file.v] ... [file.ml[ig]?] ... [file.ml{lib,pack}]\
-\n  ... [any] ... [-extra[-phony] result dependencies command]\
 \n  ... [-I dir] ... [-R physicalpath logicalpath]\
 \n  ... [-Q physicalpath logicalpath] ... [VARIABLE = value]\
-\n  ...  [-arg opt] ... [-opt|-byte] [-no-install] [-f file] [-o file]\
+\n  ... [-arg opt] ... [-docroot path] [-f file] [-o file]\
+\n  ... [-generate-meta-for-package project-name]\
 \n  [-h] [--help]\
 \n";
   output_string stderr "\
@@ -48,16 +48,6 @@ let usage_coq_makefile () =
 \n[file.ml[ig]?]: Objective Caml file to be compiled\
 \n[file.ml{lib,pack}]: ocamlbuild-style file that describes a Objective Caml\
 \n  library/module\
-\n[any] : subdirectory that should be \"made\" and has a Makefile itself\
-\n  to do so. Very fragile and discouraged.\
-\n[-extra result dependencies command]: add target \"result\" with command\
-\n  \"command\" and dependencies \"dependencies\". If \"result\" is not\
-\n  generic (do not contains a %), \"result\" is built by _make all_ and\
-\n  deleted by _make clean_.\
-\n[-extra-phony result dependencies command]: add a PHONY target \"result\"\
-\n with command \"command\" and dependencies \"dependencies\". Note that\
-\n _-extra-phony foo bar \"\"_ is a regular way to add the target \"bar\" as\
-\n as a dependencies of an already defined target \"foo\".\
 \n[-I dir]: look for Objective Caml dependencies in \"dir\"\
 \n[-R physicalpath logicalpath]: look for Coq dependencies recursively\
 \n  starting from \"physicalpath\". The logical path associated to the\
@@ -66,15 +56,13 @@ let usage_coq_makefile () =
 \n  \"physicalpath\". The logical path associated to the physical path\
 \n  is \"logicalpath\".\
 \n[VARIABLE = value]: Add the variable definition \"VARIABLE=value\"\
-\n[-byte]: compile with byte-code version of coq\
-\n[-opt]: compile with native-code version of coq\
 \n[-arg opt]: send option \"opt\" to coqc\
-\n[-install opt]: where opt is \"user\" to force install into user directory,\
-\n  \"none\" to build a makefile with no install target or\
-\n  \"global\" to force install in $COQLIB directory\
+\n[-docroot path]: Install the documentation in this folder, relative to\
+\n  \"user-contrib\".\
 \n[-f file]: take the contents of file as arguments\
 \n[-o file]: output should go in file file (recommended)\
 \n	Output file outside the current directory is forbidden.\
+\n[-generate-meta-for-package project-name]: generate META.project-name.\
 \n[-h]: print this usage summary\
 \n[--help]: equivalent to [-h]\n";
   exit 1
@@ -120,26 +108,95 @@ let read_whole_file s =
     close_in ic;
     Buffer.contents b
 
+(* Use this for quoting contents of variables which never appears as target or
+ * pattern. *)
+let makefile_quote s =
+  let out = Buffer.create 16 in
+  Buffer.add_string out "'";
+  String.iter
+    (fun c ->
+      match c with
+      | '$' -> Buffer.add_string out "$$"
+      | '#' -> Buffer.add_string out "\\#"
+      | '\'' -> Buffer.add_string out "'\\''"
+      | _ -> Buffer.add_char out c
+    )
+  s;
+  Buffer.add_string out "'";
+  Buffer.contents out
+
 let quote s = if String.contains s ' ' || CString.is_empty s then "'" ^ s ^ "'" else s
 
-let generate_makefile oc conf_file local_file dep_file args project =
-  let coqlib = Envars.coqlib () in
-  let makefile_template =
-    let template = Filename.concat "tools" "CoqMakefile.in" in
-    Filename.concat coqlib template in
+let generate_makefile oc conf_file local_file local_late_file dep_file args project =
+  let env = Boot.Env.init () in
+  (* XX coq makefile should ship files on its own dir *)
+  let cmf_dir = Boot.Env.tool env "" in
+  let makefile_template = Boot.Path.relative cmf_dir "CoqMakefile.in" in
+  if not (Boot.Path.exists makefile_template) then
+    begin
+      let makefile_template = Boot.Path.to_string makefile_template in
+      Format.eprintf "Error: cannot find %s" makefile_template;
+      exit 1
+    end;
+  let makefile_template = Boot.Path.to_string makefile_template in
   let s = read_whole_file makefile_template in
   let s = List.fold_left
     (* We use global_substitute to avoid running into backslash issues due to \1 etc. *)
     (fun s (k,v) -> Str.global_substitute (Str.regexp_string k) (fun _ -> v) s) s
     [ "@CONF_FILE@", conf_file;
       "@LOCAL_FILE@", local_file;
+      "@LOCAL_LATE_FILE@", local_late_file;
       "@DEP_FILE@", dep_file;
       "@COQ_VERSION@", Coq_config.version;
       "@PROJECT_FILE@", (Option.default "" project.project_file);
       "@COQ_MAKEFILE_INVOCATION@",String.concat " " (List.map quote args);
     ] in
   output_string oc s
-;;
+
+let generate_meta_file p =
+  try
+    match p.meta_file with
+    | Absent -> p
+    | Generate proj ->
+        let cmname = List.map (fun { thing } -> thing)
+            (files_by_suffix p.files [".mllib"; ".mlpack"]) in
+        let dir, cmname =
+          match cmname with
+          | [] -> Printf.eprintf "In order to generate a META file one needs an .mlpack or .mllib file\n"; exit 1
+          | [x] -> Filename.dirname x, Filename.(basename @@ chop_extension x)
+          | _ -> Printf.eprintf "Automatic META generation only works for one .mlpack or .mllib file, since you have more you need to write the META file by hand\n"; exit 1 in
+        let f = dir ^ "/META." ^ proj in
+        let oc = open_out f in
+        let meta : _ format = {|
+package "plugin" (
+  directory = "."
+  requires = "coq-core.plugins.ltac"
+  archive(byte) = "%s.cma"
+  archive(native) = "%s.cmxa"
+  plugin(byte) = "%s.cma"
+  plugin(native) = "%s.cmxs"
+)
+directory = "."
+|}
+        in
+        let meta = Printf.sprintf meta cmname cmname cmname cmname in
+        output_string oc meta;
+        close_out oc;
+        { p with meta_file = Present f }
+    | Present f ->
+        let ext = Filename.extension f in
+        if ext = ".in" then
+          let meta_file = Filename.chop_extension f in
+          let oc = open_out meta_file in
+          (* META generation is just a renaming for now, we lack some metadata *)
+          output_string oc (read_whole_file f);
+          close_out oc;
+          { p with meta_file = Present meta_file }
+        else
+          p (* already a META.package file *)
+  with Sys_error e ->
+    Printf.eprintf "Error: %s\n" e;
+    exit 1
 
 let section oc s =
   let pad = String.make (76 - String.length s) ' ' in
@@ -152,37 +209,6 @@ let section oc s =
   fprintf oc "%s\n\n" sharps
 ;;
 
-let clean_tgts = ["clean"; "cleanall"; "archclean"]
-
-let generate_conf_extra_target oc sps =
-  let pr_path { target; dependencies; phony; command } =
-    let target = if target = "all" then "custom-all" else target in
-    if phony then fprintf oc ".PHONY: %s\n" target;
-    if not (is_genrule target) && not phony then begin
-      fprintf oc "post-all::\n\t$(MAKE) -f $(SELF) %s\n" target;
-      if not phony then
-        fprintf oc "clean::\n\trm -f %s\n" target;
-    end;
-    fprintf oc "%s %s %s\n\t%s\n\n"
-       target
-       (if List.mem target clean_tgts then ":: " else ": ")
-       dependencies
-       command
-  in
-    if sps <> [] then
-      section oc "Extra targets. (-extra and -extra-phony, DEPRECATED)";
-    List.iter (forget_source > pr_path) sps
-
-let generate_conf_subdirs oc sds =
-  if sds <> [] then section oc "Subdirectories. (DEPRECATED)";
-  let iter f = List.iter (forget_source > f) in
-  iter (fprintf oc ".PHONY:%s\n") sds;
-  iter (fprintf oc "post-all::\n\tcd \"%s\" && $(MAKE) all\n") sds;
-  iter (fprintf oc "clean::\n\tcd \"%s\" && $(MAKE) clean\n") sds;
-  iter (fprintf oc "archclean::\n\tcd \"%s\" && $(MAKE) archclean\n") sds;
-  iter (fprintf oc "install-extra::\n\tcd \"%s\" && $(MAKE) install\n") sds
-
-
 let generate_conf_includes oc { ml_includes; r_includes; q_includes } =
   section oc "Path directives (-I, -R, -Q).";
   let module S = String in
@@ -214,26 +240,50 @@ let windrive s =
 
 let generate_conf_coq_config oc =
   section oc "Coq configuration.";
-  let src_dirs = Coq_config.all_src_dirs in
-  Envars.print_config ~prefix_var_name:"COQMF_" oc src_dirs;
-  fprintf oc "COQMF_WINDRIVE=%s\n" (windrive Coq_config.coqlib)
+  Envars.print_config ~prefix_var_name:"COQMF_" oc;
+  let env = Boot.Env.init () in
+  let coqlib = Boot.Env.(coqlib env |> Path.to_string) in
+  (* XXX: FIXME, why does this variable needs the root lib *)
+  fprintf oc "COQMF_WINDRIVE=%s\n" (windrive coqlib)
 ;;
 
-let generate_conf_files oc
-  { v_files; mli_files; mlg_files; ml_files; mllib_files; mlpack_files; }
+let check_metafile p =
+  if files_by_suffix p.files [".mlpack"; ".mllib"] <> [] && p.meta_file = Absent then begin
+    eprintf "Warning: it is recommended you provide a META.package-name file\n";
+    eprintf "Warning: since you build plugins. See also -generate-meta-for-package.\n";
+  end
+
+let write_coqbin oc =
+  fprintf oc "%s\n"
+    "COQBIN?=\n\
+    ifneq (,$(COQBIN))\n\
+    # add an ending /\n\
+    COQBIN:=$(COQBIN)/\n\
+    endif\n\
+    COQMKFILE ?= \"$(COQBIN)coq_makefile\""
+
+let generate_conf_files oc p
 =
   let module S = String in
-  let map = map_sourced_list in
+  let fout varname suffix =
+    fprintf oc "COQMF_%s := $(filter %%%s, $(COQMF_SOURCES))\n" varname suffix;
+  in
   section oc "Project files.";
-  fprintf oc "COQMF_VFILES = %s\n"      (S.concat " " (map quote v_files));
-  fprintf oc "COQMF_MLIFILES = %s\n"    (S.concat " " (map quote mli_files));
-  fprintf oc "COQMF_MLFILES = %s\n"     (S.concat " " (map quote ml_files));
-  fprintf oc "COQMF_MLGFILES = %s\n"    (S.concat " " (map quote mlg_files));
-  fprintf oc "COQMF_MLPACKFILES = %s\n" (S.concat " " (map quote mlpack_files));
-  fprintf oc "COQMF_MLLIBFILES = %s\n"  (S.concat " " (map quote mllib_files));
-  let cmdline_vfiles = filter_cmdline v_files in
-  fprintf oc "COQMF_CMDLINE_VFILES = %s\n" (S.concat " " (List.map quote cmdline_vfiles));
-;;
+
+  let cmdline_vfiles = p.cmd_line_files in
+  fprintf oc "COQMF_CMDLINE_VFILES := %s\n" (S.concat " " (map_sourced_list quote cmdline_vfiles));
+  let proj_arg = match p.project_file with
+    | Some pfile -> Printf.sprintf "-f %s" pfile
+    | None -> ""
+  in
+  fprintf oc "COQMF_SOURCES := $(shell $(COQMKFILE) -sources-of %s $(COQMF_CMDLINE_VFILES))\n" proj_arg;
+  fout "VFILES" ".v";
+  fout "MLIFILES" ".mli";
+  fout "MLFILES" ".ml";
+  fout "MLGFILES" ".mlg";
+  fout "MLPACKFILES" ".mlpack";
+  fout "MLLIBFILES" ".mllib";
+  fprintf oc "COQMF_METAFILE = %s\n"  (match p.meta_file with Present x -> x | _ -> "")
 
 let rec all_start_with prefix = function
   | [] -> true
@@ -248,45 +298,55 @@ let rec logic_gcd acc = function
       then logic_gcd (acc @ [hd]) (tl :: List.map List.tl rest)
       else acc
 
-let generate_conf_doc oc { defs; q_includes; r_includes } =
+let generate_conf_doc oc { docroot; q_includes; r_includes } =
   let includes = List.map (forget_source > snd) (q_includes @ r_includes) in
   let logpaths = List.map (String.split_on_char '.') includes in
   let gcd = logic_gcd [] logpaths in
   let root =
-    if gcd = [] then
-      if not (List.exists (fun x -> fst x.thing = "INSTALLDEFAULTROOT") defs) then begin
+    match docroot with
+    | None ->
+      if gcd = [] then
          let destination = "orphan_" ^ (String.concat "_" includes) in
-         eprintf "Warning: no common logical root\n";
-         eprintf "Warning: in such case INSTALLDEFAULTROOT must be defined\n";
-         eprintf "Warning: the install-doc target is going to install files\n";
+         eprintf "Warning: No common logical root.\n";
+         eprintf "Warning: In this case the -docroot option should be given.\n";
+         eprintf "Warning: Otherwise the install-doc target is going to install files\n";
          eprintf "Warning: in %s\n" destination;
          destination
-      end else "$(INSTALLDEFAULTROOT)"
-    else String.concat Filename.dir_sep gcd in
+      else String.concat Filename.dir_sep gcd
+    | Some p -> p in
   Printf.fprintf oc "COQMF_INSTALLCOQDOCROOT = %s\n" (quote root)
 
+let generate_conf_native oc native_compiler =
+  section oc "Native compiler.";
+  let flag = match native_compiler with
+  | None -> ""
+  | Some NativeYes -> "yes"
+  | Some NativeNo -> "no"
+  | Some NativeOndemand -> "ondemand"
+  in
+  Printf.fprintf oc "COQMF_COQPROJECTNATIVEFLAG = %s\n" flag
+
+
 let generate_conf_defs oc { defs; extra_args } =
   section oc "Extra variables.";
   List.iter (forget_source > (fun (k,v) -> Printf.fprintf oc "%s = %s\n" k v)) defs;
   Printf.fprintf oc "COQMF_OTHERFLAGS = %s\n"
-    (String.concat " " (List.map forget_source extra_args))
+    (String.concat " " (List.map (forget_source > makefile_quote) extra_args))
 
 let generate_conf oc project args  =
   fprintf oc "# This configuration file was generated by running:\n";
   fprintf oc "# %s\n\n" (String.concat " " (List.map quote args));
+  write_coqbin oc;
   generate_conf_files oc project;
   generate_conf_includes oc project;
   generate_conf_coq_config oc;
+  generate_conf_native oc project.native_compiler;
   generate_conf_defs oc project;
   generate_conf_doc oc project;
-  generate_conf_extra_target oc project.extra_targets;
-  generate_conf_subdirs oc project.subdirs;
 ;;
 
 let ensure_root_dir
-  ({ ml_includes; r_includes; q_includes;
-     v_files; ml_files; mli_files; mlg_files;
-     mllib_files; mlpack_files } as project)
+  ({ ml_includes; r_includes; q_includes; files } as project)
   =
   let exists f = List.exists (forget_source > f) in
   let here = Sys.getcwd () in
@@ -294,9 +354,7 @@ let ensure_root_dir
   if exists (fun { canonical_path = x } -> x = here) ml_includes
   || exists (fun ({ canonical_path = x },_) -> is_prefix x here) r_includes
   || exists (fun ({ canonical_path = x },_) -> is_prefix x here) q_includes
-  || (not_tops v_files &&
-      not_tops mli_files && not_tops mlg_files &&
-      not_tops ml_files && not_tops mllib_files && not_tops mlpack_files)
+  || not_tops files
   then
     project
   else
@@ -307,24 +365,6 @@ let ensure_root_dir
         r_includes = source (here_path, "Top") :: r_includes }
 ;;
 
-let warn_install_at_root_directory
-  ({ q_includes; r_includes; } as project)
-=
-  let open CList in
-  let inc_top_p =
-    map_filter
-      (fun {thing=({ path } ,ldir)} -> if ldir = "" then Some path else None)
-      (r_includes @ q_includes) in
-  let files = all_files project in
-  let bad = filter (fun f -> mem (Filename.dirname f.thing) inc_top_p) files in
-  if bad <> [] then begin
-    eprintf "Warning: No file should be installed at the root of Coq's library.\n";
-    eprintf "Warning: No logical path (-R, -Q) applies to these files:\n";
-    List.iter (fun x -> eprintf "Warning:  %s\n" x.thing) bad;
-    eprintf "\n";
-  end
-;;
-
 let check_overlapping_include { q_includes; r_includes } =
   let pwd = Sys.getcwd () in
   let aux = function
@@ -340,6 +380,14 @@ let check_overlapping_include { q_includes; r_includes } =
     aux (q_includes @ r_includes)
 ;;
 
+let check_native_compiler = function
+| None -> ()
+| Some flag ->
+  match Coq_config.native_compiler, flag with
+  | Coq_config.NativeOff, (NativeYes | NativeOndemand) ->
+    eprintf "Warning: native compilation is globally deactivated by the configure flag\n"
+  | _ -> ()
+
 let chop_prefix p f =
   let len_p = String.length p in
   let len_f = String.length f in
@@ -396,6 +444,11 @@ let _ =
     | "-destination-of" :: tgt :: rest -> Some tgt, rest
     | _ -> None, args in
 
+  (* -sources-of and -destination-of must be the first parameter *)
+  let only_sources, args = match args with
+    | "-sources-of" :: rest -> true, rest
+    | _ -> false, args in
+
   let project =
     let warning_fn x = Format.eprintf "%s@\n%!" x in
     try cmdline_args_to_project ~warning_fn ~curdir:Filename.current_dir_name args
@@ -406,6 +459,12 @@ let _ =
     exit 0
   end;
 
+  if only_sources then begin
+    let paths = String.concat " " (List.map (fun i -> i.thing) project.files) in
+    Printf.printf "%s" paths;
+    exit 0
+  end;
+
   if project.makefile = None then
     eprintf "Warning: Omitting -o is deprecated\n\n";
     (* We want to know the name of the Makefile (say m) in order to
@@ -413,33 +472,23 @@ let _ =
 
   let conf_file = Option.default "CoqMakefile" project.makefile ^ ".conf" in
   let local_file = Option.default "CoqMakefile" project.makefile ^ ".local" in
+  let local_late_file = Option.default "CoqMakefile" project.makefile ^ ".local-late" in
   let dep_file = "." ^ Option.default "CoqMakefile" project.makefile ^ ".d" in
 
-  if project.extra_targets <> [] then begin
-    eprintf "Warning: -extra and -extra-phony are deprecated.\n";
-    eprintf "Warning: Write the extra targets in %s.\n\n" local_file;
-  end;
-
-  if project.subdirs <> [] then begin
-    eprintf "Warning: Subdirectories are deprecated.\n";
-    eprintf "Warning: Use double colon rules in %s.\n\n" local_file;
-  end;
-
   let project = ensure_root_dir project in
 
-  if project.install_kind <> (Some CoqProject_file.NoInstall) then begin
-    warn_install_at_root_directory project;
-  end;
-
   check_overlapping_include project;
 
-  Envars.set_coqlib ~fail:(fun x -> Printf.eprintf "Error: %s\n" x; exit 1);
+  check_native_compiler project.native_compiler;
+
+  check_metafile project;
+
+  let project = generate_meta_file project in
 
   let ocm = Option.cata open_out stdout project.makefile in
-  generate_makefile ocm conf_file local_file dep_file (prog :: args) project;
+  generate_makefile ocm conf_file local_file local_late_file dep_file (prog :: args) project;
   close_out ocm;
   let occ = open_out conf_file in
   generate_conf occ project (prog :: args);
   close_out occ;
   exit 0
-
