@@ -42,13 +42,13 @@ let is_keyword =
       "Mutual"; "Parameter"; "Parameters"; "Print"; "Printing"; "All"; "Proof"; "Proof with"; "Qed";
       "Record"; "Recursive"; "Remark"; "Require"; "Save"; "Scheme"; "Assumptions"; "Axioms"; "Universes";
       "Induction"; "for"; "Sort"; "Section"; "Show"; "Structure"; "Syntactic"; "Syntax"; "Tactic"; "Theorem";
-      "Search"; "SearchHead"; "SearchPattern"; "SearchRewrite";
+      "Search"; "SearchPattern"; "SearchRewrite";
       "Set"; "Types"; "Undo"; "Unset"; "Variable"; "Variables"; "Context";
-      "Notation"; "Reserved Notation"; "Tactic Notation";
+      "Notation"; "Reserved Notation"; "Tactic Notation"; "Number Notation"; "String Notation"; "Enable Notation"; "Disable Notation";
       "Delimit"; "Bind"; "Open"; "Scope"; "Inline";
       "Implicit Arguments"; "Add"; "Strict";
       "Typeclasses"; "Instance"; "Global Instance"; "Class"; "Instantiation";
-      "subgoal"; "subgoals"; "vm_compute";
+      "goal"; "goals"; "vm_compute";
       "Opaque"; "Transparent"; "Time";
       "Extraction"; "Extract";
       "Variant";
@@ -74,7 +74,7 @@ let is_tactic =
       "info"; "field"; "specialize"; "evar"; "solve"; "instantiate"; "info_auto"; "info_eauto";
       "quote"; "eexact"; "autorewrite";
       "destruct"; "destruction"; "destruct_call"; "dependent"; "elim"; "extensionality";
-      "f_equal"; "generalize"; "generalize_eqs"; "generalize_eqs_vars"; "induction"; "rename"; "move"; "omega";
+      "f_equal"; "generalize"; "generalize_eqs"; "generalize_eqs_vars"; "induction"; "rename"; "move";
       "set"; "assert"; "do"; "repeat";
       "cut"; "assumption"; "exact"; "split"; "subst"; "try"; "discriminate";
       "simpl"; "unfold"; "red"; "compute"; "at"; "in"; "by";
@@ -313,7 +313,7 @@ module Latex = struct
 
   let start_verbatim inline =
     if inline then printf "\\texttt{"
-    else printf "\\begin{verbatim}"
+    else printf "\\begin{verbatim}\n"
 
   let stop_verbatim inline =
     if inline then printf "}"
@@ -479,10 +479,6 @@ module Latex = struct
 
   let end_coq () = printf "\\end{coqdoccode}\n"
 
-  let start_code () = end_doc (); start_coq ()
-
-  let end_code () = end_coq (); start_doc ()
-
   let section_kind = function
     | 1 -> "\\section{"
     | 2 -> "\\subsection{"
@@ -632,11 +628,11 @@ module Html = struct
   let stop_quote () = start_quote ()
 
   let start_verbatim inline =
-    if inline then printf "<tt>"
-    else printf "<pre>"
+    if inline then printf "<code>"
+    else printf "<pre>\n"
 
   let stop_verbatim inline =
-    if inline then printf "</tt>"
+    if inline then printf "</code>"
     else printf "</pre>\n"
 
   let url addr name =
@@ -738,7 +734,7 @@ module Html = struct
 
   let end_doc () = in_doc := false;
     stop_item ();
-    if not !raw_comments then printf "\n</div>\n"
+    if not !raw_comments then printf "</div>\n"
 
   let start_emph () = printf "<i>"
 
@@ -754,10 +750,6 @@ module Html = struct
 
   let end_comment () = printf "*)</span>"
 
-  let start_code () = end_doc (); start_coq ()
-
-  let end_code () = end_coq (); start_doc ()
-
   let start_inline_coq () =
     if !inline_notmono then printf "<span class=\"inlinecodenm\">"
                        else printf "<span class=\"inlinecode\">"
@@ -791,7 +783,7 @@ module Html = struct
            printf "  <td class=\"infrule\">%s</td>\n" (replace_spaces line)) in
     let end_assumption () =
           (printf "  <td></td>\n";
-           printf "</td>\n") in
+           printf "</tr>\n") in
     let rec print_assumptions hyps =
           match hyps with
           | []                 -> start_assumption "&nbsp;&nbsp;"
@@ -838,7 +830,7 @@ module Html = struct
       printf "<a id=\"%s_%c\"></a><h2>%s %s</h2>\n" idx c (display_letter c) cat;
       List.iter
         (fun (id,(text,link,t)) ->
-           let id' = prepare_entry id t in
+           let id' = escaped (prepare_entry id t) in
            printf "<a href=\"%s\">%s</a> %s<br/>\n" link id' text) l;
       printf "<br/><br/>"
     end
@@ -1069,9 +1061,6 @@ module TeXmacs = struct
   let start_comment () = ()
   let end_comment () = ()
 
-  let start_code () = in_doc := true; printf "<\\code>\n"
-  let end_code () = in_doc := false; printf "\n</code>"
-
   let section_kind = function
     | 1 -> "section"
     | 2 -> "subsection"
@@ -1181,9 +1170,6 @@ module Raw = struct
   let start_coq () = ()
   let end_coq () = ()
 
-  let start_code () = end_doc (); start_coq ()
-  let end_code () = end_coq (); start_doc ()
-
   let section_kind =
     function
       | 1 -> "* "
@@ -1240,9 +1226,6 @@ let end_comment = select Latex.end_comment Html.end_comment TeXmacs.end_comment
 let start_coq = select Latex.start_coq Html.start_coq TeXmacs.start_coq Raw.start_coq
 let end_coq = select Latex.end_coq Html.end_coq TeXmacs.end_coq Raw.end_coq
 
-let start_code = select Latex.start_code Html.start_code TeXmacs.start_code Raw.start_code
-let end_code = select Latex.end_code Html.end_code TeXmacs.end_code Raw.end_code
-
 let start_inline_coq =
   select Latex.start_inline_coq Html.start_inline_coq TeXmacs.start_inline_coq Raw.start_inline_coq
 let end_inline_coq =
