@@ -61,7 +61,7 @@ let interp_vernac ~check ~interactive ~state ({CAst.loc;_} as com) =
       let doc, nsid, ntip = Stm.add ~doc:state.doc ~ontop:state.sid (not !Flags.quiet) com in
 
       (* Main STM interaction *)
-      if ntip <> `NewTip then
+      if ntip <> Stm.NewAddTip then
         anomaly (str "vernac.ml: We got an unfocus operation on the toplevel!");
 
       (* Force the command  *)
@@ -72,7 +72,15 @@ let interp_vernac ~check ~interactive ~state ({CAst.loc;_} as com) =
       let (reraise, info) = Exninfo.capture reraise in
       (* XXX: In non-interactive mode edit_at seems to do very weird
          things, so we better avoid it while we investigate *)
-      if interactive then ignore(Stm.edit_at ~doc:state.doc state.sid);
+      let reraise = if interactive then begin
+          (* Exceptions don't carry enough state to print themselves (typically missing the nametab)
+             so we need to print before resetting to an older state. See eg #16745 *)
+          let reraise = UserError (CErrors.print reraise) in
+          ignore(Stm.edit_at ~doc:state.doc state.sid);
+          reraise
+        end
+        else reraise
+      in
       let info = begin
         match Loc.get_loc info with
         | None   -> Option.cata (Loc.add_loc info) info loc
@@ -81,15 +89,16 @@ let interp_vernac ~check ~interactive ~state ({CAst.loc;_} as com) =
       Exninfo.iraise (reraise, info)
 
 (* Load a vernac file. CErrors are annotated with file and location *)
-let load_vernac_core ~echo ~check ~interactive ~state file =
+let load_vernac_core ~echo ~check ~interactive ~state ?ldir file =
   (* Keep in sync *)
   let in_chan = open_utf8_file_in file in
   let in_echo = if echo then Some (open_utf8_file_in file) else None in
   let input_cleanup () = close_in in_chan; Option.iter close_in in_echo in
 
-  let in_pa =
-    Pcoq.Parsable.make ~loc:(Loc.initial (Loc.InFile file))
-      (Stream.of_channel in_chan) in
+  let dirpath = Option.cata (fun ldir -> Some Names.DirPath.(to_string ldir))
+      None ldir in
+  let in_pa = Pcoq.Parsable.make ~loc:Loc.(initial (InFile {dirpath; file}))
+      (Gramlib.Stream.of_channel in_chan) in
   let open State in
 
   (* ids = For beautify, list of parsed sids *)
@@ -110,7 +119,7 @@ let load_vernac_core ~echo ~check ~interactive ~state file =
       let state =
         Flags.silently (interp_vernac ~check ~interactive ~state) ast in
 
-      loop state (state.sid :: ids)
+      (loop [@ocaml.tailcall]) state (state.sid :: ids)
   in
   try loop state []
   with any ->   (* whatever the exception *)
@@ -167,8 +176,8 @@ let beautify_pass ~doc ~comments ~ids ~filename =
 
 (* Main driver for file loading. For now, we only do one beautify
    pass. *)
-let load_vernac ~echo ~check ~interactive ~state filename =
-  let ostate, ids, comments = load_vernac_core ~echo ~check ~interactive ~state filename in
+let load_vernac ~echo ~check ~interactive ~state ?ldir filename =
+  let ostate, ids, comments = load_vernac_core ~echo ~check ~interactive ~state ?ldir filename in
   (* Pass for beautify *)
   if !Flags.beautify then beautify_pass ~doc:ostate.State.doc ~comments ~ids:(List.rev ids) ~filename;
   (* End pass *)
