@@ -9,7 +9,6 @@
 (************************************************************************)
 
 (*i*)
-module CVars = Vars
 open Names
 open EConstr
 open CErrors
@@ -25,105 +24,165 @@ open Class_tactics
 open Libobject
 
 module RelDecl = Context.Rel.Declaration
-module NamedDecl = Context.Named.Declaration
 (*i*)
 
-let set_typeclass_transparency c local b =
-  let locality = if local then Goptions.OptLocal else Goptions.OptGlobal in
+let set_typeclass_transparency ~locality c b =
   Hints.add_hints ~locality [typeclasses_db]
-    (Hints.HintsTransparencyEntry (Hints.HintsReferences [c], b))
+    (Hints.HintsTransparencyEntry (Hints.HintsReferences c, b))
 
-let classes_transparent_state () =
-  try
-    Hints.Hint_db.transparent_state (Hints.searchtable_map typeclasses_db)
-  with Not_found -> TransparentState.empty
+let warn_deprecated_tc_transparency_without_locality =
+  CWarnings.create ~name:"deprecated-typeclasses-transparency-without-locality" ~category:"deprecated"
+    Pp.(fun () -> strbrk
+  "The default value for Typeclasses Opaque and Typeclasses \
+   Transparent locality is currently \"local\" in a section and \
+   \"global\" otherwise, but is scheduled to change in a future \
+   release. For the time being, adding typeclass transparency hints outside of \
+   sections without specifying an explicit locality attribute is \
+   therefore deprecated. It is recommended to use \"export\" whenever \
+   possible. Use the attributes #[local], #[global] and #[export] \
+   depending on your choice. For example: \"#[export] Typeclasses Transparent foo.\". \
+   This is supported since Coq 8.15.")
 
-let () =
-  Hook.set Typeclasses.classes_transparent_state_hook classes_transparent_state
+let default_tc_transparency_locality () =
+  if Global.sections_are_opened () then Hints.Local
+  else
+    let () = warn_deprecated_tc_transparency_without_locality () in
+    Hints.SuperGlobal
 
-let add_instance_hint inst path ~locality info poly =
+let tc_transparency_locality = Attributes.hint_locality ~default:default_tc_transparency_locality
+
+let set_typeclass_transparency_com ~locality refs b =
+  let refs = List.map
+      (fun x -> Tacred.evaluable_of_global_reference
+          (Global.env ())
+          (Smartlocate.global_with_alias x))
+      refs
+  in
+  set_typeclass_transparency ~locality refs b
+
+let add_instance_hint inst path ~locality info =
      Flags.silently (fun () ->
        Hints.add_hints ~locality [typeclasses_db]
           (Hints.HintsResolveEntry
-             [info, poly, false, Hints.PathHints path, inst])) ()
+             [info, false, Hints.PathHints path, inst])) ()
 
-let is_local_for_hint i =
-  match i.is_global with
-  | None -> true  (* i.e. either no Global keyword not in section, or in section *)
-  | Some n -> n <> 0 (* i.e. in a section, declare the hint as local
-                        since discharge is managed by rebuild_instance which calls again
-                        add_instance_hint; don't ask hints to take discharge into account
-                        itself *)
+(* short names without opening all Hints *)
+type locality = Hints.hint_locality = Local | Export | SuperGlobal
+
+type instance_obj = {
+  inst_class : GlobRef.t;
+  inst_info: hint_info;
+  (* Sections where the instance should be redeclared,
+     None for discard, Some 0 for none. *)
+  inst_global: Hints.hint_locality;
+  inst_impl: GlobRef.t;
+}
 
 let add_instance_base inst =
-  let poly = Global.is_polymorphic inst.is_impl in
-  let locality = if is_local_for_hint inst then Goptions.OptLocal else Goptions.OptGlobal in
-  add_instance_hint (Hints.IsGlobRef inst.is_impl) [inst.is_impl] ~locality
-    inst.is_info poly;
-  List.iter (fun (path, pri, c) ->
-    let h = Hints.IsConstr (EConstr.of_constr c, Univ.ContextSet.empty) [@ocaml.warning "-3"] in
-    add_instance_hint h path
-                ~locality pri poly)
-    (build_subclasses ~check:(not (isVarRef inst.is_impl))
-       (Global.env ()) (Evd.from_env (Global.env ())) inst.is_impl inst.is_info)
-
-let mk_instance cl info glob impl =
-  let global =
-    if glob then Some (Lib.sections_depth ())
-    else None
+  let locality = match inst.inst_global with
+  | Local -> Local
+  | SuperGlobal ->
+    (* i.e. in a section, declare the hint as local since discharge is managed
+       by rebuild_instance which calls again add_instance_hint; don't ask hints
+       to take discharge into account itself *)
+    if Global.sections_are_opened () then Local
+    else SuperGlobal
+  | Export ->
+    (* Same as above for export *)
+    if Global.sections_are_opened () then Local
+    else Export
   in
-  if match global with Some n -> n>0 && isVarRef impl | _ -> false then
-    CErrors.user_err (Pp.str "Cannot set Global an instance referring to a section variable.");
-  { is_class = cl.cl_impl;
-    is_info = info ;
-    is_global = global ;
-    is_impl = impl }
+  add_instance_hint (Hints.hint_globref inst.inst_impl) [inst.inst_impl] ~locality
+    inst.inst_info
 
 (*
  * instances persistent object
  *)
-let cache_instance (_, i) =
-  load_instance i
+let perform_instance i =
+  let i = { is_class = i.inst_class; is_info = i.inst_info; is_impl = i.inst_impl } in
+  Typeclasses.load_instance i
+
+let cache_instance inst = perform_instance inst
+
+let load_instance _ inst = match inst.inst_global with
+| Local -> assert false
+| SuperGlobal -> perform_instance inst
+| Export -> ()
+
+let open_instance i inst = match inst.inst_global with
+| Local -> assert false
+| SuperGlobal -> perform_instance inst
+| Export -> if Int.equal i 1 then perform_instance inst
 
 let subst_instance (subst, inst) =
   { inst with
-      is_class = fst (subst_global subst inst.is_class);
-      is_impl = fst (subst_global subst inst.is_impl) }
+      inst_class = fst (subst_global subst inst.inst_class);
+      inst_impl = fst (subst_global subst inst.inst_impl) }
 
-let discharge_instance (_, inst) =
-  match inst.is_global with
-  | None -> None
-  | Some n ->
-    assert (not (isVarRef inst.is_impl));
-    Some
-    { inst with
-      is_global = Some (pred n);
-      is_class = inst.is_class;
-      is_impl = inst.is_impl }
-
-let is_local i = (i.is_global == None)
+let discharge_instance inst =
+  match inst.inst_global with
+  | Local -> None
+  | SuperGlobal | Export ->
+    assert (not (isVarRef inst.inst_impl));
+    Some inst
 
 let rebuild_instance inst =
   add_instance_base inst;
   inst
 
-let classify_instance inst =
-  if is_local inst then Dispose
-  else Substitute inst
+let classify_instance inst = match inst.inst_global with
+| Local -> Dispose
+| SuperGlobal | Export -> Substitute
 
-let instance_input : instance -> obj =
+let instance_input : instance_obj -> obj =
   declare_object
     { (default_object "type classes instances state") with
       cache_function = cache_instance;
-      load_function = (fun _ x -> cache_instance x);
-      open_function = simple_open (fun _ x -> cache_instance x);
+      load_function = load_instance;
+      open_function = simple_open ~cat:Hints.hint_cat open_instance;
       classify_function = classify_instance;
       discharge_function = discharge_instance;
       rebuild_function = rebuild_instance;
       subst_function = subst_instance }
 
-let add_instance i =
-  Lib.add_anonymous_leaf (instance_input i);
+let warn_deprecated_instance_without_locality =
+  let open Pp in
+  CWarnings.create ~name:"deprecated-instance-without-locality" ~category:"deprecated"
+    ~default:CWarnings.AsError
+    (fun () -> strbrk "The default value for instance locality is currently \
+    \"local\" in a section and \"global\" otherwise, but is scheduled to change \
+    in a future release. For the time being, adding instances outside of sections \
+    without specifying an explicit locality attribute is therefore deprecated. It is \
+    recommended to use \"export\" whenever possible. Use the attributes \
+    #[local], #[global] and #[export] depending on your choice. For example: \
+    \"#[export] Instance Foo : Bar := baz.\"")
+
+let default_locality () =
+  if Global.sections_are_opened () then Local
+  else
+    let () = warn_deprecated_instance_without_locality () in
+    SuperGlobal
+
+let instance_locality =
+  Attributes.hint_locality ~default:default_locality
+
+let add_instance cl info global impl =
+  let () = match global with
+    | Local -> ()
+    | SuperGlobal ->
+      if Global.sections_are_opened () && isVarRef impl then
+        CErrors.user_err (Pp.str "Cannot set Global an instance referring to a section variable.")
+    | Export ->
+      if Global.sections_are_opened () && isVarRef impl then
+        CErrors.user_err (Pp.str "The export attribute cannot be applied to an instance referring to a section variable.")
+  in
+  let i = {
+    inst_class = cl.cl_impl;
+    inst_info = info ;
+    inst_global = global ;
+    inst_impl = impl;
+  } in
+  Lib.add_leaf (instance_input i);
   add_instance_base i
 
 let warning_not_a_class =
@@ -144,76 +203,45 @@ let declare_instance ?(warn = false) env sigma info local glob =
   let info = Option.default {hint_priority = None; hint_pattern = None} info in
   match class_of_constr env sigma (EConstr.of_constr ty) with
   | Some (rels, ((tc,_), args) as _cl) ->
-    assert (not (isVarRef glob) || local);
-    add_instance (mk_instance tc info (not local) glob)
+    add_instance tc info local glob
   | None -> if warn then warning_not_a_class (glob, ty)
 
 (*
  * classes persistent object
  *)
 
-let cache_class (_,c) = load_class c
+let cache_class c = load_class c
 
 let subst_class (subst,cl) =
   let do_subst_con c = Mod_subst.subst_constant subst c
   and do_subst c = Mod_subst.subst_mps subst c
   and do_subst_gr gr = fst (subst_global subst gr) in
   let do_subst_ctx = List.Smart.map (RelDecl.map_constr do_subst) in
-  let do_subst_context (grs,ctx) =
-    List.Smart.map (Option.Smart.map do_subst_gr) grs,
-    do_subst_ctx ctx in
-  let do_subst_projs projs = List.Smart.map (fun (x, y, z) ->
-    (x, y, Option.Smart.map do_subst_con z)) projs in
+  let do_subst_meth m =
+    let c = Option.Smart.map do_subst_con m.meth_const in
+    if c == m.meth_const then m
+    else
+    {
+      meth_name = m.meth_name;
+      meth_info = m.meth_info;
+      meth_const = c;
+    }
+  in
+  let do_subst_projs projs = List.Smart.map do_subst_meth projs in
   { cl_univs = cl.cl_univs;
     cl_impl = do_subst_gr cl.cl_impl;
-    cl_context = do_subst_context cl.cl_context;
+    cl_context = do_subst_ctx cl.cl_context;
     cl_props = do_subst_ctx cl.cl_props;
     cl_projs = do_subst_projs cl.cl_projs;
     cl_strict = cl.cl_strict;
     cl_unique = cl.cl_unique }
 
-let discharge_class (_,cl) =
-  let open CVars in
-  let repl = Lib.replacement_context () in
-  let rel_of_variable_context ctx = List.fold_right
-    ( fun decl (ctx', subst) ->
-        let decl' = decl |> NamedDecl.map_constr (substn_vars 1 subst) |> NamedDecl.to_rel_decl in
-        (decl' :: ctx', NamedDecl.get_id decl :: subst)
-    ) ctx ([], []) in
-  let discharge_rel_context (subst, usubst) n rel =
-    let rel = Context.Rel.map (Cooking.expmod_constr repl) rel in
-    let fold decl (ctx, k) =
-      let map c = subst_univs_level_constr usubst (substn_vars k subst c) in
-      RelDecl.map_constr map decl :: ctx, succ k
-    in
-    let ctx, _ = List.fold_right fold rel ([], n) in
-    ctx
-  in
-  let abs_context cl =
-    let open GlobRef in
-    match cl.cl_impl with
-      | VarRef _ | ConstructRef _ -> assert false
-      | ConstRef cst -> Lib.section_segment_of_constant cst
-      | IndRef (ind,_) -> Lib.section_segment_of_mutual_inductive ind in
-  let discharge_context ctx' subst (grs, ctx) =
-    let env = Global.env () in
-    let sigma = Evd.from_env env in
-    let grs' =
-      let newgrs = List.map (fun decl ->
-                             match decl |> RelDecl.get_type |> EConstr.of_constr |> class_of_constr env sigma with
-                             | None -> None
-                             | Some (_, ((tc,_), _)) -> Some tc.cl_impl)
-                            ctx'
-      in
-      grs @ newgrs
-    in grs', discharge_rel_context subst 1 ctx @ ctx' in
+let discharge_class cl =
   try
-    let info = abs_context cl in
-    let ctx = info.Section.abstr_ctx in
-    let ctx, subst = rel_of_variable_context ctx in
-    let usubst, cl_univs' = Lib.discharge_abstract_universe_context info cl.cl_univs in
-    let context = discharge_context ctx (subst, usubst) cl.cl_context in
-    let props = discharge_rel_context (subst, usubst) (succ (List.length (fst cl.cl_context))) cl.cl_props in
+    let info = Lib.section_segment_of_reference cl.cl_impl in
+    let info, _, cl_univs' = Cooking.lift_poly_univs info cl.cl_univs in
+    let nprops = List.length cl.cl_props in
+    let props, context = List.chop nprops (Discharge.cook_rel_context info (cl.cl_props @ cl.cl_context)) in
     let discharge_proj x = x in
     { cl_univs = cl_univs';
       cl_impl = cl.cl_impl;
@@ -229,7 +257,7 @@ let discharge_class (_,cl) =
 let rebuild_class cl =
   try
     let cst = Tacred.evaluable_of_global_reference (Global.env ()) cl.cl_impl in
-      set_typeclass_transparency cst false false; cl
+      set_typeclass_transparency ~locality:Hints.Local [cst] false; cl
   with e when CErrors.noncritical e -> cl
 
 let class_input : typeclass -> obj =
@@ -237,25 +265,13 @@ let class_input : typeclass -> obj =
     { (default_object "type classes state") with
       cache_function = cache_class;
       load_function = (fun _ -> cache_class);
-      open_function = simple_open (fun _ -> cache_class);
-      classify_function = (fun x -> Substitute x);
+      classify_function = (fun x -> Substitute);
       discharge_function = (fun a -> Some (discharge_class a));
       rebuild_function = rebuild_class;
       subst_function = subst_class }
 
 let add_class cl =
-  Lib.add_anonymous_leaf (class_input cl)
-
-let add_class env sigma cl =
-  add_class cl;
-  List.iter (fun (n, inst, body) ->
-      match inst with
-      | Some (Backward, info) ->
-        (match body with
-         | None -> CErrors.user_err Pp.(str "Non-definable projection can not be declared as a subinstance")
-         | Some b -> declare_instance ~warn:true env sigma (Some info) false (GlobRef.ConstRef b))
-      | _ -> ())
-    cl.cl_projs
+  Lib.add_leaf (class_input cl)
 
 let intern_info {hint_priority;hint_pattern} =
   let env = Global.env() in
@@ -271,12 +287,11 @@ let existing_instance glob g info =
   let env = Global.env() in
   let sigma = Evd.from_env env in
   let instance, _ = Typeops.type_of_global_in_context env c in
-  let _, r = Term.decompose_prod_assum instance in
-    match class_of_constr env sigma (EConstr.of_constr r) with
-      | Some (_, ((tc,u), _)) -> add_instance (mk_instance tc info glob c)
+  let ctx, r = Term.decompose_prod_assum instance in
+    match class_of_constr (Environ.push_rel_context ctx env) sigma (EConstr.of_constr r) with
+      | Some (_, ((tc,u), _)) -> add_instance tc info glob c
       | None -> user_err ?loc:g.CAst.loc
-                         ~hdr:"declare_instance"
-                         (Pp.str "Constant does not build instances of a declared type class.")
+                  (Pp.str "Constant does not build instances of a declared type class.")
 
 (* Declare everything in the parameters as implicit, and the class instance as well *)
 
@@ -308,81 +323,84 @@ let instance_hook info global ?hook cst =
   let info = intern_info info in
   let env = Global.env () in
   let sigma = Evd.from_env env in
-  declare_instance env sigma (Some info) (not global) cst;
+  declare_instance env sigma (Some info) global cst;
   (match hook with Some h -> h cst | None -> ())
 
-let declare_instance_constant info global impargs ?hook name udecl poly sigma term termtype =
+let declare_instance_constant iinfo global impargs ?hook name udecl poly sigma term termtype =
   let kind = Decls.(IsDefinition Instance) in
-  let scope = Declare.Global Declare.ImportDefaultBehavior in
-  let kn = Declare.declare_definition ~name ~kind ~scope ~impargs
-      ~opaque:false ~poly sigma ~udecl ~types:(Some termtype) ~body:term in
-  instance_hook info global ?hook kn
+  let cinfo = Declare.CInfo.make ~name ~impargs ~typ:(Some termtype) () in
+  let info = Declare.Info.make ~kind ~poly ~udecl () in
+  let kn = Declare.declare_definition ~cinfo ~info ~opaque:false ~body:term sigma in
+  instance_hook iinfo global ?hook kn
 
-let do_declare_instance sigma ~global ~poly k u ctx ctx' pri udecl impargs subst name =
+let do_declare_instance sigma ~locality ~poly k u ctx ctx' pri udecl impargs subst name =
   let subst = List.fold_left2
       (fun subst' s decl -> if is_local_assum decl then s :: subst' else subst')
-      [] subst (snd k.cl_context)
+      [] subst k.cl_context
   in
   let (_, ty_constr) = instance_constructor (k,u) subst in
   let termtype = it_mkProd_or_LetIn ty_constr (ctx' @ ctx) in
   let sigma, entry = Declare.prepare_parameter ~poly sigma ~udecl ~types:termtype in
   let cst = Declare.declare_constant ~name
       ~kind:Decls.(IsAssumption Logical) (Declare.ParameterEntry entry) in
-  DeclareUniv.declare_univ_binders (GlobRef.ConstRef cst) (Evd.universe_binders sigma);
   let cst = (GlobRef.ConstRef cst) in
   Impargs.maybe_declare_manual_implicits false cst impargs;
-  instance_hook pri global cst
+  instance_hook pri locality cst
 
-let declare_instance_program env sigma ~global ~poly name pri impargs udecl term termtype =
+let declare_instance_program pm env sigma ~locality ~poly name pri impargs udecl term termtype =
   let hook { Declare.Hook.S.scope; dref; _ } =
     let cst = match dref with GlobRef.ConstRef kn -> kn | _ -> assert false in
     let pri = intern_info pri in
     let env = Global.env () in
     let sigma = Evd.from_env env in
-    declare_instance env sigma (Some pri) (not global) (GlobRef.ConstRef cst)
+    declare_instance env sigma (Some pri) locality (GlobRef.ConstRef cst)
   in
   let obls, _, term, typ = RetrieveObl.retrieve_obligations env name sigma 0 term termtype in
   let hook = Declare.Hook.make hook in
   let uctx = Evd.evar_universe_context sigma in
-  let scope, kind = Declare.Global Declare.ImportDefaultBehavior, Decls.Instance in
-  let _ : DeclareObl.progress =
-    Obligations.add_definition ~name ~term ~udecl ~scope ~poly ~kind ~hook ~impargs ~uctx typ obls
-  in ()
+  let kind = Decls.IsDefinition Decls.Instance in
+  let cinfo = Declare.CInfo.make ~name ~typ ~impargs () in
+  let info = Declare.Info.make  ~udecl ~poly ~kind ~hook () in
+  let pm, _ =
+    Declare.Obls.add_definition ~pm ~cinfo ~info ~term ~uctx obls
+  in pm
 
-let declare_instance_open sigma ?hook ~tac ~global ~poly id pri impargs udecl ids term termtype =
+let declare_instance_open sigma ?hook ~tac ~locality ~poly id pri impargs udecl ids term termtype =
   (* spiwack: it is hard to reorder the actions to do
      the pretyping after the proof has opened. As a
      consequence, we use the low-level primitives to code
      the refinement manually.*)
-  let gls = List.rev (Evd.future_goals sigma) in
-  let sigma = Evd.reset_future_goals sigma in
+  let future_goals, sigma = Evd.pop_future_goals sigma in
+  let gls = List.rev (Evd.FutureGoals.comb future_goals) in
+  let sigma = Evd.push_future_goals sigma in
   let kind = Decls.(IsDefinition Instance) in
-  let hook = Declare.Hook.(make (fun { S.dref ; _ } -> instance_hook pri global ?hook dref)) in
-  let info = Lemmas.Info.make ~hook ~kind () in
+  let hook = Declare.Hook.(make (fun { S.dref ; _ } -> instance_hook pri locality ?hook dref)) in
+  let info = Declare.Info.make ~hook ~kind ~udecl ~poly () in
   (* XXX: We need to normalize the type, otherwise Admitted / Qed will fails!
      This is due to a bug in proof_global :( *)
   let termtype = Evarutil.nf_evar sigma termtype in
-  let lemma = Lemmas.start_lemma ~name:id ~poly ~udecl ~info ~impargs sigma termtype in
+  let cinfo = Declare.CInfo.make ~name:id ~impargs ~typ:termtype () in
+  let lemma = Declare.Proof.start ~cinfo ~info sigma in
   (* spiwack: I don't know what to do with the status here. *)
   let lemma =
     match term with
     | Some term ->
       let init_refine =
-        Tacticals.New.tclTHENLIST [
+        Tacticals.tclTHENLIST [
           Refine.refine ~typecheck:false (fun sigma -> sigma, term);
           Proofview.Unsafe.tclNEWGOALS (CList.map Proofview.with_empty_state gls);
-          Tactics.New.reduce_after_refine;
+          Tactics.reduce_after_refine;
         ]
       in
-      let lemma, _ = Lemmas.by init_refine lemma in
+      let lemma, _ = Declare.Proof.by init_refine lemma in
       lemma
     | None ->
-      let lemma, _ = Lemmas.by (Tactics.auto_intros_tac ids) lemma in
+      let lemma, _ = Declare.Proof.by (Tactics.auto_intros_tac ids) lemma in
       lemma
   in
   match tac with
   | Some tac ->
-    let lemma, _ = Lemmas.by tac lemma in
+    let lemma, _ = Declare.Proof.by tac lemma in
     lemma
   | None ->
     lemma
@@ -391,7 +409,7 @@ let do_instance_subst_constructor_and_ty subst k u ctx =
   let subst =
     List.fold_left2 (fun subst' s decl ->
       if is_local_assum decl then s :: subst' else subst')
-    [] subst (k.cl_props @ snd k.cl_context)
+    [] subst (k.cl_props @ k.cl_context)
   in
   let (app, ty_constr) = instance_constructor (k,u) subst in
   let termtype = it_mkProd_or_LetIn ty_constr ctx in
@@ -425,9 +443,9 @@ let do_instance_type_ctx_instance props k env' ctx' sigma ~program_mode subst =
              let rest' = List.filter (fun v -> not (is_id v)) rest
              in
              let {CAst.loc;v=mid} = get_id loc_mid in
-             List.iter (fun (n, _, x) ->
-                 if Name.equal n (Name mid) then
-                   Option.iter (fun x -> Dumpglob.add_glob ?loc (GlobRef.ConstRef x)) x) k.cl_projs;
+             List.iter (fun m ->
+                 if Name.equal m.meth_name (Name mid) then
+                   Option.iter (fun x -> Dumpglob.add_glob ?loc (GlobRef.ConstRef x)) m.meth_const) k.cl_projs;
              c :: props, rest'
            with Not_found ->
              ((CAst.make @@ CHole (None(* Some Evar_kinds.GoalEvar *), Namegen.IntroAnonymous, None)) :: props), rest
@@ -458,7 +476,7 @@ let interp_props ~program_mode env' cty k u ctx ctx' subst sigma = function
     let term = it_mkLambda_or_LetIn def ctx in
     term, termtype, sigma
 
-let do_instance_interactive env env' sigma ?hook ~tac ~global ~poly cty k u ctx ctx' pri decl imps subst id opt_props =
+let do_instance_interactive env env' sigma ?hook ~tac ~locality ~poly cty k u ctx ctx' pri decl imps subst id opt_props =
   let term, termtype, sigma = match opt_props with
     | Some props ->
       on_pi1 (fun x -> Some x)
@@ -476,19 +494,19 @@ let do_instance_interactive env env' sigma ?hook ~tac ~global ~poly cty k u ctx
       term, termtype, sigma
   in
   Flags.silently (fun () ->
-      declare_instance_open sigma ?hook ~tac ~global ~poly
+      declare_instance_open sigma ?hook ~tac ~locality ~poly
         id pri imps decl (List.map RelDecl.get_name ctx) term termtype)
     ()
 
-let do_instance env env' sigma ?hook ~global ~poly cty k u ctx ctx' pri decl imps subst id props =
+let do_instance env env' sigma ?hook ~locality ~poly cty k u ctx ctx' pri decl imps subst id props =
   let term, termtype, sigma =
     interp_props ~program_mode:false env' cty k u ctx ctx' subst sigma props
   in
   let termtype, sigma = do_instance_resolve_TC termtype sigma env in
   Pretyping.check_evars_are_solved ~program_mode:false env sigma;
-  declare_instance_constant pri global imps ?hook id decl poly sigma term termtype
+  declare_instance_constant pri locality imps ?hook id decl poly sigma term termtype
 
-let do_instance_program env env' sigma ?hook ~global ~poly cty k u ctx ctx' pri decl imps subst id opt_props =
+let do_instance_program ~pm env env' sigma ?hook ~locality ~poly cty k u ctx ctx' pri decl imps subst id opt_props =
   let term, termtype, sigma =
     match opt_props with
     | Some props ->
@@ -501,16 +519,13 @@ let do_instance_program env env' sigma ?hook ~global ~poly cty k u ctx ctx' pri
       term, termtype, sigma in
   let termtype, sigma = do_instance_resolve_TC termtype sigma env in
   if not (Evd.has_undefined sigma) && not (Option.is_empty opt_props) then
-    declare_instance_constant pri global imps ?hook id decl poly sigma term termtype
+    let () = declare_instance_constant pri locality imps ?hook id decl poly sigma term termtype in
+    pm
   else
-    declare_instance_program  env sigma ~global ~poly id pri imps decl term termtype
+    declare_instance_program pm env sigma ~locality ~poly id pri imps decl term termtype
 
-let interp_instance_context ~program_mode env ctx ~generalize pl tclass =
-  let sigma, decl = Constrexpr_ops.interp_univ_decl_opt env pl in
-  let tclass =
-    if generalize then CAst.make @@ CGeneralization (Glob_term.MaxImplicit, Some AbsPi, tclass)
-    else tclass
-  in
+let interp_instance_context ~program_mode env ctx pl tclass =
+  let sigma, decl = interp_univ_decl_opt env pl in
   let sigma, (impls, ((env', ctx), imps)) = interp_context_evars ~program_mode env sigma ctx in
   let flags = Pretyping.{ all_no_fail_flags with program_mode } in
   let sigma, (c', imps') = interp_type_evars_impls ~flags ~impls env' sigma tclass in
@@ -521,7 +536,7 @@ let interp_instance_context ~program_mode env ctx ~generalize pl tclass =
   let u_s = EInstance.kind sigma u in
   let cl = Typeclasses.typeclass_univ_instance (k, u_s) in
   let args = List.map of_constr args in
-  let cl_context = List.map (Termops.map_rel_decl of_constr) (snd cl.cl_context) in
+  let cl_context = List.map (Termops.map_rel_decl of_constr) cl.cl_context in
   let _, args =
     List.fold_right (fun decl (args, args') ->
         match decl with
@@ -533,10 +548,10 @@ let interp_instance_context ~program_mode env ctx ~generalize pl tclass =
   let sigma = resolve_typeclasses ~filter:Typeclasses.all_evars ~fail:true env sigma in
   sigma, cl, u, c', ctx', ctx, imps, args, decl
 
-let new_instance_common ~program_mode ~generalize env instid ctx cl =
+let new_instance_common ~program_mode env instid ctx cl =
   let ({CAst.loc;v=instid}, pl) = instid in
   let sigma, k, u, cty, ctx', ctx, imps, subst, decl =
-    interp_instance_context ~program_mode env ~generalize ctx pl cl
+    interp_instance_context ~program_mode env ctx pl cl
   in
   (* The name generator should not be here *)
   let id =
@@ -549,43 +564,39 @@ let new_instance_common ~program_mode ~generalize env instid ctx cl =
   let env' = push_rel_context ctx env in
   id, env', sigma, k, u, cty, ctx', ctx, imps, subst, decl
 
-let new_instance_interactive ?(global=false)
-    ~poly instid ctx cl
-    ?(generalize=true) ?(tac:unit Proofview.tactic option) ?hook
+let new_instance_interactive ~locality ~poly instid ctx cl
+    ?(tac:unit Proofview.tactic option) ?hook
     pri opt_props =
   let env = Global.env() in
   let id, env', sigma, k, u, cty, ctx', ctx, imps, subst, decl =
-    new_instance_common ~program_mode:false ~generalize env instid ctx cl in
-  id, do_instance_interactive env env' sigma ?hook ~tac ~global ~poly
+    new_instance_common ~program_mode:false env instid ctx cl in
+  id, do_instance_interactive env env' sigma ?hook ~tac ~locality ~poly
     cty k u ctx ctx' pri decl imps subst id opt_props
 
-let new_instance_program ?(global=false)
-    ~poly instid ctx cl opt_props
-    ?(generalize=true) ?hook pri =
+let new_instance_program ~locality ~pm ~poly instid ctx cl opt_props ?hook pri =
   let env = Global.env() in
   let id, env', sigma, k, u, cty, ctx', ctx, imps, subst, decl =
-    new_instance_common ~program_mode:true ~generalize env instid ctx cl in
-  do_instance_program env env' sigma ?hook ~global ~poly
-    cty k u ctx ctx' pri decl imps subst id opt_props;
-  id
+    new_instance_common ~program_mode:true env instid ctx cl in
+  let pm =
+    do_instance_program ~pm env env' sigma ?hook ~locality ~poly
+      cty k u ctx ctx' pri decl imps subst id opt_props in
+  pm, id
 
-let new_instance ?(global=false)
-    ~poly instid ctx cl props
-    ?(generalize=true) ?hook pri =
+let new_instance ~locality ~poly instid ctx cl props ?hook pri =
   let env = Global.env() in
   let id, env', sigma, k, u, cty, ctx', ctx, imps, subst, decl =
-    new_instance_common ~program_mode:false ~generalize env instid ctx cl in
-  do_instance env env' sigma ?hook ~global ~poly
+    new_instance_common ~program_mode:false env instid ctx cl in
+  do_instance env env' sigma ?hook ~locality ~poly
     cty k u ctx ctx' pri decl imps subst id props;
   id
 
-let declare_new_instance ?(global=false) ~program_mode ~poly instid ctx cl pri =
+let declare_new_instance ~locality ~program_mode ~poly instid ctx cl pri =
   let env = Global.env() in
   let ({CAst.loc;v=instid}, pl) = instid in
   let sigma, k, u, cty, ctx', ctx, imps, subst, decl =
-    interp_instance_context ~program_mode ~generalize:false env ctx pl cl
+    interp_instance_context ~program_mode env ctx pl cl
   in
-  do_declare_instance sigma ~global ~poly k u ctx ctx' pri decl imps subst instid
+  do_declare_instance sigma ~locality ~poly k u ctx ctx' pri decl imps subst instid
 
 let refine_att =
   let open Attributes in
@@ -593,3 +604,10 @@ let refine_att =
   attribute_of_list ["refine",single_key_parser ~name:"refine" ~key:"refine" ()] >>= function
   | None -> return false
   | Some () -> return true
+
+module Internal =
+struct
+let add_instance cl info glob r =
+  let glob = if glob then SuperGlobal else Local in
+  add_instance cl info glob r
+end
