@@ -18,7 +18,7 @@ let smart_global r =
   Dumpglob.add_glob ?loc:r.loc gr;
   gr
 
-let cache_bidi_hints (_name, (gr, ohint)) =
+let cache_bidi_hints (gr, ohint) =
   match ohint with
   | None -> Pretyping.clear_bidirectionality_hint gr
   | Some nargs -> Pretyping.add_bidirectionality_hint gr nargs
@@ -30,11 +30,11 @@ let subst_bidi_hints (subst, (gr, ohint as orig)) =
   let gr' = Globnames.subst_global_reference subst gr in
   if gr == gr' then orig else (gr', ohint)
 
-let discharge_bidi_hints (_name, (gr, ohint)) =
+let discharge_bidi_hints (gr, ohint) =
   if Globnames.isVarRef gr && Lib.is_in_section gr then None
   else
-    let vars = Lib.variable_section_segment_of_reference gr in
-    let n = List.length vars in
+    let vars = Lib.section_instance gr in
+    let n = Array.length vars in
     Some (gr, Option.map ((+) n) ohint)
 
 let inBidiHints =
@@ -42,7 +42,7 @@ let inBidiHints =
   declare_object { (default_object "BIDIRECTIONALITY-HINTS" ) with
                    load_function = load_bidi_hints;
                    cache_function = cache_bidi_hints;
-                   classify_function = (fun o -> Substitute o);
+                   classify_function = (fun o -> Substitute);
                    subst_function = subst_bidi_hints;
                    discharge_function = discharge_bidi_hints;
                  }
@@ -103,7 +103,7 @@ let vernac_arguments ~section_local reference args more_implicits flags =
   let sr = smart_global reference in
   let inf_names =
     let ty, _ = Typeops.type_of_global_in_context env sr in
-    Impargs.compute_implicits_names env sigma (EConstr.of_constr ty)
+    List.map pi1 (Impargs.compute_implicits_names env sigma (EConstr.of_constr ty))
   in
   let prev_names =
     try Arguments_renaming.arguments_names sr with Not_found -> inf_names
@@ -116,12 +116,12 @@ let vernac_arguments ~section_local reference args more_implicits flags =
   (* Checks *)
 
   let err_extra_args names =
-    CErrors.user_err ~hdr:"vernac_declare_arguments"
+    CErrors.user_err
       Pp.(strbrk "Extra arguments: " ++
           prlist_with_sep pr_comma Name.print names ++ str ".")
   in
   let err_missing_args names =
-    CErrors.user_err ~hdr:"vernac_declare_arguments"
+    CErrors.user_err
       Pp.(strbrk "The following arguments are not declared: " ++
           prlist_with_sep pr_comma Name.print names ++ str ".")
   in
@@ -129,9 +129,9 @@ let vernac_arguments ~section_local reference args more_implicits flags =
   let rec check_extra_args extra_args =
     match extra_args with
     | [] -> ()
-    | { notation_scope = None } :: _ ->
-      CErrors.user_err Pp.(str"Extra arguments should specify a scope.")
-    | { notation_scope = Some _ } :: args -> check_extra_args args
+    | { notation_scope = [] } :: _ ->
+      CErrors.user_err Pp.(str"Extra arguments should specify scopes.")
+    | { notation_scope = _ :: _ } :: args -> check_extra_args args
   in
 
   let args, scopes =
@@ -150,7 +150,7 @@ let vernac_arguments ~section_local reference args more_implicits flags =
   if Option.cata (fun n -> n > num_args) false nargs_before_bidi then
     CErrors.user_err Pp.(str "The \"&\" modifier should be put before any extra scope.");
 
-  let scopes_specified = List.exists Option.has_some scopes in
+  let scopes_specified = List.exists ((<>) []) scopes in
 
   if scopes_specified && clear_scopes_flag then
     CErrors.user_err Pp.(str "The \"clear scopes\" flag is incompatible with scope annotations.");
@@ -210,15 +210,7 @@ let vernac_arguments ~section_local reference args more_implicits flags =
       | Some (o,n) ->
         strbrk "Flag \"rename\" expected to rename " ++ Name.print o ++
         strbrk " into " ++ Name.print n ++ str "."
-    in CErrors.user_err ~hdr:"vernac_declare_arguments" msg
-  end;
-
-  let duplicate_names =
-    List.duplicates Name.equal (List.filter ((!=) Anonymous) names)
-  in
-  if not (List.is_empty duplicate_names) then begin
-    CErrors.user_err Pp.(strbrk "Some argument names are duplicated: " ++
-                         prlist_with_sep pr_comma Name.print duplicate_names)
+    in CErrors.user_err msg
   end;
 
   let implicits =
@@ -226,16 +218,15 @@ let vernac_arguments ~section_local reference args more_implicits flags =
   in
   let implicits = implicits :: more_implicits in
 
-  let implicits = List.map (List.map snd) implicits in
   let implicits_specified = match implicits with
-    | [l] -> List.exists (function Glob_term.Explicit -> false | _ -> true) l
+    | [l] -> List.exists (function _, Glob_term.Explicit -> false | _ -> true) l
     | _ -> true in
 
   if implicits_specified && clear_implicits_flag then
-    CErrors.user_err Pp.(str "The \"clear implicits\" flag is incompatible with implicit annotations");
+    CErrors.user_err Pp.(str "The \"clear implicits\" flag must be omitted if implicit annotations are given.");
 
   if implicits_specified && default_implicits_flag then
-    CErrors.user_err Pp.(str "The \"default implicits\" flag is incompatible with implicit annotations");
+    CErrors.user_err Pp.(str "The \"default implicits\" flag is incompatible with implicit annotations.");
 
   let rargs =
     Util.List.map_filter (function (n, true) -> Some n | _ -> None)
@@ -274,7 +265,7 @@ let vernac_arguments ~section_local reference args more_implicits flags =
   end;
 
   if scopes_specified || clear_scopes_flag then begin
-    let scopes = List.map (Option.map (fun {loc;v=k} ->
+    let scopes = List.map (List.map (fun {loc;v=k} ->
         try ignore (Notation.find_scope k); k
         with CErrors.UserError _ ->
           Notation.find_delimiters_scope ?loc k)) scopes
@@ -305,14 +296,14 @@ let vernac_arguments ~section_local reference args more_implicits flags =
     if section_local then
       Pretyping.add_bidirectionality_hint sr n
     else
-      Lib.add_anonymous_leaf (inBidiHints (sr, Some n))
+      Lib.add_leaf (inBidiHints (sr, Some n))
   end;
 
   if clear_bidi_hint then begin
     if section_local then
       Pretyping.clear_bidirectionality_hint sr
     else
-      Lib.add_anonymous_leaf (inBidiHints (sr, None))
+      Lib.add_leaf (inBidiHints (sr, None))
   end;
 
   if not (renaming_specified ||
