@@ -13,7 +13,6 @@ open CErrors
 open Util
 open Constr
 open Context
-open Vars
 open Declare
 open Names
 open Libnames
@@ -31,14 +30,14 @@ module RelDecl = Context.Rel.Declaration
 
 open Coqlib
 
-let init_constant sigma rf = Evarutil.new_global sigma rf
+let init_constant sigma rf = Evd.fresh_global sigma rf
 let fix_sub_ref () = lib_ref "program.wf.fix_sub"
 let measure_on_R_ref () = lib_ref "program.wf.mr"
-let well_founded sigma = init_constant sigma (lib_ref "core.wf.well_founded")
+let well_founded sigma = init_constant (Global.env ()) sigma (lib_ref "core.wf.well_founded")
 
 let mkSubset sigma name typ prop =
   let open EConstr in
-  let sigma, app_h = Evarutil.new_global sigma (delayed_force build_sigma).typ in
+  let sigma, app_h = Evd.fresh_global (Global.env ()) sigma (delayed_force build_sigma).typ in
   sigma, mkApp (app_h, [| typ; mkLambda (make_annot name Sorts.Relevant, typ, prop) |])
 
 let make_qref s = qualid_of_string s
@@ -59,7 +58,7 @@ let get_sigmatypes sigma ~sort ~predsort =
     | PropF, TypeF -> "sig", TypeF
     | TypeF, (PropF|TypeF) -> "sigT", TypeF
   in
-  let sigma, ty = Evarutil.new_global sigma (lib_ref ("core."^which^".type")) in
+  let sigma, ty = Evd.fresh_global (Global.env ()) sigma (lib_ref ("core."^which^".type")) in
   let uinstance = snd (destRef sigma ty) in
   let intro = mkRef (lib_ref ("core."^which^".intro"), uinstance) in
   let p1 = mkRef (lib_ref ("core."^which^".proj1"), uinstance) in
@@ -109,13 +108,12 @@ let telescope env sigma l =
 let nf_evar_context sigma ctx =
   List.map (map_constr (fun c -> Evarutil.nf_evar sigma c)) ctx
 
-let build_wellfounded (recname,pl,bl,arityc,body) poly r measure notation =
+let build_wellfounded pm (recname,pl,bl,arityc,body) poly ?typing_flags ?using r measure notation =
   let open EConstr in
   let open Vars in
-  let lift_rel_context n l = Termops.map_rel_context_with_binders (liftn n) l in
   Coqlib.check_required_library ["Coq";"Program";"Wf"];
   let env = Global.env() in
-  let sigma, udecl = Constrexpr_ops.interp_univ_decl_opt env pl in
+  let sigma, udecl = interp_univ_decl_opt env pl in
   let sigma, (_, ((env', binders_rel), impls)) = interp_context_evars ~program_mode:true env sigma bl in
   let len = List.length binders_rel in
   let top_env = push_rel_context binders_rel env in
@@ -131,16 +129,15 @@ let build_wellfounded (recname,pl,bl,arityc,body) poly r measure notation =
   let relargty =
     let error () =
       user_err ?loc:(constr_loc r)
-               ~hdr:"Command.build_wellfounded"
-                    (Printer.pr_econstr_env env sigma rel ++ str " is not an homogeneous binary relation.")
+        (Printer.pr_econstr_env env sigma rel ++ str " is not an homogeneous binary relation.")
     in
-      try
-        let ctx, ar = Reductionops.splay_prod_n env sigma 2 relty in
-          match ctx, EConstr.kind sigma ar with
-          | [LocalAssum (_,t); LocalAssum (_,u)], Sort s
-              when Sorts.is_prop (ESorts.kind sigma s) && Reductionops.is_conv env sigma t u -> t
-          | _, _ -> error ()
-      with e when CErrors.noncritical e -> error ()
+    try
+      let ctx, ar = Reductionops.splay_prod_n env sigma 2 relty in
+      match ctx, EConstr.kind sigma ar with
+      | [LocalAssum (_,t); LocalAssum (_,u)], Sort s
+        when Sorts.is_prop (ESorts.kind sigma s) && Reductionops.is_conv env sigma t u -> t
+      | _, _ -> error ()
+    with e when CErrors.noncritical e -> error ()
   in
   let sigma, measure = interp_casted_constr_evars ~program_mode:true binders_env sigma measure relargty in
   let sigma, wf_rel, wf_rel_fun, measure_fn =
@@ -148,7 +145,7 @@ let build_wellfounded (recname,pl,bl,arityc,body) poly r measure notation =
       it_mkLambda_or_LetIn measure letbinders,
       it_mkLambda_or_LetIn measure binders
     in
-    let sigma, comb = Evarutil.new_global sigma (delayed_force measure_on_R_ref) in
+    let sigma, comb = Evd.fresh_global (Global.env ()) sigma (delayed_force measure_on_R_ref) in
     let wf_rel = mkApp (comb, [| argtyp; relargty; rel; measure |]) in
     let wf_rel_fun x y =
       mkApp (rel, [| subst1 x measure_body;
@@ -167,7 +164,7 @@ let build_wellfounded (recname,pl,bl,arityc,body) poly r measure notation =
     sigma, wfa :: [arg]
   in
   let _intern_env = push_rel_context intern_bl env in
-  let sigma, proj = Evarutil.new_global sigma (delayed_force build_sigma).Coqlib.proj1 in
+  let sigma, proj = Evd.fresh_global (Global.env ()) sigma (delayed_force build_sigma).Coqlib.proj1 in
   let wfargpred = mkLambda (make_annot (Name argid') Sorts.Relevant, argtyp, wf_rel_fun (mkRel 1) (mkRel 3)) in
   let projection = (* in wfarg :: arg :: before *)
     mkApp (proj, [| argtyp ; wfargpred ; mkRel 1 |])
@@ -182,7 +179,7 @@ let build_wellfounded (recname,pl,bl,arityc,body) poly r measure notation =
                                       intern_fun_arity_prod) in
   let sigma, curry_fun =
     let wfpred = mkLambda (make_annot (Name argid') Sorts.Relevant, argtyp, wf_rel_fun (mkRel 1) (mkRel (2 * len + 4))) in
-    let sigma, intro = Evarutil.new_global sigma (delayed_force build_sigma).Coqlib.intro in
+    let sigma, intro = Evd.fresh_global (Global.env ()) sigma (delayed_force build_sigma).Coqlib.intro in
     let arg = mkApp (intro, [| argtyp; wfpred; lift 1 make; mkRel 1 |]) in
     let app = mkApp (mkRel (2 * len + 2 (* recproof + orig binders + current binders *)), [| arg |]) in
     let rcurry = mkApp (rel, [| measure; lift len measure |]) in
@@ -201,8 +198,8 @@ let build_wellfounded (recname,pl,bl,arityc,body) poly r measure notation =
     in
     let newimpls = Id.Map.singleton recname
         (Constrintern.extend_internalization_data interning_data
-           (Some (ExplByName (Id.of_string "recproof"), Impargs.Manual, (true, false)))
-           None) in
+           (Some ((Name (Id.of_string "recproof"),1,None), Impargs.Manual, (true, false)))
+           []) in
     interp_casted_constr_evars ~program_mode:true (push_rel_context ctx env) sigma
       ~impls:newimpls body (lift 1 top_arity)
   in
@@ -211,7 +208,7 @@ let build_wellfounded (recname,pl,bl,arityc,body) poly r measure notation =
   (* XXX: Previous code did parallel evdref update, so possible old
      weak ordering semantics may bite here. *)
   let sigma, def =
-    let sigma, h_a_term = Evarutil.new_global sigma (delayed_force fix_sub_ref) in
+    let sigma, h_a_term = Evd.fresh_global (Global.env ()) sigma (delayed_force fix_sub_ref) in
     let sigma, h_e_term = Evarutil.new_evar env sigma
         ~src:(Loc.tag @@ Evar_kinds.QuestionMark {
             Evar_kinds.default_question_mark with Evar_kinds.qm_obligation=Evar_kinds.Define false;
@@ -231,7 +228,7 @@ let build_wellfounded (recname,pl,bl,arityc,body) poly r measure notation =
       (* XXX: Mutating the evar_map in the hook! *)
       (* XXX: Likely the sigma is out of date when the hook is called .... *)
       let hook sigma { Declare.Hook.S.dref; _ } =
-        let sigma, h_body = Evarutil.new_global sigma dref in
+        let sigma, h_body = Evd.fresh_global (Global.env ()) sigma dref in
         let body = it_mkLambda_or_LetIn (mkApp (h_body, [|make|])) binders_rel in
         let ty = it_mkProd_or_LetIn top_arity binders_rel in
         let ty = EConstr.Unsafe.to_constr ty in
@@ -259,23 +256,30 @@ let build_wellfounded (recname,pl,bl,arityc,body) poly r measure notation =
   let evars, _, evars_def, evars_typ =
     RetrieveObl.retrieve_obligations env recname sigma 0 def typ
   in
+  let using =
+    let terms = List.map EConstr.of_constr [evars_def; evars_typ] in
+    Option.map (fun using -> Proof_using.definition_using env sigma ~using ~terms) using
+  in
   let uctx = Evd.evar_universe_context sigma in
-  ignore(Obligations.add_definition ~name:recname ~term:evars_def ~udecl
-           ~poly evars_typ ~uctx evars ~hook)
+  let cinfo = Declare.CInfo.make ~name:recname ~typ:evars_typ ?using () in
+  let info = Declare.Info.make ~udecl ~poly ~hook ?typing_flags () in
+  let pm, _ =
+    Declare.Obls.add_definition ~pm ~cinfo ~info ~term:evars_def ~uctx evars in
+  pm
 
 let out_def = function
   | Some def -> def
   | None -> user_err Pp.(str "Program Fixpoint needs defined bodies.")
 
 let collect_evars_of_term evd c ty =
-  let evars = Evar.Set.union (Evd.evars_of_term evd c) (Evd.evars_of_term evd ty) in
-  Evar.Set.fold (fun ev acc -> Evd.add acc ev (Evd.find_undefined evd ev))
-  evars (Evd.from_ctx (Evd.evar_universe_context evd))
+  Evar.Set.union (Evd.evars_of_term evd c) (Evd.evars_of_term evd ty)
 
-let do_program_recursive ~scope ~poly fixkind fixl =
-  let cofix = fixkind = DeclareObl.IsCoFixpoint in
+let do_program_recursive ~pm ~scope ~poly ?typing_flags ?using fixkind fixl =
+  let cofix = fixkind = Declare.Obls.IsCoFixpoint in
   let (env, rec_sign, udecl, evd), fix, info =
-    interp_recursive ~cofix ~program_mode:true fixl
+    let env = Global.env () in
+    let env = Environ.update_typing_flags ?typing_flags env in
+    interp_recursive env ~cofix ~program_mode:true fixl
   in
     (* Program-specific code *)
     (* Get the interesting evars, those that were not instantiated *)
@@ -284,13 +288,16 @@ let do_program_recursive ~scope ~poly fixkind fixl =
   let evd = nf_evar_map_undefined evd in
   let collect_evars name def typ impargs =
     (* Generalize by the recursive prototypes  *)
-    let def = nf_evar evd (Termops.it_mkNamedLambda_or_LetIn def rec_sign) in
-    let typ = nf_evar evd (Termops.it_mkNamedProd_or_LetIn typ rec_sign) in
-    let evm = collect_evars_of_term evd def typ in
+    let terms = [def; typ] in
+    let using = Option.map (fun using -> Proof_using.definition_using env evd ~using ~terms) using in
+    let def = nf_evar evd (EConstr.it_mkNamedLambda_or_LetIn evd def rec_sign) in
+    let typ = nf_evar evd (EConstr.it_mkNamedProd_or_LetIn evd typ rec_sign) in
+    let deps = collect_evars_of_term evd def typ in
     let evars, _, def, typ =
-      RetrieveObl.retrieve_obligations env name evm
-        (List.length rec_sign) def typ in
-    ({ Declare.Recthm.name; typ; impargs; args = [] }, def, evars)
+      RetrieveObl.retrieve_obligations env name evd
+        (List.length rec_sign) ~deps def typ in
+    let cinfo = Declare.CInfo.make ~name ~typ ~impargs ?using () in
+    (cinfo, def, evars)
   in
   let (fixnames,fixrs,fixdefs,fixtypes) = fix in
   let fiximps = List.map pi2 info in
@@ -300,33 +307,38 @@ let do_program_recursive ~scope ~poly fixkind fixl =
       let possible_indexes = List.map ComFixpoint.compute_possible_guardness_evidences info in
       (* XXX: are we allowed to have evars here? *)
       let fixtypes = List.map (EConstr.to_constr ~abort_on_undefined_evars:false evd) fixtypes in
+      let fixdefs = List.map (EConstr.Vars.subst_vars evd (List.rev fixnames)) fixdefs in
       let fixdefs = List.map (EConstr.to_constr ~abort_on_undefined_evars:false evd) fixdefs in
       let fixdecls = Array.of_list (List.map2 (fun x r -> make_annot (Name x) r) fixnames fixrs),
         Array.of_list fixtypes,
-        Array.of_list (List.map (subst_vars (List.rev fixnames)) fixdefs)
+        Array.of_list fixdefs
       in
       let indexes =
-        Pretyping.search_guard (Global.env ()) possible_indexes fixdecls in
+        let env = Global.env () in
+        let env = Environ.update_typing_flags ?typing_flags env in
+        Pretyping.search_guard env possible_indexes fixdecls in
+      let env = Environ.update_typing_flags ?typing_flags env in
       List.iteri (fun i _ ->
           Inductive.check_fix env
-                              ((indexes,i),fixdecls))
+            ((indexes,i),fixdecls))
         fixl
   end in
   let uctx = Evd.evar_universe_context evd in
   let kind = match fixkind with
-  | DeclareObl.IsFixpoint _ -> Decls.Fixpoint
-  | DeclareObl.IsCoFixpoint -> Decls.CoFixpoint
+  | Declare.Obls.IsFixpoint _ -> Decls.(IsDefinition Fixpoint)
+  | Declare.Obls.IsCoFixpoint -> Decls.(IsDefinition CoFixpoint)
   in
-  let ntns = List.map_append (fun { Vernacexpr.notations } -> notations ) fixl in
-  Obligations.add_mutual_definitions defs ~poly ~scope ~kind ~udecl ~uctx ntns fixkind
+  let ntns = List.map_append (fun { Vernacexpr.notations } -> List.map Metasyntax.prepare_where_notation notations ) fixl in
+  let info = Declare.Info.make ~poly ~scope ~kind ~udecl ?typing_flags () in
+  Declare.Obls.add_mutual_definitions ~pm defs ~info ~uctx ~ntns fixkind
 
-let do_fixpoint ~scope ~poly l =
+let do_fixpoint ~pm ~scope ~poly ?typing_flags ?using l =
   let g = List.map (fun { Vernacexpr.rec_order } -> rec_order) l in
     match g, l with
     | [Some { CAst.v = CWfRec (n,r) }],
       [ Vernacexpr.{fname={CAst.v=id}; univs; binders; rtype; body_def; notations} ] ->
         let recarg = mkIdentC n.CAst.v in
-        build_wellfounded (id, univs, binders, rtype, out_def body_def) poly r recarg notations
+        build_wellfounded pm (id, univs, binders, rtype, out_def body_def) poly ?typing_flags ?using r recarg notations
 
     | [Some { CAst.v = CMeasureRec (n, m, r) }],
       [Vernacexpr.{fname={CAst.v=id}; univs; binders; rtype; body_def; notations }] ->
@@ -339,20 +351,19 @@ let do_fixpoint ~scope ~poly l =
           user_err Pp.(str"Measure takes only two arguments in Program Fixpoint.")
         | _, _ -> r
       in
-        build_wellfounded (id, univs, binders, rtype, out_def body_def) poly
+        build_wellfounded pm (id, univs, binders, rtype, out_def body_def) poly ?typing_flags ?using
           (Option.default (CAst.make @@ CRef (lt_ref,None)) r) m notations
 
     | _, _ when List.for_all (fun ro -> match ro with None | Some { CAst.v = CStructRec _} -> true | _ -> false) g ->
       let annots = List.map (fun fix ->
           Vernacexpr.(ComFixpoint.adjust_rec_order ~structonly:true fix.binders fix.rec_order)) l in
-      let fixkind = DeclareObl.IsFixpoint annots in
+      let fixkind = Declare.Obls.IsFixpoint annots in
       let l = List.map2 (fun fix rec_order -> { fix with Vernacexpr.rec_order }) l annots in
-      do_program_recursive ~scope ~poly fixkind l
-
+      do_program_recursive ~pm ~scope ~poly ?typing_flags ?using fixkind l
     | _, _ ->
-        user_err ~hdr:"do_fixpoint"
-          (str "Well-founded fixpoints not allowed in mutually recursive blocks")
+      CErrors.user_err
+        (str "Well-founded fixpoints not allowed in mutually recursive blocks.")
 
-let do_cofixpoint ~scope ~poly fixl =
+let do_cofixpoint ~pm ~scope ~poly ?using fixl =
   let fixl = List.map (fun fix -> { fix with Vernacexpr.rec_order = None }) fixl in
-  do_program_recursive ~scope ~poly DeclareObl.IsCoFixpoint fixl
+  do_program_recursive ~pm ~scope ~poly ?using Declare.Obls.IsCoFixpoint fixl
