@@ -17,26 +17,23 @@
    declaring new schemes *)
 
 open Pp
-open CErrors
 open Util
-open Names
 open Declarations
 open Term
-open Constr
-open Inductive
-open Indrec
-open Declare
-open Libnames
 open Goptions
-open Nameops
-open Termops
-open Smartlocate
 open Vernacexpr
 open Ind_tables
 open Auto_ind_decl
 open Eqschemes
 open Elimschemes
-open Context.Rel.Declaration
+
+(** Data of an inductive scheme with name resolved *)
+type resolved_scheme = Names.Id.t CAst.t * Indrec.dep_flag * Names.inductive * Sorts.family
+
+(** flag for internal message display *)
+type internal_flag =
+  | UserAutomaticRequest (* kernel action, a message is displayed *)
+  | UserIndividualRequest   (* user action, a message is displayed *)
 
 (* Flags governing automatic synthesis of schemes *)
 
@@ -95,27 +92,26 @@ let define ~poly name sigma c types =
   let univs = Evd.univ_entry ~poly sigma in
   let entry = Declare.definition_entry ~univs ?types c in
   let kind = Decls.(IsDefinition Scheme) in
-  let kn = declare_constant ~kind ~name (DefinitionEntry entry) in
-  definition_message name;
+  let kn = Declare.declare_constant ~kind ~name (Declare.DefinitionEntry entry) in
+  Declare.definition_message name;
   kn
 
 (* Boolean equality *)
 
-let declare_beq_scheme_gen internal names kn =
-  ignore (define_mutual_scheme beq_scheme_kind internal names kn)
+let declare_beq_scheme_gen names kn =
+  ignore (define_mutual_scheme beq_scheme_kind names kn)
 
 let alarm what internal msg =
   let debug = false in
   match internal with
-  | UserAutomaticRequest
-  | InternalTacticRequest ->
+  | UserAutomaticRequest ->
     (if debug then
       Feedback.msg_debug
         (hov 0 msg ++ fnl () ++ what ++ str " not defined.")); None
-  | _ -> Some msg
+  | UserIndividualRequest -> Some msg
 
 let try_declare_scheme what f internal names kn =
-  try f internal names kn
+  try f names kn
   with e ->
   let e = Exninfo.capture e in
   let rec extract_exn = function Logic_monad.TacticFailure e -> extract_exn e | e -> e in
@@ -134,7 +130,7 @@ let try_declare_scheme what f internal names kn =
         alarm what internal
           (str "Cannot extract computational content from proposition " ++
            quote (Printer.pr_inductive (Global.env()) ind) ++ str ".")
-    | EqNotFound (ind',ind) ->
+    | EqNotFound ind' ->
         alarm what internal
           (str "Boolean equality on " ++
            quote (Printer.pr_inductive (Global.env()) ind') ++
@@ -162,6 +158,9 @@ let try_declare_scheme what f internal names kn =
            (strbrk "Unsupported constructor with an argument whose type is a non-parametric inductive type." ++
             strbrk " Type " ++ quote (Printer.pr_inductive (Global.env()) ind) ++
             str " is applied to an argument which is not a variable.")
+    | InternalDependencies ->
+         alarm what internal
+           (strbrk "Inductive types with internal dependencies in constructors not supported.")
     | e when CErrors.noncritical e ->
         alarm what internal
           (str "Unexpected error during scheme creation: " ++ CErrors.print e)
@@ -169,7 +168,7 @@ let try_declare_scheme what f internal names kn =
   in
   match msg with
   | None -> ()
-  | Some msg -> Exninfo.iraise (UserError (None, msg), snd e)
+  | Some msg -> Exninfo.iraise (CErrors.UserError msg, snd e)
 
 let beq_scheme_msg mind =
   let mib = Global.lookup_mind mind in
@@ -192,18 +191,18 @@ let declare_beq_scheme = declare_beq_scheme_with []
 (* Case analysis schemes *)
 let declare_one_case_analysis_scheme ind =
   let (mib,mip) = Global.lookup_inductive ind in
-  let kind = inductive_sort_family mip in
+  let kind = Inductive.inductive_sort_family mip in
   let dep =
     if kind == InProp then case_scheme_kind_from_prop
     else if not (Inductiveops.has_dependent_elim mib) then
       case_scheme_kind_from_type
     else case_dep_scheme_kind_from_type in
-  let kelim = elim_sort (mib,mip) in
+  let kelim = Inductive.elim_sort (mib,mip) in
     (* in case the inductive has a type elimination, generates only one
        induction scheme, the other ones share the same code with the
        appropriate type *)
   if Sorts.family_leq InType kelim then
-    define_individual_scheme dep UserAutomaticRequest None ind
+    define_individual_scheme dep None ind
 
 (* Induction/recursion schemes *)
 
@@ -227,10 +226,10 @@ let nondep_kinds_from_type =
 
 let declare_one_induction_scheme ind =
   let (mib,mip) = Global.lookup_inductive ind in
-  let kind = inductive_sort_family mip in
+  let kind = Inductive.inductive_sort_family mip in
   let from_prop = kind == InProp in
   let depelim = Inductiveops.has_dependent_elim mib in
-  let kelim = Inductiveops.sorts_below (elim_sort (mib,mip)) in
+  let kelim = Inductiveops.sorts_below (Inductive.elim_sort (mib,mip)) in
   let kelim = if Global.sprop_allowed () then kelim
     else List.filter (fun s -> s <> InSProp) kelim
   in
@@ -241,7 +240,7 @@ let declare_one_induction_scheme ind =
        else if depelim then kinds_from_type
        else nondep_kinds_from_type)
   in
-  List.iter (fun kind -> define_individual_scheme kind UserAutomaticRequest None ind)
+  List.iter (fun kind -> define_individual_scheme kind None ind)
     elims
 
 let declare_induction_schemes kn =
@@ -254,10 +253,10 @@ let declare_induction_schemes kn =
 
 (* Decidable equality *)
 
-let declare_eq_decidability_gen internal names kn =
+let declare_eq_decidability_gen names kn =
   let mib = Global.lookup_mind kn in
   if mib.mind_finite <> Declarations.CoFinite then
-    define_mutual_scheme eq_dec_scheme_kind internal names kn
+    define_mutual_scheme eq_dec_scheme_kind names kn
 
 let eq_dec_scheme_msg ind = (* TODO: mutual inductive case *)
   str "Decidable equality on " ++ quote (Printer.pr_inductive (Global.env()) ind)
@@ -276,18 +275,20 @@ let ignore_error f x =
   try f x with e when CErrors.noncritical e -> ()
 
 let declare_rewriting_schemes ind =
-  if Hipattern.is_inductive_equality ind then begin
-    define_individual_scheme rew_r2l_scheme_kind UserAutomaticRequest None ind;
-    define_individual_scheme rew_r2l_dep_scheme_kind UserAutomaticRequest None ind;
+  if Hipattern.is_inductive_equality (Global.env ()) ind then begin
+    define_individual_scheme rew_r2l_scheme_kind None ind;
+    define_individual_scheme rew_r2l_dep_scheme_kind None ind;
     define_individual_scheme rew_r2l_forward_dep_scheme_kind
-      UserAutomaticRequest None ind;
+      None ind;
     (* These ones expect the equality to be symmetric; the first one also *)
     (* needs eq *)
-    ignore_error (define_individual_scheme rew_l2r_scheme_kind UserAutomaticRequest None) ind;
+    ignore_error (define_individual_scheme rew_l2r_scheme_kind None) ind;
     ignore_error
-      (define_individual_scheme rew_l2r_dep_scheme_kind UserAutomaticRequest None) ind;
+      (define_individual_scheme sym_involutive_scheme_kind None) ind;
     ignore_error
-      (define_individual_scheme rew_l2r_forward_dep_scheme_kind UserAutomaticRequest None) ind
+      (define_individual_scheme rew_l2r_dep_scheme_kind None) ind;
+    ignore_error
+      (define_individual_scheme rew_l2r_forward_dep_scheme_kind None) ind
   end
 
 let warn_cannot_build_congruence =
@@ -298,138 +299,115 @@ let warn_cannot_build_congruence =
 let declare_congr_scheme ind =
   let env = Global.env () in
   let sigma = Evd.from_env env in
-  if Hipattern.is_equality_type env sigma (EConstr.of_constr (mkInd ind)) (* FIXME *) then begin
+  if Hipattern.is_equality_type env sigma (EConstr.of_constr (Constr.mkInd ind)) (* FIXME *) then begin
     if
       try Coqlib.check_required_library Coqlib.logic_module_name; true
       with e when CErrors.noncritical e -> false
     then
-      define_individual_scheme congr_scheme_kind UserAutomaticRequest None ind
+      define_individual_scheme congr_scheme_kind None ind
     else
       warn_cannot_build_congruence ()
   end
 
 let declare_sym_scheme ind =
-  if Hipattern.is_inductive_equality ind then
+  if Hipattern.is_inductive_equality (Global.env ()) ind then
     (* Expect the equality to be symmetric *)
-    ignore_error (define_individual_scheme sym_scheme_kind UserAutomaticRequest None) ind
+    ignore_error (define_individual_scheme sym_scheme_kind None) ind
 
 (* Scheme command *)
 
-let smart_global_inductive y = smart_global_inductive y
-let rec split_scheme env l =
+(* Boolean on scheme_type cheking if it considered dependent *)
+let sch_isdep = function
+| SchemeInduction  | SchemeElimination -> true
+| SchemeMinimality | SchemeCase        -> false
+
+(* Generate suffix for scheme given a target sort *)
+let scheme_suffix_gen {sch_type; sch_qualid; sch_sort} sort =
+  (* We check if we are working with recursion vs case schemes *)
+  let sch_isrec = match sch_type with
+    | SchemeInduction   | SchemeMinimality  -> true
+    | SchemeElimination | SchemeCase        -> false in
+  (* The _ind/_rec_/case suffix *)
+  let ind_suffix = match sch_isrec , sch_sort with
+    | true  , InSProp
+    | true  , InProp  -> "_ind"
+    | true  , _       -> "_rec"
+    | false , _       -> "_case" in
+  (* SProp and Type have an auxillary ending to the _ind suffix *)
+  let aux_suffix = match sch_sort with
+    | InSProp -> "s"
+    | InType  -> "t"
+    | _       -> "" in
+  (* Some schemes are deliminated with _dep or no_dep *)
+  let dep_suffix = match sch_isdep sch_type , sort with
+    | true  , InProp  -> "_dep"
+    | false , InSet
+    | false , InType
+    | false , InSProp -> "_nodep"
+    | _ , _           -> "" in
+  ind_suffix ^ aux_suffix ^ dep_suffix
+
+(* Resolve the names of a list of schemes using an enviornment and extract some
+important data such as the inductive type involved, whether it is a dependent
+eliminator and its sort. *)
+let rec name_and_process_schemes env l =
  match l with
-  | [] -> [],[]
-  | (Some id,t)::q -> let l1,l2 = split_scheme env q in
-    ( match t with
-      | InductionScheme (x,y,z) -> ((id,x,smart_global_inductive y,z)::l1),l2
-      | CaseScheme (x,y,z) -> ((id,x,smart_global_inductive y,z)::l1),l2
-      | EqualityScheme  x -> l1,((Some id,smart_global_inductive x)::l2)
-    )
-(*
- if no name has been provided, we build one from the types of the ind
-requested
-*)
-  | (None,t)::q ->
-      let l1,l2 = split_scheme env q in
-      let names inds recs isdep y z =
-        let ind = smart_global_inductive y in
-        let sort_of_ind = inductive_sort_family (snd (lookup_mind_specif env ind)) in
-        let suffix = (
-          match sort_of_ind with
-          | InProp ->
-              if isdep then (match z with
-              | InSProp -> inds ^ "s_dep"
-              | InProp -> inds ^ "_dep"
-              | InSet  -> recs ^ "_dep"
-              | InType -> recs ^ "t_dep")
-              else ( match z with
-              | InSProp -> inds ^ "s"
-              | InProp -> inds
-              | InSet -> recs
-              | InType -> recs ^ "t" )
-          | _ ->
-              if isdep then (match z with
-              | InSProp -> inds ^ "s"
-              | InProp -> inds
-              | InSet -> recs
-              | InType -> recs ^ "t" )
-              else (match z with
-              | InSProp -> inds ^ "s_nodep"
-              | InProp -> inds ^ "_nodep"
-              | InSet -> recs ^ "_nodep"
-              | InType -> recs ^ "t_nodep")
-        ) in
-        let newid = add_suffix (Nametab.basename_of_global (GlobRef.IndRef ind)) suffix in
-        let newref = CAst.make newid in
-          ((newref,isdep,ind,z)::l1),l2
-      in
-        match t with
-        | CaseScheme (x,y,z) -> names "_case" "_case" x y z
-        | InductionScheme (x,y,z) -> names "_ind" "_rec" x y z
-        | EqualityScheme  x -> l1,((None,smart_global_inductive x)::l2)
+  | [] -> []
+  | (Some id, {sch_type; sch_qualid; sch_sort}) :: q
+  -> ((id, sch_isdep sch_type, Smartlocate.smart_global_inductive sch_qualid, sch_sort)
+    :: name_and_process_schemes env q)
+(* If no name has been provided, we build one from the types of the ind requested *)
+  | (None, {sch_type; sch_qualid; sch_sort}) :: q
+   -> let ind = Smartlocate.smart_global_inductive sch_qualid in
+      let sort_of_ind = Inductive.inductive_sort_family (snd (Inductive.lookup_mind_specif env ind)) in
+      let suffix = scheme_suffix_gen {sch_type; sch_qualid; sch_sort} sort_of_ind in
+      let newid = Nameops.add_suffix (Nametab.basename_of_global (Names.GlobRef.IndRef ind)) suffix in
+      let newref = CAst.make newid in
+      (newref, sch_isdep sch_type, ind, sch_sort) :: name_and_process_schemes env q
+
+let do_mutual_induction_scheme ?(force_mutual=false) env l =
+  let lrecnames = List.map (fun ({CAst.v},_,_,_) -> v) l in
 
-let do_mutual_induction_scheme ?(force_mutual=false) lnamedepindsort =
-  let lrecnames = List.map (fun ({CAst.v},_,_,_) -> v) lnamedepindsort
-  and env0 = Global.env() in
   let sigma, lrecspec, _ =
     List.fold_right
       (fun (_,dep,ind,sort) (evd, l, inst) ->
        let evd, indu, inst =
          match inst with
          | None ->
-            let _, ctx = Typeops.type_of_global_in_context env0 (GlobRef.IndRef ind) in
+            let _, ctx = Typeops.type_of_global_in_context env (Names.GlobRef.IndRef ind) in
             let u, ctx = UnivGen.fresh_instance_from ctx None in
             let evd = Evd.from_ctx (UState.of_context_set ctx) in
               evd, (ind,u), Some u
          | Some ui -> evd, (ind, ui), inst
        in
           (evd, (indu,dep,sort) :: l, inst))
-    lnamedepindsort (Evd.from_env env0,[],None)
+    l (Evd.from_env env,[],None)
   in
-  let sigma, listdecl = Indrec.build_mutual_induction_scheme env0 sigma ~force_mutual lrecspec in
+  let sigma, listdecl = Indrec.build_mutual_induction_scheme env sigma ~force_mutual lrecspec in
   let poly =
     (* NB: build_mutual_induction_scheme forces nonempty list of mutual inductives
        (force_mutual is about the generated schemes) *)
-    let _,_,ind,_ = List.hd lnamedepindsort in
-    Global.is_polymorphic (GlobRef.IndRef ind)
+    let _,_,ind,_ = List.hd l in
+    Global.is_polymorphic (Names.GlobRef.IndRef ind)
   in
   let declare decl fi lrecref =
-    let decltype = Retyping.get_type_of env0 sigma (EConstr.of_constr decl) in
+    let decltype = Retyping.get_type_of env sigma (EConstr.of_constr decl) in
     let decltype = EConstr.to_constr sigma decltype in
     let cst = define ~poly fi sigma decl (Some decltype) in
-    GlobRef.ConstRef cst :: lrecref
+    Names.GlobRef.ConstRef cst :: lrecref
   in
   let _ = List.fold_right2 declare listdecl lrecnames [] in
-  fixpoint_message None lrecnames
+  Declare.fixpoint_message None lrecnames
 
-let get_common_underlying_mutual_inductive env = function
-  | [] -> assert false
-  | (id,(mind,i as ind))::l as all ->
-      match List.filter (fun (_,(mind',_)) -> not (MutInd.equal mind mind')) l with
-      | (_,ind')::_ ->
-          raise (RecursionSchemeError (env, NotMutualInScheme (ind,ind')))
-      | [] ->
-          if not (List.distinct_f Int.compare (List.map snd (List.map snd all)))
-          then user_err Pp.(str "A type occurs twice");
-          mind,
-          List.map_filter
-            (function (Some id,(_,i)) -> Some (i,id.CAst.v) | (None,_) -> None) all
+let do_scheme env l =
+  let lnamedepindsort = name_and_process_schemes env l in
+  do_mutual_induction_scheme env lnamedepindsort
 
-let do_scheme l =
-  let env = Global.env() in
-  let ischeme,escheme = split_scheme env l in
-(* we want 1 kind of scheme at a time so we check if the user
-tried to declare different schemes at once *)
-    if not (List.is_empty ischeme) && not (List.is_empty escheme)
-    then
-      user_err Pp.(str "Do not declare equality and induction scheme at the same time.")
-    else (
-      if not (List.is_empty ischeme) then do_mutual_induction_scheme ischeme
-      else
-        let mind,l = get_common_underlying_mutual_inductive env escheme in
-        declare_beq_scheme_with l mind;
-        declare_eq_decidability_scheme_with l mind
-    )
+let do_scheme_equality sch id =
+  let mind,_ = Smartlocate.smart_global_inductive id in
+  let dec = match sch with SchemeBooleanEquality -> false | SchemeEquality -> true in
+  declare_beq_scheme mind;
+  if dec then declare_eq_decidability mind
 
 (**********************************************************************)
 (* Combined scheme *)
@@ -446,11 +424,11 @@ let fold_left' f = function
     [] -> invalid_arg "fold_left'"
   | hd :: tl -> List.fold_left f hd tl
 
-let mk_coq_and sigma = Evarutil.new_global sigma (Coqlib.lib_ref "core.and.type")
-let mk_coq_conj sigma = Evarutil.new_global sigma (Coqlib.lib_ref "core.and.conj")
+let mk_coq_and sigma = Evd.fresh_global (Global.env ()) sigma (Coqlib.lib_ref "core.and.type")
+let mk_coq_conj sigma = Evd.fresh_global (Global.env ()) sigma (Coqlib.lib_ref "core.and.conj")
 
-let mk_coq_prod sigma = Evarutil.new_global sigma (Coqlib.lib_ref "core.prod.type")
-let mk_coq_pair sigma = Evarutil.new_global sigma (Coqlib.lib_ref "core.prod.intro")
+let mk_coq_prod sigma = Evd.fresh_global (Global.env ()) sigma (Coqlib.lib_ref "core.prod.type")
+let mk_coq_pair sigma = Evd.fresh_global (Global.env ()) sigma (Coqlib.lib_ref "core.prod.intro")
 
 let build_combined_scheme env schemes =
   let sigma = Evd.from_env env in
@@ -461,11 +439,11 @@ let build_combined_scheme env schemes =
     let (ctx, arity) = decompose_prod ty in
     let (_, last) = List.hd ctx in
       match Constr.kind last with
-        | App (ind, args) ->
-            let ind = destInd ind in
+        | Constr.App (ind, args) ->
+            let ind = Constr.destInd ind in
             let (_,spec) = Inductive.lookup_mind_specif env (fst ind) in
               ctx, ind, spec.mind_nrealargs
-        | _ -> ctx, destInd last, 0
+        | _ -> ctx, Constr.destInd last, 0
   in
   let (c, t) = List.hd defs in
   let ctx, ind, nargs = find_inductive t in
@@ -485,23 +463,23 @@ let build_combined_scheme env schemes =
     else (mk_coq_prod, mk_coq_pair)
   in
   (* Number of clauses, including the predicates quantification *)
-  let prods = nb_prod sigma (EConstr.of_constr t) - (nargs + 1) in
+  let prods = Termops.nb_prod sigma (EConstr.of_constr t) - (nargs + 1) in
   let sigma, coqand  = mk_and sigma in
   let sigma, coqconj = mk_conj sigma in
-  let relargs = rel_vect 0 prods in
+  let relargs = Termops.rel_vect 0 prods in
   let concls = List.rev_map
     (fun (cst, t) ->
-      mkApp(mkConstU cst, relargs),
+      Constr.mkApp(Constr.mkConstU cst, relargs),
       snd (decompose_prod_n prods t)) defs in
   let concl_bod, concl_typ =
     fold_left'
       (fun (accb, acct) (cst, x) ->
-        mkApp (EConstr.to_constr sigma coqconj, [| x; acct; cst; accb |]),
-        mkApp (EConstr.to_constr sigma coqand, [| x; acct |])) concls
+        Constr.mkApp (EConstr.to_constr sigma coqconj, [| x; acct; cst; accb |]),
+        Constr.mkApp (EConstr.to_constr sigma coqand, [| x; acct |])) concls
   in
   let ctx, _ =
     list_split_rev_at prods
-      (List.rev_map (fun (x, y) -> LocalAssum (x, y)) ctx) in
+      (List.rev_map (fun (x, y) -> Context.Rel.Declaration.LocalAssum (x, y)) ctx) in
   let typ = List.fold_left (fun d c -> Term.mkProd_wo_LetIn c d) concl_typ ctx in
   let body = it_mkLambda_or_LetIn concl_bod ctx in
   let sigma = Typing.check env sigma (EConstr.of_constr body) (EConstr.of_constr typ) in
@@ -511,9 +489,11 @@ let do_combined_scheme name schemes =
   let open CAst in
   let csts =
     List.map (fun {CAst.loc;v} ->
-        let qualid = qualid_of_ident v in
+        let qualid = Libnames.qualid_of_ident v in
         try Nametab.locate_constant qualid
-        with Not_found -> user_err ?loc Pp.(pr_qualid qualid ++ str " is not declared."))
+        with
+          Not_found -> CErrors.user_err ?loc
+            Pp.(Libnames.pr_qualid qualid ++ str " is not declared."))
       schemes
   in
   let sigma,body,typ = build_combined_scheme (Global.env ()) csts in
@@ -523,9 +503,9 @@ let do_combined_scheme name schemes =
      polymorphism of the inductive block). In that case if they want
      some other polymorphism they can also manually define the
      combined scheme. *)
-  let poly = Global.is_polymorphic (GlobRef.ConstRef (List.hd csts)) in
+  let poly = Global.is_polymorphic (Names.GlobRef.ConstRef (List.hd csts)) in
   ignore (define ~poly name.v sigma body (Some typ));
-  fixpoint_message None [name.v]
+  Declare.fixpoint_message None [name.v]
 
 (**********************************************************************)
 
