@@ -13,7 +13,6 @@ open CErrors
 open Util
 
 open Names
-open Lib
 open Libobject
 
 (************************************************************************)
@@ -119,6 +118,8 @@ let libraries_filename_table = ref DPmap.empty
 (* These are the _ordered_ sets of loaded, imported and exported libraries *)
 let libraries_loaded_list = Summary.ref [] ~name:"LIBRARY-LOAD"
 
+(* Opaque proof tables *)
+
 (* various requests to the tables *)
 
 let find_library dir =
@@ -127,8 +128,8 @@ let find_library dir =
 let try_find_library dir =
   try find_library dir
   with Not_found ->
-    user_err ~hdr:"Library.find_library"
-      (str "Unknown library " ++ DirPath.print dir)
+    user_err
+      (str "Unknown library " ++ DirPath.print dir ++ str ".")
 
 let register_library_filename dir f =
   (* Not synchronized: overwrite the previous binding if one existed *)
@@ -140,12 +141,6 @@ let library_full_filename dir =
   try DPmap.find dir !libraries_filename_table
   with Not_found -> "<unavailable filename>"
 
-let overwrite_library_filenames f =
-  let f =
-    if Filename.is_relative f then Filename.concat (Sys.getcwd ()) f else f in
-  DPmap.iter (fun dir _ -> register_library_filename dir f)
-    !libraries_table
-
 let library_is_loaded dir =
   try let _ = find_library dir in true
   with Not_found -> false
@@ -155,17 +150,13 @@ let library_is_loaded dir =
 
 let register_loaded_library m =
   let libname = m.libsum_name in
-  let link () =
-    let dirname = Filename.dirname (library_full_filename libname) in
-    let prefix = Nativecode.mod_uid_of_dirpath libname ^ "." in
-    let f = prefix ^ "cmo" in
-    let f = Dynlink.adapt_filename f in
-    Nativelib.link_library (Global.env()) ~prefix ~dirname ~basename:f
-  in
   let rec aux = function
     | [] ->
-      let () = if Flags.get_native_compiler () then link () in
-      [libname]
+        if Flags.get_native_compiler () then begin
+            let dirname = Filename.dirname (library_full_filename libname) in
+            Nativelib.enable_library dirname libname
+          end;
+        [libname]
     | m'::_ as l when DirPath.equal m' libname -> l
     | m'::l' -> m' :: aux l' in
   libraries_loaded_list := aux !libraries_loaded_list;
@@ -173,23 +164,14 @@ let register_loaded_library m =
 
   let loaded_libraries () = !libraries_loaded_list
 
-(************************************************************************)
-(** {6 Tables of opaque proof terms} *)
-
-(** We now store opaque proof terms apart from the rest of the environment.
-    See the [Indirect] constructor in [Lazyconstr.lazy_constr]. This way,
-    we can quickly load a first half of a .vo file without these opaque
-    terms, and access them only when a specific command (e.g. Print or
-    Print Assumptions) needs it. *)
-
 (** Delayed / available tables of opaque terms *)
 
-type 'a table_status =
-  | ToFetch of 'a array delayed
-  | Fetched of 'a array
+type table_status =
+  | ToFetch of Opaques.opaque_disk delayed
+  | Fetched of Opaques.opaque_disk
 
 let opaque_tables =
-  ref (DPmap.empty : (Opaqueproof.opaque_proofterm table_status) DPmap.t)
+  ref (DPmap.empty : table_status DPmap.t)
 
 let add_opaque_table dp st =
   opaque_tables := DPmap.add dp st !opaque_tables
@@ -203,7 +185,7 @@ let access_table what tables dp i =
       let t =
         try fetch_delayed f
         with Faulty f ->
-          user_err ~hdr:"Library.access_table"
+          user_err
             (str "The file " ++ str f ++ str " (bound to " ++ str dir_path ++
              str ") is corrupted,\ncannot load some " ++
              str what ++ str " in it.\n")
@@ -211,15 +193,26 @@ let access_table what tables dp i =
       tables := DPmap.add dp (Fetched t) !tables;
       t
   in
-  assert (i < Array.length t); t.(i)
+  Opaques.get_opaque_disk i t
 
-let access_opaque_table dp i =
-  let what = "opaque proofs" in
-  access_table what opaque_tables dp i
+let access_opaque_table o =
+  let (sub, ci, dp, i) = Opaqueproof.repr o in
+  let ans =
+    if DirPath.equal dp (Global.current_dirpath ()) then
+      Opaques.get_current_opaque i
+    else
+      let what = "opaque proofs" in
+      access_table what opaque_tables dp i
+  in
+  match ans with
+  | None -> None
+  | Some (c, ctx) ->
+    let (c, ctx) = Discharge.cook_opaque_proofterm ci (c, ctx) in
+    let c = Mod_subst.(List.fold_right subst_mps sub c) in
+    Some (c, ctx)
 
 let indirect_accessor = {
-  Opaqueproof.access_proof = access_opaque_table;
-  Opaqueproof.access_discharge = Cooking.cook_constr;
+  Global.access_proof = access_opaque_table;
 }
 
 (************************************************************************)
@@ -229,7 +222,7 @@ type seg_sum = summary_disk
 type seg_lib = library_disk
 type seg_univ = (* true = vivo, false = vi *)
   Univ.ContextSet.t * bool
-type seg_proofs = Opaqueproof.opaque_proofterm array
+type seg_proofs = Opaques.opaque_disk
 
 let mk_library sd md digests univs =
   {
@@ -275,10 +268,10 @@ let rec intern_library ~lib_resolver (needed, contents) (dir, f) from =
   let f = match f with Some f -> f | None -> lib_resolver dir in
   let m = intern_from_file f in
   if not (DirPath.equal dir m.library_name) then
-    user_err ~hdr:"load_physical_library"
+    user_err
       (str "The file " ++ str f ++ str " contains library" ++ spc () ++
        DirPath.print m.library_name ++ spc () ++ str "and not library" ++
-       spc() ++ DirPath.print dir);
+       spc() ++ DirPath.print dir ++ str ".");
   Feedback.feedback (Feedback.FileLoaded(DirPath.to_string dir, f));
   m.library_digests, intern_library_deps ~lib_resolver (needed, contents) dir m f
 
@@ -336,77 +329,59 @@ let register_library m =
    - called at module or module type closing when a Require occurs in
      the module or module type
    - not called from a library (i.e. a module identified with a file) *)
-let load_require _ (_,(needed,modl,_)) =
+let load_require _ needed =
   List.iter register_library needed
 
-let open_require i (_,(_,modl,export)) =
-  Option.iter (fun export ->
-      let mpl = List.map (fun m -> Unfiltered, MPfile m) modl in
-      (* TODO support filters in Require *)
-      Declaremods.import_modules ~export mpl)
-    export
-
   (* [needed] is the ordered list of libraries not already loaded *)
 let cache_require o =
-  load_require 1 o;
-  open_require 1 o
+  load_require 1 o
 
-let discharge_require (_,o) = Some o
+let discharge_require o = Some o
 
 (* open_function is never called from here because an Anticipate object *)
 
-type require_obj = library_t list * DirPath.t list * bool option
+type require_obj = library_t list
 
 let in_require : require_obj -> obj =
-  declare_object {(default_object "REQUIRE") with
-       cache_function = cache_require;
-       load_function = load_require;
-       open_function = (fun _ _ -> assert false);
-       discharge_function = discharge_require;
-       classify_function = (fun o -> Anticipate o) }
+  declare_object
+    {(default_object "REQUIRE") with
+     cache_function = cache_require;
+     load_function = load_require;
+     open_function = (fun _ _ -> assert false);
+     discharge_function = discharge_require;
+     classify_function = (fun o -> Anticipate) }
 
 (* Require libraries, import them if [export <> None], mark them for export
    if [export = Some true] *)
 
 let warn_require_in_module =
-  CWarnings.create ~name:"require-in-module" ~category:"deprecated"
-                   (fun () -> strbrk "Require inside a module is" ++
-                              strbrk " deprecated and strongly discouraged. " ++
-                              strbrk "You can Require a module at toplevel " ++
-                              strbrk "and optionally Import it inside another one.")
+  CWarnings.create ~name:"require-in-module" ~category:"fragile"
+    (fun () -> strbrk "Use of “Require” inside a module is fragile." ++ spc() ++
+               strbrk "It is not recommended to use this functionality in finished proof scripts.")
 
-let require_library_from_dirpath ~lib_resolver modrefl export =
+let require_library_from_dirpath ~lib_resolver modrefl =
   let needed, contents = List.fold_left (rec_intern_library ~lib_resolver) ([], DPmap.empty) modrefl in
   let needed = List.rev_map (fun dir -> DPmap.find dir contents) needed in
-  let modrefl = List.map fst modrefl in
-  if Lib.is_module_or_modtype () then
-    begin
-      warn_require_in_module ();
-      add_anonymous_leaf (in_require (needed,modrefl,None));
-      Option.iter (fun export ->
-          (* TODO import filters *)
-          List.iter (fun m -> Declaremods.import_module Unfiltered ~export (MPfile m)) modrefl)
-        export
-    end
-  else
-    add_anonymous_leaf (in_require (needed,modrefl,export));
-  ()
+  if Lib.is_module_or_modtype () then warn_require_in_module ();
+  Lib.add_leaf (in_require needed)
 
 (************************************************************************)
 (*s Initializing the compilation of a library. *)
 
+type ('uid, 'doc) tasks = (('uid, 'doc) Stateid.request * bool) list
+
 let load_library_todo f =
   let ch = raw_intern_library f in
   let (s0 : seg_sum), _ = ObjFile.marshal_in_segment ch ~segment:"summary" in
   let (s1 : seg_lib), _ = ObjFile.marshal_in_segment ch ~segment:"library" in
   let (s2 : seg_univ option), _ = ObjFile.marshal_in_segment ch ~segment:"universes" in
-  let tasks, _ = ObjFile.marshal_in_segment ch ~segment:"tasks" in
+  let (tasks : (Opaqueproof.opaque_handle option, 'doc) tasks option), _ = ObjFile.marshal_in_segment ch ~segment:"tasks" in
   let (s4 : seg_proofs), _ = ObjFile.marshal_in_segment ch ~segment:"opaques" in
   ObjFile.close_in ch;
   System.check_caml_version ~caml:s0.md_ocaml ~file:f;
-  if tasks = None then user_err ~hdr:"restart" (str"not a .vio file");
-  if s2 = None then user_err ~hdr:"restart" (str"not a .vio file");
-  if snd (Option.get s2) then user_err ~hdr:"restart" (str"not a .vio file");
+  if tasks = None then user_err (str "Not a .vio file.");
+  if s2 = None then user_err (str "Not a .vio file.");
+  if snd (Option.get s2) then user_err (str "Not a .vio file.");
   s0, s1, Option.get s2, Option.get tasks, s4
 
 (************************************************************************)
@@ -425,6 +400,11 @@ let error_recursively_dependent_library dir =
      strbrk " to save current library because" ++
      strbrk " it already depends on a library of this name.")
 
+type 'doc todo_proofs =
+ | ProofsTodoNone (* for .vo *)
+ | ProofsTodoSomeEmpty of Future.UUIDSet.t (* for .vos *)
+ | ProofsTodoSome of Future.UUIDSet.t * (Future.UUID.t, 'doc) tasks (* for .vio *)
+
 (* We now use two different digests in a .vo file. The first one
    only covers half of the file, without the opaque table. It is
    used for identifying this version of this library : this digest
@@ -442,7 +422,7 @@ let save_library_base f sum lib univs tasks proofs =
     ObjFile.marshal_out_segment ch ~segment:"summary" (sum    : seg_sum);
     ObjFile.marshal_out_segment ch ~segment:"library" (lib    : seg_lib);
     ObjFile.marshal_out_segment ch ~segment:"universes" (univs  : seg_univ option);
-    ObjFile.marshal_out_segment ch ~segment:"tasks" (tasks  : 'tasks option);
+    ObjFile.marshal_out_segment ch ~segment:"tasks" (tasks : (Opaqueproof.opaque_handle option, 'doc) tasks option);
     ObjFile.marshal_out_segment ch ~segment:"opaques" (proofs : seg_proofs);
     ObjFile.close_out ch
   with reraise ->
@@ -452,12 +432,7 @@ let save_library_base f sum lib univs tasks proofs =
     Sys.remove f;
     Exninfo.iraise reraise
 
-type ('document,'counters) todo_proofs =
- | ProofsTodoNone (* for .vo *)
- | ProofsTodoSomeEmpty of Future.UUIDSet.t (* for .vos *)
- | ProofsTodoSome of Future.UUIDSet.t * ((Future.UUID.t,'document) Stateid.request * bool) list * 'counters (* for .vio *)
-
-let save_library_to todo_proofs ~output_native_objects dir f otab =
+let save_library_to todo_proofs ~output_native_objects dir f =
   assert(
     let expected_extension = match todo_proofs with
       | ProofsTodoNone -> ".vo"
@@ -468,22 +443,27 @@ let save_library_to todo_proofs ~output_native_objects dir f otab =
   let except = match todo_proofs with
     | ProofsTodoNone -> Future.UUIDSet.empty
     | ProofsTodoSomeEmpty except -> except
-    | ProofsTodoSome (except,l,_) -> except
+    | ProofsTodoSome (except,l) -> except
     in
-  let cenv, seg, ast = Declaremods.end_library ~output_native_objects ~except dir in
-  let opaque_table, f2t_map = Opaqueproof.dump ~except otab in
+  (* Ensure that the call below is performed with all opaques joined *)
+  let () = Opaques.Summary.join ~except () in
+  let opaque_table, f2t_map = Opaques.dump ~except () in
+  let () = assert (not (Future.UUIDSet.is_empty except) ||
+    Safe_typing.is_joined_environment (Global.safe_env ()))
+  in
+  let cenv, seg, ast = Declaremods.end_library ~output_native_objects dir in
   let tasks, utab =
     match todo_proofs with
     | ProofsTodoNone -> None, None
     | ProofsTodoSomeEmpty _except ->
       None, Some (Univ.ContextSet.empty,false)
-    | ProofsTodoSome (_except, tasks, rcbackup) ->
+    | ProofsTodoSome (_except, tasks) ->
       let tasks =
         List.map Stateid.(fun (r,b) ->
-            try { r with uuid = Future.UUIDMap.find r.uuid f2t_map }, b
-            with Not_found -> assert b; { r with uuid = -1 }, b)
+            try { r with uuid = Some (Future.UUIDMap.find r.uuid f2t_map) }, b
+            with Not_found -> assert b; { r with uuid = None }, b)
           tasks in
-      Some (tasks,rcbackup),
+      Some tasks,
       Some (Univ.ContextSet.empty,false)
     in
     let sd = {
@@ -502,7 +482,7 @@ let save_library_to todo_proofs ~output_native_objects dir f otab =
   (* Writing native code files *)
   if output_native_objects then
     let fn = Filename.dirname f ^"/"^ Nativecode.mod_uid_of_dirpath dir in
-    Nativelib.compile_library dir ast fn
+    Nativelib.compile_library ast fn
 
 let save_library_raw f sum lib univs proofs =
   save_library_base f sum lib (Some univs) None proofs
@@ -514,12 +494,3 @@ let get_used_load_paths () =
        String.Set.empty !libraries_loaded_list)
 
 let _ = Nativelib.get_load_paths := get_used_load_paths
-
-(* These commands may not be very safe due to ML-side plugin loading
-   etc... use at your own risk *)
-let extern_state s =
-  System.extern_state Coq_config.state_magic_number s (States.freeze ~marshallable:true)
-
-let intern_state s =
-  States.unfreeze (System.with_magic_number_check (System.intern_state Coq_config.state_magic_number) s);
-  overwrite_library_filenames s
