@@ -29,10 +29,11 @@ type printable =
   | PrintFullContext
   | PrintSectionContext of qualid
   | PrintInspect of int
-  | PrintGrammar of string
+  | PrintGrammar of string list
   | PrintCustomGrammar of string
+  | PrintKeywords
   | PrintLoadPath of DirPath.t option
-  | PrintModules
+  | PrintLibraries
   | PrintModule of qualid
   | PrintModuleType of qualid
   | PrintNamespace of DirPath.t
@@ -42,7 +43,7 @@ type printable =
   | PrintName of qualid or_by_notation * UnivNames.univ_name_list option
   | PrintGraph
   | PrintClasses
-  | PrintTypeClasses
+  | PrintTypeclasses
   | PrintInstances of qualid or_by_notation
   | PrintCoercions
   | PrintCoercionPaths of class_rawexpr * class_rawexpr
@@ -60,6 +61,7 @@ type printable =
   | PrintAssumptions of bool * bool * qualid or_by_notation
   | PrintStrategy of qualid or_by_notation option
   | PrintRegistered
+  | PrintNotation of Constrexpr.notation_entry * string
 
 type glob_search_where = InHyp | InConcl | Anywhere
 
@@ -75,7 +77,6 @@ type search_request =
 type searchable =
   | SearchPattern of constr_pattern_expr
   | SearchRewrite of constr_pattern_expr
-  | SearchHead of constr_pattern_expr
   | Search of (bool * search_request) list
 
 type locatable =
@@ -105,21 +106,28 @@ type search_restriction =
   | SearchOutside of qualid list
 
 type verbose_flag   = bool (* true = Verbose;       false = Silent         *)
-type coercion_flag  = bool (* true = AddCoercion    false = NoCoercion     *)
-type instance_flag  = bool option
-  (* Some true = Backward instance; Some false = Forward instance, None = NoInstance *)
+type coercion_flag  = AddCoercion | NoCoercion
+(* Remove BackInstanceWarning at end of deprecation phase
+   (this is just to print a warning when :> is used instead of ::
+   to declare instances in classes) *)
+type instance_flag  = BackInstance | BackInstanceWarning | NoInstance
 
-type export_flag    = bool (* true = Export;        false = Import         *)
+type export_flag = Lib.export_flag = Export | Import
+
+type import_categories = {
+  negative : bool;
+  import_cats : string CAst.t list;
+}
+
+type export_with_cats = export_flag * import_categories option
+
+type infix_flag     = bool (* true = Infix;         false = Notation       *)
 
 type one_import_filter_name = qualid * bool (* import inductive components *)
 type import_filter_expr =
   | ImportAll
   | ImportNames of one_import_filter_name list
 
-type onlyparsing_flag = { onlyparsing : bool }
- (* Some v = Parse only;  None = Print also.
-    If v<>Current, it contains the name of the coq version
-    which this notation is trying to be compatible with *)
 type locality_flag  = bool (* true = Local *)
 
 type option_setting =
@@ -130,28 +138,35 @@ type option_setting =
 
 (** Identifier and optional list of bound universes and constraints. *)
 
-type sort_expr = Sorts.family
-
 type definition_expr =
   | ProveBody of local_binder_expr list * constr_expr
   | DefineBody of local_binder_expr list * Genredexpr.raw_red_expr option * constr_expr
       * constr_expr option
 
+type notation_format =
+  | TextFormat of lstring
+
 type syntax_modifier =
   | SetItemLevel of string list * Notation_term.constr_as_binder_kind option * Extend.production_level
+  | SetItemScope of string list * scope_name
   | SetLevel of int
   | SetCustomEntry of string * int option
   | SetAssoc of Gramlib.Gramext.g_assoc
   | SetEntryType of string * Extend.simple_constr_prod_entry_key
   | SetOnlyParsing
   | SetOnlyPrinting
-  | SetFormat of string * lstring
+  | SetFormat of notation_format
+
+type notation_enable_modifier =
+  | EnableNotationEntry of notation_entry CAst.t
+  | EnableNotationOnly of Notationextern.notation_use
+  | EnableNotationAll
 
 type decl_notation =
   { decl_ntn_string : lstring
   ; decl_ntn_interp : constr_expr
   ; decl_ntn_scope : scope_name option
-  ; decl_ntn_modifiers : syntax_modifier list
+  ; decl_ntn_modifiers : syntax_modifier CAst.t list
   }
 
 type 'a fix_expr_gen =
@@ -168,30 +183,35 @@ type fixpoint_expr = recursion_order_expr option fix_expr_gen
 type cofixpoint_expr = unit fix_expr_gen
 
 type local_decl_expr =
-  | AssumExpr of lname * constr_expr
-  | DefExpr of lname * constr_expr * constr_expr option
+  | AssumExpr of lname * local_binder_expr list * constr_expr
+  | DefExpr of lname * local_binder_expr list * constr_expr * constr_expr option
 
 type inductive_kind = Inductive_kw | CoInductive | Variant | Record | Structure | Class of bool (* true = definitional, false = inductive *)
 type simple_binder = lident list  * constr_expr
 type class_binder = lident * constr_expr list
 type 'a with_coercion = coercion_flag * 'a
+type 'a with_coercion_instance = (coercion_flag * instance_flag) * 'a
 (* Attributes of a record field declaration *)
 type record_field_attr = {
-  rf_subclass: instance_flag; (* the projection is an implicit coercion or an instance *)
+  rf_coercion: coercion_flag; (* the projection is an implicit coercion *)
+  rf_reversible: bool option; (* coercion is reversible, if relevant *)
+  rf_instance: instance_flag; (* the projection is an instance *)
   rf_priority: int option; (* priority of the instance, if relevant *)
+  rf_locality: Goptions.option_locality; (* locality of coercion and instance *)
   rf_notation: decl_notation list;
   rf_canonical: bool; (* use this projection in the search for canonical instances *)
   }
-type constructor_expr = (lident * constr_expr) with_coercion
+type constructor_expr = (lident * constr_expr) with_coercion_instance
 type constructor_list_or_record_decl_expr =
   | Constructors of constructor_expr list
-  | RecordDecl of lident option * (local_decl_expr * record_field_attr) list
+  | RecordDecl of lident option * (local_decl_expr * record_field_attr) list * lident option
 type inductive_params_expr = local_binder_expr list * local_binder_expr list option
 (** If the option is nonempty the "|" marker was used *)
 
 type inductive_expr =
-  ident_decl with_coercion * inductive_params_expr * constr_expr option *
-    constructor_list_or_record_decl_expr
+  cumul_ident_decl with_coercion
+  * inductive_params_expr * constr_expr option
+  * constructor_list_or_record_decl_expr
 
 type one_inductive_expr =
   lident * inductive_params_expr * constr_expr option * constructor_expr list
@@ -209,10 +229,22 @@ type proof_end =
   (*                         name in `Save ident` when closing goal *)
   | Proved of opacity_flag * lident option
 
-type scheme =
-  | InductionScheme of bool * qualid or_by_notation * sort_expr
-  | CaseScheme of bool * qualid or_by_notation * sort_expr
-  | EqualityScheme of qualid or_by_notation
+type scheme_type =
+  | SchemeInduction
+  | SchemeMinimality
+  | SchemeElimination
+  | SchemeCase
+
+type equality_scheme_type =
+  | SchemeBooleanEquality
+  | SchemeEquality
+
+  (* The data of a Scheme decleration *)
+type scheme = {
+  sch_type : scheme_type ;
+  sch_qualid : Libnames.qualid Constrexpr.or_by_notation ;
+  sch_sort : Sorts.family ;
+}
 
 type section_subset_expr =
   | SsEmpty
@@ -259,14 +291,14 @@ type register_kind =
 (** {6 Types concerning the module layer} *)
 
 type module_ast_inl = module_ast * Declaremods.inline
-type module_binder = bool option * lident list * module_ast_inl
+type module_binder = export_with_cats option * lident list * module_ast_inl
 
 (** {6 The type of vernacular expressions} *)
 
 type vernac_one_argument_status = {
   name : Name.t;
   recarg_like : bool;
-  notation_scope : string CAst.t option;
+  notation_scope : string CAst.t list;
   implicit_status : Glob_term.binding_kind;
 }
 
@@ -315,18 +347,16 @@ type nonrec vernac_expr =
 
   | VernacLoad of verbose_flag * string
   (* Syntax *)
-  | VernacSyntaxExtension of bool * (lstring * syntax_modifier list)
+  | VernacReservedNotation of infix_flag * (lstring * syntax_modifier CAst.t list)
   | VernacOpenCloseScope of bool * scope_name
   | VernacDeclareScope of scope_name
   | VernacDelimiters of scope_name * string option
   | VernacBindScope of scope_name * class_rawexpr list
-  | VernacInfix of (lstring * syntax_modifier list) *
-      constr_expr * scope_name option
   | VernacNotation of
-      constr_expr * (lstring * syntax_modifier list) *
+      infix_flag * constr_expr * (lstring * syntax_modifier CAst.t list) *
       scope_name option
-  | VernacNotationAddFormat of string * string * string
   | VernacDeclareCustomEntry of string
+  | VernacEnableNotation of bool * (string, Id.t list * qualid) Util.union option * constr_expr option * notation_enable_modifier list * notation_with_optional_scope option
 
   (* Gallina *)
   | VernacDefinition of (discharge * Decls.definition_object_kind) * name_decl * definition_expr
@@ -339,19 +369,21 @@ type nonrec vernac_expr =
   | VernacFixpoint of discharge * fixpoint_expr list
   | VernacCoFixpoint of discharge * cofixpoint_expr list
   | VernacScheme of (lident option * scheme) list
+  | VernacSchemeEquality of equality_scheme_type * Libnames.qualid Constrexpr.or_by_notation
   | VernacCombinedScheme of lident * lident list
   | VernacUniverse of lident list
-  | VernacConstraint of Glob_term.glob_constraint list
+  | VernacConstraint of univ_constraint_expr list
 
   (* Gallina extensions *)
   | VernacBeginSection of lident
   | VernacEndSegment of lident
+  | VernacExtraDependency of qualid * string * Id.t option
   | VernacRequire of
-      qualid option * export_flag option * qualid list
-  | VernacImport of export_flag * (qualid * import_filter_expr) list
+      qualid option * export_with_cats option * (qualid * import_filter_expr) list
+  | VernacImport of export_with_cats * (qualid * import_filter_expr) list
   | VernacCanonical of qualid or_by_notation
   | VernacCoercion of qualid or_by_notation *
-      class_rawexpr * class_rawexpr
+      (class_rawexpr * class_rawexpr) option
   | VernacIdentityCoercion of lident * class_rawexpr * class_rawexpr
   | VernacNameSectionHypSet of lident * section_subset_expr
 
@@ -377,18 +409,14 @@ type nonrec vernac_expr =
   | VernacExistingClass of qualid (* inductive or definition name *)
 
   (* Modules and Module Types *)
-  | VernacDeclareModule of bool option * lident *
+  | VernacDeclareModule of export_with_cats option * lident *
       module_binder list * module_ast_inl
-  | VernacDefineModule of bool option * lident * module_binder list *
+  | VernacDefineModule of export_with_cats option * lident * module_binder list *
       module_ast_inl Declaremods.module_signature * module_ast_inl list
   | VernacDeclareModuleType of lident *
       module_binder list * module_ast_inl list * module_ast_inl list
   | VernacInclude of module_ast_inl list
 
-  (* Solving *)
-
-  | VernacSolveExistential of int * constr_expr
-
   (* Auxiliary file and library management *)
   | VernacAddLoadPath of { implicit : bool
                          ; physical_path : CUnix.physical_path
@@ -400,10 +428,6 @@ type nonrec vernac_expr =
   | VernacDeclareMLModule of string list
   | VernacChdir of string option
 
-  (* State management *)
-  | VernacWriteState of string
-  | VernacRestoreState of string
-
   (* Resetting *)
   | VernacResetName of lident
   | VernacResetInitial
@@ -414,8 +438,7 @@ type nonrec vernac_expr =
   | VernacRemoveHints of string list * qualid list
   | VernacHints of string list * hints_expr
   | VernacSyntacticDefinition of
-      lident * (Id.t list * constr_expr) *
-      onlyparsing_flag
+      lident * (Id.t list * constr_expr) * syntax_modifier CAst.t list
   | VernacArguments of
       qualid or_by_notation *
       vernac_argument_status list (* Main arguments status list *) *
@@ -438,11 +461,11 @@ type nonrec vernac_expr =
   | VernacSearch of searchable * Goal_select.t option * search_restriction
   | VernacLocate of locatable
   | VernacRegister of qualid * register_kind
-  | VernacPrimitive of lident * CPrimitives.op_or_type * constr_expr option
+  | VernacPrimitive of ident_decl * CPrimitives.op_or_type * constr_expr option
   | VernacComments of comment list
 
   (* Proof management *)
-  | VernacAbort of lident option
+  | VernacAbort
   | VernacAbortAll
   | VernacRestart
   | VernacUndo of int
@@ -468,6 +491,7 @@ type control_flag =
   | ControlRedirect of string
   | ControlTimeout of int
   | ControlFail
+  | ControlSucceed
 
 type vernac_control_r =
   { control : control_flag list
