@@ -30,14 +30,13 @@ type vernac_classification =
   | VtQed of vernac_qed_type
   (* A proof step *)
   | VtProofStep of {
-      parallel : [ `Yes of solving_tac * anon_abstracting_tac | `No ];
       proof_block_detection : proof_block_name option
     }
   (* Queries are commands assumed to be "pure", that is to say, they
      don't modify the interpretation state. *)
   | VtQuery
   (* Commands that change the current proof mode *)
-  | VtProofMode of string
+  | VtProofMode of Pvernac.proof_mode
   (* To be removed *)
   | VtMeta
 and vernac_start = opacity_guarantee * Names.Id.t list
@@ -52,16 +51,151 @@ and anon_abstracting_tac = bool (** abstracting anonymously its result *)
 
 and proof_block_name = string (** open type of delimiters *)
 
-type typed_vernac =
-  | VtDefault of (unit -> unit)
-  | VtNoProof of (unit -> unit)
-  | VtCloseProof of (lemma:Lemmas.t -> unit)
-  | VtOpenProof of (unit -> Lemmas.t)
-  | VtModifyProof of (pstate:Declare.Proof.t -> Declare.Proof.t)
-  | VtReadProofOpt of (pstate:Declare.Proof.t option -> unit)
-  | VtReadProof of (pstate:Declare.Proof.t -> unit)
+module InProg = struct
+  type _ t =
+    | Ignore : unit t
+    | Use : Declare.OblState.t t
 
-type vernac_command = atts:Attributes.vernac_flags -> typed_vernac
+  let cast (type a) (x:Declare.OblState.t) (ty:a t) : a =
+    match ty with
+    | Ignore -> ()
+    | Use -> x
+end
+
+module OutProg = struct
+  type _ t =
+    | No : unit t
+    | Yes : Declare.OblState.t t
+    | Push
+    | Pop
+
+  let cast (type a) (x:a) (ty:a t) (orig:Declare.OblState.t NeList.t) : Declare.OblState.t NeList.t  =
+    match ty with
+    | No -> orig
+    | Yes -> NeList.map_head (fun _ -> x) orig
+    | Push -> NeList.push Declare.OblState.empty (Some orig)
+    | Pop -> (match NeList.tail orig with Some tl -> tl | None -> assert false)
+end
+
+module InProof = struct
+  type _ t =
+    | Ignore : unit t
+    | Reject : unit t
+    | Use : Declare.Proof.t t
+    | UseOpt : Declare.Proof.t option t
+
+  let cast (type a) (x:Declare.Proof.t option) (ty:a t) : a =
+    match x, ty with
+    | _, Ignore -> ()
+    | None, Reject -> ()
+    | Some _, Reject -> CErrors.user_err (Pp.str "Command not supported (Open proofs remain).")
+    | Some x, Use -> x
+    | None, Use -> CErrors.user_err (Pp.str "Command not supported (No proof-editing in progress).")
+    | _, UseOpt -> x
+end
+
+module OutProof = struct
+  type _ t =
+    | No : unit t
+    | Close : unit t
+    | Update : Declare.Proof.t t
+    | New : Declare.Proof.t t
+
+end
+
+type ('inprog,'outprog,'inproof,'outproof) vernac_type = {
+  inprog : 'inprog InProg.t;
+  outprog : 'outprog InProg.t;
+  inproof : 'inproof InProof.t;
+  outproof : 'outproof OutProof.t;
+}
+
+type typed_vernac =
+    TypedVernac : {
+      inprog : 'inprog InProg.t;
+      outprog : 'outprog OutProg.t;
+      inproof : 'inproof InProof.t;
+      outproof : 'outproof OutProof.t;
+      run : pm:'inprog -> proof:'inproof -> 'outprog * 'outproof;
+    } -> typed_vernac
+
+let vtdefault f =
+  TypedVernac { inprog = Ignore; outprog = No; inproof = Ignore; outproof = No;
+                run = (fun ~pm:() ~proof:() ->
+                    let () = f () in
+                    (), ()) }
+
+let vtnoproof f =
+  TypedVernac { inprog = Ignore; outprog = No; inproof = Ignore; outproof = No;
+                run = (fun ~pm:() ~proof:() ->
+                    let () = f () in
+                    (), ())
+              }
+
+let vtcloseproof f =
+  TypedVernac { inprog = Use; outprog = Yes; inproof = Use; outproof = Close;
+                run = (fun ~pm ~proof ->
+                    let pm = f ~lemma:proof ~pm in
+                    pm, ())
+              }
+
+let vtopenproof f =
+  TypedVernac { inprog = Ignore; outprog = No; inproof = Ignore; outproof = New;
+                run = (fun ~pm:() ~proof:() ->
+                    let proof = f () in
+                    (), proof)
+              }
+
+let vtmodifyproof f =
+  TypedVernac { inprog = Ignore; outprog = No; inproof = Use; outproof = Update;
+                run = (fun ~pm:() ~proof ->
+                    let proof = f ~pstate:proof in
+                    (), proof)
+              }
+
+let vtreadproofopt f =
+  TypedVernac { inprog = Ignore; outprog = No; inproof = UseOpt; outproof = No;
+                run = (fun ~pm:() ~proof ->
+                    let () = f ~pstate:proof in
+                    (), ())
+              }
+
+let vtreadproof f =
+  TypedVernac { inprog = Ignore; outprog = No; inproof = Use; outproof = No;
+                run = (fun ~pm:() ~proof ->
+                    let () = f ~pstate:proof in
+                    (), ())
+              }
+
+let vtreadprogram f =
+  TypedVernac { inprog = Use; outprog = No; inproof = Ignore; outproof = No;
+                run = (fun ~pm ~proof:() ->
+                    let () = f ~pm in
+                    (), ())
+              }
+
+let vtmodifyprogram f =
+  TypedVernac { inprog = Use; outprog = Yes; inproof = Ignore; outproof = No;
+                run = (fun ~pm ~proof:() ->
+                    let pm = f ~pm in
+                    pm, ())
+              }
+
+let vtdeclareprogram f =
+  TypedVernac { inprog = Use; outprog = No; inproof = Ignore; outproof = New;
+                run = (fun ~pm ~proof:() ->
+                    let proof = f ~pm in
+                    (), proof)
+              }
+
+let vtopenproofprogram f =
+  TypedVernac { inprog = Use; outprog = Yes; inproof = Ignore; outproof = New;
+                run = (fun ~pm ~proof:() ->
+                    let pm, proof = f ~pm in
+                    pm, proof)
+              }
+
+type vernac_command = ?loc:Loc.t -> atts:Attributes.vernac_flags -> unit -> typed_vernac
 
 type plugin_args = Genarg.raw_generic_argument list
 
@@ -74,14 +208,14 @@ let vinterp_add depr s f =
   try
     Hashtbl.add vernac_tab s (depr, f)
   with Failure _ ->
-    user_err ~hdr:"vinterp_add"
+    user_err
       (str"Cannot add the vernac command " ++ str (fst s) ++ str" twice.")
 
 let vinterp_map s =
   try
     Hashtbl.find vernac_tab s
   with Failure _ | Not_found ->
-    user_err ~hdr:"Vernac Interpreter"
+    user_err
       (str"Cannot find vernac command " ++ str (fst s) ++ str".")
 
 let warn_deprecated_command =
@@ -91,7 +225,7 @@ let warn_deprecated_command =
 
 (* Interpretation of a vernac command *)
 
-let type_vernac opn converted_args ~atts =
+let type_vernac opn converted_args ?loc ~atts () =
   let depr, callback = vinterp_map opn in
   let () = if depr then
       let rules = Egramml.get_extend_vernac_rule opn in
@@ -103,7 +237,7 @@ let type_vernac opn converted_args ~atts =
       warn_deprecated_command pr;
   in
   let hunk = callback converted_args in
-  hunk ~atts
+  hunk ?loc ~atts ()
 
 (** VERNAC EXTEND registering *)
 
@@ -120,7 +254,7 @@ let declare_vernac_classifier name f =
 
 let classify_as_query = VtQuery
 let classify_as_sideeff = VtSideff ([], VtLater)
-let classify_as_proofstep = VtProofStep { parallel = `No; proof_block_detection = None}
+let classify_as_proofstep = VtProofStep { proof_block_detection = None}
 
 type (_, _) ty_sig =
 | TyNil : (vernac_command, vernac_classification) ty_sig
@@ -169,9 +303,9 @@ let rec untype_command : type r s. (r, s) ty_sig -> r -> plugin_args -> vernac_c
 let rec untype_user_symbol : type s a b c. (a, b, c) Extend.ty_user_symbol -> (s, Gramlib.Grammar.norec, a) Pcoq.Symbol.t =
   let open Extend in function
   | TUlist1 l -> Pcoq.Symbol.list1 (untype_user_symbol l)
-  | TUlist1sep (l, s) -> Pcoq.Symbol.list1sep (untype_user_symbol l) (Pcoq.Symbol.token (CLexer.terminal s)) false
+  | TUlist1sep (l, s) -> Pcoq.Symbol.list1sep (untype_user_symbol l) (Pcoq.Symbol.tokens [Pcoq.TPattern (CLexer.terminal s)]) false
   | TUlist0 l -> Pcoq.Symbol.list0 (untype_user_symbol l)
-  | TUlist0sep (l, s) -> Pcoq.Symbol.list0sep (untype_user_symbol l) (Pcoq.Symbol.token (CLexer.terminal s)) false
+  | TUlist0sep (l, s) -> Pcoq.Symbol.list0sep (untype_user_symbol l) (Pcoq.Symbol.tokens [Pcoq.TPattern (CLexer.terminal s)]) false
   | TUopt o -> Pcoq.Symbol.opt (untype_user_symbol o)
   | TUentry a -> Pcoq.Symbol.nterm (Pcoq.genarg_grammar (Genarg.ExtraArg a))
   | TUentryl (a, i) -> Pcoq.Symbol.nterml (Pcoq.genarg_grammar (Genarg.ExtraArg a)) (string_of_int i)
@@ -184,7 +318,7 @@ let rec untype_grammar : type r s. (r, s) ty_sig -> 'a Egramml.grammar_prod_item
   let symb = untype_user_symbol tu in
   Egramml.GramNonTerminal (Loc.tag (t, symb)) :: untype_grammar ty
 
-let vernac_extend ~command ?classifier ?entry ext =
+let vernac_extend ?plugin ~command ?classifier ?entry ext =
   let get_classifier (TyML (_, ty, _, cl)) = match cl with
   | Some cl -> untype_classifier ty cl
   | None ->
@@ -220,10 +354,25 @@ let vernac_extend ~command ?classifier ?entry ext =
     let f = untype_command ty f in
     let r = untype_grammar ty in
     let () = vinterp_add depr (command, i) f in
-    Egramml.extend_vernac_command_grammar (command, i) entry r
+    let () =
+      (* allow_override is a hack for Elpi Command, since it takes
+         effect at Import time it gets called multiple times.
+         Eventually we will need a better API to support this and also
+         to support backtracking over it. *)
+      Egramml.declare_vernac_command_grammar ~allow_override:(Option.is_empty plugin)
+        (command, i) entry r
+    in
+    let () = match plugin with
+      | None -> Egramml.extend_vernac_command_grammar ~undoable:false (command, i)
+      | Some plugin ->
+        Mltop.add_init_function plugin (fun () ->
+            Egramml.extend_vernac_command_grammar ~undoable:true (command, i))
+    in
+    ()
   in
   let () = declare_vernac_classifier command cl in
-  List.iteri iter ext
+  let () = List.iteri iter ext in
+  ()
 
 (** VERNAC ARGUMENT EXTEND registering *)
 
@@ -236,15 +385,18 @@ type 'a vernac_argument = {
   arg_parsing : 'a argument_rule;
 }
 
-let vernac_argument_extend ~name arg =
+let vernac_argument_extend ~plugin ~name arg =
   let wit = Genarg.create_arg name in
   let entry = match arg.arg_parsing with
   | Arg_alias e ->
     let () = Pcoq.register_grammar wit e in
     e
   | Arg_rules rules ->
-    let e = Pcoq.create_generic_entry Pcoq.utactic name (Genarg.rawwit wit) in
-    let () = Pcoq.grammar_extend e {Pcoq.pos=None; data=[(None, None, rules)]} in
+    let e = Pcoq.create_generic_entry2 name (Genarg.rawwit wit) in
+    let plugin_uid = (plugin, "vernacargextend:"^name) in
+    let () = Egramml.grammar_extend ~plugin_uid e
+        (Pcoq.Fresh (Gramlib.Gramext.First, [None, None, rules]))
+    in
     e
   in
   let pr = arg.arg_printer in
