@@ -46,14 +46,13 @@ type vernac_classification =
   | VtQed of vernac_qed_type
   (* A proof step *)
   | VtProofStep of {
-      parallel : [ `Yes of solving_tac * anon_abstracting_tac | `No ];
       proof_block_detection : proof_block_name option
     }
   (* Queries are commands assumed to be "pure", that is to say, they
      don't modify the interpretation state. *)
   | VtQuery
   (* Commands that change the current proof mode *)
-  | VtProofMode of string
+  | VtProofMode of Pvernac.proof_mode
   (* To be removed *)
   | VtMeta
 and vernac_start = opacity_guarantee * Names.Id.t list
@@ -70,16 +69,75 @@ and proof_block_name = string (** open type of delimiters *)
 
 (** Interpretation of extended vernac phrases. *)
 
-type typed_vernac =
-  | VtDefault of (unit -> unit)
-  | VtNoProof of (unit -> unit)
-  | VtCloseProof of (lemma:Lemmas.t -> unit)
-  | VtOpenProof of (unit -> Lemmas.t)
-  | VtModifyProof of (pstate:Declare.Proof.t -> Declare.Proof.t)
-  | VtReadProofOpt of (pstate:Declare.Proof.t option -> unit)
-  | VtReadProof of (pstate:Declare.Proof.t -> unit)
+module InProg : sig
+  type _ t =
+    | Ignore : unit t
+    | Use : Declare.OblState.t t
 
-type vernac_command = atts:Attributes.vernac_flags -> typed_vernac
+  val cast : Declare.OblState.t -> 'a t -> 'a
+end
+
+module OutProg : sig
+  type _ t =
+    | No : unit t
+    | Yes : Declare.OblState.t t
+    | Push
+    | Pop
+
+  val cast : 'a -> 'a t -> Declare.OblState.t NeList.t -> Declare.OblState.t NeList.t
+end
+
+module InProof : sig
+  type _ t =
+    | Ignore : unit t
+    | Reject : unit t
+    | Use : Declare.Proof.t t
+    | UseOpt : Declare.Proof.t option t
+
+  val cast : Declare.Proof.t option -> 'a t -> 'a
+end
+
+module OutProof : sig
+
+  type _ t =
+    | No : unit t
+    | Close : unit t
+    | Update : Declare.Proof.t t
+    | New : Declare.Proof.t t
+
+end
+
+type ('inprog,'outprog,'inproof,'outproof) vernac_type = {
+  inprog : 'inprog InProg.t;
+  outprog : 'outprog InProg.t;
+  inproof : 'inproof InProof.t;
+  outproof : 'outproof OutProof.t;
+}
+
+type typed_vernac =
+    TypedVernac : {
+      inprog : 'inprog InProg.t;
+      outprog : 'outprog OutProg.t;
+      inproof : 'inproof InProof.t;
+      outproof : 'outproof OutProof.t;
+      run : pm:'inprog -> proof:'inproof -> 'outprog * 'outproof;
+    } -> typed_vernac
+
+(** Some convenient typed_vernac constructors *)
+
+val vtdefault : (unit -> unit) -> typed_vernac
+val vtnoproof : (unit -> unit) -> typed_vernac
+val vtcloseproof : (lemma:Declare.Proof.t -> pm:Declare.OblState.t -> Declare.OblState.t) -> typed_vernac
+val vtopenproof : (unit -> Declare.Proof.t) -> typed_vernac
+val vtmodifyproof : (pstate:Declare.Proof.t -> Declare.Proof.t) -> typed_vernac
+val vtreadproofopt : (pstate:Declare.Proof.t option -> unit) -> typed_vernac
+val vtreadproof : (pstate:Declare.Proof.t -> unit) -> typed_vernac
+val vtreadprogram : (pm:Declare.OblState.t -> unit) -> typed_vernac
+val vtmodifyprogram : (pm:Declare.OblState.t -> Declare.OblState.t) -> typed_vernac
+val vtdeclareprogram : (pm:Declare.OblState.t -> Declare.Proof.t) -> typed_vernac
+val vtopenproofprogram : (pm:Declare.OblState.t -> Declare.OblState.t * Declare.Proof.t) -> typed_vernac
+
+type vernac_command = ?loc:Loc.t -> atts:Attributes.vernac_flags -> unit -> typed_vernac
 
 type plugin_args = Genarg.raw_generic_argument list
 
@@ -100,6 +158,7 @@ type ty_ml = TyML : bool (* deprecated *) * ('r, 's) ty_sig * 'r * 's option ->
 
 (** Wrapper to dynamically extend vernacular commands. *)
 val vernac_extend :
+  ?plugin:string ->
   command:string ->
   ?classifier:(string -> vernac_classification) ->
   ?entry:Vernacexpr.vernac_expr Pcoq.Entry.t ->
@@ -120,7 +179,7 @@ type 'a vernac_argument = {
   arg_parsing : 'a argument_rule;
 }
 
-val vernac_argument_extend : name:string -> 'a vernac_argument ->
+val vernac_argument_extend : plugin:string -> name:string -> 'a vernac_argument ->
   ('a, unit, unit) Genarg.genarg_type * 'a Pcoq.Entry.t
 
 (** {5 STM classifiers} *)
